unit AddStat;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see AdvOpt.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Graphics, Forms, Controls, StdCtrls, Buttons, ExtCtrls, Dialogs, Grids,
     SysUtils, Classes;

type
  TAddOnStatusDlg = class(TForm)
    S: TStringGrid;
    Panel: TPanel;
    DoneBtn: TButton;
    BrowseBtn: TButton;
    OpenDialog: TOpenDialog;
    RefreshBtn: TButton;
    CopyBtn: TButton;
    procedure FormCreate(Sender: TObject);
    procedure DoneBtnClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure FormDestroy(Sender: TObject);
    procedure RefreshBtnClick(Sender: TObject);
    procedure BrowseBtnClick(Sender: TObject);
    procedure CopyBtnClick(Sender: TObject);
  public
    procedure RefreshStatus;
  end;

var
  AddOnStatusDlg: TAddOnStatusDlg;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses UCompat, Clipbrd, Expr;

procedure TAddOnStatusDlg.FormCreate(Sender: TObject);
begin
  S.Cells[0,0] := 'dll';
  S.Cells[1,0] := 'Description';
  S.Cells[2,0] := 'Calls';
  S.Cells[3,0] := 'Last Error';
  S.Cells[4,0] := 'Errors';
  S.Cells[5,0] := 'Extra';
end;

procedure TAddOnStatusDlg.RefreshStatus;
var
  f: TLoadedFunction;
  info, err: ShortString;
  nErr, nCalls: Longint;
  Extra: Double;
  i: Integer;
begin
  if LoadedFunctions.Count = 0 then Exit;
  S.RowCount := LoadedFunctions.Count+1;
  for i := 0 to LoadedFunctions.Count - 1 do begin
    S.Cells[0,i+1] := LoadedFunctions[i];
    f := TLoadedFunction(LoadedFunctions.Objects[i]);
    if (f <> nil) and (@f.StatusFunc <> nil) then begin
      err := '';
      info := '';
      nCalls := -1;
      nErr := -1;
      Extra := -1;
      f.StatusFunc(err, info, nCalls, nErr, Extra);
      S.Cells[1,i+1] := info;
      S.Cells[2,i+1] := IntToStr(nCalls);
      S.Cells[3,i+1] := err;
      S.Cells[4,i+1] := IntToStr(nErr);;
      S.Cells[5,i+1] := FloatToStr(Extra);
    end else begin
      S.Cells[1,i+1] := '';
      S.Cells[2,i+1] := '';
      S.Cells[3,i+1] := '';
      S.Cells[4,i+1] := '';
      S.Cells[5,i+1] := '';
    end;
  end;
end;

procedure TAddOnStatusDlg.DoneBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TAddOnStatusDlg.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseAction := caFree;
end;

procedure TAddOnStatusDlg.FormDestroy(Sender: TObject);
begin
  AddOnStatusDlg := nil;
end;

procedure TAddOnStatusDlg.RefreshBtnClick(Sender: TObject);
begin
  RefreshStatus
end;

procedure TAddOnStatusDlg.BrowseBtnClick(Sender: TObject);
var
  FuncName: string;
  i: Integer;
  o: TLoadedFunction;
begin
  OpenDialog.Filter := 
  {$IFDEF MSWINDOWS}
    'Addons (*.dll)|*.dll'
  {$ENDIF}
  {$IFDEF UNIX}
    'Addons (lib*.so)|lib*.so'
  {$ENDIF}
    + '|All Files (*.*)|'+AllFilesMask;
  if OpenDialog.Execute then begin
    FuncName := ExtractFileName(OpenDialog.FileName);
    FuncName := Copy(FuncName, 1, Length(FuncName)-Length(ExtractFileExt(FuncName)));
    {$IFDEF LINUX}
    if Pos('lib', FuncName) = 1 then
      FuncName := Copy(FuncName, 4, Length(FuncName));
    {$ENDIF}
    if LoadedFunctions.Find(FuncName, i) then
      raise Exception.CreateFmt('%s already loaded', [FuncName]);
    o := TLoadedFunction.Create(OpenDialog.FileName);
    LoadedFunctions.AddObject(FuncName, o);
  end;
  RefreshStatus;
end;

procedure TAddOnStatusDlg.CopyBtnClick(Sender: TObject);
begin
  RefreshStatus;
  if (S.Row > 0) and (S.Row <= LoadedFunctions.Count) then
    Clipboard.AsText := S.Cells[1,S.Row];
end;

end.
