unit Gendlg;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see AdvOpt.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Messages, Graphics, Forms, Controls, Buttons, StdCtrls, Grids, ActnList,
     Classes, Expr, NGrid, Vector, MatrixVec, GraphM, AdvOpt, TextStrm,
     ExActn, PropStrm, DataSource, DataFitter, Menus;

type

  TColInfo = record
    BeforeDot, AfterDot: Integer;
    HasNegative: Boolean;
  end;

  TNLFitDlg = class;

  TInteractiveDataSourceFitter = class(TDataSourceFitter)
  private
    FNLFitDlg: TNLFitDlg;
    function GetNLFitDlg: TNLFitDlg;
  protected
    procedure SetVariable(AName: string; const Value: Double); override;
  public
    property NLFitDlg: TNLFitDlg read GetNLFitDlg;
  end;

  TInteractiveDataSource = class (TFitterDataSource)
  private
    function GetNLFitDlg: TNLFitDlg;
  protected
    function MakeDataSourceFitter: TDataSourceFitter; override;
  public
    property NLFitDlg: TNLFitDlg read GetNLFitDlg;
  end;

  { TNLFitDlg }

  TNLFitDlg = class(TForm)
    CloseBtn: TButton;
    Lbl: TLabel;
    FunctionTextEdit: TEdit;
    LabelVariables: TLabel;
    CheckFunction: TBitBtn;
    PlotFunction: TBitBtn;
    FitOnce: TBitBtn;
    Advanced: TBitBtn;
    VarInfo: TNumberGrid;
    FloatAll: TButton;
    FloatNone: TButton;
    Undo: TBitBtn;
    Redo: TBitBtn;
    CorrMat: TNumberGrid;
    LabelCorrMat: TLabel;
    MoreButton: TButton;
    ExpandFunctionText: TSpeedButton;
    FunctionTextMemo: TMemo;
    GlobalFitCheckBox: TCheckBox;
    nObsText: TEdit;
    SigmaText: TEdit;
    MorePopupMenu: TPopupMenu;
    AddonsItem: TMenuItem;
    SplineItem: TMenuItem;
    procedure FloatNoneClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure VarInfoGetCellIsFlag(Sender: TObject; Col, Row: Integer; var IsFlag: Boolean);
    procedure CheckFunctionClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PlotFunctionClick(Sender: TObject);
    procedure FitOnceClick(Sender: TObject);
    procedure FunctionTextEditChange(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure AdvancedClick(Sender: TObject);
    procedure VarInfoChangeCellQuery(Sender: TObject; Col, Row: Longint; var CanChange: Boolean);
    procedure FloatAllClick(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure UndoClick(Sender: TObject);
    procedure RedoClick(Sender: TObject);
    procedure MoreButtonClick(Sender: TObject);
    procedure ExpandFunctionTextClick(Sender: TObject);
    procedure GlobalFitCheckBoxClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure AddonsItemClick(Sender: TObject);
    procedure SplineItemClick(Sender: TObject);
  private
    UndoList: TList;
    UndoPointer: Integer;
    CurrentFitSaved: Boolean;
    ColInfo: array [0..4] of TColInfo;
    CorrColInfo: TColInfo;
    FDataSource: TInteractiveDataSource;
    FStatusWindow: TControl;
    FDataSourceFitter: TInteractiveDataSourceFitter;
    ParseNeeded: Boolean;
    SplineDlg: TForm;
    procedure SetStatusWindow(const Value: TControl);
    procedure EmptyUndoList;
    procedure AddUndoEntry;
    procedure UpdateStatus;
    procedure LoadFromUndo;
    {$IFDEF MSWINDOWS}
    procedure WMGetMinMaxInfo(var Msg: TMessage); message WM_GETMINMAXINFO;
    {$ENDIF}
    procedure SetModified(const Value: Boolean);
    function GetModified: Boolean;
    procedure SetFunctionText(const Value: string);
    function GetFunctionText: string;
    procedure SetLargeFunctionWindow(const Value: Boolean);
    function GetLargeFunctionWindow: Boolean;
  public
    AdvDlg: TAdvDlg;
    constructor Create(AOwner: TComponent); override;
    procedure ShowFitInfo(G:TGraphM; var x, y: Integer; FindWidth: Boolean);
    procedure SaveOptionsToStream(s: TTextStream);
    procedure ReadFromPropStream(p: TPropReader);
    procedure LoadAdvDlg;
    procedure UnLoadAdvDlg;
    procedure ClearVarInfo;
    procedure LoadFromFitter;
    property DataSource: TInteractiveDataSource read FDataSource { write SetDataSource};
    property DataSourceFitter: TInteractiveDataSourceFitter read FDataSourceFitter { write SetDataSource};
    property Modified: Boolean read GetModified write SetModified;
    property StatusWindow: TControl read FStatusWindow write SetStatusWindow;
  published
    property FunctionText: string read GetFunctionText write SetFunctionText;
    property LargeFunctionWindow: Boolean read GetLargeFunctionWindow write SetLargeFunctionWindow;
  end;


procedure ShowCol(G:TGraphM; Col:Integer; var x, y:Integer; First, Last: Longint;
                             FindWidth, All: Boolean; var ColInfo: TColInfo; Grid: TNumberGrid;
                             Precision: Integer);

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses {$IFDEF LCL_} Types, Math, {$ENDIF}
     SysUtils, MathSup, EigLA, UCompat, AddStat, SplineOpt;

procedure ShowCol(G:TGraphM; Col:Integer; var x, y:Integer; First, Last: Longint;
                             FindWidth, All: Boolean; var ColInfo: TColInfo; Grid: TNumberGrid;
                             Precision: Integer);
var
  r: Longint;
  dy, Extra, i, xc, xt: Integer;
  v, dx: Double;
  s: string;
  Extent, DotExtent: TSize;
begin
  dy := Round(G.CellHeight);
  dx := G.AvCharWidth;
  xc := x;
  if FindWidth then begin
    ColInfo.BeforeDot := -1;
    ColInfo.AfterDot := -1;
    ColInfo.HasNegative := False;
  end else if Col > 0 then begin
    xc := x + Round(ColInfo.BeforeDot*dx);
    G.Justify := taLeftJustify;
    Extent := G.TextExtent(Grid.Title[Col]);
    DotExtent := G.TextExtent('.');
    { Try and centre over dot }
    xt := xc - Extent.cx div 2;
    if ColInfo.AfterDot > 0 then
      Inc(xt, DotExtent.cx div 2);
    if xt < x then begin
      xt := x;
      if ColInfo.HasNegative then
        Inc(xt, Round(dx));
    end;
    if xt + Extent.cx > x + (ColInfo.BeforeDot+1+ColInfo.AfterDot)*dx then
      xt := x + Round((ColInfo.BeforeDot+1+ColInfo.AfterDot)*dx) - Extent.cx;
    G.TextI(xt, y, Grid.Title[Col]);
  end else
    Inc(xc, ColInfo.BeforeDot);
  Dec(y, dy div 4);
  for r := First to Last do
    if All or (Grid.Valid[r,1] and Grid.Valid[r,2]) then begin
      y := y - dy;
      if y < 0 then Break;
      if Col = 0 then begin
        i := G.TextExtent(Grid.Lbl[r]).cx;
        if FindWidth then begin
          if ColInfo.BeforeDot < i then ColInfo.BeforeDot := i;
        end else begin
          G.Justify := taRightJustify;
          G.TextI(xc, y, Grid.Lbl[r]);
        end;
      end else if Grid.GetCell(r, Col, v) then begin
        if Precision < 0 then
          s := FloatToStrF(v, ffFixed, -Precision+3, -Precision)
        else
          s := FloatToStrF(v, ffGeneral, Precision, 0);
        i := Pos('.', s);
        if i = 0 then i := Length(s)+1;
        if FindWidth then begin
          if v < 0 then ColInfo.HasNegative := True;
          if ColInfo.BeforeDot < i - 1 then ColInfo.BeforeDot := i - 1;
          if ColInfo.AfterDot < Length(s) - i then ColInfo.AfterDot := Length(s) - i;
        end else begin
          G.Justify := taRightJustify;
          G.TextI(xc, y, Copy(s, 1, i-1));
          if i <= Length(s) then begin
            G.Justify := taLeftJustify;
            G.TextI(xc, y, Copy(s, i, Length(s)));
          end;
        end;
      end;
    end;
  if FindWidth and (Col > 0) then begin
    Extra := Length(Grid.Title[Col]) - (ColInfo.BeforeDot+1+ColInfo.AfterDot);
    if Extra > 0 then begin
{      Inc(AfterDot[Col], Extra div 2);
      Inc(BeforeDot[Col], Length(Data.Title[Col]) - (BeforeDot[Col]+1+AfterDot[Col])); }
      Inc(ColInfo.BeforeDot, Extra);
    end;
  end;
  if Col > 0 then
    Inc(x, Round((ColInfo.BeforeDot+1+ColInfo.AfterDot)*dx))
  else
    Inc(x, ColInfo.BeforeDot);
  Dec(y, dy div 4);
end;

const MaxUndos = 50;

procedure TNLFitDlg.CheckFunctionClick(Sender: TObject);
var
  i, j, k, FirstBlank: Integer;
begin
  with DataSourceFitter do begin
    Expression := FunctionText;
    { Delete any variables not now used, and compact array }
    if VarInfo.RowCount - 1 < Variables.Count then EmptyUndoList;
    for i := 1 to VarInfo.RowCount - 1 do
      if Variables.IndexOf(VarInfo.Lbl[i]) < 0 then
        VarInfo.Lbl[i] := '';
    { Compact unused rows }
    FirstBlank := -1;
    for i := 1 to VarInfo.RowCount - 2 do
      if Length(VarInfo.Lbl[i]) = 0 then begin
        j := i+1;
        while Length(VarInfo.Lbl[j]) = 0 do begin
          Inc(j);
          if j >= VarInfo.RowCount then begin
            FirstBlank := i;
            Break;
          end;
        end;
        if FirstBlank < 0 then begin
          for k := 1 to VarInfo.ColCount - 1 do
            if VarInfo.Valid[j,k] then
              VarInfo[i,k] := VarInfo[j,k]
            else
              VarInfo.Valid[i,k] := False;
          VarInfo.Lbl[i] := VarInfo.Lbl[j];
          VarInfo.Lbl[j] := '';
        end else
          Break;
      end;
    { Add new variable names to list }
    if Variables.Count = 0 then
      ClearVarInfo
    else
      VarInfo.RowCount := Variables.Count + 1;
    { Make sure both variable lists in same order }
    for i := 1 to Variables.Count do
      if Length(VarInfo.Lbl[i]) = 0 then begin
        { New Variable }
        VarInfo.Lbl[i] := Variables[i-1];
        VarInfo[i,1] := Variables.Objects[i-1].Value;
        VarInfo.Valid[i,2] := False;
        VarInfo.Valid[i,3] := False;
        VarInfo.Valid[i,4] := True;
      end else begin
        j := Variables.IndexOf(VarInfo.Lbl[i]);
        if j <> i-1 then
          Variables.Exchange(i-1, j);
      end;
  end;
  ParseNeeded := False;
end;

procedure TNLFitDlg.FormDestroy(Sender: TObject);
begin
  EmptyUndoList;
  AdvDlg.Free;
  inherited;
end;

procedure TNLFitDlg.PlotFunctionClick(Sender: TObject);
var
  i: Integer;
begin
  with DataSourceFitter do begin
    CheckFunctionClick(Sender);
    { Load Variables }
    with Variables do
      for i := 0 to Count - 1 do
        Objects[i].Value := VarInfo[i+1, 1];
    SetFitExprs(Expr, 0);
    if AddOnStatusDlg <> nil then
      AddOnStatusDlg.RefreshStatus
  end;
  UpdateStatus;
end;

procedure TNLFitDlg.FitOnceClick(Sender: TObject);
var
  r, i, j: Longint;
begin
  with VarInfo, DataSourceFitter do begin
    if (Expr = nil) or ParseNeeded then
      CheckFunctionClick(Sender);
    UnLoadAdvDlg;
    for r := 1 to Variables.Count do begin
      Floating[r-1] := VarInfo.Valid[r,4];
      LastError[r-1] := VarInfo.Value[r,2];
      if not VarInfo.Valid[r,1] then
        Variables.Objects[r-1].Value := 0
      else
        Variables.Objects[r-1].Value := VarInfo[r,1];
      VarInfo.Valid[r,2] := False;
      VarInfo.Valid[r,3] := False;
    end;
    FitOnce;
  end;
  if AdvDlg <> nil then
    if not Assigned(DataSourceFitter.SVD) then
      AdvDlg.SVDMat.Blank
    else
      with AdvDlg.SVDMat, DataSourceFitter do begin
        RowCount := nPar+1;
        ColCount := nPar+2;
        i := 1;
        for r := 1 to Variables.Count do
          if Floating[r-1] then begin
            Title[i+1] := Variables.Objects[r-1].Name;
            Value[i, 1] := SVDw[i];
            for j := 1 to nPar do
              Value[j, i+1] := SVD[j, i];
            Inc(i);
          end;
      end;
  AddUndoEntry;
  CurrentFitSaved := False;
  with VarInfo, DataSourceFitter do begin
    CorrMat.ColCount := nPar + 1;
    CorrMat.RowCount := nPar + 1;
    if nPar > 0 then begin
      CorrMat.FixedCols := 1;
      CorrMat.FixedRows := 1;
    end else
      CorrMat.Clear;
    nObsText.Text := IntToStr(nObs);
    SigmaText.Text := DtosM(NewSD, 7, -1);
    i := 1;
    for r := 1 to Variables.Count do
      if Floating[r-1] then begin
        Value[r, 3] := Variables.Objects[r-1].Value - Value[r, 1];
        Value[r, 1] := Variables.Objects[r-1].Value;
        if b.Valid[i] then begin
          Value[r, 2] := b[i];
          for j := 1 to i do
            if b.Valid[j] then
              CorrMat[i, j] := ada[i, j];
        end else begin
          Valid[i, 2] := False;
          for j := 1 to i do
            CorrMat.Valid[i, j] := False;
        end;
        with Variables.Objects[r-1] do begin
          CorrMat.Lbl[i] := Name;
          CorrMat.Title[i] := Name;
        end;
        for j := i+1 to nPar do
          CorrMat.Valid[j, i] := False;
        Inc(i);
      end;
    SetFitExprs(Expr, nPar);
  end;
end;

procedure TNLFitDlg.FunctionTextEditChange(Sender: TObject);
begin
  CorrMat.Clear;
  ParseNeeded := True;
end;

{$IFDEF MSWINDOWS}
procedure TNLFitDlg.WMGetMinMaxInfo(var Msg: TMessage);
var
  w, wmin: Integer;
begin
  inherited;
  w := VarInfo.NaturalWidth + 2*VarInfo.Left;
  wmin := CloseBtn.Left + CloseBtn.Width + CheckFunction.Left;
  if w < wmin then w := wmin;
  Inc(w, 2*GetSystemMetrics(SM_CXFRAME));
  with PMINMAXINFO(Msg.lParam)^ do begin
    {$IFNDEF VER80}
      ptMinTrackSize.x := w;
      // ptMaxTrackSize.x := w;
    {$ELSE}
      ptMinTrackSize.x := w;
      // ptMaxTrackSize.x := w;
    {$ENDIF}
  end;
  Msg.Result := 0;
end;
{$ENDIF}

procedure TNLFitDlg.FormResize(Sender: TObject);
var
  i, BtnTop, BtnGap, TotalHeight, Margin, Change: Integer;
begin
  FunctionTextMemo.Left := FunctionTextEdit.Left;
  FunctionTextMemo.Top := FunctionTextEdit.Top;
  VarInfo.Width := VarInfo.NaturalWidth;
  Margin := VarInfo.Left;
  {$IFDEF LCL_}
    Constraints.MinWidth := VarInfo.NaturalWidth + 2*Margin;
    i := CloseBtn.Left + CloseBtn.Width + CheckFunction.Left;
    if i > Constraints.MinWidth then Constraints.MinWidth := i;
    i := GlobalFitCheckBox.Left - nObsText.Left - nObsText.Width - Margin;
    i := FloatNone.Left + FloatNone.Width + Margin - i;
    if i > Constraints.MinWidth then Constraints.MinWidth := i;
    if FunctionTextEdit.Visible then
      FloatNone.AnchorSideTop.Control := FunctionTextEdit
    else begin
      FloatNone.AnchorSideTop.Control := FunctionTextMemo;
      FunctionTextMemo.Width := ClientWidth - 2* Margin;
    end;
    FloatNone.Anchors := FloatNone.Anchors + [akTop];
    i := VarInfo.Height - CorrMat.Height;
    if Abs(i) > 20 then
      VarInfo.Height := VarInfo.Height - i div 2;
  {$ELSE}
    Change := CorrMat.Top - (VarInfo.Top + VarInfo.Height); { Current Gap }
    if Change <= 0 then Exit; { Re-entry of routine ? }
    if FunctionTextEdit.Visible then
      Inc(Change, FunctionTextEdit.Top + FunctionTextEdit.Height)  { New Position }
    else
      Inc(Change, FunctionTextMemo.Top + FunctionTextMemo.Height);
    Change := Change - VarInfo.Top; { final change }
    if Change <> 0 then begin
      LabelVariables.Top := LabelVariables.Top + Change;
      FloatNone.Top := FloatNone.Top + Change;
      FloatAll.Top := FloatAll.Top + Change;
      MoreButton.Top := MoreButton.Top + Change;
      ExpandFunctionText.Top := ExpandFunctionText.Top + Change;
      GlobalFitCheckBox.Top := GlobalFitCheckBox.Top + Change;
      nObsText.Top := nObsText.Top + Change;
      SigmaText.Top := SigmaText.Top + Change;
      VarInfo.Top := VarInfo.Top + Change;
    end;
    if NewStyleControls then
    {$IFNDEF VER80}
      i := 2 { Checked 96/32 }
    {$ELSE}
      i := 0 { Checked 96/16 }
    {$ENDIF}
    else
      i := 2;
    FloatNone.Left := VarInfo.Left + VarInfo.Width - FloatNone.Width - i + 2;
    FloatAll.Left := FloatNone.Left - FloatAll.Width;
    MoreButton.Left := FloatAll.Left - MoreButton.Width;
    ExpandFunctionText.Left := MoreButton.Left - ExpandFunctionText.Width;
    GlobalFitCheckBox.Left := ExpandFunctionText.Left - GlobalFitCheckBox.Width;
    SigmaText.Left := GlobalFitCheckBox.Left - SigmaText.Width - 2;
    nObsText.Left := SigmaText.Left - nObsText.Width - 2;
    FunctionTextEdit.Width := ClientWidth - 2* Margin;
    FunctionTextMemo.Width := ClientWidth - 2* Margin;
    BtnGap := CloseBtn.Top - (CorrMat.Top+CorrMat.Height);
    BtnTop := ClientHeight - CloseBtn.Height - BtnGap;
    TotalHeight := BtnTop - VarInfo.Top - 2*Margin - (VarInfo.Top-LabelVariables.Top);
    VarInfo.Height := TotalHeight div 2;
    LabelCorrMat.Top := VarInfo.Top + VarInfo.Height + VarInfo.Left;
    CorrMat.Top := LabelCorrMat.Top + (VarInfo.Top-LabelVariables.Top);
    CorrMat.Height := VarInfo.Height;
    CorrMat.Width := ClientWidth - 2* CorrMat.Left;
    BtnTop := CorrMat.Top+CorrMat.Height + BtnGap;
    CloseBtn.Top := BtnTop;
    for i := 0 to ControlCount-1 do
      if Controls[i] is TBitBtn then
        Controls[i].Top := BtnTop;
  {$ENDIF}
end;

procedure TNLFitDlg.AdvancedClick(Sender: TObject);
begin
  if AdvDlg = nil then
    AdvDlg := TAdvDlg.Create(Self);
  if not AdvDlg.Visible then
    LoadAdvDlg;
  AdvDlg.Show;
end;

procedure TNLFitDlg.VarInfoChangeCellQuery(Sender: TObject; Col,
  Row: Longint; var CanChange: Boolean);
begin
  CanChange := Col in [1,4];
end;

procedure TNLFitDlg.VarInfoGetCellIsFlag(Sender: TObject; Col, Row: Integer; var IsFlag: Boolean);
begin
  IsFlag := Col = 4;
end;

procedure TNLFitDlg.FloatAllClick(Sender: TObject);
var
  r: Longint;
begin
  for r := 1 to VarInfo.RowCount-1 do
    VarInfo.Valid[r, 4] := True;
end;

procedure TNLFitDlg.FloatNoneClick(Sender: TObject);
var
  r: Longint;
begin
  for r := 1 to VarInfo.RowCount-1 do
    VarInfo.Valid[r, 4] := False;
end;

procedure TNLFitDlg.FormShow(Sender: TObject);
begin
  UpdateStatus;
end;

procedure TNLFitDlg.ClearVarInfo;
begin
  VarInfo.RowCount := 2;
  VarInfo.Lbl[1] := '';
  VarInfo[1,1] := 0;
  VarInfo.Valid[1,2] := False;
  VarInfo.Valid[1,3] := False;
  VarInfo.Valid[1,4] := True;
end;

procedure TNLFitDlg.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

constructor TNLFitDlg.Create(AOwner: TComponent);
begin
  if AOwner is TInteractiveDataSourceFitter then begin
   FDataSourceFitter := AOwner as TInteractiveDataSourceFitter;
   FDataSource := FDataSourceFitter.Owner as TInteractiveDataSource;
  end;
  inherited;
  SetModified(False);
end;

procedure TNLFitDlg.EmptyUndoList;
var
  i: Integer;
begin
  if UndoList <> nil then
    for i := 0 to UndoList.Count - 1 do
      TColumnVector(UndoList[i]).Free;
  UndoList.Free;
  UndoList := nil;
  UpdateStatus;
end;

procedure TNLFitDlg.AddonsItemClick(Sender: TObject);
begin
  if AddOnStatusDlg = nil then
    AddOnStatusDlg := TAddOnStatusDlg.Create(Self);
  AddOnStatusDlg.Show;
end;

procedure TNLFitDlg.AddUndoEntry;
var
  i: Longint;
  UndoEntry: TColumnVector;
begin
  if UndoList = nil then begin
    UndoList := TList.Create;
    UndoPointer := 0;
  end else
    Inc(UndoPointer);
  with UndoList do begin
    while Count > UndoPointer+1 do begin
      TColumnVector(Last).Free;
      Delete(Count-1);
    end;
    if Count <= UndoPointer then
      if Count >= MaxUndos then
        { Rotate, discarding first }
        for i := 0 to Count-2 do
          Exchange(i, i+1)
      else
        Add(TColumnVector.Create(DataSourceFitter.Variables.Count));
    UndoPointer := Count - 1;
    UndoEntry := TColumnVector(Last);
  end;
  for i := 1 to DataSourceFitter.Variables.Count do
    UndoEntry[i] := VarInfo[i, 1];
  CurrentFitSaved := True;
  UpdateStatus;
end;

procedure TNLFitDlg.UpdateStatus;
begin
  if csDestroying in Application.ComponentState then
    Exit;
  Undo.Enabled := (UndoList <> nil) and (UndoPointer < UndoList.Count) and (UndoPointer >= 0);
  Redo.Enabled := (UndoList <> nil) and (UndoPointer < UndoList.Count-2) and (UndoPointer >= -1);
  if AddOnStatusDlg <> nil then
    AddOnStatusDlg.RefreshStatus;
  if DataSourceFitter.GlobalFit or not Assigned(DataSource) then
    Caption := 'General Least Squares Fit'
  else
    Caption := 'General Least Squares Fit - ' + DataSource.Caption;
end;

procedure TNLFitDlg.LoadFromFitter;
var
  i, n: Integer;
begin
  FunctionText := DataSourceFitter.Expression;
  CheckFunctionClick(nil);
  with DataSourceFitter do begin
    for i := 1 to Variables.Count do begin
      n := Variables.IndexOf(VarInfo.Lbl[i])+1;
      VarInfo[n,1] := Variables.Objects[i-1].Value;
      VarInfo[n,2] := LastError[i-1];
      VarInfo.Valid[n,3] := False;
      VarInfo.Valid[n,4] := Floating[i-1];
    end;
  end;
  LoadAdvDlg;
end;

procedure TNLFitDlg.LoadFromUndo;
var
  i: Longint;
begin
  if (UndoList <> nil) and (UndoPointer < UndoList.Count) and (UndoPointer >= 0) then begin
    with TColumnVector(UndoList[UndoPointer]) do
      for i := 1 to Size do begin
        VarInfo[i, 1] := Value[i];
        VarInfo.Valid[i,2] := False;
        VarInfo.Valid[i,3] := False;
      end;
    PlotFunctionClick(Self);
  end;
end;

procedure TNLFitDlg.UndoClick(Sender: TObject);
begin
  if not CurrentFitSaved then begin
    AddUndoEntry;
    Dec(UndoPointer);
  end;
  LoadFromUndo;
  Dec(UndoPointer);
  UpdateStatus;
end;

procedure TNLFitDlg.RedoClick(Sender: TObject);
begin
  Inc(UndoPointer, 2);
  UndoClick(Self);
end;

procedure TNLFitDlg.ShowFitInfo(G:TGraphM; var x, y: Integer; FindWidth: Boolean);
var
  dx, dy, iy, Col, InfoX, CorrX, TextX, t: Integer;
  procedure Txt(s: string);
  begin
    if FindWidth then begin
      if Length(s)*G.AvCharWidth > TextX then TextX := Round(Length(s)*G.AvCharWidth);
    end else
      G.TextI(0, y, s);
    Dec(y, dy);
  end;
  procedure Vline(x: Integer);
  begin
    if not FindWidth then begin
      G.Move(x, iy+dy - dy div 4);
      G.Vector(x, y);
    end;
  end;
  procedure Hline(xe: Integer);
  begin
    if not FindWidth then begin
      G.Move(x, iy - dy div 4);
      G.Vector(xe, iy - dy div 4);
    end;
  end;
begin
  dy := Round(G.CellHeight);
  dx := Round(G.AvCharWidth);
  TextX := 0;
  if LargeFunctionWindow then begin
    Txt('Fit to:');
    with FunctionTextMemo.Lines do
      for t := 0 to Count-1 do
        Txt(Strings[t]);
  end else
    Txt(Concat('Fit to: ', FunctionText));
  Txt('Variables');
  iy := y;
  InfoX := x;
  for Col := 0 to VarInfo.ColCount - 2 do begin
    y := iy;
    ShowCol(G, Col, InfoX, y, 1, VarInfo.RowCount - 1, FindWidth, True, ColInfo[Col], VarInfo, 7);
    if Col = 0 then Vline(InfoX + dx div 2);
    Inc(InfoX, dx);
  end;
  HLine(InfoX-dx);
  Dec(y, dy);
  Txt('Correlation Matrix');
  CorrX := x;
  iy := y;
  t := CorrMat.RowCount - 1;
  // if FindWidth then t := 0;
  for Col := 0 to t do begin
    y := iy;
    ShowCol(G, Col, CorrX, y, 1, CorrMat.RowCount - 1, FindWidth, True, CorrColInfo, CorrMat, -4);
    if Col = 0 then Vline(CorrX + dx div 2);
    if not FindWidth then
      with CorrColInfo do begin
        BeforeDot := 2;
        AfterDot := 4;
        HasNegative := True;
      end;
    Inc(CorrX, dx);
  end;
  HLine(CorrX-dx);
  Dec(y, dy);
  if CorrX > InfoX then
    x := CorrX
  else
    x := InfoX;
  if TextX > x then x := TextX;
end;

procedure TNLFitDlg.SplineItemClick(Sender: TObject);
begin
  if SplineDlg = nil then
    SplineDlg := TSplineDlg.Create(Self);
  SplineDlg.Show;
end;

procedure TNLFitDlg.SaveOptionsToStream(s: TTextStream);
var
  i, j: Integer;
  Prefix: string;
  procedure Wr(const Name, Value: string);
  begin
    s.Format('%s=%s', [Name, Value]);
    s.NewLine;
  end;
begin
  if LargeFunctionWindow then
    SavePropertiesToStream(self, s, '', 'LargeFunctionWindow,FunctionTextMemo.Lines')
  else
    SavePropertiesToStream(self, s, '', 'LargeFunctionWindow,FunctionText');
  Wr('Variables.Count', IntToStr(VarInfo.RowCount - 1));
  for i := 1 to VarInfo.RowCount - 1 do
    if VarInfo.Lbl[i] <> '' then begin
      FmtStr(Prefix, 'Variables.%d.', [i]);
      Wr(Concat(Prefix,'Name'), VarInfo.Lbl[i]);
      if VarInfo.Valid[i,1] then
        Wr(Concat(Prefix,'Value'), FloatToStr(VarInfo[i,1]));
      if VarInfo.Valid[i,2] then
        Wr(Concat(Prefix,'StdDev'), FloatToStr(VarInfo[i,2]));
      if VarInfo.Valid[i,3] then
        Wr(Concat(Prefix,'LastChange'), FloatToStr(VarInfo[i,3]));
      if VarInfo.Valid[i,4] then
        Wr(Concat(Prefix,'Float'), 'yes')
      else
        Wr(Concat(Prefix,'Float'), 'no')
    end;
  if (CorrMat.FixedRows > 0) and (CorrMat.Valid[1,1] or (CorrMat.RowCount > 2)) then begin
    Wr('Correlation.Size', IntToStr(CorrMat.RowCount-1));
    for i := 1 to CorrMat.RowCount - 1 do begin
      FmtStr(Prefix, 'Correlation.Row%d.', [i]);
      Wr(Concat(Prefix,'Name'), CorrMat.Lbl[i]);
      for j := 1 to i do
        if CorrMat.Valid[i,j] then
          Wr(Concat(Prefix,Format('Col%d', [j])), FloatToStr(CorrMat[i,j]));
    end;
    if AdvDlg <> nil then
      with AdvDlg.SVDMat do begin
        Wr('SVD.Rows', IntToStr(RowCount-1));
        for i := 1 to RowCount-1 do begin
          FmtStr(Prefix, 'SVD.Row%d.', [i]);
          Wr(Concat(Prefix,'Name'), Title[i+1]);
          Wr(Concat(Prefix,'Value'), FloatToStr(Value[i,1]));
          for j := 1 to RowCount-1 do
            Wr(Concat(Prefix,Format('Col%d', [j])), FloatToStr(Value[i,j+1]));
        end;
      end;
  end;
  UnLoadAdvDlg;
  SavePropertiesToStream(DataSourceFitter, s, '',
    'GlobalFit,Increment,ScaleChanges,SVDThresh,UseStdDev,LinearFit,UseSVD');
end;

procedure TNLFitDlg.ReadFromPropStream(p: TPropReader);
var
  s: string;
  Count: Integer;
  i, j: Integer;
  Prefix: string;
  f: Extended;
  b: Boolean;
begin
  if DataSourceFitter.GlobalFit and (Length(DataSources) > 1) then Exit;
  p.ReadProperties(DataSourceFitter, 'Fit',
    'GlobalFit,Increment,ScaleChanges,SVDThresh,UseStdDev,LinearFit,UseSVD');
  LoadAdvDlg;
  { FunctionText overrides FunctionTextMemo; Function is for older files }
  if p.ReadString('Fit', 'FunctionText', s)
     or p.ReadString('Fit', 'Function', s) then begin
    FunctionText := s;
    LargeFunctionWindow := False;
  end else
    p.ReadProperties(self, 'Fit', 'LargeFunctionWindow,FunctionTextMemo.Lines');
  if p.ReadInteger('Fit', 'Variables.Count', Count) then begin
    ClearVarInfo;
    VarInfo.RowCount := Count + 1;
    for i := 1 to VarInfo.RowCount - 1 do begin
      FmtStr(Prefix, 'Variables.%d.', [i]);
      if p.ReadString('Fit', Concat(Prefix,'Name'), s) then begin
        VarInfo.Lbl[i] := s;
        if p.ReadFloat('Fit', Concat(Prefix,'Value'), f) then
          VarInfo[i,1] := f;
        if p.ReadFloat('Fit', Concat(Prefix,'StdDev'), f) then
          VarInfo[i,2] := f;
        if p.ReadFloat('Fit', Concat(Prefix,'LastChange'), f) then
          VarInfo[i,3] := f;
        if p.ReadBool('Fit', Concat(Prefix,'Float'), b) then
          VarInfo.Valid[i,4] := b;
      end;
    end;
    CheckFunctionClick(Self);
    with DataSourceFitter.Variables do
      for i := 0 to Count - 1 do
        Objects[i].Value := VarInfo[i+1, 1];
    if p.ReadInteger('Fit', 'Correlation.Size', Count) then begin
      CorrMat.RowCount := Count + 1;
      CorrMat.ColCount := Count + 1;
      for i := 1 to Count do begin
        FmtStr(Prefix, 'Correlation.Row%d.', [i]);
        if p.ReadString('Fit', Concat(Prefix,'Name'), s) then begin
          CorrMat.Lbl[i] := s;
          CorrMat.Title[i] := s;
        end;
        for j := 1 to i do
          if p.ReadFloat('Fit', Concat(Prefix,Format('Col%d', [j])), f) then
            CorrMat[i,j] := f;
      end;
    end;
    if p.ReadInteger('Fit', 'SVD.Rows', Count) then begin
      if AdvDlg = nil then
        AdvDlg := TAdvDlg.Create(Self);
      if not AdvDlg.Visible then
        LoadAdvDlg;
      with AdvDlg.SVDMat do begin
        RowCount := Count + 1;
        ColCount := Count + 2;
        for i := 1 to RowCount-1 do begin
          FmtStr(Prefix, 'SVD.Row%d.', [i]);
          if p.ReadString('Fit', Concat(Prefix,'Name'), s) then
            Title[i+1] := s;
          if p.ReadFloat('Fit', Concat(Prefix,'Value'), f) then
            Value[i,1] := f;
          for j := 1 to Count do
            if p.ReadFloat('Fit', Concat(Prefix,Format('Col%d', [j])), f) then
              Value[i,j+1] := f;
        end;
      end;
    end;
  end;
end;

procedure TNLFitDlg.LoadAdvDlg;
begin
  GlobalFitCheckBox.Checked := DataSourceFitter.GlobalFit;
  if AdvDlg <> nil then
    with DataSourceFitter do begin
      AdvDlg.Increment.Value := Increment;
      AdvDlg.ScaleChanges.Value := ScaleChanges;
      AdvDlg.SVDThresh.Value := SVDThresh;
      AdvDlg.UseStdDev.Checked := UseStdDev;
      AdvDlg.LinearFit.Checked := LinearFit;
      AdvDlg.UseSVD.Checked := UseSVD;
    end;
end;

procedure TNLFitDlg.UnLoadAdvDlg;
begin
  if AdvDlg <> nil then
    with DataSourceFitter do begin
      Increment := AdvDlg.Increment.Value;
      ScaleChanges := AdvDlg.ScaleChanges.Value;
      SVDThresh := AdvDlg.SVDThresh.Value;
      UseStdDev := AdvDlg.UseStdDev.Checked;
      LinearFit := AdvDlg.LinearFit.Checked;
      UseSVD := AdvDlg.UseSVD.Checked;
    end;
end;

procedure TNLFitDlg.MoreButtonClick(Sender: TObject);
begin
  MorePopupMenu.PopupComponent := MoreButton;
  MorePopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TNLFitDlg.SetModified(const Value: Boolean);
begin
  FunctionTextEdit.Modified := Value;
  FunctionTextMemo.Modified := Value;
  VarInfo.Modified := Value;
end;

procedure TNLFitDlg.SetStatusWindow(const Value: TControl);
begin
  FStatusWindow := Value;
  VarInfo.StatusWindow := FStatusWindow;
end;

function TNLFitDlg.GetModified: Boolean;
begin
  if LargeFunctionWindow then
    Result := FunctionTextMemo.Modified
  else
    Result := FunctionTextEdit.Modified;
  Result := Result or VarInfo.Modified;
end;

procedure TNLFitDlg.ExpandFunctionTextClick(Sender: TObject);
begin
  LargeFunctionWindow := not LargeFunctionWindow;
end;

procedure TNLFitDlg.SetFunctionText(const Value: string);
begin
  if Pos(#13,Value) > 0 then
    LargeFunctionWindow := True;
  if FunctionTextEdit.Visible then
    FunctionTextEdit.Text := Value
  else
    FunctionTextMemo.Lines.Text := Value;
  {$IFDEF LCL_}
  FunctionTextEditChange(nil); { Needs to happen even if dialog has not been used }
  {$ENDIF}
end;

function TNLFitDlg.GetFunctionText: string;
begin
  if FunctionTextEdit.Visible then
    Result := FunctionTextEdit.Text
  else
    Result := FunctionTextMemo.Lines.Text;
end;

procedure TNLFitDlg.SetLargeFunctionWindow(const Value: Boolean);
begin
  if Value <> LargeFunctionWindow then begin
    if Value then begin
      FunctionTextMemo.Lines.Text := FunctionTextEdit.Text;
      FunctionTextMemo.Modified := FunctionTextEdit.Modified;
    end else begin
      FunctionTextEdit.Text := FunctionText;
      FunctionTextEdit.Modified := FunctionTextMemo.Modified;
    end;
    {$IFDEF LCL_}
    FloatNone.Anchors := FloatNone.Anchors - [akTop];
    {$ENDIF}
    FunctionTextMemo.Visible := Value;
    FunctionTextEdit.Visible := not Value;
    FormResize(Self);
  end;
end;

function TNLFitDlg.GetLargeFunctionWindow: Boolean;
begin
  Result := Assigned(FunctionTextMemo) and FunctionTextMemo.Visible
end;

{$R-}
function _Eigval(var a: array of Double): Double;
var
  i, j, p, n, na: Integer;
  Answer: TColumnVector;
  Matrix: TMatrix;
  d: DOuble;
begin
  { Figure out matrix size }
  p := High(a);
  n := Round((Sqrt(1+8*p)-1)/2);
  na := Round(a[p]);
  if (na < 1) or (na > n) or ((n*(n+1))div 2 <> p) then
    raise Exception.Create('Bad matrix size');
  Matrix := TMatrix.Create(n,n);
  try
    Answer := TColumnVector.Create(n);
    try
      for i := 1 to n do
        for j := 1 to i do begin
          Dec(p);
          d := a[p];
          Matrix[i,j] := d;
          Matrix[j,i] := d;
        end;
      Eigen(Answer, Matrix, False);
      Result := Answer[na];
    finally
      Answer.Free;
    end;
  finally
    Matrix.Free;
  end;
end;

procedure TNLFitDlg.GlobalFitCheckBoxClick(Sender: TObject);
begin
  DataSourceFitter.GlobalFit := GlobalFitCheckBox.Checked;
end;

procedure TNLFitDlg.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if DataSourceFitter.GlobalFit and (GlobalFitter <> DataSourceFitter) then
    CloseAction := caFree;
end;

{ TInteractiveDataSource }

function TInteractiveDataSource.GetNLFitDlg: TNLFitDlg;
begin
  Result := (DataSourceFitter as TInteractiveDataSourceFitter).NLFitDlg;
end;

function TInteractiveDataSource.MakeDataSourceFitter: TDataSourceFitter;
begin
  Result := TInteractiveDataSourceFitter.Create(Self);
end;

{ TInteractiveDataSourceFitter }

function TInteractiveDataSourceFitter.GetNLFitDlg: TNLFitDlg;
begin
  if not Assigned(FNLFitDlg) then begin
    FNLFitDlg := TNLFitDlg.Create(Self);
    FNLFitDlg.UpdateStatus;
  end;
  Result := FNLFitDlg;
end;

procedure TInteractiveDataSourceFitter.SetVariable(AName: string;
  const Value: Double);
var
  Index: Integer;
begin
  inherited;
  with NLFitDlg do begin
    Index := VarInfo.Lbls.IndexOf(AName);
    if Index >= 0 then
      VarInfo[Index+1,1] := Value;
  end;
end;

initialization

GlobalFunctions.AddObject('eigval', TVaradicFunction.Create({$IFDEF FPC}@{$ENDIF}_Eigval, 3, 'eigval'));
end.
