unit DrvLoader;
{ Copyright (C) 2005 C M Western; see NLFit.pas for license}
interface

{ Loader for old drive files }

uses ExpPlotLoader, Classes;

const
  mxnchan = 13;

type
  chan_type = (NONE = 0, Ix = 1,I2 = 2,I4 = 3, R4 = 4, R8 = 5);

  TFILEHEADER = packed record
    signature : array[0..3] of AnsiChar;
    nc : word;
    np : smallint;
    channel : array[0..(mxnchan)-1] of smallint;
    npoints : array[0..(mxnchan)-1] of longint;
    ctype : array[0..(mxnchan)-1] of chan_type;
  end;

  TDriveFileLoader = class(TExperimentalPlotLoader)
  private
    Header: TFILEHEADER;
    nPoints: Longint;
    procedure ReadHeader;
    procedure ReadBody;
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses SysUtils;

{ TDriveFileLoader }

const
  chan_size: array [chan_type] of Integer = (0, 1, 2, 4, 4, 8);
  chan_ArrayType: array [chan_type] of TArrayType = (tyI8, tyI8, tyI16, tyI32, tySingle, tyDouble);

function TDriveFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  Start: string[3];
  i: Integer;
  BodyPos: Int64;
begin
  Result := False;
  if Ext = '' then begin
    Stream.Read(Start[1], 3);
    Stream.Position := 0;
    if Start = '!dr' then
      Ext := '.drv'
  end;
  if Ext <> '.drv' then
    Exit;
  ReadHeader;
  BodyPos := Stream.Position;
  for i := 0 to Header.nc - 1 do
    if Header.channel[i] >= 8 then begin
      { Stop reading when when least squares channels reached }
      Header.nc := i;
      SetLength(ChannelDest, i);
      Break;
    end else
      AddIncrementCountFor(i, Header.npoints[i], chan_ArrayType[Header.ctype[i]]);
  AutoDetectFrequency;
  CheckAvailableChannels(Header.nc, nPoints, AskAboutWeights);
  Stream.Position := BodyPos;
  ReadBody;
  Result := True;
end;

procedure TDriveFileLoader.ReadBody;
var
  i: Integer;
begin
  for i := 0 to Header.nc - 1 do
    if ChannelDest[i] < 0 then
      { Skip this channel }
      Stream.Position := Stream.Position + Header.npoints[i]*chan_size[Header.ctype[i]]
    else
      ReadAndCopyArray(-1, ChannelDest[i], chan_ArrayType[Header.ctype[i]], Header.npoints[i]);
end;

procedure TDriveFileLoader.ReadHeader;
var
  i, t: Integer;
  np: array [0..mxnchan-1] of SmallInt;
begin
  Read(Header, 8);
  if Header.nc > mxnchan then
    raise Exception.Create('Bad number of channels');
  Read(Header.channel, Sizeof(SmallInt)*Header.nc);
  nPoints := Header.np;
  if Header.signature = '!drb' then
    for i := 0 to Header.nc - 1 do begin
      Header.npoints[i] := Header.np;
      Inc(Header.channel[i], (i+1)*100);
    end
  else if Header.signature = '!drc' then begin
    Read(np, Sizeof(SmallInt)*Header.nc);
    for i := 0 to Header.nc - 1 do
      Header.npoints[i] := np[i];
  end else if Header.signature = '!drd' then begin
    Read(Header.npoints, Sizeof(longint)*Header.nc);
    { Figure out maximum number of points in non-least squares channels }
    nPoints := 0;
    for i := 0 to Header.nc - 1 do
      if (Header.channel[i] < 900) and (Header.npoints[i] > nPoints) then
        nPoints := Header.npoints[i];
  end else
    raise Exception.Create('Bad Signature:' + Header.signature);
  SetLength(ChannelDest, Header.nc);
  for i := 0 to Header.nc - 1 do begin
    t := Header.channel[i] mod 100;
    if t = 8 then t := Ord(R8);
    Header.ctype[i] := chan_type(t);
    if not (Header.ctype[i] in [I2..R8]) then
      raise Exception.CreateFmt('Bad channel type %d', [t]);
    Header.channel[i] := Header.channel[i] div 100 - 1;
    ChannelDest[i] := i;
  end;
end;

initialization
  RegisterPlotLoader(TDriveFileLoader, 'Old Drive Files', '*.drv');
end.
