unit SPALoader;
{ Copyright (C) 2012 C M Western; see NLFit.pas for license}
{ Import Nicloet spectrometer .SPA files }
interface

uses ExpPlotLoader, Classes;

type

  TSPAFileLoader = class(TExperimentalPlotLoader)
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses SysUtils, Units, TextStrm, Math, MathSup;

function ReadWord(var p: PAnsiChar):Word;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PWord(p)^);
  {$ELSE}
    PWord(p)^;
  {$ENDIF}
  Inc(p,2);
end;

function ReadInteger(var p: PAnsiChar):Integer;
begin
  Result :=
  {$IFDEF FPC_BIG_ENDIAN}
    SwapEndian(PInteger(p)^);
  {$ELSE}
    PInteger(p)^;
  {$ENDIF}
  Inc(p,4);
end;

function ReadSingle(var p: PAnsiChar):Double;
var
  v: Integer;
begin
  v := ReadInteger(p);
  Result := PSingle(@v)^;
end;

{ TSPAFileLoader }

function TSPAFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  Position: Int64;
  Buffer: string[30];
  Line, Name: AnsiString;
  p, e, s: PAnsiChar;
  i, j, nSections, nPoints: Integer;
  SecPos, SecLen, SecType: Integer;
  FirstX, LastX: Double;
begin
  Result := False;
  if (Ext = '') or (Ext = '.spa') then begin
    Position := Stream.Position;
    SetLength(buffer,30);
    if (Stream.Read(Buffer[1],30) <> 30) or (Copy(buffer,1,8) <> 'Spectral') then begin
      { Unrecognized }
      Stream.Position := Position;
      Exit;
    end;
  end else
    Exit; { Wrong extension }
  Result := True;
  Stream.Position := Position;
  Line := '';
  repeat { Fill buffer with rest of file }
    i := Length(Line);
    SetLength(Line, i+100000);
    SetLength(Line, Stream.Read((PAnsiChar(Line)+i)^,100000)+i);
  until Length(Line) <> i+100000;
  p := PAnsiChar(Line);
  e := p + Length(Line);
  Inc(p,30);
  s := p;
  while Ord(p^) >= 32 do
    Inc(p);
  SetString(Name, s, p-s);
  Connector.Comment := Name;
  p := PAnsiChar(Line)+288;
  if ReadWord(p) <> 1 then
    Connector.ReportError('SPA file type not 1',False);
  Inc(p,2);
  ReadWord(p); // nSpectra
  nSections := ReadWord(p);
  for i := 0 to nSections-1 do begin
    p := PAnsiChar(Line)+288+16*i;
    SecType := ReadWord(p);
    SecPos := ReadInteger(p);
    SecLen := ReadInteger(p);
    p := PAnsiChar(Line) + SecPos;
    if (p > e) and (SecType <> 1) then
      Connector.ReportError('Invalid block',True);
    case SecType of
       2: begin
            Inc(p,4); nPoints := ReadInteger(p);
            Inc(p,8);
            FirstX := ReadSingle(p);
            LastX := ReadSingle(p);
          end;
       3: begin
            if nPoints < 0 then
              Connector.ReportError('Section 2 of SPA file not read', True);
            if SecLen <> nPoints*4 then
              Connector.ReportError(Format('Inconsistent file length: %d %d',[SecLen, nPoints*4]), False);
            CheckAvailableChannels(2, nPoints, AskAboutWeights);
            Connector.SetUnits(-1,cm1);
            Connector.SetIntensityUnits(Normalized);
            for j := 0 to nPoints - 1 do begin
              SetValue(-1,0,j,FirstX + (j-1) *(LastX - FirstX) / (nPoints-1));
              SetValue(-1,1,j,ReadSingle(p));
            end;
          end;
      27: begin
            SetString(Name, p, SecLen);
            s := PAnsiChar(Name);
            p := s + Length(s);
            while (p > s) and (p^ = #0) do
              Dec(p);
            SetLength(name, p-s+1);
            Connector.Comment := StringReplace(Name, #13#10, '', [rfReplaceAll]);
          end;
    end;
  end;

end;

initialization
  RegisterPlotLoader(TSPAFileLoader, 'Nicolet Files', '*.spa');
end.
