unit TabbedLine;
{ Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

function TryReadTabbedLine(s: PAnsiChar; V: array of PDouble; out nFields: Integer):Integer;

implementation

uses SysUtils, MathSup;

function TryReadTabbedLine(s: PAnsiChar; V: array of PDouble; out nFields: Integer):Integer;
{ Read a tab (or comma) separated set of Double values; empty fields cause value to
  be left unchanged;
  result is zero if end of string reached, otherwise 1 based offset to first char not read;
  If Newline found, routine exits with result pointing to new line;
  nFields is the number of values read (including null values) }
var
  P, B: PAnsiChar;
  i, Code: Integer;
  Buffer: ShortString;
  Value: Extended;
begin
  nFields := 0;
  Result := 0;
  if s = nil then
    Exit; { No values read }
  P := s;
  while nFields < Length(V) do begin
    while P^ = ' ' do Inc(p);
    if P^ = #0 then
      Exit
    else if P^ in [#10,#13] then
      Break
    else if P^ in [#9,','] then
      Inc(P) { Empty field }
    else begin
      { Must make a copy, as TextToFloat requires #0 terminator }
      B := @Buffer[1];
      for i := 0 to 253 do begin
        if P^ in [#0, ' ', #9, ',', #10, #13] then
          Break;
        B^ := P^;
        Inc(B);
        Inc(P);
      end;
      B^ := #0;
      if TextToFloat(PAnsiChar(@Buffer[1]), Value, fvExtended) then
        V[nFields]^ := Value
      else begin
        { Bad character in field }
        SetLength(Buffer,B-@Buffer[1]);
        if (CompareText(Buffer,'Nan') = 0) or (CompareText(Buffer,'Inf') = 0) then
          SetNan(V[nFields]^)
        else begin
          Result := (P - s) - (B-@Buffer[1]); { Point to problem item }
          Exit;
        end;
      end;
      while P^ = ' ' do Inc(p);
      if P^ in [#9,','] then
        Inc(P); { Non-empty field }
    end;
    Inc(nFields);
  end;
  if P^ = #0 then
    Result := 0
  else
    Result := P-s+1;
end;

end.
