unit CartesianDisplacementsFrm;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, VibGridFrm,
  ExtCtrls, Menus, Grids, NGrid, Split, InternalCoords, NewSpin, MNode,
  ComCtrls, StdCtrls, AlEdit, EditExtr, ScrlCtrl, GraphI, GraphM, Graphs;

type
  TCartesianDisplacementsForm = class(TVibGridForm)
    CheckItem: TMenuItem;
    CalculatefromlItem: TMenuItem;
    RemoveRTItem: TMenuItem;
    MakeExtraOperatorsItem: TMenuItem;
    procedure GridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: string);
    procedure CheckItemClick(Sender: TObject);
    procedure CalculatefromlItemClick(Sender: TObject);
    procedure RemoveRTItemClick(Sender: TObject);
    procedure GridGetCell(Sender: TObject; Col, Row: Integer; var v: Double;
      var Valid: Boolean);
    procedure MakeExtraOperatorsItemClick(Sender: TObject);
  private
    C: TCartesianDisplacements;
    function ActiveRow(Row: Integer):Boolean;
  protected
    procedure SetActiveObject(const Value: TMNode); override;
  public
    procedure PopulateForm(ACol:Integer); override;
    procedure Replot(ACol:Integer = -1); override;
  end;

function ConnectCartesianDisplacementsForm(CartesianDisplacements: TCartesianDisplacements): TComponent;

implementation

uses Math, MathSup, Mixture, Global, ChildForm, Vib, VibPlotFrm;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectCartesianDisplacementsForm(CartesianDisplacements: TCartesianDisplacements): TComponent;
begin
  Result := ConnectVibFrm(CartesianDisplacements, TCartesianDisplacementsForm);
end;

function TCartesianDisplacementsForm.ActiveRow(Row: Integer): Boolean;
begin
  Result := not(Assigned(C) and (Row <= Length(C.RowActive)) and not C.RowActive[Row-1])
end;

procedure TCartesianDisplacementsForm.CalculatefromlItemClick(Sender: TObject);
begin
  C.CalculateFromL(False,True);
  PopulateForm;
end;

procedure TCartesianDisplacementsForm.CheckItemClick(Sender: TObject);
begin
  ReShowLogForm;
  C.Prepare(True);
  C.CheckBMatrix(True,False);
  PopulateForm;
end;

procedure TCartesianDisplacementsForm.MakeExtraOperatorsItemClick(Sender: TObject);
begin
  BeginGlobalUpdate;
  try
    C.Makel(False, False, True);
  finally
    EndGlobalUpdate;
  end;
  if C.LastError <> '' then
    ShowInfo(C.LastError);
end;

procedure TCartesianDisplacementsForm.RemoveRTItemClick(Sender: TObject);
begin
  ReShowLogForm;
  C.Prepare(True);
  C.CheckBMatrix(True,True);
  PopulateForm;
end;

procedure TCartesianDisplacementsForm.GridGetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
begin
  if ActiveRow(Row) then
    inherited
  else if Assigned(C.B) and (Row > 0) and (Col > 0) and (Row <= C.B.RowCount) and (Col <= C.B.ColCount) then begin
    v := C.B[Row,Col];
    Valid := not IsNanOrInf(v);
  end;
end;

procedure TCartesianDisplacementsForm.GridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
begin
  if (Row > 0) and (Col > 0) and ActiveRow(Row) then begin
    s := '';
    IsText := True;
    inherited;
  end;
end;

procedure TCartesianDisplacementsForm.GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
begin
  if (ARow > 0) and (ACol > 0) and ActiveRow(ARow) then
    inherited;
end;

procedure TCartesianDisplacementsForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
begin
  if not ActiveRow(ARow) and Assigned(C) then begin
    if ARow > Length(C.RowActive) then
      SetLength(C.RowActive,ARow);
    C.RowActive[ARow-1] := True;
    Grid.Invalidate;
  end;
  inherited;
end;

procedure TCartesianDisplacementsForm.PopulateForm(ACol: Integer);
var
  Molecule: TVibratingMolecule;
  i, j, xyz: Integer;
begin
  if csLoading in ComponentState then
    Exit;
  Caption := ActiveObject.Name + ' for ' + (ActiveObject.Owner as TState).UniqueName;
  Molecule := C.Owner.Owner.Owner as TVibratingMolecule;
  Grid.ColCount := Max(Grid.ColCount, 3*Molecule.nNuclei+1);
  Grid.RowCount := Max(Grid.RowCount, 3*Molecule.nNuclei-5);
  if Assigned(C.B) then begin
    Grid.RowCount := Max(Grid.RowCount, C.B.RowCount+2);
    for i := 0 to High(C.RowActive) do
      if (not C.RowActive[i]) and (i < C.B.RowCount) then
        for j := 1 to C.B.ColCount do
          Grid.Value[i+1,j] := C.B[i+1,j];
  end;
  for i := 0 to Molecule.nNuclei-1 do
    for xyz := 1 to 3 do begin
      Grid.Title[i*3+xyz] := 'xyz'[xyz] + Molecule.Nuclei[i].Name;
      Grid.ColWidths[i*3+xyz] := Max(Grid.ColWidths[i*3+xyz], Grid.DefaultColWidth);
    end;
  for i := 1 to Grid.RowCount-2 do
    if Assigned(C.B) and (i <= C.B.RowCount) and Assigned(C.InternalCoords) then
      Grid.Lbl[i] := C.InternalCoords.nameR(i)
    else
      Grid.Lbl[i] := IntToStr(i);
  Grid.Lbl[Grid.RowCount-1] := '';
  inherited;
  Replot;
end;

procedure TCartesianDisplacementsForm.Replot(ACol: Integer);
begin
  if ACol < 0 then
    ACol := Grid.Row;
  if (ACol < 1) or (ACol > Length(C.Brows)) then
    ACol := -1
  else
    ACol := C.Brows[ACol-1];
  C.Prepare(False); { TODO: May need "modified" mechanism to avoid repeated calls }
  if Assigned(C.InternalCoords) then
    PlotZmatrix(C.InternalCoords.zMatrix, C.LastError+C.LastWarning, ACol, C.B, MagnificationSpin.Value, not C.InPrincipalAxis, -1)
  else
    PlotZmatrix(nil, C.LastError+C.LastWarning, ACol, C.B, MagnificationSpin.Value, not C.InPrincipalAxis, -1); { Just shows error message }
end;

procedure TCartesianDisplacementsForm.SetActiveObject(const Value: TMNode);
begin
  inherited;
  C := Value as TCartesianDisplacements;
  RowSelect := True;
end;

end.
