unit ChildForm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses SettingsForm, Classes, Forms, MNode, ComCtrls, StdCtrls, NewSpin;

type

  { TChildForm }

  TChildForm = class(TSettingsForm)
  private
    FBaseNode: TMNode;
    FCurrentMNode: TMNode;
    FCurrentMNodeName: string;
    procedure SetBaseNode(const Value: TMNode);
    procedure SetCurrentMNode(const AValue: TMNode);
  protected
    FReStuffPending: Boolean;
    FSymmetryList: TStringList;
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer); override;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); virtual; abstract;
    procedure SettingChanged(MNode: TMNode; Index: Integer); virtual; abstract;
    procedure PropNameChanged(MNode: TMNode; Index: Integer); virtual;
    procedure Stuff(MNode: TMNode; Prefix: string; ParentNode: TTreeNode);
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; virtual; abstract;
    procedure ComboBoxChange(Sender: TObject);
    class function Lockout: Integer;
    property CurrentMNode: TMNode read FCurrentMNode write SetCurrentMNode;
    property CurrentMNodeName: string read FCurrentMNodeName;
    procedure DoShow; override;
    procedure DoHide; override;
    procedure ApplyFix; override;
    procedure UpdateActions; override;
    procedure UpdateSymmetrySpin(Spin: TSpinBox; MNode: TMNode; HasFields: Boolean);
    procedure SymmetryGetIntegerValue(Sender: TObject; var NewValue:Longint; TheText: string; var StringValid: Boolean);
    procedure SymmetrySetIntegerValue(Sender: TObject; NewValue:Longint; var NewText: string);
  public
    destructor Destroy; override;
    procedure ReStuff; virtual;
    function CanReStuff:Boolean; virtual;
    procedure EndUpdate(DoReStuff:Boolean = True; Postpone: Boolean = False);
    function BaseNodeClass: TClass; virtual; abstract;
    procedure ShowItem(MNode: TMNode); virtual; abstract;
    procedure StuffSelector(MNode: TMNode); virtual;
    procedure BaseStuffSelector(MNode: TMNode); 
    procedure StuffSelectorHelper(MNode: TMNode; ComboBox: TComboBox);
    procedure SetItemIndex(ComboBox: TComboBox; MNode: TMNode);
    property BaseNode: TMNode read FBaseNode write SetBaseNode;
  end;

procedure BeginGlobalUpdate;
procedure EndGlobalUpdate(Postpone: Boolean = false);
function FindObject(TreeView: TTreeView; Data: Pointer):TTreeNode;

implementation

uses SysUtils, Mixture, FormUtils, Math, SymUtils;

var
  FLockout: Integer; { Global Variable, so all forms locked }

{ TChildForm }

destructor TChildForm.Destroy;
begin
  FBaseNode := nil;
  CurrentMNode := nil;
  FreeAndNil(FSymmetryList);
  inherited;
end;

procedure TChildForm.MNodeChange(Sender: TMnode; Change: TMNodeChange;
  Index: Integer);
  procedure NodeGone;
  begin
    if FBaseNode = Sender then
      FBaseNode := nil;
    { If node Sender is being deleted, need to detect case where Sender is
      an ancestor of CurrentMNode, as well as the obvious case of being equal }
    if Sender.IsAncestorOf(CurrentMNode) then
      CurrentMNode := Sender.Owner as TMNode;
    ReStuff;
  end;
begin
  inherited;
  if not Visible then Exit;
  if (Change <> mnNewBaseNode)
     and (not Assigned(BaseNode) or not Assigned(Sender) or (Sender.RootNode <> BaseNode.RootNode)) then Exit;
  case Change of
    mnDestroy: NodeGone;
    mnParameterChange:
               if Sender.IsAncestorOf(CurrentMNode) or CurrentMNode.IsAncestorOf(Sender)
                  or (CurrentMNode is TCorrelations)
                  or ((Sender is TMNodePair) and ((TMNodePair(Sender).Bra = CurrentMNode) or (TMNodePair(Sender).Ket = CurrentMNode))) 
                  then
                 ParameterChanged(Sender, Index);
    mnSettingChange:
               if Sender.IsAncestorOf(CurrentMNode) or CurrentMNode.IsAncestorOf(Sender) then
                 SettingChanged(Sender, Index);
    mnNameChange, mnCreate:  ReStuff;
    mnNewBaseNode: if Sender is BaseNodeClass then
                     BaseNode := Sender;
    mnPropNameChange:
               if Sender = CurrentMNode then
                 PropNameChanged(CurrentMNode, Index);
  end;
end;

procedure TChildForm.ReStuff;
var
  MNode: TMNode;
begin
  if not FReStuffPending then begin
    FReStuffPending := True;
    CurrentMNode := CurrentMNode; { Force update of saved name }
  end;
  if not CanReStuff then Exit;
  MNode := CurrentMNode; { Save current item }
  Inc(FLockout);
  try
    StuffSelector(BaseNode);
  finally
    Dec(FLockout);
  end;
  BeginGlobalUpdate;
  try
    if Assigned(MNode) and Assigned(BaseNode) then
      ShowItem(MNode)
    else if not Assigned(CurrentMNode) or Assigned(BaseNode) then
      ShowItem(CurrentMNode); { Show default item, or nothing }
  finally
    FReStuffPending := False;
    EndGlobalUpdate(False);
  end;
end;

function TChildForm.CanReStuff: Boolean;
begin
  Result := (FLockout = 0) and not (csDestroying in ComponentState);
end;

procedure TChildForm.SetBaseNode(const Value: TMNode);
begin
  CurrentMNode := nil;
  FBaseNode := Value;
  ReStuff;
end;

procedure TChildForm.SetCurrentMNode(const AValue: TMNode);
begin
  if not Assigned(BaseNode) then
    FCurrentMNode := nil
  else
    FCurrentMNode := AValue;
  if Assigned(CurrentMNode) then
    FCurrentMNodeName := CurrentMNode.UniqueName
  else
    FCurrentMNodeName := '??';
end;

procedure TChildForm.SetItemIndex(ComboBox: TComboBox; MNode: TMNode);
var
  i: Integer;
begin
  with ComboBox.Items do
    for i := 0 to Count-1 do
      if Objects[i] = MNode then begin
        ComboBox.ItemIndex := i;
        Exit;
      end;
  ComboBox.ItemIndex := -1;
end;

procedure TChildForm.Stuff(MNode: TMNode; Prefix: string; ParentNode: TTreeNode);
var
  i: Integer;
  s, c: string;
  function OptState(MNode: TMNode): string;
  begin
    if Assigned(MNode) then
      Result := MNode.Name
    else
      Result := '?';
  end;
begin
  if not Assigned(MNode) or (csDestroying in MNode.ComponentState) then Exit;
  c := Copy(MNode.ClassName,2,255) + ': ';
  if MNode is TMixture then begin
    s := MNode.Name;
    StuffObject(c, s, MNode, ParentNode);
    Prefix := '';
  end else if MNode is TMNodePair then
    with TMNodePair(MNode) do begin
      s := '<' + OptState(Bra) + '|' + PairName + '|' +
               OptState(Ket) + '>';
      StuffObject(c + Prefix, s, MNode, ParentNode);
      Prefix := Prefix + s + ' - ';
    end
  else begin
    s := MNode.Name;
    StuffObject(c + Prefix, s, MNode, ParentNode);
    Prefix := Prefix + MNode.Name + ' - ';
  end;
  for i := 0 to MNode.Count-1 do
    Stuff(MNode.Items[i], Prefix, ParentNode);
end;

procedure TChildForm.BaseStuffSelector(MNode: TMNode);
var
  E: Exception;
begin
  Inc(FLockout);
  if Assigned(MNode) and not (csDestroying in MNode.ComponentState) then begin
    E := MNode.FixupProperties;
    if Assigned(E) then begin
      Application.Hint := E.Message;
      E.Free;
    end;
  end else
    CurrentMNode := nil;
  Dec(FLockout);
end;

procedure TChildForm.StuffSelectorHelper(MNode: TMNode; ComboBox: TComboBox);
var
  Empty: Boolean;
begin
  BaseStuffSelector(MNode);
  BeginGlobalUpdate;
  try
    ComboBox.Clear;
    Empty := (MNode = nil) or (csDestroying in MNode.ComponentState);
    if not Empty then begin
      ComboBox.OnChange := {$IFDEF FPC}@{$ENDIF}ComboBoxChange;
      Stuff(MNode, '', nil);
      if ComboBox.Items.Count > 0 then
        ComboBox.ItemIndex := 0;
    end;
  finally
    EndGlobalUpdate(False);
  end;
  UpdateComboBox(ComboBox); { Restore from form setting if needed }
  ComboBoxChange(ComboBox);
end;

procedure TChildForm.StuffSelector(MNode: TMNode);
begin
  BaseStuffSelector(MNode);
end;

procedure TChildForm.ComboBoxChange(Sender: TObject);
var
  ComboBox: TComboBox;
begin
  ComboBox := Sender as TComboBox;
  Inc(FLockout);
  if FLockout = 1 then
    try
      if Assigned(FBaseNode) and (ComboBox.ItemIndex >= 0) then
        ShowItem(TMNode(ComboBox.Items.Objects[ComboBox.ItemIndex]))
      else
        ShowItem(nil)
    except
      Dec(FLockout);
      raise;
    end
  else if Assigned(FBaseNode) and (ComboBox.ItemIndex >= 0) then
    { Set default item to show if not otherwise set }
    CurrentMNode := TMNode(ComboBox.Items.Objects[ComboBox.ItemIndex]);
  Dec(FLockout);
end;

procedure TChildForm.PropNameChanged(MNode: TMNode; Index: Integer);
begin
 { Ignore }
end;

procedure BeginGlobalUpdate;
begin
  Inc(FLockout);
end;

procedure TChildForm.EndUpdate(DoReStuff: Boolean; Postpone: Boolean);
begin
  FReStuffPending := FReStuffPending or DoReStuff;
  EndGlobalUpdate(Postpone);
end;

procedure EndGlobalUpdate(Postpone: Boolean);
var
  i: Integer;
begin
  Dec(FLockout);
  if Postpone then Exit;
  if (FLockout = 0) then begin
    for i := 0 to Screen.FormCount-1 do
      if Screen.Forms[i] is TChildForm then
        if TChildForm(Screen.Forms[i]).FReStuffPending then
          TChildForm(Screen.Forms[i]).ReStuff;
  end;
end;

class function TChildForm.Lockout: Integer;
begin
  Result := FLockout;
end;

procedure TChildForm.DoShow;
begin
  FLockout := 0;
  inherited;
end;

procedure TChildForm.DoHide;
begin
  FBaseNode := nil;
  inherited DoHide;
end;

procedure TChildForm.ApplyFix;
var
  OldBaseNode: TMNode;
begin
  OldBaseNode := FBaseNode;
  inherited; { Hide may be called, but don't want to loose base node }
  FBaseNode := OldBaseNode;
end;

procedure TChildForm.UpdateActions;
begin
  inherited UpdateActions;
  if FReStuffPending and Visible then
    ReStuff;
end;

procedure TChildForm.UpdateSymmetrySpin(Spin: TSpinBox; MNode: TMNode; HasFields: Boolean);
{ Update/prepare a SpinBox for displaying symmetries }
var
  OldValue: string;
  NewValue: Integer;
begin
  OldValue := Spin.Text; { Save text, rather than number to allow for save/restore of form }
  NewValue := -1; { Default is -1 => all }
  Spin.Enabled := not HasFields;
  FreeAndNil(FSymmetryList);
  FSymmetryList := MakeSymmetryList(MNode);
  Spin.OnGetIntegerValue := {$IFDEF FPC}@{$ENDIF}SymmetryGetIntegerValue;
  Spin.OnSetIntegerValue := {$IFDEF FPC}@{$ENDIF}SymmetrySetIntegerValue;
  if Assigned(FSymmetryList) then begin
    Spin.MaxValue := FSymmetryList.Count-1;
    NewValue := FSymmetryList.IndexOf(OldValue);
    if NewValue < Spin.MinValue then
      { all not permitted in this spin box; use integer value if appropriate }
      if not TryStrToInt(OldValue, NewValue)
         or (NewValue < Spin.MinValue) or (NewValue > Spin.MaxValue) then
        NewValue := 0;
  end else
    Spin.MaxValue := 30;
  Spin.Value := NewValue;
end;

procedure TChildForm.SymmetryGetIntegerValue(Sender: TObject;
  var NewValue: Longint; TheText: string; var StringValid: Boolean);
begin
  if TheText = '' then begin
    StringValid := True;
    NewValue := -1; { Allow blank to mean 'all' in symmetry spin boxes }
  end else
    StringValid := TryStrToSymmetry(NewValue,TheText,FSymmetryList);
end;

procedure TChildForm.SymmetrySetIntegerValue(Sender: TObject;
  NewValue: Longint; var NewText: string);
begin
  NewText := SymmetryToStr(NewValue,FSymmetryList);
end;

function FindObject(TreeView: TTreeView; Data: Pointer):TTreeNode;
var
  i: Integer;
begin
  with TreeView.Items do
    for i := 0 to Count-1 do
      {$IFDEF OLDLAZARUS}
      if Items[i].Data = Data then begin
        Result := Items[i];
        Exit;
      end;
      {$ELSE}
      if Item[i].Data = Data then begin
        Result := Item[i];
        Exit;
      end;
      {$ENDIF}
  Result := nil;
end;

end.
