unit LineFrm;

{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, NGrid, ExtCtrls, Units, MNode,
  Mixture, StdCtrls, Buttons, Menus, QuantumNumber, Obs, Settings, SettingsForm;

type

  TStateDef = record
    Manifold: string;
    J:        TQuantumNumber;
    Sym:      Integer;
  end;

  TLineW = record
    Molecule:  string;
    Upper:     TStateDef;
    Lower:     TStateDef;
    Weight:    Single;
    HasFields: Boolean;
    Line:      TLine; { Note that eigenvalues not always set }
  end;

  TLineSortFunction = function (const A, B: TLineW): Integer;

  TLineListType = (llStandard, llBranch, llSimple);

  { TLineForm }

  TLineForm = class(TSettingsForm)
    Panel: TPanel;
    Grid: TNumberGrid;
    ClearButton: TButton;
    AssignButton: TButton;
    AllButton: TButton;
    NoneButton: TButton;
    CopyButton: TSpeedButton;
    CutButton: TSpeedButton;
    DeleteButton: TSpeedButton;
    DeleteUnassigned: TButton;
    MinusButton: TButton;
    Difference: TLabel;
    MoreButton: TButton;
    SortPopupMenu: TPopupMenu;
    FrequencySortItem: TMenuItem;
    UpperStateSortItem: TMenuItem;
    LowerStateSortItem: TMenuItem;
    StrengthSortItem: TMenuItem;
    WidthSortItem: TMenuItem;
    WeightSortItem: TMenuItem;
    Auto: TCheckBox;
    BranchItem: TMenuItem;
    TestFitButton: TButton;
    FitButton: TButton;
    UndoFitButton: TButton;
    SortonItem: TMenuItem;
    OptionsItem: TMenuItem;
    CopyBranchTableItem: TMenuItem;
    UpperStateCDFitItem: TMenuItem;
    LowerStateCDfitItem: TMenuItem;
    SavetofileItem: TMenuItem;
    SaveDialog: TSaveDialog;
    GridPopupMenu: TPopupMenu;
    CopySimpleFormatItem: TMenuItem;
    InStandardFormatItem: TMenuItem;
    InBranchFormatItem: TMenuItem;
    InSimpleFormatItem: TMenuItem;
    procedure UndoFitButtonClick(Sender: TObject);
    procedure BranchItemClick(Sender: TObject);
    procedure GridGetCellLabel(Sender: TObject; Col, Row: Integer;
      var s: string; var IsText: Boolean);
    procedure ClearButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure NoneButtonClick(Sender: TObject);
    procedure CutButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure CloseButtonClick(Sender: TObject);
    procedure AssignButtonClick(Sender: TObject);
    procedure GridGetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure GridChangeCellQuery(Sender: TObject; Col, Row: Integer;
      var CanChange: Boolean);
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure DeleteUnassignedClick(Sender: TObject);
    procedure MinusButtonClick(Sender: TObject);
    procedure StrengthSortItemClick(Sender: TObject);
    procedure UpperStateSortItemClick(Sender: TObject);
    procedure FrequencySortItemClick(Sender: TObject);
    procedure LowerStateSortItemClick(Sender: TObject);
    procedure WidthSortItemClick(Sender: TObject);
    procedure WeightSortItemClick(Sender: TObject);
    procedure MoreButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
    procedure OptionsItemClick(Sender: TObject);
    procedure SavetofileItemClick(Sender: TObject);
    procedure GridDrawCellContents(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
    procedure GridSetCell(Sender: TObject; ACol, ARow: Integer; var v: Double; var Valid: Boolean);
  private
    NotEmpty: Boolean;
    Lines: array of TLineW;
    nLines: Integer; { Number of valid entries in Lines - not necessarily all displayed }
    nPendingLines: Integer;
    FLongestNameLength: Integer;
    FnLinesToShow: Integer;
    FMinIntensity, FMinAssignedIntensity: Double;
    FShowZeroIntensityLines: Boolean;
    procedure InternalAddLines(First, Last: Integer);
    function nVisibleLines: Integer; { Number of entries in Lines[] that have actually been displayed - normally equal to nLines }
    function StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
    procedure LoadOptionsDialog;
    procedure UnLoadOptionsDialog;
    procedure SetLineByIndex(var Line: TLineW; const Value: string; Index: Integer);
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
  protected
    procedure UpdateActions; override;
    procedure AddExtraVariables(AFormSettings: TFormSettings); override;
    procedure WriteExtraVariables(AFormSettings: TFormSettings); override;
    procedure HandleControl(Control: TControl; var PName, Value, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure WriteSettings(AFormSettings: TFormSettings; DoPositions: Boolean; DoSettings: Boolean); override;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AddLine(ALine: TLine; AUnits: TResonanceUnits);
    procedure ArrangeForm(Screen: TRect); override;
    procedure ShowLines(Simulation: TSimulation; FilterLines: Boolean);
    procedure BringUp;
    procedure ClearTLines;
    function HasObservations:Boolean;
    procedure LoadObservations(const Source: string; ObsType: TObsType; TakeWeight: TTakeWeight);
    procedure SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
    procedure CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean; LineListType: TLineListType);
    procedure AssignLines(APosition: Double; Units: TResonanceUnits; SourceName: string; Source: TMNode; Force: Boolean);
  end;

var
  LineForm: TLineForm;

implementation

uses {$IFDEF LCL_} Types, {$ENDIF} Math, MathSup, TextAssign, Huge, ClipBrd, SlaveEdit,
     LogUnit, Global, Main, LineOpt, ResidFrm, TextStrm, WCompat, FixFnt, Worker;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

var
  CF_Measurement, CF_Difference: DWORD;

{ TLineForm }

procedure TLineForm.AddExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
  Stream: TStringStream;
begin
  inherited;
  HandleControls(LineFrmOptDlg, {$IFDEF FPC}@{$ENDIF}AddValue, AFormSettings);
  Index := AFormSettings.IndexOfSetting('QuantumNumberFormat');
  if Index < 0 then
    Index := AFormSettings.AddSetting('QuantumNumberFormat', stString, 'J');
  AFormSettings.SetSettingFromStr(Index, QuantumNumberFormatNames[Ord(GlobalMixture.QuantumNumberFormat)]);
  Index := AFormSettings.IndexOfSetting('Lines');
  if Index < 0 then
    Index := AFormSettings.AddSetting('Lines', stString, '');
  if nVisibleLines > 0 then begin
    Stream := TStringStream.Create('');
    try
      Grid.SaveToStream(Stream);
      AFormSettings.SetSettingFromStr(Index,sLineBreak+Stream.DataString);
    finally
      Stream.Free;
    end;
  end else
    AFormSettings.SetSettingFromStr(Index,'');
end;

procedure TLineForm.AddLine(ALine: TLine; AUnits: TResonanceUnits);
var
  Rect: TGridRect;
  function StateDef(Eigenvalues: TEigenvalues; AIndex: Integer; var HasFields: Boolean):TStateDef;
  begin
    with Result do
      if Assigned(Eigenvalues) then begin
        HasFields := HasFields or Eigenvalues.HasFields;
        Manifold := Eigenvalues.Manifold.Name;
        J := Eigenvalues.J;
        Sym := Eigenvalues.Sym;
      end else begin
        Manifold := '';
        J := -1;
        Sym := -1;
      end;
  end;
begin
  { Reset scrolling so that first column is in view }
  Rect := Grid.Selection;
  Rect.Left := 1;
  Rect.Right := 1;
  Grid.Selection := Rect;
  if nLines >= Length(Lines) then
    SetLength(Lines, nLines+50);
  with Lines[nLines] do begin
    Line := ALine;
    HasFields := False;
    ConvertLine(Line, AUnits, GlobalMixture.Simulations[0].PlotUnits);
    if Assigned(ALine.EigenvaluesL) then
      Molecule := ALine.EigenvaluesL.Manifold.Molecule.Name;
    Upper := StateDef(ALine.EigenvaluesU, ALine.IndexU, HasFields);
    Lower := StateDef(ALine.EigenvaluesL, ALine.IndexL, HasFields);
    SetNan(Weight);
    if GlobalMixture.ShowEstUnc and Assigned(ALine.EigenvaluesL) then begin
      Weight := EstErr(ALine.EigenvaluesU, ALine.EigenvaluesL, ALine.IndexU, ALine.IndexL);
      if Weight <= 0 then
        SetNan(Weight) { No errors set }
      else
        Weight := -ConvertWidth(Weight, ALine.Frequency, AUnits, GlobalMixture.Simulations[0].PlotUnits)
    end;
  end;
  Inc(nLines);
end;

procedure TLineForm.SortLines(SortFunction: TLineSortFunction; First, Last: Integer);
var
  N, I, J, Offset: Integer;
  IsInOrder: Boolean;
  T: TLineW;
begin
  Grid.EditorMode := False;
  { ShellSort }
  N := Last-First+1;
  Offset := N;
  while Offset > 1 do begin
    Offset := Offset shr 1;
    repeat
      IsInOrder := True;
      for J := First to First + N - Offset - 1 do begin
        I := J + Offset;
        if SortFunction(Lines[I], Lines[J]) > 0 { I > J? } then begin
          IsInOrder := False;
          T := Lines[I];
          Lines[I] := Lines[J];
          Lines[J] := T;
        end;
      end;
    until IsInOrder;
  end;
  InternalAddLines(First, Last);
end;

function ByIntensity(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Strength - B.Line.Strength);
end;

function ByFrequency(const A, B: TLineW): Integer;
begin
 Result := -Sign(A.Line.Frequency - B.Line.Frequency);
end;

function ByWidth(const A, B: TLineW): Integer;
begin
 Result := Sign(A.Line.Width - B.Line.Width);
end;

function ByWeight(const A, B: TLineW): Integer;
begin
 if IsNanOrInf(A.Weight) then
   if IsNanOrInf(B.Weight) then
     Result := 0
   else
     Result := -1
 else if IsNanOrInf(B.Weight) then
   Result := 1
 else
   Result := Sign(A.Weight - B.Weight);
end;

function ByEigenvalues(const A, B: TStateDef): Integer;
begin
  Result := CompareText(A.Manifold, B.Manifold);
  if Result = 0 then
    Result := A.J - B.J;
  if Result = 0 then
    Result := A.Sym - B.Sym;
end;

function ByUpperState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Upper, B.Upper);
 if Result = 0 then
   Result := A.Line.IndexU - B.Line.IndexU;
 Result := -Result;
end;

function ByLowerState(const A, B: TLineW): Integer;
begin
 Result := CompareText(A.Molecule, B.Molecule);
 if Result = 0 then
   Result := ByEigenvalues(A.Lower, B.Lower);
 if Result = 0 then
   Result := A.Line.IndexL - B.Line.IndexL;
 Result := -Result;
end;

procedure SplitCachedName(var Line: TLine; var Branch, TxName: string);
var
  p: Integer;
begin
  TxName := MakeTxName(Line);
  p := Pos(':',TxName);
  if Assigned(Line.EigenvaluesL) and not Assigned(Line.EigenvaluesU) then
    p := 0; { One state only - no branch to extract }
  if p > 0 then begin
    Branch := Copy(TxName,1,p-1);
    if (Length(Branch) > 0) and (Branch[Length(Branch)] = ' ') then
      Delete(Branch,Length(Branch),1);
    Delete(TxName,1,p);
    if (TxName <> ' ') and (TxName[1] = ' ') then
      Delete(TxName,1,1);
  end else
    Branch := '';
end;

function ByBranch(const A, B: TLineW): Integer;
var
  BranchA, BranchB: string;
  function RemoveBrackets(const L: TLineW):string;
  { Branch label without (J) }
  var
    Line: TLine;
    Junk: string;
    s, e: Integer;
  begin
    Line := L.Line;
    SplitCachedName(Line, Result, Junk);
    s := Pos('(',Result);
    e := Pos(')',Result);
    if (s > 0) and (e > s) then
      Delete(Result, s, e-s+2);
  end;
begin
  BranchA := RemoveBrackets(A);
  BranchB := RemoveBrackets(B);
  Result := CompareText(BranchB, BranchA);
  if Result = 0 then
    Result := B.Lower.J - A.Lower.J;
end;

procedure TLineForm.ShowLines(Simulation: TSimulation; FilterLines: Boolean);
{ Show the lines that have not been shown yet }
var
  i, f, VisibleRows: Integer;
  Rect: TGridRect;
  Problem: Exception;
begin
  if nLines <= nVisibleLines then Exit;
  Grid.BeginUpdate;
  f := nVisibleLines;
  Problem := nil;
  for i := f to nLines - 1 do
    with Lines[i].Line do
      if Assigned(EigenvaluesL) then
        Strength := Simulation.LineIntensity(Lines[i].Line, Simulation.PlotUnits, Problem);
  Problem.Free; { Easiest to ignore problems }
  if FilterLines then
    SortLines({$IFDEF FPC} @ {$ENDIF}ByIntensity, nVisibleLines, nLines-1)
  else
    SortLines({$IFDEF FPC} @ {$ENDIF}ByFrequency, nVisibleLines, nLines-1);
  if FilterLines then begin
    nLines := f + Min(nLines-f,FnLinesToShow);
    for i := nLines-1 downto f do
      if ( not FShowZeroIntensityLines and (Lines[i].Line.Strength = 0) )
        or (Lines[i].Line.Strength < FMinIntensity*Lines[f].Line.Strength) then
        nLines := i;
  end;
  if nLines = 0 then
    ClearButtonClick(nil)
  else begin
    Grid.RowCount := nLines+1;
    nPendingLines := 0;
    Rect.Top := f+1;
    Rect.Bottom := Grid.RowCount-1;
    Rect.Left := 1;
    Rect.Right := Grid.ColCount-1;
    Grid.Selection := Rect;
    VisibleRows := {$IFDEF LCLno} Grid.Height div (Grid.DefaultRowHeight + Grid.GridLineWidth) - 1 {$ELSE} Grid.VisibleRowCount {$ENDIF};
    if Rect.Bottom > VisibleRows then
      Grid.TopRow := Min(Rect.Bottom - VisibleRows + 1, Rect.Top);
  end;
  BringUp;
  Grid.EndUpdate;
end;

procedure TLineForm.InternalAddLines(First, Last: Integer);
var
  i, dn: Integer;
  procedure MakeTitle(U, L: TEigenvalues);
  begin
    if Assigned(U) and Assigned(L) and (Length(U.Basis) > 1) then
      Grid.Title[14] := U.Basis[1].TxNameKey(U.Basis[1],L.Basis[1])
    else
      Grid.Title[14] := 'Branch';
    if Assigned(L) and (Length(L.Basis) > 1) then begin
      Grid.Title[15] := L.Basis[1].StateLabelKey;
      if Assigned(U) and (Length(U.Basis) > 1) then
        Grid.Title[15] := U.Basis[1].StateLabelKey + ' - ' + Grid.Title[15];
    end else
      Grid.Title[15] := 'Name';
  end;
  procedure Load(r,c: Integer; Value: Double; ShowNegative:Boolean = False);
  begin
    if ShowNegative or (Value >= 0) then
      Grid[r,c] := Value
    else
      Grid.Valid[r,c] := False;
  end;
begin
  if First > Last then Exit;
  dn := Last - nVisibleLines + 1; { Number added }
  if dn > 0 then
    if NotEmpty then
      Grid.RowCount := Grid.RowCount + dn
    else
      Grid.RowCount := Grid.RowCount + dn - 1;
  NotEmpty := True;
  for i := First+1 to Last+1 do
    with Lines[i-1] do begin
      Load(i,3,QuantumNumberToDisplayDouble(Upper.J,GlobalMixture.QuantumNumberFormat),HasFields);
      Load(i,4,Upper.Sym);
      Load(i,5,Line.IndexU);
      Load(i,7,QuantumNumberToDisplayDouble(Lower.J,GlobalMixture.QuantumNumberFormat),HasFields);
      Load(i,8,Lower.Sym);
      Load(i,9,Line.IndexL);
      Grid[i,10] := Line.Frequency;
      Grid[i,11] := Weight;
      Grid[i,12] := Line.Strength;
      Grid[i,13] := Line.Width;
    end;
  if First = 0 then begin
    MakeTitle(Lines[0].Line.EigenvaluesU,Lines[0].Line.EigenvaluesL);
    Grid.AutoAdjustColumn(14);
  end;
  Grid.Invalidate;
end;

procedure TLineForm.BringUp;
begin
  if WindowState in [wsMinimized, wsMaximized] then
    WindowState := wsNormal;
  if not Visible then
    Show
  else
    BringToFront;
end;

procedure TLineForm.ClearButtonClick(Sender: TObject);
var
  i, nAssigned: Integer;
begin
  if Assigned(Sender) then begin
    nAssigned := 0;
    for i := 1 to nVisibleLines do
      if Grid.Valid[i,11] and (Grid.Value[i,11] > 0) then
        Inc(nAssigned);
    if (nAssigned > 0)
       and (MessageDlg(Format('%d assigned lines - are you sure you want to delete them?', [nAssigned]),
                      mtWarning, mbOKCancel, 0) <> mrOK) then
      Exit;
  end;
  Grid.RowCount := 2;
  NotEmpty := False;
  nLines := 0;
  for i := 1 to Grid.ColCount-1 do
    Grid.Valid[1,i] := False;
  Grid.ColWidths[Grid.ColCount-1] := 150;
  Grid.Invalidate;
end;

procedure TLineForm.GridDrawCellContents(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
var
  buffer: string;
  Pitch: TFontPitch;
  FontName: TFontName;
  L: Integer;
begin
  Drawn := StringRep(ARow, ACol, buffer);
  if not Drawn then
    Exit;
  if ACol = 15 then begin
    Pitch := Grid.Canvas.Font.Pitch;
    FontName := Grid.Canvas.Font.Name;
    Grid.Canvas.Font.Name := DefaultFixedName;
    {$IFDEF MSWINDOWS}
    Grid.Canvas.Font.Size := Grid.Canvas.Font.Size + 1;
    {$ENDIF}
    Grid.Canvas.Font.Pitch := fpFixed;
    L := Grid.Canvas.TextWidth(buffer) + 10;
    if FLongestNameLength < L then
      FLongestNameLength := L;
    TextInGrid(Grid.Canvas, Rect, buffer);
    {$IFDEF MSWINDOWS}
    Grid.Canvas.Font.Size := Grid.Canvas.Font.Size - 1;
    {$ENDIF}
    Grid.Canvas.Font.Name := FontName;
    Grid.Canvas.Font.Pitch := Pitch;
  end else
    TextInGrid(Grid.Canvas, Rect, buffer)
end;

procedure TLineForm.AllButtonClick(Sender: TObject);
var
  Rect: TGridRect;
begin
  Rect.Top := 1;
  Rect.Bottom := Grid.RowCount-1;
  Rect.Left := 1;
  Rect.Right := Grid.ColCount-1;
  Grid.Selection := Rect;
end;

procedure TLineForm.ArrangeForm(Screen: TRect);
var
  R, MainRect: TRect;
begin
  inherited;
  MainRect := ScreenBounds(MainForm);
  R := ScreenBounds(Self);
  R.Left := MainRect.Left;
  R.Right := MainRect.Right;
  OffsetRect(R,0,MainRect.Bottom-R.Top);
  if R.Bottom > Screen.Bottom then
    OffsetRect(R,0,Screen.Bottom-R.Bottom);
  if LogForm.Visible then
    Dec(R.Right,25);
  SetScreenBounds(Self,R);
end;

procedure TLineForm.NoneButtonClick(Sender: TObject);
begin
  Grid.CancelSelection;
end;

procedure TLineForm.CutButtonClick(Sender: TObject);
begin
  CopyButtonClick(Sender);
  DeleteButtonClick(Sender);
end;

procedure TLineForm.CopyToStream(Stream: TStream; Row1, Row2: Integer; All: Boolean; LineListType: TLineListType);
var
  F: TextFile;
  i: Integer;
  buffer: string;
  LastNQN: Integer;
  procedure ProcessColumn(j: Integer);
  begin
    case j of
    3,7: begin
           StringRep(i, j, buffer);
           if buffer = '' then
             Write(F, ' - ')
           else
             Write(F, buffer:3, ' ');
         end;
     10: Write(F, FloatToStr(Grid[i,j]):12, ' ');
     11: if Grid.Valid[i,j] then
           Write(F, FloatToStr(Grid[i,j]), ' ')
         else
           Write(F, '- ');
     12, 13: Write(F, dtosm(Grid[i,j], 6, -1):6, ' '); { Limit Strength and width to 6 characters }
     14: begin StringRep(i,j,buffer); Write(F,': ', buffer); end;
     15: begin StringRep(i,j,buffer); WriteLn(F,' : ', buffer); end;
    else
      if StringRep(i, j, buffer) then
        Write(F, buffer:3, ' ')
      else if not Grid.Valid[i,j] then
        Write(F, ' - ')
      else
        Write(F, FloatToStr(Grid[i,j]):2, ' ');
    end;
  end;
  procedure MakeStandard;
  var
    col: Integer;
  begin
    for col := 1 to Grid.ColCount-1 do
      ProcessColumn(col);
  end;
  function ExtractSimple(s: string; Manifold: TManifold):Boolean;
  var
    p, NQN: Integer;
    InWord: Boolean;
  begin
    Result := False;
    if not Assigned(Manifold) then
      Exit;
    { Given the part transition name that specifies a state, extract the "Simple Format" quantum numbers }
    { Strip manifold if present }
    if Pos(Manifold.Name+' ',s) = 1 then
      Delete(s,1,Length(Manifold.Name)+1);
    { Strip state name; logic here allows state name to include spaces }
    for p := 1 to Length(s)-1 do { -1 as blank string not allowed }
      if (s[p] = ' ') and Assigned(Manifold.StateWithName(Copy(s,1,p-1))) then begin
        s := Copy(s,p+1,Length(s));
        Result := true;
        Break;
      end;
    if not Result then
      Exit;
    Manifold.Molecule.FixSimpleFormat(s);
    buffer := buffer + s + ' ';
    { Check NQN }
    InWord := False;
    NQN := 0;
    for p := 1 to Length(s) do
      if (s[p] <> ' ') <> Inword then begin
        Inword := not Inword;
        if InWord then
          Inc(NQN);
      end;
    if LastNQN <> NQN then begin
      WriteLn(F, 'NQN ', NQN);
      LastNQN := NQN;
    end;
  end;
  procedure MakeSimple;
  var
    Pre, Upper, Lower, Post: string;
    Molecule: TMolecule;
  begin
    Molecule := GlobalMixture.FindMolecule(Lines[i].Molecule);
    if not Assigned(Molecule) then
      Exit;
    StringRep(i,15,buffer);
    SplitObsName(buffer, Pre, Upper, Lower, Post);
    buffer := '';
    if not (ExtractSimple(Upper, Molecule.ManifoldWithName(Lines[i].Upper.Manifold))
           and ExtractSimple(Lower, Molecule.ManifoldWithName(Lines[i].Lower.Manifold))) then begin
      Write(F, '#Unconverted: ');
      MakeStandard;
    end;
    Write(F, buffer);
    ProcessColumn(10); { Position }
    ProcessColumn(11); { Weight }
    ProcessColumn(12); { Intensity }
    WriteLn(F);
  end;
begin
  AssignStream(F, Stream);
  Append(F);
  LastNQN := 0;
  for i := Row1 to Row2 do begin
    if All or Grid.Valid[i,11] then
      case LineListType of
        llStandard: MakeStandard;
        llBranch:
          begin
            StringRep(i,14,buffer); Write(F, buffer, ' ');
            ProcessColumn(10); { Position }
            ProcessColumn(11); { Weight }
            ProcessColumn(12); { Intensity }
            ProcessColumn(15); { Comment }
          end;
        llSimple: MakeSimple;
      end;
  end;
  System.Close(F);
end;


constructor TLineForm.Create(AOwner: TComponent);
begin
  FnLinesToShow := 10;
  FMinIntensity := 0.05;
  FShowZeroIntensityLines := False;
  FMinAssignedIntensity := 0.25;
  inherited;
  {$IFNDEF VCL_}
  CutButton.Height := AssignButton.Height;
  CutButton.Width := AssignButton.Height;
  CopyButton.Height := AssignButton.Height;
  CopyButton.Width := AssignButton.Height;
  DeleteButton.Height := AssignButton.Height;
  DeleteButton.Width := AssignButton.Height;
  {$ENDIF}
end;

procedure TLineForm.CopyButtonClick(Sender: TObject);
var
  s: THugeMemoryStream;
  LineListType: TLineListType;
begin
  if not NotEmpty then Exit;
  if Sender = CopyBranchTableItem then
    LineListType := llBranch
  else if Sender = CopySimpleFormatItem then
    LineListType := llSimple
  else
    LineListType := llStandard;
  s := THugeMemoryStream.Create;
  try
    CopyToStream(s, Grid.Selection.Top, Grid.Selection.Bottom, True, LineListType);
    s.CopyToClipboard(CF_TEXT);
  finally
    s.Free;
  end;
end;

procedure TLineForm.DeleteButtonClick(Sender: TObject);
var
  i, d, n, e: Integer;
begin
  { Remember how many lines to recopy }
  d := Grid.Selection.Top;
  n := Grid.Selection.Bottom+1-d;
  e := Grid.RowCount - n - 2;
  for i := d-1 to e do
    Lines[i] := Lines[i+n];
  if d = 1 then
    ClearButtonClick(nil)
  else
    Grid.RowCount := d;
  nLines := e+1;
  { Copy remaining lines down }
  InternalAddLines(d-1, e);
end;

procedure TLineForm.CloseButtonClick(Sender: TObject);
begin
  Close;
end;

type
  ClipboardMeasurement = packed record
    Value: Double;
    Signature: array [0..3] of AnsiChar;
    Filename: AnsiChar;
  end;
  PClipboardMeasurement = ^ClipboardMeasurement;

procedure TLineForm.AssignButtonClick(Sender: TObject);
var
  Value: Double;
  CHandle: TGHandle;
  Clip: PClipboardMeasurement;
  FileName: string;
begin
  if ClipBoard.HasFormat(CF_TEXT) then begin
    if CF_Measurement = 0 then
      CF_Measurement := RegisterClipboardFormat('Measurement');
    if not TryStrToFloatEx(Trim(ClipBoard.AsText),Value) then
      raise EReportException.Create(ClipBoard.AsText+' is not a valid number');
    if ClipBoard.HasFormat(CF_Measurement) then begin
      CHandle := TGHandle.Create;
      try
        CHandle.SetFromClipboard(CF_Measurement);
        Clip := PClipboardMeasurement(CHandle.Base);
        if Clip^.Signature = 'MEAS' then begin
          Value := Clip^.Value;
          FileName := ExtractFileName(PAnsiChar(@Clip^.FileName));
        end;
      finally
        CHandle.Free;
      end;
    end else
      FileName := '';
    AssignLines(Value, MainForm.Simulation.PlotUnits, Filename, nil, True);
  end;
end;

procedure TLineForm.AssignLines(APosition: Double; Units: TResonanceUnits; SourceName: string;
  Source: TMNode; Force: Boolean);
var
  StrMax: Double;
  s, Branch: string;
  i, p: Integer;
begin
  if not (NotEmpty and Visible) then
    Exit;
  if not Force and (not Auto.Checked or (SourceName = 'Calculated')) then
    Exit;
  StrMax := Lines[Grid.Selection.Top-1].Line.Strength;
  for i := Grid.Selection.Top to Grid.Selection.Bottom do begin
    { Avoid assignment from calculation of same transition }
    if Assigned(Source) then
      with Lines[i-1].Line do begin
        if Assigned(EigenValuesU) and Source.IsAncestorOf(EigenvaluesU.Manifold) then
          Exit;
        if Assigned(EigenValuesL) and Source.IsAncestorOf(EigenvaluesL.Manifold) then
          Exit;
      end;
    if StrMax < Lines[i-1].Line.Strength then
      StrMax := Lines[i-1].Line.Strength;
  end;
  for i := Grid.Selection.Top to Grid.Selection.Bottom do
    if Lines[i-1].Line.Strength > StrMax*FMinAssignedIntensity then begin
      Lines[i-1].Line.Frequency := Convert(APosition,Units,MainForm.Simulation.PlotUnits);
      SplitCachedName(Lines[i-1].Line, Branch, s);
      p := Pos(':', s);
      if p > 0 then begin
        if (p > 1) and (s[p-1] = ' ') then
          Dec(p); { Trim trailing space }
        Delete(s,p,Length(s));
      end;
      if SourceName <> '' then
        s := s + ' : ' + ExtractFileName(SourceName);
      if Branch <> '' then
        s := Branch + ' : ' + s;
      Lines[i-1].Line.CachedName := s;
      Lines[i-1].Weight := 1;
    end;
  InternalAddLines(Grid.Selection.Top-1, Grid.Selection.Bottom-1);
end;

function TLineForm.StringRep(ARow, ACol: Integer; var buffer: string): Boolean;
var
  Junk: string;
  function JValue(J: TQuantumNumber; ShowNegative: Boolean):string;
  begin
    if ShowNegative or (J >= 0) then
      Result := GlobalMixture.QuantumNumberToStr(J)
    else
      Result := '';
  end;
begin
  Result := NotEmpty and (ARow > 0);
  if Result then
    case ACol of
      1: buffer := Lines[ARow-1].Molecule;
      2: buffer := Lines[ARow-1].Upper.Manifold;
      3: buffer := JValue(Lines[ARow-1].Upper.J, Lines[ARow-1].HasFields);
      6: buffer := Lines[ARow-1].Lower.Manifold;
      7: buffer := JValue(Lines[ARow-1].Lower.J, Lines[ARow-1].HasFields);
     14: SplitCachedName(Lines[ARow-1].Line, buffer, Junk);
     15: SplitCachedName(Lines[ARow-1].Line, Junk, buffer);
    else
      Result := False;
    end;
end;

procedure TLineForm.GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: String);
begin
  StringRep(ARow, ACol, Value);
end;

procedure TLineForm.GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
begin
  // Was : CanChange := CanChange and (Col in [10,11,14]);
end;

procedure TLineForm.SavetofileItemClick(Sender: TObject);
var
  Stream: TFileStream;
  r: Integer;
  LineListType: TLineListType;
begin
  if Sender = InBranchFormatItem then
    LineListType := llBranch
  else if Sender = InSimpleFormatItem then
    LineListType := llSimple
  else
    LineListType := llStandard;
  if SaveDialog.Execute then begin
    if FileExists(SaveDialog.FileName) then
      r := MessageDlg('File exists - append to file instead of overwriting?', mtWarning, [mbYes, mbNo, mbCancel], 0)
    else
      r := mrNo;
    case r of
      mrYes: begin; Stream := TFileStream.Create(SaveDialog.FileName, fmOpenReadWrite); Stream.Seek(0, soFromEnd); end;
      mrNo:  Stream := TFileStream.Create(SaveDialog.FileName, fmCreate);
    else
      Exit;
    end;
    try
      CopyToStream(Stream, 1, nLines, True, LineListType);
    finally
      Stream.Free;
    end;
    LogForm.ObsFileName.FileName := SaveDialog.FileName;
  end;
end;

procedure TLineForm.SetLineByIndex(var Line: TLineW; const Value: string; Index: Integer);
var
  v: Double;
  Branch, TxName: string;
  iv: Integer;
begin
  if Index in [10,11,12,13] then
    if not TryStrToFloatEx(Value, v) then
      SetNan(v);
  iv := 0;
  if Index in [3,4,5,7,8,9] then
    TryStrToInt(Value,iv);
  case Index of { This copy back is required for loading grid from .pgo files  }
     1: Line.Molecule := Value;
     2: Line.Upper.Manifold := Value;
     3: Line.Upper.J := iv;
     4: Line.Upper.Sym := iv;
     5: Line.Line.IndexU := iv;
     6: Line.Lower.Manifold := Value;
     7: Line.Lower.J := iv;
     8: Line.Lower.Sym := iv;
     9: Line.Line.IndexL := iv;
    10: Line.Line.Frequency := v;
    11: Line.Weight := v;
    12: Line.Line.Strength := v;
    13: Line.Line.Width := v;
    14,15:
      begin
        SplitCachedName(Line.Line, Branch, TxName);
        if Index = 14 then
          Branch := Value
        else
          TxName := Value;
        if Branch <> '' then
          TxName := Branch + ' : ' + Value;
        Line.Line.CachedName := TxName;
      end;
  end;
end;

procedure TLineForm.GridSetCell(Sender: TObject; ACol, ARow: Integer; var v: Double; var Valid: Boolean);
var
  buffer: string;
begin
  { This allows the delete key to act on cell ranges }
  if not Valid and (ARow <= Length(Lines)) and (ACol in [1,2,6,14,15])
     and StringRep(ARow, ACol, buffer) and (buffer <> '') then begin
    SetLineByIndex(Lines[ARow-1],'',ACol);
    Grid.Modified := True;
    Grid.Changed(ACol, ARow);
  end;
end;

procedure TLineForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: String);
var
  Q: TQuantumNumber;
begin
  if ACol in [3,7] then begin
    if TryStrToQuantumNumber(Value, Q, GlobalMixture.QuantumNumberFormat) then
      case ACol of
        3: Lines[ARow-1].Upper.J := Q;
        7: Lines[ARow-1].Lower.J := Q;
      end;
  end else
    SetLineByIndex(Lines[ARow-1],Value,ACol);
end;

function TLineForm.nVisibleLines: Integer;
begin
  if NotEmpty then
    Result := Grid.RowCount-1
  else
    Result := 0
end;

procedure TLineForm.OptionsItemClick(Sender: TObject);
begin
  LoadOptionsDialog;
  if LineFrmOptDlg.ShowModal = mrOk then
    UnLoadOptionsDialog;
end;

procedure TLineForm.DeleteUnassignedClick(Sender: TObject);
var
  s, e: Integer;
  r: TGridRect;
begin
  Grid.BeginUpdate;
  repeat
    e := nVisibleLines;
    while (e > 0) and Grid.Valid[e,11] and (Grid.Value[e,11] > 0) do
      Dec(e);
    s := e;
    while (s > 0) and not (Grid.Valid[s,11] and (Grid.Value[s,11] > 0)) do
      Dec(s);
    Inc(s);
    if e >= s then begin
      r.Top := s;
      r.Bottom := e;
      r.Left := 1;
      r.Right := 2;
      Grid.Selection := r;
      DeleteButtonClick(nil);
    end else
      Break;
  until False;
  Grid.EndUpdate;
end;


procedure TLineForm.MinusButtonClick(Sender: TObject);
var
  d: Double;
  Stream: THugeMemoryStream;
begin
  if Grid.Selection.Top >= Grid.Selection.Bottom then begin
    MessageDlg('The button takes the difference between the first and last selected line', mtInformation, [mbOK], 0);
    Exit;
  end;
  if CF_Difference = 0 then
    CF_Difference := RegisterClipboardFormat('Difference');
  with Grid.Selection do
    d := Abs(Grid[Bottom,10] - Grid[Top,10]);
  Difference.Caption := dtosm(d, 12, 4);
  Stream := THugeMemoryStream.Create;
  Clipboard.Open;
  try
    Clipboard.AsText := FloatToStr(d);
    Stream.Write(d, Sizeof(Double));
    Stream.CopyToClipboard(CF_Difference);
  finally
    Clipboard.Close;
    Stream.Free;
  end;
  BroadcastName('WM_New_Difference');
end;

procedure TLineForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  inherited;
  if Change = mnCacheCleared then
    ClearTLines;
end;

procedure TLineForm.StrengthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByIntensity, 0, nLines-1);
end;

procedure TLineForm.UpperStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByUpperState, 0, nLines-1);
end;

procedure TLineForm.BranchItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByBranch, 0, nLines-1);
end;

procedure TLineForm.FrequencySortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByFrequency, 0, nLines-1);
end;

procedure TLineForm.LowerStateSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByLowerState, 0, nLines-1);
end;

procedure TLineForm.WidthSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWidth, 0, nLines-1);
end;

procedure TLineForm.WriteExtraVariables(AFormSettings: TFormSettings);
var
  Index, i, P, l: Integer;
  Strings: TStringList;
  Line: TLineW;
  Buffer, Item: string;
  QuantumNumberFormat: TQuantumNumberFormat;
  Qno: TQuantumNumber;
begin
  inherited;
  HandleControls(LineFrmOptDlg, {$IFDEF FPC}@{$ENDIF}SetValue, AFormSettings);
  UnLoadOptionsDialog;
  Index := AFormSettings.IndexOfSetting('Lines');
  if Index > 0 then begin
    ClearButtonClick(nil);
    Strings := TStringList.Create;
    Strings.Text := AFormSettings.SettingToStr(Index);
    { Determine quantum number format used for writing }
    QuantumNumberFormat := qftJ;
    Index := AFormSettings.IndexOfSetting('QuantumNumberFormat');
    if Index > 0 then begin
      Buffer := AFormSettings.SettingToStr(Index);
      for Index := 0 to High(QuantumNumberFormatNames) - 1 do
        if CompareText(Buffer, QuantumNumberFormatNames[Index]) = 0 then
          QuantumNumberFormat := TQuantumNumberFormat(Index);
    end;
    Line.Line.EigenvaluesU := nil;
    Line.Line.EigenvaluesL := nil;
    Line.Line.Colour := clDefault;
    for l := 0 to Strings.Count-1 do begin
      Buffer := Strings[l];
      if Buffer <> '' then begin
        for i := 1 to 15 do begin
          P := Pos(#9, Buffer);
          if P > 0 then begin
            Item := Copy(Buffer, 1, P-1);
            Buffer := Copy(Buffer, P+1, Length(Buffer));
          end else
            Item := Buffer;
          if (i in [3,7]) and TryStrToQuantumNumber(Item, Qno, QuantumNumberFormat) then
            { Correct quantum numbers }
            Item := IntToStr(Qno);
          SetLineByIndex(Line, Item, i);
        end;
        if nLines >= Length(Lines) then
          SetLength(Lines, nLines+50);
        Lines[nLines] := Line;
        Inc(nLines);
      end;
    end;
    Strings.Free;
  end;
  InternalAddLines(0,nLines-1);
end;

procedure TLineForm.WriteSettings(AFormSettings: TFormSettings; DoPositions: Boolean; DoSettings: Boolean);
begin
  LoadOptionsDialog;
  inherited;
end;

procedure TLineForm.WeightSortItemClick(Sender: TObject);
begin
  SortLines({$IFDEF FPC}@{$ENDIF}ByWeight, 0, nLines-1);
end;

procedure TLineForm.MoreButtonClick(Sender: TObject);
begin
  SortPopupMenu.PopupComponent := MoreButton;
  SortPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TLineForm.FormDestroy(Sender: TObject);
begin
  LineForm := nil;
end;

procedure TLineForm.FitButtonClick(Sender: TObject);
var
  ObsType: TObsType;
begin
  if not NotEmpty then Exit;
  MainForm.LastObsWasLineForm := True;
  MainForm.ClearAllPeaksItemClick(nil);
  if Sender <> TestFitButton then begin
    ReShow(LogForm);
    ReShow(ResidualForm);
  end;
  case (Sender as TComponent).Tag of
    1: ObsType := obCDUpper;
    2: ObsType := obCDLower;
  else
    ObsType := obLine;
  end;
  MainForm.DoLineFit({$IFDEF FPC}@{$ENDIF}LoadObservations,'',ObsType,Sender = TestFitButton);
end;

procedure TLineForm.ClearTLines;
var
  i: Integer;
begin
  for i := 0 to High(Lines) do
    with Lines[i] do begin
      Line.EigenvaluesU := nil;
      Line.EigenvaluesL := nil;
    end;
end;

procedure TLineForm.HandleControl(Control: TControl; var PName, Value,
  PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean);
begin
  inherited;
  if (Control = LineFrmOptDlg.MinIntensity)
     or (Control = LineFrmOptDlg.MinAssignedIntensity) then
    SettingType := stDouble;
end;

function TLineForm.HasObservations: Boolean;
var
  i: Integer;
begin
  Result := False;
  if not NotEmpty then Exit;
  for i := 0 to Grid.RowCount-2 do
    if not (IsNan(Lines[i].Weight)) and (Lines[i].Weight > 0) then begin
      Result := True;
      Exit;
    end;
end;

procedure TLineForm.LoadObservations(const Source: string; ObsType: TObsType; TakeWeight: TTakeWeight);
var
  s: THugeMemoryStream;
  nObs: Integer;
begin
  s := THugeMemoryStream.Create;
  try
    CopyToStream(s, 1, Grid.RowCount-1, True, llStandard);
    s.Position := 0;
    Obs.LoadObservations(GlobalMixture, s, '<LineList>');
    if ObsType <> obLine then begin
      nObs := Length(Observations);
      AutoCommonDifferenceRange(ObsType = obCDUpper, True, 0, nObs, -1);
      SetLength(Observations, nObs);
    end;
  finally
    s.Free;
  end;
end;

procedure TLineForm.LoadOptionsDialog;
begin
  with LineFrmOptDlg do begin
    ShowLines.Value := FnLinesToShow;
    MinIntensity.Text := FloatToStr(FMinIntensity);
    DiscardZeroIntensity.Checked := not FShowZeroIntensityLines;
    MinAssignedIntensity.Text := FloatToStr(FMinAssignedIntensity);
  end;
end;

procedure TLineForm.UndoFitButtonClick(Sender: TObject);
begin
  if GlobalMixture <> nil then
    GlobalMixture.UndoFit;
end;

procedure TLineForm.UnLoadOptionsDialog;
begin
  with LineFrmOptDlg do begin
    if not TryStrToFloatEx(MinIntensity.Text, FMinIntensity) then
      raise EReportException.Create('Bad number:'+MinIntensity.Text)
    else if not TryStrToFloatEx(MinAssignedIntensity.Text, FMinAssignedIntensity) then
      raise EReportException.Create('Bad number:'+MinAssignedIntensity.Text)
    else begin
      FnLinesToShow := ShowLines.Value;
      FShowZeroIntensityLines := not DiscardZeroIntensity.Checked;
    end;
  end;
end;

procedure TLineForm.UpdateActions;
begin
  inherited UpdateActions;
  if FLongestNameLength > Grid.ColWidths[Grid.ColCount-1] then
    Grid.ColWidths[Grid.ColCount-1] := FLongestNameLength;
  FLongestNameLength := 0;
  UndoFitButton.Enabled := (GlobalMixture <> nil) and GlobalMixture.CanUndoFit
end;

procedure TLineForm.GridGetCellLabel(Sender: TObject; Col, Row: Integer;
  var s: string; var IsText: Boolean);
begin
  if (Row > 0) and (Col > 0) then
    IsText := StringRep(Row, Col, s);
end;

end.
