unit DRintensity;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses QuantumNumber;

type
  TExcitation = record
    Rank:       TQuantumNumber; { Rank of Excitation }
    Projection: TQuantumNumber; { Component of Excitation }
    DeltaJ:     TQuantumNumber; { Change in J }
    Simple:     Boolean;        { Simple equilibrium preparation }
    Circular:   Boolean;        { Circulary polarised light }
    { TODO: crossed polarizations }
  end;

const
  RandomExcitation: TExcitation = (Rank:0; Projection:0; DeltaJ:0; Simple:True; Circular:False);

function IntensityCorrection(JFinal, JMiddle: TQuantumNumber; const Excitation: TExcitation): Double;

implementation

uses Math, MathSup, nj, Mixture;

function IntensityCorrection(JFinal, JMiddle: TQuantumNumber; const Excitation: TExcitation): Double;
var
  M, JInitial, JMax: TQuantumNumber;
  t: Double;
begin
  Result := 1;
  if Excitation.Simple then
    Exit;
  Result := 0;
  M := SmallestProjection(JFinal);
  JInitial := JMiddle - Excitation.DeltaJ;
  JMax := Min(JFinal, Min(JMiddle, JInitial));
  { TODO: This assumes both Projection's = 0 }
  if Excitation.Projection <> 0 then
    raise NotYet;
  while M <= JMax do begin
    t := (JFinal+1)*(JInitial+1)*Sqr(JMiddle+One)
         *Sqr(threej(JMiddle, -M, Excitation.Rank, 0, JInitial, M))
         *Sqr(threej(JMiddle, -M, Excitation.Rank, 0, JFinal, M));
    if M = 0 then
      Result := Result + t
    else
      Result := Result + 2*t;
    Inc(M, 2);
  end;
end;

end.
