unit InternalCoordsFrm;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus, ExtCtrls, Grids, NGrid,
  Split, MNode, Mixture, zMatrix, InternalCoords, VibGridFrm, GridEx,
  GraphM, NewSpin, Buttons, ComCtrls, StdCtrls, AlEdit, EditExtr, ScrlCtrl,
  GraphI, Graphs;

type

  { TInternalCoordsForm }

  TInternalCoordsForm = class(TVibGridForm)
    CheckItem: TMenuItem;
    ShowVIBCAfileItem: TMenuItem;
    ShowFMatrixItem: TMenuItem;
    MakeFMatrixItem: TMenuItem;
    SortFMatrixbyValueItem: TMenuItem;
    SortFMatrixbyModeItem: TMenuItem;
    PrintItem: TMenuItem;
    CalculatefmatrixfromFmatrixItem: TMenuItem;
    SetfmatrixfromFmatrixItem: TMenuItem;
    MakeExtraOperatorsItem: TMenuItem;
    procedure GridGetEditStyle(Sender: TObject; ACol, ARow: Integer; var Style: TEditStyle);
    procedure GridGetPickListItems(ACol, ARow: Integer; Items: TStrings);
    procedure MagnificationSpinChange(Sender: TObject);
    procedure FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure CheckItemClick(Sender: TObject);
    procedure ShowVIBCAfileItemClick(Sender: TObject);
    procedure ShowFMatrixItemClick(Sender: TObject);
    procedure SortFMatrixbyValueItemClick(Sender: TObject);
    procedure MakeFMatrixItemClick(Sender: TObject);
    procedure SortFMatrixbyModeItemClick(Sender: TObject);
    procedure PrintItemClick(Sender: TObject);
    procedure CalculatefmatrixfromFmatrixItemClick(Sender: TObject);
    procedure SetfmatrixfromFmatrixItemClick(Sender: TObject);
    procedure MakeExtraOperatorsItemClick(Sender: TObject);
  private
    IC: TInternalCoords;
  protected
    procedure SetActiveObject(const Value: TMNode); override;
  public
    procedure PopulateForm(ACol:Integer); override;
    procedure Replot(ACol:Integer = -1); override;
    procedure ExtraPlot(zMatrix: TzMatrix; ACol, ix, iy: Integer; Original: Boolean); override;
  end;

function ConnectInternalCoordsForm(InternalCoords: TInternalCoords): TComponent;

implementation

uses Global, ChildForm, Vib, VibPlotFrm, ImportVIBCA, Worker, MNodeExpr, SymmetryCoordsFrm;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectInternalCoordsForm(InternalCoords: TInternalCoords): TComponent;
begin
  Result := ConnectVibFrm(InternalCoords, TInternalCoordsForm);
end;

{ TInternalCoordsForm }

procedure TInternalCoordsForm.FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
begin
  if (Col = 0) and (Row = 0) then
    s := 'f'
  else
    inherited;
end;

procedure TInternalCoordsForm.GridGetEditStyle(Sender: TObject; ACol, ARow: Integer; var Style: TEditStyle);
begin
  if ACol = 1 then
    Style := esPickList
  else
    Style := esSimple;
end;

procedure TInternalCoordsForm.GridGetPickListItems(ACol, ARow: Integer; Items: TStrings);
begin
  if Items.Count = 0 then begin
    Items.Add('Stretch');
    Items.Add('Bend');
    Items.Add('Plane');
    Items.Add('BondPlane');
    Items.Add('Torsion');
    Items.Add('');
  end;
end;

procedure TInternalCoordsForm.MagnificationSpinChange(Sender: TObject);
begin
  PopulateForm;
end;

procedure TInternalCoordsForm.ShowFMatrixItemClick(Sender: TObject);
begin
  ReShowLogForm;
  IC.Prepare(True);
  IC.CalculateF(True, False);
end;

procedure TInternalCoordsForm.MakeFMatrixItemClick(Sender: TObject);
begin
  IC.Prepare(True);
  IC.CalculateF(True, True);
  PopulateForm;
end;

procedure TInternalCoordsForm.ShowVIBCAfileItemClick(Sender: TObject);
begin
  ReShowLogForm;
  WriteVIBCAFile(Log, IC);
end;

procedure TInternalCoordsForm.SetfmatrixfromFmatrixItemClick(Sender: TObject);
begin
  ReShowLogForm;
  IC.CalculateFfromF(True, True);
  PopulateForm;
end;

procedure TInternalCoordsForm.SortFMatrixbyModeItemClick(Sender: TObject);
begin
  IC.SortF(False);
  PopulateForm;
end;

procedure TInternalCoordsForm.SortFMatrixbyValueItemClick(Sender: TObject);
begin
  IC.SortF(True);
  PopulateForm;
end;

procedure TInternalCoordsForm.PopulateForm(ACol:Integer);
var
  r: Integer;
begin
  if csLoading in ComponentState then
    Exit;
  Grid.OnGetEditStyle := {$IFDEF FPC}@{$ENDIF}GridGetEditStyle;
  Grid.OnGetPickListItems := {$IFDEF FPC}@{$ENDIF}GridGetPickListItems;
  Caption := ActiveObject.Name + ' for ' + (ActiveObject.Owner as TState).UniqueName;
  for r := 1 to Grid.RowCount-2 do
    if (r > Length(IC.Brows)) or (IC.Brows[r-1] < 0) then
      Grid.Lbl[r] := ''
    else
      Grid.Lbl[r] := IC.nameR(IC.Brows[r-1]);
  Grid.Lbl[Grid.RowCount-1] := '';
  inherited;
  Grid.AutoAdjustColumn(1,False,True);
  for r := 6 to Grid.ColCount-1 do
    Grid.Title[r] := ' '; { Room for a comment... }
  Replot;
end;

procedure TInternalCoordsForm.Replot(ACol: Integer);
begin
  if ACol < 0 then
    ACol := Grid.Row;
  if (ACol > 0) and (ACol <= Length(IC.Brows)) then
    ACol := IC.Brows[ACol-1]
  else
    ACol := -1;
  IC.MakeBMatrix(IC.ElectronicState,False); { TODO: may cause too many calls to MakeBMatrix }
  if IC.LastError = '' then
    IC.MakeFMatrix(IC.nInternal, 'f', False, {$IFDEF FPC}@{$ENDIF}IC.nameR);
  PlotZmatrix(IC.zMatrix, IC.LastError+IC.LastWarning, ACol, IC.B, MagnificationSpin.Value, not IC.InPrincipalAxis, -1)
end;

procedure TInternalCoordsForm.PrintItemClick(Sender: TObject);
begin
  ReShowLogForm;
  IC.Prepare(True);
  if IC.LastError <> '' then
    WriteLn(Log,'**** ',IC.LastError);
  PopulateForm;
  IC.ElectronicState.FixupProperties.Free; { Above overwrites l matrix ignoring Symmetry Coordinates, so force full update }
end;

procedure TInternalCoordsForm.MakeExtraOperatorsItemClick(Sender: TObject);
var
  S: TSymmetryCoords;
  i: Integer;
begin
  with IC.Owner do
    for i := 0 to ComponentCount - 1 do
      if (Components[i] is TSymmetryCoords) and TSymmetryCoords(Components[i]).ReallyActive then begin
        S := TSymmetryCoords(Components[i]);
      end;
  if S <> nil then
    case MessageDlg('Symmetry coordinates present - calculate using these?', mtWarning, mbYesNoCancel, 0) of
      mrYes: begin
               (ConnectSymmetryCoordsForm(S) as TSymmetryCoordsForm).MakeExtraOperatorsItemClick(nil);
               Exit;
             end;
      mrCancel: Exit;
    end;
  BeginGlobalUpdate;
  try
    IC.Makel(False, False, True);
  finally
    EndGlobalUpdate;
  end;
  if IC.LastError <> '' then
    ShowInfo(IC.LastError);
end;

procedure TInternalCoordsForm.CalculatefmatrixfromFmatrixItemClick( Sender: TObject);
begin
  ReShowLogForm;
  IC.CalculateFfromF(True, False);
end;

procedure TInternalCoordsForm.CheckItemClick(Sender: TObject);
begin
  PopulateForm;
end;

procedure TInternalCoordsForm.ExtraPlot(zMatrix: TzMatrix; ACol, ix, iy: Integer; Original: Boolean);
  procedure JoinP(n1, n2: Integer);
  var
    p1, p2: Txyz;
  begin
    with zMatrix do
      if Original then begin
        p1 := Atoms[n1].xyz;
        p2 := Atoms[n2].xyz;
      end else begin
        p1 := Atoms[n1].pxyz;
        p2 := Atoms[n2].pxyz;
      end;
    Rotate(p1, ix);
    Rotate(p2, ix);
    G.Style(1);
    G.Point(p1[ix], p1[iy]);
    G.Join(p2[ix], p2[iy]);
    G.Style(0);
  end;
  procedure Join(n1, n2: Integer);
  begin
    n1 := IC.ModeAtoms[ACol-1][n1-1];
    n2 := IC.ModeAtoms[ACol-1][n2-1];
    JoinP(n1,n2);
  end;
  procedure PP(n: Integer);
  var
    p: TXyz;
  begin
    n := IC.ModeAtoms[ACol-1][n-1];
    with zMatrix do
      if Original then
        p := Atoms[n].xyz
      else
        p := Atoms[n].pxyz;
    Rotate(p, ix);
    G.PolyPoint(p[ix], p[iy]);
  end;
var
  i, nA: Integer;
begin
  inherited ExtraPlot(zMatrix, ACol, ix, iy, Original);
  with IC do begin
    if (ACol > 0) and (ACol <= Length(IC.ModeType)) then
      case IC.ModeType[ACol-1] of
        icStretch:    Join(1,2);
        icBend:       begin Join(1,2); Join(2,3); end;
        icLinBend:    begin Join(1,2); Join(2,3); Join(1,4); Join(2,4); Join(3,4); { TODO: Plot 1234 plane }  end;
        icPlaneSym:   begin Join(1,2); Join(1,3); Join(1,4); end;
        icPlaneGen:   begin Join(1,2); G.Style(1); G.PolyStart; PP(1); PP(3); PP(4); PP(1); G.PolyEnd; G.Style(0); { Was: Join(1,3); Join(1,4); } end;
        icTorsion:    begin Join(1,2); Join(2,3); Join(3,4); end;
        icGenTorsion: begin
                        Join(1,2);
                        for i := 2 to High(IC.ModeAtoms[ACol-1]) do begin
                          nA := IC.ModeAtoms[ACol-1][i];
                          if nA < 0 then
                            JoinP(IC.ModeAtoms[ACol-1][0], -nA-1)
                          else
                            JoinP(IC.ModeAtoms[ACol-1][1], nA);
                        end;
                      end;
        icError:      { Skip } ;
      end;
  end;
end;

procedure TInternalCoordsForm.SetActiveObject(const Value: TMNode);
begin
  inherited;
  IC := Value as TInternalCoords;
  RowSelect := True;
end;

end.
