unit LevelPlotter;
{$I utils.inc}{Copyright (C) 2005-9 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE}LevelFrm,{$ENDIF}
     Classes, GraphM, MNode, Mixture, Global, pgDOM, Gmarks, LevelLister;

type

  { TLevelPlotter }

  TLevelPlotter = class({$IFNDEF CONSOLE}TLevelForm{$ELSE}TMNode{$ENDIF})
  private
    FirstPoint: Boolean;
    LogHeaderShown: Boolean;
    SymmetryIndex, ASelectIndex, BSelectIndex, IgnoreSpinsIndex,
    XQuantumNumberIndex, LabelsStyleIndex: Integer;
    DoneCreating: Boolean;
    {$IFDEF CONSOLE}
    PlottedPoints: TPlottedPoints;
    nPlottedPoints: Integer;
    FActiveObject: string;
    FSymmetryList: TStringList;
    CurrentMNode: TMNode;
    procedure SetActiveObject(const AValue: string);
    {$ENDIF}
    procedure CheckIndices;
  public
    EMin, EMax: Double;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure PlotManifold(G: TGraphM; LevelLister: TLevelLister; Manifold: TManifold; State: TState; ToLog, FieldSummary: Boolean);
    procedure PlotItem(G: TGraphM; MNode:TMNode; Log, FieldSummary: Boolean);
    function GetSetting(AIndex: Integer): Integer; {$IFNDEF CONSOLE} override; {$ENDIF}
    procedure SetSetting(AIndex: Integer; const AValue: Integer); {$IFNDEF CONSOLE} override; {$ENDIF}
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; override;
    {$IFDEF CONSOLE}
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    property ActiveObject: string read FActiveObject write SetActiveObject;
    {$ENDIF}
  published
    { Note that if the form is edited, the settings numbers must be updated to reflect the creation order on the form }
//    property Left: Integer index 1 read GetSetting write SetSetting; { 0 }
//    property Top: Integer index 2 read GetSetting write SetSetting; { 0 }
//    property Width: Integer index 3 read GetSetting write SetSetting; { 0 }
//    property Height: Integer index 4 read GetSetting write SetSetting; { 0 }
//    property Visible: Integer index 5 read GetFlag write SetSetting; { False }
//    property ActiveObject: Integer index 6 read GetSetting write SetSetting; {  }
    property CalcBoth: Boolean       index 7 read GetFlag write SetFlag; { True }
    property CalcLine: Boolean       index 8 read GetFlag write SetFlag; { False }
    property CalcPoints: Boolean     index 9 read GetFlag write SetFlag; { False }
    property JoinHighest: Boolean    index 10 read GetFlag write SetFlag; { False }
    property TrackQuanta: Boolean    index 11 read GetFlag write SetFlag; { True }
    property TrackState: Boolean     index 12 read GetFlag write SetFlag; { True }
    property ExternalField: Boolean  index 13 read GetFlag write SetFlag; { False }
    property AltPlot: Boolean        index 14 read GetFlag write SetFlag; { False }
    property Symmetry: Integer       index 15 read GetSetting write SetSetting; { all }
    property LimitJ: Boolean         index 16 read GetFlag write SetFlag; { True }
    property ASelect: Integer        index 17 read GetSetting write SetSetting; { all }
    property BSelect: Integer        index 18 read GetSetting write SetSetting; { all }
    property XMax: Integer           index 19 read GetSetting write SetSetting; { 40 }
    property XMin: Integer           index 20 read GetSetting write SetSetting; { 0 }
    property XQuantumNumber: Integer index 21 read GetSetting write SetSetting; { J }
    property JoinObs: Boolean        index 22 read GetFlag write SetFlag; { True }
    property ObsLower: Boolean       index 23 read GetFlag write SetFlag; { False }
    property ObsUpper: Boolean       index 24 read GetFlag write SetFlag; { False }
    property IgnoreSpins: Integer    index 25 read GetSetting write SetSetting; { all }
    property ExcludePerturbations: Boolean index 26 read GetFlag write SetFlag; { False }
    property LabelsStyle: Integer    index 27 read GetSetting write SetSetting; {  }
    property FieldSamples: Integer   index 28 read GetSetting write SetSetting; { 20 }
    property ESweep: Boolean         index 29 read GetFlag write SetFlag; { True }
    property BSweep: Boolean         index 30 read GetFlag write SetFlag; { False }
    property Ymax: Double            index 1 read GetParameter write SetParameter; {  1.00000000000000E+0004 }
    property Ymin: Double            index 2 read GetParameter write SetParameter; {  0.00000000000000E+0000 }
    property Reduction: Double       index 3 read GetParameter write SetParameter; {  0.00000000000000E+0000 }
    property ErrorMultiplier: Double index 4 read GetParameter write SetParameter; {  1.00000000000000E+0000 }
    property MinField: Double        index 5 read GetParameter write SetParameter; {  0.00000000000000E+0000 }
    property MaxField: Double        index 6 read GetParameter write SetParameter; {  1.00000000000000E+0004 }
    property ShiftMult: Double       index 7 read GetParameter write SetParameter; {  1.00000000000000E+0000 }
  end;

  TLabelPos = (lbNone, lbLeftOut, lbLeftIn, lbRightIn, lbRightOut, lbStart, lbEnd);

function ReadLevelPlotter(Mixture: TMixture):TLevelPlotter;

const
  LabelsStyleNames: array [0..7] of PChar = ('None', 'LeftOut', 'LeftIn', 'RightIn', 'RightOut', 'Start', 'End', nil);

implementation

uses SysUtils, Math, MathSup, CODATA, Graphics, GStyle, QuantumNumber, QuantumMath, MNodeColour,
     Obs, Worker, Progress, SymUtils, StaticField, UCompat, LapackLSQ, Units, FuncFit;

function ReadLevelPlotter(Mixture: TMixture):TLevelPlotter;
var
  Node: TDOMElement;
begin
  {$IFDEF CONSOLE}
    Result := TLevelPlotter.Create(Mixture);
    { Read TLevelPlotter object saved as a FormSettings object }
    if Assigned(Mixture.LoadedFormSettings) then begin
      Node := Mixture.LoadedFormSettings.FirstChild as TDOMElement;
      while Assigned(Node) do begin
        if ( CompareText(Node.NodeName,'FormSettings') = 0 )
           and ( CompareText(Node.GetAttribute('Name'),'LevelWindow') = 0) then begin
             Result.ReadProperties(Node,False);
             Break;
           end;
        Node := Node.NextSibling as TDOMElement;
      end;
    end;
  {$ELSE}
    Result := LevelForm as TLevelPlotter;
  {$ENDIF}
end;

{ TLevelPlotter }

{$IFDEF CONSOLE}
function TLevelPlotter.ReadProperties(Node: TDOMNode; CheckUnread: Boolean): Integer;
begin
  FActiveObject := ReadValue(Node,'ActiveObject','');
  Result:=inherited ReadProperties(Node, CheckUnread);
end;

procedure TLevelPlotter.SetActiveObject(const AValue: string);
begin
  FreeAndNil(FSymmetryList);
  FActiveObject := AValue;
  CurrentMNode := nil; { TODO: Fix }
  FSymmetryList := MakeSymmetryList(CurrentMNode);
end;
{$ENDIF}

constructor TLevelPlotter.Create(AOwner: TComponent);
begin
  {$IFDEF CONSOLE}
  SavedProperties := SavedProperties
    + ',CalcBoth=True,TrackQuanta=True,TrackState=True,ESweep=True,BSweep=False,'
    + ',ASelect=all,BSelect=all,Symmetry=all,LimitJ=True,XMax=40,XQuantumNumber=J,JoinObs=True,IgnoreSpins=all'
    + ',FieldSamples=20,Ymax=1e4,ErrorMultiplier=1,MaxField=1e4,ShiftMult=1';
  {$ENDIF}
  inherited Create(AOwner);
  CheckIndices;
  DoneCreating := True;
end;

destructor TLevelPlotter.Destroy;
begin
  {$IFDEF CONSOLE}
  FreeAndNil(FSymmetryList);
  {$ENDIF}
  inherited Destroy;
end;

function TwoLevelFunc(x: Double; Derivative: Integer; const Parameters: array of Double):Double;
begin
  Result := Sqrt(Sqr(Parameters[1]/2) + Sqr(x*Parameters[2]));
  if Parameters[2] < 0 then
    Result := -Result;
  case Derivative of
    0: Result := Result + Parameters[0];
    1: Result := 1;
    { Note care required in derivatives below to return correct sign }
    2: Result := Parameters[1]/4/Result;
    3: Result := Parameters[2]*Sqr(x)/Result;
  end;
end;

type
  TStateLabel = record
    x, y: Double;
    s: string;
    Alignment: TAlignment;
  end;

procedure TLevelPlotter.PlotManifold(G: TGraphM; LevelLister: TLevelLister; Manifold: TManifold; State: TState; ToLog, FieldSummary: Boolean);
var
  J: TQuantumNumber;
  ASelecting, BSelecting, IncludePerturbations: Boolean;
  MarkStyle: TMarkStyle;
  ShowPoints, ShowLines: Boolean;
  FirstPointOnCurve, i, nErrors: Integer;
  Matches: array of Integer;
  LastColour: TColor;
  TIndex, MaxTIndex: Integer;
  LastName: string;
  Simulation: TSimulation;
  Overflow: Exception;
  StateLabels: array of TStateLabel;
  nStateLabels: Integer;

  procedure AddStateLabel(Ax, Ay: Double; AAlignment: TAlignment; Name: string);
  begin
    Inc(nStateLabels);
    SetLength(StateLabels, nStateLabels);
    with StateLabels[nStateLabels-1] do begin
      x := Ax;
      y := Ay;
      s := Name;
      Alignment := AAlignment;
    end;
  end;

  procedure ShowStateLabels;
  var
    i: Integer;
  begin
    G.StyleP(clWhite,0,lpSolid);
    for i := 0 to High(StateLabels) do
      with StateLabels[i] do begin
        case Alignment of
          taLeftJustify:  G.Rectangle(x, x+G.Cx*Length(s), y, y+G.Cy);
          taRightJustify: G.Rectangle(x-G.Cx*Length(s), x, y, y+G.Cy);
        end;
        G.TextJ(x, y, Alignment, s);
      end;
  end;

  function XQno(Ev: TEigenValues; EvNo: Integer):TQuantumNumber;
  begin
    Result := Ev.Qno(EvNo,TQnoType(XQuantumNumber div 2));
  end;

  function AdjustY(y: Double; J: TQuantumNumber): Double;
  begin
    Result := y - Reduction*J*(J+2)/4;
  end;

  function IsSelected(Ev: TEigenValues; EvNo: Integer):Boolean;
  begin
    Result := False;
    if (Symmetry >= 0) and (Symmetry <> Ev.Manifold.Molecule.JDeAdjustedSym(Ev.DominantBasis[EvNo].F,Ev.DominantBasis[EvNo].Sym)) then
      Exit;
    if ASelecting and (Ev.DominantBasis[EvNo].Qno(qtA) <> ASelect) then
      Exit;
    if BSelecting and (Ev.DominantBasis[EvNo].Qno(qtB) <> BSelect) then
      Exit;
    Result := True;
  end;

  procedure PlotPoint(Ev: TEigenValues; EvNo: Integer; AX, E: Double); overload;
  begin
    if Manifold.nStates > 0 then begin
      if LastName = '' then
        LastName := Ev.DominantBasis[EvNo].State.Name;
      LastColour := Ev.DominantBasis[EvNo].State.FindColour;
    end else begin
      if LastName = '' then
        LastName := Manifold.Name;
      LastColour := Manifold.FindColour;
    end;
    if LastColour = clHide then
      Exit
    else if LastColour = clNone then
      LastColour := clBlack;
    if FirstPoint then begin
      Emin := E;
      EMax := E;
      FirstPoint := False;
    end else begin
      if E > EMax then EMax := E;
      if E < EMin then EMin := E;
    end;
    if Length(PlottedPoints) <= nPlottedPoints then
      SetLength(PlottedPoints, nPlottedPoints+10000);
    with PlottedPoints[nPlottedPoints] do begin
      X := AX;
      Y := E;
      Inc(nPlottedPoints);
      Index := -EvNo;
      Extra := PtrInt(Ev);
    end;
  end;

  procedure PlotPoint(Ev: TEigenValues; EvNo: Integer); overload;
  var
    QX: TQuantumNumber;
    E: Double;
  begin
    QX := XQno(Ev,EvNo);
    E := AdjustY(Ev.Energy[EvNo], QX); { TODO: Make this a choice? Last release had Ev.Qno(EvNo,qtJ) }
    if IsSelected(Ev,EvNo) then
      PlotPoint(Ev,EvNo,MapQnoExpr(XQuantumNumber,QX),E);
  end;

  procedure PlotPoints(From: Integer); { Plot buffered points }
  var
    i: Integer;
    Ev: TEigenValues;
  begin
    if ShowPoints then
      for i := From to nPlottedPoints-1 do
        with PlottedPoints[i] do begin
          Ev := TEigenValues(Extra);
          if Manifold.nStates > 0 then begin
            LastColour := Ev.DominantBasis[-Index].State.FindColour;
          end else
            LastColour := Manifold.FindColour;
          if LastColour = clNone then
            LastColour := clBlack;
          MarkStyle.Colour := LastColour;
          if AltPlot then
            MarkStyle.MarkType := MarkTypeFor(Ev.Sym)
          else
            MarkStyle.MarkType := MarkTypeFor(Ev.JDeAdjustedSym(-Index));
          G.SymbolT(X, Y, MarkStyle);
        end;
  end;

  procedure PlotCurve(Lbl: string; LabelPos:TLabelPos);
  var
    LabelFound: Boolean;
    LX, LY: Double;
    i: Integer;
  begin
    if ShowLines and (nPlottedPoints-FirstPointOnCurve > 1) then begin
      { Record first visible point for LabelStyle = Start }
      LabelFound := not (LabelPos in [lbStart,lbEnd]);
      G.StyleP(LastColour, G.LineStyle.Width, G.LineStyle.Pattern);
      for i := FirstPointOnCurve to nPlottedPoints-1 do begin
        if i = FirstPointOnCurve then
          G.ClipPoint(PlottedPoints[i].X, PlottedPoints[i].Y)
        else
          G.ClipJoin(PlottedPoints[i].X, PlottedPoints[i].Y);
        if (not LabelFound or (LabelPos = lbEnd)) and (G.OutCode(PlottedPoints[i].X, PlottedPoints[i].Y) = 0) then begin
          LabelFound := True;
          LX := PlottedPoints[i].X-G.cx/6;
          LY := PlottedPoints[i].Y-G.cy/3;
        end;
      end;
      if LabelFound and (LabelPos in [lbStart,lbEnd]) then
        AddStateLabel(LX, LY, taRightJustify, Lbl);
    end;
  end;

  function PlotObsPoint(Manifoldo: TManifold; Jo: TQuantumNumber; Symo: Shortint; Indexo: Integer;
                         Manifoldc: TManifold; Jc: TQuantumNumber; Symc: Shortint; Indexc: Integer;
                         E: Double; Number: Integer; Upper: Boolean):Boolean;
  var
    ECalc, XX: Double;
    XQ: TQuantumNumber;
    Ev: TEigenvalues;
    OtherColour: TColor;
  begin
    Result := nErrors > 50;
    XQ := Jo;
    if (Manifoldo = Manifold) and (Jo >= XMin) and (Jo <= XMax) then
      with Observations[Number] do begin
        LastColour := Colour;
        { TODO -cField: Add field search here }
        if Assigned(Manifoldc) and not (ObsType in [obELower, obEUpper]) then begin
          { Find "other" end of transition }
          Indexc := FindEigenvaluesAndIndex(Ev, Manifoldc, Jc, Symc, Indexc, Name,
                                            Upper, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField);
          if Indexc <= 0 then begin
            Inc(nErrors);
            Exit;
          end;
          E := E + Ev.E[Indexc];
          OtherColour := Ev.DominantBasis[Indexc].State.FindColour;
        end else if (ManifoldU = nil) or (ObsType = obELower) then
          E := -E; { Lower level only }
        if AltPlot then
          MarkStyle.MarkType := MarkTypeFor(Symo)
        else
          MarkStyle.MarkType := MarkTypeFor(Manifoldo.Molecule.JDeAdjustedSym(Jo,Symo));
        if JoinObs then begin
          Indexo := FindEigenvaluesAndIndex(Ev, Manifoldo, Jo, Symo, Indexo, Name,
                                            not Upper, LineNumber, FileName, RemoveSpins, QuantumNumberFormat, ExField, IncludePerturbations);
          if Indexo <= 0 then
            Inc(nErrors)
          else begin
            if not IsSelected(Ev,Indexo) then
              Exit;
            if LastColour = clNone then
              LastColour := Ev.DominantBasis[Indexo].State.FindColour;
            XQ := XQno(Ev,Indexo);
            Ecalc := Ev.E[Indexo];
            if ErrorMultiplier > 0 then
              E := (E-Ecalc)*ErrorMultiplier + Ecalc
            else
              E := -(E-Ecalc)/Weight*ErrorMultiplier + Ecalc;
            Ecalc := Adjusty(Ecalc, XQ);
          end;
        end else
          Indexo := -1;
        if (LastColour = clNone) and (OtherColour <> clNone) then
          LastColour := OtherColour;
        if LastColour = clHide then
          Exit;
        if LastColour = clNone then
          LastColour := clBlack;
        MarkStyle.Colour := LastColour;
        E := Adjusty(E,XQ); { TODO: May not adjust on correct quantum number if Join Obs not set }
        if Length(PlottedPoints) <= nPlottedPoints then
          SetLength(PlottedPoints, nPlottedPoints+10000);
        XX := MapQnoExpr(XQuantumNumber,XQ);
        with PlottedPoints[nPlottedPoints] do begin
          X := XX;
          Y := E;
          Index := Number;
          Extra := Ord(Upper);
          G.SymbolT(X, E, MarkStyle);
        end;
        Inc(nPlottedPoints);
        if Indexo > 0 then begin
          G.StyleP(LastColour, G.LineStyle.Width, G.LineStyle.Pattern);
          G.ClipPoint(XX, E);
          G.ClipJoin(XX, Ecalc);
        end;
      end;
  end;

  procedure ShowOneState(Sym: Integer; MatchState: TState; XMax: TQuantumNumber);
  var
    EvNo, MaxEvNo, i, AdjSym: Integer;
    IndexL, IndexU, nMatch, IndexT: Integer; { Numbers filtered by State }
    Match: Boolean;
    Ev: TEigenValues;
  begin
    EvNo := 1;
    MaxEvNo := 0;
    LastName := '';
    repeat
      FirstPointOnCurve := nPlottedPoints; { No points plotted yet }
      if Simulation.HasFields then
        J := -Manifold.FindJmax(XMax,-IgnoreSpins)
      else
        J := Manifold.FindJmin(XMin,-IgnoreSpins);
      while J <= XMax do begin
        AdjSym := Manifold.Molecule.JAdjustedSym(J, Sym);
        Ev := Manifold.Eigenvalues(J, AdjSym, [], Simulation.ExField, True, plNone, State, IncludePerturbations, -IgnoreSpins);
        if Assigned(MatchState) or (TIndex >= 0) then begin
          { Find the number matching this state }
          nMatch := 0;
          for i := 1 to Ev.Size do begin
            Match := not Assigned(MatchState) or (Ev.DominantBasis[i].State = MatchState);
            if TIndex >= 0 then begin
              IndexT := Ev.TrackIndex(i);
              if IndexT > MaxTIndex then
                MaxTIndex := IndexT;
              Match := Match and (TIndex = IndexT);
            end;
            if Match then begin
              Inc(nMatch);
              if nMatch > Length(Matches) then
                SetLength(Matches, nMatch+20);
              Matches[nMatch-1] := i;
            end;
          end;
          if EvNo <= nMatch then begin
            IndexL := Matches[EvNo-1];
            IndexU := Matches[nMatch-EvNo];
          end;
        end else begin
          nMatch := Ev.Size;
          IndexL := EvNo;
          IndexU := Ev.Size-EvNo+1;
        end;
        if nMatch > MaxEvNo then
          MaxEvNo := nMatch;
        if nMatch >= EvNo then
          if JoinHighest then
            PlotPoint(Ev,IndexU)
          else
            PlotPoint(Ev,IndexL);
        Inc(J, 2);
      end;
      PlotCurve(LastName,TLabelPos(LabelsStyle));
      // PlotPoints(FirstPointOnCurve);
      Inc(EvNo);
    until EvNo > MaxEvNo;
  end;

  procedure PlotStateLabels;
  var
    i: Integer;
  begin
    for i := 0 to Manifold.nStates-1 do
      with Manifold.State[i] do
        if (Origin >= G.Ymin) and (Origin <= G.Ymax) then
          case TLabelPos(LabelsStyle) of
            lbLeftOut:  AddStateLabel(G.Xmin-G.cx/6, Origin-G.cy/3, taRightJustify, Name);
            lbLeftIn:   AddStateLabel(G.Xmin+G.cx*1.5, Origin, taLeftJustify,  Name);
            lbRightIn:  AddStateLabel(G.Xmax, Origin, taRightJustify, Name);
            lbRightOut: AddStateLabel(G.Xmax+G.Cx*3, Origin, taRightJustify, Name);
          end;
  end;

  procedure SweepField;
  var
    Fields, FieldsSq, Es: array of Double;

    procedure ShowTuningRates(iZeroField: Integer);
    var
      LinPoly, QuadPoly: array [0..1] of Double;
      TwoLevel: array [0..2] of Double;
      Emin, Emax, Conv, Diff, LinErr, QuadErr, TwoErr, Change, Err, E0, OrgDelta: Double;
      FieldUnits, DipoleUnits: string;
      Cycle: Integer;
    begin
      if ESweep then begin
        FieldUnits := '(Vm-1)';
        Conv := ConversionFactor(MHz,Manifold.Mixture.Units)*1e-27/SpeedOfLight/Planck;
        DipoleUnits := 'Debye';
      end else begin
        FieldUnits := 'T';
        Conv := ConversionFactor(MHz,Manifold.Mixture.Units) * 1e-6 * BohrMagneton / Planck;
        DipoleUnits := 'muB';
      end;
      Limits(Es,Emin,Emax);
      Diff := Emax-Emin;
      PolyFit(Fields,Es,LinPoly);
      LinErr := MaxPolyErr(Fields,Es,LinPoly);
      if ToLog then
        WriteLn(Log, 'Linear: ',LinPoly[1],' ',Manifold.Mixture.UnitsName,'/',FieldUnits,DtosMA(LinPoly[1]/Conv,10,-1),' ',DipoleUnits,' (',DtosM(LinErr/Diff*100,4,-1),'%)');
      PolyFit(FieldsSq,Es,QuadPoly);
      QuadErr := MaxPolyErr(FieldsSq,Es,QuadPoly);
      if ToLog then
        WriteLn(Log, 'Quadratic: ',QuadPoly[1],' ',Manifold.Mixture.UnitsName,'/',FieldUnits,'^2 (',DtosM(QuadErr/Diff*100,4,-1),'%)');
      { Two level fit }
      TwoLevel[0] := 0;
      TwoLevel[1] := Abs(Emax-Emin)/10;
      TwoLevel[2] := LinPoly[1];
      TwoLevel[0] := -TwoLevelFunc(0,0,TwoLevel);
      if LinErr < QuadErr then
        TwoLevel[0] := TwoLevel[0] + LinPoly[0]
      else
        TwoLevel[0] := TwoLevel[0] + QuadPoly[0];
      Cycle := 0;
      TwoErr := -1;
      repeat
        OrgDelta := TwoLevel[1];
        Change := FitOnce(@TwoLevelFunc,TwoLevel,Fields,Es);
        // Write(Log, 'Two Level=',DtosM(Change,4,-1));
        if TwoLevel[1] < 0 then
          if Cycle < 8 then
            TwoLevel[1] := OrgDelta/2
          else
            TwoLevel[1] := Abs(TwoLevel[1]);
        Err := FitMaxErr(@TwoLevelFunc,TwoLevel,Fields,Es);
        // Write(Log, ' Err=',DtosM(TwoErr/Diff*100,4,-1));
        //for i := 0 to High(TwoLevel) do
        //  Write(Log, ' ',TwoLevel[i]);
        // WriteLn(Log);
        if TwoLevel[1] > Diff*10 then
          Break; { Indicates badly conditioned fit - normally close to quadratic }
        if (Cycle > 2) and (Change < 1e-12) then
          TwoErr := Err;
        Inc(Cycle);
      until (Cycle > 20) or (TwoErr > 0);
      if FieldSummary then begin
        if LinErr < QuadErr then
          E0 := Es[iZeroField]-(Poly(Fields[iZeroField],LinPoly)-Poly(0,LinPoly))
        else
          E0 := Es[iZeroField]-(Poly(Fields[iZeroField],QuadPoly)-Poly(0,QuadPoly));
        if (TwoErr > 0) and (TwoErr < Min(LinErr, QuadErr)) then
          E0 := Es[iZeroField]-(TwoLevelFunc(Fields[iZeroField],0,TwoLevel)-TwoLevelFunc(0,0,TwoLevel));
        Write(Log, E0:14:Manifold.Precision);
        Write(Log, DtosMA(LinPoly[1],15,-1), DtosMA(LinPoly[1]/Conv,10,-1), DtosMA(LinErr/Diff*100,5,-1),'%');
        Write(Log, DtosMA(QuadPoly[1],15,-1), DtosMA(QuadErr/Diff*100,5,-1),'%');
        if TwoErr >= 0 then
          Write(Log, DtosMA(TwoLevel[0],10,-1), DtosMA(TwoLevel[1],10,-1), DtosMA(TwoLevel[2],15,-1), DtosMA(TwoLevel[2]/Conv,10,-1), DtosMA(TwoErr/Diff*100,5,-1), '%');
        WriteLn(Log);
      end;
      if (TwoErr > 0) and ToLog then
        WriteLn(Log, 'Two Level: ',DtosMA(TwoLevel[0],10,-1),' ',Manifold.Mixture.UnitsName, ' ', DtosMA(TwoLevel[1],10,-1),' ',Manifold.Mixture.UnitsName, ' ',
                                   TwoLevel[2],' ',Manifold.Mixture.UnitsName,'/',FieldUnits,' ',DtosMA(TwoLevel[2]/Conv,10,-1),' ',DipoleUnits,
                                   ' (',DtosM(TwoErr/Diff*100,4,-1),'%)');
      // WriteLn(Log);
    end;

  var
    i, j, iZeroField, iEff: Integer;
    Evs: array of TEigenvalues;
    ExFields: array of TStaticFields;
    M: TQuantumNumber;
    E, Mult: Double;
    Lbl: string;
    k: Integer;
  begin
    { Potentially memory intensive, as all eigenvalues at all fields cached }
    FirstPoint := True;
    Mult := ShiftMult;
    M := XMin;
    if SmallestProjection(M) <> Manifold.HalfSpin(-IgnoreSpins) then
      Inc(M);
    PushProgress('M',XMax div 2);
    try
      while M <= XMax do begin
        SetLength(Evs, FieldSamples);
        SetLength(Es, FieldSamples);
        SetLength(Fields, FieldSamples);
        SetLength(FieldsSq, FieldSamples);
        SetLength(ExFields, FieldSamples);
        SetProgress(M div 2);
        { Choose fields }
        iZeroField := -1;
        for i := 0 to High(Fields) do begin
          Fields[i] := MinField + (MaxField-MinField)*i/High(Fields);
          if (iZeroField < 0) or (Abs(Fields[i]) < Abs(Fields[iZeroField]))  then
            iZeroField := i;
        end;
        { Shift one of the points to close to zero, if zero included in plot range }
        if MaxField * MinField <= 0 then begin
          { Shift not quite to zero to make life easier for state tracking }
          Fields[iZeroField] := (MaxField-MinField)/High(Fields)/10000;
          if Fields[iZeroField] = 0 then
            Exit; { calculation at zero field breaks, as symmetry -1 is then invalid }
        end else begin
          iZeroField := 0; { No zero field point, so use LHS as reference }
          Mult := 1;
        end;
        for i := 0 to High(Fields) do begin
          ExFields[i] := Simulation.ExField;
          FieldsSq[i] := Sqr(Fields[i]);
          if ESweep then
            ExFields[i].E := Fields[i]
          else
            ExFields[i].B := Fields[i];
          { Start parallel calculation }
          Evs[i] := Manifold.Eigenvalues(M, -1, [], ExFields[i], False, plNone, State, IncludePerturbations, -IgnoreSpins);
        end;
        PushProgress('Field',High(Fields));
        try
          for i := 0 to High(Fields) do begin
            SetProgress(i);
            Evs[i].Make(True,[meHrot]);
          end;
        finally
          PopProgress;
        end;
        for i := 1 to Evs[0].Size do
          if IsSelected(Evs[Max(0,iZeroField)],i) then begin
            FirstPointOnCurve := nPlottedPoints;
            Lbl := Evs[Max(0,iZeroField)].DominantLabel(i,GlobalMixture.QuantumNumberFormat);
            { TODO: Match states with largest overlap at adjacent fields if TrackStates.Checked }
            for j := 0 to High(Fields) do begin
              iEff := i;
              if TrackQuanta then begin
                { Look for match in state label }
                if Lbl <> Evs[j].DominantLabel(i,GlobalMixture.QuantumNumberFormat) then
                  for k := 1 to Evs[0].Size do
                    if (j <> k) and (Lbl = Evs[j].DominantLabel(k,GlobalMixture.QuantumNumberFormat)) then begin
                      iEff := k;
                      Break;
                    end;
              end;
              E := Evs[j].Energy[iEff];
              Es[j] := E;
              if iZeroField >= 0 then
                E := (E - Evs[iZeroField].Energy[i])*Mult + Evs[iZeroField].Energy[i];
              PlotPoint(Evs[j],iEff,Fields[j],E);
              if ToLog or (FieldSummary and (j = iZeroField)) then
                LevelLister.ShowLevel(Evs[j],iEff);
            end;
            PlotCurve(Lbl,TLabelPos(LabelsStyle));
            if ToLog or FieldSummary then
              ShowTuningRates(iZeroField);
            if ToLog then begin
              Writeln(Log);
            end;
          end;
          Inc(M,2);
        end;
    finally
      PopProgress;
    end;
  end;

var
  Sym, LastSym, FirstPointInManifold: Integer;
  ManXMax: TQuantumNumber;
begin
  nErrors := 0;
  Manifold.CheckProperties;
  MarkStyle := TMarkStyle.Create;
  MarkStyle.DotAtCentre := True;
  IncludePerturbations := not ExcludePerturbations;
  ShowPoints := CalcPoints or CalcBoth;
  ShowLines := CalcLine or CalcBoth;
  ASelecting := ASelect <> Low(Integer);
  BSelecting := BSelect <> Low(Integer);
  Simulation := Manifold.Mixture.Simulations[0];
  Overflow := nil;
  FirstPointInManifold := nPlottedPoints;
  nStateLabels := 0;
  if ShowPoints or ShowLines then
    if ExternalField then
      SweepField
    else begin
      ManXMax := XMax;
      if (Manifold.Jmax > 0) and LimitJ then
        ManXMax := Min(ManXMax, Manifold.Jmax);
      if ToLog then
        LevelLister.ShowManifold(Manifold,State,IncludePerturbations,Symmetry,XMin,ManXMax,-IgnoreSpins);
      if TrackQuanta then begin
        TIndex := 0;
        MaxTIndex := 0;
      end else begin
        TIndex := -1;
        MaxTIndex := -1;
      end;
      if Simulation.HasFields then
        LastSym := 0
      else
        LastSym := Manifold.nSym-1;
      repeat
        for Sym := 0 to LastSym do
          if (Symmetry < 0) or (Symmetry = Sym) then
            if TrackState and (Manifold.nStates > 1) then
              for i := 0 to Manifold.nStates-1 do
                ShowOneState(Sym,Manifold.State[i],ManXMax)
            else
              ShowOneState(Sym,nil,ManXMax);
        Inc(TIndex);
      until TIndex > MaxTIndex;
      if LabelsStyle <> 0 then
        PlotStateLabels;
    end;
  PlotPoints(FirstPointInManifold); { Plot points last so that filled circles work }

  { Now plot observed energy levels }
  if ObsUpper then
    for i := 0 to High(Observations) do
      with Observations[i] do
        if not (ObsType in [obIntensity,obParameter,obELower,obExpression])
               and PlotObsPoint(ManifoldU, JU, SymU, IndexU, ManifoldL, JL, SymL, IndexL, Frequency, i, True) then
          Break;
  if ObsLower then
    for i := 0 to High(Observations) do
      with Observations[i] do
        if not (ObsType in [obIntensity,obParameter,obEUpper,obExpression])
               and PlotObsPoint(ManifoldL, JL, SymL, IndexL, ManifoldU, JU, SymU, IndexU, -Frequency, i, False) then
          Break;
  MarkStyle.Free;
  ShowStateLabels;
  if nErrors > 0 then
    G.Title1(Format('%d problems with observations found', [nErrors]));
  if Assigned(Overflow) then begin
    WriteLn(Log, Overflow.Message + ' - fail in partition function');
    OverFlow.Free;
  end;
end;

procedure TLevelPlotter.PlotItem(G: TGraphM; MNode: TMNode; Log, FieldSummary: Boolean);
var
  LevelLister: TLevelLister;

  procedure PlotMolecule(Molecule: TMolecule; ShowPartitionFunction: Boolean);
  var
    i: Integer;
    Overflow: Exception;
    Emin: Double;
    DummyArray: TDoubleArray;
  begin
    if ShowPartitionFunction then begin
      if Molecule.Mixture.Simulations[0].Tspin >= 0 then
        SetLength(DummyArray, Molecule.nSpinSpecies)
      else
        SetLength(DummyArray,0);
      LogHeaderShown := Molecule.ActivePartitionFunction.CalculatePartitionFunction(Molecule, Molecule.Mixture.Simulations[0], Overflow, Emin, DummyArray,
                                                            MaxDouble, True, LevelLister.ShowMoleculeLabel, not LogHeaderShown, -IgnoreSpins) > 0;
      Overflow.Free; { Problem shown in log, so do not repeat }
    end else begin
      for i := 0 to Molecule.nManifolds-1 do
        PlotManifold(G, LevelLister, Molecule.Manifold[i], nil, Log, FieldSummary);
      if Log then
        LevelLister.ShowTotals;
    end;
  end;
  procedure PlotSpecies(Species: TSpecies; ShowPartitionFunction: Boolean);
  var
    i: Integer;
  begin
    for i := 0 to Species.Count-1 do
      PlotMolecule(Species.Isotopomer[i], ShowPartitionFunction);
  end;
  procedure PlotMixture(Mixture: TMixture; ShowPartitionFunction: Boolean);
  var
    i: Integer;
  begin
    for i := 0 to Mixture.nSpecies-1 do
      PlotSpecies(Mixture.Species[i], ShowPartitionFunction);
  end;
  procedure PlotState(State: TState);
  begin
    PlotManifold(G, LevelLister, State.Manifold, State, Log, FieldSummary);
  end;
  procedure Plot(ShowPartitionFunction: Boolean);
  begin
    LogHeaderShown := False;
    if MNode is TMixture then
      PlotMixture(TMixture(MNode), ShowPartitionFunction)
    else if MNode is TSpecies then
      PlotSpecies(TSpecies(MNode), ShowPartitionFunction)
    else if MNode is TMolecule then
      PlotMolecule(TMolecule(MNode), ShowPartitionFunction)
    else if MNode is TManifold then
      if ShowPartitionFunction then
        PlotMolecule(TManifold(MNode).Molecule, ShowPartitionFunction)
      else
        PlotManifold(G, LevelLister, TManifold(MNode), nil, Log, FieldSummary)
    else if MNode is TState then
      if ShowPartitionFunction then
        PlotMolecule(TState(MNode).Molecule, ShowPartitionFunction)
      else
        PlotState(TState(MNode));
  end;

var
  XAxis: string;
begin
  MNode.CheckProperties;
  MNode.RootNode.FixupProperties.Free; { Check Mixture also to acount for e.g. spin changes }
  LevelLister := nil;
  if ExternalField then
    if ESweep then
      XAxis := 'Electric Field / Vm-1'
    else
      XAxis := 'Magnetic Field / T'
  else
    XAxis := MapQnoExprName(XQuantumNumber, (MNode as TQnoInfoMNode).QnoInfo(TQnoType(XQuantumNumber div 2)).Name);

  nPlottedPoints := 0;
  G.Start;
  try
    if Log or FieldSummary then
      LevelLister := TLevelLister.Create(Self, MNode,ExternalField,ESweep,FieldSummary,MaxDouble);
    { Plot some Levels }
    if ExternalField then
      G.Map(MinField, MaxField, Ymin, Ymax)
    else
      G.Map(MapQnoExpr(XQuantumNumber,XMin), MapQnoExpr(XQuantumNumber,XMax), Ymin, Ymax);
    G.Title2(SymmetryKey(MNode, AltPlot));
    FirstPoint := True;
    if (Reduction <> 0) and not ExternalField then
      G.Scales(XAxis, '(Energy - ' + FloatToStr(Reduction) + QnoTypeNames[XQuantumNumber div 2] + '(' + QnoTypeNames[XQuantumNumber div 2] + '+1))/' + (MNode.RootNode as TMixture).UnitsName)
    else
      G.Scales(XAxis, 'Energy/' + (MNode.RootNode as TMixture).UnitsName);
    Plot(False);
    if Log and not ExternalField then begin
      Plot(True);
      LevelLister.ShowTotals;
    end;
    if Log then
      LevelLister.ShowErrors;
  finally
    G.Finish;
    LevelLister.Free;
  end;
  SetLength(PlottedPoints,nPlottedPoints);
end;

function TLevelPlotter.GetSetting(AIndex: Integer):Integer;
var
  s: string;
begin
  if not DoneCreating then
    CheckIndices;
  if AIndex = SymmetryIndex then begin
    if not TryStrToSymmetry(Result,SettingToStr(AIndex),FSymmetryList) then
      raise EReportException.Create('Bad symmetry:'+SettingToStr(AIndex));
  end else if ((AIndex = IgnoreSpinsIndex) or (AIndex = ASelectIndex) or (AIndex = BSelectIndex)) then begin
    s := SettingToStr(AIndex);
    if LowerCase(s) = 'all' then
      if AIndex = IgnoreSpinsIndex then
        Result := 0
      else
        Result := Low(Integer)
    else
      Result := inherited GetSetting(AIndex);
  end else
    Result := inherited GetSetting(AIndex);
end;

procedure TLevelPlotter.SetSetting(AIndex: Integer; const AValue: Integer);
begin
  if not DoneCreating then
    CheckIndices;
  if AIndex = SymmetryIndex then
    SetSettingFromStr(AIndex,SymmetryToStr(AValue,FSymmetryList))
  else if (AIndex = IgnoreSpinsIndex) and (GetSetting(AIndex) = 0) then
    SetSettingFromStr(AIndex,'all')
  else if ((AIndex = ASelectIndex) or (AIndex = BSelectIndex)) and (AValue <= -9999) then
    SetSettingFromStr(AIndex,'all')
  else
    inherited SetSetting(AIndex, AValue);
end;

function TLevelPlotter.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  Result := inherited GetSettingsNames(AIndex, LowestValue);
  if AIndex = XQuantumNumberIndex then
    Result := @QnoExprNames[0]
  else if AIndex = LabelsStyleIndex then
    Result := @LabelsStyleNames[0];
end;

procedure TLevelPlotter.CheckIndices;
begin
  {$IFNDEF CONSOLE}
  with FormSettings do begin
  {$ENDIF}
    if SymmetryIndex <= 0 then
      { Can't be set in Create, because needed in intialisation }
      SymmetryIndex := IndexOfSetting('Symmetry');
    if ASelectIndex <= 0 then
      ASelectIndex  :=  IndexOfSetting('ASelect');
    if BSelectIndex <= 0 then
      BSelectIndex  := IndexOfSetting('BSelect');
    if IgnoreSpinsIndex <= 0 then
      IgnoreSpinsIndex := IndexOfSetting('IgnoreSpins');
    if XQuantumNumberIndex <= 0 then
      XQuantumNumberIndex := IndexOfSetting('XQuantumNumber');
    if LabelsStyleIndex <= 0 then
      LabelsStyleIndex := IndexOfSetting('LabelsStyle');
  {$IFNDEF CONSOLE}
  end;
  {$ENDIF}
end;

end.
