unit LmatrixFrm;
{$I utils.inc}{Copyright (C) 2008-14 C M Western; see commandline.pas for license}
interface

uses  SysUtils, Variants, Classes, Graphics, Controls, Forms,
      Dialogs, WhatsThis, Vib, Grids, NGrid, MNode, ExtCtrls, Graphs,
      Split, StdCtrls, NewSpin, GraphM, MathSup, GraphI, Menus, AlEdit, EditExtr,
      ScrlCtrl, MatrixVec, ComCtrls, VibPlotFrm, LmatrixBaseFrm;

type

  TLMatrixForm = class(TLmatrixBaseForm)
    Grid: TNumberGrid;
    PrintGeometryItem: TMenuItem;
    FixCentreofMassItem: TMenuItem;
    OrthogonalizeItem: TMenuItem;
    NormalizeItem: TMenuItem;
    CheckItem: TMenuItem;
    procedure GridChanged(Sender: TObject; Col, Row: Integer);
    procedure GridSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
    procedure GridMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PrintGeometryItemClick(Sender: TObject);
    procedure FixCentreofMassItemClick(Sender: TObject);
    procedure OrthogonalizeItemClick(Sender: TObject);
    procedure OperateButtonClick(Sender: TObject);
    procedure NormalizeItemClick(Sender: TObject);
    procedure CheckItemClick(Sender: TObject);
  private
    State: TElectronicState;
    InReplot: Boolean;
  protected
    procedure SetActiveObject(const Value: TMNode); override;
    function GetActiveObject:TMNode; override;
  public
    procedure PopulateForm(ACol:Integer); override;
  end;

function ConnectLmatrixFrm(AState: TElectronicState): TComponent;

implementation

uses {$IFDEF LCL_} LResources, {$ENDIF}
     Global, Mixture, Math, GStyle, Gmarks, InternalCoords;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectLmatrixFrm(AState: TElectronicState): TComponent;
begin
  Result := ConnectVibFrm(AState, TLMatrixForm);
end;

{ TLMatrixForm }

procedure TLMatrixForm.OperateButtonClick(Sender: TObject);
begin
  OperatePopupMenu.PopupComponent := OperateButton;
  OperatePopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TLMatrixForm.CheckItemClick(Sender: TObject);
var
  E: Exception;
begin
  E := State.FixupProperties;
  if Assigned(E) then begin
    E.Free;
  end;
end;

procedure TLMatrixForm.FixCentreofMassItemClick(Sender: TObject);
begin
  State.FixCentreOfMass(Sender);
end;

function TLMatrixForm.GetActiveObject: TMNode;
begin
  Result := State;
end;

procedure TLMatrixForm.GridChanged(Sender: TObject; Col, Row: Integer);
var
  V: Double;
  p: Integer;
begin
  if Row > 0 then
    if Col = 1 then
      with State.Nucleus((Row-1) div 3) do
        SetParameter(IndexOfParameter('x') + (Row-1) mod 3,Grid[Row,Col])
    else if (Col > 1) and (Col-2 < State.Molecule.nModes) then
      with State.Mode(Col-2) do begin
        V := Grid[Row,Col]/MassWeight(State, TDisplacementSelect(DispSelect.ItemIndex), (Row-1) div 3, Col-2);
        p := FirstDynamicParameter+Row-1;
        { Test to avoid rounding errors looking like a change }
        if not SameValue(V,GetParameter(p)) then
          SetParameter(p,V);
      end;
end;

procedure TLMatrixForm.GridMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  PopulateForm;
end;

procedure TLMatrixForm.GridSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
  PopulateForm(ACol);
end;

procedure TLMatrixForm.NormalizeItemClick(Sender: TObject);
begin
  State.Normalize(Sender);
end;

procedure TLMatrixForm.OrthogonalizeItemClick(Sender: TObject);
begin
  State.Orthogonalize(Sender);
end;

procedure TLMatrixForm.PopulateForm(ACol:Integer);
var
  i, j, b, Pass: Integer;
  Nucleus: TNucleus;
  xMin, xMax, yMin, yMax: Double;
  LimitsSet, InPrin: Boolean;
  P, L: TDPointArray;
begin
  if (csLoading in ComponentState) or InReplot then
    Exit;
  try
    InReplot := True;
    if Acol < 0 then
      ACol := Grid.Col;
    Caption := 'L matrix for ' + State.UniqueName;
    InPrin := not State.PrepareMatrices(False,ShowTRCheckBox.Checked);
    Grid.RowCount := State.Molecule.nNuclei*3+1;
    if not Assigned(State.adm) then
      exit;
    Grid.ColCount := State.adm.ColCount+2;
    if State.Molecule.nNuclei = 0 then
      Exit;
    { Equilibrium geometries }
    for i := 0 to State.Molecule.nNuclei - 1 do begin
      Nucleus := State.Nucleus(i);
      for j := 0 to 2 do begin
        b := i*3+j+1;
        Grid.Lbl[b] := State.nameNucleusXYZ(b);
        Grid.Value[b,1] := Nucleus.GetParameter(Nucleus.IndexOfParameter('x')+j)
      end;
    end;
    for i := 0 to State.Molecule.nModes - 1 do begin
      Grid.Title[i+2] := State.nameMode(i+1);
      with State.Mode(i) do
        for j := 0 to 3*State.Molecule.nNuclei - 1 do
          Grid.Value[j+1,i+2] := GetParameter(FirstDynamicParameter+j)*MassWeight(State, TDisplacementSelect(DispSelect.ItemIndex), j div 3, i);
    end;
    { Blank any invalid vibrations }
    b := Min(Max(State.Molecule.nModes,3*State.Molecule.nNuclei-6),Grid.ColCount-2);
    for i := State.Molecule.nModes to b-1 do begin
      Grid.Title[i+2] := ' ';
      for j := 0 to 3*State.Molecule.nNuclei - 1 do
        Grid.Valid[j+1,i+2] := False;
    end;
    if ShowTRCheckBox.Checked then begin
      { Rotations }
      b := Max(State.Molecule.nModes,3*State.Molecule.nNuclei-6);
      for i := 1 to 3 do
        if State.Ixx[i] > 1e-10 then begin
          Grid.Title[b+2] := 'R'+'xyz'[i];
          for j := 0 to 3*State.Molecule.nNuclei - 1 do
            Grid.Value[j+1,b+2] := State.adm[j+1,b+1]*MassWeight(State, TDisplacementSelect(DispSelect.ItemIndex), j div 3, -1);
          Inc(b);
        end;
      { Translations }
      b := 3*State.Molecule.nNuclei-3;
      for i := 0 to 2 do begin
        Grid.Title[b+2] := 'T'+Char(Ord('x')+i);
        for j := 0 to 3*State.Molecule.nNuclei - 1 do
          Grid.Value[j+1,b+2] := State.adm[j+1,b+1]*MassWeight(State, TDisplacementSelect(DispSelect.ItemIndex), j div 3, -1);
        Inc(b);
      end;
    end;
    Grid.AutoAdjustColumn(0);
    LimitsSet := False;
    G.Start;
    if State.LastError <> '' then
      G.Title1(State.LastError)
    else if (ACol > 1) and (ACol-1 <= State.Molecule.nModes) then
      G.Title1(State.nameMode(ACol-1)+' '+FloatToStr(State.Mode(ACol-2).Omega));
    for pass := 0 to 1 do
      PlotAtoms(State, ACol-2, pass=0, pass=1, False, False, InPrin,
                3, P, L, xMin, xMax, yMin, yMax, LimitsSet);
    G.Finish;
  finally
    InReplot := False;
  end;
end;

procedure TLMatrixForm.PrintGeometryItemClick(Sender: TObject);
begin
  State.PrintMolecularInfo(Sender);
end;

procedure TLMatrixForm.SetActiveObject(const Value: TMNode);
begin
  State := Value as TElectronicState;
end;

end.
