unit MNodeClip;

{$I utils.inc}

interface

uses
  Classes, SysUtils, MNode, pgDOM;

function MNodeClipboardFormat: Cardinal;
function ReadXMLMNodeFromClipboard: TXMLDocument;
function PasteXMLNodeOnto(TargetMNode: TMNode):TMNode;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF LCL_} LCLType, {$ENDIF}
     UCompat, Forms, Dialogs, Huge, Mixture, ClipBrd, pgXMLRead, RenForm;

var
  MNodeClipboardFormatVar: Cardinal = 0;

function MNodeClipboardFormat: Cardinal;
begin
  if MNodeClipboardFormatVar = 0 then
    MNodeClipboardFormatVar := RegisterClipboardFormat('MNode');
  Result := MNodeClipboardFormatVar;
end;

function ReadXMLMNodeFromClipboard: TXMLDocument;
var
  Stream: THugeMemoryStream;
begin
  Result := nil;
  Stream := THugeMemoryStream.Create;
  try
    Stream.SetFromClipboard(MNodeClipboardFormat);
    ReadXMLFile(Result, TStream(Stream), '<Clipboard>');
  finally
    Stream.Free;
  end;
end;

function PasteXMLNodeOnto(TargetMNode: TMNode):TMNode;
var
  MNode, NewMNode, MNodeP: TMNode;
  XMLDocument: TXMLDocument;
  Child: TDOMNode;
  NewSubClass: TMNodeClass;
  i, nUnused, n, NewIndex, TargetIndex: Integer;
  OldName: string;
  OldNames, NewNames: TStrings;
  SkipThis: Boolean;
begin
  Result := nil;
  nUnused := 0;
  if Assigned(TargetMNode) and Clipboard.HasFormat(MNodeClipboardFormat) then begin
    (TargetMNode.RootNode as TRootMNode).BeginPaste;
    TargetIndex := TargetMNode.ComponentIndex;
    XMLDocument := ReadXMLMNodeFromClipboard;
    OldNames := nil;
    NewNames := nil;
    try
      Child := XMLDocument.FirstChild;
      if Assigned(Child) and (Child.NodeName = 'List') then
        Child := Child.FirstChild;
      OldName := '';
      while Assigned(Child) do begin
        SkipThis := False;
        NewIndex := -1;
        NewSubClass := ReadMNodeClass(Child.NodeName, nil);
        MNode := nil;
        if NewSubClass = nil then
          SkipThis := True
        else if Assigned(TargetMNode.Owner) and NewSubClass.CanBeChildOf(TargetMNode.Owner as TMNode) then begin
          MNode := TargetMNode.Owner as TMNode;
          Inc(TargetIndex);
          NewIndex := TargetIndex;
        end else if NewSubClass.CanBeChildOf(TargetMNode) then
          MNode := TargetMNode
        else if NewSubClass.InheritsFrom(TTransitionMoment) then begin
          { Find Molecule }
          MNode := TargetMNode;
          while Assigned(MNode) and not (MNode is TMolecule) do
            MNode := MNode.Owner as TMNode;
          if Assigned(MNode) then begin
            MNodeP := MNode.ChildWithClass(TTransitionMoments);
            if (MNodeP = nil) and (MNode.ChildWithClass(TTransitionMoments,0) <> nil) then begin
              SkipThis := True;
              { TODO: Could look harder - probably only one place to paste it }
              Application.MessageBox('More than one place to paste transition moment - skipping it','Paste Problem',MB_OK);
            end;
            MNode := MNodeP;
          end;
        end;
        if not SkipThis then begin
          if NewSubClass.InheritsFrom(TNucleus) then begin
            MNodeP := MNode;
            while Assigned(MNodeP) and not (MNodeP is TMolecule) do
              MNodeP := MNodeP.Owner as TMNode;
            if Assigned(MNodeP) then begin
              (MNodeP as TMolecule).nNuclei := (MNodeP as TMolecule).nNuclei+1;
              NewMNode := (MNode as TState).Nucleus((MNodeP as TMolecule).nNuclei-1);
              n := NewMNode.ReadProperties(Child)
            end else
              raise NotYet;
          end else begin
            if not Assigned(MNode) then
              raise EReportException.Create('Nowhere to paste ' + NewSubClass.ClassName + '(' + TargetMNode.ClassName + ')');
            NewMNode := MNode.ReadMNode(Child, n);
          end;
          Inc(nUnused,n);
          if NewMNode is TMNodePair then begin
            if Assigned(OldNames) then
              { Rename MNode pairs using recorded name changes }
              with TMNodePair(NewMNode) do begin
                i := OldNames.IndexOf(FBraName);
                if i >= 0 then
                  FBraName := NewNames[i];
                i := OldNames.IndexOf(FKetName);
                if i >= 0 then
                  FKetName := NewNames[i];
              end;
            if Assigned(MNode.ComponentWithName(NewMNode.Name, TMNode, NewMNode)) then
              if not RenameForm.Rename(NewMNode as TMNodePair, NewMNode, True) then
                NewMNode := nil;
          end else begin
            { Rename if a node of the same name already exists, keeping a record of the names }
            OldName := NewMNode.Name;
            while Assigned(MNode.ComponentWithName(NewMNode.Name, TMNode, NewMNode)) do
              NewMNode.Name := 'Copy of ' + NewMNode.Name;
            if OldName <> NewMNode.Name then begin
              if not Assigned(OldNames) then begin
                OldNames := TStringList.Create;
                NewNames := TStringList.Create;
              end;
              OldNames.Add(OldName);
              NewNames.Add(NewMNode.Name);
            end;
          end;
          if Assigned(NewMNode) then
            try
              if NewIndex >= 0 then
                NewMNode.ComponentIndex := NewIndex;
              RenameForm.RenameMNodePairsAsRequired(NewMNode, False);
            except
              NewMNode.Free;
              raise;
            end;
          if not Assigned(Result) then
            Result := NewMNode;
        end;
        Child := Child.NextSibling;
      end;
    finally
      XMLDocument.Free;
      OldNames.Free;
      NewNames.Free;
    end;
  end;
  if nUnused > 0 then
    ShowMessage(IntToStr(nUnused)+' unused items - see log for details');
end;

end.

