unit MNodeColour;
{ Helper functions for MNode Colour Properties }
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Graphics;

function GetColourNames(out LowestValue: Integer): PPChar; { for MNode.GetSettingsNames }
function ColourToStr(Colour: TColor): string;
function StrToColour(const s: string; AskIfUnknown: Boolean): TColor;
function IsVisibleColour(Colour: TColor):Boolean;

const
  clHide = TColor($1FFFFFFE);
  {$IFDEF FPC}
  clWebSienna = $2D52A0;
  clWebOrange = $00A5FF;
  {$ENDIF}
  MNODEColourValues: array [0..18] of TColor =
   ( clNone, clMaroon, clGreen, clOlive, clNavy, clPurple, clTeal, clGray, clWebSienna, clRed,
     clLime, clWebOrange, clBlue, clFuchsia, clAqua, clBlack, clWhite, clHide, clNone);


implementation

uses {$IFNDEF CONSOLE} Dialogs, Forms, Graphs, GraphI, {$ENDIF}
     MNode, SysUtils, UCompat, Worker;

const
  ColourNames: array [0..19] of PChar =
   ( 'None', 'Maroon', 'Green', 'Olive', 'Navy', 'Purple', 'Teal', 'Gray', 'Brown', 'Red',
     'Lime', 'Orange', 'Blue', 'Fuchsia', 'Aqua', 'Black', 'White', 'Hide', 'Other...', nil);

function IsVisibleColour(Colour: TColor):Boolean;
begin
  Result := (Colour <> clHide) and (Colour <> clWhite) and (Colour <> clDefault) and (Colour <> clNone);
end;

var
  LastColour: TColor;

function GetColourNames(out LowestValue: Integer): PPChar;
{ for MNode.GetSettingsNames }
begin
  LowestValue := 0;
  Result := @ColourNames[0];
end;

function ColourToStr(Colour: TColor): string;
var
  i: Integer;
begin
  { Special cases - Yellow -> Orange after 5.2.974 }
  if Colour = clYellow then begin
    Result := 'Yellow';
    Exit;
  end else if Colour = clSilver then begin
    Result := 'Silver';
    Exit;
  end;
  LastColour := Colour;
  for i := 0 to High(MNodeColourValues) do
    if Colour = MNodeColourValues[i] then begin
      Result := ColourNames[i];
      Exit;
    end;
  Result := Format('$%.6x', [Colour]);
  if (Colour and not clWhite) <> 0 then
    raise Exception.Create('Bad Colour:' + Result);
end;

function StrToColour(const s: string; AskIfUnknown: Boolean): TColor;
var
  i: Integer;
  {$IFNDEF CONSOLE}
  Dialog: TColorDialog;
  {$ENDIF}
begin
  if s = '' then
    Result := clNone
  else if Length(s) <= 2 then begin
    { Backward compatibility only }
    i := StrToInt(s);
    if i = 0 then
      Result := clNone
    else
      Result := {$IFDEF CONSOLE} clBlack {$ELSE} DefaultPenColour[i] {$ENDIF};
  end else begin
    for i := 0 to High(MNodeColourValues)-1 do
      if CompareText(s,ColourNames[i]) = 0 then begin
        Result := MNodeColourValues[i];
        Exit;
      end;
    { Special cases - Yellow -> Orange after 5.2.974 }
    if CompareText(s,'Yellow') = 0 then
      Result := clYellow
    else if CompareText(s,'Silver') = 0 then
      Result := clSilver
    else if TryStrToInt(s, i) then
      Result := i
    else
      Result := clNone;
    if (Result and not clWhite) <> 0 then begin
      {$IFNDEF CONSOLE}
      if AskIfUnknown then begin
        Dialog := TColorDialog.Create(Application);
        Result := LastColour;
        try
          Dialog.Color := LastColour;
          if Dialog.Execute then
            Result := Dialog.Color;
        finally
          Dialog.Free;
        end;
      end else
        raise EReportException.Create('Bad colour name:'+s);
      {$ELSE}
      WriteLn(Log, 'Bad Colour Name:'+s); { Don't crash as probably harmless in console }
      {$ENDIF}
    end;
  end;
end;

end.
