unit OverlaySumFitter;
{$I utils.inc}{Copyright (C) 2010 C M Western; see commandline.pas for license}
interface

uses MNode, Mixture, Overlay, Fitter;

type
  TComponentType = (None, Component, Floating, Target);

function FitOverlaySum(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;

implementation

uses SysUtils, UCompat, MatrixVec, ExpPlot, Constrain, Units;

type

  TExperimentalPlotArray = array of TExperimentalPlot;

  TOverlaySumFitter = class(TFitter)
  public
    Spectra: TOverlays;
    Sources: TExperimentalPlotArray;
    function CalculateAllObs(v: TColumnVector): Boolean; override;
  end;

function TOverlaySumFitter.CalculateAllObs(v: TColumnVector): Boolean;
var
  i, j, k, l, p: Integer;
  Sum, f, vi, vf: Double;
begin
  Result := True;
  l := 1;
  for i := 0 to Spectra.nPlots-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
          for j := StartPoint to EndPoint do begin
            Sum := 0;
            f := AdjustFrequency(Frequency[j]);
            for k := 0 to High(Sources) do begin
              vf := Convert(f,Units,Sources[k].Units);
              vi := Sources[k].IntensityAt(vf,p);
              vi := ConvertIntensity(vi,vf,Sources[k].Units,Sources[k].IntensityUnits,Units,IntensityUnits);
              Sum := Sum + GetParameter(FirstDynamicParameter+k)*vi;
            end;
            { TODO: Flip logic as per contour fit }
            { Logically we should be comparing to the values in the observed TExperimentalPlot.
              However the calculated values would not then depend on the scaling parameters in
              the observed spectrum, so we apply the reverse transform to the calculated values }
            v[l] := Sum/Scale + Baseline; { Inverse transform }
            Inc(l);
          end;
end;

function FitOverlaySum(Mixture: TMixture; Spectra: TOverlays; ShowContour: OnShowContour; var Parameters: TParameters; JustResiduals: Boolean; CheckDerivatives: Double):string;
  var
    Sources: TExperimentalPlotArray;
var
  f: TOverlaySumFitter;
  nObs: Integer;
  i, j, l: Integer;
begin
  nObs := 0;
  Sources := nil;
  { Find total number of observed points }
  for i := 0 to Spectra.nPlots-1 do
    if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do begin
        if Count > 0 then
          raise EReportException.Create('Intensity Correction not implemented for overlay sum fitting');
        case ComponentType of
          None: ;
          Component, OverlaySumFitter.Floating:
            begin
              SetLength(Sources,Length(Sources)+1);
              Sources[High(Sources)] := TExperimentalPlot(Spectra[i]);
            end;
          Target: Inc(nObs, EndPoint-StartPoint+1);
        end;
      end;
  ApplyConstraints;
  CheckConstraints(Parameters);
  f := TOverlaySumFitter.Create(Length(Parameters), nObs, Mixture, CheckDerivatives);
  f.Spectra := Spectra;
  f.Sources := Sources;
  { Because of inverse scaling used in CalculateObs (see comment)
    resdiduals are inversely proportional to spectrum scale factors,
    so flag as such }
  for i := 0 to High(Parameters) do
    with Parameters[i] do
      if (MNode is TExperimentalPlot) and (Index > 0) and (MNode.ParameterNames[Index] = 'Scale') then
        Inverse := True;
  try
    f.CalculateAllObs(f.omc);
    if Assigned(ShowContour) then
      ShowContour(Spectra, f.omc, True);
    if JustResiduals then Exit;
    {  Calculate Derivatives wrt each parameter }
    { TODO: Optimize for simple derivatives }
    { Note that f.weights is only used as workspace, and all weights are taken as 1 }
    f.CalculateDerivatives(Parameters);
    FreeAndNil(f.Weights);
    l := 1; { Finally work out observed - calculated }
    for i := 0 to Spectra.nPlots-1 do
     if Spectra[i] is TExperimentalPlot then
      with TExperimentalPlot(Spectra[i]) do
        if ComponentType = Target then
          for j := StartPoint to EndPoint do begin
            // Note Baseline and Scale applied to calculated, not observed
            f.omc[l] := f.omc[l] - Intensity[j];
            Inc(l);
          end;
    Result := f.DoLeastSQ(Parameters, Mixture.Precision);
  finally
    f.Free;
  end;
end;

end.
