unit CCITT;
{$H+} { Based on code from xpdf }
interface

uses Classes, PDFStream, PDFReader;

type

  { TCCITFaxDecompressionStream }

  TCCITTFaxDecompressionStream = class(TPDFStream)
  private
    K: Integer;
    Columns, Rows, row: Cardinal;
    byteAlign, endOfBlock, Black: Boolean;
    NextLine2D, EofPending: Boolean;
    a0, b1, inputBits, outputBits, buf, inputBuf: Integer;
    codingLine, refLine: array of Word;
  public
    constructor Create(Source: TStream; Params: TPDFDictionary);
    function NextByte:Byte; override;
    function getTwoDimCode: Word;
    function getWhiteCode: Word;
    function getBlackCode: Word;
    function lookBits(n: Integer):Word;
    procedure eatBits(n: Integer);
  end;


implementation

uses SysUtils, Math;

const
  twoDimPass = 0;
  twoDimHoriz = 1;
  twoDimVert0 = 2;
  twoDimVertR1 = 3;
  twoDimVertL1 = 4;
  twoDimVertR2 = 5;
  twoDimVertL2 = 6;
  twoDimVertR3 = 7;
  twoDimVertL3 = 8;
  encEof = Word(-1);
  ccittEOL = -2;

type
  CCITTCode = record
    bits: SmallInt;
    n: SmallInt;
  end;
  PCCITTCode = ^CCITTCode;

const
twoDimTable: array[0..127] of CCITTCode = (
  (bits:-1; n:-1),
  (bits:-1; n:-1),
  (bits:7; n:twoDimVertL3),                     // 0000010
  (bits:7; n:twoDimVertR3),                     // 0000011
  (bits:6; n:twoDimVertL2), (bits:6; n:twoDimVertL2),   // 000010x
  (bits:6; n:twoDimVertR2), (bits:6; n:twoDimVertR2),   // 000011x
  (bits:4; n:twoDimPass), (bits:4; n:twoDimPass),     // 0001xxx
    (bits:4; n:twoDimPass), (bits:4; n:twoDimPass),
    (bits:4; n:twoDimPass), (bits:4; n:twoDimPass),
    (bits:4; n:twoDimPass), (bits:4; n:twoDimPass),
  (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),     // 001xxxx
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
    (bits:3; n:twoDimHoriz), (bits:3; n:twoDimHoriz),
  (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),   // 010xxxx
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
    (bits:3; n:twoDimVertL1), (bits:3; n:twoDimVertL1),
  (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),   // 011xxxx
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
    (bits:3; n:twoDimVertR1), (bits:3; n:twoDimVertR1),
  (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),     // 1xxxxxx
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0),
    (bits:1; n:twoDimVert0), (bits:1; n:twoDimVert0)
);

// 11-12 bit codes (upper 7 bits are 0)
whiteTable1: array[0..31] of CCITTCode = (
  (bits:-1; n:-1),                                      // 00000
  (bits:12; n:ccittEOL),                                // 00001
  (bits:-1; n:-1),(bits:-1; n:-1),                              // 0001x
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),      // 001xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),      // 010xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),      // 011xx
  (bits:11; n:1792),(bits:11; n:1792),                  // 1000x
  (bits:12; n:1984),                                    // 10010
  (bits:12; n:2048),                                    // 10011
  (bits:12; n:2112),                                    // 10100
  (bits:12; n:2176),                                    // 10101
  (bits:12; n:2240),                                    // 10110
  (bits:12; n:2304),                                    // 10111
  (bits:11; n:1856),(bits:11; n:1856),                  // 1100x
  (bits:11; n:1920),(bits:11; n:1920),                  // 1101x
  (bits:12; n:2368),                                    // 11100
  (bits:12; n:2432),                                    // 11101
  (bits:12; n:2496),                                    // 11110
  (bits:12; n:2560)                                     // 11111
);

// 1-9 bit codes
whiteTable2: array[0..511] of CCITTCode = (
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),      // 0000000xx
  (bits:8; n:29),(bits:8; n:29),                                // 00000010x
  (bits:8; n:30),(bits:8; n:30),                                // 00000011x
  (bits:8; n:45),(bits:8; n:45),                                // 00000100x
  (bits:8; n:46),(bits:8; n:46),                                // 00000101x
  (bits:7; n:22),(bits:7; n:22),(bits:7; n:22),(bits:7; n:22),          // 0000011xx
  (bits:7; n:23),(bits:7; n:23),(bits:7; n:23),(bits:7; n:23),          // 0000100xx
  (bits:8; n:47),(bits:8; n:47),                                // 00001010x
  (bits:8; n:48),(bits:8; n:48),                                // 00001011x
  (bits:6; n:13),(bits:6; n:13),(bits:6; n:13),(bits:6; n:13),          // 000011xxx
    (bits:6; n:13),(bits:6; n:13),(bits:6; n:13),(bits:6; n:13),
  (bits:7; n:20),(bits:7; n:20),(bits:7; n:20),(bits:7; n:20),          // 0001000xx
  (bits:8; n:33),(bits:8; n:33),                                // 00010010x
  (bits:8; n:34),(bits:8; n:34),                                // 00010011x
  (bits:8; n:35),(bits:8; n:35),                                // 00010100x
  (bits:8; n:36),(bits:8; n:36),                                // 00010101x
  (bits:8; n:37),(bits:8; n:37),                                // 00010110x
  (bits:8; n:38),(bits:8; n:38),                                // 00010111x
  (bits:7; n:19),(bits:7; n:19),(bits:7; n:19),(bits:7; n:19),          // 0001100xx
  (bits:8; n:31),(bits:8; n:31),                                // 00011010x
  (bits:8; n:32),(bits:8; n:32),                                // 00011011x
  (bits:6; n:1),(bits:6; n:1),(bits:6; n:1),(bits:6; n:1),              // 000111xxx
    (bits:6; n:1),(bits:6; n:1),(bits:6; n:1),(bits:6; n:1),
  (bits:6; n:12),(bits:6; n:12),(bits:6; n:12),(bits:6; n:12),          // 001000xxx
    (bits:6; n:12),(bits:6; n:12),(bits:6; n:12),(bits:6; n:12),
  (bits:8; n:53),(bits:8; n:53),                                // 00100100x
  (bits:8; n:54),(bits:8; n:54),                                // 00100101x
  (bits:7; n:26),(bits:7; n:26),(bits:7; n:26),(bits:7; n:26),          // 0010011xx
  (bits:8; n:39),(bits:8; n:39),                                // 00101000x
  (bits:8; n:40),(bits:8; n:40),                                // 00101001x
  (bits:8; n:41),(bits:8; n:41),                                // 00101010x
  (bits:8; n:42),(bits:8; n:42),                                // 00101011x
  (bits:8; n:43),(bits:8; n:43),                                // 00101100x
  (bits:8; n:44),(bits:8; n:44),                                // 00101101x
  (bits:7; n:21),(bits:7; n:21),(bits:7; n:21),(bits:7; n:21),          // 0010111xx
  (bits:7; n:28),(bits:7; n:28),(bits:7; n:28),(bits:7; n:28),          // 0011000xx
  (bits:8; n:61),(bits:8; n:61),                                // 00110010x
  (bits:8; n:62),(bits:8; n:62),                                // 00110011x
  (bits:8; n:63),(bits:8; n:63),                                // 00110100x
  (bits:8; n:0),(bits:8; n:0),                          // 00110101x
  (bits:8; n:320),(bits:8; n:320),                              // 00110110x
  (bits:8; n:384),(bits:8; n:384),                              // 00110111x
  (bits:5; n:10),(bits:5; n:10),(bits:5; n:10),(bits:5; n:10),          // 00111xxxx
    (bits:5; n:10),(bits:5; n:10),(bits:5; n:10),(bits:5; n:10),
    (bits:5; n:10),(bits:5; n:10),(bits:5; n:10),(bits:5; n:10),
    (bits:5; n:10),(bits:5; n:10),(bits:5; n:10),(bits:5; n:10),
  (bits:5; n:11),(bits:5; n:11),(bits:5; n:11),(bits:5; n:11),          // 01000xxxx
    (bits:5; n:11),(bits:5; n:11),(bits:5; n:11),(bits:5; n:11),
    (bits:5; n:11),(bits:5; n:11),(bits:5; n:11),(bits:5; n:11),
    (bits:5; n:11),(bits:5; n:11),(bits:5; n:11),(bits:5; n:11),
  (bits:7; n:27),(bits:7; n:27),(bits:7; n:27),(bits:7; n:27),          // 0100100xx
  (bits:8; n:59),(bits:8; n:59),                                // 01001010x
  (bits:8; n:60),(bits:8; n:60),                                // 01001011x
  (bits:9; n:1472),                                     // 010011000
  (bits:9; n:1536),                                     // 010011001
  (bits:9; n:1600),                                     // 010011010
  (bits:9; n:1728),                                     // 010011011
  (bits:7; n:18),(bits:7; n:18),(bits:7; n:18),(bits:7; n:18),          // 0100111xx
  (bits:7; n:24),(bits:7; n:24),(bits:7; n:24),(bits:7; n:24),          // 0101000xx
  (bits:8; n:49),(bits:8; n:49),                                // 01010010x
  (bits:8; n:50),(bits:8; n:50),                                // 01010011x
  (bits:8; n:51),(bits:8; n:51),                                // 01010100x
  (bits:8; n:52),(bits:8; n:52),                                // 01010101x
  (bits:7; n:25),(bits:7; n:25),(bits:7; n:25),(bits:7; n:25),          // 0101011xx
  (bits:8; n:55),(bits:8; n:55),                                // 01011000x
  (bits:8; n:56),(bits:8; n:56),                                // 01011001x
  (bits:8; n:57),(bits:8; n:57),                                // 01011010x
  (bits:8; n:58),(bits:8; n:58),                                // 01011011x
  (bits:6; n:192),(bits:6; n:192),(bits:6; n:192),(bits:6; n:192),      // 010111xxx
    (bits:6; n:192),(bits:6; n:192),(bits:6; n:192),(bits:6; n:192),
  (bits:6; n:1664),(bits:6; n:1664),(bits:6; n:1664),(bits:6; n:1664),  // 011000xxx
    (bits:6; n:1664),(bits:6; n:1664),(bits:6; n:1664),(bits:6; n:1664),
  (bits:8; n:448),(bits:8; n:448),                              // 01100100x
  (bits:8; n:512),(bits:8; n:512),                              // 01100101x
  (bits:9; n:704),                                      // 011001100
  (bits:9; n:768),                                      // 011001101
  (bits:8; n:640),(bits:8; n:640),                              // 01100111x
  (bits:8; n:576),(bits:8; n:576),                              // 01101000x
  (bits:9; n:832),                                      // 011010010
  (bits:9; n:896),                                      // 011010011
  (bits:9; n:960),                                      // 011010100
  (bits:9; n:1024),                                     // 011010101
  (bits:9; n:1088),                                     // 011010110
  (bits:9; n:1152),                                     // 011010111
  (bits:9; n:1216),                                     // 011011000
  (bits:9; n:1280),                                     // 011011001
  (bits:9; n:1344),                                     // 011011010
  (bits:9; n:1408),                                     // 011011011
  (bits:7; n:256),(bits:7; n:256),(bits:7; n:256),(bits:7; n:256),      // 0110111xx
  (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),              // 0111xxxxx
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
    (bits:4; n:2),(bits:4; n:2),(bits:4; n:2),(bits:4; n:2),
  (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),              // 1000xxxxx
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
    (bits:4; n:3),(bits:4; n:3),(bits:4; n:3),(bits:4; n:3),
  (bits:5; n:128),(bits:5; n:128),(bits:5; n:128),(bits:5; n:128),      // 10010xxxx
    (bits:5; n:128),(bits:5; n:128),(bits:5; n:128),(bits:5; n:128),
    (bits:5; n:128),(bits:5; n:128),(bits:5; n:128),(bits:5; n:128),
    (bits:5; n:128),(bits:5; n:128),(bits:5; n:128),(bits:5; n:128),
  (bits:5; n:8),(bits:5; n:8),(bits:5; n:8),(bits:5; n:8),              // 10011xxxx
    (bits:5; n:8),(bits:5; n:8),(bits:5; n:8),(bits:5; n:8),
    (bits:5; n:8),(bits:5; n:8),(bits:5; n:8),(bits:5; n:8),
    (bits:5; n:8),(bits:5; n:8),(bits:5; n:8),(bits:5; n:8),
  (bits:5; n:9),(bits:5; n:9),(bits:5; n:9),(bits:5; n:9),              // 10100xxxx
    (bits:5; n:9),(bits:5; n:9),(bits:5; n:9),(bits:5; n:9),
    (bits:5; n:9),(bits:5; n:9),(bits:5; n:9),(bits:5; n:9),
    (bits:5; n:9),(bits:5; n:9),(bits:5; n:9),(bits:5; n:9),
  (bits:6; n:16),(bits:6; n:16),(bits:6; n:16),(bits:6; n:16),          // 101010xxx
    (bits:6; n:16),(bits:6; n:16),(bits:6; n:16),(bits:6; n:16),
  (bits:6; n:17),(bits:6; n:17),(bits:6; n:17),(bits:6; n:17),          // 101011xxx
    (bits:6; n:17),(bits:6; n:17),(bits:6; n:17),(bits:6; n:17),
  (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),              // 1011xxxxx
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
    (bits:4; n:4),(bits:4; n:4),(bits:4; n:4),(bits:4; n:4),
  (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),              // 1100xxxxx
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
    (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),
  (bits:6; n:14),(bits:6; n:14),(bits:6; n:14),(bits:6; n:14),          // 110100xxx
    (bits:6; n:14),(bits:6; n:14),(bits:6; n:14),(bits:6; n:14),
  (bits:6; n:15),(bits:6; n:15),(bits:6; n:15),(bits:6; n:15),          // 110101xxx
    (bits:6; n:15),(bits:6; n:15),(bits:6; n:15),(bits:6; n:15),
  (bits:5; n:64),(bits:5; n:64),(bits:5; n:64),(bits:5; n:64),          // 11011xxxx
    (bits:5; n:64),(bits:5; n:64),(bits:5; n:64),(bits:5; n:64),
    (bits:5; n:64),(bits:5; n:64),(bits:5; n:64),(bits:5; n:64),
    (bits:5; n:64),(bits:5; n:64),(bits:5; n:64),(bits:5; n:64),
  (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),              // 1110xxxxx
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
    (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),
  (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),              // 1111xxxxx
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7),
    (bits:4; n:7),(bits:4; n:7),(bits:4; n:7),(bits:4; n:7)
);

//------------------------------------------------------------------------
// black run lengths
//------------------------------------------------------------------------

// 10-13 bit codes (upper 6 bits are 0)
blackTable1: array[0..127] of CCITTCode = (
  (bits:-1; n:-1),(bits:-1; n:-1),                                      // 000000000000x
  (bits:12; n:ccittEOL),(bits:12; n:ccittEOL),                  // 000000000001x
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000001xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000010xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000011xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000100xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000101xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000110xx
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 00000000111xx
  (bits:11; n:1792),(bits:11; n:1792),(bits:11; n:1792),(bits:11; n:1792),      // 00000001000xx
  (bits:12; n:1984),(bits:12; n:1984),                          // 000000010010x
  (bits:12; n:2048),(bits:12; n:2048),                          // 000000010011x
  (bits:12; n:2112),(bits:12; n:2112),                          // 000000010100x
  (bits:12; n:2176),(bits:12; n:2176),                          // 000000010101x
  (bits:12; n:2240),(bits:12; n:2240),                          // 000000010110x
  (bits:12; n:2304),(bits:12; n:2304),                          // 000000010111x
  (bits:11; n:1856),(bits:11; n:1856),(bits:11; n:1856),(bits:11; n:1856),      // 00000001100xx
  (bits:11; n:1920),(bits:11; n:1920),(bits:11; n:1920),(bits:11; n:1920),      // 00000001101xx
  (bits:12; n:2368),(bits:12; n:2368),                          // 000000011100x
  (bits:12; n:2432),(bits:12; n:2432),                          // 000000011101x
  (bits:12; n:2496),(bits:12; n:2496),                          // 000000011110x
  (bits:12; n:2560),(bits:12; n:2560),                          // 000000011111x
  (bits:10; n:18),(bits:10; n:18),(bits:10; n:18),(bits:10; n:18),              // 0000001000xxx
    (bits:10; n:18),(bits:10; n:18),(bits:10; n:18),(bits:10; n:18),
  (bits:12; n:52),(bits:12; n:52),                                      // 000000100100x
  (bits:13; n:640),                                             // 0000001001010
  (bits:13; n:704),                                             // 0000001001011
  (bits:13; n:768),                                             // 0000001001100
  (bits:13; n:832),                                             // 0000001001101
  (bits:12; n:55),(bits:12; n:55),                                      // 000000100111x
  (bits:12; n:56),(bits:12; n:56),                                      // 000000101000x
  (bits:13; n:1280),                                            // 0000001010010
  (bits:13; n:1344),                                            // 0000001010011
  (bits:13; n:1408),                                            // 0000001010100
  (bits:13; n:1472),                                            // 0000001010101
  (bits:12; n:59),(bits:12; n:59),                                      // 000000101011x
  (bits:12; n:60),(bits:12; n:60),                                      // 000000101100x
  (bits:13; n:1536),                                            // 0000001011010
  (bits:13; n:1600),                                            // 0000001011011
  (bits:11; n:24),(bits:11; n:24),(bits:11; n:24),(bits:11; n:24),              // 00000010111xx
  (bits:11; n:25),(bits:11; n:25),(bits:11; n:25),(bits:11; n:25),              // 00000011000xx
  (bits:13; n:1664),                                            // 0000001100100
  (bits:13; n:1728),                                            // 0000001100101
  (bits:12; n:320),(bits:12; n:320),                                    // 000000110011x
  (bits:12; n:384),(bits:12; n:384),                                    // 000000110100x
  (bits:12; n:448),(bits:12; n:448),                                    // 000000110101x
  (bits:13; n:512),                                             // 0000001101100
  (bits:13; n:576),                                             // 0000001101101
  (bits:12; n:53),(bits:12; n:53),                                      // 000000110111x
  (bits:12; n:54),(bits:12; n:54),                                      // 000000111000x
  (bits:13; n:896),                                             // 0000001110010
  (bits:13; n:960),                                             // 0000001110011
  (bits:13; n:1024),                                            // 0000001110100
  (bits:13; n:1088),                                            // 0000001110101
  (bits:13; n:1152),                                            // 0000001110110
  (bits:13; n:1216),                                            // 0000001110111
  (bits:10; n:64),(bits:10; n:64),(bits:10; n:64),(bits:10; n:64),              // 0000001111xxx
    (bits:10; n:64),(bits:10; n:64),(bits:10; n:64),(bits:10; n:64)
);

// 7-12 bit codes (upper 4 bits are 0)
blackTable2: array[0..191] of CCITTCode = (
(bits:8; n:13),(bits:8; n:13),(bits:8; n:13),(bits:8; n:13),                    // 00000100xxxx
    (bits:8; n:13),(bits:8; n:13),(bits:8; n:13),(bits:8; n:13),
    (bits:8; n:13),(bits:8; n:13),(bits:8; n:13),(bits:8; n:13),
    (bits:8; n:13),(bits:8; n:13),(bits:8; n:13),(bits:8; n:13),
  (bits:11; n:23),(bits:11; n:23),                                      // 00000101000x
  (bits:12; n:50),                                              // 000001010010
  (bits:12; n:51),                                              // 000001010011
  (bits:12; n:44),                                              // 000001010100
  (bits:12; n:45),                                              // 000001010101
  (bits:12; n:46),                                              // 000001010110
  (bits:12; n:47),                                              // 000001010111
  (bits:12; n:57),                                              // 000001011000
  (bits:12; n:58),                                              // 000001011001
  (bits:12; n:61),                                              // 000001011010
  (bits:12; n:256),                                             // 000001011011
  (bits:10; n:16),(bits:10; n:16),(bits:10; n:16),(bits:10; n:16),              // 0000010111xx
  (bits:10; n:17),(bits:10; n:17),(bits:10; n:17),(bits:10; n:17),              // 0000011000xx
  (bits:12; n:48),                                              // 000001100100
  (bits:12; n:49),                                              // 000001100101
  (bits:12; n:62),                                              // 000001100110
  (bits:12; n:63),                                              // 000001100111
  (bits:12; n:30),                                              // 000001101000
  (bits:12; n:31),                                              // 000001101001
  (bits:12; n:32),                                              // 000001101010
  (bits:12; n:33),                                              // 000001101011
  (bits:12; n:40),                                              // 000001101100
  (bits:12; n:41),                                              // 000001101101
  (bits:11; n:22),(bits:11; n:22),                                      // 00000110111x
  (bits:8; n:14),(bits:8; n:14),(bits:8; n:14),(bits:8; n:14),                  // 00000111xxxx
    (bits:8; n:14),(bits:8; n:14),(bits:8; n:14),(bits:8; n:14),
    (bits:8; n:14),(bits:8; n:14),(bits:8; n:14),(bits:8; n:14),
    (bits:8; n:14),(bits:8; n:14),(bits:8; n:14),(bits:8; n:14),
  (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),                  // 0000100xxxxx
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
    (bits:7; n:10),(bits:7; n:10),(bits:7; n:10),(bits:7; n:10),
  (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),                  // 0000101xxxxx
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
    (bits:7; n:11),(bits:7; n:11),(bits:7; n:11),(bits:7; n:11),
  (bits:9; n:15),(bits:9; n:15),(bits:9; n:15),(bits:9; n:15),                  // 000011000xxx
    (bits:9; n:15),(bits:9; n:15),(bits:9; n:15),(bits:9; n:15),
  (bits:12; n:128),                                             // 000011001000
  (bits:12; n:192),                                             // 000011001001
  (bits:12; n:26),                                              // 000011001010
  (bits:12; n:27),                                              // 000011001011
  (bits:12; n:28),                                              // 000011001100
  (bits:12; n:29),                                              // 000011001101
  (bits:11; n:19),(bits:11; n:19),                                      // 00001100111x
  (bits:11; n:20),(bits:11; n:20),                                      // 00001101000x
  (bits:12; n:34),                                              // 000011010010
  (bits:12; n:35),                                              // 000011010011
  (bits:12; n:36),                                              // 000011010100
  (bits:12; n:37),                                              // 000011010101
  (bits:12; n:38),                                              // 000011010110
  (bits:12; n:39),                                              // 000011010111
  (bits:11; n:21),(bits:11; n:21),                                      // 00001101100x
  (bits:12; n:42),                                              // 000011011010
  (bits:12; n:43),                                              // 000011011011
  (bits:10; n:0),(bits:10; n:0),(bits:10; n:0),(bits:10; n:0),                  // 0000110111xx
  (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),                  // 0000111xxxxx
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12),
    (bits:7; n:12),(bits:7; n:12),(bits:7; n:12),(bits:7; n:12)
);

// 2-6 bit codes
blackTable3: array[0..63] of CCITTCode = (
  (bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),(bits:-1; n:-1),              // 0000xx
  (bits:6; n:9),                                                // 000100
  (bits:6; n:8),                                                // 000101
  (bits:5; n:7),(bits:5; n:7),                                  // 00011x
  (bits:4; n:6),(bits:4; n:6),(bits:4; n:6),(bits:4; n:6),                      // 0010xx
  (bits:4; n:5),(bits:4; n:5),(bits:4; n:5),(bits:4; n:5),                      // 0011xx
  (bits:3; n:1),(bits:3; n:1),(bits:3; n:1),(bits:3; n:1),                      // 010xxx
    (bits:3; n:1),(bits:3; n:1),(bits:3; n:1),(bits:3; n:1),
  (bits:3; n:4),(bits:3; n:4),(bits:3; n:4),(bits:3; n:4),                      // 011xxx
    (bits:3; n:4),(bits:3; n:4),(bits:3; n:4),(bits:3; n:4),
  (bits:2; n:3),(bits:2; n:3),(bits:2; n:3),(bits:2; n:3),                      // 10xxxx
    (bits:2; n:3),(bits:2; n:3),(bits:2; n:3),(bits:2; n:3),
    (bits:2; n:3),(bits:2; n:3),(bits:2; n:3),(bits:2; n:3),
    (bits:2; n:3),(bits:2; n:3),(bits:2; n:3),(bits:2; n:3),
  (bits:2; n:2),(bits:2; n:2),(bits:2; n:2),(bits:2; n:2),                      // 11xxxx
    (bits:2; n:2),(bits:2; n:2),(bits:2; n:2),(bits:2; n:2),
    (bits:2; n:2),(bits:2; n:2),(bits:2; n:2),(bits:2; n:2),
    (bits:2; n:2),(bits:2; n:2),(bits:2; n:2),(bits:2; n:2)
);

{ TCCITFaxDecompressionStream }

constructor TCCITTFaxDecompressionStream.Create(Source: TStream; Params: TPDFDictionary);
var
  Code1: Word;
begin
  inherited Create(Source);
  K := Params.FindNumber('K',0);
  Columns := Params.FindNumber('Columns');
  NextLine2D := K < 0;
  EofPending := False;
  a0 := 1;
  inputBits := 0;
  row := 0;
  byteAlign := Params.FindNumber('EncodedByteAlign',0) <> 0;
  endOfBlock := Params.FindNumber('EndOfBlock',1) <> 0;
  Black := Params.FindNumber('BlackIs1',0) <> 0;
  Rows := Params.FindNumber('Rows',0);
  SetLength(refLine, Columns+3);
  SetLength(codingLine, Columns+2);
  codingLine[0] := 0;
  codingLine[1] := Columns;
  refLine[2] := Columns;
  buf := -1;
  // skip any initial zero bits and end-of-line marker, and get the 2D
  // encoding tag
  repeat
    Code1 := lookBits(12);
    if Code1 <> 0 then
      Break;
    eatBits(1);
  until False;
  if Code1 = 1 then
    eatBits(12);
  if K > 0 then begin
    nextLine2D := lookBits(1) = 0;
    eatBits(1);
  end;
end;

function TCCITTFaxDecompressionStream.NextByte: Byte;
var
  code1, code2, code3, a0New: Word;
  bits, i: Integer;
  gotEOL: Boolean;
  ret: Byte;
begin
  if EofPending and (codingLine[a0] >= Columns) then begin
    Eof := True;
    Exit;
  end;

  // read the next row
  if codingLine[a0] >= Columns then begin
    // 2-D encoding
    if nextLine2D then begin
      i := 0;
      while codingLine[i] < columns do begin
        refLine[i] := codingLine[i];
        Inc(i);
      end;
      refLine[i + 1] := Columns;
      refLine[i] := Columns;
      b1 := 1;
      a0 := 0;
      a0New := 0;
      codingLine[a0] := 0;
      repeat
        code1 := getTwoDimCode;
        //WriteLn(code1,' ',a0,' ',b1);
        case code1 of
        twoDimPass:
          if refLine[b1] < columns then begin
            a0New := refLine[b1 + 1];
            Inc(b1,2);
          end;
        twoDimHoriz:
        begin
          if (a0 and 1) = 0 then begin
            code1 := 0;
            code2 := 0;
            repeat
              code3 := getWhiteCode;
              Inc(code1,code3);
            until code3 < 64;
            repeat
              code3 := getBlackCode;
              Inc(code2,code3);
            until code3 < 64;
          end else begin
            code1 := 0;
            code2 := 0;
            repeat
              code3 := getBlackCode;
              Inc(code1,code3);
            until code3 < 64;
            repeat
              code3 := getWhiteCode;
              Inc(code2,code3);
            until code3 < 64;
          end;
          if (code1 > 0) or (code2 > 0) then begin
            if a0New + code1 <= columns then
              codingLine[a0 + 1] := a0New + code1
            else
              codingLine[a0 + 1] := columns;
            Inc(a0);
            if codingLine[a0] + code2 <= columns then
              codingLine[a0 + 1] := codingLine[a0] + code2
            else
              codingLine[a0 + 1] := columns;
            Inc(a0);
            a0New := codingLine[a0];
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end;
        end;
        twoDimVert0:
          if refLine[b1] < columns then begin
            Inc(a0);
            a0New := refLine[b1];
            codingLine[a0] := refLine[b1];
            Inc(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end else begin
            Inc(a0);
            codingLine[a0] := Columns;
            a0New := Columns;
          end;
        twoDimVertR1:
          if refLine[b1] + Cardinal(1) < Columns then begin
            a0New := refLine[b1] + 1;
            Inc(a0);
            codingLine[a0] := a0New;
            Inc(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end else begin
            a0New := Columns;
            Inc(a0);
            codingLine[a0] := columns;
          end;
        twoDimVertL1:
          if (refLine[b1] - 1 > a0New) or ((a0 = 0) and (refLine[b1] = 1)) then begin
            a0New := refLine[b1] - 1;
            Inc(a0);
            codingLine[a0] := a0New;
            Dec(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end;
        twoDimVertR2:
          if refLine[b1] + Cardinal(2) < columns then begin
            a0New := refLine[b1] + 2;
            Inc(a0);
            codingLine[a0] := a0New;
            Inc(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end else begin
            Inc(a0);
            a0New := Columns;
            codingLine[a0] := Columns;
          end;
        twoDimVertL2:
          if (refLine[b1] - 2 > a0New) or ((a0 = 0) and (refLine[b1] = 2)) then begin
            a0New := refLine[b1] - 2;
            Inc(a0);
            codingLine[a0] := a0New;
            Dec(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end;
        twoDimVertR3:
          if refLine[b1] + Cardinal(3) < columns then begin
            a0New := refLine[b1] + 3;
            Inc(a0);
            codingLine[a0] := a0New;
            Inc(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end else begin
            a0New := Columns;
            Inc(a0);
            codingLine[a0] := columns;
          end;
        twoDimVertL3:
          if (refLine[b1] - 3 > a0New) or ((a0 = 0) and (refLine[b1] = 3)) then begin
            a0New := refLine[b1] - 3;
            Inc(a0);
            codingLine[a0] := a0New;
            Dec(b1);
            while (refLine[b1] <= a0New) and (refLine[b1] < columns) do
              Inc(b1,2);
          end;
        encEOF:
          begin
            Eof := True;
            Exit;
          end
        else
          raise Exception.CreateFmt('PDF: Bad CCITTFax code %d',[code1]);
        end;
      until codingLine[a0] >= columns;
    end else begin
      // 1-D encoding
      a0 := 0;
      codingLine[a0] := 0;
      while True do begin
        code1 := 0;
        repeat
          code3 := getWhiteCode;
          Inc(code1,code3);
        until code3 < 64;
        codingLine[a0+1] := codingLine[a0] + code1;
        Inc(a0);
        if codingLine[a0] >= columns then
          Break;
        code2 := 0;
        repeat
          code3 := getBlackCode;
          Inc(code2,code3);
        until code3 < 64;
        codingLine[a0+1] := codingLine[a0] + code2;
        Inc(a0);
        if codingLine[a0] >= columns then
          Break;
      end;
    end;

    if codingLine[a0] <> columns then
      raise Exception.CreateFmt('PDF:CCITTFax row is wrong length (%d)', [codingLine[a0]]);

    // byte-align the row
    if byteAlign then
      inputBits := inputBits and not 7;

    // check for end-of-line marker, skipping over any extra zero bits
    gotEOL := False;
    if not endOfBlock and (row = rows - 1) then
      EofPending := True
    else begin
      code1 := lookBits(12);
      while code1 = 0 do begin
        eatBits(1);
        code1 := lookBits(12);
      end;
      if code1 = 1 then begin
        eatBits(12);
        gotEOL := True;
      end else if code1 = encEof then
        EofPending := True;
    end;

    // get 2D encoding tag
    if not EofPending and (K > 0) then begin
      nextLine2D := lookBits(1) = 0;
      eatBits(1);
    end;

    // check for end-of-block marker
    if endOfBlock and gotEOL then begin
      code1 := lookBits(12);
      if code1 = 1 then begin
              eatBits(12);
      if K > 0 then begin
        lookBits(1);
        eatBits(1);
      end;
      if K >= 0 then begin
        for i := 0 to 3 do begin
          code1 := lookBits(12);
          if code1 <> 1 then
            raise Exception.Create('PDF:Bad RTC code in CCITTFax stream');
          eatBits(12);
          if K > 0 then begin
            lookBits(1);
            eatBits(1);
        end;
      end;
          end;
          EofPending := True;
  end;
    end;
    a0 := 0;
    outputBits := codingLine[1] - codingLine[0];
    if outputBits = 0 then begin
      a0 := 1;
      outputBits := codingLine[2] - codingLine[1];
    end;
    Inc(row);
  end;

  // get a byte
  if outputBits >= 8 then begin
    if (a0 and 1) = 0 then
      ret := $ff
    else
      ret := 0;
    Dec(outputBits,8);
    if outputBits = 0 then begin
      Inc(a0);
      if codingLine[a0] < columns then
        outputBits := codingLine[a0 + 1] - codingLine[a0];
    end
  end else begin
    bits := 8;
    ret := 0;
    repeat
      if outputBits > bits then begin
        i := bits;
        bits := 0;
        if (a0 and 1) = 0 then
          ret := ret or ($ff shr (8 - i));
        Dec(outputBits,i);
      end else begin
        i := outputBits;
        Dec(bits,outputBits);
        if (a0 and 1) = 0 then
          ret := ret or (($ff shr (8 - i)) shl bits);
        outputBits := 0;
        Inc(a0);
        if codingLine[a0] < columns then
          outputBits := codingLine[a0 + 1] - codingLine[a0];
      end
    until not ((bits > 0) and (codingLine[a0] < columns));
  end;
  if black then
    buf := ret xor $ff
  else
    buf := ret;
  Result := buf;
  //WriteLn(IntToHex(buf,2));
end;

function TCCITTFaxDecompressionStream.getTwoDimCode: Word;
var
  Code: Word;
  P: PCCITTCode;
  n: Integer;
begin
  if endOfBlock then begin
    code := lookBits(7);
    p := @twoDimTable[code];
    if p^.bits > 0 then begin
      eatBits(p^.bits);
      Result := p^.n;
      Exit;
    end;
  end else
    for n := 1 to 7 do begin
      code := lookBits(n);
      if n < 7 then
              code := code shl (7 - n);
      //WriteLn(Code);
      p := @twoDimTable[code];
      if p^.bits = n then begin
        eatBits(n);
        Result := p^.n;
        Exit;
      end;
    end;
  raise Exception.CreateFmt('PDF:Bad two dim code (%04x) in CCITTFax stream', [code]);
end;

function TCCITTFaxDecompressionStream.getWhiteCode: Word;
var
  Code: Word;
  P: PCCITTCode;
  n: Integer;
begin
  if endOfBlock then begin
    code := lookBits(12);
    if (code shr 5) = 0 then
      p := @whiteTable1[code]
    else
      p := @whiteTable2[code shr 3];
    if p^.bits > 0 then begin
      eatBits(p^.bits);
      Result := p^.n;
      Exit;
    end;
  end else begin
    for n := 1 to 9 do begin
      code := lookBits(n);
      if n < 9 then
        code := code shl (9 - n);
      p := @whiteTable2[code];
      if p^.bits = n then begin
        eatBits(n);
        Result := p^.n;
        Exit;
      end;
    end;
    for n := 11 to 12 do begin
      code := lookBits(n);
      if n < 12 then
        code := code shl (12 - n);
      p := @whiteTable1[code];
      if p^.bits = n then begin
        eatBits(n);
        Result := p^.n;
        Exit;
      end;
    end;
  end;
  raise Exception.CreateFmt('PDF:Bad white code (%04x) in CCITTFax stream', [code]);
end;

function TCCITTFaxDecompressionStream.getBlackCode: Word;
var
  Code: Word;
  P: PCCITTCode;
  n: Integer;
begin
  if endOfBlock then begin
    code := lookBits(13);
    if (code shr 7) = 0 then
      P := @blackTable1[code]
    else if (code shr 9) = 0 then
      p := @blackTable2[(code shr 1) - 64]
    else
      p := @blackTable3[code shr 7];
    if p^.bits > 0 then begin
      eatBits(p^.bits);
      Result := p^.n;
      Exit;
    end;
  end else begin
    for n := 2 to 6 do begin
      code := lookBits(n);
      if n < 6 then
              code := code shl (6 - n);
      P := @blackTabLE3[code];
      if p^.bits = n then begin
              eatBits(n);
        Result := p^.n;
        Exit;
      end;
    end;
    for n := 7 to 12 do begin
      code := lookBits(n);
      if n < 12 then
              code := code shl (12 - n);
      if code >= 64 then begin
              P := @blackTable2[code - 64];
        if p^.bits = n then begin
                eatBits(n);
          Result := p^.n;
          Exit;
        end;
      end;
    end;
    for n := 10 to 13 do begin
      code := lookBits(n);
      if n < 13 then
              code := code shl (13 - n);
      p := @blackTable1[code];
      if p^.bits = n then begin
              eatBits(n);
        Result := p^.n;
        Exit;
      end;
    end;
  end;
  raise Exception.CreateFmt('PDF:Bad black code (%04x) in CCITTFax stream', [code]);
end;

function TCCITTFaxDecompressionStream.lookBits(n: Integer): Word;
var
  c: Byte;
begin
  while inputBits < n do begin
    if Stream.Read(c,1) <> 1 then begin
      if inputBits = 0 then
        Result := encEof
      else
        // near the end of the stream, the caller may ask for more bits
        // than are available, but there may still be a valid code in
        // however many bits are available -- we need to return correct
        // data in this case
        Result := (inputBuf shl (n - inputBits)) and ($ffff shr (16 - n));
      Exit;
    end;
    inputBuf := (inputBuf shl 8) + c;
    Inc(inputBits,8);
  end;
  Result := (inputBuf shr (inputBits - n)) and ($ffff shr (16 - n));
end;

procedure TCCITTFaxDecompressionStream.eatBits(n: Integer);
begin
  inputBits := Max(inputBits-n,0);
end;


end.
