unit PeakFitFrm;
{$I utils.inc}{Copyright (C) 2005-14 C M Western; see commandline.pas for license}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, NGrid, StdCtrls, ExtCtrls, ChildForm, MNode, ComCtrls,
  Overlay, Buttons, AlEdit, EditExtr, NewSpin, Menus, Settings, Units, Peaks,
  Gendlg;


type
  TPeakFitForm = class(TChildForm)
    Panel: TPanel;
    Grid: TNumberGrid;
    CopyButton: TSpeedButton;
    DeleteButton: TSpeedButton;
    DeleteUnassigned: TButton;
    NLfitButton: TSpeedButton;
    Label1: TLabel;
    FitButton: TSpeedButton;
    Label2: TLabel;
    UndoFitButton: TSpeedButton;
    OtherButton: TButton;
    OtherPopupMenu: TPopupMenu;
    ClearItem: TMenuItem;
    ResetItem: TMenuItem;
    PolyOrder: TSpinBox;
    ActiveObject: TComboBox;
    AvgErrEdit: TAlignedEdit;
    SplineFitItem: TMenuItem;
    procedure OtherButtonClick(Sender: TObject);
    procedure UndoFitButtonClick(Sender: TObject);
    procedure GridMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure DeleteUnassignedClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
    procedure FormHide(Sender: TObject);
  protected
    FChanging: Boolean;
    DisableReplot: Boolean;
    LastCell: TGridCoord;
    DataSource: TInteractiveDataSource;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    procedure HandleControl(Control: TControl; var PName, Value, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer); override;
    procedure PeakChanged(PeakNumber:Integer); virtual;
    procedure PeakMeasured(Source: TMNode; Index: Integer); virtual; abstract;
  public
    AvgErr: Double;
    procedure StuffSelector(MNode: TMNode); override;
    procedure UpdateStatus;
    procedure ArrangeForm(Screen: TRect); override;
    procedure ShowPeak(AIndex: Integer); virtual; abstract;
    procedure ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent = nil); override;
    function Peaks: TPeaks; virtual; abstract;
    function HasDataSource:Boolean; virtual; abstract;
  end;

implementation

uses {$IFDEF LCL_} LResources, Types, {$ENDIF}
     Global, Expr, Math, MathSup, Main, ExpPlot, PlotDataSource, ApplyFrm,
     ResidFrm, WCompat, DataFitter;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TPeakFitForm }

procedure TPeakFitForm.ClearButtonClick(Sender: TObject);
begin
  Peaks.Clear;
  MainForm.ConsiderReplot(False,False);
end;

procedure TPeakFitForm.CopyButtonClick(Sender: TObject);
begin
  Grid.CopyToClipboard;
end;

procedure TPeakFitForm.DeleteButtonClick(Sender: TObject);
var
  d, i, j: Integer;
  yPlot: TPlot;
begin
  Grid.BeginUpdate;
  DisableReplot := True;
  try
    d := Grid.Selection.Top;
    for i := d to Grid.Selection.Bottom do
      with Peaks do begin
        if not IsNanOrInf(Peaks[d-1].Actual) then begin
          { Delete matching peak in calibration channel, if present }
          yPlot := MainForm.LastPlot(True);
          if Assigned(yPlot) and (yPlot <> OwnerPlot) and not yPlot.Hide then
            for j := yPlot.nPeaks-1 downto 0 do
              if SameValue(Convert(yPlot.Peaks.Peaks[j].Position,yPlot.Units,(OwnerPlot as TPlot).Units), Peaks[d-1].Actual) then begin
                yPlot.Peaks.DeletePeak(j);
                Break;
              end;
        end;
        DeletePeak(d-1);
      end;
  finally
    DisableReplot := False;
    Grid.EndUpdate;
  end;
  MainForm.ConsiderReplot(False,False);
end;

procedure TPeakFitForm.DeleteUnassignedClick(Sender: TObject);
var
  i: Integer;
begin
  Grid.BeginUpdate;
  DisableReplot := True;
  try
    with Peaks do
      for i := nPeaks-1 downto 0 do
        if not Valid(i) then
          DeletePeak(i);
  finally
    DisableReplot := False;
    Grid.EndUpdate;
  end;
  MainForm.ConsiderReplot(False,False);
end;

procedure TPeakFitForm.FormHide(Sender: TObject);
begin
  Grid.EditorMode := False;
  MainForm.ConsiderReplot(False,False);
end;

procedure TPeakFitForm.GridChangeCellQuery(Sender: TObject; Col, Row: Integer; var CanChange: Boolean);
begin
  CanChange := CanChange and (Col <> 3);
end;

procedure TPeakFitForm.GridMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  LastCell := Grid.MouseCoord(X,Y);
end;

procedure TPeakFitForm.HandleControl(Control: TControl; var PName, Value,
  PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
begin
  if (Control = ActiveObject) or (Control = PolyOrder) then
    inherited;
end;

procedure TPeakFitForm.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  inherited;
  if not FChanging and Visible  then
    case Change of
      mnPeakChanged: if Sender = Peaks.Owner then PeakChanged(Index);
      mnPeakMeasured:
        if (Sender = Peaks.Owner) or (Sender = Peaks.Owner.Owner) or (Sender.Owner = Peaks.Owner.Owner) then
          PeakMeasured(Sender, Index);
    end;
end;

procedure TPeakFitForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin

end;

procedure TPeakFitForm.PeakChanged(PeakNumber: Integer);
begin
  if (PeakNumber < 0) or not Grid.Enabled or (Grid.RowCount <> Peaks.nPeaks+1) then
    ShowItem(CurrentMNode)
  else
    ShowPeak(PeakNumber);
end;

procedure TPeakFitForm.SettingChanged(MNode: TMNode; Index: Integer);
begin

end;

procedure TPeakFitForm.ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent);
begin
  if TopItem = OtherPopupMenu then
    TopItem := Self;
  inherited;
end;

procedure TPeakFitForm.ArrangeForm(Screen: TRect);
var
  R, MainRect: TRect;
begin
  inherited;
  R := ScreenBounds(Self);
  MainRect := ScreenBounds(MainForm);
  OffsetRect(R,MainRect.Right-R.Left,MainRect.Top-R.Top);
  if R.Right > Screen.Right then
    OffsetRect(R,Screen.Right-R.Right,0);
  if R.Bottom > Screen.Bottom then
    OffsetRect(R,0,Screen.Bottom-R.Bottom);
  SetScreenBounds(Self,R);
end;

procedure TPeakFitForm.OtherButtonClick(Sender: TObject);
begin
  OtherPopupMenu.PopupComponent := OtherButton;
  OtherPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

procedure TPeakFitForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TPeakFitForm.UndoFitButtonClick(Sender: TObject);
begin
  if HasDataSource then begin
    DataSource.NLFitDlg.UndoClick(nil);
    UpdateStatus;
  end;
end;

procedure TPeakFitForm.UpdateStatus;
begin
  if not Visible or (csDestroying in Application.ComponentState) then
    Exit;
  if HasDataSource then
    with DataSource do begin
      UndoFitButton.Enabled := NLFitDlg.Undo.Enabled;
      if IsNan(AvgErr) then
        AvgErrEdit.Text := ''
      else
        AvgErrEdit.Text := DtosM(AvgErr, 8, -1);
    end
  else begin
    UndoFitButton.Enabled := False;
    AvgErrEdit.Text := '';
  end;
end;

end.
