unit SettingsForm;
{$I utils.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  Classes, Controls, WhatsThis, Settings, MNode, pgDOM, Mixture, StdCtrls;

type

  { TSettingsForm }


  TOnControl = procedure(Control: TControl; PName, Value, PHint: string; SettingsType: TSettingType; Data: TObject) of object;

  TSaveSettingsInfo = record
    DOMDocument: TDOMDocument;
    SavePositions: Boolean;
    SaveSettings: Boolean;
  end;
  PSaveSettingsInfo = ^TSaveSettingsInfo;

  TSettingsForm = class(THelpForm)
  private
    FFormSettings, FDefaultFormSettings: TFormSettings;
    FPrepared: Boolean;
    function GetFormSettings: TFormSettings;
    procedure SetDefaultFormSettings(const AValue: TFormSettings);
    procedure ShowInfo(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
  protected
    FComboBoxUpdated: Boolean;
    procedure ForAllControls(OnControl: TOnControl; Data: TObject);
    procedure HandleControls(Control: TWinControl; OnControl: TOnControl; Data: TObject);
    procedure AddValue(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
    procedure SetValue(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
    procedure HandleControl(Control: TControl; var PName, AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); virtual;
    function MakeFormSettings(Defaults: TFormSettings; AOwner: TComponent = nil): TFormSettings;
    procedure AddExtraVariables(AFormSettings: TFormSettings); virtual;
    procedure WriteExtraVariables(AFormSettings: TFormSettings); virtual;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); virtual;
    function  SettingsAvailable: Boolean;
    procedure Resize; override;
    function FinishLoading(Data: TObject):Boolean;
    function ClearSettings(Data: TObject):Boolean;
    function ShowInMixture(Data: TObject):Boolean;
    procedure WriteSettings(AFormSettings: TFormSettings; SavePositions: Boolean; SaveSettings: Boolean); virtual;
    function SaveSettings(Data: TObject):Boolean;
    procedure LoadSettings(AFormSettings: TFormSettings; DoPosition, DoSettings: Boolean);
    procedure UpdateComboBox(ComboBox: TComboBox; CurrentObject: TObject = nil);
    procedure SaveComboBox(ComboBox: TComboBox);
    function DoUpdateActions(Data: TObject):Boolean;
    function FormSettingsName: string; virtual;

    { Provide partial implementation of MNode parameters and settings functions }
    function GetParameter(AIndex: Integer): Double;
    procedure SetParameter(AIndex: Integer; const Value: Double); virtual;
    function GetSetting(AIndex: Integer): Integer; virtual;
    procedure SetSetting(AIndex: Integer; const AValue: Integer); virtual;
    function GetFlag(AIndex: Integer): Boolean;
    procedure SetFlag(AIndex: Integer; const AValue: Boolean);
    function SettingToStr(AIndex: Integer): string; virtual;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); virtual;
    function GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar; virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function LoadControls:Boolean;
    property FormSettings: TFormSettings read GetFormSettings;
    property DefaultFormSettings: TFormSettings read FDefaultFormSettings write SetDefaultFormSettings;
  end;

  TOnForm = function (Data: TObject):Boolean of object;

function ForAllSettingsForms(OnForm: TOnForm; Data: TObject):Boolean;

implementation

uses {$IFDEF FPC} EditBtn, {$ENDIF}
     PropHelp, Buttons, Worker, Split, SysUtils, Global, Forms, Math, MathSup,
     NewSpin, QuantumSpin, PlotFrm;

{ TSettingsForm }

procedure TSettingsForm.AddExtraVariables(AFormSettings: TFormSettings);
begin
  { None by default }
end;

procedure TSettingsForm.AddValue(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
var
  RValue: Double;
  Index: Integer;
begin
  with Data as TFormSettings do
    if SettingType = stDouble then begin
      if not TryStrToFloatEx(Value, RValue) then
        SetNan(RValue);
      if FPrepared then
        SetParameter(IndexOfParameter(PName),RValue)
      else begin
        Index := AddParameter(PName, RValue);
        ParametersHints[Index] := PHint;
      end;
    end else if FPrepared then
      SetSettingFromStr(IndexOfSetting(PName), Value)
    else begin
      Index := AddSetting(PName, SettingType, Value);
      SettingsHints[Index] := PHint;
    end;
end;

procedure TSettingsForm.SetValue(Control: TControl; PName, Value, PHint: string;
  SettingType: TSettingType; Data: TObject);
var
  Index: Integer;
  NewValue: string;
  Unused: Boolean;
begin
  with Data as TFormSettings do begin
    Index := IndexOfParameter(PName);
    if Index > 0 then
      NewValue := FloatToStr(Parameters[Index])
    else begin
      Index := IndexOfSetting(PName);
      if Index > 0 then
        NewValue := SettingToStr(Index)
      else
        Exit; { No matching parameter }
    end;
  end;
  Unused := False; { Unused }
  if Value <> NewValue then
    HandleControl(Control, PName, NewValue, PHint, Unused, SettingType, True);
end;

function TSettingsForm.ClearSettings(Data: TObject): Boolean;
begin
  FreeAndNil(FFormSettings);
  Result := True;
end;

constructor TSettingsForm.Create(AOwner: TComponent);
begin
  inherited;
  SetDefaultFormSettings(MakeFormSettings(nil));
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
end;

destructor TSettingsForm.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  FreeAndNil(FFormSettings);
  FreeAndNil(FDefaultFormSettings);
  inherited;
end;

function TSettingsForm.SaveSettings(Data: TObject): Boolean;
var
  Node: TDOMElement;
  SaveSettingsInfo: TSaveSettingsInfo;
begin
  SaveSettingsInfo := PSaveSettingsInfo(Data)^;
  WriteSettings(FormSettings,SaveSettingsInfo.SavePositions,SaveSettingsInfo.SaveSettings);
  if not Assigned(FormSettings.Owner) and (SaveSettingsInfo.SavePositions or SaveSettingsInfo.SaveSettings) then begin
    Node := FormSettings.WriteProperties(SaveSettingsInfo.DOMDocument,False);
    { Delete nodes with no data }
    if not Assigned(Node.FirstChild) and (Node.Attributes.Count <= 1) then
      SaveSettingsInfo.DOMDocument.RemoveChild(Node);
  end;
  Result := True;
end;

function TSettingsForm.SettingsAvailable: Boolean;
begin
  Result := FPrepared and Assigned(FFormSettings);
end;

procedure TSettingsForm.ShowInfo(Control: TControl; PName, Value,
  PHint: string; SettingType: TSettingType; Data: TObject);
begin
  Write(Log, Control.ClassName:20, ': ', PName:20, ' ');
  case SettingType of
    stInteger: Write(Log, 'I ');
    stDouble:  Write(Log, 'D ');
    stString:  Write(Log, 'S ');
  end;
  WriteLn(Log, Value:25, ' ', PHint);
end;

function TSettingsForm.ShowInMixture(Data: TObject): Boolean;
var
  OldFormSettings: TFormSettings;
begin
  if not Assigned(FFormSettings) or not Assigned(FFormSettings.Owner) then begin
    OldFormSettings := FFormSettings;
    FFormSettings := MakeFormSettings(OldFormSettings,GlobalMixture);
    FreeAndNil(OldFormSettings);
  end else
    WriteSettings(FormSettings, True, True);
  Result := True;
end;

function TSettingsForm.DoUpdateActions(Data: TObject):Boolean;
begin
  UpdateActions;
  Result := True;
end;

function ComboBoxValue(ComboBox: TComboBox):string;
begin
  with ComboBox do
    if ItemIndex = 0 then
      Result := '' { Null string implies the first possible item }
    else if ItemIndex > 0 then
      Result := Items[ItemIndex]
    else
      Result := Text
end;

procedure TSettingsForm.SaveComboBox(ComboBox: TComboBox);
var
  i: Integer;
begin
  i := FormSettings.IndexOfSetting(ComboBox.Name);
  if i > 0 then begin
    FormSettings.SetSettingFromStr(i, ComboBoxValue(ComboBox));
    FComboBoxUpdated := True;
  end;
end;

procedure TSettingsForm.UpdateComboBox(ComboBox: TComboBox; CurrentObject: TObject);
var
  i, NewIndex: Integer;
  s: string;
begin
  if ComboBox.Items.Count > 0 then begin
    NewIndex := 0;
    if FComboBoxUpdated then begin
      i := FormSettings.IndexOfSetting(ComboBox.Name);
      if i > 0 then begin
        s := FormSettings.SettingToStr(i);
        for i := 0 to ComboBox.Items.Count - 1 do
          if ComboBox.Items[i] = s then begin
            NewIndex := i;
            Break;
          end;
      end;
    end else if Assigned(CurrentObject) then
      for i := 0 to ComboBox.Items.Count - 1 do
        if ComboBox.Items.Objects[i] = CurrentObject then begin
          NewIndex := i;
          Break;
        end;
    ComboBox.ItemIndex := NewIndex;
  end;
end;

procedure TSettingsForm.WriteExtraVariables(AFormSettings: TFormSettings);
begin
  { None by default }
end;

procedure TSettingsForm.WriteSettings(AFormSettings: TFormSettings; SavePositions, SaveSettings: Boolean);
var
  i: Integer;
begin
  if not SaveSettings then begin
    { Reset all to defaults }
    AFormSettings.Comment := '';
    for i := 1 to AFormSettings.nSettings do
      AFormSettings.SetSettingFromStr(i,FDefaultFormSettings.SettingToStr(i));
    AFormSettings.GetParameters(FDefaultFormSettings);
  end;
  if not (Self is TPlotForm) then
    AddValue(nil, 'Visible', BoolToStr(SavePositions and Visible,True), 'Is Window Visible', stString, AFormSettings);
  if SaveSettings then begin
    ForAllControls({$IFDEF FPC}@{$ENDIF}AddValue, AFormSettings);
    AddExtraVariables(AFormSettings);
  end;
  AFormSettings.LoadFromForm(Self, SavePositions);
end;

procedure TSettingsForm.ForAllControls(OnControl: TOnControl; Data: TObject);
begin
  HandleControls(Self, OnControl, Data);
end;

function TSettingsForm.FormSettingsName: string;
begin
  Result := Copy(ClassName, 2, Length(ClassName)-5)+'Window';
end;

function TSettingsForm.GetFormSettings: TFormSettings;
begin
  if not Assigned(FFormSettings) then
    FFormSettings := MakeFormSettings(FDefaultFormSettings);
  Result := FFormSettings;
end;

procedure TSettingsForm.SetDefaultFormSettings(const AValue: TFormSettings);
begin
  FreeAndNil(FDefaultFormSettings);
  FDefaultFormSettings := AValue;
end;

function TSettingsForm.GetFlag(AIndex: Integer): Boolean;
begin
  Result := StrToBoolean(FormSettings.SettingToStr(AIndex));
end;

procedure TSettingsForm.SetFlag(AIndex: Integer; const AValue: Boolean);
begin
  if AValue then
    FormSettings.SetSettingFromStr(AIndex,'True')
  else
    FormSettings.SetSettingFromStr(AIndex,'False')
end;

function TSettingsForm.SettingToStr(AIndex: Integer): string;
begin
  Result := FormSettings.SettingToStr(AIndex);
end;

procedure TSettingsForm.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  FormSettings.SetSettingFromStr(AIndex,Str);
end;

function TSettingsForm.GetSettingsNames(AIndex: Integer; out LowestValue: Integer): PPChar;
begin
  LowestValue := 0;
  Result := nil;
end;

function TSettingsForm.GetParameter(AIndex: Integer): Double;
begin
  Result := FormSettings.GetParameter(AIndex);
end;

procedure TSettingsForm.SetParameter(AIndex: Integer; const Value: Double);
begin
  FormSettings.SetParameter(AIndex,Value);
end;

function TSettingsForm.GetSetting(AIndex: Integer): Integer;
var
  s: string;
  LowestValue, i: Integer;
  SettingsNames: PPChar;
begin
  s := SettingToStr(AIndex);
  SettingsNames := GetSettingsNames(AIndex,LowestValue);
  if Assigned(SettingsNames) then begin
    i := 0;
    while SettingsNames^ <> nil do
      if CompareText(SettingsNames^, s) = 0 then begin
        Result := i+LowestValue;
        Exit;
      end else begin
        Inc(i);
        Inc(SettingsNames);
      end;
    Result := LowestValue; { Default if not matched }
  end else
    Result := StrToInt(s);
end;

procedure TSettingsForm.SetSetting(AIndex: Integer; const AValue: Integer);
var
  LowestValue: Integer;
  SettingsNames: PPChar;
begin
  SettingsNames := GetSettingsNames(AIndex,LowestValue);
  if Assigned(SettingsNames) then begin
    Inc(SettingsNames, GetSetting(AIndex)-LowestValue);
    SetSettingFromStr(AIndex,SettingsNames^);
  end else
    FormSettings.SetSetting(AIndex,AValue);
end;

procedure TSettingsForm.HandleControl(Control: TControl;
  var PName, AValue, PHint: string;
  var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
var
  i: Integer;
begin
  Save := True;
  if Control is TQuantumSpin then
    { Quantum numbers are always written as twice actual values;
      Handle the special case that if the string 'all' appears this is
      assumed to correspond to the minimum value }
    with TQuantumSpin(Control) do
      if Write then
        if AValue = 'all' then
          Value := MinValue
        else
          Value := StrToInt(AValue)
      else begin
        AValue := Text;
        if CheckText(AValue,False) then begin
          if Assigned(OnSetIntegerValue) then
            OnSetIntegerValue(Control, Value, AValue);
          if AValue <> 'all' then
            AValue := IntToStr(Value)
        end;
      end
  else if Control is TCustomEdit then
    if Write then
      TCustomEdit(Control).Text := AValue
    else begin
      AValue := TCustomEdit(Control).Text;
      { Just save and restore text, even for TSpinBox, to allow
        for differing interpretations of text. The initial test forces
        the box to display a string value if it can }
      if Control is TSpinBox then
        with TSpinBox(Control) do
          if Assigned(OnSetIntegerValue) and CheckText(Text,False) then
            OnSetIntegerValue(Control, Value, AValue);
    end
  {$IFDEF FPC}
  else if Control is TFileNameEdit then
    if Write then begin
      TFileNameEdit(Control).FileName := AValue
    end else
      AValue := TFileNameEdit(Control).FileName
  else if Control is TCustomEditButton then
    { Filename edit in log window, for example; descends from TCustomEdit in Delphi }
    { On Carbon this seems to not trigger the correct update events, hence use the
      more specific version above }
    if Write then begin
      TCustomEditButton(Control).Text := AValue
    end else
      AValue := TCustomEditButton(Control).Text
  {$ENDIF}
  else if Control is TCheckBox then
    if Write then
      TCheckBox(Control).Checked := StrToBool(AValue)
    else
      AValue := BoolToStr(TCheckBox(Control).Checked,True)
  else if Control is TComboBox then
    with TComboBox(Control) do
      if Write then begin
        if AValue = '' then { Null string implies first item }
          ItemIndex := 0;
        for i := 0 to Items.Count - 1 do begin
          if Items[i] = AValue then begin
            ItemIndex := i;
            Break;
          end
        end;
        FComboBoxUpdated := True;
      end else begin
        { If form not visible, combobox selection will not be set,
          so check for previously restored value }
        if not (Owner as TCustomForm).Visible and FComboBoxUpdated and Assigned(FFormSettings) then begin
          i := FormSettings.IndexOfSetting(PName);
          if i > 0 then begin
            AValue := FormSettings.SettingToStr(i);
            Exit;
          end;
        end;
        AValue := ComboBoxValue(Control as TComboBox);
      end
  else if Control is TSplit then begin
//    if Write then
//      { TODO }
//    else begin
//      with TSplit(Control) do
//        if TopLeftWindow.Visible then
//          if BottomRightWindow.Visible then
//            if Cursor = crHSplit then
//              RValue := TopLeftWindow.Width*100/(TopLeftWindow.Width+BottomRightWindow.Width)
//            else
//              RValue := TopLeftWindow.Height*100/(TopLeftWindow.Height+BottomRightWindow.Height)
//          else
//            RValue := 100
//        else
//          RValue := 0;
//      AValue := FloatToStr(RValue);
//    end;
//    SettingType := stDouble;
//    Save := True;
    Save := False; { Do not handle for the time being }
  end else if Control is TSpeedButton then
    with TSpeedButton(Control) do
      if GroupIndex <> 0 then begin
        if Write then
          Down := StrToBool(AValue)
        else
          AValue := BoolToStr(Down,True);
        Save := True;
      end else
        Save := False
  else if Control is TRadioButton then
    with TRadioButton(Control) do
      if Write then
        Checked := StrToBool(AValue)
      else
        AValue := BoolToStr(Checked,True)
  else
    Save := False;
end;

procedure TSettingsForm.HandleControls(Control: TWinControl; OnControl: TOnControl; Data: TObject);
var
  i: Integer;
  PName, PHint: string;
  Value: string;
  SettingType: TSettingType;
  Save: Boolean;
begin
  with Control do
    for i := 0 to ControlCount-1 do begin
      PName := Controls[i].Name;
      if PName = '' then
        Continue;
      Value := '';
      PHint := Controls[i].Hint;
      Save := False;
      SettingType := stString;
      HandleControl(Controls[i],PName, Value, PHint, Save, SettingType, False);
      if Save then
        OnControl(Controls[i],PName, Value, PHint, SettingType, Data);
      if Controls[i] is TWinControl then
        HandleControls(TWinControl(Controls[i]),OnControl,Data);
    end;
end;

function TSettingsForm.LoadControls: Boolean;
begin
  WriteLn(Log);
  WriteLn(Log);
  WriteLn(Log, Self.ClassName);
  ForAllControls({$IFDEF FPC}@{$ENDIF}ShowInfo, nil);
  Result := True;
end;

procedure TSettingsForm.LoadSettings(AFormSettings: TFormSettings; DoPosition, DoSettings: Boolean);
var
  Index: Integer;
  {$IFDEF VCL_}
  SavedDefaultMonitor: TDefaultMonitor;
  {$ENDIF}
begin
  if DoSettings then begin
    ForAllControls({$IFDEF FPC}@{$ENDIF}SetValue, AFormSettings);
    WriteExtraVariables(AFormSettings);
  end;
  if DoPosition then begin
    AFormSettings.RestoreForm(Self);
    if not (Self is TPlotForm) then begin
      Index := AFormSettings.IndexOfSetting('Visible');
      {$IFDEF VCL_}
      SavedDefaultMonitor := DefaultMonitor;
      {$ENDIF}
      if Index > 0 then
        try
          { Temporarily defeat monitor override mechanism }
          {$IFDEF VCL_}
          DefaultMonitor := dmDesktop;
          {$ENDIF}
          Visible := StrToBool(AFormSettings.SettingToStr(Index));
        finally
          {$IFDEF VCL_}
          DefaultMonitor := SavedDefaultMonitor;
          {$ENDIF}
        end;
    end;
  end;
end;

function TSettingsForm.MakeFormSettings(Defaults: TFormSettings; AOwner: TComponent): TFormSettings;
begin
  FPrepared := False;
  Result := TFormSettings.Create(AOwner);
  Result.Name := FormSettingsName;
  WriteSettings(Result, True, True);
  FPrepared := True;
  if Assigned(Defaults) then
    Result.GetAll(Defaults);
end;

procedure TSettingsForm.MNodeChange(Sender: TMNode; Change: TMNodeChange;
  Index: Integer);
begin
  if Change = mnDestroy then begin
    if Sender = FFormSettings then
      FFormSettings := nil;
    if Sender  = FDefaultFormSettings then
      FDefaultFormSettings := nil;
  end;
end;

procedure TSettingsForm.Resize;
begin
  inherited;
  if SettingsAvailable then
    FormSettings.LoadFromForm(Self, True)
end;

function ForAllSettingsForms(OnForm: TOnForm; Data: TObject):Boolean;
var
  i: Integer;
begin
  Result := False;
  if not Assigned(Application.MainForm) then
    Exit; { Application closing }
  { Always do the main form first }
  TMethod(OnForm).Data := Application.MainForm;
  if not OnForm(Data) then
    Exit;
  for i := 0 to Screen.CustomFormCount-1 do
    if (Screen.CustomForms[i] <> Application.MainForm)
       and (Screen.CustomForms[i] is TSettingsForm) then begin
        TMethod(OnForm).Data := Screen.CustomForms[i];
        if not OnForm(Data) then
          Exit;
    end;
  Result := True;
end;

function TSettingsForm.FinishLoading(Data: TObject): Boolean;
var
  SaveSettingsInfo: TSaveSettingsInfo;
  Node: TDOMElement;
begin
  Result := True;
  if Assigned(FFormSettings) then
    Exit;
  FFormSettings := MakeFormSettings(FDefaultFormSettings);
  SaveSettingsInfo := PSaveSettingsInfo(Data)^;
  if Assigned(SaveSettingsInfo.DOMDocument) then begin
    Node := SaveSettingsInfo.DOMDocument.FirstChild as TDOMElement;
    while Assigned(Node) do
      if ( CompareText(Node.NodeName, 'FormSettings') = 0 )
         and ( CompareText(Node.GetAttribute('Name'),FormSettingsName) = 0) then begin
           FFormSettings.ReadProperties(Node, False);
           Break;
         end
      else
        Node := Node.NextSibling as TDOMElement;
  end;
  LoadSettings(FFormSettings, SaveSettingsInfo.SavePositions, SaveSettingsInfo.SaveSettings);
end;

end.

