unit SymmetryCoordsFrm;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses SysUtils, Classes, Graphics, Controls, Forms, Dialogs, VibGridFrm,
     ExtCtrls, Grids, Menus, NGrid, Split, MNode, InternalCoords, NewSpin,
  ComCtrls, StdCtrls, AlEdit, EditExtr, ScrlCtrl, GraphI, GraphM, Graphs;

type
  TSymmetryCoordsForm = class(TVibGridForm)
    CheckItem: TMenuItem;
    ApplyItem: TMenuItem;
    ShowFMatrixItem: TMenuItem;
    MakeFMatrixItem: TMenuItem;
    SortFMatrixbyValueItem: TMenuItem;
    SortFMatrixbyModeItem: TMenuItem;
    MakeExtraOperatorsItem: TMenuItem;
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
    procedure FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var ss: string; var IsText: Boolean);
    procedure FGridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
    procedure CheckItemClick(Sender: TObject);
    procedure ApplyItemClick(Sender: TObject);
    procedure ShowFMatrixItemClick(Sender: TObject);
    procedure MakeFMatrixItemClick(Sender: TObject);
    procedure SortFMatrixbyValueItemClick(Sender: TObject);
    procedure SortFMatrixbyModeItemClick(Sender: TObject);
    procedure MakeExtraOperatorsItemClick(Sender: TObject);
  private
    S: TSymmetryCoords;
  protected
    procedure SetActiveObject(const Value: TMNode); override;
  public
    procedure PopulateForm(ACol:Integer); override;
    procedure Replot(ACol:Integer = -1); override;
  end;

function ConnectSymmetryCoordsForm(SymmetryCoords: TSymmetryCoords): TComponent;

implementation

uses Mixture, Global, Math, MatrixVec, Vib, ChildForm, VibPlotFrm;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectSymmetryCoordsForm(SymmetryCoords: TSymmetryCoords): TComponent;
begin
  Result := ConnectVibFrm(SymmetryCoords, TSymmetryCoordsForm);
end;

{ TSymmetryCoordsForm }

procedure TSymmetryCoordsForm.ApplyItemClick(Sender: TObject);
begin
  S.Apply(True);
  PopulateForm;
end;

procedure TSymmetryCoordsForm.CheckItemClick(Sender: TObject);
begin
  S.MakeLMatrix(S.Owner as TElectronicState, False, False, False, True);
  PopulateForm;
end;

procedure TSymmetryCoordsForm.FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var ss: string; var IsText: Boolean);
begin
  if (Col = 0) and (Row = 0) then
    ss := 'F'
  else
    inherited;
end;

procedure TSymmetryCoordsForm.FGridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
var
  i, j, n: Integer;
begin
  if (ARow > 0) and (ACol > 0) and Assigned(S) and Assigned(S.F) and (Length(S.AF) = 0) then begin
    { Writing to matrix - copy default if present }
    SetLength(S.AF,1); { Prevent re-entry }
    { Diagonal elements first }
    for i := 1 to S.F.RowCount do begin
      FGridSetEditText(Sender, 1, i, FloatToStr(S.F[i,i]));
      FGridSetEditText(Sender, 2, i, IntToStr(i));
      FGridSetEditText(Sender, 3, i, IntToStr(i));
    end;
    { Off-diagonal }
    n := S.F.RowCount+1;
    for i := 1 to S.F.RowCount do
      for j := 1 to i-1 do
        if S.F[i,j] <> 0 then begin
          Inc(n);
          FGridSetEditText(Sender, 1, n, FloatToStr(S.F[i,j]));
          FGridSetEditText(Sender, 2, n, IntToStr(i));
          FGridSetEditText(Sender, 3, n, IntToStr(j));
        end;
  end;
  inherited;
end;

procedure TSymmetryCoordsForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
var
  i, j: Integer;
begin
  if (ARow > 0) and (ACol > 0) and Assigned(S) and Assigned(S.U) and (Length(S.A) = 0) then begin
    { Writing to matrix - copy default if present }
    SetLength(S.A,1); { Prevent re-entry }
    for i := 1 to Min(S.U.RowCount, Grid.RowCount-1) do
      for j := 1 to S.U.ColCount do
        GridSetEditText(Sender, j, i, FloatToStr(S.U[i,j]));
  end;
  inherited;
end;

procedure TSymmetryCoordsForm.PopulateForm(ACol: Integer);
  procedure StartCopyToGrid(M: TMatrix; G: TNumberGrid);
  begin
    G.BeginUpdate;
    G.ColCount := M.ColCount+2;
    G.RowCount := M.RowCount+2;
    G.Clear;
  end;
var
  r, i: Integer;
begin
  if csLoading in ComponentState then
    Exit;
  Caption := ActiveObject.Name + ' for ' + (ActiveObject.Owner as TState).UniqueName;
  if (Length(S.A) = 0) and Assigned(S.U) then begin
    StartCopyToGrid(S.U, Grid);
    for i := 1 to S.U.RowCount do
      Grid[i,i] := 1;
    Grid.EndUpdate;
  end else begin
    Grid.RowCount := Length(S.A)+2;
    r := 0;
    for i := 0 to High(S.A) do
      r := Max(r, Length(S.A[i]));
    if Assigned(S.CartesianDisplacements.InternalCoords) then
      r := Max(S.CartesianDisplacements.InternalCoords.nInternal+1,r);
    Grid.ColCount := r+1;
  end;
  if Length(S.AF) = 0 then
    FGrid.RowCount := Length(S.ComputedF)+2;
  r := Grid.ColCount-1;
  for i := 1 to Grid.ColCount-2 do begin
    if Assigned(S.CartesianDisplacements) and Assigned(S.CartesianDisplacements.InternalCoords) then
      Grid.Title[i] := S.CartesianDisplacements.InternalCoords.nameR(i)
    else
      Grid.Title[i] := 'R'+IntToStr(i);
    Grid.ColWidths[i] := Max(Grid.ColWidths[i], Grid.DefaultColWidth);
  end;
  Grid.Title[Grid.ColCount-1] := ' ';
  for i := 1 to Grid.RowCount-2 do
    Grid.Lbl[i] := S.nameS(i);
  Grid.Lbl[Grid.RowCount-1] := '';
  inherited;
  Grid.ColCount := r+1; { Overwritten in VibGidForm }
  Replot(ACol);
end;

procedure TSymmetryCoordsForm.Replot(ACol: Integer);
var
  r: Integer;
begin
  if ACol < 0 then
    r := Grid.Row
  else
    r := ACol;
  if (r < 1) or (r > Length(S.Brows)) then
    r := -1
  else
    r := S.Brows[r-1];
  S.MakeLMatrix(S.ElectronicState, False, False, False, True); { TODO: May cause too many calls to underlying routines }
  if Assigned(S.CartesianDisplacements) then
    PlotZmatrix(S.CartesianDisplacements.InternalCoords.zMatrix, S.LastError+S.LastWarning, r, S.Bs, MagnificationSpin.Value, not S.CartesianDisplacements.InPrincipalAxis, -1)
  else
    PlotZmatrix(nil, S.LastError+S.LastWarning, r, S.Bs, MagnificationSpin.Value, False, -1); { Just shows error message }
end;

procedure TSymmetryCoordsForm.SetActiveObject(const Value: TMNode);
begin
  inherited;
  S := Value as TSymmetryCoords;
  S.MakeLMatrix(S.ElectronicState, False, False, False, True);
  RowSelect := True;
end;

procedure TSymmetryCoordsForm.ShowFMatrixItemClick(Sender: TObject);
begin
  ReShowLogForm;
  S.CalculateF(True, False);
end;

procedure TSymmetryCoordsForm.MakeExtraOperatorsItemClick(Sender: TObject);
begin
  BeginGlobalUpdate;
  try
    S.MakeLMatrix(S.ElectronicState, False, False, True, False);
  finally
    EndGlobalUpdate;
  end;
  if S.LastError <> '' then
    ShowInfo(S.LastError);
end;

procedure TSymmetryCoordsForm.MakeFMatrixItemClick(Sender: TObject);
begin
  S.CalculateF(False, True);
  PopulateForm;
end;

procedure TSymmetryCoordsForm.SortFMatrixbyValueItemClick(Sender: TObject);
begin
  S.SortF(True);
  PopulateForm;
end;

procedure TSymmetryCoordsForm.SortFMatrixbyModeItemClick(Sender: TObject);
begin
  S.SortF(False);
  PopulateForm;
end;

end.
