unit VibGridFrm;
{$I utils.inc}{Copyright (C) 2013 C M Western; see commandline.pas for license}

interface

uses Types, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, Menus, LmatrixFrm,
     ComCtrls, Split, ExtCtrls, GraphM, Graphs, Grids, MNode, Mixture, NGrid, zMatrix,
     StdCtrls, MatrixVec, NewSpin, AlEdit, EditExtr, ScrlCtrl, GraphI, MNodeExpr,
     VibPlotFrm;

type

  { TVibGridForm }

  TVibGridForm = class(TVibPlotForm)
    Grid: TNumberGrid;
    TopPanel: TPanel;
    TopSplit: TSplit;
    FGrid: TNumberGrid;
    procedure FGridGetCell(Sender: TObject; Col, Row: Longint; var v: Double; var Valid: Boolean);
    procedure GridGetCell(Sender: TObject; Col, Row: Longint; var v: Double;var Valid: Boolean);
    procedure GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
    procedure GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
    procedure GridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure GridMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure GridSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
    procedure FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
    procedure FGridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
    procedure FGridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
    procedure FGridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
    procedure GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
    procedure GResize(Sender: TObject);
  private
    Node: TMNodeWithArray;
  protected
    xMin, xMax, yMin, yMax: Double;
    ColsHeadroom: Integer;
    RowSelect: Boolean; { Set true to select rows, rather then columns to plot }
    function GetActiveObject:TMNode; override;
    procedure SetActiveObject(const Value: TMNode); override;
  public
    procedure PopulateForm(ACol:Integer); overload; override;
    procedure Replot(ACol:Integer = -1); virtual; abstract;
    procedure ExtraPlot(zMatrix: TZMatrix; ACol, ix, iy: Integer; Original: Boolean); virtual;
    procedure PlotZmatrix(zMatrix: TZMatrix; const Msg: string; r: Integer; B: TMatrix; Magnification: Double; Original: Boolean; AtomEmph: Integer);
  end;

implementation

uses Math, MathSup, Gmarks;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

procedure TVibGridForm.FGridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
var
  V: Double;
begin
  if (Row > 0) and (Col > 0) and (Row <= Length(Node.AF)) and (Col <= Length(Node.AF[Row-1])) then begin
    s := Node.AF[Row-1][Col-1];
    IsText := (Col <> 1) or not TryStrToFloat(s,V); { Simple numbers fall through to GetCell }
  end;
end;

procedure TVibGridForm.FGridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
begin
  if (ARow > 0) and (ACol > 0) and (ARow <= Length(Node.AF)) and (ACol <= Length(Node.AF[ARow-1])) then
    Value := Node.AF[ARow-1][ACol-1];
end;

procedure TVibGridForm.FGridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
begin
  { Allows the delete key to work }
  if not Valid then begin
    if (Row <= Length(Node.AF)) and (Col <= Length(Node.AF[Row-1])) then begin
      Node.AF[Row-1][Col-1] := '';
      FGrid.Changed(Col, Row);
    end;
  end;
end;

procedure TVibGridForm.FGridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
begin
  if ARow > Length(Node.AF) then
    SetLength(Node.AF,ARow);
  if ACol > Length(Node.AF[ARow-1]) then
    SetLength(Node.AF[ARow-1],ACol);
  Node.AF[ARow-1][ACol-1] := Value;
  FGrid.Valid[ARow,ACol] := True; { Turns off error indication, if any }
end;

function TVibGridForm.GetActiveObject: TMNode;
begin
  Result := Node;
end;

procedure TVibGridForm.GResize(Sender: TObject);
begin
  Replot;
end;

procedure TVibGridForm.GridGetCellLabel(Sender: TObject; Col, Row: Integer; var s: string; var IsText: Boolean);
var
  V: Double;
begin
  if (Row > 0) and (Col > 0) and (Row <= Length(Node.A)) and (Col <= Length(Node.A[Row-1])) then begin
    s := Node.A[Row-1][Col-1];
    IsText := not TryStrToFloat(s,V); { Simple numbers fall through to GetCell }
  end;
end;

procedure TVibGridForm.GridGetEditText(Sender: TObject; ACol, ARow: Integer; var Value: string);
begin
  if (ARow > 0) and (ACol > 0) and (ARow <= Length(Node.A)) and (ACol <= Length(Node.A[ARow-1])) then
    Value := Node.A[ARow-1][ACol-1];
end;

procedure TVibGridForm.GridGetCell(Sender: TObject; Col, Row: Longint; var v: Double; var Valid: Boolean);
begin
  if (Row > 0) and (Col > 0) and (Row <= Length(Node.A)) and (Col <= Length(Node.A[Row-1])) then
    Valid := TryStrToFloat(Node.A[Row-1][Col-1], v);
end;

procedure TVibGridForm.FGridGetCell(Sender: TObject; Col, Row: Longint; var v: Double; var Valid: Boolean);
begin
  if (Row > 0) and (Col > 0) and (Row <= Length(Node.AF)) and (Col <= Length(Node.AF[Row-1])) then
    Valid := TryStrToFloat(Node.AF[Row-1][Col-1], v);
end;

procedure TVibGridForm.GridMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  Replot;
end;

procedure TVibGridForm.GridSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
  if Assigned(Node) and (ARow <= Length(Node.A)) and (ACol <= Length(Node.A[ARow-1])) then
    StatusBar.SimpleText := Node.A[ARow-1][ACol-1];
  if RowSelect then begin
    if ARow <> Grid.Row then
      Replot(ARow)
  end else if ACol <> Grid.Col then
    Replot(ACol);
end;

procedure TVibGridForm.GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double; var Valid: Boolean);
begin
  { Allows the delete key to work }
  if not Valid then begin
    if (Row <= Length(Node.A)) and (Col <= Length(Node.A[Row-1])) then begin
      Node.A[Row-1][Col-1] := '';
      Grid.Changed(Col, Row);
    end;
  end;
end;

procedure TVibGridForm.GridSetEditText(Sender: TObject; ACol, ARow: Integer; const Value: string);
begin
  if ARow > Length(Node.A) then
    SetLength(Node.A,ARow);
  if ACol > Length(Node.A[ARow-1]) then
    SetLength(Node.A[ARow-1],ACol);
  Node.A[ARow-1][ACol-1] := Value;
  Grid.Valid[ARow,ACol] := True; { Turns off error indication, if any }
end;

procedure TVibGridForm.PlotZmatrix(zMatrix: TZMatrix; const Msg: string; r: Integer;
  B: TMatrix; Magnification: Double; Original: Boolean; AtomEmph: Integer);
var
  d, x, y, cx, cy, dx, dy: Double;
  ix, iy, i, j, ia: Integer;
  xyz, dxyz: Txyz;
begin
  PrepRotation;
  G.Start;
  G.Title1(Msg);
  if Assigned(zMatrix) and (High(zMatrix.Atoms) > 0) then begin
    xMin := 0;
    xMax := 0;
    yMin := 0;
    yMax := 0;
    case PlotPlaneGroup.ItemIndex of
      0: begin ix := 1; iy := 2; end;
      1: begin ix := 2; iy := 3; end;
      2: begin ix := 1; iy := 3; end;
    end;
    with zMatrix do
      for i := 0 to High(Atoms) do begin
        if Original then
          xyz := Atoms[i].xyz
        else
          xyz := Atoms[i].pxyz;
        Rotate(xyz, ix);
        xMin := Min(xyz[ix], xMin);
        xMax := Max(xyz[ix], xMax);
        yMin := Min(xyz[iy], yMin);
        yMax := Max(xyz[iy], yMax);
    end;
    d := Max(xMax-xMin, yMax-yMin)/10;
    d := Max(d, 0.2);
    G.MapLin(xMin-d,xMax+d,yMin-d,yMax+d);
    if RotationAngle.Value = 0 then
      G.Scales('xyz'[ix],'xyz'[iy])
    else
      G.Scales('xyz'[ix],'');
    G.Chlen(cx,cy);
    if Original and Assigned(zMatrix.Itrans) then
      ShowPrincipalAxes(zMatrix.CofM[ix], zMatrix.CofM[iy], ix, iy, zMatrix.Itrans);
    ExtraPlot(zMatrix, r, ix, iy, Original);
    ia := 0; { Need separate index to ignore dummy atoms }
    with zMatrix do
      for i := 0 to High(Atoms) do begin
        if Original then
          xyz := Atoms[i].xyz
        else
          xyz := Atoms[i].pxyz;
        Rotate(xyz, ix);
        x := xyz[ix];
        y := xyz[iy];
        if i = AtomEmph then
          G.Symbol(x,y,5) { "Current" atom }
        else if Atoms[i].Info.Mass <= 0 then
          G.Symbol(x,y,4) { Dummy atom }
        else
          G.Symbol(x,y,3);
        G.Text(x+cx,y-cy,Atoms[i].Name);
        if (r > 0) and Assigned(B) and (r <= B.RowCount) and (3*ia+3 <= B.ColCount) and (Atoms[i].Info.Mass > 0) then begin
          for j := 1 to 3 do
            dxyz[j] := B[r, ia*3+j]*Magnification;
          Rotate(dxyz, ix);
          dx := dxyz[ix];
          dy := dxyz[iy];
          if not (IsNan(dx) or IsNan(dy)) then
            G.Arrow(x,y,x+dx,y+dy,arHead);
          Inc(ia);
        end;
      end;
  end;
  G.Finish;
end;

procedure TVibGridForm.PopulateForm(ACol:Integer);
var
  i, m: Integer;
begin
  if csLoading in ComponentState then
    Exit;
  Grid.AllowText := True;
  FGrid.AllowText := True;
  Grid.Clear;
  Grid.AutoAdjustColumn(0,False,True);
  for i := 1 to Grid.ColCount-1 do
    Grid.AutoAdjustColumn(i,i>0);
  Grid.RowCount := Max(Grid.RowCount, Length(Node.A)+2);
  m := 0;
  for i := 0 to High(Node.A) do
    m := Max(Length(Node.A[i]),m);
  Grid.ColCount := Max(Grid.ColCount, m+1+ColsHeadroom);
  if FGrid.Visible then begin
    FGrid.Invalidate;
    FGrid.RowCount := Max(FGrid.RowCount, Length(Node.AF)+2);
    for i := 2 to FGrid.ColCount-1 do
      FGrid.Title[i] := IntToStr(i-1);
  end;
end;

procedure TVibGridForm.ExtraPlot(zMatrix: TZMatrix; ACol, ix, iy: Integer; Original: Boolean);
begin
  {}
end;

procedure TVibGridForm.SetActiveObject(const Value: TMNode);
begin
  ColsHeadroom := 1;
  Node := Value as TMNodeWithArray;
end;

end.
