unit VibPart;
{ Class for calculating partition functions using a TVibratingMolecule object }
{$I utils.inc}{Copyright (C) 2011 C M Western; see commandline.pas for license}
interface

uses SysUtils, Math, MathSup, Mixture, PartFunc, Vib;

type

  TVibrationalPartitionFunction = class(TDefaultPartitionFunction)
  published
    VibMolecule: TVibratingMolecule;
    function FixupProperties: Exception; override;
    function CalculatePartitionFunction(Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception; out Emin: Double; var TheSpinSpeciesAbundance: TDoubleArray;
                                          EMax: Double = MaxDouble; Print: Boolean = false; ShowMolecule: Boolean = false; ShowHeader: Boolean = false; RemoveSpins: Byte = 0):Double; override;
    function HelpOnParameter(AIndex: Integer): string; override;
  published
    property ESwitch: Double index 1 read GetParameter write SetParameter;
  end;


implementation

uses Worker, LevelLister;

{ TVibrationalPartitionFunction }

function TVibrationalPartitionFunction.CalculatePartitionFunction(
  Molecule: TMolecule; Simulation: TSimulation; out Overflow: Exception;
  out Emin: Double; var TheSpinSpeciesAbundance: TDoubleArray;
  EMax: Double; Print, ShowMolecule, ShowHeader: Boolean; RemoveSpins: Byte): Double;
var
  Unused: TDoubleArray;
  Incl, Only: Double;
begin
  Result := inherited CalculatePartitionFunction(Molecule, Simulation, Overflow, Emin, TheSpinSpeciesAbundance, MaxDouble, Print, ShowMolecule, ShowHeader, RemoveSpins);
  if Assigned(VibMolecule) and (ESwitch > 0) then
    with VibMolecule do begin
      if Print then begin
        WriteLn(Log);
        WriteLn(Log, 'Calculating Separate Vibrational Partition function');
      end;
      { Vibrational partition function including all vibrational states }
      Incl := ActivePartitionFunction.CalculatePartitionFunction(VibMolecule,Simulation,Overflow,Emin,Unused,
                                                                 EMax, Print, ShowMolecule, ShowHeader);
      if Print then begin
        WriteLn(Log, 'Calculating Vibrational Partition function for levels up to ',FloatToStr(ESwitch));
        ShowLevelList(VibMolecule, lltByEnergy, ESwitch);
      end;
      Only := ActivePartitionFunction.CalculatePartitionFunction(VibMolecule,Simulation,Overflow,Emin,Unused,
                                                                 ESwitch, Print, ShowMolecule, ShowHeader);
      Result := Result * Incl/Only;
    end;
end;

function TVibrationalPartitionFunction.FixupProperties: Exception;
begin
  if ComponentCount = 1 then
    VibMolecule := Components[0] as TVibratingMolecule
  else
    VibMolecule := nil;
  Result := inherited FixUpProperties;
end;

function TVibrationalPartitionFunction.HelpOnParameter(AIndex: Integer): string;
begin
  if AIndex = 1 then
    Result := 'Vibronic levels with energies below this are assumed to be included in the rovibronic calculation'
  else
    Result := inherited HelpOnParameter(AIndex);
end;

initialization
  TVibrationalPartitionFunction.RegisterClass(TMolecule);
  TVibratingMolecule.RegisterClass(TVibrationalPartitionFunction);
end.
