unit VibPlotFrm;
{$I utils.inc}{Copyright (C) 2014 C M Western; see commandline.pas for license}

interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, MNode, WhatsThis, ComCtrls, StdCtrls, AlEdit, EditExtr, NewSpin,
     ExtCtrls, ScrlCtrl, GraphI, GraphM, Graphs, Split, Menus, MatrixVec, zMatrix;

type

  TVibForm = class(THelpForm)
  protected
    procedure SetActiveObject(const Value: TMNode); virtual; abstract;
    function GetActiveObject:TMNode; virtual; abstract;
    procedure MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
  public
    PopulateNeeded: Boolean;
    destructor Destroy; override;
    procedure UpdateActions; override;
    procedure PopulateForm; overload;
    procedure PopulateForm(ACol:Integer); overload; virtual; abstract;
    property ActiveObject: TMNode read GetActiveObject write SetActiveObject;
  end;

  TVibFormClass = class of TVibForm;

  TVibPlotForm = class(TVibForm)
    OperatePopupMenu: TPopupMenu;
    Panel: TPanel;
    G: TGraphD;
    InnerPanel: TPanel;
    PlotPlaneGroup: TRadioGroup;
    MagnificationLabel: TLabel;
    MagnificationSpin: TFixedSpin;
    OperateButton: TButton;
    Split: TSplit;
    StatusBar: TStatusBar;
    RotationLabel: TLabel;
    RotationAngle: TFixedSpin;
    procedure GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure PlotPlaneGroupClick(Sender: TObject);
    procedure OperateButtonClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormClose(Sender: TObject; var AnAction: TCloseAction);
  private
    sn, cs: Extended;
  public
    procedure ShowPrincipalAxes(xc, yc: Double; ix, iy: Integer; Itrans: TMatrix);
    procedure Rotate(var xyz: Txyz; Axis: Integer);
    procedure PrepRotation;
  end;

function ConnectVibFrm(MNode: TMnode; FormClass: TVibFormClass): TComponent;

implementation

uses Math, MathSup, Global;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function ConnectVibFrm(MNode: TMnode; FormClass: TVibFormClass): TComponent;
var
  Form: TVibForm;
  i: Integer;
begin
  for i := 0 to Screen.CustomFormCount - 1 do
    if Screen.CustomForms[i].ClassType = FormClass then
      if TVibForm(Screen.CustomForms[i]).ActiveObject = MNode then begin
        ReShow(Screen.CustomForms[i]);
        Result := Screen.CustomForms[i];
        Exit;
      end;
  Form := FormClass.Create(Application);
  Result := Form;
  Form.ActiveObject := MNode;
  AddMNodeNotification({$IFDEF FPC}@{$ENDIF}Form.MNodeChange);
  Form.Show;
  Form.PopulateForm;
end;

{ TVibForm }

destructor TVibForm.Destroy;
begin
  RemoveMNodeNotification({$IFDEF FPC}@{$ENDIF}MNodeChange);
  inherited;
end;

procedure TVibForm.MNodeChange(Sender: TMnode; Change: TMNodeChange; Index: Integer);
begin
  case Change of
    mnDestroy, mnDummyDestroy:
      if Sender = ActiveObject then Free;
    mnParameterChange, mnSettingChange, mnPropNameChange, mnNameChange, mnGraphicChanged:
      PopulateNeeded := True;
  end;
end;

procedure TVibForm.PopulateForm;
begin
  PopulateForm(-1);
end;

procedure TVibForm.UpdateActions;
begin
  inherited;
  if PopulateNeeded then begin
    if Assigned(ActiveObject) then
      PopulateForm;
    PopulateNeeded := False;
  end;
end;

{ TVibPlotForm }

procedure TVibPlotForm.FormClose(Sender: TObject; var AnAction: TCloseAction);
begin
  AnAction := caFree;
end;

procedure TVibPlotForm.FormShow(Sender: TObject);
begin
  InnerPanel.Height := 2*PlotPlaneGroup.Top + PlotPlaneGroup.Height;
  {$IFDEF LCLcarbon}
  MagnificationSpin.AutoSize := True;
  {$ENDIF}
  PopulateForm;
end;

procedure TVibPlotForm.GMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  s: string;
  dx, dy: Double;
  c: WideChar;
begin
  s := Concat(RoundedStr(G.LastMouseX, (G.Xmax-G.Xmin)/G.Width), '  ',
              RoundedStr(G.LastMouseY, (G.Ymax-G.Ymin)/G.Height));
  if G.AreaSelected then
    with G.SelectedMathArea do begin
      dx := Right-Left;
      dy := Top-Bottom;
      s := s + ', [' + RoundedStr(Sqrt(Sqr(dx)+Sqr(dy)), (G.Xmax-G.Xmin)/G.Width) + ', ';
      s := s + RoundedStr(RadToDeg(arctan2(-dy,dx)), 0.1);
      c := #$b0;
      s := s + c + ']';
    end;
  StatusBar.SimpleText := s;
end;

procedure TVibPlotForm.OperateButtonClick(Sender: TObject);
var
  P: TPoint;
begin
  P := OperateButton.ClientToScreen(Point(10,10));
  OperatePopupMenu.Popup(P.X, P.Y);
end;

procedure TVibPlotForm.PlotPlaneGroupClick(Sender: TObject);
begin
  PopulateForm;
end;

procedure TVibPlotForm.PrepRotation;
var
  Ang: Double;
begin
  Ang := RotationAngle.Value;
  SinCos(DegToRad(Ang),sn, cs);
end;

procedure TVibPlotForm.Rotate(var xyz: Txyz; Axis: Integer);
var
  i1, i2: Integer;
  t: Double;
begin
  i1 := Axis mod 3 + 1;
  i2 := i1 mod 3 + 1;
  t := cs*xyz[i1]+sn*xyz[i2];
  xyz[i2] := -sn*xyz[i1]+cs*xyz[i2];
  xyz[i1] := t;
end;

procedure TVibPlotForm.ShowPrincipalAxes(xc, yc: Double; ix, iy: Integer; Itrans: TMatrix);
var
  A: TColumnVector;
  i, M1, M2: Integer;
  cx, cy, Dx, Dy: Double;
  xyz: Txyz;
  j: Integer;
begin
  if not Assigned(Itrans) then
    Exit;
  PrepRotation;
  G.Chlen(cx, cy);
  A := TColumnVector.Create(3);
  for i := 1 to 3 do begin
    A.Zero;
    A[i] := 1000;
    A.PreMultiply(Itrans);
    for j := 1 to 3 do
      xyz[j] := A[j];
    Rotate(xyz, ix);
    Dx := xyz[ix];
    Dy := xyz[iy];
    if Sqr(Dx) + Sqr(dy) < 1 then
      Continue; { Axis perpendicular to plane }
    G.ClipPoint(xc-Dx,yc-Dy);
    Dx := Dx + xc;
    Dy := Dy + yc;
    G.ClipJoin(Dx, Dy);
    M1 := G.OutCode(xc, yc);
    M2 := G.OutCode(Dx, Dy);
    if (M1 and M2) = 0 then begin
      { Some of axis visible }
      G.ClpPnt(Dx,Dy,xc,yc,M2);
      if M2 <> 0 then
         G.ClpPnt(Dx,Dy,xc,yc,M2);
      G.Text(Dx+cx/2,Dy+cy/2,Char(Ord('a')+i-1));
    end;
  end;
  A.Free;
end;

end.
