unit TabTextForm;
{$I utils.inc}
interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF} Messages, SysUtils, Classes,
  Graphics, Controls, Forms, Dialogs, ComCtrls, ToolBarEx, ToolWin, LogActn,
  ExActn, ActnList, ScrlCtrl, LogControl, EditLog, Menus, ProjAbout,
  FxFntDlg, CompatMsg, Types, Recent;

{$IFDEF MSWINDOWS}
{$IFNDEF FPC}
{$DEFINE DRAWCLOSEBTN}
{$ENDIF}
{$ENDIF}

type

  TFileMemo = class(TEditLog)
  public
    { Additional per file information that must be recorded }
    LastFileAge: Integer;
    FileName: string;
    AlienNewline, PreserveNewline: Boolean;
    procedure SetFileName(AFileName: string); virtual;
    procedure ResetModified;
    function TabSheet:TTabSheet;
    procedure Close;
    procedure QueryClose(var CanClose: Boolean);
    procedure Save;
    procedure SaveAs;
    procedure SaveToFile(const AFileName: string);
    procedure CheckAge;
    procedure Reload;
    procedure LoadFile(AFileName: string; APreserveNewline: Boolean);
  end;

  { TTabbedMainText }

  TTabbedMainText = class(TForm)
    PageControl: TPageControl;
    ActionList: TActionList;
    ExSelectAll: TExSelectAll;
    LogFind: TLogFind;
    LogFindNext: TLogFindNext;
    LogFindThisWord: TLogFindThisWord;
    TextGoToLine: TTextGoToLine;
    ExClear: TExClear;
    TextFind: TLogFind;
    TextFindNext: TLogFindNext;
    ExCopy: TExCopy;
    ExCut: TExCut;
    ExPaste: TExPaste;
    ExRedo: TExRedo;
    ExUndo: TExUndo;
    TextDecrement: TTextDecrement;
    TextExchangeWithBookMark: TTextExchangeWithBookMark;
    TextGoToBookMark: TTextGoToBookMark;
    TextIncrement: TTextIncrement;
    TextPlayBackMacro: TTextPlayBackMacro;
    TextRecordMacro: TTextRecordMacro;
    TextSetBookMark: TTextSetBookMark;
    TextSortDown: TTextSortDown;
    TextSortUp: TTextSortUp;
    TextReplace: TTextReplace;
    TextRemoveTabs: TTextRemoveTabs;
    ToolBar: TToolBarEx;
    OpenButton: TToolButtonEx;
    NewButton: TToolButtonEx;
    SaveButton: TToolButtonEx;
    Spacer1: TToolButtonEx;
    CutButton: TToolButtonEx;
    CopyButton: TToolButtonEx;
    PasteButton: TToolButtonEx;
    UndoButton: TToolButtonEx;
    RedoButton: TToolButtonEx;
    Spacer2: TToolButtonEx;
    FindButton: TToolButtonEx;
    FindNextButton: TToolButtonEx;
    SortUpButton: TToolButtonEx;
    SortDownButton: TToolButtonEx;
    MainMenu: TMainMenu;
    FileItem: TMenuItem;
    FileNew: TMenuItem;
    FileOpen: TMenuItem;
    ReloadItem: TMenuItem;
    CloseItem: TMenuItem;
    Save: TMenuItem;
    SaveAs: TMenuItem;
    EditFileAtCursor: TMenuItem;
    StartFileatCursor: TMenuItem;
    Another: TMenuItem;
    N6: TMenuItem;
    Print: TMenuItem;
    PrintPreview: TMenuItem;
    FileSeparatorItem: TMenuItem;
    FileExit: TMenuItem;
    EditItem: TMenuItem;
    Undo: TMenuItem;
    Redo: TMenuItem;
    N2: TMenuItem;
    Cut: TMenuItem;
    CopyItem: TMenuItem;
    Delete: TMenuItem;
    Paste: TMenuItem;
    PasteSpecialItem: TMenuItem;
    SelectAllItem: TMenuItem;
    N3: TMenuItem;
    Find: TMenuItem;
    FindNext: TMenuItem;
    FindthisWordItem: TMenuItem;
    Replace: TMenuItem;
    N1: TMenuItem;
    GoToLineItem: TMenuItem;
    Special: TMenuItem;
    SetBookmark: TMenuItem;
    ExchangewithBookmark: TMenuItem;
    GotoBookmark: TMenuItem;
    N4: TMenuItem;
    Increment: TMenuItem;
    Decrement: TMenuItem;
    RemoveTabs: TMenuItem;
    SetTabSpacing: TMenuItem;
    N5: TMenuItem;
    SortUp: TMenuItem;
    SortDown: TMenuItem;
    Options: TMenuItem;
    RecordKeyStrokes: TMenuItem;
    PlayMacro: TMenuItem;
    FontItem: TMenuItem;
    KeepBackup: TMenuItem;
    AlienNewline: TMenuItem;
    PreserveNewline: TMenuItem;
    Slave: TMenuItem;
    HelpItem: TMenuItem;
    About: TMenuItem;
    OpenDialog: TOpenDialog;
    TabbedOpen: TAction;
    TabbedSave: TAction;
    TabbedNew: TAction;
    StatusBar: TStatusBar;
    Project: TProject;
    SaveDialog: TSaveDialog;
    FixedFontDialog: TFixedFontDialog;
    SaveAllItem: TMenuItem;
    TabPopupMenu: TPopupMenu;
    ClosePageItem: TMenuItem;
    CloseAllOtherPagesItem: TMenuItem;
    NItem: TMenuItem;
    SaveAllButton: TToolButtonEx;
    TabbedSaveAll: TAction;
    RecentFiles: TRecentFiles;
    N7: TMenuItem;
    procedure PageControlCloseTabClicked(Sender: TObject);
    procedure TabbedOpenExecute(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TabbedNewExecute(Sender: TObject);
    procedure PageControlMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PageControlMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure PageControlMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure PageControlMouseLeave(Sender: TObject);
    {$IFDEF DRAWCLOSEBTN}
    procedure PageControlDrawTab(Control: TCustomTabControl; TabIndex: Integer; const Rect: TRect; IsActive: Boolean);
    {$ENDIF}
    procedure MRUFileListMRUItemClick(Sender: TObject; AFilename: string);
    procedure TabSheetShow(Sender: TObject);
    procedure AboutClick(Sender: TObject);
    procedure PasteSpecialItemClick(Sender: TObject);
    procedure AlienNewlineClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FileExitClick(Sender: TObject);
    procedure TabbedSaveExecute(Sender: TObject);
    procedure SaveAsClick(Sender: TObject);
    procedure FixedFontDialogChange(Sender: TObject);
    procedure FontItemClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure SaveAllItemClick(Sender: TObject);
    procedure AnotherClick(Sender: TObject);
    procedure EditFileAtCursorClick(Sender: TObject);
    procedure StartFileatCursorClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure StatusBarDblClick(Sender: TObject);
    procedure StatusBarMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure SetTabSpacingClick(Sender: TObject);
    procedure TabbedSaveUpdate(Sender: TObject);
    procedure TabPopupMenuPopup(Sender: TObject);
    procedure TabPopupMenuClick(Sender: TObject);
    procedure ClosePageItemClick(Sender: TObject);
    procedure TabbedSaveAllExecute(Sender: TObject);
    procedure TabbedSaveAllUpdate(Sender: TObject);
    procedure ReloadItemClick(Sender: TObject);
  protected
    procedure UpdateActions; override;
  private
    FLastMouseX: Integer;
    InitialTabSheet: TTabSheet;
    FCloseButtons: array of TRect;
    FCloseButtonDown: Integer;
    FCloseButtonPushed: Boolean;
    {$IFDEF DRAWCLOSEBTN}
    FTabBounds: array of Integer; { Array of right edges of tabs }
    {$ENDIF}
    {$IFDEF UNIX}
    SaveAllMsg, GotoLineMsg, AskMessage: TCompatMsg;
    procedure DoAsk(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
    {$ENDIF}
    {$IFDEF DRAWCLOSEBTN}
    procedure CheckCloseButtons;
    {$ENDIF}
    {$IFDEF MSWINDOWS}
    {$IFNDEF FPC}
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    {$ENDIF}
    procedure WMSettingChange(var Message: TMessage); message WM_SETTINGCHANGE;
    {$ENDIF}
    function MakeNewTab: TTabSheet;
    function MakeNewMemo: TFileMemo;
    procedure DisplayHint(Sender: TObject);
    procedure PasteSpecial(Sender: TObject);
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
    procedure TryAndSave(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
    procedure GotoLine(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
  public
    function LoadFile(FileName: string; Memo: TFileMemo = nil): Boolean;
    function MemoOf(TabSheet: TTabSheet):TFileMemo;
    procedure SetDefaultExt(Ext: string);
    function TabWithName(const TabName: string):TTabSheet;
    function TabWithFileName(const TabName: string):TTabSheet;
    procedure SwitchToSheet(TabSheet: TTabSheet);
    function ActiveMemo: TFileMemo;
    function ExecuteAction(AnAction: TBasicAction): Boolean; override;
    procedure CloseMemo(Memo: TFileMemo);
    function FileAtCursor: string;
    {$IFDEF MSWINDOWS}
    procedure WndProc(var Msg: TMessage); override;
    {$ENDIF}
  end;

var
  TabbedMainText: TTabbedMainText;

implementation

uses {$IFDEF MSWINDOWS} ShellAPI, ShlObj, ActiveX, {$IFNDEF FPC} ShLwApi, {$ENDIF}{$ENDIF}
     {$IFDEF TIMING} Profile, ProfileForm, {$ENDIF}
     {$IFDEF DARWIN} MacOSAll, CarbonProc, {$ENDIF}
     Clipbrd, Huge, UCompat, LongName, WCompat, Math, Detach, CharType, ResActn;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{$IFDEF DARWIN}
{ TODO: Integrate into TCompatMsg }
var
  FOpenLineEventHandlerUPP: AEEventHandlerUPP;
  FSaveAllEventHandlerUPP: AEEventHandlerUPP;

function CarbonApp_LineOpen(var AEvent: AppleEvent; var {%H-}Reply: AppleEvent;
  {%H-}Data: SInt32): OSErr; {$IFDEF darwin}mwpascal;{$ENDIF}
var
  DocList: AEDescList;
  nItems: Integer;
  Keyword: AEKeyword;
  FileDesc: AEDesc;
  FileRef: FSRef;
  FileURL: CFURLRef;
  FileCFStr: CFStringRef;
  FileName: string;
  LineNumber: Integer;
const
  Sname = 'CarbonApp_LineOpen';
begin
  if OSError(AEGetParamDesc(AEvent, keyDirectObject, typeAEList, DocList{%H-}),
    Sname, 'AEGetParamDesc') then Exit;
  try
    if OSError(AECountItems(DocList, nItems{%H-}), SName, 'AECountItems') then Exit;
    if nItems <> 2 then
      Exit;
    if OSError(AEGetNthDesc(DocList, 1, typeFSRef, @Keyword, FileDesc{%H-}),
      SName, 'AEGetNthDesc') then Exit;
    if OSError(AEGetDescData(FileDesc, @FileRef, SizeOf(FSRef)),
      SName, 'AEGetDescData') then Exit;
    if OSError(AEDisposeDesc(FileDesc),
      SName, 'AEDisposeDesc') then Exit;
    FileURL := CFURLCreateFromFSRef(kCFAllocatorDefault, FileRef);
    FileCFStr := CFURLCopyFileSystemPath(FileURL, kCFURLPOSIXPathStyle);
    try
      FileName := CFStringToStr(FileCFStr);
    finally
      FreeCFString(FileURL);
      FreeCFString(FileCFStr);
    end;
    if OSError(AEGetNthPtr(DocList, 2, typeInteger, nil, nil, @LineNumber, SizeOf(LineNumber), nil),
      SName, 'AEGetNthDesc') then Exit;
    TabbedMainText.LoadFile(FileName);
    with TabbedMainText.ActiveMemo do
      ChangeCaretPos(CaretPos.X, LineNumber-1, False);
  finally
    AEDisposeDesc(DocList);
  end;
  Result := noErr;
end;

function CarbonApp_SaveAll(var AEvent: AppleEvent; var {%H-}Reply: AppleEvent;
  {%H-}Data: SInt32): OSErr; {$IFDEF darwin}mwpascal;{$ENDIF}
var
  Handled: Boolean;
begin
  TabbedMainText.TryAndSave(nil, 0, 0, Handled);
  Result := noErr;
end;

procedure ConnectDarwinHandlers;
begin
  FOpenLineEventHandlerUPP := NewAEEventHandlerUPP(AEEventHandlerProcPtr(@CarbonApp_LineOpen));
  OSError(
    AEInstallEventHandler(kCoreEventClass, SwapEndian(FourCharCode('PGOs')), FOpenLineEventHandlerUPP, 0, False),
    'ConnectDarwinHandlers', 'AEInstallEventHandler');
end;

procedure DisconnectDarwinHandlers;
begin
  DisposeAEEventHandlerUPP(FOpenLineEventHandlerUPP);
end;
{$ENDIF}

var
  WM_Save_All: Word;
  WM_GOTO_LINE: Word;
  WM_PGOPHER_ASK: Word;

{ TTabbedMainText }

procedure TTabbedMainText.AboutClick(Sender: TObject);
begin
  Project.Execute;
end;

function TTabbedMainText.ActiveMemo: TFileMemo;
begin
  Result := MemoOf(PageControl.ActivePage);
end;

procedure TTabbedMainText.AlienNewlineClick(Sender: TObject);
begin
  with ActiveMemo do
    AlienNewline := not AlienNewline;
  SwitchToSheet(nil);
end;

procedure TTabbedMainText.AnotherClick(Sender: TObject);
var
  Ext: string;
begin
  Ext := ExtractFileExt(ActiveMemo.FileName);
  Detach1(Application.Exename, Ext);
end;

procedure TTabbedMainText.CloseMemo(Memo: TFileMemo);
var
  CanClose: Boolean;
begin
  Memo.QueryClose(CanClose);
  if PageControl.PageCount = 0 then begin
    InitialTabSheet := nil;
    InitialTabSheet := MakeNewTab;
    MemoOf(InitialTabSheet).SetFileName('');
  end;
end;

function TTabbedMainText.FileAtCursor: string;
var
  Path: string;
  OK: Boolean;
  {$IFDEF UNIX}
  ECode: Cardinal;
  {$ENDIF}
begin
  Result := ActiveMemo.WordAtCursor({$IFDEF FPC}@{$ENDIF}IsNonSpaceFileChar);
  DoDirSeparators(Result);
  if Length(Result) > 0 then begin
    OK := FileExists(Result);
    if not OK then begin
      Path := ExtractFilePath(ActiveMemo.FileName)+Result;
      OK := FileExists(Path);
      if OK then
        Result := Path;
    end;
    if not OK then begin
      {$IFDEF MSWINDOWS}
      raise EFOpenError.CreateFmt('Can''t open file %s', [Result]);
      {$ENDIF}
      {$IFDEF UNIX}
      ECode := GetLastOSError;
      Application.MessageBox(PChar(Format('Can''t open file %s - %s', [Result,SysErrorMessage(ECode)])),'Error',0);
      {$ENDIF}
    end;
  end;

end;

procedure TTabbedMainText.ClosePageItemClick(Sender: TObject);
begin
  CloseMemo(ActiveMemo);
end;

procedure TTabbedMainText.DisplayHint(Sender: TObject);
begin
  StatusBar.Panels[4].Text := Application.Hint;
end;

procedure TTabbedMainText.EditFileAtCursorClick(Sender: TObject);
var
  FileName: string;
begin
  FileName := FileAtCursor;
  if FileName <> '' then begin
    StatusBar.Panels[4].Text := Concat('Opening ', FileName);
    LoadFile(FileName)
  end;
end;

function TTabbedMainText.ExecuteAction(AnAction: TBasicAction): Boolean;
begin
  if AnAction is TResultAction then
    with AnAction as TResultAction do
      if Control = ActiveMemo then begin
        StatusBar.Panels[4].Text := Hint;
        Result := True;
        Exit;
      end;
  Result := inherited ExecuteAction(AnAction);
end;

procedure TTabbedMainText.FileExitClick(Sender: TObject);
begin
  Close;
end;

procedure TTabbedMainText.FixedFontDialogChange(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to PageControl.PageCount - 1 do
    MemoOf(PageControl.Pages[i]).FixedFont := FixedFontDialog.FixedFont;
end;

procedure TTabbedMainText.FontItemClick(Sender: TObject);
begin
  FixedFontDialog.Execute;
end;

procedure TTabbedMainText.FormActivate(Sender: TObject);
var
  i: Integer;
begin
  {$IFDEF UNIX}
  AskMessage.Status := 1; { Flag this application can respond to queries }
  {$ENDIF}
  for i := 0 to PageControl.PageCount - 1 do
    MemoOf(PageControl.Pages[i]).CheckAge;
end;

procedure TTabbedMainText.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  i: Integer;
begin
  CanClose := True;
  i := PageControl.PageCount - 1;
  while CanClose and (i >= 0) do begin
    MemoOf(PageControl.Pages[i]).QueryClose(CanClose);
    Dec(i);
  end;
end;

procedure TTabbedMainText.FormCreate(Sender: TObject);
{$IFDEF TIMING}
var
  MenuItem: TMenuItem;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  WM_Save_All := RegisterWindowMessage('WM_Save_All');
  WM_GOTO_LINE := RegisterWindowMessage('WM_GOTO_LINE');
  WM_PGOPHER_ASK := RegisterWindowMessage('WM_PGOPHER_ASK');
  {$ENDIF}
  Application.OnActivate := {$IFDEF FPC}@{$ENDIF}FormActivate;
  {$IFDEF UNIX}
    SaveAllMsg := TCompatMsg.Create(Self);
    with SaveAllMsg do begin
      MessageName := 'WM_Save_All';
      OnReceived := @TryAndSave;
      WM_Save_All := MessageID;
    end;
    GotoLineMsg := TCompatMsg.Create(Self);
    with GotoLineMsg do begin
      MessageName := 'WM_GOTO_LINE';
      OnReceived := @GotoLine;
      WM_GOTO_LINE := MessageID;
    end;
    AskMessage := TCompatMsg.Create(Self);
    with AskMessage  do begin
      MessageName := 'WM_PGOPHER_ASK';
      OnReceived := @DoAsk;
      WM_PGOPHER_ASK := MessageID;
    end;
  {$ENDIF}
  {$IFDEF DARWIN}
    ConnectDarwinHandlers;
  {$ENDIF}
  {$IFDEF DRAWCLOSEBTN} { or any widget set that does not provide close buttons }
  PageControl.TabWidth := 150;
  PageControl.OwnerDraw := True;
  PageControl.OnDrawTab := PageControlDrawTab;
  {$ELSE}
  PageControl.Options := [nboShowCloseButtons];
  PageControl.OnCloseTabClicked :=  {$IFDEF FPC}@{$ENDIF}PageControlCloseTabClicked;
  {$ENDIF}
  InitialTabSheet := MakeNewTab;
  MemoOf(InitialTabSheet).SetFileName('');
  AllowCtrlC;
  Application.OnHint := {$IFDEF FPC}@{$ENDIF}DisplayHint;
  DragAcceptFiles(Self, True);
  {$IFDEF FPC}
  OnDropFiles := @FormDropFiles;
  {$ENDIF}
  PageControl.TabStop := False;
  AdjustMainMenu(MainMenu, About.OnClick);
//  OnShow := {$IFDEF FPC}@{$ENDIF}FormShow;
  {$IFDEF TIMING}
  MenuItem := TMenuItem.Create(Self);
  MenuItem.Caption := 'Timing';
  MenuItem.OnClick := @ProfilerForm.MenuHandler;
  HelpItem.Add(MenuItem);
  {$ENDIF}
end;

procedure TTabbedMainText.FormDestroy(Sender: TObject);
begin
  if HandleAllocated then
    DragAcceptFiles(Self, False);
  {$IFDEF DARWIN}
    DisconnectDarwinHandlers;
  {$ENDIF}
end;

procedure TTabbedMainText.FormDropFiles(Sender: TObject; const FileNames: array of String);
var
  i: Integer;
begin
  for i := 0 to High(FileNames) do
    LoadFile(FileNames[i]);
end;

procedure TTabbedMainText.GotoLine(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
begin
  try
    with ActiveMemo do
      ChangeCaretPos(CaretPos.X, lParam-1, False);
    Handled := True;
  except
  end
end;

function TTabbedMainText.LoadFile(FileName: string; Memo: TFileMemo): Boolean;
var
  TabSheet: TTabSheet;
  Buffer: string;
begin
  Result := False;
  if (Length(FileName) > 0) and ( (FileName[Length(FileName)] = PathDelim) or DirectoryExists(FileName) ) then begin
    ShowMessage(name + ' is a directory');
    Exit;
  end;
  Buffer := GetLongName(ExpandFileName(FileName));
  TabSheet := TabWithFileName(Buffer);
  if Assigned(TabSheet) then  begin
    SwitchToSheet(TabSheet);
    Exit;
  end;
  if (Memo = nil) and not FileExists(FileName) then begin
    {$IFDEF MSWINDOWS}
    if IDNO = Application.MessageBox(
           AddressOfFirstChar(Format('File %s does not exist. Create?'#0, [FileName])),
           'Warning', mb_YesNo or mb_IconQuestion or mb_DefButton1) then
    {$ENDIF}
    {$IFDEF UNIX}
    if mrNo = MessageDlg(Format('File %s does not exist. Create?', [FileName]),
                         mtWarning, mbYesNo, 0) then
    {$ENDIF}
      Exit;
    Memo := MakeNewMemo;
  end else begin
    if Memo = nil then
      Memo := MakeNewMemo;
    Memo.LoadFile(FileName, PreserveNewline.Checked);
  end;
  Result := True;
  Memo.SetFileName(ExpandFileName(FileName));
  Memo.ResetModified;
  Memo.StartOfText(False);
  InitialTabSheet := nil;
  SwitchToSheet(Memo.Owner as TTabSheet);
end;

function TTabbedMainText.MakeNewMemo: TFileMemo;
begin
  Result := MemoOf(MakeNewTab);
end;

function TTabbedMainText.MakeNewTab: TTabSheet;
begin
  if Assigned(InitialTabSheet) and not MemoOf(InitialTabSheet).Modified and (MemoOf(InitialTabSheet).Lines.Count = 0) then
    Result := InitialTabSheet
  else begin
    Result := TTabSheet.Create(Self);
    Result.Parent := PageControl;
    Result.OnShow := {$IFDEF FPC}@{$ENDIF}TabSheetShow;
    Result.TabStop := False;
    with TFileMemo.Create(Result) do begin
      Parent := Result;
      Align := alClient;
      BevelOuter := bvLowered;
      WantTabs := True;
      FixedFontDialog := Self.FixedFontDialog;
      FixedFont := FixedFontDialog.FixedFont;
      HideSelection := False;
      MaxLines := 0;
    end;
    Result.PageControl := PageControl;
  end;
end;

function TTabbedMainText.MemoOf(TabSheet: TTabSheet): TFileMemo;
begin
  Result := TabSheet.Controls[0] as TFileMemo;
end;

procedure TTabbedMainText.MRUFileListMRUItemClick(Sender: TObject; AFilename: string);
begin
  LoadFile(AFilename)
end;

{$IFDEF DRAWCLOSEBTN}
procedure TTabbedMainText.CheckCloseButtons;
var
  OldCount, i: Integer;
begin
  if Length(FCloseButtons) <> PageControl.PageCount then begin
    OldCount := High(FCloseButtons);
    Setlength(FCloseButtons, PageControl.PageCount);
    for i := OldCount+1 to PageControl.PageCount - 1 do
      FCloseButtons[i] := Rect(0,0,0,0);
    OldCount := High(FTabBounds);
    Setlength(FTabBounds, PageControl.PageCount);
    for i := OldCount+1 to PageControl.PageCount - 1 do
      FTabBounds[i] := 0;
  end;
  if Length(FTabBounds) <> PageControl.PageCount then begin
    OldCount := High(FTabBounds);
    Setlength(FTabBounds, PageControl.PageCount);
    for i := OldCount+1 to PageControl.PageCount - 1 do
      FTabBounds[i] := 0;
  end;
end;

procedure TTabbedMainText.PageControlDrawTab(Control: TCustomTabControl;
  TabIndex: Integer; const Rect: TRect; IsActive: Boolean);
const
  CloseBtnSize = 14;
var
  TabCaption: TPoint;
  BtnRect: TRect;
  DrawState: Cardinal;
  Size: TSize;
  s: string;
begin
  CheckCloseButtons;
  FTabBounds[TabIndex] := Rect.Right;
  
  TabCaption.Y := Rect.Top + 3;

  if IsActive then begin
    BtnRect.Top := Rect.Top + 4;
    BtnRect.Right := Rect.Right - 5;
    TabCaption.X := Rect.Left + 6;
  end else begin
    BtnRect.Top := Rect.Top + 3;
    BtnRect.Right := Rect.Right - 5;
    TabCaption.X := Rect.Left + 3;
  end;

  BtnRect.Bottom := BtnRect.Top + CloseBtnSize;
  BtnRect.Left := BtnRect.Right - CloseBtnSize;

  PageControl.Canvas.FillRect(Rect);
  s := PageControl.Pages[TabIndex].Caption;
  Size := PageControl.Canvas.TextExtent(s);
  if Size.cx > BtnRect.Left - TabCaption.X then begin
    s := ExtractFileName(s);
    Size := PageControl.Canvas.TextExtent(s);
    if Size.cx > BtnRect.Left - TabCaption.X then begin
      s := Copy(s,1,Length(s)-Length(ExtractFileExt(s)));
      while PageControl.Canvas.TextExtent(s).cx > BtnRect.Left - TabCaption.X  do
        s := Copy(s,1,Length(s)-1);
    end;
  end;

  PageControl.Canvas.TextOut(TabCaption.X, TabCaption.Y, s);

  if (FCloseButtonDown = TabIndex) and FCloseButtonPushed then
    DrawState := DFCS_CAPTIONCLOSE + DFCS_PUSHED
  else
    DrawState := DFCS_CAPTIONCLOSE;
  Windows.DrawFrameControl(PageControl.Canvas.Handle, BtnRect, DFC_CAPTION, DrawState);
  // Alternative if themes used
  {
    Dec(BtnRect.Left);
    if (FCloseButtonMouseDownIndex = TabIndex) and FCloseButtonShowPushed then
      CloseBtnDrawDetails := ThemeServices.GetElementDetails(twCloseButtonPushed)
    else
      CloseBtnDrawDetails := ThemeServices.GetElementDetails(twCloseButtonNormal);
    ThemeServices.DrawElement(PageControl.Canvas.Handle, CloseBtnDrawDetails,BtnRect);
  }
  FCloseButtons[TabIndex] := BtnRect;
end;
{$ENDIF}

procedure TTabbedMainText.PageControlMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
begin
  {$IFDEF DRAWCLOSEBTN}
    CheckCloseButtons;
    if Button = mbLeft then
      for i := 0 to PageControl.PageCount - 1 do
        if PtInRect(FCloseButtons[i], Point(X,Y)) then begin
          FCloseButtonDown := i;
          FCloseButtonPushed := True;
          PageControl.Repaint;
        end;
    if Button = mbRight then begin
      i := 0;
      while (i < High(FTabBounds)) and (X > FTabBounds[i]) do
        Inc(i);
      if i <> PageControl.ActivePageIndex then
        SwitchToSheet(PageControl.Pages[i]);
    end;
  {$ENDIF}
  ActiveMemo.SetFocus;
end;

procedure TTabbedMainText.PageControlMouseLeave(Sender: TObject);
begin
  {$IFDEF DRAWCLOSEBTN}
    if FCloseButtonPushed then begin
      FCloseButtonPushed := False;
      PageControl.Invalidate;
    end;
  {$ENDIF}
  PageControl.Hint := ActiveMemo.FileName;
end;

procedure TTabbedMainText.PageControlMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
begin
  {$IFDEF DRAWCLOSEBTN}
    if (ssLeft in Shift) and (FCloseButtonDown >= 0) then
      if FCloseButtonPushed <> PtInRect(FCloseButtons[FCloseButtonDown], Point(X,Y)) then begin
        FCloseButtonPushed := not FCloseButtonPushed;
        PageControl.Invalidate;
      end;
    i := 0;
    while (i < High(FTabBounds)) and (X > FTabBounds[i]) do
      Inc(i);
  {$ELSE}
    i := PageControl.TabIndexAtClientPos(Types.Point(X,Y));
  {$ENDIF}
  if i > 0 then
    PageControl.Hint := MemoOf(PageControl.Pages[i]).FileName;
end;

procedure TTabbedMainText.PageControlMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  {$IFDEF DRAWCLOSEBTN}
    if (Button = mbLeft) and (FCloseButtonDown >= 0)
        and PtInRect(FCloseButtons[FCloseButtonDown], Point(X,Y)) then begin
      PageControlCloseTabClicked(PageControl.Pages[FCloseButtonDown]);
    end;
  {$ENDIF}
end;

procedure TTabbedMainText.PasteSpecial(Sender: TObject);
var
  GHandle: TGHandle;
  Memo: TFileMemo;
  CFmt: Word;
  {$IFDEF MSWINDOWS}
  dataObj: IDataObject;
  fetc: TFormatEtc;
  stgm, stgmf: TStgMedium;
  strm: IStream;
  fgd: PFileGroupDescriptorW;
  i: Integer;
  Buffer: AnsiString;
  FileName: AnsiString;
  FileSize: LongInt;
  nb: LongInt;
  FormatEnumerator: IEnumFormatEtc;
  SourceFormatEtc: TFormatEtc;
  GetNum, GotNum: longInt;
  EnumName: string;
  Cida: PIDA;
  PIDL: PItemIDList;
  Shf, Child: IShellFolder;
  StrRet: TStrRet;
  {$ENDIF}
begin
  CFmt := Clipboard.Formats[(Sender as TComponent).Tag];
  Memo := MemoOf(PageControl.ActivePage);
  GHandle := TGHandle.Create;
  try
    {$IFDEF MSWINDOWS}{$IFNDEF FPC}
    {$IFNDEF UNICODE}
    if (CFmt = RegisterClipboardFormat(CFSTR_SHELLIDLIST)) then begin
      if (OleGetClipboard(dataObj) = S_OK) and (dataObj <> nil) then begin
        fetc.cfFormat := CFmt;
        fetc.ptd := nil;
        fetc.dwAspect := DVASPECT_CONTENT;
        fetc.lindex := -1;
        fetc.tymed := TYMED_HGLOBAL;
        if (dataObj.GetData(fetc,stgm) = S_OK) and (stgm.tymed = TYMED_HGLOBAL) then begin
          Cida := GlobalLock(stgm.hGlobal);
          PIDL := PItemIDList(PAnsiChar(Cida) + Cida.aoffset[0]);
          SetLength(Buffer,MAX_PATH);
          SHGetPathFromIDList(PIDL,PAnsiChar(Buffer));
          SetLength(Buffer,StrLen(PAnsiChar(Buffer)));
          Memo.Lines.Add(Buffer);
          SHGetDesktopFolder(Shf);
          Shf.BindToObject(PIDL, nil, IID_IShellFolder,Pointer(shf));
          for i := 1 to Cida.cidl do begin
            {$R-}
            PIDL := PItemIDList(PAnsiChar(Cida) + Cida.aoffset[i]);
            {$R+}
            SetLength(Buffer,MAX_PATH);
            Shf.GetDisplayNameOf(PIDL, SHGDN_NORMAL, StrRet);
            StrRetToBuf(@StrRet, PIDL, PAnsiChar(Buffer), Length(Buffer));
            SetLength(Buffer,StrLen(PAnsiChar(Buffer)));
            Memo.Lines.Add(Buffer);
          end;
        end;
        GlobalUnlock(stgm.hGlobal);
        ReleaseStgMedium(stgm);
      end;
    end else
    {$ENDIF}
    if (CFmt = RegisterClipboardFormat(CFSTR_FILEDESCRIPTORW)) then begin
      if (OleGetClipboard(dataObj) = S_OK) and (dataObj <> nil) then begin
        FillChar(fetc,SizeOf(fetc),0);
        fetc.cfFormat := CFmt;
        fetc.ptd := nil;
        fetc.dwAspect := DVASPECT_CONTENT;
        fetc.lindex := -1;
        fetc.tymed := TYMED_HGLOBAL;
        if (dataObj.GetData(fetc,stgm) = S_OK) and (stgm.tymed = TYMED_HGLOBAL) then begin
//          if (dataObj.EnumFormatEtc(DATADIR_GET, FormatEnumerator) = S_OK)
//              and (FormatEnumerator.Reset = S_OK) then begin
//            GetNum := 1; // Get one format at a time.
//            while (FormatEnumerator.Next(GetNum, SourceFormatEtc, @GotNum) = S_OK) and (GetNum = GotNum) do begin
//              SetLength(EnumName,255);
//              SetLength(EnumName,GetClipboardFormatName(SourceFormatEtc.cfFormat,PChar(EnumName),Length(EnumName)));
//              Memo.Lines.Add(IntToStr(SourceFormatEtc.cfFormat)+' '+EnumName+' '+IntToStr(SourceFormatEtc.lindex)+' '+IntToStr(SourceFormatEtc.tymed));
//            end;
//          end;
          fgd := GlobalLock(stgm.hGlobal);
          try
            for i := 0 to fgd.cItems-1 do begin
              {$R-}
              {$IFDEF UNICODE}
                FileName := fgd.fgd[i].cFileName;
              {$ELSE}
                SetLength(FileName, 3*MAX_PATH);
                SetLength(FileName, UnicodeToUtf8(PChar(FileName),@fgd.fgd[i].cFileName,Length(FileName))-1);
              {$ENDIF}
              FileSize := fgd.fgd[i].nFileSizeLow;
              Memo.Lines.Add(FileName);
              {$R+}
              fetc.cfFormat := RegisterClipboardFormat(CFSTR_FILECONTENTS);
              fetc.tymed := TYMED_HGLOBAL ; // or TYMED_ISTREAM or TYMED_ISTORAGE;
              fetc.lindex := i;
              if (dataObj.GetData(fetc,stgmf) = S_OK) then begin
                case stgmf.tymed of
                  TYMED_HGLOBAL:  Memo.Lines.Add('HGLOBAL');
                  TYMED_ISTREAM:
                    begin
                      strm := IStream(stgmf.stm);
                      SetLength(Buffer, FileSize);
                      // Should repeat until all read
                      strm.Read(@Buffer[1],Length(Buffer),@nb);
                      Memo.SelText := Buffer;
                    end;
                  TYMED_ISTORAGE: Memo.Lines.Add('ISTORAGE');
                end;
                ReleaseStgMedium(stgmf);
              end;
            end;
          finally
            GlobalUnlock(stgm.hGlobal);
            ReleaseStgMedium(stgm);
          end;
        end;
      end;
    end else if CFmt = RegisterClipboardFormat(CFSTR_FILEDESCRIPTOR) then begin
      if (OleGetClipboard(dataObj) = S_OK) and (dataObj <> nil) then begin
        fetc.cfFormat := CFmt;
        fetc.ptd := nil;
        fetc.dwAspect := DVASPECT_CONTENT;
        fetc.lindex := -1;
        fetc.tymed := TYMED_HGLOBAL;
        if (dataObj.GetData(fetc,stgm) = S_OK) and (stgm.tymed = TYMED_HGLOBAL) then begin
          fgd := GlobalLock(stgm.hGlobal);
          try
            for i := 0 to fgd.cItems-1 do begin
              {$R-}
              Memo.Lines.Add(fgd.fgd[i].cFileName);
              {$R+}
              fetc.cfFormat := RegisterClipboardFormat(CFSTR_FILECONTENTS);
              fetc.tymed := TYMED_HGLOBAL ; // or TYMED_ISTREAM or TYMED_ISTORAGE;
              fetc.lindex := i;
              if dataObj.GetData(fetc,stgmf) = S_OK then begin
                case stgmf.tymed of
                  TYMED_HGLOBAL:  Memo.Lines.Add('HGLOBAL');
                  TYMED_ISTREAM:
                    begin
                      strm := IStream(stgmf.stm);
                      SetLength(Buffer, fgd.fgd[i].nFileSizeLow*4);
                      // Should repeat until all read
                      strm.Read(@Buffer[1],fgd.fgd[i].nFileSizeLow,@nb);
                      SetLength(Buffer,nb);
                      Memo.SelText := Buffer;
                    end;
                  TYMED_ISTORAGE: Memo.Lines.Add('ISTORAGE');
                end;
                ReleaseStgMedium(stgmf);
              end;
            end;
          finally
            GlobalUnlock(stgm.hGlobal);
            ReleaseStgMedium(stgm);
          end;
        end;
      end;
    end else
    {$ENDIF}{$ENDIF}
    begin
      GHandle.SetFromClipboard(CFmt);
      Memo.SetSelTextBuf(GHandle.Item[0]);
    end;
  finally
    GHandle.Free;
  end;
end;

procedure TTabbedMainText.PasteSpecialItemClick(Sender: TObject);
var
  i, n, nF: Integer;
  MenuItem: TMenuItem;
  Buffer: string;
begin
  nF := Clipboard.FormatCount;
  while PasteSpecialItem.Count > nF do
    PasteSpecialItem.Delete(PasteSpecialItem.Count-1);
  while PasteSpecialItem.Count < nF do begin
    MenuItem := TMenuItem.Create(Self);
    PasteSpecialItem.Add(MenuItem);
  end;
  for i := 0 to nF - 1 do begin
    SetLength(Buffer,255);
    {$IFDEF MSWINDOWS}
    n := GetClipboardFormatName(Clipboard.Formats[i],PChar(Buffer),Length(Buffer));
    if n = 0 then
      case Clipboard.Formats[i] of
        Windows.CF_TEXT: Buffer := 'TEXT';
        Windows.CF_BITMAP: Buffer := 'BITMAP';
        Windows.CF_METAFILEPICT: Buffer := 'METAFILEPICT';
        CF_SYLK: Buffer := 'SYLK';
        CF_DIF: Buffer := 'DIF';
        CF_TIFF: Buffer := 'TIFF';
        CF_OEMTEXT: Buffer := 'OEMTEXT';
        CF_DIB: Buffer := 'DIB';
        CF_PALETTE: Buffer := 'PALETTE';
        CF_PENDATA: Buffer := 'PENDATA';
        CF_RIFF: Buffer := 'RIFF';
        CF_WAVE: Buffer := 'WAVE';
        CF_UNICODETEXT: Buffer := 'UNICODETEXT';
        CF_ENHMETAFILE: Buffer := 'ENHMETAFILE';
        CF_HDROP: Buffer := 'HDROP';
        CF_LOCALE: Buffer := 'LOCALE ';
        {$IFNDEF FPC}
        CF_DIBV5: Buffer := 'DIBV5';
        {$ENDIF}
        CF_OWNERDISPLAY: Buffer := 'OWNERDISPLAY';
        CF_DSPTEXT: Buffer := 'DSPTEXT';
        CF_DSPBITMAP: Buffer := 'DSPBITMAP';
        CF_DSPMETAFILEPICT: Buffer := 'DSPMETAFILEPICT';
        CF_DSPENHMETAFILE: Buffer := 'DSPENHMETAFILE';
      else
        Buffer := '';
      end
    else
      SetLength(Buffer,n);
    {$ELSE}
    Buffer := '';
    {$ENDIF}
    with PasteSpecialItem.Items[i] do begin
      Tag := i;
      OnClick := {$IFDEF FPC}@{$ENDIF}PasteSpecial;
      Caption := Format('%s'#9'%d',[Buffer, Clipboard.Formats[i]])
    end;
  end;
end;

procedure TTabbedMainText.ReloadItemClick(Sender: TObject);
begin
  ActiveMemo.Reload;
end;

procedure TTabbedMainText.SaveAllItemClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to PageControl.PageCount - 1 do
    with MemoOf(PageControl.Pages[i]) do
      if Modified then
        Save;
end;

procedure TTabbedMainText.SaveAsClick(Sender: TObject);
begin
  ActiveMemo.SaveAs;
end;

procedure TTabbedMainText.SetDefaultExt(Ext: string);
var
  i: Integer;
  e, p, ee: PChar;
  Filter: string;
begin
  if Length(Ext) = 0 then Exit;
  if Ext[1] <> '.' then Ext := Concat('.', Ext);
  Ext := Concat('*', Ext);
  OpenDialog.DefaultExt := Copy(Ext, 3, Length(Ext));
  Filter := AnsiLowerCase(OpenDialog.Filter);
  p := @Filter[1];
  i := 1;
  repeat
    p := StrScan(p, {$IFDEF MSWINDOWS}'|'{$ENDIF}{$IFDEF UNIX}'('{$ENDIF});
    if p <> nil then begin
      e := p + 1;
      p := StrScan(e, {$IFDEF MSWINDOWS}'|'{$ENDIF}{$IFDEF UNIX}')'{$ENDIF});
      if p <> nil then begin
        p^ := #0;
        Inc(p);
      end;
      repeat
        ee := StrScan(e, ';');
        if ee <> nil then
          ee^ := #0;
        if StrIComp(e, @Ext[1]) = 0 then begin
          OpenDialog.FilterIndex := i;
          Exit;
        end;
        e := ee + 1;
      until ee = nil;
    end;
    Inc(i);
  until p = nil;
  {$IFDEF UNIX}
  OpenDialog.DefaultExt := ''; { This only works properly if it is in the filter list }
  {$ENDIF}
  OpenDialog.FilterIndex := i-1;
end;

procedure TTabbedMainText.SetTabSpacingClick(Sender: TObject);
var
  s: string;
  i: Integer;
begin
  s := IntToStr(ActiveMemo.TabSpacing);
  if InputQuery('Set Tab Spacing', 'Tab spacing in characters', s) then
    for i := 0 to PageControl.PageCount - 1 do
      MemoOf(PageControl.Pages[i]).TabSpacing := SysUtils.StrToInt(s);
end;

procedure TTabbedMainText.StartFileatCursorClick(Sender: TObject);
var
  FileName, Dir: string;
begin
  FileName := FileAtCursor;
  if Length(FileName) > 0 then begin
    GetDir(0, Dir);
    {$IFDEF MSWINDOWS}
    if ShellExecute(0, nil, PChar(FileName), nil, PChar(Dir), SW_SHOW) <= 32 then
    // TODO: Equivalent of ShellExecute for Linux
    {$ENDIF}
      EditFileatCursorClick(Sender);
  end;
end;

procedure TTabbedMainText.StatusBarDblClick(Sender: TObject);
var
  x: Integer;
begin
  x := FLastMouseX - StatusBar.Panels[0].Width - StatusBar.Panels[1].Width;
  if x > 0 then begin
    Dec(x, StatusBar.Panels[2].Width);
    if x < 0 then
      with ActiveMemo do
        Modified := not Modified
    else if x < StatusBar.Panels[3].Width then
      TextRecordMacro.Execute;
  end;
end;

procedure TTabbedMainText.StatusBarMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FLastMouseX := X;
end;

procedure TTabbedMainText.SwitchToSheet(TabSheet: TTabSheet);
begin
  if Assigned(TabSheet) then
    PageControl.ActivePage := TabSheet
  else
    TabSheet := PageControl.ActivePage;
  TabSheetShow(TabSheet);
end;

procedure TTabbedMainText.TabbedNewExecute(Sender: TObject);
var
  TabSheet: TTabSheet;
begin
  TabSheet := MakeNewTab;
  MemoOf(TabSheet).SetFileName('');
  SwitchToSheet(TabSheet);
end;

procedure TTabbedMainText.TabbedOpenExecute(Sender: TObject);
var
  i: Integer;
begin
  if OpenDialog.Execute then
    for i := 0 to OpenDialog.Files.Count-1 do
      LoadFile(OpenDialog.Files[i]);
end;

procedure TTabbedMainText.PageControlCloseTabClicked(Sender: TObject);
var
  Memo: TFileMemo;
begin
  Memo := MemoOf(Sender as TTabSheet);
  FCloseButtonDown := -1;
  CloseMemo(Memo);
  PageControl.Repaint;
end;

procedure TTabbedMainText.TabbedSaveAllExecute(Sender: TObject);
var
  i: Integer;
begin
  for i := 0 to PageControl.PageCount - 1 do
    with MemoOf(PageControl.Pages[i]) do
      if Modified then
        if FileName = '' then
          SaveAs
        else
          Save;
end;

procedure TTabbedMainText.TabbedSaveAllUpdate(Sender: TObject);
var
  i: Integer;
begin
  if Slave.Checked then
    for i := 0 to PageControl.PageCount - 1 do
      with MemoOf(PageControl.Pages[i]) do
        if Modified then begin
          TabbedSaveAll.Enabled := True;
          {$IFDEF UNIX}
          SaveAllMsg.Status := 1;
          {$ENDIF}
          Exit;
        end;
  TabbedSaveAll.Enabled := False;
  {$IFDEF UNIX}
  SaveAllMsg.Status := 0;
  {$ENDIF}
end;

procedure TTabbedMainText.TabbedSaveExecute(Sender: TObject);
begin
  ActiveMemo.Save;
end;

procedure TTabbedMainText.TabbedSaveUpdate(Sender: TObject);
begin
  TabbedSave.Enabled := ActiveMemo.Modified;
end;

procedure TTabbedMainText.TabPopupMenuClick(Sender: TObject);
begin
  SwitchToSheet(PageControl.Pages[(Sender as TComponent).Tag]);
end;

procedure TTabbedMainText.TabPopupMenuPopup(Sender: TObject);
var
  i: Integer;
  M: TMenuItem;
begin
  for i := TabPopupMenu.Items.Count - 1 downto Nitem.MenuIndex+1 do
    TabPopupMenu.Items[i].Free;
  for i := 0 to PageControl.PageCount - 1 do begin
    M := TMenuItem.Create(TabPopupMenu);
    TabPopupMenu.Items.Add(M);
    with M do begin
      Caption := PageControl.Pages[i].Caption;
      Tag := i;
      OnClick := {$IFDEF FPC}@{$ENDIF}TabPopupMenuClick;
      Checked := i = PageControl.ActivePageIndex;
    end;
  end;
end;

procedure TTabbedMainText.TabSheetShow(Sender: TObject);
var
  Memo: TFileMemo;
begin
  Memo := MemoOf(Sender as TTabSheet);
  PageControl.Hint := Memo.FileName;
  if Visible then { LCL issue ? }
    Memo.SetFocus;
  AlienNewline.Checked := Memo.AlienNewline;
  {$IFDEF DRAWCLOSEBTN}
  SetLength(FTabBounds,0);
  {$ENDIF}
  if Memo.FileName <> '' then
    Application.Title :=  ExtractFileName(Memo.FileName)+' - Tabbed Slave Editor'
  else
    Application.Title := 'Tabbed Slave Editor';
  Caption := Application.Title;
  PageControl.Invalidate;
end;

function TTabbedMainText.TabWithFileName(const TabName: string): TTabSheet;
var
  i: Integer;
begin
  for i := 0 to PageControl.PageCount - 1 do
    if SysUtils.CompareText(MemoOf(PageControl.Pages[i]).Filename, TabName) = 0 then begin
      Result := PageControl.Pages[i];
      Exit;
    end;
  Result := nil;
end;

function TTabbedMainText.TabWithName(const TabName: string): TTabSheet;
var
  i: Integer;
begin
  for i := 0 to PageControl.PageCount - 1 do
    if SysUtils.CompareText(PageControl.Pages[i].Caption, TabName) = 0 then begin
      Result := PageControl.Pages[i];
      Exit;
    end;
  Result := nil;
end;

procedure TTabbedMainText.TryAndSave(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
var
  i: Integer;
begin
  Handled := True; { Assume no action required }
  if Slave.Checked then
    for i := 0 to PageControl.PageCount - 1 do
      with MemoOf(PageControl.Pages[i]) do
        if Modified then
          if Filename = '' then
            Handled := False { Name not set, so can't save }
          else
            try
              Save;
            except
              Handled := False; { Fail during save }
              raise;
            end;
  {$IFDEF UNIX}
  if Handled then
    SaveAllMsg.Status := 0; { Don't need to wait }
  {$ENDIF}
end;

procedure TTabbedMainText.UpdateActions;
var
  Memo: TFileMemo;
begin
  inherited;
  if not Assigned(PageControl.ActivePage) then
    Exit; { During creation/destruction }
  {$IFNDEF MSWINDOWS}
  with StatusBar do
    if SimplePanel then begin
      SimplePanel := False;
      Panels[1].Width := Canvas.TextWidth('9999 ');
      Panels[0].Width := 2*Panels[1].Width;
      Panels[2].Width := Canvas.TextWidth('Modified ');
      Panels[3].Width := Canvas.TextWidth('Recording ');
    end;
  {$ENDIF}
  Memo := MemoOf(PageControl.ActivePage);
  StatusBar.Panels[0].Text := IntToStr(Memo.CaretPos.y+1)+ ' ';
  StatusBar.Panels[1].Text := IntToStr(Memo.CaretPos.x) + ' ';
  if Memo.Modified then
    StatusBar.Panels[2].Text := 'Modified'
  else
    StatusBar.Panels[2].Text := '';
  if Memo.Recording then
    StatusBar.Panels[3].Text := 'Recording'
  else
    StatusBar.Panels[3].Text := '';
end;

{$IFDEF MSWINDOWS}
{$IFNDEF FPC}
procedure TTabbedMainText.WMDropFiles(var Message: TMessage);
begin
  HandleWMDropFiles(Message, FormDropFiles);
end;
{$ENDIF}

procedure TTabbedMainText.WMSettingChange(var Message: TMessage);
begin
  inherited;
  Monitor; { VCL bug work around - Check for number of monitors changing }
end;

procedure TTabbedMainText.WndProc(var Msg: TMessage);
var
  Handled: Boolean;
  WName: AnsiString;
  Data: PCopyDataStruct;
  TabSheet: TTabSheet;
begin
  if Msg.Msg = WM_GOTO_LINE then
    GotoLine(Self, Msg.wParam, Msg.lParam, Handled)
  else if Msg.Msg = WM_PGOPHER_ASK then begin
    if InSendMessage then begin
      ReplyMessage(272727);
      Sleep(0);
    end;
  end else if Msg.Msg = WM_COPYDATA then begin
    Data := PCopyDataStruct(Msg.LParam);
    if (Data^.dwData = WM_PGOPHER_ASK) and (Data^.cbData < 400) then
      try
        SetLength(WName, Data^.cbData-1);
        Move(Data^.lpData^,WName[1],Data^.cbData);
        TabSheet := TabWithFileName(WName);
        if Assigned(TabSheet) then begin
          SwitchToSheet(TabSheet);
          if InSendMessage then begin
            ReplyMessage(727272);
            Sleep(0);
          end;
        end else if InSendMessage then begin
          ReplyMessage(272727);
          Sleep(0);
        end;
      except
      end
    else if (Data^.dwData = WM_GOTO_LINE) and (Data^.cbData < 400) then
      try
        SetLength(WName, Data^.cbData-1);
        Move(Data^.lpData^,WName[1],Data^.cbData);
        LoadFile(WName);
      except
      end
  end else if Msg.Msg = WM_Save_All then begin
    Msg.result := 0;
    try
      TryAndSave(Self, 0, 0, Handled);
    finally
      if not Handled then
        if InSendMessage then begin
          ReplyMessage(272727);
          Sleep(0);
        end;
    end;
  end else
    inherited WndProc(Msg);       { dispatch normally }
end;
{$ELSE}
procedure TTabbedMainText.DoAsk(Sender: TObject; wParam, lParam: Integer; var Handled: Boolean);
var
  TabSheet: TTabSheet;
  WName: AnsiString;
begin
  {$IFDEF LCLcarbon}
  Exit; { TODO: impement inter-app messaging }
  {$ELSE}
  Handled := True; { Assume no action required }
  WName := GetRootWindowString(WM_PGOPHER_ASK);
  if WName <> '' then
    if lParam = 0 { Query } then begin
      TabSheet := TabWithFileName(WName);
      if Assigned(TabSheet) then begin
        SwitchToSheet(TabSheet);
        AskMessage.Status := 2; { Found }
        Exit;
      end;
    end else
      LoadFile(WName);
  AskMessage.Status := 1; { Not Found }
  {$ENDIF}
end;
{$ENDIF}

{ TFileMemo }

procedure TFileMemo.CheckAge;
begin
  if (LastFileAge <> -1) and (FileAge(Filename) > LastFileAge) then begin
    LastFileAge := FileAge(Filename); { Prevent re-entry }
    if MessageDlg(Filename+' on disk has changed. Reload from disk?', mtWarning, [mbYes, mbNo], 0) = mrYes then
      Reload;
  end;
end;

procedure TFileMemo.Close;
begin
  TabbedMainText.RecentFiles.AddItem(FileName);
  Owner.Free;
end;

procedure TFileMemo.LoadFile(AFileName: string; APreserveNewline: Boolean);
var
  Stream: TFileStream;
  Buffer: AnsiString;
  nSep: Integer;
begin
  Stream := nil;
  try
    Stream := TFileStream.Create(AFileName, fmOpenRead or fmShareDenyWrite);
    if APreserveNewline then begin
      { Guess line end from first few bytes }
      SetLength(Buffer, 255);
      SetLength(Buffer, Stream.Read(Buffer[1], Length(Buffer)));
      nSep := 0;
      if Pos(#10, Buffer) > 0 then Inc(nSep);
      if Pos(#13, Buffer) > 0 then Inc(nSep);
      Stream.Position := 0;
    end else
      nSep := NewLineLength;
    AlienNewline := nSep <> NewLineLength;
    RecordUndo := False;
    Lines.LoadFromStream(Stream);
    APreserveNewline := PreserveNewline;
    ResetModified;
  finally
    RecordUndo := True;
    Stream.Free;
  end;
end;

procedure TFileMemo.QueryClose(var CanClose: Boolean);
var
  MsgResult: Word;
begin
  CanClose := not Modified;
  if Modified then begin
    MsgResult := MessageDlg(Format('File %s has been modified. Save file?',
                            [FileName]), mtWarning, mbYesNoCancel, 0);
    CanClose := False;
    case MsgResult of
      mrYes:
        try
          Save;
          CanClose := not Modified;
        except
          {Required to work around a kylix bug}
          Application.HandleException(self);
        end;
      mrNo: CanClose := True;
    end;
  end;
  if CanClose then
    Close;
end;

procedure TFileMemo.Reload;
begin
  LoadFile(FileName, PreserveNewline);
end;

procedure TFileMemo.ResetModified;
begin
  Modified := False;
  ClearUndo;
  if FileName = '' then
    LastFileAge := -1
  else
    LastFileAge := FileAge(FileName);
end;

procedure TFileMemo.Save;
begin
  if FileName = '' then
    SaveAs
  else begin
    SaveToFile(Filename);
    ResetModified;
  end;
end;

procedure TFileMemo.SaveAs;
var
  OtherTab: TTabSheet;
  OtherMemo: TFileMemo;
begin
  with TabbedMainText.SaveDialog do begin
    Filter := TabbedMainText.OpenDialog.Filter;
    FilterIndex := TabbedMainText.OpenDialog.FilterIndex;
    DefaultExt := ExtractFileExt(FileName);
  end;
  TabbedMainText.SaveDialog.FileName := FileName;
  if TabbedMainText.SaveDialog.Execute then begin
    OtherTab := TabbedMainText.TabWithName(ExpandFileName(TabbedMainText.SaveDialog.FileName));
    if Assigned(OtherTab) then begin
      OtherMemo := TabbedMainText.MemoOf(OtherTab);
      if mrYes = MessageDlg(Format('File %s is already open in another window, and has been modified. Continue?',
                                  [TabbedMainText.SaveDialog.FileName]), mtWarning, mbYesNo, 0) then
        OtherMemo.Close
      else
        Exit;
    end;
    SaveToFile(TabbedMainText.SaveDialog.FileName);
    SetFileName(TabbedMainText.SaveDialog.FileName);
    ResetModified;
    if TabSheet = TabbedMainText.PageControl.ActivePage then
      TabbedMainText.SwitchToSheet(TabSheet);
  end;
end;

procedure TFileMemo.SaveToFile(const AFileName: string);
begin
  Lines.SaveToFileEnh(AFileName, TabbedMainText.KeepBackup.Checked, AlienNewline);
end;

procedure TFileMemo.SetFileName(AFileName: string);
var
  i: Integer;
  NewCaption: string;
  MainWindow: TTabbedMainText;
begin
  MainWindow := Owner.Owner as TTabbedMainText;
  Filename := AFileName;
  if Length(FileName) = 0 then begin
    NewCaption := '[Untitled]';
    i := 0;
    while MainWindow.TabWithName(NewCaption) <> nil do begin
      Inc(i);
      NewCaption := Format('[Untitled %d]', [i]);
    end;
    TabSheet.Caption := NewCaption;
    LastFileAge := -1;
  end else begin
    {$IFDEF LCL_}
      TabSheet.Caption := ExtractFileName(GetLongName(FileName));
    {$ELSE}
      TabSheet.Caption := GetLongName(FileName);
    {$ENDIF}
  end;
//  SetDefaultExt(ExtractFileExt(Caption));
//  SetCurrentDir(PChar(ExtractFilePath(FileName)));
end;

function TFileMemo.TabSheet: TTabSheet;
begin
  Result := Owner as TTabSheet;
end;

end.
