unit CEig;
{$I utils.inc}
interface

{ Eigenvalues of an Hermitiian matrix }

uses MathSup, EigLA, MatrixVec, ComplexMatrix;

procedure CEigen(Eigenvals: TColumnVector; Matrix: TComplexMatrix; EvReqd: Boolean; Offset: Integer = 0);

function zheevr( { Simplified interface to LAPACK routine }
  n: Integer;
  var w: Double;
  var a, z: Complex;
  WantEigenVectors: Boolean = True;
  UpperTriangle: Boolean = False;
  Range: TRangeWanted = rwAll;
  VL: Double = 0;  VU:Double = 0;
  IL: Integer = 0; IU: Integer = 0;
  AbsTol:Double = 0
): Integer;

implementation

uses SysUtils, Profile, Lapack, Vector, Worker;

function zheevr( { Simplified interface to LAPACK routine }
  n: Integer;
  var w: Double;
  var a, z: Complex;
  WantEigenVectors: Boolean = True;
  UpperTriangle: Boolean = False;
  Range: TRangeWanted = rwAll;
  VL: Double = 0;  VU:Double = 0;
  IL: Integer = 0; IU: Integer = 0;
  AbsTol:Double = 0
): Integer;
var
  jobz, rnge, uplo, ss: AnsiChar;
  isuppz, iwork: array of Integer;
  work: array of Complex;
  rwork: array of Double;
  twork: Double;
  info, tiwork, lwork, liwork, lrwork: Integer;
  cwork: Complex;
begin
  Result := 0;
  if WantEigenVectors then
    jobz := 'V'
  else
    jobz := 'N';
  case Range of
    rwAll: rnge := 'A';
    rwInterval: rnge := 'V';
    rwIndex: rnge := 'I'
  end;
  if UpperTriangle then
    uplo := 'U'
  else
    uplo := 'L';
  if AbsTol = 0 then begin
    ss := 'S';
    AbsTol := dlamch_(ss, 1);
  end;
  if Range <> rwAll then
    SetLength(isuppz, n);
  lwork := -1;
  liwork := -1;
  lrwork := -1;
  zheevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, tiwork, cwork, lwork, twork, lrwork, tiwork, liwork, info, 1, 1, 1);
  if info = 0 then begin
    SetLength(work, Round(cwork.re));
    liwork := tiwork;
    SetLength(iwork, liwork);
    SetLength(rwork, Round(twork));
    lwork := Length(work);
    lrwork := Length(rwork);
    if (Range = rwAll) or ((Range = rwIndex) and (IU-IL = n-1)) then begin
      SetLength(isuppz, 2*n);
      zheevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, isuppz[0], work[0], lwork, rwork[0], lrwork, iwork[0], liwork, info, 1, 1, 1)
    end else
      zheevr_(jobz, rnge, uplo, n, a, n, VL, VU, IL, IU, AbsTol, Result, w, z, n, tiwork,    work[0], lwork, rwork[0], lrwork, iwork[0], liwork, info, 1, 1, 1);
  end;
  if info < 0 then
    raise Exception.CreateFmt('zheevr: internal error %d', [-info])
  else if info > 0 then
    Result := -Info;
end;

var
 EigenValuesTime: string = 'Complex Matrix Diagonalization';

procedure CEigen(Eigenvals: TColumnVector; Matrix: TComplexMatrix; EvReqd: Boolean; Offset: Integer = 0);
var
  info: Integer;
  EigenVectors: TVector;
  PEigenVectors: Pcomplex;
  PEigenVals: PDouble;
begin
  if Matrix.ColCount = 0 then exit;
  if Matrix.Mode = cmReal then
    Eigen(Eigenvals, Matrix, EvReqd, Offset)
  else begin
    TStart(EigenValuesTime);
    Matrix.Mode := cmNComplex;
    if EvReqd then begin
      EigenVectors := TComplexMatrix.Create(Matrix.RowCount, Matrix.ColCount, cmNComplex);
      PEigenVectors := Pcomplex(EigenVectors.FirstItem);
    end else begin
      EigenVectors := nil;
      PEigenVectors := nil; { Separate variable to avoid using EigenVectors below }
    end;
    PEigenVals := Eigenvals.FirstItem;
    Inc(PEigenVals, Offset);
    try
      info := zheevr(Matrix.ColCount, PEigenvals^, Pcomplex(Matrix.FirstItem)^, PEigenVectors^, EvReqd);
      if info < 0 then
        raise Exception.CreateFmt('Eigen: internal error in zheevr %d', [-info])
      else if info <> Matrix.RowCount then
        raise Exception.CreateFmt('Eigen: Only %d of %d eigenvalues found', [info, Matrix.RowCount]);
    finally
      if Assigned(EigenVectors) then
        Matrix.AssignAndNil(EigenVectors);
    end;
    TEnd(EigenValuesTime);
  end;
end;

end.

