unit CODATA;
{ 2010 CODATA recommended values from http://physics.nist.gov/ }
interface

const
  { 2006 CODATA recommended values in comments }
  ge: Double = 2.00231930436153 { 2.0023193043622 };
  BohrMagneton: Double = 927.400968e-26 { 927.400915e-26} ; { e hcross / 2 / me }
  SpeedOfLight: Double = 299792458.0;
  Planck: Double = 6.62606957e-34 { 6.62606896e-34 };
  Avagadro: Double = 6.02214129e23 { 6.02214179e23 };
  ElectronCharge: Double = 1.602176565e-19;
  ElectronMass: Double = 9.10938291e-31;
  BoltzmannConstant: Double = 1.3806488e-23;
  BohrRadius: Double = 0.52917721092e-10 { 0.52917720859e-10 };
  MeOverMp: Double = 5.4461702178e-4 { 5.4461702177e-4 };
  MnOverMp: Double = 1.00137841917;
  GravitationalConstant: Double = 6.67384e-11;
  FundamentalConstantsSource: string = 'CODATA 2010';

var
  { Computed derived constants to avoid discrepancies }
  hCross: Double;
  eVTocm1: Double; { 8065.54429 8065.54465 }
  KelvinToMHz: Double; { 20836.618 20836.644 }
  cm1toJ: Double; { 19.8644568326930306e-24 = 100 h c }
  amuAngstoms2toSI: Double; { 16.6053892103218986e-48 }

implementation

uses MathSup;

initialization
  hCross := Planck / Two / Pi;
  eVTocm1 := ElectronCharge / Planck / SpeedOfLight / 100;
  KelvinToMHz := BoltzmannConstant / Planck / 1000000;
  cm1toJ := SpeedOfLight * Planck * 100;
  amuAngstoms2toSI := 1e-23 / Avagadro;
end.
