unit CharType;

interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
  {$IFDEF MSWINDOWS} uses Windows; {$ENDIF}
  {$IFDEF UNIX}
  type Bool = Boolean;
  function IsCharAlpha(ch: Char): Bool;
  function IsCharAlphaNumeric(ch: Char): Bool;
  { IsCharAlphaNumeric and IsCharAlpha are available in WinProcs/Windows unit }
  {$ENDIF}
  type
  CharFilter = function (c: Char):Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
  function IsFileChar(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
  function IsNonSpaceFileChar(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
  function IsCharIdentifier(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}

  function UpperChar(c: Char): Char;
  function LowerChar(c: Char): Char;

implementation

{$IFDEF UNIX}
// TODO: generalize to other charsets
function IsCharAlpha(ch: Char): Bool;
begin
  Result := ch in ['A'..'Z', 'a'..'z', '_'];
end;

function IsCharAlphaNumeric(ch: Char): Bool;
begin
  Result := ch in ['A'..'Z', 'a'..'z', '_', '0'..'9'];
end;

function LowerChar(c: Char): Char;
begin
  Result := c;
  if Result in ['A'..'Z'] then
    Result := Char(Ord(Result)+Ord('a')-Ord('A'));
end;

function UpperChar(c: Char): Char;
begin
  Result := c;
  if Result in ['a'..'z'] then
    Result := Char(Ord(Result)+Ord('A')-Ord('a'));
end;
{$ENDIF}

function IsFileChar(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
begin
  Result := (c {$IFNDEF VER80} >= {$ELSE} > {$ENDIF} ' ') and (c <> '"') and (c <> '''') and (c <> '<')
            and (c <> '>') and (c <> ',') ;
end;

function IsNonSpaceFileChar(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
begin
  Result := (c <> ' ') and IsFileChar(c);
end;

function IsCharIdentifier(c: Char): Bool; {$IFDEF WIN32} stdcall; {$ENDIF}
begin
  Result := (c = '_') or IsCharAlphaNumeric(c);
end;

{$IFDEF MSWINDOWS}
function UpperChar(c: Char): Char;
begin
  {$IFDEF UNICODE}
    Result := Char(CharUpper(PChar(Integer(Ord(c))))); { Note that the absence of an address is correct }
  {$ELSE}
    Result := Char(AnsiUpper(PChar(Integer(Ord(c))))); { Note that the absence of an address is correct }
  {$ENDIF}
end;

function LowerChar(c: Char): Char;
begin
  {$IFDEF UNICODE}
    Result := Char(CharLower(PChar(Integer(Ord(c))))); { Note that the absence of an address is correct }
  {$ELSE}
    Result := Char(AnsiLower(PChar(Integer(Ord(c)))));
  {$ENDIF}
end;
{$ENDIF}
end.
