unit FuncFit;
{$I utils.inc}{ Copyright (C) 2011 C M Western; see UCompat.pas for license}
interface

{ Stirpped down version of LeastSQ routine, using analytic derivatives }

type
  TFitableFunction = function(x: Double; Derivative: Integer; const Parameters: array of Double):Double;

function FitOnce(f: TFitableFunction; var Parameters: array of Double; const x, y: array of Double):Double;
function FitMaxErr(f: TFitableFunction; var Parameters: array of Double; const x, y: array of Double):Double;

implementation

uses Math, MatrixVec, LapackLSQ;

function FitOnce(f: TFitableFunction; var Parameters: array of Double; const x, y: array of Double):Double;
var
  omc: TColumnVector;
  A: TMatrix;
  i, j: Integer;
begin
  A := nil;
  omc := TColumnVector.Create(Length(x));
  try
    A := TMatrix.Create(Length(x), Length(Parameters));
    for i := 0 to High(x) do begin
      omc[i+1] := y[i] - f(x[i],0,Parameters);
      for j := 1 to Length(Parameters) do
        A[i+1,j] := f(x[i],j,Parameters);
    end;
//    A.Write('A');
//    omc.Write('omc');
    dgels(A,omc);
    Result := 0;
    for i := 0 to High(Parameters) do begin
      Result := Max(Result, Abs(omc[i+1])/(Max(Abs(Parameters[i]), Max(Abs(omc[i+1]), 1))));
      Parameters[i] := Parameters[i] + omc[i+1];
    end;
  finally
    A.Free;
    omc.Free;
  end;
end;

function FitMaxErr(f: TFitableFunction; var Parameters: array of Double; const x, y: array of Double):Double;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to High(x) do
    Result := Max(Result, Abs(y[i]-f(x[i],0,Parameters)));
end;

end.

