unit IntVec;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

uses Huge;

type

  TIntVector = class(TGVectorN)
  private
    function GetValue(i:Longint):Longint;
    procedure SetValue(i:Longint; v:Longint);
    procedure SetSize(n:Longint);
  public
    constructor Create(ASize: Longint);
    procedure Clear;
    procedure Move(Old, ANew:Longint);
    property Size read GetSize write SetSize;
    property Value[i:Longint]:Longint read GetValue write SetValue; default;
    procedure ClearRange(First, Count:Longint);
    function FirstItem: PLongint;
    function IndexOf(AValue: Longint): Longint;
  end;

  TIntColumnVector = class(TIntVector) { As TVector, but first index is 1 }
  public
    constructor Create(ASize: Longint);
  end;

implementation

uses SysUtils;

constructor TIntVector.Create(ASize: Longint);
begin
  inherited Create(SizeOf(Longint));
  SetSize(ASize);
end;

constructor TIntColumnVector.Create(ASize: Longint);
begin
  inherited Create(ASize);
  Offset := 1;
end;

function TIntVector.FirstItem: PLongint;
begin
  Result := Base;
end;

procedure TIntVector.ClearRange(First, Count:Longint);
{$IFNDEF VER80}
begin
  FillChar(Plongint(Self.Item[First])^, (Count-First)*SizeOf(Longint), 0);
end;
{$ELSE}
var
  p: Pointer;
  n: Longint;
begin
  p := Self.Item[First];
  Count := Count * 4;
  while Count > 0 do begin
    n := 65536-Ofs(p^);
    if n > Count then n := Count;
    if n > 32768 then n := 32768;
    FillChar(p^, n, 0);
    Dec(Count, n);
    HugeInc(p, n);
    Count := Count - n;
  end;
end;
{$ENDIF}

procedure TIntVector.SetSize(n:Longint);
var
  Old: Longint;
begin
  Old := Size;
  inherited Size := n;
  if Size > Old then
    ClearRange(Old+Offset, Size-Old);
end;

procedure TIntVector.Clear;
begin
  ClearRange(Offset, Size);
end;

function TIntVector.GetValue(i:Longint):Longint;
begin
  Result := PLongint(Item[i])^;
end;

procedure TIntVector.SetValue(i:Longint; v:Longint);
begin
  PLongint(Item[i])^ := v;
end;

procedure TIntVector.Move(Old, ANew:Longint);
var
  i: Longint;
  step: Integer;
  vold: Integer;
begin
  if Old > ANew then
    step := -1
  else
    step := 1;
  i := Old;
  vold := Value[i];
  while i <> ANew do begin
    Value[i] := Value[i+step];
    i := i + step;
  end;
  Value[ANew] := vold
end;

function TIntVector.IndexOf(AValue: Integer): Longint;
var
  p: PLongInt;
  n: LongInt;
begin
  n := Size;
  p := FirstItem;
  while (n > 0) and (p^ <> AValue) do begin
    Inc(p);
    Dec(n);
  end;
  if n > 0 then
    Result := Size - n + Offset
  else
    Result := Offset - 1;
end;

end.
