unit LongName;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

function GetLongName(const name: string): string;
{ Try and get the long name form of the given file }

implementation

{$IFDEF UNIX}
uses SysUtils;

function GetLongName(const name: string): string;
{$IFNDEF FPC}
var
  Match: TFilenameCaseMatch;
{$ENDIF}
begin
{$IFNDEF FPC}
  Result := ExpandFileNameCase(name, Match);
  if (Match = mkNone) or (Match = mkAmbiguous) then
{$ENDIF}
{TODO: This for FPC?}
    Result := name;
end;
{$ENDIF}

{$IFDEF MSWINDOWS}
uses Windows, SysUtils;

function GetLongName(const name: string): string;
var
  h: Cardinal;
  w: WIN32_FIND_DATA;
begin
  h := Windows.FindFirstFile(Pchar(name), w);
  if h <> INVALID_HANDLE_VALUE then begin
    Windows.FindClose(h);
    Result := ExtractFilePath(name);
    if not ( (Result = '\\') or (Result[Length(Result)] = ':')
             or ( (Length(Result) = 3) and (Copy(Result, 2, 2) = ':\') ) ) then begin
      Result := GetLongName(Copy(Result, 1, Length(Result)-1));
      Result :=  Result + '\';
    end;
    Result :=  Result + w.cFileName;
  end else
    Result := name;
end;
{$ENDIF}

{$IFDEF VER80}
uses Wintypes, WinProcs, SysUtils;

var
  FreeLibrary32W: function(handle:Longint):bool;
  GetProcAddress32W: function(module:Longint; funcname:PChar):Pointer;
  LoadLibraryEx32W: function(libname:PChar; hFile, dwFlags:Longint):Longint;

  CallProc32W2: function(p1, p2: Longint; lpProcAddress:Pointer; fAddressConvert, nParams: Longint):Longint;
  CallProc32W1: function(p: Longint; lpProcAddress:Pointer; fAddressConvert, nParams: Longint):Longint;

  State: (Connected, Win31, NotConnected);
  pFindFirstFile, pFindNextFile, pFindClose: Pointer;

const
  INVALID_HANDLE_VALUE = -1;
  MAX_PATH = 260;

type
  tFILETIME=record
    dwLowDateTime,
    dwHighDateTime:longint;
  end;
  pFILETIME = ^tFILETIME;

  WIN32_FIND_DATA = record
    dwFileAttributes: longint;
    ftCreationTime: TFileTime;
    ftLastAccessTime: TFileTime;
    ftLastWriteTime: TFileTime;
    nFileSizeHigh: longint;
    nFileSizeLow: longint;
    dwReserved0: longint;
    dwReserved1: longint;
    cFileName: array[0..MAX_PATH - 1] of Char;
    cAlternateFileName: array[0..13] of Char;
  end;
  pWIN32_FIND_DATA=^WIN32_FIND_DATA;

procedure Connect;
var
  hinstKernel: THandle;
  hKernel32: Longint;
begin
  if State = NotConnected then begin
    State := Win31; { Assume error }
    hinstKernel := LoadLibrary('KERNEL');
    if (hinstKernel < HINSTANCE_ERROR) then exit;

    @CallProc32W2       := GetProcAddress(hinstKernel, 'CALLPROC32W');
    @FreeLibrary32W    := GetProcAddress(hinstKernel, 'FREELIBRARY32W');
    @LoadLibraryEx32W  := GetProcAddress(hinstKernel, 'LOADLIBRARYEX32W');
    @GetProcAddress32W := GetProcAddress(hinstKernel, 'GETPROCADDRESS32W');
    @CallProc32W1 := @CallProc32W2;

    if (@CallProc32W2=nil) or (@FreeLibrary32W=nil) or (@LoadLibraryEx32W=nil) or
       (@GetProcAddress32W=nil) then
      Exit;

    hKernel32 := LoadLibraryEx32W('kernel32', 0, 0);
    if hKernel32 <> 0 then begin
      pFindFirstFile := GetProcAddress32W(hKernel32, 'FindFirstFileA');
      pFindNextFile := GetProcAddress32W(hKernel32, 'FindNextFileA');
      pFindClose := GetProcAddress32W(hKernel32, 'FindClose');
      FreeLibrary32W(hKernel32); { Assume nobody unloads it... }
      if (pFindFirstFile = nil) or (pFindNextFile = nil) or (pFindClose = nil) then
        Exit;
    end;

    State := Connected;
  end;
end;

function GetLongName(const name: string): string;
var
  Handle: Longint;
  w: WIN32_FIND_DATA;
  n: Integer;
  dir: string;
begin
  if State = NotConnected then Connect;
  Result := AnsiLowerCase(name);
  if (Length(Result) < 255) and (State = Connected) then begin
    Result[Length(Result)+1] := #0;
    Handle := CallProc32W2(Longint(@Result[1]), Longint(@w), pFindFirstFile, 3, 2);
    if Handle <> INVALID_HANDLE_VALUE then begin
      CallProc32W1(Handle, pFindClose, 0, 1);
      dir := ExtractFilePath(Result);
      if not ( (dir = '\\') or (dir[Length(dir)] = ':')
               or ( (Length(dir) = 3) and (Copy(dir, 2, 2) = ':\') ) ) then begin
        dir := GetLongName(Copy(dir, 1, Length(dir)-1));
        AppendStr(dir, '\');
      end;
      if Length(dir) + StrLen(@w.cFileName[0]) <= 255 then
        Result := Concat(dir, StrPas(@w.cFileName[0]));
    end;
  end;
end;

begin
  State := NotConnected;
{$ENDIF}
end.

