{ Loosely based on Z_Prof tool by A. Baars (antonie.baars@wxs.nl) }

unit Profile;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface

procedure TStart(var s:string);
procedure TEnd(var s:string);

procedure CalculateOverhead(mSecMin: Integer = 0 { Minimum number of milliseconds to time for });
procedure ClearTiming;
procedure ComputeTotalTimings;
procedure WriteTimingInfo(var F: TextFile);

procedure SetTrackMemory(Track: Boolean);

function IsTrackingMemory: Boolean;

type
  TMark = record
    Passes: Longint;
    Start, PrevTime, TotalTime, Fastest: int64;
    StartAllocs, StartFrees, Allocs, Frees: Integer;
    Active: boolean;
  end;

var
  Marks: array { ThreadIndex } of array { TimingName } of TMark;
  TimingNames: array of string;
  ThreadNames: array of string;
  OverheadCyclesInteger: Integer; { Time taken between null TStart() .. TEnd() }
  OverheadCycles: Double;         { Time taken between null TStart() .. TEnd() }
  ExternalOverheadCycles: Double; { Time to perform null TStart() .. TEnd() }
  MarkLevel: Integer;
  ShowTimingAtEnd: Boolean = False;
  TimeChanged: Boolean = False;

procedure GrowMarks(Increment: Integer; LockHeld: Boolean = False);

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} { For TRTLCriticalSection }
     Worker, UCompat, Clock, Math, SysUtils;

var
  TrackingMemory: Boolean = False;
  nAllocations: Integer = 0;
  nFree: Integer = 0;
  MemoryManagerEx: {$IFNDEF FPC}TMemoryManagerEx{$ELSE}TMemoryManager{$ENDIF};

type
{$IFDEF FPC}
  {$IF defined(VER2) and (defined(VER2_1) or defined(VER2_2))}
    PtrMem = PtrInt;
  {$ELSE}
    PtrMem = PtrUInt;
  {$IFEND}
{$ELSE}
  PtrMem = Integer;
{$ENDIF}


function IsTrackingMemory: Boolean;
begin
  Result := TrackingMemory;
end;

function TrackGetMem(Size: PtrMem): Pointer;
begin
  Inc(nAllocations);
  Result := MemoryManagerEx.GetMem(Size);
end;

function TrackFreeMem(P: Pointer): PtrMem;
begin
  Inc(nFree);
  Result := MemoryManagerEx.FreeMem(P);
end;

function TrackReallocMem({$IFDEF FPC}var{$ENDIF}P: Pointer; Size: PtrMem): Pointer;
begin
  Inc(nAllocations);
  Result := MemoryManagerEx.ReallocMem(P, Size);
end;

function TrackAllocMem(Size: PtrMem): Pointer;
begin
  Inc(nAllocations);
  Result := MemoryManagerEx.AllocMem(Size);
end;

{$IFDEF FPC}
function TrackFreememSize(p:pointer; Size:PtrMem):PtrMem;
begin
  Inc(nFree);
  Result := MemoryManagerEx.FreememSize(p,Size);
end;
{$ENDIF}

procedure SetTrackMemory(Track: Boolean);
var
  M: {$IFNDEF FPC}TMemoryManagerEx{$ELSE}TMemoryManager{$ENDIF};
begin
  if TrackingMemory <> Track then
    if Track then begin
      GetMemoryManager(MemoryManagerEx);
      M := MemoryManagerEx;
      M.GetMem := @TrackGetMem;
      M.FreeMem := @TrackFreeMem;
      M.AllocMem := @TrackAllocMem;
      {$IFDEF FPC}
      M.FreememSize := @TrackFreememSize;
      {$ENDIF}
      M.ReallocMem := @TrackReallocMem;
      SetMemoryManager(M);
    end else begin
      SetMemoryManager(MemoryManagerEx);
    end;
  TrackingMemory := Track;
end;

var
  TimingLock: TRTLCriticalSection;
  TimingLockAvailable: Boolean = False;

procedure GrowMarks(Increment: Integer; LockHeld: Boolean = False);
var
  I, Old: Integer;
begin
  if not LockHeld and TimingLockAvailable then
    EnterCriticalSection(TimingLock);
  try
    SetLength(TimingNames, Length(TimingNames)+Increment);
    { Make sure all arrays of TMark have the same length }
    for I := 0 to High(Marks) do begin
      Old := Length(Marks[I]);
      SetLength(Marks[I], Length(TimingNames));
      if Length(TimingNames) > Old then
        FillChar(Marks[I][Old], (Length(TimingNames)-Old)*SizeOf(TMark), 0);
    end;
  finally
    if not LockHeld and TimingLockAvailable then
      LeaveCriticalSection(TimingLock);
  end;

end;

function TSection(var s:string):Integer;
var
  Stub: string;
begin
  if TimingLockAvailable then
    EnterCriticalSection(TimingLock);
  try
    if (Length(s) <> 4) or (s[4] <> #0) then begin
      { Search through current entries }
      Result := 0;
      while Result <= High(TimingNames) do begin
        if (s = TimingNames[Result]) or (TimingNames[Result] = '') then
          Break;
        Inc(Result);
      end;
      if Result > High(TimingNames) then
        GrowMarks(20, True);
      if TimingNames[Result] = '' then begin
        { New entry }
        TimingNames[Result] := s;
      end;
      SetLength(Stub,4);
      PInteger(@Stub[1])^ := Result;
      {$IFDEF UNICODE}
      PInteger(@Stub[3])^ := 0;
      {$ENDIF}
      s := Stub;
      Stub := ''; { Necessary for race avoidance }
    end else
      Result := PInteger(s)^; { Another process did it }
  finally
    if TimingLockAvailable then
      LeaveCriticalSection(TimingLock);
  end;
end;

{.$DEFINE DEBUG_TIME}

procedure TStart(var s:string);
var
  m: Integer;
begin
  if Length(s) <> 4 then
    m := TSection(s)
  else
    m := PInteger(s)^;
  if m > High(Marks[ThreadIndex]) then
    m := TSection(s); { Marks being expanded; force wait }
  with Marks[ThreadIndex][m] do
    {$IFDEF DEBUG_TIME}
    if Active then begin
      Active := False;
      raise Exception.CreateFmt('TStart(%s)', [TimingNames[m]])
    end else
    {$ENDIF}
    begin
      if TrackingMemory then begin
        StartAllocs := nAllocations;
        StartFrees := nFree;
      end;
      Active := True;
      Inc(MarkLevel);
      Start := Clocks();
    end;
end;

procedure TEnd(var s: string);
var
  t: int64;
  m: Integer;
begin
  t := Clocks();
  if Length(s) <> 4 then
    m := TSection(s)
  else
    m := PInteger(s)^;
  if m > High(Marks[ThreadIndex]) then
    m := TSection(s); { Marks being expanded; force wait }
  with Marks[ThreadIndex][m] do
    {$IFDEF DEBUG_TIME}
    if not Active then
      raise Exception.CreateFmt('TEnd(%d)', [m])
    else
    {$ENDIF}
    begin
      PrevTime := t - Start - OverheadCyclesInteger;
      Active := False;
      TotalTime := TotalTime + PrevTime;
      if (PrevTime < Fastest) or (Passes = 0) then
        Fastest := PrevTime;
      Inc(Passes);
      Dec(MarkLevel);
      if MarkLevel <= 0 then begin
        TimeChanged := True;
        MarkLevel := 0;
      end;
      if TrackingMemory then begin
        Inc(Frees, nFree - StartFrees);
        Inc(Allocs, nAllocations - StartAllocs);
      end;
    end;
end;

procedure CalculateOverhead(mSecMin: Integer{ Minimum number of milliseconds to time for });
var
  s, s1: string;
  i, n, nCycles: Integer;
  Start: TDateTime;
begin
  nCycles := 10000;
  repeat
    Start := Now;
    s1 := '__Overhead';
    n := TSection(s1);
    s := '';
    TSection(s); { Use a blank entry }
    OverheadCyclesInteger := 0;
    TStart(s1);
    for i := 1 to nCycles do begin
      TStart(s); TEnd(s);
    end;
    TEnd(s1);
    with Marks[ThreadIndex][TSection(s)] do begin
      OverheadCycles := TotalTime/nCycles;
      OverheadCyclesInteger := Round(OverheadCycles);
      Passes := 0;
      TotalTime := 0;
    end;
    ExternalOverheadCycles := Marks[ThreadIndex][n].PrevTime / nCycles;
    TimingNames[n] := '';
    TimingNames[n+1] := '';
    nCycles := nCycles*10;
  until (nCycles > 1000000) or ((Now-Start)*SecsPerDay >= mSecMin/1000);
end;

procedure ComputeTotalTimings;
var
  i, t: Integer;
begin
  { Compute overall totals }
  for i := 0 to High(TimingNames) do begin
    FillChar(Marks[0][i], SizeOf(TMark), 0);
    Marks[0][i].Fastest := Marks[1][i].Fastest;
    for t := 1 to High(ThreadNames) do begin
      Inc(Marks[0][i].Passes, Marks[t][i].Passes);
      Inc(Marks[0][i].TotalTime, Marks[t][i].TotalTime);
      if Marks[0][i].Fastest > Marks[t][i].Fastest then
        Marks[0][i].Fastest := Marks[t][i].Fastest;
      Inc(Marks[0][i].Allocs, Marks[t][i].Allocs);
      Inc(Marks[0][i].Frees, Marks[t][i].Frees);
    end;
  end;
end;

procedure WriteTimingInfo(var F: TextFile);
var
  Musec: Double;
  procedure ShowEntry(Mark: TMark; ShowPrev: Boolean; const DisplayName: string);
  begin
    with Mark do begin
      Write(F, Passes:7, ' ', Fastest/Musec:10:1, ' ');
      if ShowPrev then
        Write(F, PrevTime/Musec:10:1)
      else
        Write(F, ' ':10);
      Passes := Max(Passes,1); { Avoid divide by zero in displaying zero }
      Write(F, ' ', TotalTime/Passes/Musec:10:1, ' ', TotalTime/Musec/1000:10:3, ' ');
      if TrackingMemory then
        Write(F, Allocs:8, Frees:8, ' ');
      Write(F, DisplayName);
      if Active then
        WriteLn(F, ' (unclosed)')
      else
        WriteLn(F);
    end;
  end;
var
  i, t: Integer;
begin
  Musec := ClockFrequency/1e6;
  CalculateOverhead;
  ComputeTotalTimings;
  Write(F, 'Timing (microseconds) using a ');
  if ClockFrequency > 1e6 then
    WriteLn(F, MuSec:10:1,' MHz Clock')
  else
    WriteLn(F, Round(ClockFrequency),' Hz Clock');
  Write(F, '  Count    Fastest       Last    Average Total/mSec');
  if TrackingMemory then
    Write(F, '   Allocs   Frees');
  WriteLn(F);
  for i := 0 to High(TimingNames) do begin
    if TimingNames[i] = '' then
      Break;
    if Length(ThreadNames) = 2 then begin
      if Marks[1][i].Passes > 0 then
        ShowEntry(Marks[1][i], True, TimingNames[i]);
    end else begin
      if Marks[0][i].Passes > 0 then begin
        for t := 1 to High(ThreadNames) do
          if Marks[t][i].Passes > 0 then
            ShowEntry(Marks[t][i],True,TimingNames[i]+':'+ThreadNames[t]);
        ShowEntry(Marks[0][i],False,TimingNames[i]);
        Writeln(F);
      end;
    end;
  end;
  WriteLn(F, 'Overhead/ns:', OverheadCycles/Musec*1000:5:0, ExternalOverheadCycles/Musec*1000:5:0);
end;

procedure ClearTiming;
var
  I: Integer;
begin
  for I := 0 to High(Marks) do
    if Length(Marks[I]) > 0 then
      FillChar(Marks[I][0], Length(Marks[I])*SizeOf(TMark), 0);
end;

initialization
  InitCriticalSection(TimingLock);
  TimingLockAvailable := True;
finalization
  if ShowTimingAtEnd and (Length(Marks) > 0) then
    WriteTimingInfo(Output);
  TimingLockAvailable := False;
  DoneCriticalSection(TimingLock);
end.
