unit PropHelp;

{$I utils.inc}

interface

uses TypInfo;

function PropertyToStr(AObject: TObject; PropInfo: PPropInfo): string; overload;
function PropertyToStr(AObject: TObject; const PropName: string): string; overload;
procedure SetPropertyFromStr(AObject: TObject; PropInfo: PPropInfo; const s: string);  overload;
procedure SetPropertyFromStr(AObject: TObject; const PropName: string; const s: string); overload;
function TryStrToBoolean(s: String; out Value: Boolean): Boolean;
function StrToBoolean(s: String): Boolean;

implementation

uses SysUtils;

{$IFNDEF FPC}
  {$DEFINE PTP}
{$ENDIF}

function GetPropType(PropInfo: PPropInfo): PTypeInfo;
begin
  Result := PropInfo^.PropType{$IFDEF PTP}^{$ENDIF};
end;

function FloatPropertyToStr(AObject: TObject; PropInfo: PPropInfo): string;
const
  Precisions: array[TFloatType] of Integer = (7, 15, 18, 18, 19);
begin
  Result := FloatToStrF(GetFloatProp(AObject, PropInfo), ffGeneral,
                        Precisions[GetTypeData(GetPropType(PropInfo))^.FloatType], 0);
end;

function PropertyToStr(AObject: TObject; const PropName: string): string;
var
  PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(AObject, PropName);
  if not Assigned(PropInfo) then
    raise Exception.Create(PropName + ' not found in ' + AObject.ClassName);
  Result := PropertyToStr(AObject, PropInfo);
end;

function PropertyToStr(AObject: TObject; PropInfo: PPropInfo): string;
begin
  Result := '';
  case PropInfo^.PropType^.Kind of
    tkInteger: Result := IntToStr(GetOrdProp(AObject, PropInfo));
    {$IFDEF FPC}
    tkBool:    if GetOrdProp(AObject, PropInfo) <> 0 then
                 Result := 'True'
               else
                 Result := 'False';
    {$ENDIF}
    tkWChar,
    tkChar: Result := Char(GetOrdProp(AObject, PropInfo));
    tkEnumeration: Result := GetEnumProp(AObject, PropInfo);
    tkFloat: Result := FloatPropertyToStr(AObject, PropInfo);
    tkInt64: Result := IntToStr(GetInt64Prop(AObject, PropInfo));
    tkWString,
    tkLString,
{$IFDEF FPC}
    tkAString,
{$ENDIF}
    tkString: Result := GetStrProp(AObject, PropInfo);
    tkVariant: Result := GetVariantProp(AObject, PropInfo);
    tkSet: Result := SetToString(PropInfo, GetOrdProp(AObject, PropInfo), True);
  else
    raise Exception.CreateFmt('Unknown property type for %s', [PropInfo^.Name]);
{   tkClass: Not implemented }
  end;
end;

function TryStrToBoolean(s: String; out Value: Boolean): Boolean;
begin
  Result := True;
  if (s = '') or (s[1] in ['n', 'N', 'f', 'F', '0']) then
    Value := False
  else if (s[1] in ['y', 'Y', 't', 'T', '1']) or (s = '-1') then
    Value := True
  else
    Result := False;
end;

function StrToBoolean(s: String): Boolean;
begin
  if not TryStrToBoolean(s,Result) then
    raise Exception.Create('Bad true/false expression: ' + s);
end;

procedure SetPropertyFromStr(AObject: TObject; const PropName: string; const s: string);
var
  PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(AObject, PropName);
  if not Assigned(PropInfo) then
    raise Exception.Create(PropName + ' not found in ' + AObject.ClassName);
  SetPropertyFromStr(AObject, PropInfo, s);
end;

procedure SetPropertyFromStr(AObject: TObject; PropInfo: PPropInfo; const s: string);
begin
  {$IFDEF DELPHI}
  if PropInfo^.PropType^.Name = 'Boolean' then
    { Be more flexible in accepting boolean values }
    SetOrdProp(AObject, PropInfo, Integer(StrToBoolean(s)))
  else
  {$ENDIF}
  case PropInfo^.PropType^.Kind of
    tkInteger:  SetOrdProp(AObject, PropInfo, StrToInt(s));
    {$IFDEF FPC}
    tkBool:     SetOrdProp(AObject, PropInfo, Integer(StrToBoolean(s)));
    {$ENDIF}
    tkWChar:    SetOrdProp(AObject, PropInfo, Longint(s[1]));
    tkChar:     SetOrdProp(AObject, PropInfo, Integer(s[1]));
    tkEnumeration: SetEnumProp(AObject, PropInfo, s);
    tkFloat:    SetFloatProp(AObject, PropInfo, StrToFloat(s));
    tkInt64:    SetInt64Prop(AObject, PropInfo, StrToInt64(s));
    tkLString,
    tkWString,
    {$IFDEF FPC}
    tkAString,
    {$ENDIF}
    tkString:   SetStrProp(AObject, PropInfo, s);
    tkVariant:  SetVariantProp(AObject, PropInfo, s);
    tkSet:      SetSetProp(AObject, PropInfo, s);
  else
    raise Exception.CreateFmt('Unknown property type for %s', [PropInfo^.Name]);
{   tkClass: Not implemented }
  end;
end;

end.
