unit SavitzkyGolay;

interface

uses MathSup;

function SmoothingCoefficients(
  Leading, Trailing: Integer; { Gives set of weights to compute Data[0] given Data[-Leading..Trailing] }
  Derivative: Integer = 0;    { Order of derivative; 0 for smoothing only }
  PolyOrder:Integer = 2       { Order of polynomial to fit to }
): TDoubleArray;

implementation

uses LUDecomp, Math;

function SmoothingCoefficients(
  Leading, Trailing: Integer; { Gives set of weights to compute Data[0] given Data[-Leading..Trailing] }
  Derivative: Integer = 0;    { Order of derivative; 0 for smoothing only }
  PolyOrder:Integer = 2       { Order of polynomial to fit to }
): TDoubleArray;
{ Return a set of coefficients for smoothing (or producing a derivative), based
  on the method described in A Savitzky and MJE Golay, Anal Chem, 36, 1627 (1964).
  Algorithm is based on that given in Numerical Recipes }
var
  a,b: array of Double;
  ipiv: array of Integer;
  i, j, m, Info: Integer;
  Sum, Fac: Double;
  function Index(i,j:Integer):Integer;
  begin
    Result := j*(PolyOrder+1)+i;
  end;
begin
  PolyOrder := Max(PolyOrder,Derivative);
  SetLength(a,Sqr(PolyOrder+1));
  for i := 0 to 2*PolyOrder do begin
    if i > 0 then
      Sum := 0
    else
      Sum := 1;
    for j := 1 to Leading do
      Sum := Sum + IntPower(-j,i);
    for j := 1 to Trailing do
      Sum := Sum + IntPower(j,i);
    m := Min(i,2*PolyOrder-i);
    j := -m;
    while j <= m do begin
      a[Index((i+j) div 2,(i-j) div 2)] := sum;
      Inc(j,2);
    end;
  end;
  SetLength(ipiv,PolyOrder+1);
  Info := dgetrf(PolyOrder+1,PolyOrder+1,a[0],PolyOrder+1,ipiv[0]);
  if Info = 0 then begin
    SetLength(b,PolyOrder+1);
    FillChar(b[0],(PolyOrder+1)*Sizeof(Double),0);
    b[Derivative] := 1;
    for i := 2 to Derivative do
      b[Derivative] := b[Derivative]*i;
    Info := dgetrs(eqNormal, PolyOrder+1, 1, a[0], PolyOrder+1, ipiv[0], b[0], Length(b));
    if Info = 0 then begin
      SetLength(Result,Leading+Trailing+1);
      for i := -Leading to Trailing do begin
        Sum := b[0];
        Fac := 1;
        for j := 1 to PolyOrder do begin
          Fac := Fac * i;
          Sum := Sum + b[j]*Fac;
        end;
        Result[i+Leading] := Sum;
      end;
    end;
  end;
end;

end.
