unit AlEdit;
{ Add Alignment to a TEdit Control }
{.$DEFINE SETEDITRECT} { Required so that multiline behaviour never shows }
{ For CLX_, Alignment is available anyway }
{ TODO: Implement alignment for LCL_ (may not be possible with GTK) }

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses {$IFDEF MSWINDOWS} Windows, Messages, {$ENDIF}
     Graphics, Controls, stdctrls,
     SysUtils, Classes;

type

  { TAlignedEdit }

  TAlignedEdit = class(TEdit)
  private
    FnVisibleDigits: Integer;
    FAlignment: TAlignment;
    procedure SetAlignment(Value: TAlignment);
    {$IFDEF VCL_}
    procedure WMGetDlgCode(var Message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    {$ENDIF}
    function GetnVisibleDigits: Integer;
  protected
    {$IFDEF VCL_}
    procedure CreateParams(var Params: TCreateParams); override;
    {$ENDIF}
    {TODO : FontChanged in LCL}
    procedure KeyPress(var Key: Char); override;
    procedure Resize; override;
  public
    constructor Create(AOwner: TComponent); override;
    property nVisibleDigits: Integer read GetnVisibleDigits;
  published
    property Alignment: TAlignment read FAlignment write SetAlignment default taLeftJustify;
  end;
  
implementation

uses {$IFDEF LCL_} LCLType, Types, LCLIntf, {$ENDIF} UCompat;

{ TAlignedEdit }

procedure TAlignedEdit.SetAlignment(Value: TAlignment);
begin
  if Value <> FAlignment then begin
    FAlignment := Value;
    {$IFDEF MSWINDOWS}
    if HandleAllocated then
      RecreateWnd{$IFDEF FPC}(Self){$ENDIF};
    { Note that under GTK/LCL_ this looses any text present }
    {$ENDIF}
  end;
end;


function TAlignedEdit.GetnVisibleDigits: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  ZeroSize: TSize;
begin
  if FnVisibleDigits < 0 then begin
    { Get Text Metrics of current font }
    DC := GetDC(0);
    SaveFont := SelectObject(DC, Font.Handle);
    GetTextExtentPoint(DC, AddressOfFirstChar('0123456789'), 10, ZeroSize);
    FnVisibleDigits := MulDiv(ClientWidth, 10, ZeroSize.cx);
    SelectObject(DC, SaveFont);
    ReleaseDC(0, DC);
  end;
  Result := FnVisibleDigits;
end;

procedure TAlignedEdit.KeyPress(var Key: Char);
begin
  if Ord(Key) = VK_RETURN then
    Key := Chr(0) { Always discard return, as it is not meant to be a multiline control }
  else if Key <> Chr(0) then
    inherited KeyPress(Key);
end;

procedure TAlignedEdit.Resize;
begin
  inherited Resize;
  FnVisibleDigits := -1;
end;

constructor TAlignedEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  AutoSize := False;
  FnVisibleDigits := -1;
  FAlignment := taLeftJustify;
end;

{$IFDEF VCL_}
procedure TAlignedEdit.CreateParams(var Params: TCreateParams);
const
  Alignments:array[TAlignment] of DWORD = (ES_LEFT,ES_RIGHT,ES_CENTER);
begin
  inherited CreateParams(Params);
  { Windows 2k (and possibly others) shows a strange interaction here }
  { If ES_RIGHT set, ES_AUTOVSCROLL must also be set }
  Params.Style := (Params.Style and not ES_AUTOVSCROLL) or ES_MULTILINE or ES_AUTOHSCROLL or Alignments[FAlignment];
  if FAlignment = taRightJustify then
    Params.Style := Params.Style or ES_AUTOVSCROLL;
  // Params.Style := Params.Style and not ES_WANTRETURN; { But this is ineffective }
end;
{$ENDIF}

{$IFDEF VCL_}
procedure TAlignedEdit.CMFontChanged(var Message: TMessage);
begin
  inherited;
  FnVisibleDigits := -1;
end;

procedure TAlignedEdit.WMGetDlgCode(var Message: TWMGetDlgCode);
begin
  inherited;
  Message.Result := Message.Result and not DLGC_WANTALLKEYS;
end;
{$ENDIF}

end.

