unit FileNameEdit;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses Controls, Dialogs, Buttons, ExtDlgs,
     SysUtils, Classes, EditExtr;

type

  TFileDialogKind = (dkOpen, dkSave, dkOpenPicture, dkSavePicture);

  TFileNameEdit = class(TCustomEditButton)
  private
    FDialog: TOpenDialog;
    FDialogKind: TFileDialogKind;
    procedure CreateDialog;
    procedure SpeedButtonClick(Sender: TObject);
    procedure SetDialogKind(const Value: TFileDialogKind);
    function GetOptions: TOpenOptions;
    procedure SetOptions(const Value: TOpenOptions);
    {$IFDEF VCL_}
    function GetOptionsEx: TOpenOptionsEx;
    procedure SetOptionsEx(const Value: TOpenOptionsEx);
    {$ENDIF}
    function GetDefaultExt: string;
    procedure SetDefaultExt(const Value: string);
    function GetFileName: TFileName;
    procedure SetFileName(const Value: TFileName);
    function GetFilter: string;
    function GetFilterIndex: Integer;
    function GetInitialDir: string;
    procedure SetFilter(const Value: string);
    procedure SetFilterIndex(const Value: Integer);
    procedure SetInitialDir(const Value: string);
    function GetTitle: string;
    procedure SetTitle(const Value: string);
  protected
    function CreateExtraControl:TControl; override;
  public
    constructor Create(AOwner: TComponent); override;
    function Execute: Boolean;
    property Dialog: TOpenDialog read FDialog;
  published
    property DialogKind: TFileDialogKind read FDialogKind write SetDialogKind default dkOpen;

    property DefaultExt: string read GetDefaultExt write SetDefaultExt;
    property FileName: TFileName read GetFileName write SetFileName;
    property Filter: string read GetFilter write SetFilter;
    property FilterIndex: Integer read GetFilterIndex write SetFilterIndex default 1;
    property InitialDir: string read GetInitialDir write SetInitialDir;
    property Options: TOpenOptions read GetOptions write SetOptions
             default [{$IFDEF VCL_}ofHideReadOnly,{$ENDIF}ofEnableSizing];
    {$IFDEF VCL_}
    property OptionsEx: TOpenOptionsEx read GetOptionsEx write SetOptionsEx default [];
    {$ENDIF}
    property DialogTitle: string read GetTitle write SetTitle;
  end;

implementation

uses {$IFDEF VCL_}  CommDlg, {$ENDIF}
     {$IFDEF LCL_}  LResources, {$ENDIF}
     UCompat;

{ Note that these contain the glyph used at runtime, *not* the
  icon for the palette }
{$IFNDEF FPC}
{$R FilEditR.dcr}
{$ENDIF}

{ TFileNameEdit }

constructor TFileNameEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  CreateDialog;
end;

function TFileNameEdit.CreateExtraControl: TControl;
begin
  Result := inherited CreateExtraControl;
  with SpeedButton do begin
    OnClick := {$IFDEF FPC}@{$ENDIF}SpeedButtonClick;
    {$IFDEF FPC}
      Glyph.LoadFromLazarusResource('FILEEDIT');
    {$ELSE}
      Glyph.LoadFromResourceName(HInstance, 'FILEEDIT');
    {$ENDIF}
  end;
end;

procedure TFileNameEdit.CreateDialog;
var
  NewDialog: TOpenDialog;
begin
  case FDialogKind of
    dkOpen: NewDialog := TOpenDialog.Create(Self);
{$IFDEF VCL_}
    dkOpenPicture: NewDialog := TOpenPictureDialog.Create(Self);
    dkSavePicture: NewDialog := TSavePictureDialog.Create(Self);
{$ENDIF}
    else {dkSave} NewDialog := TSaveDialog.Create(Self);
  end;
  try
    if FDialog <> nil then begin
      with NewDialog do begin
        DefaultExt := FDialog.DefaultExt;
        FileName := FDialog.FileName;
        Filter := FDialog.Filter;
        FilterIndex := FDialog.FilterIndex;
        InitialDir := FDialog.InitialDir;
        Options := FDialog.Options;
        {$IFDEF VCL_}
        OptionsEx := FDialog.OptionsEx;
        {$ENDIF}
        Title := FDialog.Title;
      end;
      FDialog.Free;
    end else begin
      NewDialog.Filter := 'All Files (*.*)|'+AllFIlesMask;
    end;
  finally
    FDialog := NewDialog;
  end;
end;

function TFileNameEdit.GetFileName: TFileName;
begin
  Result := Text;
end;

procedure TFileNameEdit.SetFileName(const Value: TFileName);
begin
  Text := Value;
  Dialog.FileName := Value;
end;

procedure TFileNameEdit.SpeedButtonClick(Sender: TObject);
begin
  Execute;
end;

function TFileNameEdit.Execute: Boolean;
var
  s: string;
begin
  s := Text;
  DoDirSeparators(s);
  Dialog.FileName := s;
  Result := Dialog.Execute;
  {$IFDEF VCL_}
  if CommDlgExtendedError = FNERR_INVALIDFILENAME then begin
    { If file does not exist, dialog box can fail }
    Dialog.FileName := ExtractFileName(s);
    Result := Dialog.Execute;
  end;
  {$ENDIF}
  if Result then
    Text := Dialog.FileName;
end;

procedure TFileNameEdit.SetDialogKind(const Value: TFileDialogKind);
begin
  if FDialogKind <> Value then begin
    FDialogKind := Value;
    CreateDialog;
  end;
end;

function TFileNameEdit.GetOptions: TOpenOptions;
begin
  Result := Dialog.Options;
end;

procedure TFileNameEdit.SetOptions(const Value: TOpenOptions);
begin
  Dialog.Options := Value;
end;

{$IFDEF VCL_}
function TFileNameEdit.GetOptionsEx: TOpenOptionsEx;
begin
  Result := Dialog.OptionsEx;
end;

procedure TFileNameEdit.SetOptionsEx(const Value: TOpenOptionsEx);
begin
  Dialog.OptionsEx := Value;
end;
{$ENDIF}

function TFileNameEdit.GetDefaultExt: string;
begin
  Result := Dialog.DefaultExt;
end;

procedure TFileNameEdit.SetDefaultExt(const Value: string);
begin
  Dialog.DefaultExt := Value;
end;

function TFileNameEdit.GetFilter: string;
begin
  Result := Dialog.Filter;
end;

function TFileNameEdit.GetFilterIndex: Integer;
begin
  Result := Dialog.FilterIndex;
end;

function TFileNameEdit.GetInitialDir: string;
begin
  Result := Dialog.InitialDir;
end;

procedure TFileNameEdit.SetFilter(const Value: string);
begin
  Dialog.Filter := Value;
end;

procedure TFileNameEdit.SetFilterIndex(const Value: Integer);
begin
  Dialog.FilterIndex := Value;
end;

procedure TFileNameEdit.SetInitialDir(const Value: string);
begin
  Dialog.InitialDir := Value;
end;

function TFileNameEdit.GetTitle: string;
begin
  Result := Dialog.Title;
end;

procedure TFileNameEdit.SetTitle(const Value: string);
begin
  Dialog.Title := Value;
end;

{$IFDEF LCL_}
initialization
  {$I FilEditR.lrs}
{$ENDIF}

end.
