unit FixDlg;

interface
{$I utils.inc}

uses
  Graphics, Controls, Forms, Dialogs, StdCtrls, Buttons,
  SysUtils, Classes, FixFnt;

type
  TFixedFontForm = class(TForm)
    Faces: TListBox;
    Label1: TLabel;
    Sizes: TComboBox;
    Label2: TLabel;
    Bold: TCheckBox;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    Sample: TMemo;
    Label3: TLabel;
    procedure FormShow(Sender: TObject);
    procedure FacesKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FacesMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SizesChange(Sender: TObject);
    procedure BoldMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BoldKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure BoldClick(Sender: TObject);
  private
    CurrentIndex: Integer;
    CurrentFace: Integer;
    BestIndex, BestHeight, BestWidth: Integer;
    FFixedFont: TFixedFont;
    procedure SetFixedFont(const Value: TFixedFont); { For Matching previous font }
    procedure DoFaceChange;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure CheckSizeChange;
    procedure DoSizeChange;
    procedure CheckFaceChange;
    property FixedFont:TFixedFont read FFixedFont write SetFixedFont;
  end;

var
  FixedFontForm: TFixedFontForm;

implementation

uses {$IFDEF MSWINDOWS} Windows {$ENDIF}
     {$IFDEF LCLgtk2} gtk2, Pango, Gtk2Proc, Gtk2Def, glib2 {$ENDIF}
     {$IFDEF LCLgtk3} LazGtk3, LazPango1, Gtk3procs, LazGLib2 {$DEFINE LCLgtk2} {$ENDIF}
     {$IFDEF LCLcarbon} LCLIntf, LCLType {$ENDIF}
     {$IFDEF LCLcocoa} LCLIntf {$ENDIF}
     ;
{$IFDEF VCL_}
{$R *.dfm}
{$ENDIF}
{$IFDEF LCL_}
{$R *.lfm}
{$ENDIF}

{$IFDEF MSWINDOWS}
function EnumFacesProc (var lf:TLOGFONT; var tm:TTextMetric;
                        FontType: Cardinal; Form:TFixedFontForm):Integer; {$IFDEF WIN32} stdcall {$ELSE} far {$ENDIF};
var
  Name: string;
begin
  if (tm.tmPitchAndFamily and TMPF_FIXED_PITCH) = 0 then begin
    Name := lf.lfFaceName;
    SetLength(Name, StrLen(lf.lfFaceName));
    if Form.Faces.Items.IndexOf(Name) < 0 then
      Form.Faces.Items.Add(Name);
  end;
  Result := 1;
end;
{$ENDIF}

{$IFDEF LCLcarbon}
function EnumFontProc(var lf: TEnumLogFontEx; var Metric: TNewTextMetricEx;
    FontType: Longint; Data: LParam):Longint; stdcall;
var
  Name: string;
  Form: TFixedFontForm;
begin
  Form := TFixedFontForm(Data);
  Name := lf.elfLogFont.lfFaceName;
  SetLength(Name, StrLen(PChar(Name)));
  if Form.Faces.Items.IndexOf(Name) < 0 then
    Form.Faces.Items.Add(Name);
  Result := 1;
end;

{$ENDIF}

procedure TFixedFontForm.FormShow(Sender: TObject);
  {$IFDEF MSWINDOWS}
  var
    proc: TFarProc;
  {$ENDIF}
  {$IFDEF LCLgtk2}
  var
    Widget : PGTKWidget;
    Context : PPangoContext;
    families : PPPangoFontFamily;
    Tmp: AnsiString;
    I, N: Integer;
  {$ENDIF}
  {$IFDEF LCLcarbon}
  var
    LogFont: TLogFont;
  {$ENDIF}
begin
  CurrentFace := -1;
  Faces.Clear;
  {$IFDEF MSWINDOWS}
    {$IFDEF FPC}
      EnumFonts(Canvas.Handle, nil, ENUMFONTSPROC(@EnumFacesProc), Pointer(Self));
    {$ELSE}
      proc := MakeProcInstance(@EnumFacesProc, HInstance);
      EnumFonts(Canvas.Handle, nil, proc, Pointer(Self));
      FreeProcInstance(proc);
    {$ENDIF}
  {$ENDIF}
  {$IFDEF LCLgtk2}
  Widget := GetStyleWidget(lgsDefault);
  if Widget = nil then
    exit;
  Context := gtk_widget_get_pango_context(Widget);
  if Context = nil then
    exit;
  families := nil;
  pango_context_list_families(Context, @families, @n);
  for I := 0 to N - 1 do
    if (families[I] <> nil) and pango_font_family_is_monospace(families[I]) then begin
      Tmp := StrPas(pango_font_family_get_name(families[I]));
      if Tmp <> '' then
        if Faces.Items.IndexOf(Tmp) < 0 then
          Faces.Items.Append(Tmp);
   end;
  if (families <> nil) then
    g_free(families);
  {$ENDIF}
  {$IFDEF LCLcarbon}
  FillChar(LogFont, SizeOf(LogFont), 0);
  LogFont.lfCharSet := DEFAULT_CHARSET;
  EnumFontFamiliesEx(Canvas.Handle, @LogFont, @EnumFontProc, PtrUInt(Self), 0);
  {$ENDIF}
  CurrentFace := Faces.Items.IndexOf(FixedFont.Name);
  if CurrentFace < 0 then begin
    { Font Not Found }
    CurrentFace := 0;
    FFixedFont.Name := Faces.Items[0];
  end;
  Faces.ItemIndex := CurrentFace;
  DoFaceChange;
end;

procedure AddSize(h, w: Integer; Form:TFixedFontForm);
var
  Name: string;
begin
  if w = 0 then
    Name := IntToStr(h)
  else
    Name := Format('%2d x %2d', [w, h]);
  with Form.Sizes.Items do
    if IndexOf(Name) < 0 then
      Add(Name);
  { Select closest to current font size; look for something smaller if no exact match }
  if h <= Form.FixedFont.Height then
    if (h > Form.BestHeight) or (Form.BestIndex = 0)
        or ((h = Form.BestHeight) and (w <= Form.FixedFont.Width) and (w > Form.BestWidth)) then begin
      Form.BestHeight := h;
      Form.BestWidth := w;
      Form.BestIndex := Form.Sizes.Items.Count - 1;
    end;
end;

procedure AddStandardSizes(Form:TFixedFontForm);
var
  i: Integer;
begin
  { make a size list up }
  for i := 8 to 12 do AddSize(i, 0, Form);
  for i := 7 to 14 do AddSize(2*i, 0, Form);
  AddSize(36, 0, Form);
  AddSize(48, 0, Form);
  AddSize(72, 0, Form);
end;

{$IFDEF MSWINDOWS}
function EnumSizesProc (var lf:TLogFont; var tm:TTextMetric;
                        FontType: Cardinal; Form:TFixedFontForm):Integer; {$IFDEF WIN32} stdcall {$ELSE} far {$ENDIF};
begin
  Result := 1;
  if (lf.lfItalic or lf.lfUnderline or lf.lfStrikeOut) = 0 then
    if (FontType and TRUETYPE_FONTTYPE) <> 0 then begin
      AddStandardSizes(Form);
      Result := 0; { no need to list this one further }
    end else { non-scalable font }
      AddSize (lf.lfHeight, lf.lfWidth, Form);
end;
{$ENDIF}

procedure TFixedFontForm.DoSizeChange;
begin
  CurrentIndex := Sizes.ItemIndex;
  FFixedFont.Bold := Bold.Checked;
  FFixedFont.Size := Sizes.Items[CurrentIndex];
  FFixedFont.CopyTo(Sample.Font);
  Sample.Clear;
  Sample.Lines.Add(FFixedFont.Name
             + ' Sample Aa Bb Cc Dd Ee Ff Gg 0123456789 The Small Brown Fox');
end;

procedure TFixedFontForm.CheckSizeChange;
begin
  if (Sizes.ItemIndex >= 0) and ((Sizes.ItemIndex <> CurrentIndex) or (FFixedFont.Bold <> Bold.Checked)) then
    DoSizeChange;
end;

procedure TFixedFontForm.CheckFaceChange;
begin
  if CurrentFace <> Faces.ItemIndex then begin
    CurrentFace := Faces.ItemIndex;
    FFixedFont.Name := Faces.Items[CurrentFace];
    DoFaceChange;
  end;
end;

procedure TFixedFontForm.FacesKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  CheckFaceChange;
end;

procedure TFixedFontForm.FacesMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  CheckFaceChange;
end;

procedure TFixedFontForm.SizesChange(Sender: TObject);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.SetFixedFont(const Value: TFixedFont);
begin
  FFixedFont.Assign(Value);
end;

constructor TFixedFontForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFixedFont := TFixedFont.Create;
end;

destructor TFixedFontForm.Destroy;
begin
  FFixedFont.Free;
  inherited Destroy;
end;

procedure TFixedFontForm.Loaded;
begin
  inherited Loaded;
  {$IFNDEF LCL_}
  // Moved from resource file as not present on all systems
  BorderIcons := [biSystemMenu];
  Faces.IntegralHeight := True;
  {$ENDIF}
end;

procedure TFixedFontForm.BoldMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.BoldKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.DoFaceChange;
{$IFDEF MSWINDOWS}
  var
    proc: TFarProc;
    AName: string;
{$ENDIF}
begin
  Sizes.Clear;
  CurrentIndex := -1;
  BestIndex := 0;
  BestHeight := 0;
  BestWidth := 0;
  {$IFDEF MSWINDOWS}
    {$IFDEF FPC}
      EnumFonts(Canvas.Handle, PChar(AName), ENUMFONTSPROC(@EnumSizesProc), Pointer(Self));
    {$ELSE}
      proc := MakeProcInstance(@EnumSizesProc, HInstance);
      EnumFonts(Canvas.Handle, PChar(AName), proc, Pointer(Self));
      FreeProcInstance(proc);
    {$ENDIF}
  {$ENDIF}
  with Sizes.Items do
    if Count = 0 then { No sizes - make some up }
      AddStandardSizes(Self);
  Sizes.ItemIndex := BestIndex;
  DoSizeChange;
end;

procedure TFixedFontForm.BoldClick(Sender: TObject);
begin
  CheckSizeChange;
end;

end.
