unit GStyle;

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses Graphics, Gmarks, Classes;

const
  MAXSTYLE = 9;
  MAXPATLENGTH = 6;

type

  TPat = record
    length: Byte;
    scale:  Byte;
    Segments: array [0..MAXPATLENGTH-1] of ShortInt;
  end;

  TGStyle = class(TPersistent)
  private
    FColour: TColor;
    FOnChange: TNotifyEvent;
    procedure SetColour(AColour: TColor);
  protected
    procedure Changed;
  public
    constructor Create;
  published
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Colour: TColor read FColour write SetColour nodefault;
  end;

  TLinePattern = (lpSolid, lpDash, lpDot, lpDashDot, lpDashDotDot, lpLongDash, lpShortDash,
                  lpLongShort, lpLongShortShort, lpEndPoints);

  TLineStyle = class(TGStyle)
  private
    FWidth: Single;
    FPattern: TLinePattern;
    procedure SetWidth(AWidth: Single);
    procedure SetPattern(APattern: TLinePattern);
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
  published
    property Width: Single read FWidth write SetWidth;
    property Pattern: TLinePattern read FPattern write SetPattern nodefault;
  end;

  TMarkStyle = class(TGStyle)
  private
    FSize: Single;
    FMarkType: TMarkType;
    FDotAtCentre: Boolean;
    procedure SetSize(ASize: Single);
    procedure SetMarkType(AType: TMarkType);
    procedure SetDotAtCentre(Dot: Boolean);
  public
    constructor Create;
    procedure Assign(Source: TPersistent); override;
  published
    property Size: Single read FSize write SetSize;
    property MarkType: TMarkType read FMarkType write SetMarkType default mtDiamond;
    property DotAtCentre: Boolean read FDotAtCentre write SetDotAtCentre default True;
  end;

const
  pats: array [0..MAXSTYLE] of TPat =
( (length:1; scale: 1; Segments:(  1, 0, 0, 0, 0, 0)), (* 0 - solid line *)
  (length:2; scale:12; Segments:(  6, 6, 0, 0, 0, 0)), (* 1 - - - - *)
  (length:2; scale:15; Segments:( -1, 3, 0, 0, 0, 0)), (* 2 . . . . *)
  (length:4; scale:12; Segments:(  6, 3,-1, 3, 0, 0)), (* 3 -.-.-. *)
  (length:6; scale:12; Segments:(  6, 2,-1, 2,-1, 2)), (* 4 -..-..-.. *)
  (length:2; scale:12; Segments:( 10, 2, 0, 0, 0, 0)), (* 5 -- -- -- -- *)
  (length:2; scale:12; Segments:(  2,10, 0, 0, 0, 0)), (* 6 -  -  -  - *)
  (length:4; scale:24; Segments:( 10, 5, 4, 5, 0, 0)), (* 7 --  -  --  - *)
  (length:6; scale:24; Segments:( 10, 5, 4, 5, 4, 5)), (* 8 --  -  - --  - - *)
  (length:0; scale: 1; Segments:(  0, 0, 0, 0, 0, 0))  (* MAXSTYLE Endpoints only *)
);

implementation

{ TGStyle }

procedure TGStyle.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

constructor TGStyle.Create;
begin
  inherited Create;
  FColour := clBlack;
end;

procedure TGStyle.SetColour(AColour: TColor);
begin
  if Colour <> AColour then begin
    FColour := AColour;
    Changed;
  end;
end;

{ TLineStyle }

procedure TLineStyle.Assign(Source: TPersistent);
begin
  if Source is TLineStyle then begin
    Width := TLineStyle(Source).Width;
    Colour := TLineStyle(Source).Colour;
    Pattern := TLineStyle(Source).Pattern;
  end else
    inherited Assign(Source);
end;

constructor TLineStyle.Create;
begin
  inherited Create;
  FWidth := 0;
  FPattern := lpSolid;
end;

procedure TLineStyle.SetWidth(AWidth: Single);
begin
  if Width <> AWidth then begin
    FWidth := AWidth;
    Changed;
  end;
end;

procedure TLineStyle.SetPattern(APattern: TLinePattern);
begin
  if Pattern <> APattern then begin
    FPattern := APattern;
    Changed;
  end;
end;

{ TMarkStyle }

constructor TMarkStyle.Create;
begin
  inherited Create;
  FSize := 2;
  FMarkType := mtDiamond;
  FDotAtCentre := True;
end;

procedure TMarkStyle.Assign(Source: TPersistent);
begin
  if Source is TMarkStyle then begin
    Colour := TMarkStyle(Source).Colour;
    Size := TMarkStyle(Source).Size;
    MarkType := TMarkStyle(Source).MarkType;
    DotAtCentre := TMarkStyle(Source).DotAtCentre;
  end else
    inherited Assign(Source);
end;

procedure TMarkStyle.SetSize(ASize: Single);
begin
  if Size <> ASize then begin
    FSize := ASize;
    Changed;
  end;
end;

procedure TMarkStyle.SetDotAtCentre(Dot: Boolean);
begin
  if DotAtCentre <> Dot then begin
    FDotAtCentre := Dot;
    Changed;
  end;
end;

procedure TMarkStyle.SetMarkType(AType: TMarkType);
begin
  if Ord(AType) > Ord(High(TMarkType)) then
    AType:= High(TMarkType);
  if MarkType <> AType then begin
    FMarkType := AType;
    Changed;
  end;
end;

end.

