unit LogActn;

interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
uses
  {$IFNDEF CLX_} ActnList, {$ELSE} QActnList, {$ENDIF}
  Classes, ExActn, EditLog, LogControl;

type

  TLogAction = class(TActionBase)
  public
    function HandlesTarget(Target: TObject): Boolean; override;
    procedure UpdateTarget(Target: TObject); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure ExecuteTextTarget(Target: TLog); virtual; abstract;
    procedure UpdateTextTarget(Target: TLog); virtual;
  end;

  TTextAction = class(TActionBase)
  public
    function HandlesTarget(Target: TObject): Boolean; override;
    procedure UpdateTarget(Target: TObject); override;
    procedure ExecuteTarget(Target: TObject); override;
    procedure ExecuteTextTarget(Target: TEditLog); virtual; abstract;
    procedure UpdateTextTarget(Target: TEditLog); virtual;
  end;

  TLogFind = class(TLogAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
  end;

  TLogFindNext = class(TLogFind)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
  end;

  TLogFindThisWord = class(TLogFindNext)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
  end;

  TTextReplace = class(TLogFindNext)
  public
    function HandlesTarget(Target: TObject): Boolean; override;
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
  end;

  TTextSortUp = class(TLogAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
    procedure UpdateTextTarget(Target: TLog); override;
  end;

  TTextSortDown = class(TLogAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
    procedure UpdateTextTarget(Target: TLog); override;
  end;

  TTextSetBookMark = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextGoToBookMark = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextExchangeWithBookMark = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextGoToLine = class(TLogAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TLog); override;
  end;

  TTextIncrement = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextDecrement = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextRecordMacro = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextPlayBackMacro = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextRemoveTabs = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

  TTextSetTabSpacing = class(TTextAction)
  public
    constructor Create(AOwner: TComponent); override;
    procedure ExecuteTextTarget(Target: TEditLog); override;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses
  {$IFNDEF CLX_} { Windows, } Forms, Menus, Dialogs,
  {$ELSE} QForms, QMenus, QDialogs, Types, {$ENDIF}
  {$IFDEF LCL_} LCLType, {$ENDIF}
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  CharType, LogFind, SysUtils, ResActn;

{$IFDEF SINGLE}
procedure Register;
begin
    RegisterActions('Text', [TLogFind, TLogFindNext, TLogFindThisWord, TTextReplace,
                             TTextIncrement, TTextDecrement,
                             TTextRecordMacro, TTextPlayBackMacro, TTextSortUp, TTextSortDown,
                             TTextSetBookMark, TTextGoToBookMark, TTextExchangeWithBookMark, TTextGoToLine,
                             TTextRemoveTabs], nil);
end;
{$ENDIF}

{ TLogFind }

constructor TLogFind.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('F'), [ssCtrl]);
  Hint := 'Find|Search for text';
  ImageIndex := 12;
  Caption := '&Find...';
end;

procedure TLogFind.ExecuteTextTarget(Target: TLog);
begin
  Target.ShowFindDialog;
end;

{ TLogFindNext }

constructor TLogFindNext.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(VK_F3, []);
  Hint := 'Find Next|Find next occurence of text';
  ImageIndex := 13;
  Caption := 'Find Ne&xt';
end;

procedure TLogFindNext.ExecuteTextTarget(Target: TLog);
begin
  Target.FindNext;
end;

{ TLogFindThisWord }

constructor TLogFindThisWord.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(VK_F2, []);
  Hint := 'Find Word|Find next selected or current word';
  ImageIndex := -1;
  Caption := 'Find &Word';
end;

procedure TLogFindThisWord.ExecuteTextTarget(Target: TLog);
begin
  Target.FindThisWord;
end;

{ TTextReplace }

constructor TTextReplace.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetShortCut(Word('R'), [ssCtrl]);
  Hint := 'Replace|Show replace dialog box';
  ImageIndex := -1;
  Caption := '&Replace...';
end;

procedure TTextReplace.ExecuteTextTarget(Target: TLog);
begin
  Target.ShowReplaceDialog;
end;

function TTextReplace.HandlesTarget(Target: TObject): Boolean;
begin
  if Target is TEditLog then
    Result := inherited HandlesTarget(Target)
  else
    Result := False;
end;

{ TTextSortUp }

constructor TTextSortUp.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Sort Up|Sort in ascending order of selected columns';
  ImageIndex := 19;
  Caption := 'Sort &Up';
end;

procedure TTextSortUp.ExecuteTextTarget(Target: TLog);
begin
  Target.SortOnSelectedColumns(True);
end;

procedure TTextSortUp.UpdateTextTarget(Target: TLog);
begin
  Enabled := Target.CanSortOnSelectedColumns;
end;

{ TTextSortDown }

constructor TTextSortDown.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Sort Down|Sort in descending order of selected columns';
  ImageIndex := 20;
  Caption := 'Sort Do&wn';
end;

procedure TTextSortDown.ExecuteTextTarget(Target: TLog);
begin
  Target.SortOnSelectedColumns(False);
end;

procedure TTextSortDown.UpdateTextTarget(Target: TLog);
begin
  Enabled := Target.CanSortOnSelectedColumns;
end;

{ TLogAction }

procedure TLogAction.ExecuteTarget(Target: TObject);
begin
  SetResultForControl(Target as TLog, '');
  ExecuteTextTarget(Target as TLog);
end;

function TLogAction.HandlesTarget(Target: TObject): Boolean;
begin
  Result := ((Control <> nil) and (Target = Control) or
    (Control = nil) and (Target is TLog)) and TLog(Target).Focused;
end;

procedure TLogAction.UpdateTarget(Target: TObject);
begin
  UpdateTextTarget(Target as TLog);
end;

procedure TLogAction.UpdateTextTarget(Target: TLog);
begin
  Enabled := True;
end;

{ TTextAction }

procedure TTextAction.ExecuteTarget(Target: TObject);
begin
  SetResultForControl(Target as TEditLog, '');
  ExecuteTextTarget(Target as TEditLog);
end;

function TTextAction.HandlesTarget(Target: TObject): Boolean;
begin
  Result := ((Control <> nil) and (Target = Control) or
    (Control = nil) and (Target is TEditLog)) and TEditLog(Target).Focused;
end;

procedure TTextAction.UpdateTarget(Target: TObject);
begin
  UpdateTextTarget(Target as TEditLog);
end;

procedure TTextAction.UpdateTextTarget(Target: TEditLog);
begin
  Enabled := True;
end;

{ TTextSetBookMark }

constructor TTextSetBookMark.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Set Bookmark|Set Bookmark at current location';
  Caption := 'Set &Bookmark';
  SetShortCut(Word('B'), [ssCtrl]);
end;

procedure TTextSetBookMark.ExecuteTextTarget(Target: TEditLog);
begin
  Target.SetBookMarkHere;
end;

{ TTextGoToBookMark }

constructor TTextGoToBookMark.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Goto Bookmark';
  Caption := 'Goto B&ookmark';
  SetShortCut(Word('Q'), [ssCtrl]);
end;

procedure TTextGoToBookMark.ExecuteTextTarget(Target: TEditLog);
begin
  Target.GotoBookMark;
end;

{ TTextExchangeWithBookMark }

constructor TTextExchangeWithBookMark.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'E&xchange with Bookmark|Exchange cursor and bookmark positions';
  Caption := 'E&xchange with Bookmark';
  SetShortCut(Word('B'), [ssShift, ssCtrl]);
end;

procedure TTextExchangeWithBookMark.ExecuteTextTarget(Target: TEditLog);
begin
  Target.ExchangeCursorAndBookMark;
end;

{ TTextGoToLine }

constructor TTextGoToLine.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Go To Line|Ask for line number to move to';
  Caption := '&Go To Line...';
  SetShortCut(Word('G'), [ssCtrl]);
end;

procedure TTextGoToLine.ExecuteTextTarget(Target: TLog);
begin
  Target.ShowGotoLineDialog;
end;

{ TTextIncrement }

constructor TTextIncrement.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Increment|Increment number under cursor';
  Caption := '&Increment';
  SetShortCut(VK_F11, []);
end;

procedure TTextIncrement.ExecuteTextTarget(Target: TEditLog);
begin
  if not Target.IncrementNumberAtCursor(1) then
    SetResultForControl(Target, 'No number to increment');
end;

{ TTextDecrement }

constructor TTextDecrement.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Decrement|Decrement number under cursor';
  Caption := '&Decrement';
  SetShortCut(VK_F11, [ssShift]);
end;

procedure TTextDecrement.ExecuteTextTarget(Target: TEditLog);
begin
  if not Target.IncrementNumberAtCursor(-1) then
    SetResultForControl(Target, 'No number to decrement');
end;

{ TTextRecordMacro }

constructor TTextRecordMacro.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Record|Start recording keystrokes';
  Caption := '&Record KeyStrokes';
  SetShortCut(VK_F12, [ssShift]);
end;

procedure TTextRecordMacro.ExecuteTextTarget(Target: TEditLog);
begin
  Target.Recording := not Target.Recording;
  if Target.Recording then begin
    Hint := 'Stop Recording|Stop recording keystrokes';
    Caption := 'Stop &Recording';
  end else begin
    Hint := 'Record|Start recording keystrokes';
    Caption := '&Record KeyStrokes';
  end
end;

{ TTextPlayBackMacro }

constructor TTextPlayBackMacro.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Playback|Play back previously recorded keystrokes';
  Caption := '&Play Macro';
  SetShortCut(VK_F12, []);
end;

procedure TTextPlayBackMacro.ExecuteTextTarget(Target: TEditLog);
begin
  Target.PlayBack;
end;

{ TTextRemoveTabs }

constructor TTextRemoveTabs.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Remove Tabs|Replace tabs with spaces in selected region';
  Caption := '&Remove Tabs';
end;

procedure TTextRemoveTabs.ExecuteTextTarget(Target: TEditLog);
begin
  Target.RemoveTabs;
end;

{ TTextSetTabSpacing }

constructor TTextSetTabSpacing.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Hint := 'Set Tab spacing|Set tab spacing';
  Caption := 'Set Tab Spacing...';
end;

procedure TTextSetTabSpacing.ExecuteTextTarget(Target: TEditLog);
var
  s: string;
begin
  s := IntToStr(Target.TabSpacing);
  if InputQuery('Set Tab Spacing', 'Tab spacing in characters', s) then
    Target.TabSpacing := StrToInt(s);
end;

end.
