unit NGrid;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses Messages, Graphics, Controls, Forms, Dialogs, Menus, StdCtrls, Grids,
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     SysUtils, Classes, Huge, MathSup, MatrixVec, TextStrm, GridEx;

type
  TOnChangedEvent = procedure (Sender: TObject; Col, Row: Longint) of object;
  TOnCellEvent = procedure (Sender: TObject; Col, Row: Longint; var v:Double; var Valid:Boolean) of object;
  TChangeCellQueryEvent = procedure (Sender: TObject; Col, Row: Longint; var CanChange: Boolean) of object;
  TOnGetCellLabelEvent = procedure (Sender: TObject; Col, Row: Longint; var s:string; var IsText:Boolean) of object;
  TOnGetCellIsFlagEvent = procedure (Sender: TObject; Col, Row: Longint; var IsFlag: Boolean) of object;
  TDrawCellContentsEvent = procedure (Sender: TObject; ACol, ARow: Longint;
    Rect: TRect; State: TGridDrawState; var Drawn: Boolean) of object;
  {$IFDEF LCL_}
  TMovedEvent = procedure (Sender: TObject; FromIndex, ToIndex: Longint) of object;
  TDrawCellEvent = TOnDrawCell;
  TSelectCellEvent = TOnSelectCellEvent;
  {$ENDIF}

  { TNumberGrid }

  TNumberGrid = class(TCustomGrid)
  private
    FMatrix: TMatrix;
    FCopyLabels: Boolean;
    InAreaSelection:(asNone, asCol, asRow, asAll);
    FOnColumnMoved: TMovedEvent;
    FOnDrawCell: TDrawCellEvent;
    FOnDrawCellContents: TDrawCellContentsEvent;
    {$IFNDEF LCL_}
    FOnGetEditMask: TGetEditEvent;
    {$ENDIF}
    FOnGetEditText: TGetEditEvent;
    FOnRowMoved: TMovedEvent;
    FOnSelectCell: TSelectCellEvent;
    FOnGetCell: TOnCellEvent;
    FOnGetCellLabel: TOnGetCellLabelEvent;
    FOnGetCellIsFlag: TOnGetCellIsFlagEvent;
    FOnSetCell: TOnCellEvent;
    FOnSetEditText: TSetEditEvent;
    FOnTopLeftChanged: TNotifyEvent;
    FOnChangeCellQuery: TChangeCellQueryEvent;
    FTitles: TStringList;
    FLbls: TStringList;
    FOnChanged: TOnChangedEvent;
    FStatusWindow: TControl;
    LastCol, LastRow:Longint; { Last Cell for error entry }
    LastString:string;
    SelectingAnchor:TGridCoord;
    SavedOptions:TGridOptions;
    OriginalText:string;
    OriginalModified:Boolean;
    InhibitNextInvalidate: Boolean;
    LastCell: TGridCoord; { Last left button press }
    FNGEditor: TWinControl;
    FAllowText: Boolean;
    FOnGetEditStyle: TGetEditStyleEvent;
    FOnGetPickListItems: TOnGetPickListItems;
    FOnEditButtonClick: TNotifyEvent;
    {$IFDEF OLDLCL_}
    FDropDownEditor: TComboBox;
    {$ENDIF}
    procedure TrackSelectedArea(X, Y: Integer);
    {$IFDEF VCL_} { TODO: WINLCL also? }
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
    {$ENDIF}
    function GetTextN: TCaption;
    procedure SetTextN(const AValue: TCaption);
  protected
    FUpdateCount: Integer;
    CurrentZeroWidth: Integer;
    EndPending: Boolean;
    {$IFNDEF LCL_}
    function CreateEditor: TInplaceEdit; override;
    function GetEditStyle(ACol, ARow: Longint): TEditStyle; override;
    function GetEditMask(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF}; override;
    {$ELSE}
    procedure SelectEditor; override;
    {$ENDIF}
    procedure CheckSizes; virtual;
    procedure ColumnMoved(FromIndex, ToIndex: Longint); {$IFNDEF LCL_} override; {$ENDIF}
    procedure RowMoved(FromIndex, ToIndex: Longint); {$IFNDEF LCL_} override; {$ENDIF}
    {$IFDEF LCL_}
    procedure ColRowMoved(IsColumn: Boolean; FromIndex,ToIndex: Integer); override;
    // procedure EditorDoGetValue; override;
    procedure FocusCell(ACol, ARow: Longint; UpdateSelection: Boolean);
    procedure DrawFocusRect(aCol,aRow: Integer; ARect: TRect); override;
    procedure DoCopyToClipboard; override;
    procedure DoCutToClipboard; override;
    procedure DoPasteFromClipboard; override;
    {$IFDEF OLDLCL_}
    function CreateDropDownEditor: TComboBox;
    procedure EditorDoSetValue; override;
    {$ENDIF}
    {$ENDIF}
    procedure DrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState); override;
    function GetEditText(ACol, ARow: Longint): {$IFNDEF CLX_} string; {$ELSE} WideString; {$ENDIF} override;
    function CanEditShow: Boolean; override;
    function CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean; override;
    function SelectCell(ACol, ARow: Longint): Boolean; override;
    procedure SetEditText(ACol, ARow: Longint; const s: {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF}); override;
    procedure TopLeftChanged; override;
    procedure SetTitles(l:TStringList);
    function GetTitle(i:Longint):string;
    procedure SetTitle(i:Longint; s:string);
    procedure SetLbls(l:TStringList);
    function GetLbl(i:Longint):string;
    procedure SetLbl(i:Longint; s:string);
    function GetValid(ARow, ACol:Longint):Boolean;
    procedure SetValid(ARow, ACol:Longint; IsValid:Boolean);
    function GetValue(ARow, ACol:Longint):Double;
    procedure SetValue(ARow, ACol:Longint; const v:Double);
    procedure SizeChanged(OldColCount, OldRowCount: Longint); override;
    procedure ShowStatus(ACol, ARow: Longint);
    procedure SetCell(ARow, ACol: Longint; v:Double; IsValid:Boolean); virtual;
    function GetPointer(ARow, ACol:Longint):Pointer;
    property Item[ARow, ACol: Longint]:Pointer read GetPointer;
    function CanChange(ARow, ACol:Longint):Boolean; virtual;
    procedure KeyPress(var Key: Char); override;
    procedure DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
  public
    Modified:Boolean;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean = False; UseContents: Boolean = False);
    procedure Blank;
    function DataExtent:TGridRect;
    function NaturalWidth: Integer;
    procedure DefaultDrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
    procedure SaveAreaToStream(Stream:TStream; Area:TGridRect; WithLabels:Boolean = False);
    procedure SaveToStream(Stream:TStream);
    procedure OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint; IncludeNonNumeric: Boolean = False);
    procedure Loaded; override;
    procedure LoadFromStream(Stream:TStream);
    procedure Changed(ACol, ARow: Longint); virtual;
    procedure ClearArea(Area:TGridRect);
    procedure ClearColumn(ACol: Longint);
    procedure CancelSelection;
    procedure Clear;
    procedure CopyToClipboard(WithLabels:Boolean); overload;
    procedure SelectAll;
    procedure MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure SetFocus; override;
    function GetCell(ARow, ACol: Longint; out v:Double):Boolean; virtual;
    function GetCellLabel(ARow, ACol: Longint; out s: string):Boolean;
    function GetCellIsFlag(ARow, ACol: Longint):Boolean;
    procedure ExchangeRows(i, j: Longint);
    procedure Update; override;
    property Canvas;
    property Col;
    property ColWidths;
    property EditorMode;
    property InplaceEditor;
    property GridHeight;
    property GridWidth;
    property LeftCol;
    property Selection;
    property Row;
    property RowHeights;
    {$IFNDEF LCL_}
    property TabStops;
    {$ENDIF}
    property TopRow;

    property Title[i:LongInt]:string read GetTitle write SetTitle;
    property Lbl[i:LongInt]:string read GetLbl write SetLbl;
    property Valid[ARow, ACol:Longint]:Boolean read GetValid write SetValid;
    property Value[ARow, ACol:Longint]:Double read GetValue write SetValue; default;
    {$IFDEF OLDLCL_}
    property DropDownEditor: TComboBox read FDropDownEditor;
    {$ENDIF}

    procedure Invalidate; override;
    procedure InvalidateCol(ACol: Longint);
    procedure BeginUpdate;
    procedure EndUpdate;
  published
    procedure ClearSelection;
    procedure CopyToClipboard; overload;
    procedure PasteFromClipboard;
    function CanCopyToClipboard: Boolean;
    function CanPasteFromClipboard: Boolean;

    property Align;
    property Anchors;
    {$IFNDEF LCL_}
    property BorderStyle;
    {$ENDIF}
    property Caption read GetTextN write SetTextN;
    property Color;
    property ColCount;
    {$IFDEF MSWINDOWS}
    {$IFNDEF FPC}
    property Ctl3D;
    property ParentCtl3D;
    {$ENDIF}
    property DragCursor;
    {$ENDIF}
    property DefaultColWidth;
    property DefaultRowHeight;
    property DefaultDrawing;
    property DragMode;
    property Enabled;
    property FixedColor;
    property FixedCols;
    property FixedRows;
    property Font;
    property GridLineWidth;
    property Options;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property RowCount;
    property ScrollBars;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property VisibleColCount;
    property VisibleRowCount;
    property OnClick;
    property OnColumnMoved: TMovedEvent read FOnColumnMoved write FOnColumnMoved;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnDrawCell: TDrawCellEvent read FOnDrawCell write FOnDrawCell;
    property OnDrawCellContents: TDrawCellContentsEvent read FOnDrawCellContents write FOnDrawCellContents;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    {$IFNDEF LCL_}
    property OnGetEditMask: TGetEditEvent read FOnGetEditMask write FOnGetEditMask;
    {$ENDIF}
    property OnGetEditText: TGetEditEvent read FOnGetEditText write FOnGetEditText;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnRowMoved: TMovedEvent read FOnRowMoved write FOnRowMoved;
    property OnSelectCell: TSelectCellEvent read FOnSelectCell write FOnSelectCell;
    property OnSetEditText: TSetEditEvent read FOnSetEditText write FOnSetEditText;
    property OnTopLeftChanged: TNotifyEvent read FOnTopLeftChanged write FOnTopLeftChanged;
    property OnGetCellLabel: TOnGetCellLabelEvent read FOnGetCellLabel write FOnGetCellLabel;
    property OnGetCellIsFlag: TOnGetCellIsFlagEvent read FOnGetCellIsFlag write FOnGetCellIsFlag;

    property OnChanged: TOnChangedEvent read FOnChanged write FOnChanged;
    property CopyLabels: Boolean read FCopyLabels write FCopyLabels;
    property Titles:TStringList read FTitles write SetTitles;
    property Lbls:TStringList read FLbls write SetLbls;
    property StatusWindow: TControl read FStatusWindow write FStatusWindow;
    property OnGetCell: TOnCellEvent read FOnGetCell write FOnGetCell;
    property OnSetCell: TOnCellEvent read FOnSetCell write FOnSetCell;
    property OnChangeCellQuery: TChangeCellQueryEvent read FOnChangeCellQuery write FOnChangeCellQuery;

    property AllowText: Boolean read FAllowText write FAllowText default False;
    property OnGetEditStyle: TGetEditStyleEvent read FOnGetEditStyle write FOnGetEditStyle;
    property OnGetPickListItems: TOnGetPickListItems read FOnGetPickListItems write FOnGetPickListItems;
    property OnEditButtonClick: TNotifyEvent read FOnEditButtonClick write FOnEditButtonClick;
  end;

{ Helper prcedure for filling grid }
procedure TextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Alignment: TAlignment = taLeftJustify);
procedure ColouredTextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Colour: TColor; Alignment: TAlignment = taLeftJustify);
procedure NumberInGrid(Canvas: TCanvas; ARect: TRect; Value: Double; var CurrentZeroWidth: Integer; Alignment: TAlignment = taLeftJustify);
function NaturalWidth(Grid: TCustomGrid): Integer;

implementation

uses {$IFDEF LCL_} LCLType, Types, LCLIntf, {$ENDIF}
     ClipBrd, PropHelp, HugeClip, Math, Vector, WCompat, UCompat;

{$IFDEF SINGLE}
{$R NGrid.d32}
{$ENDIF}

type

  TNGInplaceEdit = class({$IFDEF LCL_}TStringCellEditor{$ELSE}TInplaceEditList{$ENDIF})
  private
    Grid:TNumberGrid;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
  public
    constructor CreateWithGrid(AOwner:TNumberGrid);
    property ReadOnly;
  published
    procedure PasteFromClipboard; { Implementation relies on this being a non-virtual function }
  end;


constructor TNumberGrid.Create(AOwner: TComponent);
begin
  FTitles := TStringList.Create;
  FLbls := TStringList.Create;
  inherited Create(AOwner);
  {$IFDEF LCL_}
    ExtendedSelect := True;
  {$ENDIF}
  LastCol := -1;
  CheckSizes;
end;

destructor TNumberGrid.Destroy;
begin
  FMatrix.Free;
  FTitles.Free;
  FLbls.Free;
  {$IFDEF OLDLCL_}
  if Assigned(DropDownEditor) then
    FreeAndNil(FDropDownEditor);
  {$ENDIF}
  {$IFDEF LCL_}
  FNGEditor.Free;
  {$ENDIF}
  inherited Destroy;
end;

procedure TNumberGrid.Loaded;
begin
  inherited Loaded;
  CheckSizes;
end;

procedure TNumberGrid.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TNumberGrid.EndUpdate;
begin
  Dec(FUpdateCount);
  if FUpdateCount = 0 then Invalidate;
end;


procedure TNumberGrid.InvalidateCol(ACol: Longint);
begin
  if FUpdateCount = 0 then
    Invalidate;
{ inherited InvalidateCol(ACol) is broken in WIN32 and does not exist in DELPHI 1 }
end;

procedure TNumberGrid.Changed(ACol, ARow: Longint);
begin
  if FUpdateCount = 0 then
    if ACol < 0 then
      if ARow < 0 then
        Invalidate
      else
        InvalidateRow(ARow)
    else if ARow < 0 then
      InvalidateCol(ACol)
    else
      InvalidateCell(ACol, ARow);
  if (ARow = Row) and (ACol = Col) then ShowStatus(ACol, ARow);
  if Assigned(FOnChanged) and not (csDesigning in ComponentState) then
    FOnChanged(Self, ACol, ARow);
end;

procedure TNumberGrid.Update;
begin
  CheckSizes;
  inherited Update;
end;

procedure TNumberGrid.CheckSizes;
var
  j: Longint;
  AChange: Boolean;
begin
  AChange := not Assigned(FMatrix)
             or (FMatrix.RowCount <> RowCount - FixedRows)
             or (FMatrix.ColCount <> ColCount - FixedCols);
  if not Assigned(FMatrix) then begin
    FMatrix := TMatrix.Create(RowCount-FixedRows, ColCount-FixedCols);
    FMatrix.Clear;
  end else
    FMatrix.Resize(RowCount - FixedRows, ColCount - FixedCols, True);
  if Titles.Count >= ColCount then begin
    for j := Titles.Count-1 downto ColCount-1 do
      Titles.Delete(j);
    AChange := True;
  end;
  if Lbls.Count >= RowCount then begin
    for j := Lbls.Count-1 downto RowCount-1 do
      Lbls.Delete(j);
    AChange := True;
  end;
  if AChange then
    Invalidate;
end;

procedure TNumberGrid.SizeChanged(OldColCount, OldRowCount: Longint);
begin
  inherited SizeChanged(OldColCount, OldRowCount);
  CheckSizes;
end;

procedure TNumberGrid.CancelSelection;
var
  s:TGridRect;
begin
  s.Left := Col; s.Right := Col;
  s.Bottom := Row; s.Top := Row;
  Selection := s;
end;

constructor TNGInPlaceEdit.CreateWithGrid(AOwner:TNumberGrid);
begin
  Create(nil); { So it doesn't show in the designer }
  Grid := AOwner;
  AutoSize := True;
end;

{$IFDEF LCL_}
procedure TNumberGrid.SelectEditor;
var
  aEditor: TWinControl;
  Style: TEditStyle;
  List: TStrings;
  i: Integer;
begin
  Style := esSimple;
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Style);
  if Style = esSimple then begin
    if not Assigned(FNGEditor) then begin
      FNGEditor := TNGInplaceEdit.CreateWithGrid(Self);
      FNGEditor.Visible := False;
    end;
    aEditor := FNGEditor;
  end else begin
    {$IFDEF OLDLCL_}
    if not Assigned(DropDownEditor) then begin
      FDropDownEditor := CreateDropDownEditor;
      FDropDownEditor.Visible := False;
    end;
    aEditor:= DropDownEditor;
    {$ENDIF}
    aEditor := EditorByStyle(cbsPickList);
    if (goEditing in Options) and Assigned(OnSelectEditor) then
      OnSelectEditor(Self, Col, Row, aEditor);
    List := TStringList.Create;
    try
      { To avoid crashes, only write to dropdown list if it has changed }
      DoGetPicklistItems(Col, Row, List);
      with aEditor as TCustomComboBox do begin
        if Items.Count <> List.Count then
          Items.Assign(List)
        else
          for i := 0 to List.Count-1 do
            Items[i] := List[i];
        DropDownCount := List.Count;
        AutoSize := True;
      end;
    finally
      List.Free;
    end;
 end;
 if aEditor<>Editor then
   Editor:=aEditor;
end;

{$IFDEF OLDLCL_}
function TNumberGrid.CreateDropDownEditor: TComboBox;
begin
  Result := TStringGridCellComboBox.Create(Self);
  Result.Style := csDropDownList;
  Result.Parent := Self.Parent;
end;
{$ENDIF}
{$ELSE}
function TNumberGrid.CreateEditor: TInplaceEdit;
begin
  Result := TNGInplaceEdit.CreateWithGrid(Self);
//  Result := TInplaceEditList.Create(Self);
  TInplaceEditList(Result).OnGetPickListitems := DoGetPicklistItems;
  TInplaceEditList(Result).OnEditButtonClick := FOnEditButtonClick;
  FNGEditor := Result;
end;

function TNumberGrid.GetEditStyle(ACol, ARow: Longint): TEditStyle;
begin
  Result := inherited GetEditStyle(ACol, ARow);
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Result);
end;

function TNumberGrid.GetEditMask(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF};
begin
  Result := '';
  if Assigned(FOnGetEditMask) then FOnGetEditMask(Self, ACol, ARow, Result);
end;
{$ENDIF}

procedure TNGInplaceEdit.PasteFromClipboard;
begin
  Grid.PasteFromClipboard; { Relies on PasteFromClipboard being a non-virtual function }
end;

procedure TNGInplaceEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  case Key of
    VK_ESCAPE:
      begin
        Text := Grid.OriginalText;
        Grid.EditorMode := False;
        Grid.Modified := Grid.OriginalModified;
        Key := 0;
      end;
    VK_LEFT:
      if (SelLength = 0) and (SelStart = 0) then begin
        Grid.KeyDown(Key, Shift);
        Key := 0;
      end;
    VK_RIGHT:
      if SelStart = Length(Text) then begin
        Grid.KeyDown(Key, Shift);
        Key := 0;
      end;
  else
    inherited KeyDown(Key, Shift);
  end;
end;

{$IFDEF LCL_}
procedure TNumberGrid.ColRowMoved(IsColumn: Boolean; FromIndex, ToIndex: Integer);
begin
  inherited ColRowMoved(IsColumn, FromIndex, ToIndex);
  if IsColumn then
    ColumnMoved(FromIndex, ToIndex)
  else
    RowMoved(FromIndex, ToIndex)
end;
{$ENDIF}

procedure TNumberGrid.ColumnMoved(FromIndex, ToIndex: Longint);
begin
  FMatrix.MoveColumn(FromIndex, ToIndex);
  if Assigned(FOnColumnMoved) then FOnColumnMoved(Self, FromIndex, ToIndex);
  Changed(-1, -1);
end;

procedure TNumberGrid.RowMoved(FromIndex, ToIndex: Longint);
begin
  FMatrix.MoveRow(FromIndex,ToIndex);
  if Assigned(FOnRowMoved) then FOnRowMoved(Self, FromIndex, ToIndex);
  Changed(-1, -1);
end;

{$IFDEF LCL_}
{$IFDEF OLDLCL_}
procedure TNumberGrid.EditorDoSetValue;
var
  Msg: TGridMessage;
  List: TStringList;
  i: Integer;
begin
  if Assigned(Editor) and (Editor = DropDownEditor) then begin
    Msg.LclMsg.Msg:=GM_SETGRID; { Lockout Changes while list being made up }
    Msg.Grid:=nil;
    Editor.Dispatch(Msg);
    List := TStringList.Create;
    try
      { To avoid crashes, only write to dropdown list if it has changed }
      DoGetPicklistItems(Col, Row, List);
      if DropDownEditor.Items.Count <> List.Count then
        DropDownEditor.Items.Assign(List)
      else
        for i := 0 to List.Count-1 do
          DropDownEditor.Items[i] := List[i]
    finally
      List.Free;
    end;
    Msg.Grid:=Self;
    Editor.Dispatch(Msg);
  end;
  inherited EditordoSetValue;
end;
{$ENDIF}

procedure TNumberGrid.FocusCell(ACol, ARow: Longint; UpdateSelection: Boolean);
begin
  if not UpdateSelection then
    SelectActive := False;
  MoveExtend(False, ACol, ARow);
end;

procedure TNumberGrid.DrawFocusRect(aCol, aRow: Integer; ARect: TRect);
var
  OldColour: TColor;
begin
  inherited DrawFocusRect(aCol, aRow, ARect);
  // Copied from TDrawGrid
  // Draw focused cell if we have the focus
  if Self.Focused Or ((goEditing in Options)and(goAlwaysShowEditor in Options) and ((Editor=nil) or not Editor.Focused)) then
  begin
    OldColour := Canvas.Pen.Color;
    Canvas.Pen.Color:=FocusColor;
    Canvas.Pen.Style:=psDot;
    if goRowSelect in Options then begin
      Canvas.MoveTo(GCache.FixedWidth+1, aRect.Top);
      Canvas.LineTo(GCache.MaxClientXY.x-2, aRect.Top);
      Canvas.LineTo(GCache.MaxClientXY.x-2, aRect.Bottom-2);
      Canvas.LineTo(GCache.FixedWidth+1, aRect.Bottom-2);
      Canvas.LineTo(GCache.FixedWidth+1, aRect.Top+1);
    end else begin
      Canvas.MoveTo(aRect.Left, aRect.Top);
      Canvas.LineTo(ARect.Right-2,aRect.Top);
      Canvas.LineTo(aRect.Right-2,aRect.bottom-2);
      Canvas.LineTo(aRect.Left, aRect.Bottom-2);
      Canvas.Lineto(aRect.left, aRect.top+1);
    end;
    Canvas.Pen.Style:=psSolid;
    Canvas.Pen.Color:=OldColour;
  end;
end;

procedure TNumberGrid.DoCopyToClipboard;
begin
  CopyToClipboard;
end;

procedure TNumberGrid.DoCutToClipboard;
begin
  CopyToClipboard;
  ClearSelection;
end;

procedure TNumberGrid.DoPasteFromClipboard;
begin
  PasteFromClipboard;
end;

{$ENDIF}

procedure TNumberGrid.SetEditText(ACol, ARow: Longint; const s: {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF});
var
  AValue: Double;
  OK, Flg: Boolean;
begin
  LastCol := -1;
  if Length(s) <> 0 then begin
    if GetCellIsFlag(ARow, ACol) then begin
      OK := TryStrToBoolean(s, Flg);
      SetValid(ARow, ACol, Flg and OK);
    end else begin
      OK := TryStrToFloat(Trim(s), AValue);
      if OK then
        SetCell(ARow, ACol, AValue, True)
    end;
    if not OK then begin
      SetValid(ARow, ACol, False);
      LastCol := ACol; { Save text, to allow user to edit it }
      LastRow := ARow;
      LastString := s;
    end
  end else
    SetValid(ARow, ACol, False);
  if Assigned(FOnSetEditText) then FOnSetEditText(Self, ACol, ARow, s);
end;

function TNumberGrid.GetEditText(ACol, ARow: Longint): {$IFNDEF CLX_} string {$ELSE} WideString {$ENDIF};
var
  v: Double;
begin
  Result := '';
  if EditorMode and ((Selection.Left <> ACol) or (Selection.Right <> ACol)
                      or (Selection.Top <> ARow) or (Selection.Bottom <> ARow)) then
    CancelSelection; { If typing, then only one cell should be selected }
  if (LastCol <> ACol) or (LastRow <> ARow) then begin
    if (LastCol >= 0) then begin
      InvalidateCell(LastCol, LastRow);
      LastCol := -1;
    end;
    if GetCellIsFlag(ARow, ACol) then
      GetCellLabel(ARow, ACol, Result)
    else if GetCell(ARow, ACol, v) then
      Result := FloatToStr(v);
    if Assigned(InplaceEditor) and (InplaceEditor is TNGInplaceEdit) then
      TNGInplaceEdit(InplaceEditor).ReadOnly := not CanChange(ARow, ACol);
  end else
    Result := LastString;
  if Assigned(FOnGetEditText) then FOnGetEditText(Self, ACol, ARow, Result);
  OriginalText := Result;
  OriginalModified := Modified;
end;

function TNumberGrid.CanEditShow: Boolean;
begin
  Result := inherited CanEditShow and not GetCellIsFlag(Row, Col);
  if Result and Assigned(FOnChangeCellQuery) then begin
    FOnChangeCellQuery(Self, Col, Row, Result);
  end;
  if not Result then EditorMode := False;
end;

{$IFDEF UNIX}
type TOpenControl = class(TControl)
  property Caption;
  end;
{$ENDIF}

procedure TNumberGrid.ShowStatus(ACol, ARow: Longint);
var
  v: Double;
  Buffer: String;
begin
  if Assigned(FStatusWindow) then begin
    if not GetCellLabel(ARow, ACol, Buffer) then
      if GetCell(ARow, ACol, v) then
        Buffer := FloatToStr(v)
      else
        Buffer := '';
{$IFDEF MSWINDOWS}
    FStatusWindow.SetTextBuf(PChar(Buffer));
{$ENDIF}
{$IFDEF UNIX}
    TOpenControl(FStatusWindow).Caption := Buffer; // TODO: Check
{$ENDIF}
  end;
end;

function TNumberGrid.CanGridAcceptKey(Key: Word; Shift: TShiftState): Boolean;
var
  r, nr, c, nc: Integer;
  FirstCellSet, FirstTime: Boolean;
  v: Double;
begin
  Result := False;
  if Key = VK_END then begin
    if ssCtrl in Shift then begin
      with DataExtent.BottomRight do
        FocusCell(X, Y, not (ssShift in Shift));
    end else
      EndPending := True;
      Exit;
  end else if Key = VK_DELETE then
    ClearSelection
  else if Shift*[ssShift, ssAlt, ssCtrl] = [ssCtrl] then
    case Key of
      Ord('C'), VK_INSERT: CopyToClipboard;
      Ord('V'): PasteFromClipboard;
      Ord('X'): begin CopyToClipboard; ClearSelection; end;
    else
      Result := True;
    end
  else if EndPending and (Key in [VK_DOWN, VK_UP, VK_LEFT, VK_RIGHT]) then begin
    { Find target cell }
    c := Col;
    r := Row;
    FirstCellSet := GetCell(r, c, v);
    FirstTime := True;
    repeat
      nc := c;
      nr := r;
      case Key of
        VK_DOWN:  if r >= RowCount-1 then Break else Inc(nr);
        VK_UP:    if r <= FixedRows then Break else Dec(nr);
        VK_LEFT:  if c <= FixedCols then Break else Dec(nc);
        VK_RIGHT: if c >= ColCount-1 then Break else Inc(nc);
      end;
      if GetCell(nr, nc, v) xor FirstCellSet then
        if FirstTime then
          FirstCellSet := not FirstCellSet
        else
          Break;
      FirstTime := False;
      c := nc;
      r := nr;
    until False;
    FocusCell(c, r, not (ssShift in Shift));
  end else
    Result := True;
  EndPending := False;
end;

function TNumberGrid.SelectCell(ACol, ARow: Longint): Boolean;
begin
  Result := True;
  if Assigned(FOnSelectCell) then FOnSelectCell(Self, ACol, ARow, Result);
  if Result then
    ShowStatus(ACol, ARow);
end;

procedure TNumberGrid.DrawCell(ACol, ARow: Longint; ARect: TRect;
  AState: TGridDrawState);
begin
  {$IFDEF LCL_}
  inherited DrawCell(ACol, ARow, ARect, AState);
  {$ENDIF}
  if DefaultDrawing then DefaultDrawCell(ACol, ARow, ARect, AState);
  if Assigned(FOnDrawCell) and not (csDesigning in ComponentState) then FOnDrawCell(Self, ACol, ARow, ARect, AState);
end;

procedure TNumberGrid.TopLeftChanged;
begin
  inherited TopLeftChanged;
  if Assigned(FOnTopLeftChanged) then FOnTopLeftChanged(Self);
end;

procedure TNumberGrid.SetTitles(l:TStringList);
begin
  FTitles.Assign(l);
  Changed(-1, 0);
end;

function TNumberGrid.GetTitle(i:Longint):string;
begin
  if (i > 0) and (i <= FTitles.Count) then
    Result := FTitles[i-1]
  else if i = 0 then
    Result := Caption
  else
    Result := '';
end;

procedure TNumberGrid.SetTitle(i:Longint; s:string);
begin
  if (i > FTitles.Count) and (i <= ColCount-1) then
    while FTitles.Count < i do FTitles.Add('');
  if (i > 0) and (i <= FTitles.Count) then
    FTitles[i-1] := s
  else if i = 0 then
    Caption := s;
  Changed(i, 0);
end;

procedure TNumberGrid.SetLbls(l:TStringList);
begin
  FLbls.Assign(l);
  Changed(0, -1);
end;

function TNumberGrid.GetLbl(i:Longint):string;
begin
  if (i > 0) and (i <= Lbls.Count) then
    Result := Lbls[i-1]
  else if i = 0 then
    Result := Caption
  else
    Result := '';
end;

procedure TNumberGrid.SetLbl(i:Longint; s:string);
begin
  if (i > Lbls.Count) and (i <= RowCount-1) then
    while Lbls.Count < i do Lbls.Add('');
  if (i > 0) and (i <= Lbls.Count) then
    Lbls[i-1] := s
  else if i = 0 then
    Caption := s;
  Changed(0, i);
end;

function TNumberGrid.GetTextN: TCaption;
begin
  Result := inherited Caption;
end;

procedure TNumberGrid.SetTextN(const AValue: TCaption);
begin
  inherited Caption := AValue;
  Changed(0, 0);
end;

procedure TNumberGrid.AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean; UseContents: Boolean);
var
  r: Integer;
  s: string;
  MaxSize, sz: TSize;
begin
  r := 0;
  MaxSize.cx := 0;
  MaxSize.cy := 0;
  { Autosize on column and row headings only }
  while (r < RowCount) and ( UseContents or (r < FixedRows) or (aCol < FixedCols) ) do begin
    if GetCellLabel(r, aCol, s) then begin
      sz := Canvas.TextExtent(s);
      with MaxSize do begin
        cx := Max(cx,sz.cx);
        cy := Max(cy,sz.cy);
      end;
    end;
    Inc(r);
  end;
  if MaxSize.cx = 0 then
    MaxSize.cx := DefaultColWidth
  else
    Inc(MaxSize.cx, 6);
  if not GrowOnly or (ColWidths[aCol] < MaxSize.cx) then
    ColWidths[aCol] := MaxSize.cx;
  {$IFDEF LCLcarbon}
  Inc(MaxSize.cy,4);
  if MaxSize.cy > DefaultRowHeight then
    DefaultRowHeight := MaxSize.cy;
  {$ENDIF}
end;

{$IFDEF VCL_}
{ // TODO: Linux? }
procedure TNumberGrid.CMFontChanged(var Message: TMessage);
begin
  inherited;
  CurrentZeroWidth := 0; { Flag that actual value needs to be determined }
end;

procedure TNumberGrid.WMChar(var Msg: TWMChar);
begin
  if (goEditing in Options) and (Char(Msg.CharCode) in [^H, #32..#255])
     and GetCellIsFlag(Row,Col) then
    DoKeyPress(Msg) { Do not try to send key to inplace editor, which would loose it }
  else
    inherited;
end;
{$ENDIF}

function TNumberGrid.GetCellLabel(ARow, ACol: Longint;  out s: string): Boolean;
begin
  Result := True;
  if (ARow = LastRow) and (ACol = LastCol) then
    s := 'Bad Value'
  else if ARow < FixedRows then begin
    if ARow >= FixedRows - 1 then
      s := Title[ACol]
    else
      s := '';
    if (Length(s) = 0) and (ACol <> 0) then
      s := Char(Ord('A')+Acol-1);
  end else if (ACol < FixedCols)
              or ((ACol = 0) and (FixedCols = 0) and (Lbls.Count > 0)) then begin
    if (ACol = Max(FixedCols-1,0)) and (ARow > 0) and (ARow <= Lbls.Count) then
      s := Lbls[ARow-1]
    else if ARow = 0 then
      s := Caption
    else
      s := IntToStr(ARow);
  end else if GetCellIsFlag(ARow,ACol) then
    if Valid[ARow, ACol] then
      s := 'yes'
    else
      s := 'no'
  else begin
    Result := False;
    s := '';
  end;
  if Assigned(FOnGetCellLabel) and not (csDesigning in ComponentState) then
    FOnGetCellLabel(Self, ACol, ARow, s, Result);
end;

function TNumberGrid.GetCellIsFlag(ARow, ACol: Longint):Boolean;
begin
  Result := False;
  if Assigned(FOnGetCellIsFlag) then
    FOnGetCellIsFlag(Self, ACol, ARow, Result);
end;

{ Helper procedures for filling grid }
procedure TextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Alignment: TAlignment);
var
  uFormat: UINT;
  x: Integer;
{$IFDEF LCL_}
  TextStyle: TTextStyle;
{$ENDIF}
begin
  {$IFDEF VCL_}
    {$IFDEF UNICODE}
      Canvas.TextRect(ARect, ARect.Left + 2, ARect.Top + 2, Text);
    {$ELSE}
      uFormat := GetTextAlign(Canvas.Handle);
      case Alignment of
        taLeftJustify:  x := ARect.Left + 2;
        taRightJustify: begin x := ARect.Right - 2; SetTextAlign(Canvas.Handle, uFormat or TA_RIGHT); end;
        taCenter: begin x := (ARect.Right+ARect.Left) div 2; SetTextAlign(Canvas.Handle, uFormat or TA_CENTER); end;
      end;
      ExtTextOut(Canvas.Handle, x, ARect.Top + 2,
                 ETO_CLIPPED or ETO_OPAQUE, @ARect, AddressOfFirstChar(Text), Length(Text), nil);
      if Alignment <> taLeftJustify then
        SetTextAlign(Canvas.Handle, uFormat);
    {$ENDIF}
  {$ENDIF}
  {$IFDEF LCL_}
  Dec(ARect.Right, 2);
  Dec(ARect.Bottom, 2);
  TextStyle := Canvas.TextStyle;
  TextStyle.Alignment := taLeftJustify;
  TextStyle.Layout := tlTop;
  TextStyle.Clipping := True;
  TextStyle.WordBreak := False;
  TextStyle.SingleLine := True;
  TextStyle.Opaque := True;
  //Canvas.FillRect(ARect); // LCL bug(?) means above fails, while this is OK, and TextRect calls FillRect anyway.
  //                        // Problem seems to be that GTK2 fails of SetBKMode(OPAQUE);
  Canvas.TextRect(ARect, ARect.Left + 2, ARect.Top + 2, Text, TextStyle);
  {$ENDIF}
  {$IFDEF CLX_}
  Canvas.FillRect(ARect);
  Canvas.TextRect(ARect, ARect.Left + 2, ARect.Top + 2, Text,
                  Integer(AlignmentFlags_AlignLeft) or Integer(AlignmentFlags_AlignTop));
  {$ENDIF}
end;

procedure ColouredTextInGrid(Canvas: TCanvas; ARect: TRect; Text: string; Colour: TColor; Alignment: TAlignment);
var
  OldStyle: TBrushStyle;
  OldColour: TColor;
  {$IFDEF VCL_}
  Current: COLORREF;
  {$ENDIF}
begin
  {$IFDEF VCL_}
  Current := GetBkColor(Canvas.Handle);
  SetBkColor(Canvas.Handle, clRed);
  {$ENDIF}
  OldColour := Canvas.Brush.Color;
  OldStyle := Canvas.Brush.Style;
  Canvas.Brush.Color := clRed;
  Canvas.Brush.Style := bsSolid;
  TextInGrid(Canvas, ARect, Text);
  { This restore currently fails under gtk2 - ? widgetset bug }
  Canvas.Brush.Color := OldColour;
  Canvas.Brush.Style := OldStyle;
  {$IFDEF VCL_}
  SetBkColor(Canvas.Handle, Current);
  {$ENDIF}
end;

procedure NumberInGrid(Canvas: TCanvas; ARect: TRect; Value: Double; var CurrentZeroWidth: Integer; Alignment: TAlignment);
var
  nChar: Integer;
  buffer: string;
  ZeroSize: TSize;
begin
  if CurrentZeroWidth = 0 then begin
    {$IFDEF VCL_}
    GetTextExtentPoint(Canvas.Handle, AddressOfFirstChar('0123456789'), 10, ZeroSize);
    {$ELSE}
    ZeroSize := Canvas.TextExtent('0123456789');
    {$ENDIF}
    CurrentZeroWidth := ZeroSize.cx;
  end;
  nChar := (ARect.Right - ARect.Left - 4) * 10 div CurrentZeroWidth;
  buffer := DtosM(Value, nchar, -1);
  TextInGrid(Canvas, ARect, buffer, Alignment);
end;

procedure TNumberGrid.DefaultDrawCell(ACol, ARow: Longint; ARect: TRect; AState: TGridDrawState);
var
  buffer: string;
  v: Double;
  Drawn: Boolean;
begin
  Drawn := False;
  if Assigned(FOnDrawCellContents) then
    FOnDrawCellContents(Self,Acol,ARow,ARect,AState,Drawn);
  if Drawn then
    Exit;
  if (ARow = LastRow) and (ACol = LastCol) then begin
    ColouredTextInGrid(Canvas, ARect, LastString, clRed);
    Exit;
  end else if not GetCellLabel(ARow, ACol, buffer) then
    if GetCell(ARow, ACol, v) then begin
      NumberInGrid(Canvas, ARect, v, CurrentZeroWidth);
      Exit;
    end else
      buffer := '';
  TextInGrid(Canvas, ARect, buffer);
end;

function TNumberGrid.GetPointer(ARow, ACol:Longint):Pointer;
begin
  Result := FMatrix.Item[ARow-FixedRows+1, ACol-FixedCols+1];
end;

function TNumberGrid.CanChange(ARow, ACol:Longint):Boolean;
begin
  Result := (ARow >= FixedRows) and (ACol >= FixedCols);
end;

function TNumberGrid.GetCell(ARow, ACol:Longint; out v:Double):Boolean;
var
  p: ^Double;
begin
  p := Item[ARow, ACol];
  Result := not IsNanOrInf(p^);
  Move(p^, v, SizeOf(Double));  { So that specific Invalid values can be seen }
  if Assigned(FOnGetCell) then FOnGetCell(Self, ACol, ARow, v, Result);
end;

procedure TNumberGrid.SetCell(ARow, ACol:Longint; v:Double; IsValid:Boolean);
var
  p: ^Double;
  OldV: Double;
  OldIsValid, Change: Boolean;
begin
  if not CanChange(ARow, ACol) then Exit;
  p := Item[ARow, ACol];
  OldIsValid := not IsNanOrInf(p^);
  if OldIsValid then
    OldV := p^;
  if Assigned(FOnSetCell) then FOnSetCell(Self, ACol, ARow, v, IsValid);
  { Note: To Set a specific invalid value, set IsValid := True
    and v to the required value }
  if IsValid then
    Move(v, p^, SizeOf(Double))
  else
    SetNan(p^);
  IsValid := not IsNanOrInf(p^);  { Check for a specific invalid value }
  Change := OldIsValid <> IsValid;
  if (LastCol = ACol) and (LastRow = ARow) then begin
    Change := True; { Turn off error indication }
    LastCol := -1;
  end;
  if not Change and IsValid then
    Change := v <> OldV;
  if Change then begin
    Modified := True;
    Changed(ACol, ARow);
  end;
end;

function TNumberGrid.GetValid(ARow, ACol:Longint):Boolean;
var
  v: Double;
begin
  Result := GetCell(ARow, ACol, v);
end;

procedure TNumberGrid.SetValid(ARow, ACol:Longint; IsValid:Boolean);
begin
  SetCell(ARow, ACol, 0, IsValid);
end;

function TNumberGrid.GetValue(ARow, ACol:Longint):Double;
begin
  GetCell(ARow, ACol, Result);
end;

procedure TNumberGrid.SetValue(ARow, ACol:Longint; const v:Double);
begin;
  SetCell(ARow, ACol, v, True);
end;

procedure TNumberGrid.ExchangeRows(i, j: Longint);
var
  ACol: Longint;
  v1, v2: Double;
begin
  for ACol := 1 to ColCount - 1 do
    if GetCell(i, ACol, v1) then
      if GetCell(j, ACol, v2) then begin
        Value[i, ACol] := v2;
        Value[j, ACol] := v1;
      end else begin
        Value[j, ACol] := v1;
        Valid[i, ACol] := False;
      end
    else if GetCell(j, ACol, v2) then begin
      Value[i, ACol] := v2;
      Valid[j, ACol] := False;
    { else no action } end;
end;

procedure TNumberGrid.Invalidate;
begin
  { Fudge to prevent some excess Invalidates }
  if not InhibitNextInvalidate then
    if FUpdateCount = 0 then
      inherited Invalidate;
  InhibitNextInvalidate := False;
end;

procedure TNumberGrid.SetFocus;
begin
  { Windows/MDI bug fix? - be more forcefull in getting focus }
  inherited;
  {$IFDEF MSWINDOWS}
  if Windows.GetFocus <> Handle then
    Windows.SetFocus(Handle);
  {$ENDIF}
end;

procedure TNumberGrid.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell: TGridCoord;
  XFixed, YFixed:Boolean;
  NewSelection:TGridRect;
begin
  SavedOptions := [];
  InAreaSelection := asNone;
  Cell := MouseCoord(X, Y);
  if Button = mbLeft then
    LastCell := Cell;
  if (Cell.X <> Col) or (Cell.Y <> Row) then EditorMode := False;
  if (Button = mbLeft) and not EditorMode and (goRangeSelect in Options)
     and not (ssCtrl in Shift) and not (ssAlt in Shift) then begin
    if (Cell.X <> Col) or (Cell.Y <> Row) then begin
      XFixed := (Cell.X < FixedCols) and (Cell.X >= 0);
      YFixed := (Cell.Y < FixedRows) and (Cell.Y >= 0);
      if XFixed or YFixed then begin
        SetFocus; { Required in LCL }
        if XFixed then begin
          if not YFixed then begin
            if not (goRowMoving in Options)then begin
              InAreaSelection := asRow;
              NewSelection.Right := FixedCols;
              NewSelection.Left := ColCount-1;
              NewSelection.Top := Cell.Y;
              NewSelection.Bottom := Cell.Y;
            end else
              InAreaSelection := asNone;
          end else begin
            InAreaSelection := asAll;
            NewSelection := DataExtent;
            Cell := NewSelection.TopLeft;
            NewSelection.TopLeft := NewSelection.BottomRight;
            NewSelection.BottomRight := Cell;
          end;
        end else begin { YFixed }
          if not (goColMoving in Options) then begin
            NewSelection.Left := Cell.X;
            NewSelection.Right := Cell.X;
            NewSelection.Bottom := FixedRows;
            NewSelection.Top := DataExtent.Bottom;
            InAreaSelection := asCol;
          end else
            InAreaSelection := asNone;
        end;
        if InAreaSelection <> asNone then begin
          InvalidateCell(Col, Row);
          SelectingAnchor := NewSelection.BottomRight;
          Selection := NewSelection;
          InvalidateCell(Col, Row);
        end;
      end else begin
        {$IFNDEF LCL_} { LCL does this anyway using ExtendedSelect }
          InhibitNextInvalidate := goEditing in Options;
          SavedOptions := Options*[goEditing];
          Options := Options - [goEditing];	{ Allow Selection of cells by mouse drag }
        {$ENDIF}
{       if ssShift in Shift then begin
          NewSelection := Selection;
          NewSelection.BottomRight := SelectingAnchor;
          Selection := NewSelection;
        end; }
      end;
    end;
  end else if (Button = mbRight) and not EditorMode then begin
    { Right click should ensure selection is under mouse }
    NewSelection := Selection;
    { Click test should include fixed cols }
    if NewSelection.Left = FixedCols then
      NewSelection.Left := 0;
    if NewSelection.Top = FixedRows then
      NewSelection.Top := 0;
    if (Cell.Y < NewSelection.Top) or (Cell.Y > NewSelection.Bottom)
       or (Cell.X < NewSelection.Left) or (Cell.X > NewSelection.Right) then
      if Cell.X < FixedCols then
        if Cell.Y < FixedRows then
          FocusCell(FixedCols, FixedRows, True)
        else
          FocusCell(FixedCols, Cell.Y, True)
      else if Cell.Y < FixedRows then
        FocusCell(Cell.X, FixedRows, True)
      else
        FocusCell(Cell.X, Cell.Y, True)
  end;
  inherited MouseDown(Button, Shift, X, Y);
  if (InAreaSelection <> asNone) and ((FGridState = gsRowSizing) or (FGridState = gsColSizing)) then begin
    CancelSelection;
    InAreaSelection := asNone;
  end;
end;

procedure TNumberGrid.TrackSelectedArea(X, Y: Integer);
var
  Cell:TGridCoord;
  NewSelection:TGridRect;
begin
  Cell := MouseCoord(X, Y);
  NewSelection.BottomRight := SelectingAnchor;
  case InAreaSelection of
   	asCol:
      if (Cell.X >= FixedCols) and (Cell.X < ColCount) and (NewSelection.Right <> Cell.X) then begin
        NewSelection.Left := Cell.X;
        NewSelection.Top := Selection.Bottom;
        InvalidateCell(Col, Row);
        Selection := NewSelection;
      end;
    asRow:
      if (Cell.Y >= FixedRows) and (Cell.Y < RowCount) and (NewSelection.Bottom <> Cell.Y) then begin
        NewSelection.Top := Cell.Y;
        NewSelection.Left := Selection.Right;
        InvalidateCell(Col, Row);
        Selection := NewSelection;
      end;
  end;
end;

procedure TNumberGrid.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove(Shift, X, Y);
  if InAreaSelection <> asNone then TrackSelectedArea(X, Y);
end;

procedure TNumberGrid.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  Cell: TGridCoord;
  ChangeOK: Boolean;
begin
  inherited MouseUp(Button, Shift, X, Y);
  InhibitNextInvalidate := SavedOptions <> [];
  Options := Options + SavedOptions;
  SavedOptions := [];
  if InAreaSelection <> asNone then TrackSelectedArea(X, Y);
  InAreaSelection := asNone;
  if Button = mbLeft then begin
    Cell := MouseCoord(X,Y);
    if (Cell.X >= FixedCols) and (Cell.Y >= FixedRows) and  (Cell.X = LastCell.X) and (Cell.Y = LastCell.Y) and GetCellIsFlag(Cell.Y,Cell.X) then begin
      ChangeOK := True;
      if Assigned(FOnChangeCellQuery) then
        FOnChangeCellQuery(Self, Cell.X, Cell.Y, ChangeOK);
      if ChangeOK then
        Valid[Cell.Y,Cell.X] := not Valid[Cell.Y,Cell.X];
    end;
  end;
end;

procedure TNumberGrid.KeyPress(var Key: Char);
var
  V, ChangeOK: Boolean;
begin
  inherited KeyPress(Key);
  if GetCellIsFlag(Row,Col) then begin
    V := not Valid[Row,Col];
    if (Key = ' ') or TryStrToBoolean(Key,V) then begin
      ChangeOK := True;
      if Assigned(FOnChangeCellQuery) then
        FOnChangeCellQuery(Self, Col, Row, ChangeOK);
      if ChangeOK then
        Valid[Row,Col] := V;
      Key := #0;
    end;
  end;
end;

procedure TNumberGrid.DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
begin
  if Assigned (OnGetPickListItems) then
    OnGetPickListItems(Col, Row, AItems);
end;

function TNumberGrid.DataExtent:TGridRect;
var
  r,c:Longint;
  s: string;
begin
  Result.Right := FixedCols;
  Result.Left := ColCount-1;
  Result.Bottom := FixedRows;
  Result.Top := RowCount-1;
  for r := FixedRows to RowCount-1 do
    for c := FixedCols to ColCount-1 do begin
      if not GetValid(r,c) then
        { Not a number - check for non null label }
        if not GetCellLabel(r,c,s) or (s = '') then
          Continue;
      if Result.Right < c then Result.Right := c;
      if Result.Left > c then Result.Left := c;
      if Result.Top > r then Result.Top := r;
      if Result.Bottom < r then Result.Bottom := r;
    end;
end;

procedure TNumberGrid.SelectAll;
begin
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).SelectAll
  else begin
    InvalidateCell(Col, Row);
    Selection := DataExtent;
    InvalidateCell(Col, Row);
  end;
end;

procedure TNumberGrid.ClearArea(Area:TGridRect);
var
  r,c:Longint;
begin
  EditorMode := False;
  for r := Area.Top to Area.Bottom do
    for c := Area.Left to Area.Right do
      Valid[r,c] := False;
end;

procedure TNumberGrid.ClearSelection;
var
  Writeable: Boolean;
  r,c:Longint;
begin
  if not (goEditing in Options) then Exit;
  { Simulate DEL key }
  if EditorMode and (InplaceEditor is TCustomEdit) then
    with InplaceEditor as TCustomEdit do begin
      if SelLength <= 0 then
        SelLength := 1;
      ClearSelection;
    end
  else
    for r := Selection.Top to Selection.Bottom do
      for c := Selection.Left to Selection.Right do begin
        Writeable := True;
        if Assigned(FOnChangeCellQuery) then
          FOnChangeCellQuery(Self, c, r, Writeable);
        if Writeable then
          Valid[r,c] := False;
      end;
end;

procedure TNumberGrid.ClearColumn(ACol: Longint);
begin
  EditorMode := False;
  {$IFDEF MAKEFROMCOLVECTS}
  FMatrix.Column[ACol].Clear;
  {$ELSE}
  FMatrix.ClearRange(FMatrix.Index(1,ACol), FMatrix.RowCount);
  {$ENDIF}
  Changed(ACol, -1);
end;

procedure TNumberGrid.Clear;
{$IFDEF MAKEFROMCOLVECTS}
var
  c:Longint;
{$ENDIF}
begin
  EditorMode := False;
{$IFDEF MAKEFROMCOLVECTS}
  for c := 1 to ColCount-1 do
    FMatrix.Column[c].Clear;
{$ELSE}
  FMatrix.Clear;
{$ENDIF}
  Modified := False;
  Changed(-1, -1);
end;

procedure TNumberGrid.Blank;
begin
  RowCount := 2;
  ColCount := 2;
  Title[1] := ' ';
  Lbl[1] := ' ';
  Valid[1,1] := False;
  Caption := '';
end;

procedure TNumberGrid.SaveAreaToStream(Stream:TStream; Area:TGridRect; WithLabels:Boolean);
var
  s:TTextStream;
  r,c:Longint;
  v: Double;
  buffer: string;
begin
  if (Area.Left <= FixedCols) and WithLabels and (Area.Left <> Area.Right) then
    Area.Left := 0;
  if (Area.Top <= FixedRows) and WithLabels and (Area.Top <> Area.Bottom) then
    Area.Top := 0;
  s := nil;
  try
    s := TTextStream.Create(Stream);
    for r := Area.Top to Area.Bottom do begin
      for c := Area.Left to Area.Right do begin
        if GetCellLabel(r, c, buffer) then
          s.PutString(buffer)
        else if GetCell(r, c, v) then
          s.PutFloat(v);
        if c <> Area.Right then s.Tab;
      end;
      s.Newline;
    end;
  finally
    s.Free;
  end;
end;

procedure TNumberGrid.SaveToStream(Stream:TStream);
begin
  SaveAreaToStream(Stream, DataExtent);
end;

procedure TNumberGrid.OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint; IncludeNonNumeric: Boolean);
var
  s:TTextStream;
  r,c:Longint;
  Writeable: Boolean;
  V: Double;
  nValues: Integer;
begin
  s := nil;
  try
    BeginUpdate;
    s := TTextStream.Create(Stream);
    s.NULLisEOF := True;
    s.ItemSeparators := [isTab,isComma];
    s.NoParseQuotes := IncludeNonNumeric; { With this setting, copy and paste of strings containing quotes between grids works }
    r := StartRow;
    repeat
      S.GetFreeItem;
      if (S.AtEndOfFile and (S.Token = '')) then
        Continue;
      if not IncludeNonNumeric and not S.IsRealTokenOrBlank(V) then
        Continue; { Skip lines without numbers }
      nValues := 0;
      for c := StartCol to ColCount-1 do begin
        Writeable := True;
        if Assigned(FOnChangeCellQuery) then
          FOnChangeCellQuery(Self, c, r, Writeable);
        if Writeable then begin
          if r >= RowCount then
            RowCount := RowCount * 2;
          if IncludeNonNumeric then
            SetEditText(c, r, S.Token)
          else
            Value[r,c] := V;
        end;
        if IncludeNonNumeric or not IsNan(V) then
          Inc(nValues);
        if S.AtEndOfLine then
          Break; { End of data on line }
        S.GetFreeItem;
        if not IncludeNonNumeric and not S.IsRealTokenOrBlank(V) then
          Continue; { end of data on line }
      end;
      if IncludeNonNumeric or (nValues > 0) then
        Inc(r); { Skip lines with no numbers on }
    until S.NextLine;
  finally
    EndUpdate;
    s.Free;
  end;
end;

procedure TNumberGrid.LoadFromStream(Stream:TStream);
begin
  OffsetLoadFromStream(Stream, 1, 1, AllowText);
end;

function TNumberGrid.CanCopyToClipboard: Boolean;
begin
  if EditorMode then
    Result := (InplaceEditor <> nil) and (InplaceEditor is TCustomEdit) and (TCustomEdit(InplaceEditor).SelLength > 0)
  else
    Result := True; { Always one cell selected }
end;

procedure TNumberGrid.CopyToClipboard(WithLabels:Boolean);
var
  s:THugeMemoryStream;
begin
  if EditorMode and InplaceEditor.Visible and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).CopyToClipboard
  else begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      SaveAreaToStream(s, Selection, WithLabels or CopyLabels);
      s.CopyToClipboard(CF_TEXT);
    finally
      s.Free;
    end;
  end;
end;

procedure TNumberGrid.CopyToClipboard;
begin
  CopyToClipboard(False);
end;

function TNumberGrid.CanPasteFromClipboard: Boolean;
begin
  Result := (goEditing in Options) and ClipboardHasText;
end;

procedure TNumberGrid.PasteFromClipboard;
var
  s: THugeMemoryStream;
  r, c: Longint;
  function MoreThanOneWord: Boolean;
  var
    p: PAnsiChar;
  begin
    p := s.Memory;
    while (p^ <> #0) and (p^ <= ' ') do Inc(p);
    while (p^ <> #0) and (p^ > ' ') and (p <> ',') do Inc(p);
    while (p^ <> #0) and (p^ <= ' ') do Inc(p);
    Result := p^ > ' ';
  end;
begin
  if ClipboardHasText then begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      s.SetFromClipboard(CF_TEXT);
      { As a heuristic, assume anything with more than one word
        should be pasted into grid as a whole }
      if not EditorMode or MoreThanOneWord then begin
        EditorMode := False;
        r := Min(Selection.Top, Selection.Bottom);
        c := Min(Selection.Left, Selection.Right);
        OffsetLoadFromStream(s, r, c, AllowText);
        Exit;
      end;
    finally
      s.Free;
    end;
  end;
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).PasteFromClipboard
end;

procedure TNumberGrid.MouseToCell(X, Y: Integer; var ACol, ARow: Longint);
var
  Coord: TGridCoord;
begin
  Coord := MouseCoord(X, Y);
  ACol := Coord.X;
  ARow := Coord.Y;
end;

function TNumberGrid.NaturalWidth: Integer;
begin
  Result := NGrid.NaturalWidth(Self);
end;

type
  TOpenGrid = class(TCustomGrid);

function NaturalWidth(Grid: TCustomGrid): Integer;
var
  i: Longint;
begin
  with TOpenGrid(Grid) do begin
    {$IFNDEF CLX_}
    if {$IFDEF LCLcarbonOFF}True or{$ENDIF} (VisibleRowCount < RowCount-1) then
      Result := GetSystemMetrics(SM_CXVSCROLL)
    {$ELSE}
    if VScrollBar.Visible then
      Result := VScrollBar.Width
    {$ENDIF}
    else
      Result := 0;
    {$IFNDEF LCL_}
      if BorderStyle = bsSingle then
          Inc(Result, 4);
      {$IFDEF UNIX}
      Inc(Result, -2);
      {$ENDIF}
    {$ELSE}
      Dec(Result, (ColCount-5)*GridLineWidth);
    {$ENDIF}
    for i := 0 to ColCount - 1 do
        Inc(Result, ColWidths[i]+GridLineWidth);
  end;
end;

end.

