unit Prevwdlg;

interface
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Messages, Graphics, Controls,
                 Forms, Dialogs, ExtCtrls, StdCtrls, Buttons, Printers,
  {$ELSE}        QGraphics, QControls, QForms, QDialogs, QExtCtrls, QStdCtrls,
                 QButtons, QPrinters, Types, {$ENDIF}
  {$IFDEF LCL_} PrintersDlgs, {$ENDIF}
  Handles, Classes, SysUtils, NewSpin, AlEdit, EditExtr, ExpCan;

type

  TPrintPreviewOption = (ppShowPrint, ppShowPrintSetup, ppShowLandPort, ppShowPages, ppShowCentre,
                         ppShowDefault, ppShowZoom, ppShowMaximize, ppShowClose,
                         ppAreaIsFirstPage);
  TPrintPreviewOptions = set of TPrintPreviewOption;
  { Note usage of ppAreaIsFirstPage:
    If set, then Pages is the number of pages to show in the preview,
    and MaxPages is the number of pages to print by default.
    If ppAreaIsFirstPage is clear, then the document is stretched over Pages pages, and
    MaxPages is the maximum number of pages the document can be stretched over. }

  TShowPreviewEvent = procedure (Sender:TObject; Canvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                                 APage: Integer; OnPrinter:Boolean) of object;

  TLongPoint = record
    x: LongInt;
    y: LongInt;
  end;

  { TPrintPreviewForm }

  TPrintPreviewForm = class(TForm)
    SpeedBar: TPanel;
    CloseBtn: TBitBtn;
    PrinterSetupBtn: TBitBtn;
    StatusBar: TPanel;
    Page: TPaintBox;       
    PageImage: TPaintBox;
    SpinPages: TSpinBox;
    PagesLabel: TLabel;
    PrintBtn: TBitBtn;
    Centre: TBitBtn;
    Default: TBitBtn;
    PageWidth: TFixedSpin;
    Label2: TLabel;
    Label3: TLabel;
    PageHeight: TFixedSpin;
    Label4: TLabel;
    PageLeft: TFixedSpin;
    Label5: TLabel;
    Label6: TLabel;
    PageTop: TFixedSpin;
    StretchHandle: TStretchHandle;
    LandPortBtn: TBitBtn;
    Zoom: TBitBtn;
    ScrollBox: TScrollBox;
    Maximize: TBitBtn;
    Settings: TBitBtn;
    ShrinkBtn: TBitBtn;
    procedure FormShow(Sender: TObject);
    procedure PageImagePaint(Sender: TObject);
    procedure PagePaint(Sender: TObject);
    procedure CloseBtnClick(Sender: TObject);
    procedure PrinterSetupClick(Sender: TObject);
    procedure StretchHandleResized(Sender: TObject);
    procedure SpinPagesChange(Sender: TObject);
    procedure StretchHandleMoved(Sender: TObject);
    procedure PrintClick(Sender: TObject);
    procedure CentreClick(Sender: TObject);
    procedure DefaultClick(Sender: TObject);
    procedure WidthChange(Sender: TObject);
    procedure LeftChange(Sender: TObject);
    procedure TopChange(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure LandPortBtnClick(Sender: TObject);
    procedure ZoomClick(Sender: TObject);
    procedure MaximizeClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure SettingsClick(Sender: TObject);
    procedure ShrinkBtnClick(Sender: TObject);
  private
    FXSize, FYSize,                        { Page size in mm }
    FXMargin, FYMargin: Double;            { Printing Offset in mm; set very negative to centre }
    DeviceXSize, DeviceYSize,              { as above, but in device units, including margins }
    DeviceXMargin, DeviceYMargin: Longint;
    FOrientation: TPrinterOrientation;
    FPages: Integer;
    FDefaultDrawing:Boolean;
    FOnShowPreview: TShowPreviewEvent;
    FOnChange: TNotifyEvent;
    PixPerInch:TPoint;      { Device units per inch }
    { Printer Information (All in device units) }
    PaperSize: TPoint;          { Size of 1 page }
    TotalPaperSize: TLongPoint; { Size of all pages }
    TopLeft:TPoint;
    PrintableAreaSize:TPoint;
    PreviewScale:Double;    { Screen Pixels per printer inch }
    PhysicalPrinterDetails: string;
    LockoutDoChange: Boolean;
    FAspectRatio: Double;
    FOptions: TPrintPreviewOptions;
    FMaxPages: Integer;
    FDefaultFraction: Double;
    FZoomFactor: Double;
    FOnSettingsClick: TNotifyEvent;
    {$IFNDEF CLX_}
    FPrinterSetupDialog: TPrinterSetupDialog;
    FPrintDialog: TPrintDialog;
    {$ENDIF}
    {$IFDEF MSWINDOWS}
    procedure WMEraseBkgnd(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    {$ENDIF}
    { Convert Printer device units to inches }
    function Xinch(x:Double):Double;
    function Yinch(y:Double):Double;
    { Convert Printer device units to preview screen units }
    function Xscale(x:Double):Longint;
    function Yscale(y:Double):Longint;
    procedure UpdatePageSize(p:TPrinter; UpdateControls: Boolean);
    procedure SetXSize(v:Double);
    procedure SetYSize(v:Double);
    procedure SetXMargin(v:Double);
    procedure SetYMargin(v:Double);
    procedure SetOnShowPreview(AOnShowPreview:TShowPreviewEvent);
    procedure SetPages(v:Integer);
    procedure SetOrientation(AOrientation: TPrinterOrientation);
    procedure SetAspectRatio(const Value: Double);
    procedure SetOptions(const Value: TPrintPreviewOptions);
    procedure PositionButtons;
    procedure SetMaxPages(const Value: Integer);
    procedure SetLandPortBtn;
    procedure ProcessPreviewRect(ACanvas: TCanvas; APage, OffsetX, OffsetY: Integer; r:TRect);
    procedure SetZoomFactor(const Value: Double);
    procedure GetPrinterInfo(Printer: TPrinter);
    procedure SetOnSettingsClick(const Value: TNotifyEvent);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double);
    procedure Changed;
    procedure Resize; override;
    procedure DoPrint;
    procedure Print;
    Procedure PrinterSetup;
  published
    { Size of image in mm }
    property XSize:Double read FXSize write SetXSize;
    property YSize:Double read FYSize write SetYSize;
    { Position of image in mm; set to -2*LargeMargin to centre;
      set to  2*LargeMargin to centre at one end if image is less than 1/2 full page }
    property XMargin:Double read FXMargin write SetXMargin;
    property YMargin:Double read FYMargin write SetYMargin;
    property Orientation: TPrinterOrientation read FOrientation write SetOrientation;
    property DefaultDrawing:Boolean read FDefaultDrawing write FDefaultDrawing;
    property OnShowPreview:TShowPreviewEvent read FOnShowPreview write SetOnShowPreview;
    property OnChange:TNotifyEvent read FOnChange write FOnChange;
    property Pages:Integer read FPages write SetPages default 1;
    property MaxPages:Integer read FMaxPages write SetMaxPages default 99;
    { Width / Height of image }
    property AspectRatio: Double read FAspectRatio write SetAspectRatio;
    property Options: TPrintPreviewOptions read FOptions write SetOptions;
    property DefaultFraction: Double read FDefaultFraction write FDefaultFraction;
    property ZoomFactor: Double read FZoomFactor write SetZoomFactor;
    property OnSettingsClick: TNotifyEvent read FOnSettingsClick write SetOnSettingsClick;
  end;

var
  PrintPreviewForm: TPrintPreviewForm;

const
  LargeMargin = 1000000.0;

implementation

uses {$IFDEF LCL_} ExportG, LCLType, LCLProc, {$ENDIF}
     CanCan;

{$IFNDEF LCL_}
{$R *.DFM}
{$ELSE}
{$R *.lfm}
{$ENDIF}

 const
   mmPerInch = 25.4;

constructor TPrintPreviewForm.Create(AOwner: TComponent);
begin
  FOptions := [ppShowPrint, ppShowPrintSetup, ppShowLandPort, ppShowPages,
               ppShowCentre, ppShowDefault, ppShowZoom, ppShowClose];
  FPages := 1;
  FMaxPages := 99;
  FDefaultFraction := 0.9;
  FZoomFactor := 1;
  LockoutDoChange := True;
  DefaultClick(Self);
  inherited Create(AOwner);
  LockoutDoChange := False;
  FDefaultDrawing := False;
  PositionButtons;
  SetLandPortBtn;
end;

destructor TPrintPreviewForm.Destroy;
begin
  if PrintPreviewForm = self then
    PrintPreviewForm := nil;
  inherited;
end;

procedure TPrintPreviewForm.SetPages(v:Integer);
begin
  if (v <> FPages) and (v > 0) and (v <= MaxPages) then begin
    FPages := v;
    SpinPages.Value := v;
    Changed;
  end;
end;

procedure TPrintPreviewForm.SetOrientation(AOrientation: TPrinterOrientation);
begin
  if FOrientation <> AOrientation then begin
    FOrientation := AOrientation;
    SetLandPortBtn;
    Changed;
  end;
end;

procedure TPrintPreviewForm.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
  UpdatePageSize(Printer,True);
  if HandleAllocated then Invalidate;
end;

procedure TPrintPreviewForm.Resize;
begin
  inherited Resize;
  UpdatePageSize(Printer,False);
end;

procedure TPrintPreviewForm.SetMargins(AXSize, AYSize, AXMargin, AYMargin:Double);
begin
  GetPrinterInfo(Printer);
  if AXSize < 0 then AXSize := DefaultFraction*Xinch(PrintableAreaSize.x)*mmPerInch;
  if AYSize < 0 then AYSize := DefaultFraction*Yinch(PrintableAreaSize.y)*mmPerInch;
  if AspectRatio > 0 then
    { Shrink to ensure correct aspect ratio }
    if AspectRatio * AYSize > AXSize then
      AYSize := AXSize / AspectRatio
    else
      AXSize := AYSize * AspectRatio;
  if (AXSize <> FXSize) or (AYSize <> FYSize) or (AXMargin <> FXMargin) or (AYMargin <> FYMargin) then begin
    FXSize := AXSize;
    FYSize := AYSize;
    FXMargin := AXMargin;
    FYMargin := AYMargin;
    Changed;
  end;
end;

procedure TPrintPreviewForm.SetXSize(v:Double);
begin
  SetMargins(v, YSize, XMargin, YMargin);
end;

procedure TPrintPreviewForm.SetYSize(v:Double);
begin
  SetMargins(XSize, v, XMargin, YMargin);
end;

procedure TPrintPreviewForm.SetXMargin(v:Double);
begin
  SetMargins(XSize, YSize, v, YMargin);
end;

procedure TPrintPreviewForm.SetYMargin(v:Double);
begin
  SetMargins(XSize, YSize, XMargin, v);
end;

function TPrintPreviewForm.Xinch(x:Double):Double;
begin
  Result := x / PixPerInch.x;
end;

function TPrintPreviewForm.Yinch(y:Double):Double;
begin
  Result := y / PixPerInch.y;
end;

function TPrintPreviewForm.Xscale(x:Double):Longint;
begin
  Result := Round(PreviewScale * x / PixPerInch.x);
end;

function TPrintPreviewForm.Yscale(y:Double):Longint;
begin
  Result := Round(PreviewScale * y / PixPerInch.y);
end;

procedure TPrintPreviewForm.SetOnShowPreview(AOnShowPreview:TShowPreviewEvent);
begin
  FOnShowPreview := AOnShowPreview;
  Invalidate;
end;

function TotalImageLength(Centre:Boolean; Length, Margin:LongInt; Page, Printable, Offset:Integer):LongInt;
{ All Parameters in device units }
var
  p: LongInt;
begin
  if not Centre then begin
    p := (Margin - Offset) mod Page;
    if p < 0 then p := p + Page;
    { Position of start on first non-blank page }
    Result := Page - p;
    if p < Printable then
      Length := Length - (Printable-p);
    if Length > 0 then begin
      p := Length div Printable;
      Result := Result + p*Page;
      Length := Length - p*Printable;
      if Length = 0 then
        Result := Result - (Page-Printable)    { Exactly on page boundary }
      else
        Result := Result + Length;
    end else
      Result := Length + Printable - p;
    { At this point, result includes region to end of last page beyond margins }
  end else
    { Centre Image }
    Result := 2 * TotalImageLength(False, Length div 2, 0, Page, Printable, Offset);
end;

procedure TPrintPreviewForm.GetPrinterInfo(Printer: TPrinter);
{$IFDEF CLX_}
var
  sz: TSize;
{$ENDIF}
{$IFDEF LCL_}
var
  PaperRect: TPaperRect;
  procedure SwapI(var a, b: Longint);
  var
    c: Longint;
  begin
    c := a;
    a := b;
    b := c;
  end;
{$ENDIF}
begin
  Printer.Orientation := Orientation;
  {$IFDEF VCL_}
  { TODO: This under WINLCL }
  (* Printer Information *)
  PixPerInch.x := GetDeviceCaps(Printer.Handle, LOGPIXELSX);
  PixPerInch.y := GetDeviceCaps(Printer.Handle, LOGPIXELSY);
  (* Page Info (All in device units) *)
  Escape(Printer.Handle, GETPHYSPAGESIZE, 0, Nil, @PaperSize);
  Escape(Printer.Handle, GETPRINTINGOFFSET, 0, Nil, @TopLeft);
  PrintableAreaSize.x :=  GetDeviceCaps(Printer.Handle, HORZRES);
  PrintableAreaSize.y :=  GetDeviceCaps(Printer.Handle, VERTRES);
  {$ENDIF}
  {$IFDEF CLX_}
  if Printer.PrintAdapter is TQPrintAdapter then
    TQPrintAdapter(Printer.PrintAdapter).FullPage := False;
  (* Printer Information *)
  PixPerInch.x := Printer.XDPI;
  PixPerInch.y := Printer.YDPI;
  sz := Printer.Margins;
  TopLeft.X := sz.cx;
  TopLeft.Y := sz.cy;
  {$ENDIF}
  {$IFDEF LCL_}
  PixPerInch.x := Printer.XDPI;
  PixPerInch.y := Printer.YDPI;
  PaperRect := Printer.PaperSize.PaperRect;
  PaperSize := PaperRect.PhysicalRect.BottomRight;
  TopLeft := PaperRect.WorkRect.TopLeft;
  with PaperRect.WorkRect do
    PrintableAreaSize := Point(Right-Left,Bottom-Top);
  if Printer.Orientation <> poLandscape then begin
    SwapI(PaperSize.x,PaperSize.y);
    SwapI(TopLeft.X,TopLeft.Y);
    SwapI(PrintableAreaSize.X,PrintableAreaSize.Y);
  end;
  {$ENDIF}
end;

procedure TPrintPreviewForm.UpdatePageSize(p:TPrinter; UpdateControls: Boolean);
var
  V, H, t, NewHeight, NewWidth, NewLeft, NewTop: Integer;
  PapSiz: Longint;
  CentreX, CentreY: Boolean;
  function Min(a, b:Double):Double;
  begin
    if a < b then
      Result := a
    else
      Result := b;
  end;
  function Cen(a:Integer): Integer;
  begin
    if a < 0 then
      Result := StatusBar.Height
    else
      Result := a div 2;
  end;
begin
  GetPrinterInfo(p);

  TotalPaperSize.x := PaperSize.x;
  TotalPaperSize.y := PaperSize.y;
  if Orientation = poLandscape then
    TotalPaperSize.x := TotalPaperSize.x * Pages
  else
    TotalPaperSize.y := TotalPaperSize.y * Pages;
  if not HandleAllocated then Exit;
  (* Determine scale factor for screen image *)
  V := ScrollBox.ClientHeight - StatusBar.Height; { - 2*StatusBar.Height - SpeedBar.Height; }
  H := ScrollBox.ClientWidth - StatusBar.Height;
  if V <= 0 then V := 1;
  if H <= 0 then H := 1;
  PreviewScale := ZoomFactor * Min(V/YInch(TotalPaperSize.y), H/XInch(TotalPaperSize.x));

  NewHeight := Yscale(TotalPaperSize.y);
  NewWidth := Xscale(TotalPaperSize.x);
  ScrollBox.AutoScroll := False;
  Page.SetBounds(Cen(ScrollBox.ClientWidth - NewWidth), { Left }
                 Cen(ScrollBox.ClientHeight - NewHeight), { Top }
                 NewWidth, NewHeight);
  ScrollBox.AutoScroll := True;

  if (StretchHandle.ChildCount > 1) or (StretchHandle.Children[0] <> PageImage) then
    StretchHandle.Detach;

  DeviceXSize := Round(XSize/mmPerInch*PixPerInch.x);
  DeviceYSize := Round(YSize/mmPerInch*PixPerInch.y);
  CentreX := Abs(XMargin) >= LargeMargin;
  CentreY := Abs(YMargin) >= LargeMargin;
  if not CentreX then DeviceXMargin := Round(XMargin/mmPerInch*PixPerInch.x);
  if not CentreY then DeviceYMargin := Round(YMargin/mmPerInch*PixPerInch.y);
  if not (ppAreaIsFirstPage in Options) and (Pages > 1) then
    if Orientation = poLandscape then
      DeviceXSize := TotalImageLength(CentreX, DeviceXSize, DeviceXMargin, PaperSize.x, PrintableAreaSize.x, TopLeft.x)
    else
      DeviceYSize := TotalImageLength(CentreY, DeviceYSize, DeviceYMargin, PaperSize.y, PrintableAreaSize.y, TopLeft.y);
  { Centre, or shift to one end if less than 1/2 full page }
  PapSiz := TotalPaperSize.x;
  if (Orientation = poLandscape) and (ppAreaIsFirstPage in Options) then
    PapSiz := PaperSize.x;
  if CentreX then begin
    if XMargin >= LargeMargin then begin
      t := Trunc(PapSiz/DeviceXSize);
      DeviceXMargin := (PapSiz - t*DeviceXSize) div (t+1) ;
    end else
      DeviceXMargin := (PapSiz - DeviceXSize) div 2 ;
    if ((Pages = 1) or (ppAreaIsFirstPage in Options))
       and (DeviceXSize <= PrintableAreaSize.x) then begin
      if DeviceXMargin < TopLeft.x then
        DeviceXMargin := TopLeft.x;
      if DeviceXMargin + DeviceXSize > TopLeft.x + PrintableAreaSize.x then
        DeviceXMargin := TopLeft.x + PrintableAreaSize.x - DeviceXSize;
    end;
  end;
  PapSiz := TotalPaperSize.y;
  if (Orientation <> poLandscape) and (ppAreaIsFirstPage in Options) then
    PapSiz := PaperSize.y;
  if CentreY then begin
    if YMargin >= LargeMargin then begin
      t := Trunc(PapSiz/DeviceYSize);
      DeviceYMargin := (PapSiz - t*DeviceYSize) div (t+1) ;
    end else
      DeviceYMargin := (PapSiz - DeviceYSize) div 2 ;
    if ((Pages = 1) or (ppAreaIsFirstPage in Options))
       and (DeviceYSize <= PrintableAreaSize.y) then begin
      if DeviceYMargin < TopLeft.y then
        DeviceYMargin := TopLeft.y;
      if DeviceYMargin + DeviceYSize > TopLeft.y + PrintableAreaSize.y then
        DeviceYMargin := TopLeft.y + PrintableAreaSize.y - DeviceYSize;
    end;
  end;
  NewWidth := XScale(DeviceXSize);
  NewHeight := YScale(DeviceYSize);
  NewLeft := Page.Left + XScale(DeviceXMargin);
  NewTop := Page.Top + YScale(DeviceYMargin);
  { Ifdef required by SETPARENT define in Handles.pas }
  if (NewLeft <> PageImage.Left{$IFDEF LCLcarbon}+StretchHandle.Left{$ENDIF}) or (NewWidth <> PageImage.Width)
     or (NewTop <> PageImage.Top{$IFDEF LCLcarbon}+StretchHandle.Top{$ENDIF}) or (NewHeight <> PageImage.Height)
     or (StretchHandle.ChildCount = 0) then begin
    { To avoid autosize loop, only Detach/Attach if anything changed }
    StretchHandle.Detach;
    PageImage.SetBounds(NewLeft, NewTop, NewWidth, NewHeight);
  end;
  if StretchHandle.ChildCount = 0 then
    StretchHandle.Attach(PageImage);
  SpeedBar.BringToFront;  { Ensure PageImage and handles do not obscure the buttons }
  StatusBar.BringToFront;
  StretchHandle.TabStop := True;

  PhysicalPrinterDetails := Format(
  'Printable Width %f mm out of %f mm, Height %f mm out of %f mm, Margins: Left %f mm, Top %f mm',
                              [Xinch(PrintableAreaSize.x)*mmperInch, Xinch(PaperSize.x)*mmperInch,
                               Yinch(PrintableAreaSize.y)*mmperInch, Yinch(PaperSize.y)*mmperInch,
                               Xinch(TopLeft.x)*mmperInch, Yinch(TopLeft.y)*mmperInch]);

  if UpdateControls then begin
    LockoutDoChange := True;
    try
      PageWidth.Value := XSize;
      PageHeight.Value := YSize;
      PageLeft.Value := DeviceXMargin/PixPerInch.x*mmPerInch;
      PageTop.Value := DeviceYMargin/PixPerInch.y*mmPerInch;
    finally
      LockoutDoChange := False;
    end;
  end;
  Page.Invalidate;
end;

function NextRect(var l, r:Integer; var Pos, Len, Off:Longint; Length, Margin:LongInt;
                  Page, Printable, Offset:Integer; Scale:Double):Boolean;
var
  p, w: Longint;
  function WScale(v:Longint):Integer;
  begin
    Result := Round(v*Scale);
  end;
begin
  if Len < 0 then begin
    Len := Length;
    Pos := Margin;
  end;
  p := (Pos - Offset) mod Page;   { Position within page }
  if p < 0 then
    p := p + Page;
  if p >= Printable then begin
    Pos := Pos + Page - p;
    p := 0;
  end;
  l := WScale(Pos-Margin);
  w := Printable - p;
  if w > Len then w := Len;
  r := WScale(Pos-Margin+w);
  Off := WScale(Pos-Margin-(Length-Len));
  Len := Len - w;
  Pos := Pos + w;
  Result := w > 0;
end;

procedure TPrintPreviewForm.ProcessPreviewRect(ACanvas: TCanvas; APage, OffsetX, OffsetY: Integer; r:TRect);
var
{$IFDEF MSWINDOWS}
  s: Integer;
{  WindowOrg: TPoint; }
{$ENDIF}
  TheCanvas: TExportableCanvas;
begin
  TheCanvas := nil;
  if DefaultDrawing then
    ACanvas.FillRect(r);
  if Assigned(FOnShowPreview) then begin
    {$IFDEF MSWINDOWS}
    s := SaveDC(ACanvas.Handle);
    {$ENDIF}
    try
      {$IFDEF MSWINDOWS}
      IntersectClipRect(ACanvas.Handle, r.Left, r.Top, r.Right, r.Bottom);
      {$ENDIF}
      {$IFDEF UNIX}
      { Omit unless really needed:
          Canvas.SetClipRect(r);
        as r needs to be offset by co-ordinate transform in painter }
      {$ENDIF}
      {$IFDEF NOLONGER}
          {$IFDEF WIN32}
          { In Win32 Delphi seems to offset the window origin rather than the viewport origin.
            Offsetting the viewport is essential for TGraph to work, as it sets the origin }
          SetWindowOrgEx(ACanvas.Handle, 0, 0, @WindowOrg);
          OffsetViewPortOrgEx(ACanvas.Handle, OffsetX-WindowOrg.x{+PageImage.Left}, OffsetY-WindowOrg.y{+PageImage.Top},
                              PChar(nil)^);
          {$ELSE}
          OffsetViewPortOrgEx(ACanvas.Handle, OffsetX, OffsetY, nil);
          {$ENDIF}
      {$ENDIF}
      TheCanvas := TCanvasCanvas.Create(ACanvas);
      FOnShowPreview(Self, TheCanvas, OffsetX, OffsetY, PageImage.ClientWidth, PageImage.ClientHeight, APage, False);
    finally
      FreeAndNil(TheCanvas);
      {$IFDEF MSWINDOWS}
      RestoreDC(ACanvas.Handle, s);
      ACanvas.Refresh;
      {$ENDIF}
      {$IFDEF UNIX}
      { Omit: Canvas.ResetClipRegion; see above }
      {$ENDIF}
    end;
  end;
end;

procedure TPrintPreviewForm.PageImagePaint(Sender: TObject);
var
  r: TRect;
  Pos, Len, OffsetX, OffsetY, CPage: Longint;
  More: Boolean;
begin
  {$IFDEF CLX_}
  PageImage.Canvas.Start;
  try
  {$ENDIF}
  r := PageImage.Canvas.ClipRect;
  OffsetX := 0;
  OffsetY := 0;
  CPage := 1;
  if (Pages > 1) and not (ppAreaIsFirstPage in Options) then begin
    Len := -1;
    repeat
      if Orientation = poLandscape then
        More := NextRect(r.Left, r.Right, Pos, Len, OffsetX, DeviceXSize, DeviceXMargin,
                         PaperSize.x, PrintableAreaSize.x, TopLeft.x, PreviewScale / PixPerInch.x)
      else
        More := NextRect(r.Top, r.Bottom, Pos, Len, OffsetY, DeviceYSize, DeviceYMargin,
                         PaperSize.y, PrintableAreaSize.y, TopLeft.y, PreviewScale / PixPerInch.y);
      if More then ProcessPreviewRect(PageImage.Canvas, CPage, OffsetX, OffsetY, r);
    until not More;
  end else
    ProcessPreviewRect(PageImage.Canvas, CPage, OffsetX, OffsetY, r);
  {$IFDEF CLX_}
  finally
    PageImage.Canvas.Stop;
  end;
  {$ENDIF}
end;

procedure TPrintPreviewForm.FormShow(Sender: TObject);
begin
  UpdatePageSize(Printer,True);
end;

procedure TPrintPreviewForm.PagePaint(Sender: TObject);
var
  r, Clip: TRect;
  i, x, y: Integer;
begin
  with Page.Canvas do
   try
    {$IFDEF CLX_}
    Start;
    {$ENDIF}
    Rectangle(0, 0, Page.Width, Page.Height);
    r.Left := Xscale(TopLeft.x);
    r.Right := Xscale(PrintableAreaSize.x) + Xscale(TopLeft.x);
    r.Top := Yscale(TopLeft.y);
    r.Bottom := Yscale(PrintableAreaSize.y) + Yscale(TopLeft.y);
    x := Xscale(PaperSize.x);
    y := Yscale(PaperSize.y);
    for i := 0 to Pages-1 do begin
      if (i > 0) and (ppAreaIsFirstPage in Options) then begin
        Clip.Left := Xscale(DeviceXMargin-TopLeft.x) + r.Left;
        Clip.Right := Clip.Left + Xscale(DeviceXSize);
        Clip.Top := Yscale(DeviceYMargin-TopLeft.y) + r.Top;
        Clip.Bottom := Clip.Top + YScale(DeviceYSize);
        ProcessPreviewRect(Page.Canvas, i+1, Clip.Left, Clip.Top, Clip)
      end;
      Pen.Style := psDot;
      MoveTo(0, r.Top);
      LineTo(Page.ClientWidth, r.Top);
      MoveTo(0, r.Bottom);
      LineTo(Page.ClientWidth, r.Bottom);
      MoveTo(r.Left, 0);
      LineTo(r.Left, Page.ClientHeight);
      MoveTo(r.Right, 0);
      LineTo(r.Right, Page.ClientHeight);
      Pen.Style := psSolid;
      if Orientation = poLandscape then begin
        if i > 0 then begin
          MoveTo(x*i, 0);
          LineTo(x*i, Page.ClientHeight);
        end;
        OffsetRect(r, x, 0);
      end else begin
        if i > 0 then begin
          MoveTo(0, y*i);
          LineTo(Page.ClientWidth, y*i);
        end;
        OffsetRect(r, 0, y)
      end;
    end;
   finally
    {$IFDEF CLX_}
    Stop;
    {$ENDIF}
   end;
end;

procedure TPrintPreviewForm.CloseBtnClick(Sender: TObject);
begin
  Close;
end;

procedure TPrintPreviewForm.PrinterSetupClick(Sender: TObject);
begin
  PrinterSetup;
end;

procedure TPrintPreviewForm.PrinterSetup;
begin
  Printer.Orientation := Orientation;
  {$IFNDEF CLX_}
  if FPrinterSetupDialog = nil then
    FPrinterSetupDialog := TPrinterSetupDialog.Create(Self);
  if FPrinterSetupDialog.Execute then
  {$ELSE}
  if Printer.ExecuteSetup then
  {$ENDIF}
    if Orientation <> Printer.Orientation then
      Orientation := Printer.Orientation
    else
      UpdatePageSize(Printer,True);
end;

function PrintableImageLength(Length, Margin:Double; Page, Printable, Offset:Integer):Longint; { All in device units }
var
  p, l: LongInt;
begin
  l := Round(Length);
  { Position of start on first non-blank page }
  p := (Round(Margin) - Offset) mod Page;
  if p < 0 then p := p + Page;
  l := l - (Page - p);
  if p < Printable then begin
    Result := Printable - p;
    if (l < 0) and (-l > Page-Printable) then
      Result := l + Page - p;  { Entirely in first page }
  end else
    Result := 0;
  if l > 0 then begin
    p := l div Page;
    Result := Result + p*Printable;
    l := l - Page*p;
    if l > Printable then l := Printable;
    Result := Result + l;
  end;
end;

procedure TPrintPreviewForm.StretchHandleResized(Sender: TObject);
var
  W, H, X, Y: Double;
begin
  W := PageImage.Width / PreviewScale * PixPerInch.x;
  H := PageImage.Height / PreviewScale * PixPerInch.y;
  X := (PageImage.Left-Page.Left) / PreviewScale * PixPerInch.x;
  Y := (PageImage.Top-Page.Top) / PreviewScale * PixPerInch.y;
  if Pages > 1 then
    if Orientation = poLandscape then
      W := PrintableImageLength(W, X, PaperSize.x, PrintableAreaSize.x, TopLeft.x)
    else
      H := PrintableImageLength(H, Y, PaperSize.y, PrintableAreaSize.y, TopLeft.y);
  FXMargin := -LargeMargin; { Force UpdatePageSize }
  SetMargins(mmPerInch*W/PixPerInch.x, mmPerInch*H/PixPerInch.y,
             mmPerInch*X/PixPerInch.x, mmPerInch*Y/PixPerInch.y);
end;

procedure TPrintPreviewForm.SpinPagesChange(Sender: TObject);
begin
  Pages := SpinPages.Value;
end;

procedure TPrintPreviewForm.StretchHandleMoved(Sender: TObject);
begin
  FXMargin := -LargeMargin; { Force UpdatePageSize }
  SetMargins(XSize, YSize,
             mmPerInch*(PageImage.Left-Page.Left)/PreviewScale,
             mmPerInch*(PageImage.Top-Page.Top)/PreviewScale);
end;

procedure TPrintPreviewForm.PrintClick(Sender: TObject);
begin
  Print;
end;

procedure TPrintPreviewForm.Print;
begin
  if FPrintDialog = nil then
    FPrintDialog := TPrintDialog.Create(Self);
  FPrintDialog.MinPage := 1;
  if ppAreaIsFirstPage in Options then
    FPrintDialog.MaxPage := MaxPages
  else
    FPrintDialog.MaxPage := Pages;
  FPrintDialog.FromPage := 1;
  FPrintDialog.ToPage := FPrintDialog.MaxPage;
  if FPrintDialog.MaxPage > 1 then
    FPrintDialog.Options := FPrintDialog.Options + [poPageNums]
  else
    FPrintDialog.Options := FPrintDialog.Options - [poPageNums];
  FPrintDialog.PrintRange := prAllPages;
  if FPrintDialog.Execute then
  {$IFDEF CLX_}
  if Printer.ExecuteSetup then
  {$ENDIF}
    DoPrint;
end;

function PrinterRect(n:Integer; var l, r, Off, pg:Integer; var Pos, Len:Longint; Length, Margin:LongInt;
                     Page, Printable, Offset:Integer):Boolean;
begin
  if Len < 0 then begin
    Len := Length;
    Pos := Offset;
    pg := 0;
  end;
  repeat
    Inc(pg);
    r := Printable;
    if Pos < Margin then
      { Might be first page }
      if Pos + Printable < Margin then
        l := r { Nothing on this page - before start}
      else { First page }
        l := Margin - Pos
    else
      l := 0;
    if r-l > Len then r := l + Len;
    Len := Len - (r-l);
    Pos := Pos + Page;
  until pg = n;
  Off := r + Len - Length;
  Result := r > l;
end;

procedure TPrintPreviewForm.DoPrint;
var
  {$IFDEF MSWINDOWS} s, {$ENDIF}
  CPage, pg : Integer;
  Pos, Len: Longint;
  OffsetX, OffsetY: Integer;
  r: TRect;
  Some: Boolean;
  MaxPage:Integer;
  TheCanvas: TExportableCanvas;
begin
  TheCanvas := nil;
  if not Assigned(FOnShowPreview) then Exit;
  OffsetX := DeviceXMargin - TopLeft.x;
  OffsetY := DeviceYMargin - TopLeft.y;
  Len := -1;
  UpdatePageSize(Printer,True);
  Printer.Title := Application.Title;
  Printer.BeginDoc;
  {$IFDEF VCL_}
  { TGraph plotting is not re-entrant, so disable the abort processing
    to ensure that a graph window is not replotted while printing;
    this means a print progess dialog is not possible }
  SetAbortProc(Printer.Handle,nil);
  {$ENDIF}
  try
    {$IFDEF VCL_}
    if FPrintDialog = nil then
      FPrintDialog := TPrintDialog.Create(Self);
    MaxPage := FPrintDialog.ToPage;
    for CPage := FPrintDialog.FromPage to MaxPage do begin
    {$ELSE}
    if ppAreaIsFirstPage in Options then
      MaxPage := MaxPages
    else
      MaxPage := Pages;
    for CPage := 1 to MaxPage do begin
    {$ENDIF}
      r.Left := 0;
      r.Top := 0;
      r.Right := PrintableAreaSize.x;
      r.Bottom := PrintableAreaSize.y;
      if ppAreaIsFirstPage in Options then begin
        Some := True;
        Len := 1; { Force NewPage }
      end else
        if Orientation = poLandscape then
          Some := PrinterRect(CPage, r.Left, r.Right, OffsetX, pg, Pos, Len, Round(XSize/mmPerInch*PixPerInch.x), DeviceXMargin,
                              PaperSize.x, PrintableAreaSize.x, TopLeft.x)
        else
          Some := PrinterRect(CPage, r.Top, r.Bottom, OffsetY, pg, Pos, Len, Round(YSize/mmPerInch*PixPerInch.y), DeviceYMargin,
                              PaperSize.y, PrintableAreaSize.y, TopLeft.y);
      if Some then
        with Printer do begin
          {$IFDEF MSWINDOWS}
          s := SaveDC(Canvas.Handle);
          IntersectClipRect(Canvas.Handle, r.Left, r.Top, r.Right, r.Bottom);
          {$ENDIF}
          {$IFDEF UNIX}
          {$ENDIF}
          {$IFDEF NOLONGER}
            {$IFDEF WIN32}
            OffsetViewPortOrgEx(Canvas.Handle, OffsetX, OffsetY, PChar(nil)^);
            {$ELSE}
            OffsetViewPortOrgEx(Canvas.Handle, OffsetX, OffsetY, nil);
            {$ENDIF}
          {$ENDIF}
          try
            TheCanvas := TCanvasCanvas.Create(Canvas);
            FOnShowPreview(Self, TheCanvas, OffsetX, OffsetY,
                           Round(XSize/mmPerInch*PixPerInch.x),
                           Round(YSize/mmPerInch*PixPerInch.y), CPage, True);
          finally
            TheCanvas.Free;
          end;
          if not Printing then Break;
          {$IFDEF MSWINDOWS}
          RestoreDC(Canvas.Handle, s);
          {$ENDIF}
          {$IFNDEF LCL_}
          Canvas.Refresh; { should this be CLX_ only? }
          {$ENDIF}
        end;
      if (Len > 0) and (CPage <> MaxPage) then
        Printer.NewPage;
    end;
  finally
    if Printer.Printing then
      Printer.EndDoc;
{    if PrintProgressDlg <> nil then
      PrintProgressDlg.Hide; }
  end;
end;

procedure TPrintPreviewForm.CentreClick(Sender: TObject);
begin
  SetMargins(XSize, YSize, -2*LargeMargin, -2*LargeMargin);
end;

procedure TPrintPreviewForm.DefaultClick(Sender: TObject);
begin
  if not (ppAreaIsFirstPage in Options) then Pages := 1;
  SetMargins(-1, -1, 2*LargeMargin, 2*LargeMargin);
end;

procedure TPrintPreviewForm.WidthChange(Sender: TObject);
begin
  if not LockoutDoChange then
    SetMargins(PageWidth.Value, PageHeight.Value, XMargin, YMargin);
end;

procedure TPrintPreviewForm.LeftChange(Sender: TObject);
begin
  if not LockoutDoChange then
    XMargin := PageLeft.Value;
end;

procedure TPrintPreviewForm.TopChange(Sender: TObject);
begin
  if not LockoutDoChange then
    YMargin := PageTop.Value;
end;

procedure TPrintPreviewForm.FormMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if Button = mbRight then
  {$IFNDEF CLX_}
    Application.MessageBox(@PhysicalPrinterDetails[1], 'Physical Printer Details', MB_ICONINFORMATION or MB_OK);
  {$ELSE}
    Application.MessageBox(PhysicalPrinterDetails, 'Physical Printer Details');
  {$ENDIF}
end;

{$IFDEF MSWINDOWS}
procedure TPrintPreviewForm.WMEraseBkgnd(var Message: TWMEraseBkgnd);
begin
  if IsIconic(Handle) then inherited else begin
    SaveDC(Message.DC); { Not sure this bit will work, as scroll box now used }
    ExcludeClipRect(Message.DC, Page.Left, Page.Top, Page.Left + Page.Width, Page.Top + Page.Height);
    FillRect(Message.DC, ClientRect, Brush.Handle);
    RestoreDC(Message.DC, -1);
    Message.Result := 1;
  end;
end;
{$ENDIF}

procedure TPrintPreviewForm.SetAspectRatio(const Value: Double);
begin
  if Value <> AspectRatio then begin
    FAspectRatio := Value;
    StretchHandle.AspectRatio := Value;
  end;
end;

procedure TPrintPreviewForm.LandPortBtnClick(Sender: TObject);
begin
  if Orientation = poPortrait then
    Orientation := poLandscape
  else
    Orientation := poPortrait;
  DefaultClick(Sender);
end;

procedure TPrintPreviewForm.SetOptions(const Value: TPrintPreviewOptions);
begin
  if Options <> Value then begin
    FOptions := Value;
    PositionButtons;
  end;
end;

procedure TPrintPreviewForm.SetOnSettingsClick(const Value: TNotifyEvent);
begin
  if Assigned(FOnSettingsClick) <> Assigned(Value) then begin
    FOnSettingsClick := Value;
    PositionButtons;
  end else
    FOnSettingsClick := Value;
end;

const
  nSpeedBtn = 11;
  ControlFlags: array [1..nSpeedBtn] of TPrintPreviewOption =
                  (ppShowPrint, ppShowPrintSetup, ppShowLandPort, ppShowPages, ppShowCentre,
                   ppShowDefault, ppShowZoom, ppShowZoom, ppShowMaximize,
                   ppShowClose {Settings - ignored use Event instead}, ppShowClose);
procedure TPrintPreviewForm.PositionButtons;
var
  i, x: Integer;
  SpeedBtns: array [1..nSpeedBtn] of TWinControl;
begin
  x := PrintBtn.Left;
  with SpeedBar do
    for i := 0 to ControlCount-1 do
      if (Controls[i] is TWinControl) and not (Controls[i] is TLabel) then
        if TWinControl(Controls[i]).TabOrder+1 in [1..nSpeedBtn] then
          SpeedBtns[TWinControl(Controls[i]).TabOrder+1] := TWinControl(Controls[i]);
  for i := Low(SpeedBtns) to High(SpeedBtns) do
    if ( (SpeedBtns[i] <> Settings) and (ControlFlags[i] in Options))
       or ((SpeedBtns[i] = Settings) and Assigned(FOnSettingsClick) ) then begin
      {$IFDEF LCLcarbon}
      if ControlFlags[i] = ppShowPrintSetup then begin
        SpeedBtns[i].Visible := False;
        Continue;
      end;
      {$ENDIF}
      if ControlFlags[i] = ppShowPages then
        with PagesLabel do begin
          Visible := True;
          Left := x;
          Inc(x, Width+2);
        end;
      with SpeedBtns[i] do begin
          Visible := True;
          Left := x;
          Inc(x, Width+4);
      end;
      if ControlFlags[i] = ppShowPages then
        Inc(x, SpinPages.ExtraWidth);
    end else begin
      SpeedBtns[i].Visible := False;
    end;
  PagesLabel.Visible := SpinPages.Visible;
  Constraints.MinHeight := (SpeedBar.Height + StatusBar.Height)*2;
  Constraints.MinWidth := x;
end;

procedure TPrintPreviewForm.SetMaxPages(const Value: Integer);
begin
  FMaxPages := Value;
  if Pages > MaxPages then
    Pages := MaxPages;
  SpinPages.MaxValue := FMaxPages;
  if MaxPages = 1 then
    Options := Options - [ppShowPages]
  else
    Options := Options + [ppShowPages];
end;

procedure TPrintPreviewForm.SetLandPortBtn;
begin
  if Orientation = poPortrait then
    LandPortBtn.Hint := 'Landscape|Rotate Paper'
  else
    LandPortBtn.Hint := 'Portrait|Rotate Paper';
end;

procedure TPrintPreviewForm.ZoomClick(Sender: TObject);
begin
  if ZoomFactor > 8 then
    ZoomFactor := 1
  else
    ZoomFactor := ZoomFactor * 2;
end;

procedure TPrintPreviewForm.MaximizeClick(Sender: TObject);
begin
  SetMargins(Xinch(PrintableAreaSize.x)*mmperInch,
             Yinch(PrintableAreaSize.y)*mmperInch,
             Xinch(TopLeft.x)*mmperInch,
             Yinch(TopLeft.y)*mmperInch);
end;

procedure TPrintPreviewForm.SetZoomFactor(const Value: Double);
begin
  if Value <> ZoomFactor then begin
    FZoomFactor := Value;
    Changed;
  end;
end;

procedure TPrintPreviewForm.ShrinkBtnClick(Sender: TObject);
begin
  if ZoomFactor < 0.1 then
    ZoomFactor := 1
  else
    ZoomFactor := ZoomFactor / 2;
end;

procedure TPrintPreviewForm.FormClose(Sender: TObject;
  var CloseAction: TCloseAction);
begin
  CloseAction := caFree;
end;

procedure TPrintPreviewForm.SettingsClick(Sender: TObject);
begin
  if Assigned(FOnSettingsClick) then
    FOnSettingsClick(Self);
end;

end.

