unit ProjAbout;
{ Automatic version number control. This is based on TProject by
  T S Eriksen, though almost all of the code has been moved to a
  design time only unit. }
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses
  {$IFDEF WIN32} Windows, {$ENDIF}
  StdCtrls, Buttons, Controls, ExtCtrls, Forms, Classes;

type
  TFileType = (ftProgram,ftLibrary);

{CMW}

  { TAboutBox }

  TAboutBox = class(TForm)
    Panel: TPanel;
    ProgramIcon: TImage;
    ProductName: TLabel;
    Description: TLabel;
    Details: TLabel;
    OKButton: TButton;
    procedure FormShow(Sender: TObject);
  end;

  TProject = class(TComponent)
  private
    { user defined properties }
    FResFile        : string; { Resourcefile name }
    FCompany        : string; { stored in Ini-file }
    FCopyright      : string; { stored in Ini-file }
    FComment        : string; { default : <comments about this file/product> }
    FDescription    : string; { default : <Description of file> }
    FSpecialBuild   : string; { default : <How this build differs from a standard build> }
    FPrivateBuild   : string; { default : <Private information> }
    FProductname    : string; { default : <name of product this file ships with> }
    FProductVersion : string; { default : <0.0.0.0> }
    FOriginalName   : string; { default : <name of .exe file > }
    FTrademarks     : string; { default : <Legal trademarks used in this build> }
    FLanguage       : string; { default : UK_English }
    FCharset        : string; { default : Multilinqual }
    FFileType       : TFileType; { default : ffProgram }
    { Auto defined properties }
    { Internalname, }         { TProject.name }
    { Fileversion , }         { TProject.version + TProject.build }
    { FileFlags }             { ffprerease when version < 1.0, special/private build }
    { property data buffers etc }
    FAutoNaming  : Boolean;   { Styrer automatisk navngivning af project }
    // FProjectPath : string;    { hvor de skal kopieres fra }
    FBuild       : Integer;   { build nr, incr. ved component write state}
    FVersion     : string;    { Version number }
    FBuildDate   : string;
  protected
    { Protected declarations }
    procedure   Loaded; override;
    procedure   WriteState(Writer:TWriter);    override;
  public
    constructor Create(AComponent:TComponent); override;
    procedure   Execute(VersionInfo: string = '');
    { Public declarations }
  published
    { Published declarations }
    property AutoNaming  : Boolean   read FAutoNaming  write FAutoNaming;
    property Build       : Integer   read FBuild       write FBuild default 0;
    property Version     : string    read FVersion     write FVersion;
    { version resource properties }
    // property ProjectPath : string   read FProjectPath    write FProjectPath;
    property Company       : string read FCompany        write FCompany;
    property Copyright     : string read FCopyright      write FCopyright;
    property Comment       : string read FComment        write FComment;
    property Description   : string read FDescription    write FDescription;
    property SpecialBuild  : string read FSpecialBuild   write FSpecialBuild;
    property PrivateBuild  : string read FPrivateBuild   write FPrivateBuild;
    property Productname   : string read FProductName    write FProductname;
    property ProductVersion: string read FProductVersion write FProductVersion;
    property OriginalName  : string read FOriginalName   write FOriginalName;
    property Trademarks    : string read FTrademarks     write FTrademarks;
    property Language      : string read FLanguage       write FLanguage;
    property Charset       : string read FCharset        write FCharset;
    property FileType      : TFileType read FFileType    write FFileType;
    property ResourceFile  : string    read FResFile     write FResFile;
    property BuildDate     : string    read FBuildDate   write FBuildDate;
  end;

  TProjectHelper = class
    procedure DoWriteState(Project:TProject); virtual; abstract;
    procedure DoCreate(Project:TProject); virtual; abstract;
  end;

var
  DesignProjectHelper: TProjectHelper;

implementation

Uses
  Math, SysUtils, Sysinfo, UCompat, Graphics, Types;

{CMW}
{$IFNDEF LCL_}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function CheckStr(const s:string):string;
begin
  if (s = '') or (s[1] = '<') then
    Result := ''
  else
    Result := s;
end;

procedure TProject.Execute(VersionInfo: string); {CMW}
var
  About:TAboutBox;
  ProdName:string;
  procedure Add(s: string);
  var
    c: string;
    {$IFDEF FPC}
    i: Integer;
    {$ENDIF}
  begin
    if s <> '' then begin
      {$IFDEF FPC}
      {Strip UTF8 characters for the time being}
      for i := 1 to Length(s) do
        if Ord(S[i]) > 128 then
          s[i] := ' ';
      {$ENDIF}
      c := About.Details.Caption;
      if c <> '' then
        About.Details.Caption := Concat(c, LineEnd, s)
      else
        About.Details.Caption := s;
    end;
  end;
var
  Icon: TIcon;
  wh: Integer;
begin
  About := TAboutBox.Create(Application);
  try
    ProdName := CheckStr(ProductName);
    if ProdName = '' then
      ProdName := Application.Title;
    About.ProductName.Caption := ProdName;
    if not Assigned(Application.Icon) or Application.Icon.Empty then begin
      if Assigned(Application.MainForm) and Assigned(TForm(Application.MainForm).Icon) and not TForm(Application.MainForm).Icon.Empty then
      Icon := TForm(Application.MainForm).Icon;
    end else
      Icon := Application.Icon;
    About.ProgramIcon.Picture.Icon := Icon;
    {$IFDEF FPC}
    with About.ProgramIcon do begin
      wh := Min(Width,Height);
      with Picture.Icon do
        Current := GetBestIndexForSize(Size(wh,wh));
    end;
    {$ENDIF}
    if VersionInfo = '' then
      VersionInfo := Concat(Version, '.', IntToStr(Build), ' ', BuildDate);
    About.Description.Caption := Concat(CheckStr(Description), LineEnd, VersionInfo);
    About.Details.Caption := '';
    Add(CheckStr(Comment));
    Add(CheckStr(Copyright));
    Add(CheckStr(Company));
    Add(Concat(LineEnd, SysInfo.DetermineCpu(True)));
    About.ShowModal;
  finally
    About.Free;
  end;
end;

{ protected procedures }

procedure TProject.WriteState(Writer:TWriter);
begin
  { When something on a form has changed all components (on it) will
    be saved when next compiled - before that save, we'll just
    increase the Build number }
  if csDesigning in ComponentState then begin
    Inc(FBuild);                        { increase build number }
    DateTimeToString(FBuildDate, 'd mmm yyyy h:nn', Now);
    if DesignProjectHelper <> nil then
      DesignProjectHelper.DoWriteState(Self);
  end;
  Inherited WriteState(Writer);
end;

procedure TProject.Loaded;
var
  s : string;
  a : integer;
begin
  inherited Loaded;
  If (csDesigning in ComponentState) then begin
    { Denne blok udfres kun ved
      1) Load af projekt
      2) Visning af form/modul hvor TProject er lagret  }
    Inc(FBuild);                     { increase build number }
  end else begin
    { Denne blok udfrers kun ved
      1) Start af applikation }
    If (AutoNaming) then begin
      application.Title:=Name+' ver. '+Version+'.'+
                         IntToStr(Build);
      s:=application.title;
      For a:=1 to length(s) do if s[a]='_' then s[a]:=' ';
      application.title:=s;
    end;
  end;
end;

constructor TProject.Create(AComponent:TComponent);
begin
  Inherited Create(AComponent);
  { Only do this if designing - if runtime, it's not nesseary }
  If (csDesigning in ComponentState) and (DesignProjectHelper <> nil) then
    DesignProjectHelper.DoCreate(Self);
end;

{ TAboutBox }

procedure TAboutBox.FormShow(Sender: TObject);
var
  Extra: Integer;
begin
  Description.Width := Panel.Width - Description.Left - 6;
  Description.AutoSize := True;
  Details.Top := Max(ProgramIcon.Top+ProgramIcon.Height, Description.Top+Description.Height)+5;
  Details.Anchors := [akLeft, akTop];
  {$IFNDEF FPC}
    Details.AutoSize := True;
  {$ENDIF}
  Extra := Details.Top + Details.Height + 8 - Panel.Height;
  if Extra > 0 then
    Height := Height + Extra;
  Details.Anchors := Details.Anchors + [akBottom];
end;

end.


