unit ScrlCtrl;
{ Control with optional scoll bars; implmented to provide more control
  over scroll bars, and to smooth over the inability to add scroll bars
  to an arbitrary control in Kylix }
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Classes,
     {$IFNDEF CLX_} Messages, Forms, StdCtrls, ExtCtrls, Controls, Graphics,
     {$ELSE} Types, QGraphics, QControls, QStdCtrls, QExtCtrls, QForms, {$ENDIF}
     {$IFDEF LCL} LCLType, {$ENDIF}
     WCompat;

type

  TScrollableControl = class;

  TControlScrollBarEx = class(TPersistent)
  private
    FControl: TScrollableControl;
    FKind: TScrollBarKind;
    FPosition: Longint;
    FRange: Longint;
    FPageSize: Integer;
    FTracking: Boolean;
    FPositionChanging: Boolean;
    {$IFDEF MSWINDOWS}
    fnBar: Integer;
    {$ENDIF}
    {$IFDEF UNIX}
    FScrollBar: TScrollBar;
    ForceOff: Boolean;
    procedure CheckScrollBar;
    procedure ScrollBarScroll(Sender: TObject);
    function GetVisible: Boolean;
    function GetSize: Integer;
    {$ENDIF}
    procedure SetPosition(Value: Longint);
    procedure SetRange(Value: Longint);
    procedure SetPageSize(const Value: Integer);
  public
    ScrollJump: Integer; { How many scroll bar steps to move on a simple click } { TODO: Linux }
    ScaleFactor: Integer; { From scroll bar changes to pixels to scroll }
    {$IFDEF MSWINDOWS}
    procedure ScrollMessage(var Msg: TWMScroll);
    {$ENDIF}
    constructor Create(AControl: TScrollableControl; AKind: TScrollBarKind);
    destructor Destroy; override;
    property Control:TScrollableControl read FControl;
    property Kind: TScrollBarKind read FKind;
    property Position: Longint read FPosition write SetPosition;
    property Range: Longint read FRange write SetRange;
    property PageSize: Integer read FPageSize write SetPageSize;
    {$IFDEF UNIX}
    property Visible: Boolean read GetVisible;
    property Size: Integer read GetSize;
    {$ENDIF}
  published
    property Tracking: Boolean read FTracking write FTracking default True;
  end;

  { TScrollableControl }

  TScrollableControl = class(TCustomControl)
  private
    FBevelInner: TPanelBevel;
    FBevelOuter: TPanelBevel;
    FBevelWidth: TBevelWidth;
    FHorzScrollBar: TControlScrollBarEx;
    FVertScrollBar: TControlScrollBarEx;
    FScrollBars: TScrollStyle;
    FPaintRect: TRect;                  { ClientRect less borders and scroll bar }
    {$IFDEF UNIX}
    FCornerWidget: TPanel;
    FPartInvalidateCount: Integer;
    {$ENDIF}
    {$IFDEF VCL_}
    FBorderStyle: TBorderStyle;
    procedure SetBorderStyle(const AValue: TBorderStyle);
    {$ENDIF}
    {$IFDEF MSWINDOWS}
    procedure WMHScroll(var Message: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var Message: TWMVScroll); message WM_VSCROLL;
    {$ENDIF}
    procedure SetBevelInner(const AValue: TPanelBevel);
    procedure SetBevelOuter(const AValue: TPanelBevel);
    procedure SetBevelWidth(const AValue: TBevelWidth);
    procedure WMMouseWheel(var Msg: TWMMouseWheel); message WM_MOUSEWHEEL;
    procedure SetTopMargin(const Value: Integer);
    procedure SetLeftMargin(const Value: Integer);
    function GetPaintWidth: Integer;
    function GetPaintHeight: Integer;
  protected
    FTopMargin: Integer;
    FLeftMargin: Integer;
    {$IFDEF MSWINDOWS}
    procedure CreateParams(var Params: TCreateParams); override;
    procedure WndProc(var Msg: TMessage); override;
    {$ENDIF}
    {$IFDEF UNIX}
    function GetClientRect: TRect; override;
    {$ENDIF}
    function BorderSize: Integer;
    procedure AdjustClientRect(var Rect: TRect); override;
    procedure PaintBorder;
    procedure Paint; override;
    procedure Resize; override;
    procedure SetScrollBars(Value: TScrollStyle);
    procedure UpdatePaintRect;
    procedure UpdateScrollBars; virtual;
    property PaintWidth: Integer read GetPaintWidth;
    property PaintHeight: Integer read GetPaintHeight;
  public
    property PaintRect: TRect read FPaintRect;
    procedure ScrollBy(DeltaX, DeltaY: Integer); {$IFDEF CLX_} override; {$ENDIF}
    {$IFDEF UNIX}
    procedure Invalidate; override;
    procedure InvalidateRect(Rect: TRect; DOEraseBackground: Boolean);
    {$ENDIF}
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property BevelInner: TPanelBevel read FBevelInner write SetBevelInner default bvNone;
    property BevelOuter: TPanelBevel read FBevelOuter write SetBevelOuter default bvRaised;
    property BevelWidth: TBevelWidth read FBevelWidth write SetBevelWidth default 1;
    {$IFDEF VCL_}
    property BorderStyle: TBorderStyle read FBorderStyle write SetBorderStyle default bsNone;
    {$ENDIF}
  published
    property HorzScrollBar:TControlScrollBarEx read FHorzScrollBar;
    property VertScrollBar:TControlScrollBarEx read FVertScrollBar;
    property TopMargin: Integer read FTopMargin write SetTopMargin default 0;
    property LeftMargin: Integer read FLeftMargin write SetLeftMargin default 0;
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssBoth;
    property PopupMenu;
    property OnContextPopup;
    property OnMouseWheel;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
  end;

implementation

uses {$IFDEF CLX_} Qt, {$ENDIF} {$IFDEF LCL_} LCLIntf, GraphType, {$ENDIF} SysUtils;

{ TControlScrollBarEx }

constructor TControlScrollBarEx.Create(AControl: TScrollableControl; AKind: TScrollBarKind);
begin
  inherited Create;
  ScaleFactor := 1;
  ScrollJump := 1;
  FKind := AKind;
  FControl := AControl;
  FPosition := 0;
  FRange := 0;
  Tracking := True;
  {$IFDEF MSWINDOWS}
  if Kind = sbHorizontal then
    fnBar := SB_HORZ
  else
    fnBar := SB_VERT;
  {$ENDIF}
end;

destructor TControlScrollBarEx.Destroy;
begin
  inherited Destroy;
end;

procedure TControlScrollBarEx.SetPosition(Value: Longint);
var
  ScrollAmount: Longint;
  LargeScroll: Boolean;
begin
  if Value > Range then
    Value := Range
  else if Value < 0 then
    Value := 0;
  if (Value <> Position) and not FPositionChanging then
   try
    FPositionChanging := True;
    ScrollAmount := (FPosition - Value)*ScaleFactor;
    if Kind = sbHorizontal then
      LargeScroll := Abs(ScrollAmount) > Control.ClientWidth div 2
    else
      LargeScroll := Abs(ScrollAmount) > Control.ClientHeight div 2;
    {$IFDEF CLX_}
    if not LargeScroll and (Control.FPartInvalidateCount > 0) then
      Control.Update; { Invalid areas are not shifted by scroll, so paint
                    before position variable changes }
    {$ENDIF}
    FPosition := Value;
    if LargeScroll then
      Control.Invalidate
    else
      if Kind = sbHorizontal then begin
        Control.ScrollBy(ScrollAmount, 0);
      end else
        Control.ScrollBy(0, ScrollAmount);
    {$IFDEF MSWINDOWS}
    SetScrollPos(Control.Handle, fnBar, Position, True);
    {$ENDIF}
    {$IFDEF UNIX}
    CheckScrollBar;
    {$ENDIF}
    {*** may need PositionEditWindow here }
   finally
    FPositionChanging := False;
   end;
end;

{$IFDEF UNIX}
function TControlScrollBarEx.GetSize: Integer;
begin
  if FScrollBar <> nil then
    if Kind = sbHorizontal then
      Result := FScrollBar.Height
    else
      Result := FScrollBar.Width
  else
    Result := 0;
end;

procedure TControlScrollBarEx.ScrollBarScroll(Sender: TObject);
begin
  Position := FScrollBar.Position;
  // ScrollPos := Position; { In case its changed }
end;

procedure TControlScrollBarEx.CheckScrollBar;
{ Create scroll bar if necessary, and sync range+position }
var
  SavedPageSize, newRange: Integer;
  procedure PositionBar;
  begin
    if Kind = sbHorizontal then
      FScrollBar.SetBounds(Control.BorderSize, Control.Height-FScrollBar.Height-Control.BorderSize,
                           Control.ClientWidth-2*Control.BorderSize, FScrollBar.Height)
    else
      FScrollBar.SetBounds(Control.Width-FScrollBar.Width-Control.BorderSize, Control.BorderSize,
                           FScrollBar.Width, Control.ClientHeight-2*Control.BorderSize);
  end;
begin
  if not ForceOff and (Range > 0) then begin
    {$IFDEF LCL_}
    newRange := Range + PageSize - 1;
    {$ELSE}
    newRange := Range;
    {$ENDIF}
    if FScrollBar = nil then begin
      FScrollBar := TScrollBar.Create(Control);
      with FScrollBar do begin
        Parent := Control;
        Kind := FKind;
        SavedPageSize := PageSize;
        FPageSize := -1;
        PageSize := SavedPageSize;
        SetParams(Position, 0, newRange, SavedPageSize);
        {$IFDEF LCLcarbon}
        if FKind = sbVertical then
          Width := GetSystemMetrics(SM_CXVSCROLL);
        {$ENDIF}
        PositionBar;
        TabStop := False;
        OnChange := {$IFDEF FPC}@{$ENDIF}ScrollBarScroll;
        Color := clBtnFace;
      end;
    end else begin
      FScrollBar.SetParams(Position, 0, newRange, PageSize);
      {$IFDEF LCL_}
      FScrollBar.PageSize := PageSize;
      {$ENDIF}
      PositionBar;
    end;
    FScrollBar.Show;
  end else if FScrollBar <> nil then
    FScrollBar.Hide;
end;

function TControlScrollBarEx.GetVisible: Boolean;
begin
  Result := not ForceOff and (Range > 0);
end;

{$ENDIF}

{$IFDEF MSWINDOWS}
procedure TControlScrollBarEx.ScrollMessage(var Msg: TWMScroll);
{$IFDEF WIN32}
var
  si: TScrollInfo;
{$ENDIF}
begin
  case Msg.ScrollCode of
    SB_LINEUP: SetPosition(Position - ScrollJump);
    SB_LINEDOWN: SetPosition(Position + ScrollJump);
    SB_PAGEUP: SetPosition(Position - PageSize + 2);
    SB_PAGEDOWN: SetPosition(Position + PageSize - 2);
    SB_THUMBPOSITION, SB_THUMBTRACK:
{$IFDEF WIN32}
      if Range > 32767 then begin
        si.cbSize := SizeOf(si);
        si.fMask := SIF_TRACKPOS;
        GetScrollInfo(Control.Handle, fnBar, si);
        SetPosition(si.nTrackPos)
      end else
{$ENDIF}
        SetPosition(Msg.Pos);
    SB_TOP: SetPosition(0);
    SB_BOTTOM: SetPosition(Range);
    SB_ENDSCROLL: begin end;
  end;
end;
{$ENDIF}

procedure TControlScrollBarEx.SetPageSize(const Value: Integer);
{$IFDEF MSWINDOWS}
var
  Rnge: Integer;
{$ENDIF}
begin
  if PageSize <> Value then begin
    FPageSize := Value;
    if FPageSize <= 0 then FPageSize := 1; { Force sane value } 
    {$IFDEF UNIX}
    if FScrollBar <> nil then
      if PageSize < 3 then
        FScrollBar.LargeChange := 2
      else
        FScrollBar.LargeChange := PageSize - 2;
    {$ENDIF}
    {$IFDEF MSWINDOWS}
    Rnge := Range;
    FRange := -1;
    SetRange(Rnge); { Force scroll bar update }
    {$ENDIF}
  end
end;

procedure TControlScrollBarEx.SetRange(Value: Longint);
{$IFDEF WIN32}
var
  ScrollInfo: TScrollInfo;
{$ENDIF}
begin
  if Value < 0 then Value := 0;
  if Value <> Range then begin
    FRange := Value;
    if Position > Value then
      SetPosition(Value); { To avoid re-entrancy problems, change position first }
  {$IFDEF MSWINDOWS}
   {$IFDEF WIN32}
    ScrollInfo.cbSize := SizeOf(ScrollInfo);
    ScrollInfo.fMask := SIF_ALL;
    ScrollInfo.nMin := 0;
    if Value = 0 then
      ScrollInfo.nMax := 0
    else
      ScrollInfo.nMax := Value + PageSize - 1;
    ScrollInfo.nPage := PageSize;
    ScrollInfo.nPos := Position;
    SetScrollInfo(Control.Handle, fnBar, ScrollInfo, TRUE);
   {$ELSE}
    SetScrollRange(Control.Handle, fnBar, 0, Value, TRUE);
   {$ENDIF}
  {$ENDIF}
  {$IFDEF UNIX}
    CheckScrollBar;
  {$ENDIF}
  end;
end;

{ TScrollableControl }

constructor TScrollableControl.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FScrollBars := ssBoth;
  ControlStyle := ControlStyle + [csFramed, csCaptureMouse, csOpaque] - [csSetCaption];
  FHorzScrollBar := TControlScrollBarEx.Create(Self, sbHorizontal);
  FVertScrollBar := TControlScrollBarEx.Create(Self, sbVertical);
  FBevelOuter := bvRaised;
  FBevelInner := bvNone;
  FBevelWidth := 1;
end;

destructor TScrollableControl.Destroy;
begin
  FreeAndNil(FHorzScrollBar);
  FreeAndNil(FVertScrollBar);   
  {$IFDEF UNIX}
  FreeAndNil(FCornerWidget);
  {$ENDIF}
  inherited Destroy;
end;

procedure TScrollableControl.PaintBorder;
var
  Rect: TRect;
  {$IFNDEF LCL_}
  TopColor, BottomColor: TColor;
  procedure AdjustColors(Bevel: TPanelBevel);
  begin
    TopColor := clBtnHighlight;
    if Bevel = bvLowered then TopColor := clBtnShadow;
    BottomColor := clBtnShadow;
    if Bevel = bvLowered then BottomColor := clBtnHighlight;
  end;
  {$ENDIF}
begin
  Rect := inherited GetClientRect;
  if BorderStyle = bsSingle then begin
    // May want Frame3D(Canvas, Rect, Color, Color, BorderWidth); force Color = clBlack?
    Canvas.Rectangle(Rect);
    InflateRect(Rect, -1, -1);
  end;
  if BevelOuter <> bvNone then begin
    {$IFDEF LCL_}
    Canvas.Frame3D(Rect, BevelWidth, BevelOuter);
    InflateRect(Rect, -1, -1);
    {$ELSE}
    AdjustColors(BevelOuter);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
    {$ENDIF}
  end;
  if BevelInner <> bvNone then begin
    {$IFDEF LCL_}
    Canvas.Frame3D(Rect, BevelWidth, BevelInner);
    {$ELSE}
    AdjustColors(BevelInner);
    Frame3D(Canvas, Rect, TopColor, BottomColor, BevelWidth);
    {$ENDIF}
  end;
end;

function TScrollableControl.BorderSize: Integer;
begin
  Result := BorderWidth;
  if BevelOuter <> bvNone then Inc(Result, BevelWidth);
  if BevelInner <> bvNone then Inc(Result, BevelWidth);
end;

procedure TScrollableControl.AdjustClientRect(var Rect: TRect);
var
  BevelSize: Integer;
begin
  inherited AdjustClientRect(Rect);
  BevelSize := BorderSize;
  InflateRect(Rect, -BevelSize, -BevelSize);
end;

procedure TScrollableControl.Paint;
var
  CRect: TRect;
begin
  {$IFDEF UNIX}
  FPartInvalidateCount := 0;
  {$ENDIF}
  PaintBorder;
  if TopMargin > 0 then
    Canvas.FillRect(Rect(PaintRect.Left-LeftMargin, PaintRect.Top-TopMargin, PaintRect.Right, PaintRect.Top));
  if LeftMargin > 0 then
    Canvas.FillRect(Rect(PaintRect.Left-LeftMargin, PaintRect.Top, PaintRect.Left, PaintRect.Bottom));
  IntersectRect(CRect, Canvas.ClipRect, PaintRect);
  {$IFDEF CLX_}
  Canvas.SetClipRect(CRect);
  {$ELSE}
  IntersectClipRect(Canvas.Handle, CRect.Left, CRect.Top, CRect.Right, CRect.Bottom);
  {$ENDIF}
end;

procedure TScrollableControl.SetScrollBars(Value: TScrollStyle);
begin
  if FScrollBars <> Value then
  begin
    FScrollBars := Value;
    {$IFDEF MSWINDOWS}
    if HandleAllocated then
      RecreateWnd{$IFDEF LCL}(Self){$ENDIF};
    {$ENDIF}
    {$IFDEF UNIX}
    VertScrollBar.ForceOff := Value in [ssNone, ssHorizontal, ssAutoHorizontal];
    HorzScrollBar.ForceOff := Value in [ssNone, ssVertical, ssAutoHorizontal];
    {$ENDIF}
  end;
end;

procedure TScrollableControl.UpdateScrollBars;
begin
 {$IFDEF UNIX}
  if not( Assigned(VertScrollBar) and Assigned(HorzScrollBar) ) then
    Exit;
  if VertScrollBar.Visible and HorzScrollBar.Visible then begin
    if FCornerWidget = nil then begin
      FCornerWidget := TPanel.Create(Self);
      FCornerWidget.Parent := Self;
      with FCornerWidget do begin
        Color := clBtnFace;
        BorderStyle := bsNone;
        BevelInner := bvNone;
        BevelOuter := bvNone;
      end;
    end;
    FCornerWidget.SetBounds(ClientWidth-BorderSize, ClientHeight-BorderSize, VertScrollBar.Size, HorzScrollBar.Size);
    FCornerWidget.Show;
  end else if (FCornerWidget <> nil) and FCornerWidget.Visible then
    FCornerWidget.Hide;
  HorzScrollBar.CheckScrollBar;
  VertScrollBar.CheckScrollBar;
  {$ENDIF}
  UpdatePaintRect;
end;

{$IFDEF MSWINDOWS}
procedure TScrollableControl.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  with params do begin
    if FScrollBars in [ssVertical, ssBoth] then Style := Style or WS_VSCROLL;
    if FScrollBars in [ssHorizontal, ssBoth] then Style := Style or WS_HSCROLL;
  end;
{ if FBorderStyle = bsSingle then
    Params.Style := Params.Style or WS_BORDER; }
  { Params.ExStyle := Params.ExStyle or $200 } { WS_EX_CLIENTEDGE };
end;

procedure TScrollableControl.WndProc(var Msg: TMessage);
begin
  if Msg.Msg = uMSH_MOUSEWHEEL then begin
    { Translate W95 style message }
    Msg.Msg := WM_MOUSEWHEEL;
    {$IFDEF FPC}
      Msg.wParamhi := Msg.wParamlo;
      Msg.wParamlo := 0;
    {$ELSE}
      {$ifdef WIN32}
      TWMMouseWheel(Msg).WheelDelta := Msg.WParam;
      TWMMouseWheel(Msg).Keys := 0;
      {$endif}
    {$ENDIF}
  end;
  inherited WndProc(Msg);	{ dispatch normally }
end;

procedure TScrollableControl.WMHScroll(var Message: TWMHScroll);
begin
  if (Message.ScrollBar = 0) then
    HorzScrollBar.ScrollMessage(Message)
  else
    inherited;
end;

procedure TScrollableControl.WMVScroll(var Message: TWMVScroll);
begin
  if (Message.ScrollBar = 0) then
    VertScrollBar.ScrollMessage(Message)
  else
    inherited;
end;
{$ENDIF}

{ TODO: Wheel handling under linux }
procedure TScrollableControl.WMMouseWheel(var Msg: TWMMouseWheel);
var
  i:Integer;
begin
  i := WheelScrollLines * Msg.WheelDelta div WHEEL_DELTA;
  if VertScrollBar.Range > 0 then
    VertScrollBar.Position := VertScrollBar.Position - i*VertScrollBar.ScrollJump
  else if HorzScrollBar.Range > 0 then
    HorzScrollBar.Position := HorzScrollBar.Position - i*HorzScrollBar.ScrollJump;
end;

procedure TScrollableControl.SetBevelInner(const AValue: TPanelBevel);
begin
  if FBevelInner = AValue then Exit;
  FBevelInner := AValue;
  Invalidate;
end;

procedure TScrollableControl.SetBevelOuter(const AValue: TPanelBevel);
begin
  if FBevelOuter = AValue then Exit;
  FBevelOuter := AValue;
  Invalidate;
end;

procedure TScrollableControl.SetBevelWidth(const AValue: TBevelWidth);
begin
  if FBevelWidth = AValue then Exit;
  FBevelWidth := AValue;
  Invalidate;
end;

{$IFDEF VCL_}
procedure TScrollableControl.SetBorderStyle(const AValue: TBorderStyle);
begin
  if AValue = FBorderStyle then Exit;
  FBorderStyle := AValue;
  Invalidate
end;
{$ENDIF}

procedure TScrollableControl.ScrollBy(DeltaX, DeltaY: Integer);
{$IFDEF MSWINDOWS}
var
  WRect: TRect;
{$ENDIF}
begin
  if HandleAllocated and Visible then begin
    {$IFDEF MSWINDOWS}
    { Don't use FPaintRect as invalidated regions are not shifted
      if scroll rect not null; }
    ScrollWindow(Handle, DeltaX, DeltaY, nil {FPaintRect}, nil);
    { To make up for this, invalidate borders }
    if DeltaX <> 0 then begin
      WRect := Rect(0, 0, PaintRect.Left, Height);
      if DeltaX > 0 then
        Inc(WRect.Right, DeltaX);
      InvalidateRect(Handle, @WRect, FALSE);
      WRect := Rect(PaintRect.Right, 0, Width, Height);
      if DeltaX < 0 then
        Inc(WRect.Left, DeltaX);
      InvalidateRect(Handle, @WRect, FALSE);
    end;
    if DeltaY <> 0 then begin
      WRect := Rect(0, 0, Width, PaintRect.Top);
      if DeltaY > 0 then
        Inc(WRect.Bottom, DeltaY);
      InvalidateRect(Handle, @WRect, FALSE);
      WRect := Rect(0, PaintRect.Bottom, Width, Height);
      if DeltaY < 0 then
        Inc(WRect.Top, DeltaY);
      InvalidateRect(Handle, @WRect, FALSE);
    end;
    {$ENDIF}
    {$IFDEF CLX_}
    FPartInvalidateCount := 1;
    QWidget_scroll(Handle, DeltaX, DeltaY, @FPaintRect);
    {$ENDIF}
    {$IFDEF LCL_}
    Invalidate; { TODO : Find out how to scroll }
    {$ENDIF}
  end;
end;

procedure TScrollableControl.UpdatePaintRect;
begin
  if not HandleAllocated then Exit;
  FPaintRect := ClientRect;
  AdjustClientRect(FPaintRect); // Previously: InflateRect(FPaintRect, -BorderSize, -BorderSize);
  Inc(FPaintRect.Left, LeftMargin);
  Inc(FPaintRect.Top, TopMargin);
  { Fudge to prevent divide by zero }
  if PaintHeight < 1 then Dec(FPaintRect.Left);
  if PaintWidth < 1 then Dec(FPaintRect.Top);
end;

procedure TScrollableControl.Resize;
begin
  inherited Resize;
  UpdateScrollBars;
end;

{$IFDEF UNIX}
procedure TScrollableControl.Invalidate;
begin
  inherited;
  FPartInvalidateCount := 0;
end;

procedure TScrollableControl.InvalidateRect(Rect: TRect; DOEraseBackground: Boolean);
begin
  {$IFDEF LCL_}
  LCLIntf.InvalidateRect(Handle, @Rect, DoEraseBackground);
  {$ELSE}
  inherited InvalidateRect(Rect, DoEraseBackground);
  {$ENDIF}
  Inc(FPartInvalidateCount);
end;

function TScrollableControl.GetClientRect: TRect;
begin
  Result := inherited GetClientRect;
  if Assigned(HorzScrollBar) and HorzScrollBar.Visible then
    Dec(Result.Bottom, HorzScrollBar.Size);
  if Assigned(VertScrollBar) and VertScrollBar.Visible then
    Dec(Result.Right, VertScrollBar.Size);
end;
{$ENDIF}

procedure TScrollableControl.SetTopMargin(const Value: Integer);
begin
  FTopMargin := Value;
  Invalidate;
  UpdatePaintRect;
end;

procedure TScrollableControl.SetLeftMargin(const Value: Integer);
begin
  FLeftMargin := Value;
  Invalidate;
  UpdatePaintRect;
end;

function TScrollableControl.GetPaintWidth: Integer;
begin
  Result := PaintRect.Right - PaintRect.Left;
end;

function TScrollableControl.GetPaintHeight: Integer;
begin
  Result := PaintRect.Bottom - PaintRect.Top;
end;


end.
