unit SlaveEdit;
{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface
{$IFDEF XBASED} uses X; {$ENDIF}


type
  TWindowMessage = {$IFDEF XBASED} TAtom {$ELSE} Cardinal {$ENDIF};

procedure LaunchEditor(FileName: string; LineNumber: Integer = -1);
function SaveAllWindows: Boolean;
procedure BroadcastNotification(WindowMessage: TWindowMessage);
function BroadcastName(const Name: string):TWindowMessage;
function RegisterMessage(const Name: string):TWindowMessage;

implementation

uses {$IFDEF MSWINDOWS} ShellAPI, Windows, Messages, {$ENDIF}
     {$IFDEF XBASED}
       Xlib, Xatom,
      {$IFDEF LCLgtk} GTKProc, {$ENDIF}
      {$IFDEF LCLgtk2} Gtk2Proc, gdk2x, Gtk2Def, {$ENDIF}
     {$ENDIF}
     {$IFDEF DARWIN}
       MacOSAll, IniFiles, Process,
       {$IFDEF LCLcarbon} CarbonProc, {$ENDIF}
       {$IFDEF LCLcocoa} CocoaAll, {$ENDIF}
     {$ENDIF}
     CompatMsg, UCompat, Controls, Dialogs, SysUtils, Forms, Detach;

{$IFDEF LCLcocoa}
procedure DarwinLaunchEditor(FileName: string; LineNumber: Integer = -1);
begin
  { TODO }
end;
{$ENDIF}

{$IFDEF LCLcarbon} // Maybe IFDEF DARWIN
//function CheckRunning(AppSig: AnsiString): Boolean;
//var
//  Event: AppleEvent;
//  s: String;
//  FinderSig: OSType;
//begin
//  //for i := 1 to 4 do
//  //  Sig[i] := AppSig[5-i];
//  //if OSError(AEBuildAppleEvent(kAEMiscStandards, kAEActivate,
//  //                             typeApplSignature, @Sig[1], 4,
//  //                             kAutoGenerateReturnID, kAnyTransactionID, Event, nil,
//  //                             '&subj:''null''()' // &'subj':null(), &'csig':magn(65536)
//  //                             ), 'CheckRunning', 'AEBuildAppleEvent') then Exit;
//  FinderSig := SwapEndian(FourCharCode('MACS'));
//  s := '''----'':obj{''form'':enum(''ID  ''), ''want'':type(''appf''),''seld'':"'+AppSig+'", ''from'':''null''()}';
//  if OSError(AEBuildAppleEvent(kAEMiscStandards, kAEActivate, //kAECoreSuite, kAEGetData,
//                               typeApplSignature, @FinderSig, SizeOf(FinderSig),
//                               kAutoGenerateReturnID, kAnyTransactionID, Event, nil,
//                               PChar(s)
//                               ), 'CheckRunning', 'AEBuildAppleEvent') then Exit;
//  AESendMessage(Event, nil, kAEWaitReply, kAEDefaultTimeout);
//  AEDisposeDesc(Event);
//end;

function DarwinSlaveOpen(FileName: string; LineNumber: Integer):Boolean;
var
  fileCF: CFStringRef;
  URLref: CFURLRef;
  fileFS: FSRef;
  fileAlias: AliasHandle;
  Event: AppleEvent;
  appSig: OSType;
  status: OSStatus;
begin
  Result := False;
  CreateCFString(FileName, fileCF);
  URLref := CFURLCreateWithFileSystemPath(kCFAllocatorDefault, fileCF, kCFURLPOSIXPathStyle, True);
  if not CFURLGetFSRef(URLref, fileFS) then Exit;
  if OSError(FSNewAlias(nil, fileFS, fileAlias), 'DarwinSlaveOpen', 'FSNewAlias') then Exit;
  appSig := SwapEndian(FourCharCode('PGOs'));
  if LineNumber < 0 then begin
    if OSError(AEBuildAppleEvent(kCoreEventClass, kAEOpen,
                                 typeApplSignature, @appSig, SizeOf(appSig),
                                 kAutoGenerateReturnID, kAnyTransactionID, Event, nil,
                                 '''----'':[alis(@@)]', fileAlias
                                 ), 'DarwinSlaveOpen', 'AEBuildAppleEvent') then Exit;
  end else begin
    if OSError(AEBuildAppleEvent(kCoreEventClass, appSig,
                                 typeApplSignature, @appSig, SizeOf(appSig),
                                 kAutoGenerateReturnID, kAnyTransactionID, Event, nil,
                                 '''----'':[alis(@@), long(@)]', fileAlias, LineNumber
                                 ), 'DarwinSlaveOpen', 'AEBuildAppleEvent') then Exit;
  end;
  status := AESendMessage(Event, nil, kAENoReply, kAEDefaultTimeout);
  AEDisposeDesc(Event);
  case status of
    noErr: Result := True;
    procNotFound: { Need to start application first };
  else
    OSError(status, 'DarwinSlaveOpen', 'AESendMessage');
  end;
end;

function DarwinFindSlave:string;
var
  p: Integer;
begin
  Result := Application.ExeName;
  { Look in same directory }
  p := Pos('/Contents/MacOS/', Result);
  if p > 0 then begin
    Dec(p);
    while (p > 0) and (Result[p] <> '/') do
      Dec(p);
    Delete(Result,p+1,Length(Result));
    Result := Result + 'tabslave.app';
    if DirectoryExists(Result) then
      Exit;
    { Special case for normal compilation location }
    Dec(p);
    while (p > 0) and (Result[p] <> '/') do
      Dec(p);
    Result := Copy(Result, 1, p) + 'tabslave' + Copy(Result, p+1, Length(Result));
    if DirectoryExists(Result) then
      Exit;
  end;
  Result := 'tabslave.app';
end;

procedure DarwinLaunchEditor(FileName: string; LineNumber: Integer = -1);
var
  InIFile: TCustomIniFile;
  s: AnsiString;
  args: array [0..1] of string;
begin
  FileName := ExpandFileName(FileName);
  IniFile := GetIniFile('Slave', False);
  try
    s := InIFile.ReadString('Editor', 'Command', '');
    if s <> '' then begin
      if not FileExists(s) then
        raise EReportException.Create('Script file '+s+' not found');
      if LineNumber <= 0 then
        RunCommand(s, FileName, s)
      else begin
        args[0] := FileName;
        args[1] := IntToStr(LineNumber);
        RunCommand(s, args, s);
      end;
      if s <> '' then
        raise EReportException.Create('Script reports:'+s);
    end else begin
      { Run slave editor }
      { First - is it running? }
      if not DarwinSlaveOpen(FileName, LineNumber) then begin
        { Not running, so try and start it }
        RunCommand('Open -a '+DarwinFindSlave, s);
        DarwinSlaveOpen(FileName, LineNumber);
      end;
    end;
  finally
    IniFile.Free;
  end;
end;

procedure DarwinSaveAll;
var
  appSig: OSType;
  Event: AppleEvent;
  InIFile: TCustomIniFile;
  s: AnsiString;
begin
  IniFile := GetIniFile('Slave', False);
  try
    s := InIFile.ReadString('Editor', 'SaveAll', '');
    if s <> '' then begin
      if not FileExists(s) then
        raise EReportException.Create('Save All Script file '+s+' not found');
      RunCommand(s, s);
      if s <> '' then
        raise EReportException.Create('Save All Script reports:'+s);
    end else begin
      appSig := SwapEndian(FourCharCode('PGOs'));
      if OSError(AEBuildAppleEvent(kAEMiscStandards, SwapEndian(FourCharCode('PGOa')),
                                   typeApplSignature, @appSig, SizeOf(appSig),
                                   kAutoGenerateReturnID, kAnyTransactionID, Event, nil,
                                   '&subj:''null''()'
                                   ), 'DarwinSaveAll', 'AEBuildAppleEvent') then Exit;
      AESendMessage(Event, nil, kAEWaitReply, kAEDefaultTimeout);
      AEDisposeDesc(Event);
    end;
  finally
    InIFile.Free;
  end;
end;

{$ENDIF}

function FindSlave:String;
var
  s: string;
begin
  Result := 'tabslave'{$IFDEF UNICODE}+'U'{$ENDIF}{$IFDEF MSWINDOWS} + '.exe'{$ENDIF};
  if not FileExists(Result) then begin
    s := ExtractFilePath(Application.ExeName);
    if FileExists(s + Result) then
      Result := ExpandFileName(s + Result)
    else begin
      s := s + '..' + PathDelim + 'tabslave' + PathDelim;
      if FileExists(s + Result) then
        Result := ExpandFileName(s + Result)
    end;
  end else
    Result := ExpandFileName(Result);
end;

type

  TWindowEnumerator = class
  protected
    ResultHWND: TWnd;
    function CallBack(hWnd: TWnd): Boolean; overload; virtual;
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; overload; virtual; abstract;
  public
    function DoEnumerate: TWnd;
  end;

  TWindowFinder = class(TWindowEnumerator)
  protected
    FText: string;
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; override;
  public
    function FindWindow(Text: string): TWnd;
  end;

  TExWindowFinder = class(TWindowFinder)
  private
    {$IFDEF MSWINDOWS}
    Data: TCopyDataStruct;
    WName: AnsiString;
    {$ENDIF}
  protected
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; override;
  public
    TabSlaveWindow: TWnd;
    function FindWindow(Text: string): TWnd;
  end;

  TSaveAllWindows = class(TWindowEnumerator)
  public
    First: Boolean;
  protected
    function CallBack(hWnd: TWnd): Boolean; override;
    {$IFDEF XBASED}
    function StartSave: TWnd;
    function FinishSave: TWnd;
    {$ENDIF}
  end;

procedure BringWindowToFront(hWnd: TWnd);
begin
  {$IFDEF MSWINDOWS}
    {$IFDEF VER80}
      SetActiveWindow(hWnd);
    {$ELSE}
      SetForegroundWindow(hWnd);
    {$ENDIF}
    ShowWindow(hWnd, SW_RESTORE);
  {$ENDIF}
  {$IFDEF XBASED}
    XRaiseWindow(X11Display, hWnd);
    XSetInputFocus(X11Display, hWnd, RevertToNone, CurrentTime);
    XSync(X11Display, False);
  {$ENDIF}
end;

{$IFDEF XBASED}
procedure PostMessage(Window: TWnd; Msg: TWindowMessage; wParam, lParam: Integer);
var
  Event: TXClientMessageEvent;
begin
  Event.display := X11Display;
  Event._type := ClientMessage;
  Event.message_type := Msg;
  Event.format := 32;
  Event.data.l[0] := wParam;
  Event.data.l[1] := lParam;
  Event.window := Window;
  XSendEvent(X11Display, Window, False, NoEventMask, Pointer(@Event));
end;
{$ENDIF}

{$IFDEF DARWIN}
procedure PostMessage(Window: TWnd; Msg: TWindowMessage; wParam, lParam: Integer);
begin
  { TODO -cLCLcarbon : Implement }
end;
{$ENDIF}

var
  WM_GOTO_LINE: TWindowMessage = 0;
  WM_PGOPHER_ASK: TWindowMessage = 0;

procedure LaunchEditor(FileName: string; LineNumber: Integer);
var
  Window: TWnd;
  Handle: Integer;
  WName: AnsiString;
  {$IFDEF MSWINDOWS}
  Data: TCopyDataStruct;
  r: {$IFNDEF FPC}{$IF CompilerVersion >= 23}UInt64{$ELSE}DWORD{$IFEND}{$ELSE}LongWord{$ENDIF}; { Declared length may be wrong }
  {$ENDIF}
  procedure SendLineNumber;
  begin
    PostMessage(Window, WM_GOTO_LINE, 0, LineNumber);
  end;
  function FindWindow:Boolean;
  begin
    with TExWindowFinder.Create do begin
      Window := FindWindow(FileName);
      Result := Window <> TWnd(0);
      if not Result then
        Window := TabSlaveWindow;
      Free;
    end;
  end;
begin
  if FileName = '' then begin
    ShowMessage('Please choose a name for the file before editing it');
    Exit;
  end;
  FileName := ExpandFileName(FileName);
  {$IFDEF DARWIN}
  DarwinLaunchEditor(FileName, LineNumber);
  {$ELSE}
  if WM_GOTO_LINE = 0 then
    WM_GOTO_LINE := RegisterMessage('WM_GOTO_LINE');
  DoDirSeparators(FileName);
  if FindWindow then begin
    BringWindowToFront(Window);
    if LineNumber > 0 then
      SendLineNumber;
    Exit;
  end;
  if not FileExists(FileName) then
    if MessageDlg(Format('File %s does not exist. Create it?', [FileName]),
                  mtConfirmation, [mbYes, mbNo], 0) = mrYes then begin
      Handle := FileCreate(FileName);
      if Handle >= 0 then
        FileClose(Handle)
      else
        RaiseLastOSError;
    end else
      Exit;
  if Window <> TWnd(0) then begin
    { Send file to TabSlave window for opening }
    {$IFDEF MSWINDOWS}
    WName := FileName;
    Data.dwData := WM_GOTO_LINE;
    Data.cbData := Length(WName)+1;
    Data.lpData := PAnsiChar(WName);
    // Note that PostMessage fails here (implicit use of pointers?)
    SendMessageTimeout(Window, WM_COPYDATA, Application.MainFormHandle, NativeInt(@Data), SMTO_ABORTIFHUNG or SMTO_BLOCK, 2000, {$IFNDEF FPC}{$IF CompilerVersion >= 23}@{$IFEND}{$ENDIF}r);
    {$ENDIF}
    {$IFDEF XBASED}
    XChangeProperty(X11Display, RootWindow, WM_PGOPHER_ASK, XA_STRING, 8, PropModeReplace, @FileName[1], Length(FileName));
    XFlush(X11Display);
    PostMessage(Window, WM_PGOPHER_ASK, 0, 1); { Now ask for load }
    {$ENDIF}
  end else begin
    {$IFDEF MSWINDOWS}
      { Try default application associated with file }
      if ShellExecute(0, nil, PChar(ExtractFileName(FileName)), nil, PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
        { else try slave }
        if ShellExecute(0, nil, PChar(FindSlave), PChar(ExtractFileName(FileName)), PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
          { use notepad as a last resort }
          if WinExec(PAnsiChar(AnsiString('notepad ' + FileName)), SW_SHOW) < 32 then
            RaiseLastOSError;
    {$ENDIF}
    {$IFDEF UNIX}
    if not Detach1(FindSlave, FileName) then
      if not Detach1('gedit', FileName) then
        if not Detach1('kedit', FileName) then
          if not Detach1('xedit', FileName) then
            RaiseLastOSError;
    {$ENDIF}
  end;
  if LineNumber > 0 then begin
    Sleep(0); { Hang about waiting for the window to turn up }
    if not FindWindow then begin
      Sleep(100);
      if not FindWindow then begin
        Sleep(1000);
        if not FindWindow then
          Exit;
      end;
    end;
    SendLineNumber;
  end;
  {$ENDIF}
end;

{ TWindowEnumerator }

{$IFDEF MSWINDOWS}
function EnumWindowsProc(hWnd: TWnd; lParam: LPARAM): BOOL; { export; } stdcall;
begin
  Result := TWindowEnumerator(lParam).CallBack(hWnd);
end;
{$ENDIF}

{$IFDEF XBASED}
{ Duplicate logic of XmuClientWindow }
function ClientWindow(Window: TWindow): TWindow;
var
  WM_STATE: TAtom;
  function GetProperty(AWindow: TWindow): Boolean;
  var
    AType: TAtom;
    Format: Integer;
    nItems, After: DWord;
    data: PChar;
  begin
    AType := None;
    XGetWindowProperty(X11Display, AWindow, WM_STATE, 0, 0, Ord(False), AnyPropertyType,
                       @AType, @Format, @nItems, @After, Pointer(@data));
    Result := Atype <> 0;
  end;
  function SearchChildren(AWindow: TWindow): TWindow;
  var
    Root, Parent: TWindow;
    Children: PWindow;
    nChildren: Cardinal;
    i: Integer;
  begin
    Result := 0;
    if (XQueryTree(X11Display, AWindow, @Root, @Parent, @Children, @nChildren) = 0) or not Assigned(Children) then
      Exit;
    try
      for i := 1 to nChildren do
        if GetProperty(Children[i-1]) then begin
          Result := Children[i-1];
          Exit;
        end;
      for i := 1 to nChildren do begin
        Result := SearchChildren(Children[i-1]);
        if Result <> 0 then
          Exit;
        end;
     finally
       XFree(children);
     end;
  end;
begin
  Result := Window; { Default is current window }
  WM_STATE := XInternAtom(X11Display, 'WM_STATE', True);
  if WM_STATE = 0 then Exit;
  if GetProperty(Window) then Exit; { On current Window }
  Result := SearchChildren(Window);
  if Result = 0 then
    Result := Window;
end;
{$ENDIF}

function TWindowEnumerator.CallBack(hWnd: TWnd): Boolean;
var
  buffer: string;
  {$IFDEF XBASED}
  Name: PChar;
  {$ELSE}
  n: Integer;
  {$ENDIF}
begin
  Result := True; { Continue enumeration by default }
  {$IFDEF MSWINDOWS}
    if not IsWindowVisible(hWnd) then Exit;
    n := GetWindowTextLength(hWnd);
    if n <= 0 then Exit;
    SetLength(buffer, n);
    n := GetWindowText(hWnd, PChar(buffer), Length(buffer)+1);
    if n <= 0 then Exit;
    SetLength(buffer, n);
  {$ENDIF}
  {$IFDEF XBASED}
  if (XFetchName(X11Display, hWnd, @Name) <> 0) and Assigned(Name) then begin
    buffer := Name;
    XFree(Name);
  end else
    Exit; { No name - skip }
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
  Result := CallBack(hWnd, buffer);
end;

function TWindowEnumerator.DoEnumerate: TWnd;
{$IFDEF XBASED}
var
  Root, Parent: TWindow;
  Children: PWindow;
  nChildren: Cardinal;
  i: Integer;
{$ENDIF}
begin
  ResultHWND := TWnd(0);
  {$IFDEF MSWINDOWS}
    EnumWindows(@EnumWindowsProc, LPARAM(self));
  {$ENDIF}
  {$IFDEF XBASED}
    // XSync(X11Display, False);
    if (XQueryTree(X11Display, RootWindow, @Root, @Parent, @Children, @nChildren) <> 0) and Assigned(Children) then
      try
        for i := 1 to nChildren do
          if not CallBack(ClientWindow(Children[i-1])) then
            Break;
      finally
        XFree(Children);
      end;
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
  Result := ResultHWND;
end;

{ TWindowFinder }

function TWindowFinder.CallBack(hWnd: TWnd; WindowName: string): Boolean;
var
  p: Integer;
  Tail: string;
begin
  Result := True;
  p := Pos(FText, AnsiLowerCase(WindowName));
  if p = 0 then Exit;
  if p < Length(WindowName) - Length(FText) + 1 then begin
    { Not at the end of the string - look for "file - application name" }
    Tail := Copy(WindowName, p+Length(FText), 3);
    if Tail <> ' - ' then
      Exit;
  end;
  ResultHWND := hWnd;
  Result := False;
end;

function TWindowFinder.FindWindow(Text: string): TWnd;
begin
  FText := AnsiLowerCase(ExtractFileName(Text));
  Result := DoEnumerate;
end;

{ TExWindowFinder }

function TExWindowFinder.CallBack(hWnd: TWnd; WindowName: string): Boolean;
{$IFDEF MSWINDOWS}
  var
    r: {$IFNDEF FPC}{$IF CompilerVersion >= 23}UInt64{$ELSE}DWORD{$IFEND}{$ELSE}LongWord{$ENDIF};
{$ENDIF}
{$IFDEF XBASED}
  var
    Prop: Integer;
{$ENDIF}
begin
  Result := True;
  {$IFDEF MSWINDOWS}
    r := 0; { Fix for 32/64 bit issues }
    if SendMessageTimeout(hWnd, WM_PGOPHER_ASK, 0, 0, SMTO_ABORTIFHUNG or SMTO_BLOCK, 1000, {$IFNDEF FPC}{$IF CompilerVersion >= 23}@{$IFEND}{$ENDIF}r) = 0 then
      r := 0; { Call failed or timed out }
    if r = 272727 then begin
      if SendMessageTimeout(hWnd, WM_COPYDATA, 0, NativeInt(@Data), SMTO_ABORTIFHUNG or SMTO_BLOCK, 1000, {$IFNDEF FPC}{$IF CompilerVersion >= 23}@{$IFEND}{$ENDIF}r) = 0 then
        r := 0; { Call failed or timed out }
      if r = 727272 then begin
        ResultHWND := hWnd;
        Result := False;
      end else if r = 272727 then
        if TabSlaveWindow = 0 then
          TabSlaveWindow := hWnd;
    end else
  {$ENDIF}
  {$IFDEF XBASED}
    Prop := GetWindowInteger(hWnd, WM_PGOPHER_ASK);
    if Prop <> 0 then begin
      if FText <> '' then begin
        TabSlaveWindow := hWnd; { First Cycle }
        PostMessage(hWnd, WM_PGOPHER_ASK, 0, 0); { Query for file }
      end else if Prop = 2 then begin
        { Found }
        ResultHWND := hWnd;
        Result := False;
      end
    end else if FText <> '' then { Older style only on first cycle }
  {$ENDIF}
      Result := inherited CallBack(hWnd, WindowName);
end;

function TExWindowFinder.FindWindow(Text: string): TWnd;
begin
  TabSlaveWindow := TWnd(0);
  Result := TWnd(0);
  if WM_PGOPHER_ASK = 0 then
    WM_PGOPHER_ASK := RegisterMessage('WM_PGOPHER_ASK');
  FText := Text;
  {$IFDEF MSWINDOWS}
    WName := FText;
    Data.dwData := WM_PGOPHER_ASK;
    Data.cbData := Length(WName)+1;
    Data.lpData := PAnsiChar(WName);
    Result := DoEnumerate;
  {$ENDIF}
  FText := AnsiLowerCase(ExtractFileName(Text));
  {$IFDEF XBASED}
    XChangeProperty(X11Display, RootWindow, WM_PGOPHER_ASK, XA_STRING, 8, PropModeReplace, @Text[1], Length(Text));
    XFlush(X11Display);
    Result := DoEnumerate;
    if (Result = 0) and (TabSlaveWindow <> 0) then begin
      FText := ''; { Go round again and collect results }
      XFlush(X11Display);
      Sleep(10);
      Result := DoEnumerate;
    end;
  {$ENDIF}
end;

{ TSaveAllWindows }

var
  WM_Save_All: TWindowMessage = 0;

function TSaveAllWindows.CallBack(hWnd: TWnd): Boolean;
var
{$IFDEF MSWINDOWS}
  r: Cardinal;
{$ENDIF}
{$IFDEF XBASED}
  Prop: Integer;
{$ENDIF}
{$IFDEF DARWIN}
  i: Integer;
{$ENDIF}
begin
  Result := True;
  {$IFDEF MSWINDOWS}
    {$IFDEF OLDER}
    r := SendMessage(hWnd, WM_Save_All, 0, 0);
    {$ELSE}
    if SendMessageTimeout(hWnd, WM_Save_All, 0, 0, SMTO_ABORTIFHUNG or SMTO_BLOCK, 1000, {$IFNDEF FPC}{$IF CompilerVersion >= 23}@{$IFEND}{$ENDIF}r) = 0 then
      r := 0; { Call failed or timed out }
    {$ENDIF}
    if r = 272727 then begin
      ResultHWND := hWnd;
      Result := False;
    end;
  {$ENDIF}
  {$IFDEF XBASED}
    Prop := GetWindowInteger(hWnd, WM_Save_All);
    { Does window have WM_Save_All property set? }
    if Prop <> 0 then begin
      ResultHWND := hWnd;
      if First then
        PostMessage(hWnd, WM_Save_All, MainWindow, 0)
      else
        Result := False;
    end;
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
end;

{$IFDEF XBASED}
function TSaveAllWindows.StartSave: TWnd;
begin
  First := True;
  Result := DoEnumerate;
  XFlush(X11Display);
end;

function TSaveAllWindows.FinishSave: TWnd;
var
  Start: TDateTime;
begin
  First := False;
  Start := Now;
  repeat
    XFlush(X11Display);
    Application.ProcessMessages;
    Result := DoEnumerate;
  until (Result = 0) or (Now - Start > 2/SecsPerDay);
end;
{$ENDIF}

function SaveAllWindows: Boolean;
var
  Window: TWnd;
begin
  Result := True; { Assume OK }
  if WM_Save_All = 0 then
    WM_Save_All := RegisterMessage('WM_Save_All');
  with TSaveAllWindows.Create do
    try
      {$IFDEF XBASED}
      Window := StartSave;
      if Window <> 0 then
        Window := FinishSave;
      {$ENDIF}
      {$IFDEF MSWINDOWS}
      Window := DoEnumerate;
      {$ENDIF}
      {$IFDEF LCLcarbon}
      Window := nil;
      DarwinSaveAll;
      {$ENDIF}
      if Window <> TWnd(0) then begin
        BringWindowToFront(Window);
        Result := False;
      end;
    finally
      Free;
    end;
end;

procedure BroadcastNotification(WindowMessage: TWindowMessage);
begin
  {$IFDEF MSWINDOWS}
  PostMessage(HWND_BROADCAST, WindowMessage, 0, 0);
  {$ENDIF}
  { TODO: UNIX }
  { TODO -cLCLcarbon : Implement }
end;

function BroadcastName(const Name: string):TWindowMessage;
begin
  Result := RegisterMessage(Name);
  BroadcastNotification(Result);
end;

function RegisterMessage(const Name: string):TWindowMessage;
begin
  Result :=
  {$IFDEF MSWINDOWS}
    RegisterWindowMessage(PChar(Name));
  {$ENDIF}
  {$IFDEF XBASED}
    XInternAtom(X11Display, PChar(Name), False);
  {$ENDIF}
  {$IFDEF DARWIN}
    0;
  {$ENDIF}
end;

end.
