unit TextWin;

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

interface

uses
  Messages, Graphics, Controls, Forms, Dialogs, Menus, ExtCtrls, ActnList, ImgList, ComCtrls,
  SysUtils, Classes, LogBase, ExActn, FxFntDlg, LogControl,
  EditLog, LogActn, PrintPre, TextPre, ScrlCtrl;

type

  { TTextChild }

  TTextChild = class(TLogBaseChild)
    ExPaste: TExPaste;
    ExRedo: TExRedo;
    ExUndo: TExUndo;
    N2: TMenuItem;
    Undo: TMenuItem;
    Redo: TMenuItem;
    Replace: TMenuItem;
    Special: TMenuItem;
    SetBookmark: TMenuItem;
    ExchangewithBookmark: TMenuItem;
    GotoBookmark: TMenuItem;
    N4: TMenuItem;
    Increment: TMenuItem;
    Decrement: TMenuItem;
    N5: TMenuItem;
    SortUp: TMenuItem;
    SortDown: TMenuItem;
    Memo: TEditLog;
    Options: TMenuItem;
    RecordKeyStrokes: TMenuItem;
    PlayMacro: TMenuItem;
    FontItem: TMenuItem;
    FixedFontDialog: TFixedFontDialog;
    StatusBar: TStatusBar;
    FileItem: TMenuItem;
    FileNew: TMenuItem;
    OpenDialog: TOpenDialog;
    FileOpen: TMenuItem;
    Save: TMenuItem;
    SaveAs: TMenuItem;
    N6: TMenuItem;
    FileExit: TMenuItem;
    EditFileAtCursor: TMenuItem;
    ReloadItem: TMenuItem;
    CloseItem: TMenuItem;
    SaveDialog: TSaveDialog;
    Paste: TMenuItem;
    AlienNewline: TMenuItem;
    KeepBackup: TMenuItem;
    StartFileatCursor: TMenuItem;
    RemoveTabs: TMenuItem;
    FileSeparatorItem: TMenuItem;
    Print: TMenuItem;
    PrintPreview: TMenuItem;
    TextPrintPreview: TTextPrintPreview;
    TextDecrement: TTextDecrement;
    TextExchangeWithBookMark: TTextExchangeWithBookMark;
    TextGoToBookMark: TTextGoToBookMark;
    TextIncrement: TTextIncrement;
    TextPlayBackMacro: TTextPlayBackMacro;
    TextRecordMacro: TTextRecordMacro;
    TextSetBookMark: TTextSetBookMark;
    TextSortDown: TTextSortDown;
    TextSortUp: TTextSortUp;
    TextReplace: TTextReplace;
    TextRemoveTabs: TTextRemoveTabs;
    PreserveNewline: TMenuItem;
    SetTabSpacing: TMenuItem;
    procedure FontItemClick(Sender: TObject);
    procedure StatusBarDblClick(Sender: TObject);
    procedure StatusBarMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
    procedure NewClick(Sender: TObject);
    procedure FileOpenClick(Sender: TObject);
    procedure SaveClick(Sender: TObject);
    procedure SaveAsClick(Sender: TObject);
    procedure FileExitClick(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure EditFileAtCursorClick(Sender: TObject);
    procedure ReloadItemClick(Sender: TObject);
    procedure CloseItemClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FixedFontDialogChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure AlienNewlineClick(Sender: TObject);
    procedure KeepBackupClick(Sender: TObject);
    procedure StartFileatCursorClick(Sender: TObject);
    procedure PrintClick(Sender: TObject);
    procedure PrintPreviewClick(Sender: TObject);
    procedure PreserveNewlineClick(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    FLastMouseX: Integer;
    FFormUsed: Boolean;
    function GetFormUsed: Boolean;
    {$IFDEF VCL_}
    procedure WMDropFiles(var Message: TMessage); message WM_DROPFILES;
    {$ENDIF}
    procedure FormDropFiles(Sender: TObject; const FileNames: array of String);
  protected
    procedure UpdateActions; override;
  public
    LastFileAge: Integer;
    constructor Create(AOwner: TComponent); override;
    procedure ResetModified;
    function LoadFile(FileName: string; OldFileName: string = ''): Boolean;
    procedure SetDefaultExt(Ext: string);
    procedure SetFileName(FileName: string; OldFileName: string = ''); virtual;
    function ExecuteAction(AnAction: TBasicAction): Boolean; override;
    procedure SaveToFile(FileName: string);
    property FormUsed: Boolean read GetFormUsed write FFormUsed;
  end;

var
  TextChild: TTextChild;
  DefaultTextChildFilter: string;

function ChildWithName(const ChildName: string):TCustomForm;
procedure SaveAllTextChildren;
function AnyTextChildModified: Boolean;
function ActiveTextChild: TTextChild;
function ActiveOrOnlyTextChild: TTextChild;
function AnyTextChild: TTextChild;
procedure AskAndOpenFileInTextChild;
procedure DropFilesToTextChild(const Files: array of string);
function LoadFileInto(FileName: string; TextChild: TTextChild): TTextChild;

implementation

{$IFNDEF LCL_}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

uses ResActn,RTLConsts,
     {$IFDEF MSWINDOWS} Windows, ShellAPI, { Consts, } {$ENDIF}
     WCompat, UCompat, Detach, CharType, ImageMod, LongName;

{$IFNDEF LCL_}
type TOpenCustomForm = class(TCustomForm);
{$ENDIF}

function ChildWithName(const ChildName: string):TCustomForm;
var
  i: Integer;
begin
  Result := nil;
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  if TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm then
    with TOpenCustomForm(Application.MainForm) do
      for i := 0 to MDIChildCount-1 do
        if AnsiCompareText(ChildName, MDIChildren[i].Caption) = 0 then begin
          Result := MDIChildren[i];
          Break;
        end;
  {$ENDIF}
end;

procedure SaveAllTextChildren;
var
  i: Integer;
begin
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  if TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm then
    with TOpenCustomForm(Application.MainForm) do
      for i := 0 to MDIChildCount-1 do
        if MDIChildren[i] is TTextChild then
          with MDIChildren[i] as TTextChild do
            if Memo.Modified then
              SaveClick(nil);
  {$ENDIF}
end;

function AnyTextChildModified: Boolean;
var
  i: Integer;
begin
  Result := False;
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  if TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm then
    with TOpenCustomForm(Application.MainForm) do
      for i := 0 to MDIChildCount-1 do
        if MDIChildren[i] is TTextChild then
          with MDIChildren[i] as TTextChild do
            if Memo.Modified then begin
              Result := True;
              Exit;
            end;
  {$ENDIF}
end;

function ActiveTextChild: TTextChild;
begin
  Result := nil;
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  with TOpenCustomForm(Application.MainForm) do
    if (FormStyle = fsMDIForm) and (ActiveMDIChild is TTextChild) then
      Result := ActiveMDIChild as TTextChild
    else
  {$ENDIF}
      if Application.MainForm is TTextChild then
        Result := Application.MainForm as TTextChild;
end;

function ActiveOrOnlyTextChild: TTextChild;
var
  i: Integer;
begin
  Result := ActiveTextChild;
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  if (Result = nil) and (TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm) then
    with TOpenCustomForm(Application.MainForm) do
      for i := 0 to MDIChildCount-1 do
        if MDIChildren[i] is TTextChild then
          if Result <> nil then begin
            Result := nil; { More than one TextChild window, so return none }
            Break;
          end else
            Result := MDIChildren[i] as TTextChild;
  {$ENDIF}
end;


function AnyTextChild: TTextChild;
var
  i: Integer;
begin
  Result := ActiveTextChild;
  {$IFNDEF LCL_} { TODO: MDI under LCL? }
  if (Result = nil) and (TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm) then
    with TOpenCustomForm(Application.MainForm) do
      for i := 0 to MDIChildCount-1 do
        if MDIChildren[i] is TTextChild then begin
          Result := MDIChildren[i] as TTextChild;
          Break;
        end;
  {$ENDIF}
end;

procedure AskAndOpenFileInTextChild;
var
  TextChild: TTextChild;
begin
  TextChild := AnyTextChild;
  if TextChild <> nil then
    TextChild.FileOpenClick(nil)
  else begin
    TextChild := TTextChild.Create(Application);
    TextChild.FileOpenClick(nil);
    if TextChild.OpenDialog.FileName <> '' then
      TextChild.Show { Probably Redundant }
    else
      TextChild.Free;
  end
end;

{$IFDEF VCL_}
procedure TTextChild.WMDropFiles(var Message: TMessage);
begin
  HandleWMDropFiles(Message, FormDropFiles);
end;
{$ENDIF}

procedure TTextChild.FormDropFiles(Sender: TObject; const FileNames: array of String);
begin
  DropFilesToTextChild(FileNames);
end;

procedure DropFilesToTextChild(const Files: array of string);
var
  n, i: Integer;
  CanClose: Boolean;
  TextChild: TTextChild;
begin
  for n := 0 to High(Files) do begin
    TextChild := AnyTextChild;
    if (TextChild <> nil) and (TextChild.FormStyle <> fsMDIChild) then
      with TextChild do begin
        CanClose := True;
        FormCloseQuery(TextChild, CanClose);
        if CanClose then begin
          LoadFile(Files[n]);
          for i := n+1 to High(Files) do
            Detach1(ApplicationExecutable, Files[i]);
        end;
        Exit;
      end
    else
      LoadFileInto(Files[n], TextChild);
  end;
end;

function LoadFileInto(FileName: string; TextChild: TTextChild): TTextChild;
{ Loads file, possibly into given TextChild;
  return True TextChild actually used }
var
  OtherChild: TCustomForm;
  r: Integer;
begin
  Result := nil;
  OtherChild := ChildWithName(FileName);
  if OtherChild <> nil then begin
    {$IFDEF MSWINDOWS}
    r := Application.MessageBox(
           AddressOfFirstChar(Format('File %s is already open. Do you want to open a new window?'#0, [FileName])),
           'Warning', mb_YesNoCancel or mb_IconQuestion or mb_DefButton1);
    {$ELSE}
    r := MessageDlg(
           Format('File %s is already open. Do you want to open a new window?', [FileName]),
           mtWarning, mbYesNoCancel, 0);
    {$ENDIF}
    case r of
      mrCancel { IDCANCEL } : Exit;
      mrNo { IDNO }: with OtherChild do begin
              {$IFNDEF LCL_} { TODO: MDI under LCL? }
              if TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm then
                {$IFDEF MSWINDOWS}
                SendMessage(Application.MainForm.ClientHandle, WM_MDIACTIVATE, Handle, 0)
                {$ENDIF}
              else
              {$ENDIF}
                SetFocus;
              if OtherChild is TTextChild then
                Result := OtherChild as TTextChild;
              Exit;
            end;
    end;
  end;
  if (TextChild <> nil) and not TextChild.FormUsed then begin
    if TextChild.LoadFile(FileName) then begin
      TextChild.SetFocus;
      Result := TextChild;
    end
  end else begin
    {$IFNDEF LCL_} { TODO: MDI under LCL? }
    if TOpenCustomForm(Application.MainForm).FormStyle = fsMDIForm then begin
      Result := TTextChild.Create(Application);
      if not Result.LoadFile(FileName) then
        FreeAndNil(Result);
    end else 
    {$ENDIF}
    begin
      Result := nil; { Lauch another application }
      Detach1(ApplicationExecutable, FileName);
    end;
  end;
end;

procedure TTextChild.FontItemClick(Sender: TObject);
begin
  FixedFontDialog.FixedFont := Memo.FixedFont;
  if FixedFontDialog.Execute then begin
    Memo.FixedFont := FixedFontDialog.FixedFont;
    FixedFontDialog.FixedFont.CopyTo(TextPrintPreview.Font);
  end;
end;

procedure TTextChild.StatusBarDblClick(Sender: TObject);
var
  x: Integer;
begin
  x := FLastMouseX - StatusBar.Panels[0].Width - StatusBar.Panels[1].Width;
  if x > 0 then begin
    Dec(x, StatusBar.Panels[2].Width);
    if x < 0 then
      Memo.Modified := not Memo.Modified
    else if x < StatusBar.Panels[3].Width then
      TextRecordMacro.Execute;
  end;
end;

procedure TTextChild.StatusBarMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FLastMouseX := X;
end;

type TDeTabStringList = class(TStrings)
  private
    FTextChild: TTextChild;
  protected
    function Get(Index: Integer): string; override;
    function GetCount: Integer; override;
  public
    constructor Create(TextChild: TTextChild);
    procedure Clear; override;
    procedure Delete(Index: Integer); override;
    procedure Insert(Index: Integer; const S: string); override;
  end;

procedure TTextChild.FormCreate(Sender: TObject);
begin
  inherited;
  if DefaultTextChildFilter = '' then
    DefaultTextChildFilter := OpenDialog.Filter
  else
    OpenDialog.Filter := DefaultTextChildFilter;
  SetFileName('');
  FileItem.GroupIndex := 0; { Not loaded properly }
  TextPrintPreview.Lines := TDeTabStringList.Create(Self);
  DragAcceptFiles(Self, True);
  {$IFNDEF VCL_}
  OnDropFiles := @FormDropFiles;
  {$ENDIF}
end;

procedure TTextChild.SetDefaultExt(Ext:string);
var
  i: Integer;
  e, p, ee: PChar;
  Filter: string;
begin
  if Length(Ext) = 0 then Exit;
  if Ext[1] <> '.' then Ext := Concat('.', Ext);
  Ext := Concat('*', Ext);
  OpenDialog.DefaultExt := Copy(Ext, 3, Length(Ext));
  Filter := AnsiLowerCase(OpenDialog.Filter);
  p := @Filter[1];
  i := 1;
  repeat
    p := StrScan(p, {$IFDEF MSWINDOWS}'|'{$ENDIF}{$IFDEF UNIX}'('{$ENDIF});
    if p <> nil then begin
      e := p + 1;
      p := StrScan(e, {$IFDEF MSWINDOWS}'|'{$ENDIF}{$IFDEF UNIX}')'{$ENDIF});
      if p <> nil then begin
        p^ := #0;
        Inc(p);
      end;
      repeat
        ee := StrScan(e, ';');
        if ee <> nil then
          ee^ := #0;
        if StrIComp(e, @Ext[1]) = 0 then begin
          OpenDialog.FilterIndex := i;
          Exit;
        end;
        e := ee + 1;
      until ee = nil;
    end;
    Inc(i);
  until p = nil;
  {$IFDEF UNIX}
  OpenDialog.DefaultExt := ''; { This only works properly if it is in the filter list }
  {$ENDIF}
  OpenDialog.FilterIndex := i-1;
end;

procedure TTextChild.SetFileName(FileName:string; OldFileName: string);
var
  i: Integer;
  NewCaption: string;
begin
  OpenDialog.Filename := FileName;
  if Length(FileName) = 0 then begin
    NewCaption := '[Untitled]';
    i := 0;
    while (FormStyle = fsMDIChild) and
          (ChildWithName(NewCaption) <> nil ) do begin
      Inc(i);
      NewCaption := Format('[Untitled %d]', [i]);
    end;
    Caption := NewCaption;
    LastFileAge := -1;
  end else begin
    Caption := GetLongName(FileName);
  end;
  SetDefaultExt(ExtractFileExt(Caption));
  SetCurrentDir(PChar(ExtractFilePath(FileName)));
end;

function TTextChild.LoadFile(FileName:string; OldFileName: string): Boolean;
var
  Stream: TFileStream;
  Buffer: AnsiString;
  nSep: Integer;
begin
  Result := False;
  if (Length(FileName) > 0) and ( (FileName[Length(FileName)] = PathDelim) or DirectoryExists(FileName) ) then begin
    ShowMessage(name + ' is a directory');
    Exit;
  end;
  if not FileExists(FileName) then begin
    {$IFDEF MSWINDOWS}
    if IDNO = Application.MessageBox(
           AddressOfFirstChar(Format('File %s does not exist. Create?'#0, [FileName])),
           'Warning', mb_YesNo or mb_IconQuestion or mb_DefButton1) then
    {$ENDIF}
    {$IFDEF UNIX}
    if mrNo = MessageDlg(Format('File %s does not exist. Create?', [FileName]),
                         mtWarning, mbYesNo, 0) then
    {$ENDIF}
      Exit;
    Memo.Clear;
    LastFileAge := -1;
  end else begin
    Stream := nil;
    try
      Stream := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
      if PreserveNewline.Checked then begin
        { Guess line end from first few bytes }
        SetLength(Buffer, 255);
        SetLength(Buffer, Stream.Read(Buffer[1], Length(Buffer)));
        nSep := 0;
        if Pos(#10, Buffer) > 0 then Inc(nSep);
        if Pos(#13, Buffer) > 0 then Inc(nSep);
        AlienNewline.Checked := nSep <> NewLineLength;
        Stream.Position := 0;
      end;
      Memo.RecordUndo := False;
      Memo.Lines.LoadFromStream(Stream);
    finally
      Memo.RecordUndo := True;
      Stream.Free;
    end;
  end;
  Result := True;
  SetFileName(ExpandFileName(FileName), OldFileName);
  ResetModified;
  Memo.StartOfText(False);
  FFormUsed := True;
end;

procedure TTextChild.FileOpenClick(Sender: TObject);
var
  CanClose: Boolean;
  OldFileName: string;
  i: Integer;
  SelfUsed: Boolean;
begin
  OldFileName := OpenDialog.FileName;
  SelfUsed := False;
  if FormStyle = fsMDIChild then
    try
      if OpenDialog.Execute then
        for i := 0 to OpenDialog.Files.Count-1 do
          SelfUsed := (LoadFileInto(OpenDialog.Files[i], Self) = self) or SelfUsed;
    finally
      if not SelfUsed then OpenDialog.FileName := OldFileName;
    end
  else begin
    CanClose := True;
    FormCloseQuery(Self, CanClose);
    if CanClose then
      with OpenDialog do
        if Execute then begin
          LoadFile(OpenDialog.Files[0],OldFileName);
          for i := 1 to OpenDialog.Files.Count-1 do begin
            StatusBar.Panels[4].Text := Concat('Opening ', OpenDialog.Files[i]);
            Detach1(ApplicationExecutable, OpenDialog.Files[i]);
          end;
        end;
  end;
end;

procedure TTextChild.SaveClick(Sender: TObject);
begin
  if OpenDialog.Filename <> '' then begin
    SaveToFile(OpenDialog.Filename);
    ResetModified;
    FFormUsed := True;
  end else
    SaveAsClick(Sender);
end;

procedure TTextChild.SaveAsClick(Sender: TObject);
begin
  with SaveDialog do begin
    Filter := OpenDialog.Filter;
    FilterIndex := OpenDialog.FilterIndex;
    DefaultExt := OpenDialog.DefaultExt;
  end;
  if SaveDialog.Execute then begin
    SaveToFile(SaveDialog.FileName);
    SetFileName(SaveDialog.FileName);
    ResetModified;
  end;
end;

procedure TTextChild.FileExitClick(Sender: TObject);
begin
  if (FormStyle = fsMDIChild) then
    Application.MainForm.Close
  else
    Close;
end;

procedure TTextChild.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  MsgResult: Word;
begin
  if Memo.Modified then begin
    MsgResult := MessageDlg(Format('File %s has been modified. Save file?',
    [Caption]), mtWarning, mbYesNoCancel, 0);
    CanClose := False;
    case MsgResult of
      mrYes:
        try
          SaveClick(Sender);
          CanClose := not Memo.Modified;
        except
          {Required to work around a kylix bug}
          Application.HandleException(self);
        end;
      mrNo: CanClose := True;
    end;
  end;
end;

procedure TTextChild.ResetModified;
begin
  Memo.Modified := False;
  Memo.ClearUndo;
  if OpenDialog.FileName = '' then
    LastFileAge := -1
  else
    LastFileAge := FileAge(OpenDialog.FileName);
end;

procedure TTextChild.NewClick(Sender: TObject);
var
  CanClose: Boolean;
begin
  if (FormStyle = fsMDIChild) then
    TTextChild.Create(Application).SetDefaultExt(ExtractFileExt(OpenDialog.DefaultExt))
  else begin
    CanClose := True;
    FormCloseQuery(Self, CanClose);
    if CanClose then begin
      Memo.Clear;
      SetFileName('');
      ResetModified;
      FFormUsed := False;
    end;
  end;
end;

procedure TTextChild.EditFileAtCursorClick(Sender: TObject);
var
  FileName: string;
  ECode: Cardinal;
begin
  FileName := Memo.WordAtCursor({$IFDEF FPC}@{$ENDIF}IsNonSpaceFileChar);
  DoDirSeparators(FileName);
  if Length(FileName) > 0 then
    if FileExists(FileName) then begin
      StatusBar.Panels[4].Text := Concat('Opening ', FileName);
      if FormStyle = fsMDIChild then
        LoadFileInto(FileName, nil)
      else
        Detach1(ApplicationExecutable, FileName)
    end else begin
      {$IFDEF MSWINDOWS}
      raise EFOpenError.CreateFmt('Can''t open file %s', [FileName]);
      {$ENDIF}
      {$IFDEF UNIX}
      ECode := GetLastOSError;
      Application.MessageBox(PChar(Format('Can''t open file %s - %s', [FileName,SysErrorMessage(ECode)])),'Error',0);
      {$ENDIF}
    end;
end;

procedure TTextChild.ReloadItemClick(Sender: TObject);
begin
  if OpenDialog.Filename <> '' then
    LoadFile(OpenDialog.Filename);
end;

function TTextChild.GetFormUsed: Boolean;
begin
  Result := FFormUsed or Memo.Modified;
end;

procedure TTextChild.CloseItemClick(Sender: TObject);
begin
  Close;
end;

procedure TTextChild.UpdateActions;
begin
  inherited;
  StatusBar.Panels[0].Text := IntToStr(Memo.CaretPos.y+1)+ ' ';
  StatusBar.Panels[1].Text := IntToStr(Memo.CaretPos.x) + ' ';
  if Memo.Modified then
    StatusBar.Panels[2].Text := 'Modified'
  else
    StatusBar.Panels[2].Text := '';
  if Memo.Recording then
    StatusBar.Panels[3].Text := 'Recording'
  else
    StatusBar.Panels[3].Text := '';
end;

constructor TTextChild.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  StatusBar.SimplePanel := False;
  {$IFNDEF MSWINDOWS}
  with StatusBar do
    if SimplePanel then begin
      SimplePanel := False;
      Panels[1].Width := Canvas.TextWidth('9999 ');
      Panels[0].Width := 2*Panels[1].Width;
      Panels[2].Width := Canvas.TextWidth('Modified ');
      Panels[3].Width := Canvas.TextWidth('Recording ');
    end;
  {$ENDIF}
  {$IFDEF LCL_}
  Application.OnActivate := @FormActivate;
  {$ENDIF}
  SetTabSpacing.Action := TTextSetTabSpacing.Create(AOwner);
end;

procedure TTextChild.FormDestroy(Sender: TObject);
begin
  if HandleAllocated then
    DragAcceptFiles(Self, False);
  TextPrintPreview.Lines.Free;
  inherited;
end;

function TTextChild.ExecuteAction(AnAction: TBasicAction): Boolean;
begin
  if AnAction is TResultAction then
    with AnAction as TResultAction do
      if Control = Memo then begin
        StatusBar.Panels[4].Text := Hint;
        Result := True;
        Exit;
      end;
  Result := inherited ExecuteAction(AnAction);
end;

procedure TTextChild.FixedFontDialogChange(Sender: TObject);
begin
  inherited;
  Memo.FixedFont := FixedFontDialog.FixedFont;
end;

procedure TTextChild.FormShow(Sender: TObject);
begin
  inherited;
  if not FixedFontDialog.Local then
    Memo.FixedFont := FixedFontDialog.FixedFont;
end;

procedure TTextChild.AlienNewlineClick(Sender: TObject);
begin
  inherited;
  AlienNewline.Checked := not AlienNewline.Checked;
end;

procedure TTextChild.SaveToFile(FileName: string);
begin
  Memo.Lines.SaveToFileEnh(FileName, KeepBackup.Checked, AlienNewline.Checked);
end;

procedure TTextChild.KeepBackupClick(Sender: TObject);
begin
  inherited;
  KeepBackup.Checked := not KeepBackup.Checked;
end;

procedure TTextChild.PreserveNewlineClick(Sender: TObject);
begin
  inherited;
  PreserveNewline.Checked := not PreserveNewline.Checked;
end;

procedure TTextChild.StartFileatCursorClick(Sender: TObject);
var
  FileName, Dir: string;
begin
  FileName := Memo.WordAtCursor({$IFDEF FPC}@{$ENDIF}IsNonSpaceFileChar);
  if Length(FileName) > 0 then begin
    GetDir(0, Dir);
    {$IFDEF MSWINDOWS}
    if ShellExecute(0, nil, PChar(FileName), nil, PChar(Dir), SW_SHOW) <= 32 then
    // TODO: Equivalent of ShellExecute for Linux
    {$ENDIF}
      EditFileatCursorClick(Sender);
  end;
end;

procedure TTextChild.PrintClick(Sender: TObject);
begin
  inherited;
  Memo.FixedFont.CopyTo(TextPrintPreview.Font);
  TextPrintPreview.Print;
end;

procedure TTextChild.PrintPreviewClick(Sender: TObject);
begin
  inherited;
  Memo.FixedFont.CopyTo(TextPrintPreview.Font);
  TextPrintPreview.ShowModal;
end;

{ TDeTabStringList }

procedure TDeTabStringList.Clear;
begin
end;

constructor TDeTabStringList.Create(TextChild: TTextChild);
begin
  FTextChild := TextChild;
  inherited Create;
end;

procedure TDeTabStringList.Delete(Index: Integer);
begin
end;

function TDeTabStringList.Get(Index: Integer): string;
begin
  Result := FTextChild.Memo.LineWithoutTabs(Index);
end;

function TDeTabStringList.GetCount: Integer;
begin
  Result := FTextChild.Memo.Lines.Count;
end;

procedure TDeTabStringList.Insert(Index: Integer; const S: string);
begin
end;

procedure TTextChild.FormActivate(Sender: TObject);
begin
  inherited;
  if (LastFileAge <> -1) and (FileAge(OpenDialog.Filename) > LastFileAge) then begin
    LastFileAge := FileAge(OpenDialog.Filename); { Prevent re-entry }
    if MessageDlg('File on disk has changed. Reload from disk?', mtWarning, [mbYes, mbNo], 0) = mrYes then
      ReloadItemClick(nil);
  end;
end;

end.
