unit WCompat;

{$I utils.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
{ Contains various compatibility routines to cope with differences in operating systems, compilers
  and windowing routines }
interface

{$IFDEF LCLgtk}
{$IFNDEF BSD}
  {.$DEFINE XINERAMA} { TODO: SHould be dynamic load }
{$ENDIF}
{$ELSE}
{$IFNDEF BSD}
  {$DEFINE XINERAMA}
{$ENDIF}
{$ENDIF}

uses
  Classes, Forms, Messages, Graphics, Types, Menus,
  {$IFDEF VCL_} ShellAPI, {$ENDIF}
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFDEF FPC} LMessages, {$ENDIF}
  {$IFDEF DELPHI}{$IF CompilerVersion >= 24.0}System.UITypes,{$IFEND}{$ENDIF}
  SysUtils;


{$IFDEF UNIX}
  {$IFDEF XBASED}
  function ReadNetFrame(Form: TCustomForm):TRect;
  {$ENDIF}
  function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
  function GetRValue(Colour: TColor):Byte;
  function GetGValue(Colour: TColor):Byte;
  function GetBValue(Colour: TColor):Byte;
{$ENDIF}

{$IFDEF DELPHI}
  {$IF CompilerVersion >= 24.0}
  { In VCL anyway, but remove deprecation warning }
  type TScrollStyle = System.UITypes.TScrollStyle;
  const
    ssNone = System.UITypes.TScrollStyle.ssNone;
    ssHorizontal = System.UITypes.TScrollStyle.ssHorizontal;
    ssVertical = System.UITypes.TScrollStyle.ssVertical;
    ssBoth = System.UITypes.TScrollStyle.ssBoth;
  {$IFEND}
{$ENDIF}

procedure DragAcceptFiles(Form: TCustomForm; Enabled: Boolean);
procedure AdjustMainMenu(MainMenu: TMainMenu; OnAbout: TNotifyEvent);
procedure BringAllToFront;

function ClipboardHasText: Boolean;

function WheelScrollLines: Integer;
{$IFDEF VER90} {$define ADDWHeel} {$ENDIF}
{$IFDEF VER80} {$define ADDWHeel} {$ENDIF}
{$IFDEF MSWINDOWS}{$IFDEF LCL_}  {$define ADDWHeel} {$ENDIF}{$ENDIF}
{$IFDEF ADDWHeel}
const
  WHEEL_DELTA = 120;      { Default value for rolling one notch }
  WM_MOUSEWHEEL = $20A;  { message that will be supported by the OS }
{$ENDIF}

{$IFDEF LCL_}
type
  TWMMouseWheel = TLMMouseEvent;
{$IFNDEF ADDWHeel}
const
  WHEEL_DELTA = 120;      { Default value for rolling one notch }
  WM_MOUSEWHEEL = $20A;  { message that will be supported by the OS }
{$ENDIF}
{$ENDIF}

type
  TStringArray = array of String;

{$IFDEF MSWINDOWS}
var
  uMSH_MOUSEWHEEL:Word;

function ConvertHDROP(Drop: HDROP): TStringArray;

type
  TOnDropFiles = procedure (Sender: TObject; const FileNames: array of String) of object;

{$IFDEF VCL_}
procedure HandleWMDropFiles(var Message: TMessage; OnDropFiles: TOnDropFiles);
{$ENDIF}
{$ENDIF}

function ScreenAreaFor(Form: TCustomForm):TRect;
function ScreenBounds(Form: TCustomForm):TRect;
procedure SetScreenBounds(Form: TCustomForm; Rect: TRect);

implementation

uses
    Controls, ClipBrd,
    {$IFDEF MSWINDOWS}
      UCompat,
    {$ENDIF}
    {$IFDEF XBASED}
      ctypes, X, xlib, xatom, CompatMsg,
      {$IFDEF LCLgtk1} gtk, gdk, glib, {$IFDEF XINERAMA} xinerama, {$ENDIF} {$ENDIF}
      {$IFDEF LCLgtk2} gtk2, gdk2, glib2, {$ENDIF}
      {$IFDEF LCLgtk3} LazGdk3, {$ENDIF}
    {$ENDIF}
    {$IFDEF LCL_}
      LCLType,
    {$ENDIF}
    Math;

{$IFDEF UNIX}

{$IFDEF CLX}
procedure CutFirstDirectory(var S: TFileName); { From Windows FileCtrl }
var
  Root: Boolean;
  P: Integer;
begin
  if S = '\' then
    S := '';
  begin
    if S[1] = '\' then begin
      Root := True;
      Delete(S, 1, 1);
    end else
      Root := False;
    if S[1] = '.' then
      Delete(S, 1, 4);
    P := AnsiPos('\',S);
    if P <> 0 then begin
      Delete(S, 1, P);
      S := '...\' + S;
    end else
      S := '';
    if Root then
      S := '\' + S;
  end;
end;

{ Linux equivalent to FileCtrl.MinimizeName
  - Trim directory names until the resulting string fits on the given Canvas }
function MinimizeName(const Filename: TFileName; Canvas: TCanvas; MaxLen: Integer): TFileName;
var
  Dir: TFileName;
  Name: TFileName;
begin
  Result := FileName;
  Dir := ExtractFilePath(Result);
  Name := ExtractFileName(Result);

  while (Dir <> '') and (Canvas.TextWidth(Result) > MaxLen) do begin
    if Dir = '\...\' then
      Dir := '...\'
    else
      CutFirstDirectory(Dir);
    Result := Dir + Name;
  end;
end;

function ClipboardHasText: Boolean;
begin
  Result := Clipboard.Provides('text/plain');
end;
{$ENDIF}

{$IFDEF FPC}

{$ENDIF}
{$ENDIF}

{$IFDEF testselection}
function GetPrimarySelection: string;
begin
  XConvertSelection(QtDisplay, XA_PRIMARY, XA_STRING, {None} XA_STRING,
                    QWidget_winId(Application.MainForm.Handle), CurrentTime);
  Result := '';
end;
{$ENDIF}

function ClipboardHasText: Boolean;
begin
  Result := Clipboard.HasFormat(CF_TEXT);
end;

procedure BringAllToFront;
var
  CurrentForm: TCustomForm;
  i: Integer;
begin
  CurrentForm := Screen.ActiveCustomForm;
  if not Assigned(CurrentForm) then
    Exit;
  for i := 0 to Screen.CustomFormCount-1 do
    with Screen.CustomForms[i] do
      if Visible then
        BringToFront;
  CurrentForm.BringToFront;
end;

{$IFDEF DARWIN}
type
  TMinimizeMenuItem = class(TMenuItem)
  public
    procedure Click; override;
  end;

procedure TMinimizeMenuItem.Click;
begin
  with Screen.ActiveForm do
    if WindowState = wsMinimized then
      WindowState := wsNormal
    else
      WindowState := wsMinimized;
end;

type
  TZoomMenuItem = class(TMenuItem)
  public
    procedure Click; override;
  end;

procedure TZoomMenuItem.Click;
begin
  with Screen.ActiveForm do
    if WindowState <> wsNormal then
      WindowState := wsNormal
    else
      WindowState := wsMaximized;
end;

type
  TBFMenuItem = class(TMenuItem)
  public
    procedure Click; override;
  end;

procedure TBFMenuItem.Click;
begin
  BringAllToFront;
end;

type
  TCycleMenuItem = class(TMenuItem)
  public
    procedure Click; override;
  end;

procedure TCycleMenuItem.Click;
var
  CurrentForm: TCustomForm;
  i, n: Integer;
begin
  CurrentForm := Screen.ActiveCustomForm;
  if not Assigned(CurrentForm) then
    Exit;
  for i := 0 to Screen.CustomFormCount-1 do
    if Screen.CustomForms[i] = CurrentForm then begin
      n := i;
      Break;
    end;
  i := n+1;
  repeat
    if i >= Screen.CustomFormCount then
      i := 0;
    if i = n then
      Break; { No other forms visible }
    if Screen.CustomForms[i].Visible then begin
       Screen.CustomForms[i].BringToFront;
       Break;
    end;
    Inc(i);
  until False;
end;
{$ENDIF}

procedure AdjustMainMenu(MainMenu: TMainMenu; OnAbout: TNotifyEvent);
{$IFDEF DARWIN}
var
  AppMenu, MenuItem, WindowMenu: TMenuItem;
  WindowMenuAt: Integer;
  procedure AdjustShortCuts(M: TMenuItem);
  var
    Key: Word;
    Shift: TShiftState;
    i: Integer;
  begin
    for i := 0 to M.Count-1 do
      AdjustShortCuts(M.Items[i]);
    ShortCutToKey(M.ShortCut, Key, Shift);
    if Shift = [ssCtrl] then
      case Key of
        VK_M, VK_Q, VK_H: { Clash with standard short cuts };
      else
        M.ShortCut := ShortCut(Key, [ssMeta]);
      end;
  end;
{$ENDIF}
  {$IFDEF LCL}
  procedure MakeTransparent(MenuItem: TMenuItem);
  var
    i: Integer;
  begin
    if MenuItem.HasBitmap then
      MenuItem.Bitmap.Transparent := True;
    for i := 0 to MenuItem.Count-1 do
      MakeTransparent(MenuItem.Items[i]);
  end;
  {$ENDIF}
begin
  {$IFDEF DARWIN}
  if Assigned(OnAbout) then begin
    AppMenu := TMenuItem.Create(MainMenu.Owner);  // Application menu
    AppMenu.Caption := #$EF#$A3#$BF;    // Unicode Apple logo char
    MainMenu.Items.Insert(0, AppMenu);
    MenuItem := TMenuItem.Create(MainMenu.Owner);
    MenuItem.Caption := 'About ' + Application.Title;
    MenuItem.OnClick := OnAbout;
    AppMenu.Add(MenuItem);
  end;
  WindowMenuAt := MainMenu.Items.Count-1;
  if MainMenu.Items[WindowMenuAt].Caption = 'Help' then
    Dec(WindowMenuAt);
  WindowMenu := TMenuItem.Create(MainMenu.Owner);
  WindowMenu.Caption := 'Window';
  MenuItem := TMinimizeMenuItem.Create(MainMenu.Owner);
  MenuItem.Caption := 'Minimize';
  MenuItem.ShortCut := ShortCut(VK_M, [ssMeta]);
  WindowMenu.Add(MenuItem);
  MenuItem := TZoomMenuItem.Create(MainMenu.Owner);
  MenuItem.Caption := 'Zoom';
  WindowMenu.Add(MenuItem);
  MenuItem := TCycleMenuItem.Create(MainMenu.Owner);
  MenuItem.Caption := 'Cycle Through Windows';
  MenuItem.ShortCut := ShortCut(VK_OEM_3, [ssMeta]);
  WindowMenu.Add(MenuItem);
  MenuItem := TMenuItem.Create(MainMenu.Owner);
  MenuItem.Caption := '-';
  WindowMenu.Add(MenuItem);
  MenuItem := TBFMenuItem.Create(MainMenu.Owner);
  MenuItem.Caption := 'Bring All to Front';
  WindowMenu.Add(MenuItem);
  MainMenu.Items.Insert(WindowMenuAt, WindowMenu);
  AdjustShortCuts(MainMenu.Items);
  {$ENDIF}
  {$IFDEF LCL}
  MakeTransparent(MainMenu.Items);
  {$ENDIF}
end;

{$IFNDEF VCL_}
function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
begin
  Result := Red or (Green shl 8) or (Blue shl 16);
end;

function GetRValue(Colour: TColor):Byte;
begin
  Result := Colour and 255;
end;

function GetGValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 8) and 255;
end;

function GetBValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 16) and 255;
end;
{$ENDIF}

function WheelScrollLines: Integer;
begin
  Result :=
  {$IFDEF VCL_}
    Mouse.WheelScrollLines;
  if Result = 0 then
    Result := 3;
  {$ELSE}
    3;
  {$ENDIF}
end;

{$IFDEF MSWINDOWS}
function ConvertHDROP(Drop: HDROP): TStringArray;
var
  i, nFiles, l: Integer;
begin
  nFiles := DragQueryFile(Drop, UINT(MinusOne), nil, 0);
  SetLength(Result, nFiles);
  for i := 0 to High(Result) do begin
    l := DragQueryFile(Drop, i, nil, 0); { Length of name }
    SetLength(Result[i],l);
    DragQueryFile(Drop, i, @Result[i][1], l+1);
  end;
end;

{$IFDEF VCL_}
procedure HandleWMDropFiles(var Message: TMessage; OnDropFiles: TOnDropFiles);
var
  Files: TStringArray;
begin
  try
    Files := ConvertHDROP(Message.wParam);
  finally
    DragFinish(Message.wParam);
    Message.Result := 0;
  end;
  if Assigned(OnDropFiles) then
    OnDropFiles(nil, Files);
end;

procedure DragAcceptFiles(Form: TCustomForm; Enabled: Boolean);
begin
  ShellAPI.DragAcceptFiles(Form.Handle, Enabled);
end;
{$ENDIF}
{$ENDIF}

{$IFDEF FPC}
procedure DragAcceptFiles(Form:TCustomForm; Enabled: Boolean);
begin
  Form.AllowDropFiles := Enabled;
end;
{$ENDIF}

{$IFDEF XBASED}
var
  _NET_FRAME_EXTENTS: TAtom = 0;
  LastNetFrame: TRect;

function ReadNetFrame(Form: TCustomForm):TRect;
var
  AtomType: TAtom;
  ActualFormat: Longint;
  Nitems, BytesLeft: Cardinal;
  Data: {$IFDEF CPU64} PQWord {$ELSE} PCardinal {$ENDIF};
  Wnd: TWnd;
begin
  if _NET_FRAME_EXTENTS = 0 then
    _NET_FRAME_EXTENTS := XInternAtom(X11Display, '_NET_FRAME_EXTENTS', False);
  Result := LastNetFrame;
  if Form.HandleAllocated then begin
     Wnd := XHandle(Form.Handle);
     if (Wnd <> 0) and (XGetWindowProperty(X11Display, XHandle(Form.Handle), _NET_FRAME_EXTENTS,
                                           0, 4, Ord(False), XA_CARDINAL, @AtomType, @ActualFormat, @Nitems, @BytesLeft, @Data) = Success)
        and (Data <> nil) then begin
      Result := Rect(Data^, (Data+2)^, (Data+1)^, (Data+3)^);
      XFree(Data);
      LastNetFrame := Result;
    end;
  end;
end;

var
  _NET_WM_DESKTOP: TAtom = 0;
  _NET_NUMBER_OF_DESKTOPS: TAtom = 0;
  _NET_WORKAREA: TAtom = 0;

function ReadWorkArea(Form: TCustomForm):TRect;
var
  nDesktops, Desktop: Cardinal;
  AtomType: TAtom;
  ActualFormat: cint;
  Nitems, BytesLeft: culong;
  Data, P: Pculong;
  Wnd: TWnd;
begin
  if _NET_WM_DESKTOP = 0 then begin
    _NET_WM_DESKTOP := XInternAtom(X11Display, '_NET_WM_DESKTOP', False);
    _NET_NUMBER_OF_DESKTOPS := XInternAtom(X11Display, '_NET_NUMBER_OF_DESKTOPS', False);
    _NET_WORKAREA := XInternAtom(X11Display, '_NET_WORKAREA', False);
  end;
  if (XGetWindowProperty(X11Display, RootWindow, _NET_NUMBER_OF_DESKTOPS,
                        0, 1, False, XA_CARDINAL, @AtomType, @ActualFormat, @Nitems, @BytesLeft, @Data) = Success)
     and (Data <> nil) then begin
    nDesktops := Data^;
    XFree(Data);
    Desktop := 0;
    if Assigned(Form) and Form.HandleAllocated then begin
       Wnd := XHandle(Form.Handle);
       if (Wnd <> 0) and (XGetWindowProperty(X11Display, Wnd, _NET_WM_DESKTOP,
                                             0, 1, False, XA_CARDINAL, @AtomType, @ActualFormat, @Nitems, @BytesLeft, @Data) = Success)
          and (Data <> nil) then begin
         Desktop := Data^;
         XFree(Data);
      end;
    end;
    Desktop := Min(Desktop, nDesktops-1);
    if (XGetWindowProperty(X11Display, RootWindow, _NET_WORKAREA,
                          0, (Desktop+1)*4, False, XA_CARDINAL, @AtomType, @ActualFormat, @Nitems, @BytesLeft, @Data) = Success)
       and (Data <> nil) then begin
      P := Data;
      Inc(P, Desktop);
      Result := Rect(P^, (P+1)^, (P+2)^, (P+3)^);
      XFree(Data);
      Exit;
    end;
  end;
  Result := Rect(0, 0, Screen.Width, Screen.Height);
end;

var
  LoadMonitorsCalled: Boolean = False;

var Monitors: array of TRect;

{$IFNDEF LCLgtk1}
function LoadMonitors: Boolean;
var
  nMonitors: cint;
  S: TGdkRectangle;
  opcode, firstevent, firsterror: cint;
  i: Integer;
begin
  if not LoadMonitorsCalled then begin
    nMonitors := gdk_screen_get_n_monitors(gdk_screen_get_default());
    if nMonitors > 0 then begin
      SetLength(Monitors, nMonitors);
      for i := 0 to nMonitors-1 do begin
        gdk_screen_get_monitor_geometry(gdk_screen_get_default(),i,@S);
        with S do
          Monitors[i] := Rect(x, y, x+width, y+height);
      end;
    end;
    LoadMonitorsCalled := True;
  end;
  Result := Length(Monitors) > 0;
end;
{$ENDIF}

{$IFDEF LCLgtk1}
{ Try and use the Xinerama extension}
{$IFDEF XINERAMA}
function LoadMonitors: Boolean;
var
  nMonitors: cint;
  XineramaScreenInfo, S: PXineramaScreenInfo;
  opcode, firstevent, firsterror: cint;
  i: Integer;
begin
  if not LoadMonitorsCalled then begin
    if XQueryExtension(gdk_display, 'XINERAMA', @opcode, @firstevent, @firsterror)
       and XineramaIsActive(gdk_display) then begin
      XineramaScreenInfo := XineramaQueryScreens(gdk_display, @nMonitors);
      if XineramaScreenInfo <> nil then begin
        if (nMonitors > 0) and (nMonitors < 10) then begin
          SetLength(Monitors, nMonitors);
          S := XineramaScreenInfo;
          for i := 0 to nMonitors-1 do begin
            with S^ do
              Monitors[i] := Rect(x_org, y_org, x_org+width, y_org+height);
            Inc(S);
          end;
        end;
        XFree(XineramaScreenInfo);
      end;
    end;
    LoadMonitorsCalled := True;
  end;
  Result := Length(Monitors) > 0;
end;
{$ENDIF XINERAMA}
{$ENDIF LCLgtk}

function MonitorFromWindow(Form: TCustomForm):Integer;
var
  i, lastarea, area: Integer;
  R: TRect;
begin
  Result := 0; { Default }
  lastarea := 0;
  for i := 0 to High(Monitors) do
    if IntersectRect(R, ScreenBounds(Form), Monitors[i]) then begin
      area := (R.Right-R.Left)*(R.Bottom-R.Top);
      if area > lastarea then begin
        Result := i;
        lastarea := area;
      end;
  end;
end;

{$ENDIF}

{$IFDEF MSWINDOWS}
type
  HMONITOR = THandle;
  MONITORINFO = record
    cbSize: DWORD;
    rcMonitor: TRECT;
    rcWork: TRECT;
    dwFlags:DWORD;
  end;
  LPMONITORINFO = ^MONITORINFO;

const
  { MONITOR_DEFAULTTONULL = 0; }
  { MONITOR_DEFAULTTOPRIMARY = 1; }
  MONITOR_DEFAULTTONEAREST = 2;
  
//function GetMonitorInfo(hMonitor:HMONITOR; lpmi: LPMONITORINFO):BOOL; stdcall; external 'user32.dll' name 'GetMonitorInfoA';
//function MonitorFromWindow(hwnd: HWND; dwFlags: DWORD):HMONITOR; stdcall; external 'user32.dll' name 'MonitorFromWindow';

type
  TGetMonitorInfo = function (hMonitor:HMONITOR; lpmi: LPMONITORINFO):BOOL; stdcall;
  TMonitorFromWindow = function (hwnd: HWND; dwFlags: DWORD):HMONITOR; stdcall;

var
  GetMonitorInfo: TGetMonitorInfo = nil;
  MonitorFromWindow: TMonitorFromWindow = nil;

{$ENDIF}

function ScreenAreaFor(Form: TCustomForm):TRect;
{$IFDEF MSWINDOWS}
var
  Info: MONITORINFO;
  Monitor: HMONITOR;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
    {$IFDEF VCL_Off}
      if Assigned(Form) then
        Result := Screen.MonitorFromWindow(Form.Handle).WorkareaRect
      else
        Result := Screen.DesktopRect;
    {$ELSE}
      if Assigned(Form) and ({$IFNDEF FPC}@{$ENDIF}GetMonitorInfo <> nil) then begin
        Monitor := MonitorFromWindow(Form.Handle, MONITOR_DEFAULTTONEAREST);
        if Monitor <> 0 then begin
          Info.cbSize := Sizeof(Info);
          if not GetMonitorInfo(Monitor, @Info) then
            RaiseLastOSError;
          Result := Info.rcWork;
          Exit;
        end;
      end;
      with Result do begin
        Left := 0;
        Top := 0;
        Right := Screen.Width;
        Bottom := Screen.Height;
      end;
    {$ENDIF}
  {$ELSE}
    {$IFDEF XINERAMA}
      Result := ReadWorkArea(Form);
      if Assigned(Form) and LoadMonitors then
        IntersectRect(Result, Result, Monitors[MonitorFromWindow(Form)]);
    {$ELSE}
      {$IFDEF DARWIN}
        if Assigned(Form) and (Form.HandleAllocated) then
          Result := Screen.MonitorFromWindow(Form.Handle).WorkareaRect
        else if Assigned(Application.MainForm) and Application.MainForm.HandleAllocated then
          Result := ScreenAreaFor(Application.MainForm)
        else if Screen.MonitorCount > 0 then
          Result := Screen.Monitors[0].WorkareaRect
        else
          Result := Rect(0, 20, Screen.Width, Screen.Height-20); { Guess title bar height }
      {$ELSE}
        Result := Rect(0, 0, Screen.Width, Screen.Height);
      {$ENDIF}
    {$ENDIF}
  {$ENDIF}
end;

function ScreenBounds(Form: TCustomForm):TRect;
{$IFDEF UNIX}
var
  Frame: TRect;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  Windows.GetWindowRect(Form.Handle,Result);
  {$ELSE}
  Result := Form.BoundsRect;
  {$ENDIF}
  {$IFDEF XBASED}
  Frame := ReadNetFrame(Form);
  Inc(Result.Right,Frame.Right+Frame.Left);
  Inc(Result.Bottom,Frame.Bottom+Frame.Top);
  {$ENDIF}
end;

procedure SetScreenBounds(Form: TCustomForm; Rect: TRect);
{$IFDEF XBASED}
var
  Frame: TRect;
{$ENDIF}
begin
  {$IFDEF MSWINDOWS}
  Windows.MoveWindow(Form.Handle,Rect.Left,Rect.Top,Rect.Right-Rect.Left,Rect.Bottom-Rect.Top,True);
  Exit;
  {$ENDIF}
  {$IFDEF XBASED}
  Frame := ReadNetFrame(Form);
  Dec(Rect.Right,Frame.Right+Frame.Left);
  Dec(Rect.Bottom,Frame.Bottom+Frame.Top);
  {$ENDIF}
  Form.BoundsRect := Rect;
end;

{$IFDEF MSWINDOWS}
const
  MSH_MOUSEWHEEL = 'MSWHEEL_ROLLMSG';
  { wParam = wheel rotation expressed in multiples of WHEEL_DELTA
    lParam is the mouse coordinates }
    
initialization
  uMSH_MOUSEWHEEL := RegisterWindowMessage(MSH_MOUSEWHEEL);
  { Dynamic load, as not available on earliest Win9X. Note that no adjustment needed for 64 bit }
  GetMonitorInfo := TGetMonitorInfo(GetProcAddress(GetModuleHandle('user32'),'GetMonitorInfo'));
  if {$IFNDEF FPC}@{$ENDIF}GetMonitorInfo = nil then
    GetMonitorInfo := TGetMonitorInfo(GetProcAddress(GetModuleHandle('user32'),'GetMonitorInfoA'));
  if {$IFNDEF FPC}@{$ENDIF}GetMonitorInfo <> nil then
    MonitorFromWindow := TMonitorFromWindow(GetProcAddress(GetModuleHandle('user32'),'MonitorFromWindow'));
{$ENDIF}
end.


