unit Atomic;
{$I utils.inc}{Copyright (C) 2006 C M Western; see commandline.pas for license}
interface

{ Various atomic calibration lines from  the
Ralchenko, Yu., Jou, F.-C., Kelleher, D.E., Kramida, A.E., Musgrove, A., Reader, J., Wiese, W.L., and Olsen, K. (2005).
NIST Atomic Spectra Database (version 3.0.3), [Online]. Available: http://physics.nist.gov/asd3 [2006, June 18]. National Institute of Standards and Technology, Gaithersburg, MD.
  NIST Atomic Spectra Database [Ver. 3.0.3],
  Yu. Ralchenko, F.-C. Jou, D.E. Kelleher, A.E. Kramida,
  A. Musgrove, J. Reader, W.L. Wiese, and K. Olsen }
{ An atlas of optogalvanic transitions in Neon.
      RAL report, RAL-91-069 (1991)
            S.H. Ashworth and J.M. Brown. }

uses Mixture;

type

  TAtomicLineList = record
    Name: String;
    LineList: TLineList;
    IgnoreIntensities: Boolean;
  end;

var
  AtomicLineLists: array of TAtomicLineList;

implementation

uses CalSource, Units, Graphics, AirVac, SysUtils;

type
  TAtomicLine = record
    Position: Double;
    Intensity: Integer;
    Charge: Byte;
  end;

const
{ Observed Wavelength Vac /nm }
Neon: array [0..1827] of TAtomicLine = (
(Position:52.174;Intensity:120;Charge:4),
(Position:52.182;Intensity:140;Charge:4),
(Position:54.113;Intensity:80;Charge:4),
(Position:54.207;Intensity:100;Charge:4),
(Position:54.389;Intensity:150;Charge:4),
(Position:56.842;Intensity:400;Charge:5),
(Position:56.976;Intensity:250;Charge:5),
(Position:56.983;Intensity:500;Charge:5),
(Position:57.211;Intensity:250;Charge:5),
(Position:57.234;Intensity:800;Charge:5),
(Position:57.32775;Intensity:1;Charge:1),
(Position:57.33754;Intensity:1;Charge:1),
(Position:57.34056;Intensity:1;Charge:1),
(Position:57.34889;Intensity:1;Charge:1),
(Position:57.35257;Intensity:1;Charge:1),
(Position:57.36252;Intensity:1;Charge:1),
(Position:57.36679;Intensity:1;Charge:1),
(Position:57.37873;Intensity:1;Charge:1),
(Position:57.38383;Intensity:1;Charge:1),
(Position:57.39821;Intensity:1;Charge:1),
(Position:57.40447;Intensity:1;Charge:1),
(Position:57.42208;Intensity:1;Charge:1),
(Position:57.42982;Intensity:1;Charge:1),
(Position:57.45172;Intensity:1;Charge:1),
(Position:57.46144;Intensity:1;Charge:1),
(Position:57.48912;Intensity:1;Charge:1),
(Position:57.50158;Intensity:1;Charge:1),
(Position:57.53703;Intensity:1;Charge:1),
(Position:57.55347;Intensity:1;Charge:1),
(Position:57.58544;Intensity:1;Charge:1),
(Position:57.59530;Intensity:1;Charge:1),
(Position:57.59822;Intensity:1;Charge:1),
(Position:57.60045;Intensity:1;Charge:1),
(Position:57.60684;Intensity:1;Charge:1),
(Position:57.61032;Intensity:1;Charge:1),
(Position:57.62048;Intensity:1;Charge:1),
(Position:57.62255;Intensity:1;Charge:1),
(Position:57.62481;Intensity:1;Charge:1),
(Position:57.63685;Intensity:1;Charge:1),
(Position:57.64184;Intensity:1;Charge:1),
(Position:57.65645;Intensity:1;Charge:1),
(Position:57.65658;Intensity:1;Charge:1),
(Position:57.66262;Intensity:1;Charge:1),
(Position:57.68048;Intensity:1;Charge:1),
(Position:57.68083;Intensity:1;Charge:1),
(Position:57.68643;Intensity:1;Charge:1),
(Position:57.68816;Intensity:1;Charge:1),
(Position:57.71034;Intensity:1;Charge:1),
(Position:57.71080;Intensity:1;Charge:1),
(Position:57.71687;Intensity:1;Charge:1),
(Position:57.72024;Intensity:1;Charge:1),
(Position:57.74803;Intensity:1;Charge:1),
(Position:57.74872;Intensity:1;Charge:1),
(Position:57.76038;Intensity:1;Charge:1),
(Position:57.79649;Intensity:1;Charge:1),
(Position:57.79740;Intensity:1;Charge:1),
(Position:57.80710;Intensity:1;Charge:1),
(Position:57.81262;Intensity:1;Charge:1),
(Position:57.85123;Intensity:1;Charge:1),
(Position:57.86049;Intensity:1;Charge:1),
(Position:57.86175;Intensity:1;Charge:1),
(Position:57.88220;Intensity:1;Charge:1),
(Position:57.94712;Intensity:1;Charge:1),
(Position:57.94880;Intensity:1;Charge:1),
(Position:57.97704;Intensity:1;Charge:1),
(Position:57.98406;Intensity:1;Charge:1),
(Position:58.05113;Intensity:1;Charge:1),
(Position:58.06887;Intensity:1;Charge:1),
(Position:58.07131;Intensity:1;Charge:1),
(Position:58.11215;Intensity:1;Charge:1),
(Position:58.24687;Intensity:1;Charge:1),
(Position:58.25059;Intensity:1;Charge:1),
(Position:58.25977;Intensity:1;Charge:1),
(Position:58.31257;Intensity:1;Charge:1),
(Position:58.36891;Intensity:1;Charge:1),
(Position:58.52472;Intensity:1;Charge:1),
(Position:58.53040;Intensity:1;Charge:1),
(Position:58.63138;Intensity:1;Charge:1),
(Position:58.72127;Intensity:1;Charge:1),
(Position:58.91792;Intensity:1;Charge:1),
(Position:58.99114;Intensity:1;Charge:1),
(Position:59.00109;Intensity:1;Charge:1),
(Position:59.18306;Intensity:1;Charge:1),
(Position:59.59200;Intensity:30;Charge:1),
(Position:59.87056;Intensity:20;Charge:1),
(Position:59.88897;Intensity:10;Charge:1),
(Position:60.00365;Intensity:20;Charge:1),
(Position:60.27263;Intensity:40;Charge:1),
(Position:61.56283;Intensity:50;Charge:1),
(Position:61.86716;Intensity:50;Charge:1),
(Position:61.91023;Intensity:40;Charge:1),
(Position:62.68232;Intensity:60;Charge:1),
(Position:62.97388;Intensity:60;Charge:1),
(Position:73.58962;Intensity:300;Charge:1),
(Position:74.37195;Intensity:120;Charge:1),
(Position:113.185;Intensity:100;Charge:2),
(Position:125.503;Intensity:20;Charge:3),
(Position:125.568;Intensity:110;Charge:3),
(Position:125.719;Intensity:160;Charge:3),
(Position:168.168;Intensity:120;Charge:2),
(Position:168.836;Intensity:180;Charge:2),
(Position:188.811;Intensity:100;Charge:2),
(Position:188.971;Intensity:100;Charge:2),
(Position:190.749;Intensity:200;Charge:2),
(Position:191.608;Intensity:500;Charge:2),
(Position:193.003;Intensity:300;Charge:2),
(Position:193.883;Intensity:200;Charge:2),
(Position:194.546;Intensity:100;Charge:2),
(Position:200.766;Intensity:80;Charge:2),
(Position:201.909;Intensity:65;Charge:4),
(Position:202.284;Intensity:110;Charge:4),
(Position:202.621;Intensity:80;Charge:2),
(Position:208.613;Intensity:150;Charge:2),
(Position:208.762;Intensity:200;Charge:3),
(Position:209.009;Intensity:300;Charge:3),
(Position:209.311;Intensity:240;Charge:3),
(Position:209.311;Intensity:240;Charge:3),
(Position:209.621;Intensity:400;Charge:3),
(Position:209.678;Intensity:180;Charge:2),
(Position:209.692;Intensity:120;Charge:2),
(Position:216.190;Intensity:200;Charge:3),
(Position:216.445;Intensity:300;Charge:3),
(Position:218.157;Intensity:200;Charge:3),
(Position:220.457;Intensity:30;Charge:4),
(Position:221.004;Intensity:200;Charge:3),
(Position:221.254;Intensity:200;Charge:3),
(Position:221.445;Intensity:240;Charge:3),
(Position:221.676;Intensity:300;Charge:3),
(Position:222.150;Intensity:10;Charge:4),
(Position:222.811;Intensity:75;Charge:5),
(Position:223.310;Intensity:110;Charge:5),
(Position:224.618;Intensity:65;Charge:5),
(Position:225.872;Intensity:250;Charge:4),
(Position:226.027;Intensity:65;Charge:5),
(Position:226.278;Intensity:175;Charge:4),
(Position:226.391;Intensity:240;Charge:3),
(Position:226.409;Intensity:65;Charge:5),
(Position:226.524;Intensity:110;Charge:4),
(Position:226.561;Intensity:200;Charge:3),
(Position:226.641;Intensity:250;Charge:5),
(Position:228.650;Intensity:550;Charge:4),
(Position:229.385;Intensity:30;Charge:4),
(Position:229.420;Intensity:250;Charge:4),
(Position:235.156;Intensity:250;Charge:4),
(Position:235.324;Intensity:450;Charge:4),
(Position:235.868;Intensity:700;Charge:4),
(Position:236.340;Intensity:250;Charge:4),
(Position:236.400;Intensity:250;Charge:4),
(Position:236.621;Intensity:110;Charge:4),
(Position:237.288;Intensity:350;Charge:4),
(Position:238.493;Intensity:65;Charge:4),
(Position:238.568;Intensity:350;Charge:4),
(Position:241.346;Intensity:300;Charge:3),
(Position:241.367;Intensity:240;Charge:3),
(Position:241.451;Intensity:200;Charge:3),
(Position:241.451;Intensity:200;Charge:3),
(Position:247.415;Intensity:200;Charge:3),
(Position:256.256;Intensity:80;Charge:1),
(Position:256.289;Intensity:80;Charge:2),
(Position:256.789;Intensity:90;Charge:2),
(Position:257.532;Intensity:80;Charge:1),
(Position:257.532;Intensity:80;Charge:1),
(Position:259.025;Intensity:20;Charge:1),
(Position:259.081;Intensity:800;Charge:3),
(Position:259.144;Intensity:100;Charge:1),
(Position:259.144;Intensity:100;Charge:1),
(Position:259.193;Intensity:30;Charge:1),
(Position:259.438;Intensity:600;Charge:3),
(Position:259.534;Intensity:20;Charge:1),
(Position:259.599;Intensity:500;Charge:1),
(Position:259.646;Intensity:400;Charge:3),
(Position:261.081;Intensity:300;Charge:3),
(Position:261.419;Intensity:240;Charge:3),
(Position:261.437;Intensity:300;Charge:1),
(Position:261.437;Intensity:300;Charge:1),
(Position:261.4705;Intensity:20;Charge:1),
(Position:261.504;Intensity:50;Charge:1),
(Position:261.665;Intensity:200;Charge:3),
(Position:261.740;Intensity:250;Charge:1),
(Position:261.980;Intensity:50;Charge:1),
(Position:261.980;Intensity:50;Charge:1),
(Position:262.055;Intensity:30;Charge:1),
(Position:262.188;Intensity:80;Charge:1),
(Position:262.368;Intensity:150;Charge:1),
(Position:262.389;Intensity:80;Charge:2),
(Position:263.067;Intensity:80;Charge:2),
(Position:263.686;Intensity:90;Charge:2),
(Position:263.908;Intensity:80;Charge:2),
(Position:263.949;Intensity:200;Charge:3),
(Position:264.076;Intensity:150;Charge:1),
(Position:264.186;Intensity:200;Charge:3),
(Position:264.326;Intensity:80;Charge:1),
(Position:264.326;Intensity:80;Charge:1),
(Position:264.489;Intensity:80;Charge:2),
(Position:264.630;Intensity:500;Charge:1),
(Position:264.649;Intensity:300;Charge:1),
(Position:264.821;Intensity:2000;Charge:1),
(Position:264.855;Intensity:80;Charge:1),
(Position:264.900;Intensity:150;Charge:1),
(Position:264.935;Intensity:250;Charge:1),
(Position:265.180;Intensity:500;Charge:1),
(Position:265.831;Intensity:150;Charge:1),
(Position:265.831;Intensity:150;Charge:1),
(Position:266.863;Intensity:20;Charge:1),
(Position:266.992;Intensity:30;Charge:1),
(Position:267.6070;Intensity:2000;Charge:1),
(Position:267.644;Intensity:2000;Charge:1),
(Position:267.7816;Intensity:10;Charge:1),
(Position:267.8185;Intensity:150;Charge:1),
(Position:267.867;Intensity:20;Charge:1),
(Position:267.870;Intensity:600;Charge:3),
(Position:267.870;Intensity:600;Charge:3),
(Position:267.944;Intensity:500;Charge:3),
(Position:268.1481;Intensity:10;Charge:1),
(Position:268.7540;Intensity:150;Charge:1),
(Position:270.1356;Intensity:80;Charge:1),
(Position:270.1482;Intensity:20;Charge:1),
(Position:270.2440;Intensity:20;Charge:1),
(Position:270.2567;Intensity:20;Charge:1),
(Position:270.3362;Intensity:30;Charge:1),
(Position:270.512;Intensity:20;Charge:1),
(Position:270.754;Intensity:30;Charge:1),
(Position:272.5579;Intensity:30;Charge:1),
(Position:273.2167;Intensity:30;Charge:1),
(Position:273.2337;Intensity:30;Charge:1),
(Position:273.342;Intensity:10;Charge:1),
(Position:273.5564;Intensity:20;Charge:1),
(Position:273.5978;Intensity:30;Charge:1),
(Position:273.650;Intensity:80;Charge:1),
(Position:273.6984;Intensity:50;Charge:1),
(Position:274.434;Intensity:150;Charge:1),
(Position:275.663;Intensity:150;Charge:1),
(Position:275.663;Intensity:150;Charge:1),
(Position:275.946;Intensity:30;Charge:1),
(Position:276.0138;Intensity:20;Charge:1),
(Position:276.3140;Intensity:30;Charge:1),
(Position:276.374;Intensity:80;Charge:2),
(Position:276.7181;Intensity:30;Charge:1),
(Position:276.810;Intensity:30;Charge:1),
(Position:276.859;Intensity:20;Charge:1),
(Position:277.58707;Intensity:50;Charge:1),
(Position:278.224;Intensity:30;Charge:1),
(Position:278.224;Intensity:30;Charge:1),
(Position:278.250;Intensity:30;Charge:1),
(Position:278.289;Intensity:20;Charge:1),
(Position:279.284;Intensity:90;Charge:2),
(Position:279.31442;Intensity:200;Charge:1),
(Position:279.3484;Intensity:30;Charge:1),
(Position:279.504;Intensity:80;Charge:2),
(Position:279.5421;Intensity:50;Charge:1),
(Position:279.5916;Intensity:350;Charge:1),
(Position:279.6437;Intensity:10;Charge:1),
(Position:279.67862;Intensity:80;Charge:1),
(Position:280.063;Intensity:20;Charge:1),
(Position:281.031;Intensity:100;Charge:2),
(Position:281.55210;Intensity:200;Charge:1),
(Position:282.6090;Intensity:100;Charge:1),
(Position:282.6445;Intensity:80;Charge:1),
(Position:282.8421;Intensity:30;Charge:1),
(Position:283.37560;Intensity:80;Charge:1),
(Position:283.60734;Intensity:150;Charge:1),
(Position:284.341;Intensity:150;Charge:1),
(Position:284.45;Intensity:10;Charge:1),
(Position:284.7327;Intensity:20;Charge:1),
(Position:285.5445;Intensity:10;Charge:1),
(Position:286.2911;Intensity:80;Charge:1),
(Position:287.35059;Intensity:350;Charge:1),
(Position:288.1135;Intensity:30;Charge:1),
(Position:288.2124;Intensity:10;Charge:1),
(Position:288.2697;Intensity:20;Charge:1),
(Position:290.744;Intensity:80;Charge:2),
(Position:290.767;Intensity:80;Charge:2),
(Position:291.091;Intensity:90;Charge:2),
(Position:291.126;Intensity:90;Charge:2),
(Position:291.199;Intensity:80;Charge:2),
(Position:291.23232;Intensity:250;Charge:1),
(Position:291.40266;Intensity:2000;Charge:1),
(Position:291.4270;Intensity:20;Charge:1),
(Position:291.597;Intensity:80;Charge:2),
(Position:292.648;Intensity:80;Charge:2),
(Position:293.01828;Intensity:150;Charge:1),
(Position:293.296;Intensity:80;Charge:2),
(Position:293.35832;Intensity:1000;Charge:1),
(Position:294.151;Intensity:80;Charge:2),
(Position:294.5436;Intensity:20;Charge:1),
(Position:294.690;Intensity:90;Charge:2),
(Position:294.7593;Intensity:20;Charge:1),
(Position:294.81626;Intensity:2000;Charge:1),
(Position:294.99117;Intensity:100;Charge:1),
(Position:295.01839;Intensity:150;Charge:1),
(Position:295.3390;Intensity:50;Charge:1),
(Position:295.658;Intensity:150;Charge:2),
(Position:295.8157;Intensity:80;Charge:1),
(Position:295.8157;Intensity:80;Charge:1),
(Position:296.411;Intensity:150;Charge:2),
(Position:296.805;Intensity:150;Charge:2),
(Position:297.397;Intensity:100;Charge:2),
(Position:297.55873;Intensity:3000;Charge:1),
(Position:297.63919;Intensity:350;Charge:1),
(Position:298.033;Intensity:100;Charge:2),
(Position:298.06782;Intensity:500;Charge:1),
(Position:298.15151;Intensity:400;Charge:1),
(Position:298.17951;Intensity:500;Charge:1),
(Position:298.35400;Intensity:3000;Charge:1),
(Position:299.33024;Intensity:2000;Charge:1),
(Position:299.5123;Intensity:30;Charge:1),
(Position:300.255;Intensity:150;Charge:2),
(Position:301.30131;Intensity:500;Charge:1),
(Position:301.38355;Intensity:500;Charge:1),
(Position:301.819;Intensity:120;Charge:2),
(Position:301.82337;Intensity:500;Charge:1),
(Position:302.7794;Intensity:150;Charge:1),
(Position:302.7794;Intensity:150;Charge:1),
(Position:302.790;Intensity:300;Charge:2),
(Position:302.9306;Intensity:30;Charge:1),
(Position:302.9306;Intensity:30;Charge:1),
(Position:302.974;Intensity:300;Charge:2),
(Position:303.12057;Intensity:500;Charge:1),
(Position:303.167;Intensity:100;Charge:2),
(Position:303.534;Intensity:120;Charge:2),
(Position:303.680;Intensity:100;Charge:2),
(Position:303.860;Intensity:100;Charge:2),
(Position:304.047;Intensity:100;Charge:2),
(Position:304.498;Intensity:100;Charge:2),
(Position:304.645;Intensity:100;Charge:2),
(Position:304.68332;Intensity:70;Charge:1),
(Position:304.845;Intensity:120;Charge:2),
(Position:305.523;Intensity:100;Charge:2),
(Position:305.557;Intensity:100;Charge:2),
(Position:305.82797;Intensity:3000;Charge:1),
(Position:306.000;Intensity:100;Charge:2),
(Position:306.338;Intensity:100;Charge:2),
(Position:306.419;Intensity:100;Charge:2),
(Position:306.45858;Intensity:2000;Charge:1),
(Position:306.6559;Intensity:50;Charge:1),
(Position:306.8105;Intensity:250;Charge:1),
(Position:306.8105;Intensity:250;Charge:1),
(Position:307.178;Intensity:100;Charge:2),
(Position:307.242;Intensity:100;Charge:2),
(Position:307.662;Intensity:100;Charge:2),
(Position:307.78700;Intensity:2000;Charge:1),
(Position:307.97750;Intensity:1000;Charge:1),
(Position:308.00753;Intensity:1000;Charge:1),
(Position:308.907;Intensity:120;Charge:2),
(Position:309.299;Intensity:100;Charge:2),
(Position:309.380;Intensity:120;Charge:2),
(Position:309.491;Intensity:100;Charge:2),
(Position:309.600;Intensity:100;Charge:2),
(Position:309.803;Intensity:100;Charge:2),
(Position:310.2307;Intensity:70;Charge:1),
(Position:311.888;Intensity:100;Charge:2),
(Position:311.906;Intensity:120;Charge:2),
(Position:312.71028;Intensity:2000;Charge:1),
(Position:314.224;Intensity:300;Charge:2),
(Position:314.463;Intensity:100;Charge:2),
(Position:314.86252;Intensity:250;Charge:1),
(Position:314.95226;Intensity:1000;Charge:1),
(Position:314.959;Intensity:100;Charge:2),
(Position:315.43238;Intensity:1000;Charge:1),
(Position:316.535;Intensity:100;Charge:2),
(Position:316.657;Intensity:100;Charge:2),
(Position:316.84929;Intensity:500;Charge:1),
(Position:318.966;Intensity:100;Charge:2),
(Position:319.550;Intensity:120;Charge:2),
(Position:319.951;Intensity:500;Charge:2),
(Position:320.7125;Intensity:10;Charge:1),
(Position:320.8833;Intensity:100;Charge:1),
(Position:320.8833;Intensity:100;Charge:1),
(Position:320.989;Intensity:60;Charge:2),
(Position:321.029;Intensity:120;Charge:2),
(Position:321.467;Intensity:120;Charge:2),
(Position:321.526;Intensity:150;Charge:2),
(Position:321.912;Intensity:150;Charge:2),
(Position:322.575;Intensity:120;Charge:2),
(Position:323.050;Intensity:120;Charge:2),
(Position:323.100;Intensity:200;Charge:2),
(Position:323.135;Intensity:120;Charge:2),
(Position:323.295;Intensity:120;Charge:2),
(Position:323.330;Intensity:150;Charge:2),
(Position:324.434;Intensity:100;Charge:2),
(Position:324.504;Intensity:100;Charge:2),
(Position:324.928;Intensity:100;Charge:2),
(Position:325.130;Intensity:100;Charge:2),
(Position:329.868;Intensity:150;Charge:2),
(Position:331.069;Intensity:150;Charge:2),
(Position:332.068;Intensity:300;Charge:2),
(Position:332.470;Intensity:1000;Charge:2),
(Position:332.811;Intensity:150;Charge:2),
(Position:333.012;Intensity:100;Charge:2),
(Position:333.580;Intensity:200;Charge:2),
(Position:334.536;Intensity:150;Charge:2),
(Position:334.641;Intensity:300;Charge:2),
(Position:334.679;Intensity:150;Charge:2),
(Position:335.27129;Intensity:250;Charge:1),
(Position:335.598;Intensity:200;Charge:2),
(Position:335.878;Intensity:120;Charge:2),
(Position:336.157;Intensity:200;Charge:2),
(Position:336.313;Intensity:120;Charge:2),
(Position:336.368;Intensity:100;Charge:2),
(Position:336.819;Intensity:120;Charge:2),
(Position:337.07756;Intensity:5000;Charge:1),
(Position:337.08750;Intensity:7000;Charge:1),
(Position:337.277;Intensity:100;Charge:2),
(Position:337.66184;Intensity:500;Charge:1),
(Position:337.919;Intensity:500;Charge:2),
(Position:338.939;Intensity:150;Charge:2),
(Position:338.991;Intensity:120;Charge:2),
(Position:339.377;Intensity:300;Charge:2),
(Position:340.580;Intensity:100;Charge:2),
(Position:340.793;Intensity:120;Charge:2),
(Position:341.413;Intensity:100;Charge:2),
(Position:341.789;Intensity:120;Charge:2),
(Position:341.867;Intensity:120;Charge:2),
(Position:341.88834;Intensity:5000;Charge:1),
(Position:341.89856;Intensity:500;Charge:1),
(Position:342.48939;Intensity:500;Charge:1),
(Position:342.967;Intensity:120;Charge:2),
(Position:344.86902;Intensity:2000;Charge:1),
(Position:345.17528;Intensity:500;Charge:1),
(Position:345.51838;Intensity:1000;Charge:1),
(Position:345.760;Intensity:100;Charge:2),
(Position:346.031;Intensity:100;Charge:2),
(Position:346.15147;Intensity:1000;Charge:1),
(Position:346.53307;Intensity:1000;Charge:1),
(Position:346.75709;Intensity:2000;Charge:1),
(Position:347.35649;Intensity:5000;Charge:1),
(Position:348.052;Intensity:150;Charge:2),
(Position:348.172;Intensity:200;Charge:2),
(Position:348.293;Intensity:200;Charge:2),
(Position:349.90641;Intensity:1000;Charge:1),
(Position:350.22171;Intensity:2000;Charge:1),
(Position:351.17248;Intensity:500;Charge:1),
(Position:351.61953;Intensity:2000;Charge:1),
(Position:352.14780;Intensity:10000;Charge:1),
(Position:354.386;Intensity:120;Charge:2),
(Position:355.882;Intensity:120;Charge:2),
(Position:356.222;Intensity:100;Charge:2),
(Position:356.39727;Intensity:150;Charge:1),
(Position:356.952;Intensity:250;Charge:2),
(Position:357.520;Intensity:100;Charge:2),
(Position:357.563;Intensity:200;Charge:2),
(Position:359.45517;Intensity:5000;Charge:1),
(Position:359.46640;Intensity:3000;Charge:1),
(Position:360.11966;Intensity:1000;Charge:1),
(Position:361.02082;Intensity:500;Charge:1),
(Position:363.47001;Intensity:1000;Charge:1),
(Position:364.497;Intensity:150;Charge:2),
(Position:366.511;Intensity:200;Charge:2),
(Position:368.32905;Intensity:1000;Charge:1),
(Position:368.67844;Intensity:1000;Charge:1),
(Position:369.526;Intensity:200;Charge:2),
(Position:370.22780;Intensity:400;Charge:1),
(Position:371.068;Intensity:150;Charge:2),
(Position:371.414;Intensity:250;Charge:2),
(Position:372.817;Intensity:250;Charge:2),
(Position:375.52818;Intensity:500;Charge:1),
(Position:376.6889;Intensity:50;Charge:1),
(Position:376.733;Intensity:800;Charge:2),
(Position:376.9118;Intensity:50;Charge:1),
(Position:377.0520;Intensity:70;Charge:1),
(Position:377.0725;Intensity:50;Charge:1),
(Position:377.820;Intensity:1000;Charge:2),
(Position:381.951;Intensity:100;Charge:2),
(Position:383.084;Intensity:120;Charge:2),
(Position:388.3798;Intensity:20;Charge:1),
(Position:388.8236;Intensity:10;Charge:1),
(Position:389.0529;Intensity:50;Charge:1),
(Position:390.0828;Intensity:20;Charge:1),
(Position:394.4656;Intensity:20;Charge:1),
(Position:398.5192;Intensity:20;Charge:1),
(Position:398.5380;Intensity:70;Charge:1),
(Position:399.6428;Intensity:10;Charge:1),
(Position:399.6851;Intensity:10;Charge:1),
(Position:399.9725;Intensity:10;Charge:1),
(Position:400.0394;Intensity:10;Charge:1),
(Position:401.4887;Intensity:10;Charge:1),
(Position:401.5130;Intensity:20;Charge:1),
(Position:402.1151;Intensity:20;Charge:1),
(Position:403.8403;Intensity:50;Charge:1),
(Position:403.8756;Intensity:150;Charge:1),
(Position:403.8837;Intensity:50;Charge:1),
(Position:404.3469;Intensity:100;Charge:1),
(Position:404.3784;Intensity:500;Charge:1),
(Position:404.6805;Intensity:20;Charge:1),
(Position:406.5184;Intensity:500;Charge:1),
(Position:406.5977;Intensity:150;Charge:1),
(Position:406.9984;Intensity:300;Charge:1),
(Position:407.0392;Intensity:300;Charge:1),
(Position:407.0538;Intensity:50;Charge:1),
(Position:408.0511;Intensity:20;Charge:1),
(Position:408.1300;Intensity:500;Charge:1),
(Position:411.3042;Intensity:10;Charge:1),
(Position:411.3260;Intensity:150;Charge:1),
(Position:411.3854;Intensity:200;Charge:1),
(Position:411.4026;Intensity:100;Charge:1),
(Position:412.8105;Intensity:20;Charge:1),
(Position:412.9237;Intensity:300;Charge:1),
(Position:413.1677;Intensity:200;Charge:1),
(Position:413.22266;Intensity:700;Charge:1),
(Position:416.59820;Intensity:500;Charge:1),
(Position:416.7265;Intensity:300;Charge:1),
(Position:417.5143;Intensity:20;Charge:1),
(Position:417.55433;Intensity:700;Charge:1),
(Position:417.63966;Intensity:600;Charge:1),
(Position:417.6665;Intensity:400;Charge:1),
(Position:419.7597;Intensity:150;Charge:1),
(Position:419.92847;Intensity:700;Charge:1),
(Position:420.4454;Intensity:20;Charge:1),
(Position:420.4454;Intensity:20;Charge:1),
(Position:420.4454;Intensity:20;Charge:1),
(Position:422.093;Intensity:150;Charge:2),
(Position:423.3515;Intensity:10;Charge:1),
(Position:423.504;Intensity:100;Charge:2),
(Position:425.0734;Intensity:20;Charge:1),
(Position:425.185;Intensity:120;Charge:2),
(Position:425.3615;Intensity:20;Charge:1),
(Position:425.3972;Intensity:20;Charge:1),
(Position:425.7696;Intensity:20;Charge:1),
(Position:425.7696;Intensity:20;Charge:1),
(Position:426.0938;Intensity:10;Charge:1),
(Position:426.3679;Intensity:20;Charge:1),
(Position:426.8487;Intensity:10;Charge:1),
(Position:426.8925;Intensity:50;Charge:1),
(Position:426.92098;Intensity:700;Charge:1),
(Position:427.09240;Intensity:700;Charge:1),
(Position:427.14270;Intensity:500;Charge:1),
(Position:427.58647;Intensity:500;Charge:1),
(Position:427.6370;Intensity:10;Charge:1),
(Position:427.67622;Intensity:700;Charge:1),
(Position:428.0054;Intensity:50;Charge:1),
(Position:428.0483;Intensity:150;Charge:1),
(Position:428.4447;Intensity:100;Charge:1),
(Position:428.9748;Intensity:50;Charge:1),
(Position:429.1006;Intensity:20;Charge:1),
(Position:429.3184;Intensity:20;Charge:1),
(Position:429.3184;Intensity:20;Charge:1),
(Position:429.3184;Intensity:20;Charge:1),
(Position:430.4458;Intensity:300;Charge:1),
(Position:430.4906;Intensity:10;Charge:1),
(Position:430.5166;Intensity:50;Charge:1),
(Position:430.74621;Intensity:700;Charge:1),
(Position:431.1342;Intensity:20;Charge:1),
(Position:431.5323;Intensity:10;Charge:1),
(Position:431.5909;Intensity:300;Charge:1),
(Position:431.7222;Intensity:150;Charge:1),
(Position:432.0049;Intensity:50;Charge:1),
(Position:432.0726;Intensity:10;Charge:1),
(Position:432.2707;Intensity:20;Charge:1),
(Position:432.8482;Intensity:100;Charge:1),
(Position:433.53453;Intensity:700;Charge:1),
(Position:433.74460;Intensity:500;Charge:1),
(Position:433.9420;Intensity:20;Charge:1),
(Position:434.1476;Intensity:20;Charge:1),
(Position:434.1640;Intensity:20;Charge:1),
(Position:434.6701;Intensity:20;Charge:1),
(Position:434.6701;Intensity:20;Charge:1),
(Position:434.6984;Intensity:10;Charge:1),
(Position:434.7258;Intensity:150;Charge:1),
(Position:435.8523;Intensity:20;Charge:1),
(Position:435.9143;Intensity:50;Charge:1),
(Position:436.0041;Intensity:20;Charge:1),
(Position:436.3916;Intensity:300;Charge:1),
(Position:436.4454;Intensity:20;Charge:1),
(Position:436.4750;Intensity:700;Charge:1),
(Position:436.6932;Intensity:10;Charge:1),
(Position:437.109;Intensity:120;Charge:2),
(Position:437.3025;Intensity:20;Charge:1),
(Position:437.3386;Intensity:300;Charge:1),
(Position:437.6226;Intensity:20;Charge:1),
(Position:437.6226;Intensity:20;Charge:1),
(Position:437.8984;Intensity:20;Charge:1),
(Position:437.8984;Intensity:20;Charge:1),
(Position:438.063;Intensity:70;Charge:2),
(Position:438.078;Intensity:150;Charge:2),
(Position:438.2451;Intensity:300;Charge:1),
(Position:438.629;Intensity:100;Charge:2),
(Position:439.322;Intensity:200;Charge:2),
(Position:439.5604;Intensity:150;Charge:1),
(Position:439.5604;Intensity:150;Charge:1),
(Position:439.5604;Intensity:150;Charge:1),
(Position:439.6008;Intensity:150;Charge:1),
(Position:439.6243;Intensity:10;Charge:1),
(Position:439.6541;Intensity:10;Charge:1),
(Position:439.6791;Intensity:500;Charge:1),
(Position:439.7204;Intensity:10;Charge:1),
(Position:439.8410;Intensity:10;Charge:1),
(Position:439.923;Intensity:150;Charge:2),
(Position:439.9371;Intensity:50;Charge:1),
(Position:440.3611;Intensity:20;Charge:1),
(Position:440.3817;Intensity:10;Charge:1),
(Position:440.4222;Intensity:10;Charge:1),
(Position:440.6819;Intensity:20;Charge:1),
(Position:441.054;Intensity:150;Charge:2),
(Position:441.0858;Intensity:200;Charge:1),
(Position:441.3524;Intensity:200;Charge:1),
(Position:441.446;Intensity:100;Charge:2),
(Position:441.4801;Intensity:150;Charge:1),
(Position:441.6381;Intensity:50;Charge:1),
(Position:441.8057;Intensity:500;Charge:1),
(Position:442.1799;Intensity:10;Charge:1),
(Position:442.263;Intensity:100;Charge:2),
(Position:442.27969;Intensity:500;Charge:1),
(Position:442.37624;Intensity:3000;Charge:1),
(Position:442.60490;Intensity:3000;Charge:1),
(Position:442.6643;Intensity:1500;Charge:1),
(Position:442.8998;Intensity:300;Charge:1),
(Position:442.9224;Intensity:150;Charge:1),
(Position:442.976;Intensity:100;Charge:2),
(Position:442.987;Intensity:100;Charge:2),
(Position:443.0654;Intensity:10;Charge:1),
(Position:443.214;Intensity:150;Charge:2),
(Position:443.218;Intensity:150;Charge:2),
(Position:443.37611;Intensity:200;Charge:1),
(Position:443.4643;Intensity:100;Charge:1),
(Position:443.49688;Intensity:700;Charge:1),
(Position:443.6339;Intensity:50;Charge:1),
(Position:444.1610;Intensity:150;Charge:1),
(Position:444.1610;Intensity:150;Charge:1),
(Position:444.2059;Intensity:20;Charge:1),
(Position:444.2137;Intensity:10;Charge:1),
(Position:444.6226;Intensity:300;Charge:1),
(Position:444.6919;Intensity:10;Charge:1),
(Position:444.7786;Intensity:10;Charge:1),
(Position:445.4233;Intensity:150;Charge:1),
(Position:445.4503;Intensity:50;Charge:1),
(Position:445.4574;Intensity:20;Charge:1),
(Position:445.4778;Intensity:10;Charge:1),
(Position:445.5535;Intensity:50;Charge:1),
(Position:445.6815;Intensity:150;Charge:1),
(Position:445.7631;Intensity:10;Charge:1),
(Position:445.830;Intensity:120;Charge:2),
(Position:446.1427;Intensity:1000;Charge:1),
(Position:446.4109;Intensity:20;Charge:1),
(Position:446.69077;Intensity:500;Charge:1),
(Position:446.7298;Intensity:50;Charge:1),
(Position:446.7756;Intensity:20;Charge:1),
(Position:446.80656;Intensity:700;Charge:1),
(Position:446.8745;Intensity:10;Charge:1),
(Position:447.2226;Intensity:50;Charge:1),
(Position:447.2226;Intensity:50;Charge:1),
(Position:447.3501;Intensity:10;Charge:1),
(Position:447.3501;Intensity:10;Charge:1),
(Position:447.6387;Intensity:50;Charge:1),
(Position:447.6912;Intensity:1000;Charge:1),
(Position:448.2080;Intensity:150;Charge:1),
(Position:448.4448;Intensity:1500;Charge:1),
(Position:448.93518;Intensity:3000;Charge:1),
(Position:449.30318;Intensity:800;Charge:1),
(Position:449.3098;Intensity:500;Charge:1),
(Position:449.3392;Intensity:50;Charge:1),
(Position:449.3672;Intensity:300;Charge:1),
(Position:449.3949;Intensity:150;Charge:1),
(Position:449.4369;Intensity:50;Charge:1),
(Position:449.4960;Intensity:500;Charge:1),
(Position:450.0262;Intensity:20;Charge:1),
(Position:450.1105;Intensity:50;Charge:1),
(Position:450.1105;Intensity:50;Charge:1),
(Position:450.1444;Intensity:500;Charge:1),
(Position:451.1435;Intensity:150;Charge:1),
(Position:451.2774;Intensity:200;Charge:1),
(Position:451.2774;Intensity:200;Charge:1),
(Position:451.6157;Intensity:700;Charge:1),
(Position:451.6288;Intensity:20;Charge:1),
(Position:451.6677;Intensity:300;Charge:1),
(Position:451.8203;Intensity:500;Charge:1),
(Position:451.9003;Intensity:1000;Charge:1),
(Position:452.399;Intensity:100;Charge:2),
(Position:452.7033;Intensity:700;Charge:1),
(Position:452.7446;Intensity:500;Charge:1),
(Position:452.7954;Intensity:150;Charge:1),
(Position:452.7954;Intensity:150;Charge:1),
(Position:452.8995;Intensity:150;Charge:1),
(Position:452.9243;Intensity:10;Charge:1),
(Position:453.0746;Intensity:300;Charge:1),
(Position:453.3666;Intensity:10;Charge:1),
(Position:453.75722;Intensity:1500;Charge:1),
(Position:453.89491;Intensity:3000;Charge:1),
(Position:453.90268;Intensity:10000;Charge:1),
(Position:453.95750;Intensity:3000;Charge:1),
(Position:454.0441;Intensity:500;Charge:1),
(Position:454.16531;Intensity:500;Charge:1),
(Position:454.5776;Intensity:500;Charge:1),
(Position:454.7003;Intensity:10;Charge:1),
(Position:454.8493;Intensity:100;Charge:1),
(Position:454.9003;Intensity:150;Charge:1),
(Position:455.1916;Intensity:10;Charge:1),
(Position:455.3874;Intensity:300;Charge:1),
(Position:455.5692;Intensity:100;Charge:1),
(Position:455.5838;Intensity:50;Charge:1),
(Position:455.6101;Intensity:400;Charge:1),
(Position:455.6669;Intensity:300;Charge:1),
(Position:455.7975;Intensity:20;Charge:1),
(Position:456.3728;Intensity:10;Charge:1),
(Position:456.7168;Intensity:600;Charge:1),
(Position:456.8110;Intensity:400;Charge:1),
(Position:456.8419;Intensity:150;Charge:1),
(Position:456.9125;Intensity:100;Charge:1),
(Position:457.034;Intensity:100;Charge:2),
(Position:457.43425;Intensity:50;Charge:1),
(Position:457.4839;Intensity:500;Charge:1),
(Position:457.5180;Intensity:300;Charge:1),
(Position:457.63442;Intensity:3000;Charge:1),
(Position:457.7140;Intensity:200;Charge:1),
(Position:458.3319;Intensity:1500;Charge:1),
(Position:458.3389;Intensity:150;Charge:1),
(Position:458.37362;Intensity:1500;Charge:1),
(Position:458.3840;Intensity:150;Charge:1),
(Position:458.4264;Intensity:50;Charge:1),
(Position:458.7161;Intensity:100;Charge:1),
(Position:458.7430;Intensity:20;Charge:1),
(Position:459.4530;Intensity:500;Charge:1),
(Position:459.6536;Intensity:500;Charge:1),
(Position:460.5385;Intensity:150;Charge:1),
(Position:460.5970;Intensity:10;Charge:1),
(Position:460.6228;Intensity:50;Charge:1),
(Position:461.0656;Intensity:300;Charge:1),
(Position:461.1201;Intensity:1500;Charge:1),
(Position:461.5684;Intensity:1000;Charge:1),
(Position:461.8204;Intensity:50;Charge:1),
(Position:461.9130;Intensity:700;Charge:1),
(Position:462.9095;Intensity:20;Charge:1),
(Position:462.96075;Intensity:1500;Charge:1),
(Position:462.9756;Intensity:300;Charge:1),
(Position:463.7423;Intensity:700;Charge:1),
(Position:463.79346;Intensity:700;Charge:1),
(Position:463.8273;Intensity:500;Charge:1),
(Position:464.0890;Intensity:300;Charge:1),
(Position:464.1742;Intensity:700;Charge:1),
(Position:464.4482;Intensity:50;Charge:1),
(Position:464.5231;Intensity:20;Charge:1),
(Position:464.6134;Intensity:400;Charge:1),
(Position:464.67188;Intensity:3000;Charge:1),
(Position:464.7186;Intensity:10;Charge:1),
(Position:465.1206;Intensity:700;Charge:1),
(Position:465.3404;Intensity:300;Charge:1),
(Position:465.5002;Intensity:500;Charge:1),
(Position:465.76973;Intensity:3000;Charge:1),
(Position:466.24103;Intensity:1500;Charge:1),
(Position:466.4397;Intensity:400;Charge:1),
(Position:466.4824;Intensity:200;Charge:1),
(Position:466.7960;Intensity:500;Charge:1),
(Position:466.86709;Intensity:1000;Charge:1),
(Position:467.2192;Intensity:700;Charge:1),
(Position:467.9527;Intensity:3000;Charge:1),
(Position:467.99202;Intensity:500;Charge:1),
(Position:468.0445;Intensity:1500;Charge:1),
(Position:468.16770;Intensity:1000;Charge:1),
(Position:468.2510;Intensity:500;Charge:1),
(Position:468.3240;Intensity:200;Charge:1),
(Position:468.3456;Intensity:200;Charge:1),
(Position:468.4221;Intensity:100;Charge:1),
(Position:468.4549;Intensity:50;Charge:1),
(Position:468.5075;Intensity:300;Charge:1),
(Position:468.8983;Intensity:1000;Charge:1),
(Position:468.9503;Intensity:20;Charge:1),
(Position:469.2893;Intensity:150;Charge:1),
(Position:469.6677;Intensity:200;Charge:1),
(Position:469.8257;Intensity:50;Charge:1),
(Position:470.1784;Intensity:50;Charge:1),
(Position:470.38464;Intensity:1500;Charge:1),
(Position:470.57113;Intensity:15000;Charge:1),
(Position:471.01770;Intensity:12000;Charge:1),
(Position:471.13829;Intensity:10000;Charge:1),
(Position:471.18061;Intensity:300;Charge:1),
(Position:471.33817;Intensity:15000;Charge:1),
(Position:471.34590;Intensity:150;Charge:1),
(Position:471.34590;Intensity:150;Charge:1),
(Position:471.4119;Intensity:100;Charge:1),
(Position:471.56587;Intensity:700;Charge:1),
(Position:471.64440;Intensity:300;Charge:1),
(Position:471.65772;Intensity:300;Charge:1),
(Position:471.6663;Intensity:15000;Charge:1),
(Position:471.89284;Intensity:700;Charge:1),
(Position:472.28546;Intensity:700;Charge:1),
(Position:472.3471;Intensity:50;Charge:1),
(Position:472.4035;Intensity:150;Charge:1),
(Position:472.5132;Intensity:700;Charge:1),
(Position:472.5132;Intensity:700;Charge:1),
(Position:472.5484;Intensity:50;Charge:1),
(Position:472.6467;Intensity:700;Charge:1),
(Position:475.09037;Intensity:3000;Charge:1),
(Position:475.20112;Intensity:300;Charge:1),
(Position:475.3131;Intensity:300;Charge:1),
(Position:475.40612;Intensity:5000;Charge:1),
(Position:475.4452;Intensity:10;Charge:1),
(Position:475.5770;Intensity:1000;Charge:1),
(Position:476.0059;Intensity:1500;Charge:1),
(Position:478.1674;Intensity:3000;Charge:1),
(Position:478.2221;Intensity:300;Charge:1),
(Position:478.2576;Intensity:20;Charge:1),
(Position:478.5359;Intensity:20;Charge:1),
(Position:479.02646;Intensity:10000;Charge:1),
(Position:479.09371;Intensity:1000;Charge:1),
(Position:479.15586;Intensity:5000;Charge:1),
(Position:479.2067;Intensity:300;Charge:1),
(Position:480.1453;Intensity:150;Charge:1),
(Position:480.2418;Intensity:20;Charge:1),
(Position:480.3705;Intensity:100;Charge:1),
(Position:480.4568;Intensity:10;Charge:1),
(Position:481.0844;Intensity:100;Charge:1),
(Position:481.14084;Intensity:1500;Charge:1),
(Position:481.19837;Intensity:1000;Charge:1),
(Position:481.5684;Intensity:500;Charge:1),
(Position:481.8246;Intensity:10;Charge:1),
(Position:481.89850;Intensity:3000;Charge:1),
(Position:482.01314;Intensity:1500;Charge:1),
(Position:482.1284;Intensity:700;Charge:1),
(Position:482.32693;Intensity:3000;Charge:1),
(Position:482.45203;Intensity:1000;Charge:1),
(Position:482.4718;Intensity:500;Charge:1),
(Position:482.6877;Intensity:500;Charge:1),
(Position:482.8687;Intensity:10000;Charge:1),
(Position:482.8936;Intensity:3000;Charge:1),
(Position:483.0637;Intensity:50;Charge:1),
(Position:483.86655;Intensity:5000;Charge:1),
(Position:484.3919;Intensity:100;Charge:1),
(Position:484.4294;Intensity:500;Charge:1),
(Position:484.6499;Intensity:150;Charge:1),
(Position:484.7121;Intensity:50;Charge:1),
(Position:485.0885;Intensity:300;Charge:1),
(Position:485.2856;Intensity:600;Charge:1),
(Position:485.4011;Intensity:1000;Charge:1),
(Position:486.0962;Intensity:150;Charge:1),
(Position:486.44394;Intensity:1000;Charge:1),
(Position:486.57317;Intensity:300;Charge:1),
(Position:486.68656;Intensity:1000;Charge:1),
(Position:486.7835;Intensity:800;Charge:1),
(Position:486.83784;Intensity:700;Charge:1),
(Position:486.96364;Intensity:700;Charge:1),
(Position:488.4767;Intensity:150;Charge:1),
(Position:488.62812;Intensity:10000;Charge:1),
(Position:488.64614;Intensity:1000;Charge:1),
(Position:488.9730;Intensity:50;Charge:1),
(Position:489.3456;Intensity:5000;Charge:1),
(Position:489.3594;Intensity:100;Charge:1),
(Position:489.9292;Intensity:700;Charge:1),
(Position:490.0381;Intensity:500;Charge:1),
(Position:492.9611;Intensity:700;Charge:1),
(Position:493.2320;Intensity:500;Charge:1),
(Position:494.0420;Intensity:1000;Charge:1),
(Position:494.6367;Intensity:1000;Charge:1),
(Position:495.67734;Intensity:1500;Charge:1),
(Position:495.84169;Intensity:10000;Charge:1),
(Position:495.8505;Intensity:1500;Charge:1),
(Position:497.49433;Intensity:1000;Charge:1),
(Position:497.6148;Intensity:500;Charge:1),
(Position:497.7349;Intensity:100;Charge:1),
(Position:498.1014;Intensity:50;Charge:1),
(Position:499.6323;Intensity:1500;Charge:1),
(Position:499.7603;Intensity:20;Charge:1),
(Position:499.8876;Intensity:150;Charge:1),
(Position:499.9896;Intensity:100;Charge:1),
(Position:500.1790;Intensity:30;Charge:1),
(Position:500.4957;Intensity:20;Charge:1),
(Position:500.65549;Intensity:5000;Charge:1),
(Position:500.67429;Intensity:500;Charge:1),
(Position:501.2401;Intensity:250;Charge:1),
(Position:501.6586;Intensity:50;Charge:1),
(Position:502.4271;Intensity:250;Charge:1),
(Position:503.27515;Intensity:2500;Charge:1),
(Position:503.29119;Intensity:20;Charge:1),
(Position:503.74060;Intensity:350;Charge:1),
(Position:503.89975;Intensity:30;Charge:1),
(Position:503.91560;Intensity:5000;Charge:1),
(Position:504.3004;Intensity:10;Charge:1),
(Position:504.4259;Intensity:150;Charge:1),
(Position:504.7223;Intensity:150;Charge:1),
(Position:504.8015;Intensity:30;Charge:1),
(Position:505.43532;Intensity:250;Charge:1),
(Position:506.0561;Intensity:20;Charge:1),
(Position:507.54604;Intensity:30;Charge:1),
(Position:507.56152;Intensity:350;Charge:1),
(Position:507.80123;Intensity:350;Charge:1),
(Position:508.01850;Intensity:150;Charge:1),
(Position:508.1799;Intensity:1500;Charge:1),
(Position:508.2776;Intensity:150;Charge:1),
(Position:508.53944;Intensity:250;Charge:1),
(Position:509.1740;Intensity:80;Charge:1),
(Position:510.04733;Intensity:250;Charge:1),
(Position:510.61237;Intensity:350;Charge:1),
(Position:511.50974;Intensity:750;Charge:1),
(Position:511.79290;Intensity:1500;Charge:1),
(Position:511.84505;Intensity:350;Charge:1),
(Position:512.19327;Intensity:250;Charge:1),
(Position:512.3293;Intensity:20;Charge:1),
(Position:512.36838;Intensity:1500;Charge:1),
(Position:512.37886;Intensity:1500;Charge:1),
(Position:512.9709;Intensity:20;Charge:1),
(Position:514.4698;Intensity:50;Charge:1),
(Position:514.63717;Intensity:5000;Charge:1),
(Position:514.6444;Intensity:5000;Charge:1),
(Position:514.65685;Intensity:350;Charge:1),
(Position:515.15189;Intensity:350;Charge:1),
(Position:515.33962;Intensity:750;Charge:1),
(Position:515.58630;Intensity:500;Charge:1),
(Position:515.81037;Intensity:500;Charge:1),
(Position:516.03389;Intensity:500;Charge:1),
(Position:516.49230;Intensity:100;Charge:1),
(Position:518.3763;Intensity:20;Charge:1),
(Position:519.00572;Intensity:1500;Charge:1),
(Position:519.27680;Intensity:350;Charge:1),
(Position:519.45713;Intensity:1500;Charge:1),
(Position:519.4670;Intensity:1500;Charge:1),
(Position:520.53452;Intensity:1500;Charge:1),
(Position:520.8015;Intensity:30;Charge:1),
(Position:521.03152;Intensity:700;Charge:1),
(Position:521.20180;Intensity:500;Charge:1),
(Position:521.57907;Intensity:350;Charge:1),
(Position:522.38057;Intensity:500;Charge:1),
(Position:523.54842;Intensity:500;Charge:1),
(Position:527.55070;Intensity:400;Charge:1),
(Position:528.15546;Intensity:500;Charge:1),
(Position:529.96633;Intensity:1500;Charge:1),
(Position:530.62339;Intensity:700;Charge:1),
(Position:531.62637;Intensity:300;Charge:1),
(Position:531.82837;Intensity:250;Charge:1),
(Position:532.2030;Intensity:20;Charge:1),
(Position:532.7878;Intensity:750;Charge:1),
(Position:533.21548;Intensity:5;Charge:1),
(Position:533.22604;Intensity:6000;Charge:1),
(Position:533.47918;Intensity:500;Charge:1),
(Position:533.7194;Intensity:100;Charge:1),
(Position:534.25794;Intensity:10000;Charge:1),
(Position:534.4186;Intensity:10;Charge:1),
(Position:534.44909;Intensity:100;Charge:1),
(Position:534.47696;Intensity:6000;Charge:1),
(Position:535.06916;Intensity:1500;Charge:1),
(Position:535.5002;Intensity:50;Charge:1),
(Position:535.66534;Intensity:1500;Charge:1),
(Position:535.68288;Intensity:50;Charge:1),
(Position:535.69130;Intensity:1500;Charge:1),
(Position:535.9510;Intensity:100;Charge:1),
(Position:536.15028;Intensity:1500;Charge:1),
(Position:536.19179;Intensity:350;Charge:1),
(Position:536.37247;Intensity:250;Charge:1),
(Position:536.77092;Intensity:250;Charge:1),
(Position:537.38049;Intensity:750;Charge:1),
(Position:537.6470;Intensity:500;Charge:1),
(Position:538.47426;Intensity:250;Charge:1),
(Position:540.20631;Intensity:20000;Charge:1),
(Position:541.162;Intensity:50;Charge:1),
(Position:541.41537;Intensity:2500;Charge:1),
(Position:542.00647;Intensity:1500;Charge:1),
(Position:542.1662;Intensity:500;Charge:1),
(Position:543.51616;Intensity:2500;Charge:1),
(Position:544.8634;Intensity:80;Charge:1),
(Position:545.00234;Intensity:1500;Charge:1),
(Position:549.59423;Intensity:500;Charge:1),
(Position:550.88742;Intensity:250;Charge:1),
(Position:551.2707;Intensity:30;Charge:1),
(Position:551.3016;Intensity:150;Charge:1),
(Position:552.216;Intensity:30;Charge:1),
(Position:553.52158;Intensity:750;Charge:1),
(Position:554.01894;Intensity:500;Charge:1),
(Position:556.06416;Intensity:350;Charge:1),
(Position:556.39863;Intensity:1500;Charge:1),
(Position:556.43110;Intensity:5000;Charge:1),
(Position:556.45980;Intensity:750;Charge:1),
(Position:557.75877;Intensity:350;Charge:1),
(Position:558.7456;Intensity:50;Charge:1),
(Position:559.08991;Intensity:500;Charge:1),
(Position:559.270;Intensity:80;Charge:1),
(Position:565.41352;Intensity:750;Charge:1),
(Position:565.75955;Intensity:750;Charge:1),
(Position:565.82287;Intensity:5000;Charge:1),
(Position:566.41204;Intensity:750;Charge:1),
(Position:568.6224;Intensity:250;Charge:1),
(Position:569.13951;Intensity:1500;Charge:1),
(Position:571.69265;Intensity:350;Charge:1),
(Position:572.04663;Intensity:1500;Charge:1),
(Position:572.08114;Intensity:5000;Charge:1),
(Position:572.11167;Intensity:750;Charge:1),
(Position:574.98929;Intensity:5000;Charge:1),
(Position:575.02391;Intensity:700;Charge:1),
(Position:576.21862;Intensity:700;Charge:1),
(Position:576.56511;Intensity:30;Charge:1),
(Position:576.60175;Intensity:7000;Charge:1),
(Position:577.19070;Intensity:500;Charge:1),
(Position:580.56993;Intensity:750;Charge:1),
(Position:580.60590;Intensity:5000;Charge:1),
(Position:581.30179;Intensity:3000;Charge:1),
(Position:581.82346;Intensity:500;Charge:1),
(Position:582.17694;Intensity:5000;Charge:1),
(Position:583.05223;Intensity:750;Charge:1),
(Position:585.41101;Intensity:20000;Charge:1),
(Position:587.00431;Intensity:750;Charge:1),
(Position:587.37726;Intensity:750;Charge:1),
(Position:587.44552;Intensity:5000;Charge:1),
(Position:588.35252;Intensity:10000;Charge:1),
(Position:589.99633;Intensity:200;Charge:1),
(Position:590.37300;Intensity:30;Charge:1),
(Position:590.40980;Intensity:500;Charge:1),
(Position:590.44193;Intensity:50;Charge:1),
(Position:590.80661;Intensity:500;Charge:1),
(Position:591.5272;Intensity:2500;Charge:1),
(Position:592.05469;Intensity:2500;Charge:1),
(Position:592.06691;Intensity:80;Charge:1),
(Position:593.60964;Intensity:750;Charge:1),
(Position:594.09609;Intensity:500;Charge:1),
(Position:594.64810;Intensity:5000;Charge:1),
(Position:596.32743;Intensity:700;Charge:1),
(Position:596.71236;Intensity:5000;Charge:1),
(Position:596.78317;Intensity:350;Charge:1),
(Position:597.62823;Intensity:5000;Charge:1),
(Position:597.71896;Intensity:6000;Charge:1),
(Position:598.40324;Intensity:80;Charge:1),
(Position:598.95660;Intensity:1500;Charge:1),
(Position:599.33073;Intensity:750;Charge:1),
(Position:600.25896;Intensity:1000;Charge:1),
(Position:603.16667;Intensity:10000;Charge:1),
(Position:604.3686;Intensity:150;Charge:1),
(Position:604.78090;Intensity:500;Charge:1),
(Position:606.62150;Intensity:500;Charge:1),
(Position:607.60193;Intensity:10000;Charge:1),
(Position:609.78506;Intensity:3000;Charge:1),
(Position:611.97122;Intensity:150;Charge:1),
(Position:613.01461;Intensity:1000;Charge:1),
(Position:614.4208;Intensity:1000;Charge:1),
(Position:614.47629;Intensity:10000;Charge:1),
(Position:615.20006;Intensity:1000;Charge:1),
(Position:615.78417;Intensity:500;Charge:1),
(Position:616.52994;Intensity:10000;Charge:1),
(Position:617.45238;Intensity:150;Charge:1),
(Position:617.65916;Intensity:700;Charge:1),
(Position:617.69930;Intensity:500;Charge:1),
(Position:618.38567;Intensity:1500;Charge:1),
(Position:618.48684;Intensity:50;Charge:1),
(Position:619.07774;Intensity:700;Charge:1),
(Position:619.47799;Intensity:500;Charge:1),
(Position:620.46902;Intensity:150;Charge:1),
(Position:620.74945;Intensity:1000;Charge:1),
(Position:621.55950;Intensity:1500;Charge:1),
(Position:621.90013;Intensity:10000;Charge:1),
(Position:622.74574;Intensity:500;Charge:1),
(Position:624.84574;Intensity:1000;Charge:1),
(Position:625.1322;Intensity:50;Charge:1),
(Position:625.4462;Intensity:20;Charge:1),
(Position:626.05196;Intensity:1000;Charge:1),
(Position:626.82285;Intensity:10000;Charge:1),
(Position:627.47491;Intensity:700;Charge:1),
(Position:627.77686;Intensity:500;Charge:1),
(Position:629.54853;Intensity:1000;Charge:1),
(Position:630.65329;Intensity:1000;Charge:1),
(Position:631.54315;Intensity:1500;Charge:1),
(Position:632.99145;Intensity:3000;Charge:1),
(Position:633.26400;Intensity:1500;Charge:1),
(Position:633.61791;Intensity:10000;Charge:1),
(Position:635.36094;Intensity:1000;Charge:1),
(Position:636.67560;Intensity:1000;Charge:1),
(Position:638.47560;Intensity:10000;Charge:1),
(Position:640.2845;Intensity:1000;Charge:1),
(Position:640.4018;Intensity:20000;Charge:1),
(Position:641.15187;Intensity:1500;Charge:1),
(Position:642.34794;Intensity:1000;Charge:1),
(Position:644.64930;Intensity:1500;Charge:1),
(Position:650.83255;Intensity:15000;Charge:1),
(Position:653.46872;Intensity:1000;Charge:1),
(Position:660.07754;Intensity:10000;Charge:1),
(Position:660.47244;Intensity:1000;Charge:1),
(Position:664.18432;Intensity:100;Charge:1),
(Position:664.263;Intensity:50;Charge:1),
(Position:665.39294;Intensity:1500;Charge:1),
(Position:666.8733;Intensity:1000;Charge:1),
(Position:668.01205;Intensity:5000;Charge:1),
(Position:671.88974;Intensity:700;Charge:1),
(Position:672.29897;Intensity:20;Charge:1),
(Position:673.98920;Intensity:700;Charge:1),
(Position:676.14479;Intensity:150;Charge:1),
(Position:693.13787;Intensity:100000;Charge:1),
(Position:702.59870;Intensity:34000;Charge:1),
(Position:703.43521;Intensity:85000;Charge:1),
(Position:705.32365;Intensity:2200;Charge:1),
(Position:706.10543;Intensity:10000;Charge:1),
(Position:706.6710;Intensity:80;Charge:1),
(Position:711.42683;Intensity:110;Charge:1),
(Position:714.05078;Intensity:55;Charge:1),
(Position:717.59154;Intensity:77000;Charge:1),
(Position:721.519;Intensity:150;Charge:2),
(Position:723.718;Intensity:150;Charge:2),
(Position:724.71631;Intensity:77000;Charge:1),
(Position:730.68549;Intensity:89;Charge:1),
(Position:734.596;Intensity:150;Charge:2),
(Position:744.09469;Intensity:60000;Charge:1),
(Position:747.44962;Intensity:3100;Charge:1),
(Position:749.09335;Intensity:32000;Charge:1),
(Position:749.416;Intensity:100;Charge:2),
(Position:752.489;Intensity:150;Charge:2),
(Position:753.78488;Intensity:28000;Charge:1),
(Position:754.61211;Intensity:13000;Charge:1),
(Position:762.343;Intensity:50;Charge:1),
(Position:772.67492;Intensity:60;Charge:1),
(Position:774.287;Intensity:120;Charge:2),
(Position:783.51833;Intensity:56;Charge:1),
(Position:784.12088;Intensity:230;Charge:1),
(Position:784.21426;Intensity:7;Charge:1),
(Position:792.838;Intensity:120;Charge:2),
(Position:792.92978;Intensity:300;Charge:1),
(Position:793.91790;Intensity:1300;Charge:1),
(Position:794.53654;Intensity:7900;Charge:1),
(Position:794.63256;Intensity:200;Charge:1),
(Position:802.632;Intensity:3;Charge:1),
(Position:804.400;Intensity:3;Charge:1),
(Position:807.828;Intensity:1;Charge:1),
(Position:808.46801;Intensity:5700;Charge:1),
(Position:808.656;Intensity:100;Charge:2),
(Position:809.531;Intensity:3;Charge:1),
(Position:812.07818;Intensity:3800;Charge:1),
(Position:813.11461;Intensity:1200;Charge:1),
(Position:813.86432;Intensity:17000;Charge:1),
(Position:825.09501;Intensity:310;Charge:1),
(Position:826.16499;Intensity:3300;Charge:1),
(Position:826.708;Intensity:100;Charge:2),
(Position:826.83491;Intensity:7200;Charge:1),
(Position:826.93890;Intensity:990;Charge:1),
(Position:830.26062;Intensity:29000;Charge:1),
(Position:830.38388;Intensity:1900;Charge:1),
(Position:831.729;Intensity:100;Charge:2),
(Position:836.80455;Intensity:4600;Charge:1),
(Position:837.441;Intensity:100;Charge:2),
(Position:837.86610;Intensity:6600;Charge:1),
(Position:837.99093;Intensity:76000;Charge:1),
(Position:841.94727;Intensity:2700;Charge:1),
(Position:842.07398;Intensity:26000;Charge:1),
(Position:846.56824;Intensity:3700;Charge:1),
(Position:848.67736;Intensity:1300;Charge:1),
(Position:849.76932;Intensity:69000;Charge:1),
(Position:854.70427;Intensity:1600;Charge:1),
(Position:857.37082;Intensity:2900;Charge:1),
(Position:858.52609;Intensity:1600;Charge:1),
(Position:859.36184;Intensity:41000;Charge:1),
(Position:863.70190;Intensity:35000;Charge:1),
(Position:863.76897;Intensity:740;Charge:1),
(Position:864.94164;Intensity:6000;Charge:1),
(Position:865.67600;Intensity:64000;Charge:1),
(Position:865.78995;Intensity:7600;Charge:1),
(Position:867.064;Intensity:100;Charge:2),
(Position:868.18775;Intensity:13000;Charge:1),
(Position:868.43062;Intensity:15000;Charge:1),
(Position:870.65028;Intensity:2900;Charge:1),
(Position:876.9944;Intensity:160;Charge:1),
(Position:877.40664;Intensity:10000;Charge:1),
(Position:878.11430;Intensity:2100;Charge:1),
(Position:878.30336;Intensity:57000;Charge:1),
(Position:878.44132;Intensity:230;Charge:1),
(Position:878.61661;Intensity:43000;Charge:1),
(Position:879.49201;Intensity:260;Charge:1),
(Position:883.33316;Intensity:550;Charge:1),
(Position:885.62980;Intensity:27000;Charge:1),
(Position:886.77399;Intensity:2100;Charge:1),
(Position:886.81905;Intensity:15000;Charge:1),
(Position:889.46730;Intensity:12;Charge:1),
(Position:889.80;Intensity:3;Charge:1),
(Position:891.789;Intensity:4;Charge:1),
(Position:892.19496;Intensity:6400;Charge:1),
(Position:892.99;Intensity:3;Charge:1),
(Position:893.17018;Intensity:18;Charge:1),
(Position:894.39681;Intensity:8;Charge:1),
(Position:896.4788;Intensity:4;Charge:1),
(Position:897.11;Intensity:3;Charge:1),
(Position:899.10240;Intensity:1800;Charge:1),
(Position:903.94792;Intensity:9;Charge:1),
(Position:904.93;Intensity:1;Charge:1),
(Position:904.93;Intensity:1;Charge:1),
(Position:905.1570;Intensity:4;Charge:1),
(Position:905.4909;Intensity:3;Charge:1),
(Position:905.5127;Intensity:7;Charge:1),
(Position:907.5523;Intensity:5;Charge:1),
(Position:908.195;Intensity:100;Charge:2),
(Position:910.46;Intensity:1;Charge:1),
(Position:910.603;Intensity:4;Charge:1),
(Position:911.78;Intensity:1;Charge:1),
(Position:911.78;Intensity:1;Charge:1),
(Position:915.11829;Intensity:12000;Charge:1),
(Position:919.63;Intensity:1;Charge:1),
(Position:919.63;Intensity:1;Charge:1),
(Position:920.42841;Intensity:8900;Charge:1),
(Position:922.25900;Intensity:6000;Charge:1),
(Position:922.41109;Intensity:2200;Charge:1),
(Position:922.92230;Intensity:1800;Charge:1),
(Position:927.80644;Intensity:910;Charge:1),
(Position:929.011;Intensity:200;Charge:2),
(Position:930.34053;Intensity:7700;Charge:1),
(Position:931.31381;Intensity:830;Charge:1),
(Position:931.65288;Intensity:2700;Charge:1),
(Position:932.90663;Intensity:6900;Charge:1),
(Position:934.31;Intensity:3;Charge:1),
(Position:935.59;Intensity:4;Charge:1),
(Position:935.59;Intensity:4;Charge:1),
(Position:937.58796;Intensity:1500;Charge:1),
(Position:937.98004;Intensity:7;Charge:1),
(Position:941.333;Intensity:9;Charge:1),
(Position:941.490;Intensity:6;Charge:1),
(Position:942.79655;Intensity:4800;Charge:1),
(Position:943.55961;Intensity:66;Charge:1),
(Position:944.64;Intensity:3;Charge:1),
(Position:944.785;Intensity:4;Charge:1),
(Position:945.66;Intensity:1;Charge:1),
(Position:946.18060;Intensity:2800;Charge:1),
(Position:948.92849;Intensity:5000;Charge:1),
(Position:950.920;Intensity:4;Charge:1),
(Position:951.10;Intensity:7;Charge:1),
(Position:953.67793;Intensity:6100;Charge:1),
(Position:955.00241;Intensity:2800;Charge:1),
(Position:957.6628;Intensity:2;Charge:1),
(Position:957.964;Intensity:120;Charge:2),
(Position:964.48;Intensity:1;Charge:1),
(Position:966.80709;Intensity:18000;Charge:1),
(Position:972.75;Intensity:1;Charge:1),
(Position:979.08;Intensity:3;Charge:1),
(Position:981.155;Intensity:100;Charge:2),
(Position:982.6147;Intensity:3;Charge:1),
(Position:984.0204;Intensity:13;Charge:1),
(Position:984.0204;Intensity:13;Charge:1),
(Position:990.001;Intensity:4;Charge:1),
(Position:990.177;Intensity:3;Charge:1),
(Position:990.177;Intensity:3;Charge:1),
(Position:990.3309;Intensity:30;Charge:1),
(Position:990.3309;Intensity:30;Charge:1),
(Position:990.5052;Intensity:25;Charge:1),
(Position:991.7914;Intensity:13;Charge:1),
(Position:992.1321;Intensity:6;Charge:1),
(Position:992.1321;Intensity:6;Charge:1),
(Position:993.9577;Intensity:10;Charge:1),
(Position:994.1077;Intensity:16;Charge:1),
(Position:994.6866;Intensity:9;Charge:1),
(Position:994.7785;Intensity:4;Charge:1),
(Position:994.7785;Intensity:4;Charge:1),
(Position:995.0788;Intensity:9;Charge:1),
(Position:996.6337;Intensity:7;Charge:1),
(Position:997.69;Intensity:3;Charge:1),
(Position:1000.8343;Intensity:13;Charge:1),
(Position:1001.0129;Intensity:20;Charge:1),
(Position:1001.1429;Intensity:5;Charge:1),
(Position:1003.99;Intensity:3;Charge:1),
(Position:1004.17;Intensity:3;Charge:1),
(Position:1021.3634;Intensity:2;Charge:1),
(Position:1022.7461;Intensity:4;Charge:1),
(Position:1024.85213;Intensity:16;Charge:1),
(Position:1029.82378;Intensity:420;Charge:1),
(Position:1043.54497;Intensity:13;Charge:1),
(Position:1043.54497;Intensity:13;Charge:1),
(Position:1056.53029;Intensity:8000;Charge:1),
(Position:1062.35735;Intensity:780;Charge:1),
(Position:1067.6794;Intensity:19;Charge:1),
(Position:1069.3386;Intensity:55;Charge:1),
(Position:1069.3386;Intensity:55;Charge:1),
(Position:1076.1151;Intensity:15;Charge:1),
(Position:1076.3218;Intensity:14;Charge:1),
(Position:1076.3218;Intensity:14;Charge:1),
(Position:1076.6972;Intensity:150;Charge:1),
(Position:1076.9036;Intensity:110;Charge:1),
(Position:1078.3484;Intensity:69;Charge:1),
(Position:1078.9241;Intensity:16;Charge:1),
(Position:1078.9241;Intensity:16;Charge:1),
(Position:1079.3818;Intensity:2;Charge:1),
(Position:1080.10009;Intensity:6100;Charge:1),
(Position:1080.9318;Intensity:54;Charge:1),
(Position:1081.1089;Intensity:74;Charge:1),
(Position:1081.7717;Intensity:40;Charge:1),
(Position:1082.2783;Intensity:92;Charge:1),
(Position:1084.11868;Intensity:37;Charge:1),
(Position:1084.74479;Intensity:9400;Charge:1),
(Position:1088.9259;Intensity:58;Charge:1),
(Position:1089.1374;Intensity:85;Charge:1),
(Position:1089.4134;Intensity:25;Charge:1),
(Position:1102.38978;Intensity:78;Charge:1),
(Position:1104.70248;Intensity:64;Charge:1),
(Position:1105.27483;Intensity:200;Charge:1),
(Position:1106.3837;Intensity:29;Charge:1),
(Position:1112.33233;Intensity:73;Charge:1),
(Position:1113.75587;Intensity:49;Charge:1),
(Position:1114.14832;Intensity:55;Charge:1),
(Position:1114.60715;Intensity:26000;Charge:1),
(Position:1116.32695;Intensity:270;Charge:1),
(Position:1118.05855;Intensity:49000;Charge:1),
(Position:1129.60569;Intensity:65;Charge:1),
(Position:1130.15353;Intensity:46;Charge:1),
(Position:1130.69830;Intensity:140;Charge:1),
(Position:1130.76410;Intensity:100;Charge:1),
(Position:1133.27281;Intensity:41;Charge:1),
(Position:1133.67906;Intensity:37;Charge:1),
(Position:1136.97838;Intensity:110;Charge:1),
(Position:1139.35520;Intensity:15000;Charge:1),
(Position:1141.22575;Intensity:8800;Charge:1),
(Position:1152.58996;Intensity:33000;Charge:1),
(Position:1152.81755;Intensity:17000;Charge:1),
(Position:1153.95029;Intensity:9100;Charge:1),
(Position:1160.47129;Intensity:2600;Charge:1),
(Position:1161.72599;Intensity:13000;Charge:1),
(Position:1169.12022;Intensity:2800;Charge:1),
(Position:1177.00137;Intensity:15000;Charge:1),
(Position:1179.22713;Intensity:13000;Charge:1),
(Position:1179.31165;Intensity:3200;Charge:1),
(Position:1198.3060;Intensity:20;Charge:1),
(Position:1198.81939;Intensity:7400;Charge:1),
(Position:1199.9852;Intensity:18;Charge:1),
(Position:1200.1097;Intensity:5;Charge:1),
(Position:1206.96364;Intensity:23000;Charge:1),
(Position:1239.2373;Intensity:3;Charge:1),
(Position:1241.2164;Intensity:4;Charge:1),
(Position:1243.3906;Intensity:18;Charge:1),
(Position:1245.67756;Intensity:75;Charge:1),
(Position:1246.27992;Intensity:4300;Charge:1),
(Position:1246.75265;Intensity:160;Charge:1),
(Position:1247.6881;Intensity:23;Charge:1),
(Position:1247.6881;Intensity:23;Charge:1),
(Position:1249.01478;Intensity:450;Charge:1),
(Position:1249.01478;Intensity:450;Charge:1),
(Position:1252.36597;Intensity:21;Charge:1),
(Position:1254.1172;Intensity:15;Charge:1),
(Position:1256.31982;Intensity:48;Charge:1),
(Position:1257.44446;Intensity:130;Charge:1),
(Position:1257.72631;Intensity:96;Charge:1),
(Position:1257.72631;Intensity:96;Charge:1),
(Position:1258.0790;Intensity:8;Charge:1),
(Position:1258.3586;Intensity:6;Charge:1),
(Position:1258.80450;Intensity:1200;Charge:1),
(Position:1259.08693;Intensity:850;Charge:1),
(Position:1259.84506;Intensity:1600;Charge:1),
(Position:1260.42251;Intensity:32;Charge:1),
(Position:1260.4740;Intensity:13;Charge:1),
(Position:1260.67659;Intensity:18;Charge:1),
(Position:1260.76255;Intensity:550;Charge:1),
(Position:1262.11211;Intensity:150;Charge:1),
(Position:1263.4480;Intensity:6;Charge:1),
(Position:1264.37754;Intensity:500;Charge:1),
(Position:1264.61981;Intensity:710;Charge:1),
(Position:1265.44433;Intensity:330;Charge:1),
(Position:1266.18089;Intensity:470;Charge:1),
(Position:1266.80709;Intensity:98;Charge:1),
(Position:1266.80709;Intensity:98;Charge:1),
(Position:1268.70027;Intensity:300;Charge:1),
(Position:1269.26745;Intensity:6500;Charge:1),
(Position:1272.2276;Intensity:23;Charge:1),
(Position:1273.0267;Intensity:11;Charge:1),
(Position:1274.97132;Intensity:540;Charge:1),
(Position:1275.26124;Intensity:710;Charge:1),
(Position:1275.62108;Intensity:120;Charge:1),
(Position:1275.62108;Intensity:120;Charge:1),
(Position:1275.91401;Intensity:170;Charge:1),
(Position:1276.34399;Intensity:37;Charge:1),
(Position:1277.0074;Intensity:6;Charge:1),
(Position:1277.0074;Intensity:6;Charge:1),
(Position:1277.30181;Intensity:1600;Charge:1),
(Position:1278.0147;Intensity:16;Charge:1),
(Position:1285.6550;Intensity:5;Charge:1),
(Position:1286.7610;Intensity:7;Charge:1),
(Position:1286.8249;Intensity:11;Charge:1),
(Position:1289.06881;Intensity:14;Charge:1),
(Position:1291.55459;Intensity:8400;Charge:1),
(Position:1298.36239;Intensity:25;Charge:1),
(Position:1305.8359;Intensity:12;Charge:1),
(Position:1306.2387;Intensity:6;Charge:1),
(Position:1309.9978;Intensity:6;Charge:1),
(Position:1312.9761;Intensity:6;Charge:1),
(Position:1313.1271;Intensity:12;Charge:1),
(Position:1314.9041;Intensity:5;Charge:1),
(Position:1322.28579;Intensity:4500;Charge:1),
(Position:1325.4823;Intensity:8;Charge:1),
(Position:1330.0183;Intensity:6;Charge:1),
(Position:1334.3362;Intensity:10;Charge:1),
(Position:1339.2952;Intensity:7;Charge:1),
(Position:1353.0772;Intensity:5;Charge:1),
(Position:1358.8803;Intensity:5;Charge:1),
(Position:1374.2491;Intensity:4;Charge:1),
(Position:1387.0096;Intensity:9;Charge:1),
(Position:1391.1976;Intensity:5;Charge:1),
(Position:1397.4792;Intensity:3;Charge:1),
(Position:1401.6752;Intensity:4;Charge:1),
(Position:1401.9732;Intensity:6;Charge:1),
(Position:1404.6946;Intensity:9;Charge:1),
(Position:1407.7958;Intensity:5;Charge:1),
(Position:1428.7508;Intensity:7;Charge:1),
(Position:1430.47431;Intensity:130;Charge:1),
(Position:1434.60814;Intensity:120;Charge:1),
(Position:1435.72729;Intensity:51;Charge:1),
(Position:1438.8045;Intensity:5;Charge:1),
(Position:1450.38851;Intensity:18;Charge:1),
(Position:1493.38864;Intensity:110;Charge:1),
(Position:1493.5264;Intensity:7;Charge:1),
(Position:1497.4865;Intensity:15;Charge:1),
(Position:1498.8949;Intensity:22;Charge:1),
(Position:1499.04149;Intensity:530;Charge:1),
(Position:1506.31048;Intensity:19;Charge:1),
(Position:1507.82883;Intensity:140;Charge:1),
(Position:1507.91629;Intensity:48;Charge:1),
(Position:1508.8034;Intensity:4;Charge:1),
(Position:1508.94329;Intensity:94;Charge:1),
(Position:1511.0582;Intensity:13;Charge:1),
(Position:1512.9862;Intensity:5;Charge:1),
(Position:1514.42356;Intensity:350;Charge:1),
(Position:1517.6111;Intensity:6;Charge:1),
(Position:1517.84581;Intensity:32;Charge:1),
(Position:1518.0482;Intensity:4;Charge:1),
(Position:1519.38748;Intensity:63;Charge:1),
(Position:1519.47634;Intensity:99;Charge:1),
(Position:1519.50832;Intensity:270;Charge:1),
(Position:1519.67882;Intensity:48;Charge:1),
(Position:1523.48765;Intensity:5300;Charge:1),
(Position:1535.23836;Intensity:160;Charge:1),
(Position:1537.42789;Intensity:74;Charge:1),
(Position:1539.4233;Intensity:17;Charge:1),
(Position:1541.18032;Intensity:250;Charge:1),
(Position:1541.3269;Intensity:6;Charge:1),
(Position:1543.5339;Intensity:4;Charge:1),
(Position:1545.5085;Intensity:11;Charge:1),
(Position:1545.54506;Intensity:110;Charge:1),
(Position:1547.04528;Intensity:25;Charge:1),
(Position:1550.3719;Intensity:23;Charge:1),
(Position:1550.5133;Intensity:13;Charge:1),
(Position:1560.84777;Intensity:65;Charge:1),
(Position:1576.5356;Intensity:4;Charge:1),
(Position:1580.6965;Intensity:4;Charge:1),
(Position:1581.6501;Intensity:7;Charge:1),
(Position:1602.71469;Intensity:130;Charge:1),
(Position:1604.9882;Intensity:5;Charge:1),
(Position:1610.28840;Intensity:47;Charge:1),
(Position:1625.7112;Intensity:5;Charge:1),
(Position:1626.8035;Intensity:7;Charge:1),
(Position:1626.86908;Intensity:44;Charge:1),
(Position:1627.2797;Intensity:8;Charge:1),
(Position:1627.34004;Intensity:63;Charge:1),
(Position:1635.13887;Intensity:40;Charge:1),
(Position:1640.97373;Intensity:150;Charge:1),
(Position:1642.8149;Intensity:15;Charge:1),
(Position:1647.3492;Intensity:19;Charge:1),
(Position:1647.92536;Intensity:81;Charge:1),
(Position:1653.26019;Intensity:34;Charge:1),
(Position:1659.60411;Intensity:34;Charge:1),
(Position:1661.1556;Intensity:30;Charge:1),
(Position:1661.39757;Intensity:89;Charge:1),
(Position:1663.85935;Intensity:35;Charge:1),
(Position:1667.1664;Intensity:16;Charge:1),
(Position:1679.33780;Intensity:54;Charge:1),
(Position:1686.62554;Intensity:37;Charge:1),
(Position:1691.4677;Intensity:11;Charge:1),
(Position:1698.86133;Intensity:67;Charge:1),
(Position:1711.6810;Intensity:5;Charge:1),
(Position:1716.66222;Intensity:1800;Charge:1),
(Position:1723.8892;Intensity:7;Charge:1),
(Position:1729.8963;Intensity:7;Charge:1),
(Position:1731.5238;Intensity:4;Charge:1),
(Position:1796.6073;Intensity:13;Charge:1),
(Position:1803.45708;Intensity:160;Charge:1),
(Position:1804.07373;Intensity:1900;Charge:1),
(Position:1808.8119;Intensity:4500;Charge:1),
(Position:1808.8201;Intensity:910;Charge:1),
(Position:1821.52793;Intensity:250;Charge:1),
(Position:1822.60624;Intensity:1600;Charge:1),
(Position:1823.19929;Intensity:1300;Charge:1),
(Position:1825.24309;Intensity:87;Charge:1),
(Position:1825.83042;Intensity:64;Charge:1),
(Position:1828.16323;Intensity:14000;Charge:1),
(Position:1828.76064;Intensity:10000;Charge:1),
(Position:1830.89656;Intensity:6800;Charge:1),
(Position:1836.41077;Intensity:1900;Charge:1),
(Position:1837.6458;Intensity:9;Charge:1),
(Position:1838.90058;Intensity:360;Charge:1),
(Position:1838.98458;Intensity:6400;Charge:1),
(Position:1839.41888;Intensity:480;Charge:1),
(Position:1839.49582;Intensity:8600;Charge:1),
(Position:1840.78607;Intensity:3900;Charge:1),
(Position:1842.74320;Intensity:6300;Charge:1),
(Position:1842.8231;Intensity:26;Charge:1),
(Position:1846.36807;Intensity:1300;Charge:1),
(Position:1846.36807;Intensity:1300;Charge:1),
(Position:1848.08447;Intensity:4100;Charge:1),
(Position:1859.6617;Intensity:6900;Charge:1),
(Position:1860.2776;Intensity:9500;Charge:1),
(Position:1862.3992;Intensity:1600;Charge:1),
(Position:1863.0245;Intensity:2000;Charge:1),
(Position:1866.0699;Intensity:9;Charge:1),
(Position:1868.1179;Intensity:87;Charge:1),
(Position:1868.455;Intensity:14;Charge:1),
(Position:1868.7339;Intensity:130;Charge:1),
(Position:1890.3986;Intensity:30;Charge:1),
(Position:1894.2723;Intensity:150;Charge:1),
(Position:1894.9817;Intensity:6;Charge:1),
(Position:1911.6409;Intensity:5;Charge:1),
(Position:1957.9094;Intensity:260;Charge:1),
(Position:1958.2455;Intensity:790;Charge:1),
(Position:1977.7860;Intensity:32;Charge:1),
(Position:1988.2735;Intensity:5;Charge:1),
(Position:2014.0224;Intensity:1;Charge:1),
(Position:2035.5771;Intensity:630;Charge:1),
(Position:2035.9404;Intensity:43;Charge:1),
(Position:2037.2201;Intensity:2;Charge:1),
(Position:2041.7199;Intensity:8;Charge:1),
(Position:2042.1587;Intensity:37;Charge:1),
(Position:2042.5447;Intensity:3;Charge:1),
(Position:2085.4446;Intensity:11;Charge:1),
(Position:2090.1599;Intensity:7;Charge:1),
(Position:2091.0236;Intensity:29;Charge:1),
(Position:2096.6936;Intensity:2;Charge:1),
(Position:2101.4614;Intensity:11;Charge:1),
(Position:2101.9261;Intensity:8;Charge:1),
(Position:2102.3342;Intensity:3;Charge:1),
(Position:2104.7013;Intensity:2700;Charge:1),
(Position:2119.1366;Intensity:4;Charge:1),
(Position:2122.5899;Intensity:11;Charge:1),
(Position:2122.5899;Intensity:11;Charge:1),
(Position:2133.1283;Intensity:26;Charge:1),
(Position:2133.6274;Intensity:20;Charge:1),
(Position:2137.455;Intensity:15;Charge:1),
(Position:2139.2749;Intensity:10;Charge:1),
(Position:2139.2749;Intensity:10;Charge:1),
(Position:2139.6970;Intensity:15;Charge:1),
(Position:2140.1796;Intensity:12;Charge:1),
(Position:2142.0939;Intensity:30;Charge:1),
(Position:2142.7883;Intensity:46;Charge:1),
(Position:2145.7276;Intensity:13;Charge:1),
(Position:2146.4337;Intensity:17;Charge:1),
(Position:2150.471;Intensity:20;Charge:1),
(Position:2150.9615;Intensity:90;Charge:1),
(Position:2153.4557;Intensity:8;Charge:1),
(Position:2153.934;Intensity:18;Charge:1),
(Position:2156.9002;Intensity:20;Charge:1),
(Position:2156.9002;Intensity:20;Charge:1),
(Position:2158.2413;Intensity:34;Charge:1),
(Position:2158.5760;Intensity:32;Charge:1),
(Position:2160.2349;Intensity:36;Charge:1),
(Position:2160.2349;Intensity:36;Charge:1),
(Position:2162.6639;Intensity:15;Charge:1),
(Position:2162.6639;Intensity:15;Charge:1),
(Position:2162.7560;Intensity:12;Charge:1),
(Position:2162.7560;Intensity:12;Charge:1),
(Position:2163.8469;Intensity:23;Charge:1),
(Position:2163.8469;Intensity:23;Charge:1),
(Position:2164.5598;Intensity:3;Charge:1),
(Position:2164.5598;Intensity:3;Charge:1),
(Position:2164.5598;Intensity:3;Charge:1),
(Position:2164.5598;Intensity:3;Charge:1),
(Position:2165.9251;Intensity:15;Charge:1),
(Position:2167.3267;Intensity:23;Charge:1),
(Position:2167.3267;Intensity:23;Charge:1),
(Position:2167.467;Intensity:6;Charge:1),
(Position:2171.4039;Intensity:2900;Charge:1),
(Position:2217.7292;Intensity:78;Charge:1),
(Position:2225.3432;Intensity:1300;Charge:1),
(Position:2243.4265;Intensity:1300;Charge:1),
(Position:2247.2920;Intensity:540;Charge:1),
(Position:2253.6528;Intensity:8500;Charge:1),
(Position:2266.7971;Intensity:1300;Charge:1),
(Position:2269.3959;Intensity:210;Charge:1),
(Position:2310.6784;Intensity:2500;Charge:1),
(Position:2326.6619;Intensity:3800;Charge:1),
(Position:2337.9343;Intensity:5000;Charge:1),
(Position:2357.1764;Intensity:3400;Charge:1),
(Position:2364.2934;Intensity:17000;Charge:1),
(Position:2370.8130;Intensity:1200;Charge:1),
(Position:2371.4099;Intensity:74;Charge:1),
(Position:2371.5599;Intensity:5900;Charge:1),
(Position:2391.8541;Intensity:170;Charge:1),
(Position:2395.7931;Intensity:11000;Charge:1),
(Position:2396.115;Intensity:45;Charge:1),
(Position:2396.2964;Intensity:4600;Charge:1),
(Position:2397.8372;Intensity:220;Charge:1),
(Position:2398.4701;Intensity:6000;Charge:1),
(Position:2409.3528;Intensity:200;Charge:1),
(Position:2409.8982;Intensity:46;Charge:1),
(Position:2410.5148;Intensity:1100;Charge:1),
(Position:2415.6486;Intensity:210;Charge:1),
(Position:2416.2547;Intensity:12;Charge:1),
(Position:2416.8025;Intensity:2000;Charge:1),
(Position:2422.5535;Intensity:140;Charge:1),
(Position:2425.6224;Intensity:2800;Charge:1),
(Position:2429.3066;Intensity:12;Charge:1),
(Position:2431.6420;Intensity:38;Charge:1),
(Position:2437.1661;Intensity:7400;Charge:1),
(Position:2437.633;Intensity:18;Charge:1),
(Position:2437.8260;Intensity:3800;Charge:1),
(Position:2439.0011;Intensity:360;Charge:1),
(Position:2439.5228;Intensity:37;Charge:1),
(Position:2445.4531;Intensity:1900;Charge:1),
(Position:2445.9078;Intensity:12;Charge:1),
(Position:2445.9775;Intensity:240;Charge:1),
(Position:2446.6068;Intensity:3300;Charge:1),
(Position:2447.1606;Intensity:370;Charge:1),
(Position:2448.2800;Intensity:7;Charge:1),
(Position:2451.036;Intensity:4;Charge:1),
(Position:2453.2498;Intensity:55;Charge:1),
(Position:2460.6763;Intensity:9;Charge:1),
(Position:2477.140;Intensity:7;Charge:1),
(Position:2478.3248;Intensity:1700;Charge:1),
(Position:2479.6287;Intensity:5;Charge:1),
(Position:2490.2752;Intensity:21;Charge:1),
(Position:2491.0521;Intensity:780;Charge:1),
(Position:2492.9689;Intensity:3;Charge:1),
(Position:2493.5696;Intensity:2900;Charge:1),
(Position:2494.2298;Intensity:46;Charge:1),
(Position:2500.6628;Intensity:170;Charge:1),
(Position:2507.1216;Intensity:35;Charge:1),
(Position:2508.4899;Intensity:7;Charge:1),
(Position:2516.8567;Intensity:1300;Charge:1),
(Position:2517.6147;Intensity:3;Charge:1),
(Position:2519.5185;Intensity:25;Charge:1),
(Position:2523.4824;Intensity:300;Charge:1),
(Position:2528.4125;Intensity:100;Charge:1),
(Position:2537.6263;Intensity:15;Charge:1),
(Position:2540.0128;Intensity:280;Charge:1),
(Position:2543.8750;Intensity:7;Charge:1),
(Position:2546.6860;Intensity:12;Charge:1),
(Position:2549.8245;Intensity:6;Charge:1),
(Position:2551.0427;Intensity:6;Charge:1),
(Position:2553.1295;Intensity:4600;Charge:1),
(Position:2554.7659;Intensity:4;Charge:1),
(Position:2558.8111;Intensity:14;Charge:1),
(Position:2562.8812;Intensity:7;Charge:1),
(Position:2563.8336;Intensity:42;Charge:1),
(Position:2575.306;Intensity:4;Charge:1),
(Position:2575.337;Intensity:10;Charge:1),
(Position:2575.337;Intensity:10;Charge:1),
(Position:2576.0661;Intensity:10;Charge:1),
(Position:2576.0661;Intensity:10;Charge:1),
(Position:2578.6665;Intensity:91;Charge:1),
(Position:2579.4008;Intensity:85;Charge:1),
(Position:2584.5550;Intensity:48;Charge:1),
(Position:2586.1933;Intensity:130;Charge:1),
(Position:2587.3469;Intensity:38;Charge:1),
(Position:2587.8656;Intensity:18;Charge:1),
(Position:2587.9642;Intensity:56;Charge:1),
(Position:2588.7950;Intensity:38;Charge:1),
(Position:2594.8961;Intensity:32;Charge:1),
(Position:2596.0581;Intensity:21;Charge:1),
(Position:2598.2412;Intensity:10;Charge:1),
(Position:2598.2412;Intensity:10;Charge:1),
(Position:2607.2823;Intensity:52;Charge:1),
(Position:2608.0017;Intensity:68;Charge:1),
(Position:2608.8520;Intensity:13;Charge:1),
(Position:2609.5809;Intensity:17;Charge:1),
(Position:2613.1288;Intensity:98;Charge:1),
(Position:2613.1288;Intensity:98;Charge:1),
(Position:2616.2321;Intensity:15;Charge:1),
(Position:2616.2321;Intensity:15;Charge:1),
(Position:2616.2321;Intensity:15;Charge:1),
(Position:2617.8206;Intensity:230;Charge:1),
(Position:2617.8206;Intensity:230;Charge:1),
(Position:2621.1856;Intensity:120;Charge:1),
(Position:2621.1856;Intensity:120;Charge:1),
(Position:2621.3615;Intensity:90;Charge:1),
(Position:2621.3615;Intensity:90;Charge:1),
(Position:2622.8172;Intensity:110;Charge:1),
(Position:2622.8172;Intensity:110;Charge:1),
(Position:2624.3848;Intensity:15;Charge:1),
(Position:2624.3848;Intensity:15;Charge:1),
(Position:2627.6978;Intensity:160;Charge:1),
(Position:2627.6978;Intensity:160;Charge:1),
(Position:2628.267;Intensity:25;Charge:1),
(Position:2628.267;Intensity:25;Charge:1),
(Position:2628.282;Intensity:10;Charge:1),
(Position:2686.8106;Intensity:1000;Charge:1),
(Position:2752.8250;Intensity:140;Charge:1),
(Position:2758.0984;Intensity:930;Charge:1),
(Position:2782.6375;Intensity:15;Charge:1),
(Position:2797.9570;Intensity:240;Charge:1),
(Position:2839.3944;Intensity:570;Charge:1),
(Position:2854.0970;Intensity:310;Charge:1),
(Position:2875.2113;Intensity:81;Charge:1),
(Position:2929.5268;Intensity:13;Charge:1),
(Position:2939.5843;Intensity:13;Charge:1),
(Position:2945.5855;Intensity:130;Charge:1),
(Position:2949.5606;Intensity:14;Charge:1),
(Position:2967.6059;Intensity:22;Charge:1),
(Position:2972.2119;Intensity:61;Charge:1),
(Position:2981.2553;Intensity:6;Charge:1),
(Position:2993.9522;Intensity:4;Charge:1),
(Position:2994.9038;Intensity:9;Charge:1),
(Position:3012.7143;Intensity:1;Charge:1),
(Position:3013.5096;Intensity:2;Charge:1),
(Position:3013.8002;Intensity:1;Charge:1),
(Position:3017.3468;Intensity:10;Charge:1),
(Position:3020.8732;Intensity:620;Charge:1),
(Position:3026.7823;Intensity:41;Charge:1),
(Position:3027.5862;Intensity:17;Charge:1),
(Position:3030.8501;Intensity:44;Charge:1),
(Position:3037.1786;Intensity:20;Charge:1),
(Position:3042.5647;Intensity:15;Charge:1),
(Position:3047.2757;Intensity:9;Charge:1),
(Position:3047.5223;Intensity:12;Charge:1),
(Position:3060.3119;Intensity:28;Charge:1),
(Position:3060.3346;Intensity:52;Charge:1),
(Position:3063.9710;Intensity:21;Charge:1),
(Position:3067.5320;Intensity:23;Charge:1),
(Position:3072.0023;Intensity:53;Charge:1),
(Position:3092.8308;Intensity:16;Charge:1),
(Position:3097.5572;Intensity:7;Charge:1),
(Position:3101.1644;Intensity:25;Charge:1),
(Position:3111.0469;Intensity:11;Charge:1),
(Position:3122.394;Intensity:3;Charge:1),
(Position:3186.8616;Intensity:78;Charge:1),
(Position:3217.9408;Intensity:1;Charge:1),
(Position:3243.3342;Intensity:7;Charge:1),
(Position:3270.054;Intensity:4;Charge:1),
(Position:3318.2139;Intensity:830;Charge:1),
(Position:3332.5251;Intensity:5;Charge:1),
(Position:3334.1790;Intensity:230;Charge:1),
(Position:3336.1478;Intensity:1700;Charge:1),
(Position:3352.0419;Intensity:47;Charge:1),
(Position:3362.8670;Intensity:14;Charge:1),
(Position:3368.6444;Intensity:2;Charge:1),
(Position:3371.7548;Intensity:8;Charge:1),
(Position:3372.2372;Intensity:3;Charge:1),
(Position:3381.3686;Intensity:3;Charge:1),
(Position:3384.9869;Intensity:3;Charge:1),
(Position:3390.9054;Intensity:450;Charge:1),
(Position:3391.2263;Intensity:440;Charge:1),
(Position:3392.2350;Intensity:1200;Charge:1),
(Position:3395.2413;Intensity:7;Charge:1),
(Position:3399.6462;Intensity:4;Charge:1),
(Position:3414.0648;Intensity:360;Charge:1),
(Position:3427.6169;Intensity:5;Charge:1),
(Position:3438.3050;Intensity:3;Charge:1),
(Position:3448.0836;Intensity:590;Charge:1),
(Position:3449.9275;Intensity:240;Charge:1),
(Position:3456.444;Intensity:1;Charge:1),
(Position:3467.3121;Intensity:6;Charge:1),
(Position:3478.9486;Intensity:380;Charge:1),
(Position:3521.780;Intensity:2;Charge:1),
(Position:3551.7017;Intensity:120;Charge:1),
(Position:3584.4581;Intensity:790;Charge:1),
(Position:3620.9395;Intensity:33;Charge:1),
(Position:3648.1630;Intensity:96;Charge:1),
(Position:3717.674;Intensity:3;Charge:1),
(Position:3718.2250;Intensity:42;Charge:1),
(Position:3739.6549;Intensity:4;Charge:1),
(Position:3774.6247;Intensity:130;Charge:1),
(Position:3900.7087;Intensity:10;Charge:1),
(Position:3901.9950;Intensity:74;Charge:1),
(Position:3913.6736;Intensity:170;Charge:1),
(Position:3913.7029;Intensity:65;Charge:1),
(Position:3920.7468;Intensity:3;Charge:1),
(Position:3932.4462;Intensity:66;Charge:1),
(Position:3934.1564;Intensity:48;Charge:1),
(Position:3938.6615;Intensity:3;Charge:1),
(Position:3940.346;Intensity:4;Charge:1),
(Position:3945.7774;Intensity:640;Charge:1),
(Position:3947.4992;Intensity:430;Charge:1),
(Position:3951.8767;Intensity:12;Charge:1),
(Position:3956.6317;Intensity:260;Charge:1),
(Position:3963.9886;Intensity:240;Charge:1),
(Position:3965.4346;Intensity:350;Charge:1),
(Position:3966.9971;Intensity:280;Charge:1),
(Position:3969.9524;Intensity:75;Charge:1),
(Position:3980.9128;Intensity:160;Charge:1),
(Position:3981.7158;Intensity:180;Charge:1),
(Position:3984.0733;Intensity:150;Charge:1),
(Position:3994.4171;Intensity:29;Charge:1),
(Position:3994.4482;Intensity:20;Charge:1),
(Position:4007.4299;Intensity:280;Charge:1),
(Position:4009.1275;Intensity:390;Charge:1),
(Position:4013.8549;Intensity:65;Charge:1),
(Position:4015.5856;Intensity:97;Charge:1),
(Position:4025.4424;Intensity:350;Charge:1),
(Position:4025.4792;Intensity:540;Charge:1),
(Position:4027.6140;Intensity:3;Charge:1),
(Position:4029.1835;Intensity:140;Charge:1),
(Position:4029.1835;Intensity:140;Charge:1),
(Position:4031.4409;Intensity:6;Charge:1),
(Position:4031.4409;Intensity:6;Charge:1),
(Position:4035.6667;Intensity:1900;Charge:1),
(Position:4035.6667;Intensity:1900;Charge:1),
(Position:4042.5964;Intensity:1300;Charge:1),
(Position:4042.9447;Intensity:660;Charge:1),
(Position:4042.9447;Intensity:660;Charge:1),
(Position:4042.9682;Intensity:300;Charge:1),
(Position:4045.7207;Intensity:690;Charge:1),
(Position:4045.7474;Intensity:340;Charge:1),
(Position:4052.2294;Intensity:85;Charge:1),
(Position:4052.2553;Intensity:59;Charge:1),
(Position:4052.2553;Intensity:59;Charge:1),
(Position:4056.4426;Intensity:1500;Charge:1),
(Position:4056.4426;Intensity:1500;Charge:1),
(Position:4058.7308;Intensity:180;Charge:1),
(Position:4058.7308;Intensity:180;Charge:1),
(Position:4065.2823;Intensity:7;Charge:1),
(Position:4065.2823;Intensity:7;Charge:1),
(Position:4081.8029;Intensity:10;Charge:1),
(Position:4082.758;Intensity:3;Charge:1),
(Position:4082.8183;Intensity:5;Charge:1),
(Position:4085.1083;Intensity:2;Charge:1),
(Position:4092.9375;Intensity:7;Charge:1),
(Position:4093.511;Intensity:3;Charge:1),
(Position:4093.910;Intensity:2;Charge:1),
(Position:4094.160;Intensity:2;Charge:1),
(Position:4094.5547;Intensity:3;Charge:1),
(Position:4095.0274;Intensity:8;Charge:1),
(Position:4096.1383;Intensity:14;Charge:1),
(Position:4099.2010;Intensity:8;Charge:1),
(Position:4105.495;Intensity:9;Charge:1),
(Position:4117.232;Intensity:5;Charge:1),
(Position:4218.2976;Intensity:140;Charge:1),
(Position:4433.580;Intensity:5;Charge:1),
(Position:4549.3929;Intensity:14;Charge:1),
(Position:4579.6921;Intensity:26;Charge:1),
(Position:4640.25;Intensity:10;Charge:1),
(Position:4640.25;Intensity:10;Charge:1),
(Position:4645.27;Intensity:10;Charge:1),
(Position:4645.27;Intensity:10;Charge:1),
(Position:4645.27;Intensity:10;Charge:1),
(Position:4645.27;Intensity:10;Charge:1),
(Position:4647.54;Intensity:10;Charge:1),
(Position:4647.54;Intensity:10;Charge:1),
(Position:4715.9797;Intensity:27;Charge:1),
(Position:4717.9303;Intensity:11;Charge:1),
(Position:4724.8165;Intensity:10;Charge:1),
(Position:4758.8616;Intensity:9;Charge:1),
(Position:4856.52;Intensity:30;Charge:1),
(Position:4904.51;Intensity:20;Charge:1),
(Position:4917.62;Intensity:10;Charge:1),
(Position:5006.26;Intensity:10;Charge:1),
(Position:5081.56;Intensity:10;Charge:1),
(Position:5116.59;Intensity:30;Charge:1),
(Position:5147.66;Intensity:20;Charge:1),
(Position:5167.80;Intensity:20;Charge:1),
(Position:5170.90;Intensity:20;Charge:1),
(Position:5171.94;Intensity:20;Charge:1),
(Position:5179.37;Intensity:20;Charge:1),
(Position:5180.41;Intensity:40;Charge:1),
(Position:5180.41;Intensity:40;Charge:1),
(Position:5232.76;Intensity:60;Charge:1),
(Position:5232.76;Intensity:60;Charge:1),
(Position:5300.21;Intensity:10;Charge:1),
(Position:5324.02;Intensity:20;Charge:1),
(Position:5324.02;Intensity:20;Charge:1),
(Position:5327.24;Intensity:10;Charge:1),
(Position:5376.57;Intensity:10;Charge:1),
(Position:5404.78;Intensity:20;Charge:1),
(Position:5493.10;Intensity:20;Charge:1)
);

Iron: array [0..1635] of TAtomicLine = (
(Position:50.242;Intensity:10;Charge:4),
(Position:50.535;Intensity:11;Charge:4),
(Position:50.669;Intensity:11;Charge:4),
(Position:52.569;Intensity:17;Charge:4),
(Position:52.629;Intensity:15;Charge:4),
(Position:52.657;Intensity:10;Charge:4),
(Position:52.663;Intensity:13;Charge:4),
(Position:53.091;Intensity:10;Charge:4),
(Position:53.178;Intensity:10;Charge:4),
(Position:53.555;Intensity:10;Charge:4),
(Position:53.661;Intensity:14;Charge:4),
(Position:53.674;Intensity:10;Charge:4),
(Position:53.710;Intensity:15;Charge:4),
(Position:53.726;Intensity:13;Charge:4),
(Position:53.779;Intensity:14;Charge:4),
(Position:53.794;Intensity:13;Charge:4),
(Position:53.844;Intensity:10;Charge:4),
(Position:54.420;Intensity:10;Charge:4),
(Position:54.622;Intensity:10;Charge:4),
(Position:54.880;Intensity:10;Charge:4),
(Position:55.032;Intensity:11;Charge:4),
(Position:55.177;Intensity:10;Charge:4),
(Position:55.214;Intensity:13;Charge:4),
(Position:55.274;Intensity:11;Charge:4),
(Position:55.426;Intensity:10;Charge:4),
(Position:55.566;Intensity:10;Charge:4),
(Position:57.288;Intensity:10;Charge:4),
(Position:57.676;Intensity:10;Charge:4),
(Position:57.976;Intensity:10;Charge:4),
(Position:60.753;Intensity:14;Charge:4),
(Position:60.880;Intensity:13;Charge:4),
(Position:60.965;Intensity:10;Charge:4),
(Position:72.881;Intensity:6;Charge:3),
(Position:73.000;Intensity:5;Charge:3),
(Position:73.771;Intensity:5;Charge:3),
(Position:73.926;Intensity:5;Charge:3),
(Position:80.755;Intensity:9;Charge:3),
(Position:80.786;Intensity:8;Charge:3),
(Position:80.884;Intensity:8;Charge:3),
(Position:81.128;Intensity:8;Charge:3),
(Position:81.338;Intensity:10;Charge:3),
(Position:83.805;Intensity:8;Charge:3),
(Position:84.428;Intensity:10;Charge:3),
(Position:84.541;Intensity:9;Charge:3),
(Position:84.742;Intensity:8;Charge:3),
(Position:85.972;Intensity:8;Charge:3),
(Position:86.176;Intensity:8;Charge:3),
(Position:86.183;Intensity:10;Charge:3),
(Position:87.346;Intensity:8;Charge:3),
(Position:89.076;Intensity:9;Charge:3),
(Position:89.117;Intensity:10;Charge:3),
(Position:89.144;Intensity:8;Charge:3),
(Position:89.942;Intensity:8;Charge:3),
(Position:95.033;Intensity:10;Charge:3),
(Position:98.137;Intensity:10;Charge:3),
(Position:98.388;Intensity:10;Charge:3),
(Position:98.582;Intensity:8;Charge:3),
(Position:99.123;Intensity:9;Charge:3),
(Position:101.0260;Intensity:1;Charge:7),
(Position:101.6072;Intensity:0;Charge:7),
(Position:101.725;Intensity:9;Charge:3),
(Position:101.774;Intensity:8;Charge:3),
(Position:101.829;Intensity:8;Charge:3),
(Position:103.212;Intensity:8;Charge:3),
(Position:105.527;Intensity:12;Charge:2),
(Position:106.387;Intensity:8;Charge:3),
(Position:106.836;Intensity:15;Charge:2),
(Position:107.160;Intensity:15;Charge:2),
(Position:107.3953;Intensity:4;Charge:7),
(Position:108.0637;Intensity:2;Charge:7),
(Position:108.0736;Intensity:0;Charge:7),
(Position:108.7861;Intensity:2;Charge:7),
(Position:109.5343;Intensity:6;Charge:7),
(Position:109.689;Intensity:15;Charge:2),
(Position:109.912;Intensity:12;Charge:2),
(Position:111.209;Intensity:18;Charge:2),
(Position:111.7580;Intensity:6;Charge:7),
(Position:112.199;Intensity:12;Charge:2),
(Position:112.253;Intensity:9;Charge:3),
(Position:112.286;Intensity:12;Charge:2),
(Position:112.488;Intensity:9;Charge:3),
(Position:112.802;Intensity:8;Charge:3),
(Position:112.807;Intensity:12;Charge:2),
(Position:113.043;Intensity:12;Charge:2),
(Position:113.341;Intensity:15;Charge:2),
(Position:113.368;Intensity:12;Charge:2),
(Position:113.864;Intensity:12;Charge:2),
(Position:114.1435;Intensity:10;Charge:7),
(Position:114.233;Intensity:12;Charge:2),
(Position:114.323;Intensity:12;Charge:2),
(Position:114.495;Intensity:18;Charge:2),
(Position:114.741;Intensity:12;Charge:2),
(Position:114.829;Intensity:15;Charge:2),
(Position:115.116;Intensity:12;Charge:2),
(Position:115.4992;Intensity:4;Charge:7),
(Position:116.3879;Intensity:5;Charge:7),
(Position:116.6183;Intensity:8;Charge:7),
(Position:117.3915;Intensity:1;Charge:7),
(Position:118.0823;Intensity:4;Charge:7),
(Position:120.8375;Intensity:4;Charge:7),
(Position:122.6653;Intensity:5;Charge:7),
(Position:123.9690;Intensity:5;Charge:7),
(Position:124.4442;Intensity:0;Charge:7),
(Position:126.3844;Intensity:2;Charge:7),
(Position:126.744;Intensity:12;Charge:2),
(Position:127.200;Intensity:12;Charge:2),
(Position:131.786;Intensity:400;Charge:5),
(Position:131.835;Intensity:300;Charge:5),
(Position:132.041;Intensity:300;Charge:5),
(Position:132.134;Intensity:300;Charge:5),
(Position:132.149;Intensity:300;Charge:5),
(Position:132.327;Intensity:400;Charge:5),
(Position:133.040;Intensity:400;Charge:5),
(Position:133.2381;Intensity:6;Charge:7),
(Position:134.561;Intensity:300;Charge:5),
(Position:135.901;Intensity:400;Charge:5),
(Position:136.128;Intensity:300;Charge:5),
(Position:136.145;Intensity:300;Charge:5),
(Position:136.182;Intensity:600;Charge:5),
(Position:136.308;Intensity:300;Charge:5),
(Position:136.364;Intensity:300;Charge:5),
(Position:136.557;Intensity:300;Charge:5),
(Position:137.102;Intensity:12;Charge:2),
(Position:137.359;Intensity:700;Charge:5),
(Position:137.367;Intensity:600;Charge:5),
(Position:137.412;Intensity:300;Charge:5),
(Position:137.634;Intensity:500;Charge:5),
(Position:137.646;Intensity:300;Charge:5),
(Position:137.856;Intensity:500;Charge:5),
(Position:138.568;Intensity:300;Charge:5),
(Position:138.794;Intensity:800;Charge:5),
(Position:139.797;Intensity:400;Charge:5),
(Position:140.024;Intensity:600;Charge:5),
(Position:140.239;Intensity:800;Charge:5),
(Position:140.667;Intensity:400;Charge:5),
(Position:140.682;Intensity:500;Charge:5),
(Position:140.725;Intensity:400;Charge:5),
(Position:140.903;Intensity:300;Charge:5),
(Position:140.922;Intensity:300;Charge:5),
(Position:140.945;Intensity:600;Charge:5),
(Position:141.520;Intensity:400;Charge:5),
(Position:141.812;Intensity:300;Charge:5),
(Position:142.046;Intensity:600;Charge:5),
(Position:142.573;Intensity:12;Charge:4),
(Position:143.057;Intensity:800;Charge:5),
(Position:143.143;Intensity:13;Charge:4),
(Position:144.053;Intensity:800;Charge:5),
(Position:144.079;Intensity:300;Charge:5),
(Position:144.222;Intensity:400;Charge:5),
(Position:144.662;Intensity:800;Charge:5),
(Position:144.885;Intensity:700;Charge:5),
(Position:144.993;Intensity:400;Charge:5),
(Position:145.556;Intensity:300;Charge:5),
(Position:145.616;Intensity:700;Charge:5),
(Position:145.983;Intensity:500;Charge:5),
(Position:146.073;Intensity:400;Charge:5),
(Position:146.263;Intensity:500;Charge:5),
(Position:146.468;Intensity:700;Charge:5),
(Position:146.538;Intensity:500;Charge:5),
(Position:146.665;Intensity:400;Charge:5),
(Position:146.900;Intensity:500;Charge:5),
(Position:147.320;Intensity:12;Charge:4),
(Position:147.560;Intensity:300;Charge:5),
(Position:147.947;Intensity:500;Charge:5),
(Position:148.953;Intensity:12;Charge:4),
(Position:149.518;Intensity:12;Charge:4),
(Position:150.517;Intensity:10;Charge:3),
(Position:152.660;Intensity:13;Charge:4),
(Position:153.026;Intensity:13;Charge:4),
(Position:153.263;Intensity:14;Charge:4),
(Position:153.291;Intensity:13;Charge:4),
(Position:153.386;Intensity:15;Charge:4),
(Position:153.395;Intensity:13;Charge:4),
(Position:153.658;Intensity:14;Charge:4),
(Position:153.829;Intensity:12;Charge:4),
(Position:153.863;Intensity:10;Charge:3),
(Position:154.216;Intensity:13;Charge:4),
(Position:154.270;Intensity:14;Charge:4),
(Position:154.640;Intensity:12;Charge:4),
(Position:155.020;Intensity:12;Charge:3),
(Position:155.235;Intensity:12;Charge:4),
(Position:155.271;Intensity:12;Charge:4),
(Position:155.422;Intensity:300;Charge:5),
(Position:156.246;Intensity:12;Charge:4),
(Position:156.379;Intensity:12;Charge:2),
(Position:156.626;Intensity:13;Charge:4),
(Position:156.827;Intensity:14;Charge:4),
(Position:157.018;Intensity:12;Charge:4),
(Position:157.042;Intensity:12;Charge:4),
(Position:157.124;Intensity:12;Charge:4),
(Position:157.720;Intensity:12;Charge:4),
(Position:157.776;Intensity:12;Charge:4),
(Position:158.062;Intensity:12;Charge:2),
(Position:159.062;Intensity:12;Charge:4),
(Position:159.151;Intensity:13;Charge:4),
(Position:159.205;Intensity:13;Charge:4),
(Position:159.667;Intensity:12;Charge:4),
(Position:159.801;Intensity:13;Charge:4),
(Position:160.050;Intensity:12;Charge:4),
(Position:160.058;Intensity:13;Charge:4),
(Position:160.121;Intensity:10;Charge:3),
(Position:160.167;Intensity:13;Charge:4),
(Position:160.208;Intensity:12;Charge:4),
(Position:160.318;Intensity:13;Charge:4),
(Position:160.373;Intensity:13;Charge:4),
(Position:160.488;Intensity:13;Charge:4),
(Position:160.568;Intensity:13;Charge:4),
(Position:160.597;Intensity:15;Charge:4),
(Position:160.698;Intensity:13;Charge:4),
(Position:160.846;Intensity:18;Charge:2),
(Position:160.910;Intensity:17;Charge:4),
(Position:160.983;Intensity:14;Charge:4),
(Position:161.047;Intensity:13;Charge:4),
(Position:161.120;Intensity:13;Charge:4),
(Position:161.364;Intensity:13;Charge:4),
(Position:161.402;Intensity:15;Charge:4),
(Position:161.464;Intensity:13;Charge:4),
(Position:161.500;Intensity:13;Charge:4),
(Position:161.561;Intensity:12;Charge:4),
(Position:161.668;Intensity:16;Charge:4),
(Position:161.768;Intensity:14;Charge:4),
(Position:161.847;Intensity:12;Charge:2),
(Position:161.902;Intensity:14;Charge:4),
(Position:162.091;Intensity:12;Charge:4),
(Position:162.116;Intensity:13;Charge:4),
(Position:162.157;Intensity:14;Charge:4),
(Position:162.168;Intensity:15;Charge:2),
(Position:162.338;Intensity:13;Charge:4),
(Position:162.353;Intensity:13;Charge:4),
(Position:162.647;Intensity:15;Charge:4),
(Position:162.690;Intensity:14;Charge:4),
(Position:162.854;Intensity:13;Charge:4),
(Position:162.915;Intensity:15;Charge:2),
(Position:163.018;Intensity:13;Charge:4),
(Position:163.108;Intensity:17;Charge:4),
(Position:163.112;Intensity:15;Charge:2),
(Position:163.208;Intensity:12;Charge:4),
(Position:163.240;Intensity:14;Charge:4),
(Position:163.401;Intensity:13;Charge:4),
(Position:163.540;Intensity:18;Charge:2),
(Position:163.632;Intensity:15;Charge:2),
(Position:163.807;Intensity:12;Charge:4),
(Position:163.830;Intensity:12;Charge:4),
(Position:163.940;Intensity:15;Charge:2),
(Position:163.940;Intensity:14;Charge:4),
(Position:164.004;Intensity:16;Charge:4),
(Position:164.016;Intensity:14;Charge:4),
(Position:164.176;Intensity:12;Charge:2),
(Position:164.187;Intensity:15;Charge:4),
(Position:164.288;Intensity:12;Charge:4),
(Position:164.709;Intensity:15;Charge:4),
(Position:164.716;Intensity:12;Charge:2),
(Position:165.158;Intensity:15;Charge:4),
(Position:165.290;Intensity:15;Charge:4),
(Position:165.341;Intensity:13;Charge:4),
(Position:165.611;Intensity:13;Charge:4),
(Position:165.665;Intensity:15;Charge:4),
(Position:165.782;Intensity:12;Charge:4),
(Position:165.843;Intensity:12;Charge:4),
(Position:166.010;Intensity:14;Charge:4),
(Position:166.157;Intensity:12;Charge:4),
(Position:166.232;Intensity:13;Charge:4),
(Position:166.252;Intensity:13;Charge:4),
(Position:166.354;Intensity:13;Charge:4),
(Position:166.809;Intensity:13;Charge:4),
(Position:166.961;Intensity:12;Charge:4),
(Position:167.074;Intensity:12;Charge:2),
(Position:167.104;Intensity:14;Charge:4),
(Position:167.286;Intensity:12;Charge:4),
(Position:167.368;Intensity:13;Charge:4),
(Position:167.566;Intensity:14;Charge:4),
(Position:167.678;Intensity:12;Charge:4),
(Position:167.712;Intensity:12;Charge:4),
(Position:168.136;Intensity:13;Charge:4),
(Position:168.195;Intensity:12;Charge:4),
(Position:168.769;Intensity:15;Charge:4),
(Position:169.888;Intensity:15;Charge:4),
(Position:170.040;Intensity:12;Charge:4),
(Position:170.204;Intensity:12;Charge:2),
(Position:170.493;Intensity:12;Charge:4),
(Position:170.981;Intensity:13;Charge:4),
(Position:171.141;Intensity:15;Charge:4),
(Position:171.276;Intensity:14;Charge:4),
(Position:171.711;Intensity:12;Charge:4),
(Position:171.790;Intensity:14;Charge:4),
(Position:171.816;Intensity:14;Charge:4),
(Position:171.842;Intensity:12;Charge:4),
(Position:171.946;Intensity:14;Charge:4),
(Position:172.271;Intensity:14;Charge:4),
(Position:172.406;Intensity:14;Charge:4),
(Position:172.426;Intensity:12;Charge:4),
(Position:172.563;Intensity:16;Charge:4),
(Position:176.108;Intensity:13;Charge:4),
(Position:176.138;Intensity:12;Charge:2),
(Position:176.492;Intensity:12;Charge:4),
(Position:176.736;Intensity:12;Charge:4),
(Position:178.526;Intensity:20;Charge:2),
(Position:178.674;Intensity:20;Charge:2),
(Position:178.807;Intensity:18;Charge:2),
(Position:179.210;Intensity:13;Charge:4),
(Position:179.693;Intensity:13;Charge:4),
(Position:180.532;Intensity:12;Charge:4),
(Position:182.042;Intensity:12;Charge:4),
(Position:182.798;Intensity:13;Charge:4),
(Position:184.024;Intensity:12;Charge:4),
(Position:186.042;Intensity:12;Charge:4),
(Position:186.964;Intensity:12;Charge:4),
(Position:186.983;Intensity:10;Charge:3),
(Position:187.423;Intensity:12;Charge:4),
(Position:187.799;Intensity:12;Charge:3),
(Position:188.205;Intensity:10;Charge:3),
(Position:188.676;Intensity:12;Charge:3),
(Position:189.067;Intensity:13;Charge:3),
(Position:189.398;Intensity:11;Charge:3),
(Position:189.546;Intensity:20;Charge:3),
(Position:190.758;Intensity:10;Charge:3),
(Position:191.406;Intensity:19;Charge:3),
(Position:191.508;Intensity:15;Charge:3),
(Position:192.279;Intensity:15;Charge:3),
(Position:192.601;Intensity:10;Charge:3),
(Position:192.630;Intensity:18;Charge:3),
(Position:193.039;Intensity:15;Charge:3),
(Position:193.151;Intensity:14;Charge:3),
(Position:193.4538;Intensity:30;Charge:1),
(Position:193.7269;Intensity:25;Charge:1),
(Position:193.734;Intensity:14;Charge:3),
(Position:193.890;Intensity:10;Charge:3),
(Position:194.348;Intensity:14;Charge:3),
(Position:194.534;Intensity:12;Charge:3),
(Position:194.6988;Intensity:50;Charge:1),
(Position:195.033;Intensity:10;Charge:3),
(Position:195.101;Intensity:12;Charge:3),
(Position:195.1571;Intensity:25;Charge:1),
(Position:195.259;Intensity:30;Charge:1),
(Position:195.265;Intensity:11;Charge:3),
(Position:195.3005;Intensity:30;Charge:1),
(Position:195.332;Intensity:13;Charge:3),
(Position:195.349;Intensity:10;Charge:3),
(Position:195.422;Intensity:10;Charge:3),
(Position:195.7823;Intensity:60;Charge:1),
(Position:195.858;Intensity:11;Charge:3),
(Position:196.0144;Intensity:60;Charge:1),
(Position:196.032;Intensity:13;Charge:3),
(Position:196.125;Intensity:30;Charge:1),
(Position:196.2111;Intensity:50;Charge:1),
(Position:196.311;Intensity:12;Charge:2),
(Position:198.750;Intensity:15;Charge:3),
(Position:199.161;Intensity:14;Charge:3),
(Position:199.407;Intensity:13;Charge:3),
(Position:199.556;Intensity:12;Charge:3),
(Position:199.642;Intensity:12;Charge:3),
(Position:206.221;Intensity:10;Charge:3),
(Position:206.890;Intensity:12;Charge:3),
(Position:207.965;Intensity:14;Charge:3),
(Position:208.4786;Intensity:100;Charge:1),
(Position:208.501;Intensity:10;Charge:3),
(Position:209.080;Intensity:12;Charge:3),
(Position:209.815;Intensity:15;Charge:3),
(Position:209.836;Intensity:12;Charge:3),
(Position:210.447;Intensity:12;Charge:3),
(Position:210.799;Intensity:10;Charge:3),
(Position:215.246;Intensity:15;Charge:3),
(Position:215.839;Intensity:12;Charge:3),
(Position:215.8472;Intensity:50;Charge:1),
(Position:215.915;Intensity:12;Charge:3),
(Position:216.195;Intensity:10;Charge:3),
(Position:216.270;Intensity:15;Charge:2),
(Position:216.7453;Intensity:40;Charge:1),
(Position:216.763;Intensity:12;Charge:3),
(Position:217.172;Intensity:12;Charge:3),
(Position:217.534;Intensity:15;Charge:3),
(Position:217.8800;Intensity:300;Charge:1),
(Position:218.109;Intensity:12;Charge:3),
(Position:218.7170;Intensity:250;Charge:1),
(Position:218.7576;Intensity:60;Charge:1),
(Position:218.7879;Intensity:120;Charge:1),
(Position:219.2524;Intensity:250;Charge:1),
(Position:219.6729;Intensity:150;Charge:1),
(Position:220.1077;Intensity:80;Charge:1),
(Position:220.1411;Intensity:80;Charge:1),
(Position:220.910;Intensity:15;Charge:2),
(Position:220.954;Intensity:10;Charge:3),
(Position:221.434;Intensity:20;Charge:2),
(Position:221.895;Intensity:12;Charge:2),
(Position:222.107;Intensity:20;Charge:2),
(Position:222.252;Intensity:10;Charge:3),
(Position:222.996;Intensity:10;Charge:3),
(Position:223.312;Intensity:10;Charge:3),
(Position:223.338;Intensity:10;Charge:3),
(Position:223.660;Intensity:10;Charge:3),
(Position:223.885;Intensity:10;Charge:3),
(Position:224.224;Intensity:12;Charge:3),
(Position:224.628;Intensity:25;Charge:2),
(Position:225.1488;Intensity:50;Charge:1),
(Position:225.2572;Intensity:60;Charge:1),
(Position:225.647;Intensity:25;Charge:2),
(Position:226.0211;Intensity:300;Charge:1),
(Position:226.229;Intensity:12;Charge:3),
(Position:226.5090;Intensity:60;Charge:1),
(Position:226.7786;Intensity:80;Charge:1),
(Position:226.812;Intensity:10;Charge:3),
(Position:226.8170;Intensity:80;Charge:1),
(Position:227.1564;Intensity:50;Charge:1),
(Position:227.2772;Intensity:150;Charge:1),
(Position:227.6729;Intensity:150;Charge:1),
(Position:228.0641;Intensity:80;Charge:1),
(Position:228.4791;Intensity:150;Charge:1),
(Position:228.7956;Intensity:150;Charge:1),
(Position:229.3231;Intensity:300;Charge:1),
(Position:229.377;Intensity:10;Charge:3),
(Position:229.512;Intensity:80;Charge:1),
(Position:229.657;Intensity:15;Charge:3),
(Position:229.8495;Intensity:200;Charge:1),
(Position:229.8877;Intensity:600;Charge:1),
(Position:229.9928;Intensity:80;Charge:1),
(Position:230.0850;Intensity:300;Charge:1),
(Position:230.2393;Intensity:50;Charge:1),
(Position:230.4133;Intensity:100;Charge:1),
(Position:230.4290;Intensity:150;Charge:1),
(Position:230.9709;Intensity:120;Charge:1),
(Position:231.3815;Intensity:150;Charge:1),
(Position:231.841;Intensity:10;Charge:3),
(Position:231.993;Intensity:10;Charge:3),
(Position:232.1071;Intensity:200;Charge:1),
(Position:232.242;Intensity:10;Charge:3),
(Position:232.766;Intensity:10;Charge:3),
(Position:232.811;Intensity:100;Charge:2),
(Position:232.859;Intensity:15;Charge:2),
(Position:233.203;Intensity:100;Charge:2),
(Position:233.269;Intensity:15;Charge:2),
(Position:233.352;Intensity:300;Charge:2),
(Position:233.749;Intensity:10;Charge:3),
(Position:233.873;Intensity:200;Charge:2),
(Position:233.968;Intensity:10;Charge:3),
(Position:234.421;Intensity:600;Charge:2),
(Position:234.468;Intensity:80;Charge:2),
(Position:234.500;Intensity:150;Charge:2),
(Position:234.570;Intensity:25;Charge:2),
(Position:234.606;Intensity:50;Charge:2),
(Position:234.883;Intensity:200;Charge:2),
(Position:234.902;Intensity:250;Charge:2),
(Position:235.192;Intensity:50;Charge:2),
(Position:235.239;Intensity:15;Charge:2),
(Position:235.303;Intensity:25;Charge:2),
(Position:235.419;Intensity:30;Charge:2),
(Position:235.440;Intensity:15;Charge:2),
(Position:235.520;Intensity:50;Charge:2),
(Position:235.561;Intensity:40;Charge:2),
(Position:235.984;Intensity:200;Charge:2),
(Position:236.031;Intensity:15;Charge:2),
(Position:236.072;Intensity:150;Charge:2),
(Position:236.101;Intensity:120;Charge:2),
(Position:236.123;Intensity:30;Charge:2),
(Position:236.274;Intensity:40;Charge:2),
(Position:236.458;Intensity:60;Charge:2),
(Position:236.555;Intensity:200;Charge:2),
(Position:236.648;Intensity:80;Charge:2),
(Position:236.731;Intensity:25;Charge:2),
(Position:236.931;Intensity:80;Charge:2),
(Position:237.0180;Intensity:80;Charge:1),
(Position:237.067;Intensity:80;Charge:2),
(Position:237.122;Intensity:25;Charge:2),
(Position:237.2154;Intensity:120;Charge:1),
(Position:237.4349;Intensity:300;Charge:1),
(Position:237.446;Intensity:150;Charge:2),
(Position:237.5243;Intensity:120;Charge:1),
(Position:237.592;Intensity:60;Charge:2),
(Position:237.716;Intensity:120;Charge:2),
(Position:237.886;Intensity:20;Charge:2),
(Position:238.000;Intensity:80;Charge:2),
(Position:238.014;Intensity:20;Charge:2),
(Position:238.093;Intensity:40;Charge:2),
(Position:238.149;Intensity:120;Charge:2),
(Position:238.2562;Intensity:150;Charge:1),
(Position:238.277;Intensity:1000;Charge:2),
(Position:238.363;Intensity:20;Charge:2),
(Position:238.379;Intensity:20;Charge:2),
(Position:238.398;Intensity:60;Charge:2),
(Position:238.512;Intensity:50;Charge:2),
(Position:238.910;Intensity:40;Charge:2),
(Position:238.936;Intensity:300;Charge:2),
(Position:239.026;Intensity:8;Charge:3),
(Position:239.0701;Intensity:200;Charge:1),
(Position:239.083;Intensity:30;Charge:2),
(Position:239.150;Intensity:20;Charge:2),
(Position:239.221;Intensity:15;Charge:2),
(Position:239.331;Intensity:20;Charge:2),
(Position:239.615;Intensity:40;Charge:2),
(Position:239.635;Intensity:1000;Charge:2),
(Position:239.745;Intensity:15;Charge:2),
(Position:239.997;Intensity:300;Charge:2),
(Position:240.078;Intensity:20;Charge:2),
(Position:240.202;Intensity:15;Charge:2),
(Position:240.516;Intensity:50;Charge:2),
(Position:240.561;Intensity:800;Charge:2),
(Position:240.739;Intensity:250;Charge:2),
(Position:240.770;Intensity:80;Charge:2),
(Position:241.125;Intensity:300;Charge:2),
(Position:241.180;Intensity:200;Charge:2),
(Position:241.254;Intensity:50;Charge:2),
(Position:241.404;Intensity:150;Charge:2),
(Position:241.718;Intensity:20;Charge:2),
(Position:241.860;Intensity:80;Charge:2),
(Position:241.917;Intensity:15;Charge:2),
(Position:242.1131;Intensity:60;Charge:1),
(Position:242.343;Intensity:60;Charge:2),
(Position:242.3825;Intensity:60;Charge:1),
(Position:242.395;Intensity:40;Charge:2),
(Position:242.488;Intensity:150;Charge:2),
(Position:242.513;Intensity:15;Charge:2),
(Position:242.533;Intensity:30;Charge:2),
(Position:242.910;Intensity:120;Charge:2),
(Position:242.903;Intensity:30;Charge:2),
(Position:242.954;Intensity:25;Charge:2),
(Position:242.977;Intensity:25;Charge:2),
(Position:243.013;Intensity:20;Charge:2),
(Position:243.060;Intensity:30;Charge:2),
(Position:243.082;Intensity:120;Charge:2),
(Position:243.176;Intensity:25;Charge:2),
(Position:243.300;Intensity:80;Charge:2),
(Position:243.361;Intensity:60;Charge:2),
(Position:243.480;Intensity:25;Charge:2),
(Position:243.498;Intensity:20;Charge:2),
(Position:243.539;Intensity:20;Charge:2),
(Position:243.547;Intensity:50;Charge:2),
(Position:243.569;Intensity:50;Charge:2),
(Position:243.736;Intensity:25;Charge:2),
(Position:243.891;Intensity:8;Charge:3),
(Position:243.8921;Intensity:60;Charge:1),
(Position:244.004;Intensity:150;Charge:2),
(Position:244.048;Intensity:150;Charge:1),
(Position:244.085;Intensity:80;Charge:1),
(Position:244.116;Intensity:40;Charge:2),
(Position:244.311;Intensity:30;Charge:2),
(Position:244.331;Intensity:100;Charge:1),
(Position:244.445;Intensity:60;Charge:2),
(Position:244.4613;Intensity:250;Charge:1),
(Position:244.525;Intensity:100;Charge:2),
(Position:244.585;Intensity:50;Charge:2),
(Position:244.5953;Intensity:50;Charge:1),
(Position:244.631;Intensity:100;Charge:2),
(Position:244.654;Intensity:40;Charge:2),
(Position:244.685;Intensity:50;Charge:2),
(Position:244.721;Intensity:30;Charge:2),
(Position:244.794;Intensity:40;Charge:2),
(Position:244.807;Intensity:25;Charge:2),
(Position:244.8451;Intensity:60;Charge:1),
(Position:244.849;Intensity:30;Charge:2),
(Position:245.070;Intensity:25;Charge:2),
(Position:245.094;Intensity:25;Charge:2),
(Position:245.4219;Intensity:100;Charge:1),
(Position:245.472;Intensity:20;Charge:2),
(Position:245.532;Intensity:30;Charge:2),
(Position:245.645;Intensity:15;Charge:2),
(Position:245.664;Intensity:15;Charge:2),
(Position:245.783;Intensity:15;Charge:2),
(Position:245.8342;Intensity:1500;Charge:1),
(Position:245.952;Intensity:150;Charge:2),
(Position:245.971;Intensity:40;Charge:2),
(Position:246.118;Intensity:60;Charge:2),
(Position:246.202;Intensity:80;Charge:2),
(Position:246.260;Intensity:100;Charge:2),
(Position:246.2926;Intensity:100;Charge:1),
(Position:246.3392;Intensity:1500;Charge:1),
(Position:246.404;Intensity:50;Charge:2),
(Position:246.4475;Intensity:50;Charge:1),
(Position:246.476;Intensity:40;Charge:2),
(Position:246.565;Intensity:40;Charge:2),
(Position:246.5895;Intensity:800;Charge:1),
(Position:246.666;Intensity:50;Charge:2),
(Position:246.725;Intensity:15;Charge:2),
(Position:246.742;Intensity:60;Charge:2),
(Position:246.757;Intensity:60;Charge:2),
(Position:246.8478;Intensity:60;Charge:1),
(Position:246.904;Intensity:15;Charge:2),
(Position:246.9625;Intensity:600;Charge:1),
(Position:247.026;Intensity:60;Charge:2),
(Position:247.116;Intensity:25;Charge:2),
(Position:247.142;Intensity:80;Charge:2),
(Position:247.1712;Intensity:80;Charge:1),
(Position:247.3083;Intensity:800;Charge:1),
(Position:247.335;Intensity:40;Charge:2),
(Position:247.318;Intensity:40;Charge:2),
(Position:247.3642;Intensity:1000;Charge:1),
(Position:247.391;Intensity:200;Charge:1),
(Position:247.407;Intensity:50;Charge:2),
(Position:247.480;Intensity:30;Charge:2),
(Position:247.5562;Intensity:600;Charge:1),
(Position:247.587;Intensity:50;Charge:2),
(Position:247.629;Intensity:40;Charge:2),
(Position:247.701;Intensity:15;Charge:2),
(Position:247.7405;Intensity:60;Charge:1),
(Position:247.809;Intensity:25;Charge:2),
(Position:247.932;Intensity:60;Charge:2),
(Position:248.0229;Intensity:120;Charge:1),
(Position:248.0525;Intensity:1200;Charge:1),
(Position:248.091;Intensity:100;Charge:2),
(Position:248.180;Intensity:15;Charge:2),
(Position:248.287;Intensity:80;Charge:2),
(Position:248.307;Intensity:25;Charge:2),
(Position:248.341;Intensity:100;Charge:2),
(Position:248.362;Intensity:15;Charge:2),
(Position:248.4021;Intensity:10000;Charge:1),
(Position:248.4283;Intensity:300;Charge:1),
(Position:248.447;Intensity:15;Charge:2),
(Position:248.4935;Intensity:1000;Charge:1),
(Position:248.499;Intensity:60;Charge:2),
(Position:248.519;Intensity:30;Charge:2),
(Position:248.6740;Intensity:50;Charge:1),
(Position:248.7123;Intensity:800;Charge:1),
(Position:248.7442;Intensity:100;Charge:1),
(Position:248.7817;Intensity:100;Charge:1),
(Position:248.8121;Intensity:120;Charge:1),
(Position:248.8894;Intensity:4000;Charge:1),
(Position:248.9696;Intensity:100;Charge:1),
(Position:249.023;Intensity:80;Charge:2),
(Position:249.0501;Intensity:1000;Charge:1),
(Position:249.058;Intensity:50;Charge:2),
(Position:249.0664;Intensity:50;Charge:1),
(Position:249.1395;Intensity:3000;Charge:1),
(Position:249.146;Intensity:100;Charge:2),
(Position:249.161;Intensity:60;Charge:2),
(Position:249.1907;Intensity:2000;Charge:1),
(Position:249.215;Intensity:100;Charge:2),
(Position:249.309;Intensity:25;Charge:2),
(Position:249.401;Intensity:500;Charge:2),
(Position:249.393;Intensity:100;Charge:2),
(Position:249.463;Intensity:20;Charge:2),
(Position:249.4752;Intensity:60;Charge:1),
(Position:249.5003;Intensity:50;Charge:1),
(Position:249.662;Intensity:100;Charge:1),
(Position:249.7286;Intensity:600;Charge:1),
(Position:249.857;Intensity:50;Charge:2),
(Position:249.965;Intensity:150;Charge:1),
(Position:250.167;Intensity:40;Charge:2),
(Position:250.1886;Intensity:1000;Charge:1),
(Position:250.206;Intensity:40;Charge:2),
(Position:250.2447;Intensity:50;Charge:1),
(Position:250.314;Intensity:60;Charge:2),
(Position:250.408;Intensity:40;Charge:2),
(Position:250.432;Intensity:20;Charge:2),
(Position:250.462;Intensity:60;Charge:2),
(Position:250.685;Intensity:80;Charge:2),
(Position:250.756;Intensity:40;Charge:2),
(Position:250.8655;Intensity:500;Charge:1),
(Position:250.910;Intensity:30;Charge:2),
(Position:250.9509;Intensity:50;Charge:1),
(Position:251.1591;Intensity:1000;Charge:1),
(Position:251.252;Intensity:120;Charge:2),
(Position:251.3032;Intensity:80;Charge:1),
(Position:251.3122;Intensity:400;Charge:1),
(Position:251.514;Intensity:50;Charge:2),
(Position:251.7328;Intensity:80;Charge:1),
(Position:251.789;Intensity:50;Charge:2),
(Position:251.8419;Intensity:300;Charge:1),
(Position:251.8860;Intensity:800;Charge:1),
(Position:251.981;Intensity:60;Charge:2),
(Position:252.0387;Intensity:150;Charge:1),
(Position:252.185;Intensity:40;Charge:2),
(Position:252.258;Intensity:30;Charge:2),
(Position:252.3239;Intensity:50;Charge:1),
(Position:252.3608;Intensity:4000;Charge:1),
(Position:252.442;Intensity:200;Charge:1),
(Position:252.5052;Intensity:500;Charge:1),
(Position:252.578;Intensity:100;Charge:1),
(Position:252.615;Intensity:200;Charge:2),
(Position:252.683;Intensity:25;Charge:2),
(Position:252.705;Intensity:300;Charge:2),
(Position:252.786;Intensity:20;Charge:2),
(Position:252.8195;Intensity:2000;Charge:1),
(Position:252.846;Intensity:30;Charge:2),
(Position:252.964;Intensity:20;Charge:2),
(Position:252.984;Intensity:20;Charge:2),
(Position:252.9895;Intensity:800;Charge:1),
(Position:252.999;Intensity:25;Charge:2),
(Position:253.007;Intensity:80;Charge:1),
(Position:253.031;Intensity:250;Charge:2),
(Position:253.0597;Intensity:150;Charge:1),
(Position:253.087;Intensity:40;Charge:2),
(Position:253.1448;Intensity:200;Charge:1),
(Position:253.263;Intensity:20;Charge:2),
(Position:253.439;Intensity:120;Charge:2),
(Position:253.456;Intensity:60;Charge:1),
(Position:253.518;Intensity:100;Charge:2),
(Position:253.625;Intensity:120;Charge:2),
(Position:253.6369;Intensity:400;Charge:1),
(Position:253.743;Intensity:60;Charge:2),
(Position:253.7554;Intensity:200;Charge:1),
(Position:253.756;Intensity:200;Charge:2),
(Position:253.760;Intensity:50;Charge:2),
(Position:253.790;Intensity:50;Charge:2),
(Position:253.896;Intensity:50;Charge:2),
(Position:253.926;Intensity:40;Charge:2),
(Position:253.944;Intensity:20;Charge:2),
(Position:253.956;Intensity:100;Charge:2),
(Position:253.975;Intensity:150;Charge:2),
(Position:253.967;Intensity:100;Charge:2),
(Position:254.0120;Intensity:50;Charge:1),
(Position:254.128;Intensity:20;Charge:2),
(Position:254.142;Intensity:200;Charge:2),
(Position:254.1735;Intensity:600;Charge:1),
(Position:254.186;Intensity:80;Charge:2),
(Position:254.260;Intensity:60;Charge:2),
(Position:254.286;Intensity:300;Charge:1),
(Position:254.354;Intensity:25;Charge:2),
(Position:254.414;Intensity:60;Charge:2),
(Position:254.468;Intensity:250;Charge:1),
(Position:254.546;Intensity:150;Charge:1),
(Position:254.573;Intensity:40;Charge:2),
(Position:254.598;Intensity:40;Charge:2),
(Position:254.620;Intensity:20;Charge:2),
(Position:254.6742;Intensity:800;Charge:1),
(Position:254.720;Intensity:40;Charge:2),
(Position:254.743;Intensity:80;Charge:2),
(Position:254.763;Intensity:80;Charge:1),
(Position:254.936;Intensity:20;Charge:2),
(Position:254.951;Intensity:100;Charge:2),
(Position:254.985;Intensity:80;Charge:2),
(Position:255.016;Intensity:80;Charge:2),
(Position:255.023;Intensity:60;Charge:2),
(Position:255.0378;Intensity:600;Charge:1),
(Position:255.054;Intensity:40;Charge:2),
(Position:255.080;Intensity:60;Charge:2),
(Position:255.092;Intensity:25;Charge:2),
(Position:255.145;Intensity:50;Charge:2),
(Position:256.105;Intensity:40;Charge:2),
(Position:256.286;Intensity:25;Charge:2),
(Position:256.330;Intensity:400;Charge:2),
(Position:256.425;Intensity:200;Charge:2),
(Position:256.699;Intensity:20;Charge:2),
(Position:256.768;Intensity:60;Charge:2),
(Position:257.129;Intensity:25;Charge:2),
(Position:257.162;Intensity:30;Charge:2),
(Position:257.513;Intensity:150;Charge:2),
(Position:257.651;Intensity:50;Charge:1),
(Position:257.7463;Intensity:300;Charge:1),
(Position:257.763;Intensity:25;Charge:2),
(Position:257.869;Intensity:60;Charge:2),
(Position:258.307;Intensity:50;Charge:1),
(Position:258.314;Intensity:8;Charge:3),
(Position:258.335;Intensity:100;Charge:2),
(Position:258.531;Intensity:1500;Charge:1),
(Position:258.665;Intensity:650;Charge:2),
(Position:258.877;Intensity:90;Charge:1),
(Position:259.232;Intensity:90;Charge:2),
(Position:259.356;Intensity:30;Charge:2),
(Position:259.429;Intensity:60;Charge:1),
(Position:259.451;Intensity:90;Charge:2),
(Position:259.640;Intensity:8;Charge:3),
(Position:259.915;Intensity:650;Charge:2),
(Position:260.018;Intensity:2000;Charge:2),
(Position:260.035;Intensity:300;Charge:1),
(Position:260.612;Intensity:20;Charge:2),
(Position:260.620;Intensity:20;Charge:2),
(Position:260.6435;Intensity:60;Charge:1),
(Position:260.729;Intensity:300;Charge:2),
(Position:260.7606;Intensity:800;Charge:1),
(Position:260.787;Intensity:650;Charge:2),
(Position:261.185;Intensity:20;Charge:2),
(Position:261.265;Intensity:600;Charge:2),
(Position:261.460;Intensity:320;Charge:2),
(Position:261.793;Intensity:8;Charge:3),
(Position:261.840;Intensity:320;Charge:2),
(Position:261.8799;Intensity:250;Charge:1),
(Position:261.985;Intensity:20;Charge:2),
(Position:262.119;Intensity:90;Charge:2),
(Position:262.147;Intensity:20;Charge:2),
(Position:262.245;Intensity:40;Charge:2),
(Position:262.431;Intensity:400;Charge:1),
(Position:262.627;Intensity:50;Charge:2),
(Position:262.645;Intensity:200;Charge:2),
(Position:262.907;Intensity:150;Charge:2),
(Position:263.085;Intensity:20;Charge:2),
(Position:263.183;Intensity:250;Charge:2),
(Position:263.210;Intensity:250;Charge:2),
(Position:263.239;Intensity:50;Charge:2),
(Position:263.3022;Intensity:100;Charge:1),
(Position:263.6595;Intensity:300;Charge:1),
(Position:264.2433;Intensity:50;Charge:1),
(Position:264.4786;Intensity:200;Charge:1),
(Position:264.618;Intensity:9;Charge:3),
(Position:266.545;Intensity:60;Charge:2),
(Position:266.743;Intensity:30;Charge:2),
(Position:266.7605;Intensity:300;Charge:1),
(Position:266.7758;Intensity:60;Charge:1),
(Position:267.9858;Intensity:600;Charge:1),
(Position:268.555;Intensity:500;Charge:2),
(Position:269.0010;Intensity:400;Charge:1),
(Position:269.340;Intensity:60;Charge:2),
(Position:269.593;Intensity:10;Charge:3),
(Position:269.614;Intensity:9;Charge:3),
(Position:269.708;Intensity:50;Charge:1),
(Position:269.9907;Intensity:200;Charge:1),
(Position:270.082;Intensity:8;Charge:3),
(Position:270.193;Intensity:8;Charge:3),
(Position:270.479;Intensity:60;Charge:2),
(Position:270.6815;Intensity:80;Charge:1),
(Position:270.7385;Intensity:400;Charge:1),
(Position:270.9374;Intensity:60;Charge:1),
(Position:270.985;Intensity:20;Charge:2),
(Position:271.2459;Intensity:200;Charge:1),
(Position:271.521;Intensity:80;Charge:2),
(Position:271.702;Intensity:50;Charge:2),
(Position:271.7062;Intensity:50;Charge:1),
(Position:271.8591;Intensity:50;Charge:1),
(Position:271.868;Intensity:50;Charge:2),
(Position:271.9242;Intensity:250;Charge:1),
(Position:271.9833;Intensity:4000;Charge:1),
(Position:272.0226;Intensity:100;Charge:1),
(Position:272.1003;Intensity:50;Charge:1),
(Position:272.1709;Intensity:1500;Charge:1),
(Position:272.4385;Intensity:400;Charge:1),
(Position:272.569;Intensity:30;Charge:2),
(Position:272.5760;Intensity:150;Charge:1),
(Position:272.686;Intensity:80;Charge:1),
(Position:272.7042;Intensity:50;Charge:1),
(Position:272.819;Intensity:25;Charge:2),
(Position:272.835;Intensity:80;Charge:2),
(Position:272.8828;Intensity:200;Charge:1),
(Position:272.9628;Intensity:50;Charge:1),
(Position:272.971;Intensity:80;Charge:2),
(Position:273.154;Intensity:40;Charge:2),
(Position:273.4390;Intensity:1000;Charge:1),
(Position:273.4814;Intensity:60;Charge:1),
(Position:273.5077;Intensity:50;Charge:1),
(Position:273.6285;Intensity:500;Charge:1),
(Position:273.6422;Intensity:50;Charge:1),
(Position:273.8120;Intensity:500;Charge:1),
(Position:273.864;Intensity:120;Charge:1),
(Position:274.036;Intensity:400;Charge:2),
(Position:274.3065;Intensity:250;Charge:1),
(Position:274.3216;Intensity:800;Charge:1),
(Position:274.401;Intensity:200;Charge:2),
(Position:274.4377;Intensity:150;Charge:1),
(Position:274.4880;Intensity:200;Charge:1),
(Position:274.5339;Intensity:80;Charge:1),
(Position:274.729;Intensity:300;Charge:2),
(Position:275.013;Intensity:100;Charge:2),
(Position:275.029;Intensity:500;Charge:2),
(Position:275.0953;Intensity:1200;Charge:1),
(Position:275.194;Intensity:20;Charge:2),
(Position:275.296;Intensity:20;Charge:2),
(Position:275.410;Intensity:80;Charge:2),
(Position:275.450;Intensity:50;Charge:1),
(Position:275.4846;Intensity:150;Charge:1),
(Position:275.5240;Intensity:100;Charge:1),
(Position:275.570;Intensity:30;Charge:2),
(Position:275.654;Intensity:800;Charge:2),
(Position:275.7143;Intensity:250;Charge:1),
(Position:275.8131;Intensity:100;Charge:1),
(Position:276.063;Intensity:50;Charge:1),
(Position:276.2596;Intensity:120;Charge:1),
(Position:276.263;Intensity:150;Charge:2),
(Position:276.2842;Intensity:150;Charge:1),
(Position:276.3588;Intensity:120;Charge:1),
(Position:276.3925;Intensity:120;Charge:1),
(Position:276.448;Intensity:20;Charge:2),
(Position:276.595;Intensity:25;Charge:2),
(Position:276.7727;Intensity:80;Charge:1),
(Position:276.8339;Intensity:250;Charge:1),
(Position:277.012;Intensity:50;Charge:1),
(Position:277.017;Intensity:25;Charge:2),
(Position:277.289;Intensity:300;Charge:1),
(Position:277.405;Intensity:50;Charge:1),
(Position:277.413;Intensity:8;Charge:3),
(Position:277.551;Intensity:20;Charge:2),
(Position:277.773;Intensity:15;Charge:2),
(Position:277.889;Intensity:60;Charge:1),
(Position:277.9040;Intensity:600;Charge:1),
(Position:278.012;Intensity:40;Charge:2),
(Position:278.451;Intensity:50;Charge:2),
(Position:278.601;Intensity:30;Charge:2),
(Position:278.892;Intensity:3000;Charge:1),
(Position:279.471;Intensity:20;Charge:2),
(Position:279.860;Intensity:200;Charge:1),
(Position:280.012;Intensity:30;Charge:2),
(Position:280.5347;Intensity:400;Charge:1),
(Position:280.781;Intensity:1500;Charge:1),
(Position:281.407;Intensity:10;Charge:3),
(Position:281.4116;Intensity:2500;Charge:1),
(Position:282.4107;Intensity:300;Charge:1),
(Position:282.639;Intensity:600;Charge:1),
(Position:282.6519;Intensity:50;Charge:1),
(Position:282.9640;Intensity:120;Charge:1),
(Position:283.239;Intensity:25;Charge:2),
(Position:283.3269;Intensity:1500;Charge:1),
(Position:283.6784;Intensity:120;Charge:1),
(Position:283.8954;Intensity:200;Charge:1),
(Position:284.034;Intensity:30;Charge:2),
(Position:284.064;Intensity:20;Charge:2),
(Position:284.149;Intensity:15;Charge:2),
(Position:284.4467;Intensity:200;Charge:1),
(Position:284.4813;Intensity:1000;Charge:1),
(Position:284.6430;Intensity:100;Charge:1),
(Position:284.895;Intensity:15;Charge:2),
(Position:284.916;Intensity:15;Charge:2),
(Position:285.2635;Intensity:800;Charge:1),
(Position:285.775;Intensity:30;Charge:2),
(Position:285.918;Intensity:25;Charge:2),
(Position:287.0149;Intensity:50;Charge:1),
(Position:287.3177;Intensity:50;Charge:1),
(Position:287.5015;Intensity:80;Charge:1),
(Position:289.5353;Intensity:50;Charge:1),
(Position:289.593;Intensity:8;Charge:3),
(Position:290.332;Intensity:9;Charge:3),
(Position:290.528;Intensity:12;Charge:3),
(Position:290.665;Intensity:8;Charge:3),
(Position:290.835;Intensity:10;Charge:3),
(Position:290.855;Intensity:12;Charge:3),
(Position:291.3010;Intensity:120;Charge:1),
(Position:292.476;Intensity:8;Charge:3),
(Position:292.9864;Intensity:120;Charge:1),
(Position:293.7762;Intensity:1200;Charge:1),
(Position:294.2203;Intensity:60;Charge:1),
(Position:294.526;Intensity:12;Charge:2),
(Position:294.8738;Intensity:1000;Charge:1),
(Position:294.925;Intensity:8;Charge:3),
(Position:295.110;Intensity:60;Charge:1),
(Position:295.4803;Intensity:600;Charge:1),
(Position:295.8228;Intensity:250;Charge:1),
(Position:296.085;Intensity:80;Charge:1),
(Position:296.410;Intensity:8;Charge:3),
(Position:296.6120;Intensity:150;Charge:1),
(Position:296.7764;Intensity:1500;Charge:1),
(Position:297.023;Intensity:120;Charge:1),
(Position:297.0966;Intensity:800;Charge:1),
(Position:297.139;Intensity:15;Charge:2),
(Position:297.4000;Intensity:1200;Charge:1),
(Position:297.4103;Intensity:500;Charge:1),
(Position:298.2315;Intensity:600;Charge:1),
(Position:298.4441;Intensity:1000;Charge:1),
(Position:298.564;Intensity:60;Charge:1),
(Position:298.569;Intensity:50;Charge:2),
(Position:298.641;Intensity:13;Charge:2),
(Position:299.5300;Intensity:1000;Charge:1),
(Position:299.5375;Intensity:250;Charge:1),
(Position:300.0387;Intensity:500;Charge:1),
(Position:300.1326;Intensity:120;Charge:1),
(Position:300.1823;Intensity:800;Charge:1),
(Position:300.250;Intensity:12;Charge:3),
(Position:300.2530;Intensity:60;Charge:1),
(Position:300.352;Intensity:15;Charge:2),
(Position:300.8158;Intensity:200;Charge:1),
(Position:300.816;Intensity:12;Charge:3),
(Position:300.902;Intensity:500;Charge:1),
(Position:301.0446;Intensity:120;Charge:1),
(Position:301.405;Intensity:15;Charge:3),
(Position:301.8506;Intensity:60;Charge:1),
(Position:301.9862;Intensity:60;Charge:1),
(Position:302.089;Intensity:60;Charge:2),
(Position:302.1371;Intensity:500;Charge:1),
(Position:302.1519;Intensity:1500;Charge:1),
(Position:302.1953;Intensity:600;Charge:1),
(Position:302.4913;Intensity:500;Charge:1),
(Position:302.6519;Intensity:150;Charge:1),
(Position:302.6723;Intensity:500;Charge:1),
(Position:303.1030;Intensity:80;Charge:1),
(Position:303.2096;Intensity:60;Charge:1),
(Position:303.5367;Intensity:60;Charge:1),
(Position:303.784;Intensity:40;Charge:2),
(Position:303.8273;Intensity:800;Charge:1),
(Position:304.2522;Intensity:80;Charge:1),
(Position:304.8491;Intensity:800;Charge:1),
(Position:305.8335;Intensity:600;Charge:1),
(Position:305.9975;Intensity:1000;Charge:1),
(Position:306.8135;Intensity:250;Charge:1),
(Position:307.6613;Intensity:120;Charge:1),
(Position:309.2475;Intensity:120;Charge:1),
(Position:309.9088;Intensity:80;Charge:1),
(Position:310.0795;Intensity:100;Charge:1),
(Position:310.0868;Intensity:100;Charge:1),
(Position:310.1203;Intensity:60;Charge:1),
(Position:310.1565;Intensity:100;Charge:1),
(Position:313.734;Intensity:10;Charge:3),
(Position:315.511;Intensity:12;Charge:2),
(Position:317.501;Intensity:10;Charge:3),
(Position:317.6364;Intensity:80;Charge:1),
(Position:317.691;Intensity:10;Charge:3),
(Position:317.893;Intensity:10;Charge:3),
(Position:318.5816;Intensity:150;Charge:1),
(Position:319.2582;Intensity:250;Charge:1),
(Position:319.4149;Intensity:500;Charge:1),
(Position:319.4222;Intensity:800;Charge:1),
(Position:319.700;Intensity:12;Charge:2),
(Position:319.7852;Intensity:200;Charge:1),
(Position:320.0425;Intensity:80;Charge:1),
(Position:320.140;Intensity:60;Charge:1),
(Position:320.6324;Intensity:50;Charge:1),
(Position:321.260;Intensity:50;Charge:1),
(Position:321.281;Intensity:100;Charge:1),
(Position:321.424;Intensity:13;Charge:2),
(Position:321.4939;Intensity:200;Charge:1),
(Position:321.5325;Intensity:200;Charge:1),
(Position:321.6867;Intensity:60;Charge:1),
(Position:321.8306;Intensity:50;Charge:1),
(Position:322.0513;Intensity:80;Charge:1),
(Position:322.0696;Intensity:60;Charge:1),
(Position:322.2975;Intensity:300;Charge:1),
(Position:322.671;Intensity:600;Charge:1),
(Position:322.866;Intensity:13;Charge:2),
(Position:322.8728;Intensity:80;Charge:1),
(Position:323.135;Intensity:20;Charge:2),
(Position:323.398;Intensity:80;Charge:1),
(Position:323.4900;Intensity:50;Charge:1),
(Position:323.5547;Intensity:120;Charge:1),
(Position:323.7156;Intensity:300;Charge:1),
(Position:324.0368;Intensity:100;Charge:1),
(Position:324.5123;Intensity:80;Charge:1),
(Position:324.6942;Intensity:80;Charge:1),
(Position:325.530;Intensity:60;Charge:1),
(Position:326.5987;Intensity:80;Charge:1),
(Position:326.6559;Intensity:50;Charge:1),
(Position:326.782;Intensity:13;Charge:3),
(Position:327.1943;Intensity:50;Charge:1),
(Position:327.702;Intensity:11;Charge:3),
(Position:328.121;Intensity:50;Charge:1),
(Position:328.770;Intensity:150;Charge:1),
(Position:328.976;Intensity:10;Charge:3),
(Position:330.617;Intensity:9;Charge:3),
(Position:330.692;Intensity:120;Charge:1),
(Position:330.7295;Intensity:200;Charge:1),
(Position:334.035;Intensity:9;Charge:3),
(Position:335.6191;Intensity:400;Charge:1),
(Position:335.6481;Intensity:80;Charge:1),
(Position:337.0514;Intensity:60;Charge:1),
(Position:337.1751;Intensity:120;Charge:1),
(Position:337.9648;Intensity:50;Charge:1),
(Position:338.1081;Intensity:50;Charge:1),
(Position:338.4950;Intensity:60;Charge:1),
(Position:338.910;Intensity:12;Charge:2),
(Position:339.3278;Intensity:50;Charge:1),
(Position:339.3625;Intensity:150;Charge:1),
(Position:340.0309;Intensity:150;Charge:1),
(Position:340.5330;Intensity:80;Charge:1),
(Position:340.8436;Intensity:500;Charge:1),
(Position:341.4110;Intensity:250;Charge:1),
(Position:342.5266;Intensity:60;Charge:1),
(Position:342.8102;Intensity:500;Charge:1),
(Position:342.9731;Intensity:60;Charge:1),
(Position:344.1592;Intensity:6000;Charge:1),
(Position:344.1975;Intensity:2500;Charge:1),
(Position:344.4863;Intensity:1000;Charge:1),
(Position:344.6136;Intensity:200;Charge:1),
(Position:345.460;Intensity:15;Charge:2),
(Position:346.6853;Intensity:1200;Charge:1),
(Position:347.6445;Intensity:2000;Charge:1),
(Position:347.7697;Intensity:500;Charge:1),
(Position:349.1573;Intensity:2500;Charge:1),
(Position:349.8841;Intensity:500;Charge:1),
(Position:350.059;Intensity:9;Charge:3),
(Position:350.128;Intensity:9;Charge:3),
(Position:350.276;Intensity:10;Charge:3),
(Position:351.4822;Intensity:250;Charge:1),
(Position:352.2268;Intensity:300;Charge:1),
(Position:352.7048;Intensity:400;Charge:1),
(Position:352.7174;Intensity:100;Charge:1),
(Position:352.7245;Intensity:60;Charge:1),
(Position:352.7389;Intensity:60;Charge:1),
(Position:352.7475;Intensity:60;Charge:1),
(Position:353.4209;Intensity:100;Charge:1),
(Position:353.7567;Intensity:200;Charge:1),
(Position:354.2095;Intensity:300;Charge:1),
(Position:354.3087;Intensity:250;Charge:1),
(Position:355.4754;Intensity:80;Charge:1),
(Position:355.5941;Intensity:400;Charge:1),
(Position:355.7894;Intensity:200;Charge:1),
(Position:355.9531;Intensity:400;Charge:1),
(Position:356.6397;Intensity:1000;Charge:1),
(Position:357.1116;Intensity:1200;Charge:1),
(Position:357.127;Intensity:800;Charge:1),
(Position:357.3016;Intensity:120;Charge:1),
(Position:357.4413;Intensity:100;Charge:1),
(Position:357.4849;Intensity:60;Charge:1),
(Position:357.4908;Intensity:60;Charge:1),
(Position:358.221;Intensity:4000;Charge:1),
(Position:358.3222;Intensity:150;Charge:1),
(Position:358.5683;Intensity:150;Charge:1),
(Position:358.5952;Intensity:120;Charge:1),
(Position:358.6342;Intensity:300;Charge:1),
(Position:358.6728;Intensity:150;Charge:1),
(Position:358.706;Intensity:10;Charge:3),
(Position:358.7127;Intensity:200;Charge:1),
(Position:358.8008;Intensity:400;Charge:1),
(Position:359.5659;Intensity:100;Charge:1),
(Position:360.197;Intensity:11;Charge:3),
(Position:360.4232;Intensity:150;Charge:1),
(Position:360.491;Intensity:11;Charge:3),
(Position:360.6483;Intensity:200;Charge:1),
(Position:360.7709;Intensity:500;Charge:1),
(Position:360.9888;Intensity:1500;Charge:1),
(Position:361.119;Intensity:250;Charge:1),
(Position:361.3098;Intensity:60;Charge:1),
(Position:361.8820;Intensity:150;Charge:1),
(Position:361.9800;Intensity:1500;Charge:1),
(Position:362.2495;Intensity:200;Charge:1),
(Position:362.3037;Intensity:150;Charge:1),
(Position:362.422;Intensity:150;Charge:1),
(Position:363.2131;Intensity:100;Charge:1),
(Position:363.2498;Intensity:1200;Charge:1),
(Position:363.3076;Intensity:60;Charge:1),
(Position:363.9335;Intensity:100;Charge:1),
(Position:364.1427;Intensity:200;Charge:1),
(Position:364.4755;Intensity:80;Charge:1),
(Position:364.8881;Intensity:1500;Charge:1),
(Position:365.0546;Intensity:250;Charge:1),
(Position:365.1319;Intensity:80;Charge:1),
(Position:365.2507;Intensity:200;Charge:1),
(Position:367.1069;Intensity:120;Charge:1),
(Position:367.1134;Intensity:150;Charge:1),
(Position:367.7358;Intensity:100;Charge:1),
(Position:367.8676;Intensity:150;Charge:1),
(Position:368.0961;Intensity:1500;Charge:1),
(Position:368.3290;Intensity:200;Charge:1),
(Position:368.4103;Intensity:120;Charge:1),
(Position:368.5156;Intensity:150;Charge:1),
(Position:368.7047;Intensity:120;Charge:1),
(Position:368.8506;Intensity:500;Charge:1),
(Position:369.0527;Intensity:120;Charge:1),
(Position:369.5059;Intensity:150;Charge:1),
(Position:369.6103;Intensity:120;Charge:1),
(Position:370.2139;Intensity:150;Charge:1),
(Position:370.5516;Intensity:80;Charge:1),
(Position:370.6620;Intensity:1200;Charge:1),
(Position:370.8096;Intensity:60;Charge:1),
(Position:370.8876;Intensity:150;Charge:1),
(Position:370.8974;Intensity:300;Charge:1),
(Position:371.0301;Intensity:600;Charge:1),
(Position:371.7499;Intensity:120;Charge:1),
(Position:372.0993;Intensity:8000;Charge:1),
(Position:372.3622;Intensity:1500;Charge:1),
(Position:372.5436;Intensity:120;Charge:1),
(Position:372.6551;Intensity:60;Charge:1),
(Position:372.8153;Intensity:60;Charge:1),
(Position:372.8679;Intensity:500;Charge:1),
(Position:373.3457;Intensity:150;Charge:1),
(Position:373.4379;Intensity:1200;Charge:1),
(Position:373.5926;Intensity:5000;Charge:1),
(Position:373.6386;Intensity:120;Charge:1),
(Position:373.8194;Intensity:6000;Charge:1),
(Position:373.9369;Intensity:100;Charge:1),
(Position:374.4426;Intensity:400;Charge:1),
(Position:374.453;Intensity:80;Charge:1),
(Position:374.6626;Intensity:6000;Charge:1),
(Position:374.6964;Intensity:1200;Charge:1),
(Position:374.9327;Intensity:3000;Charge:1),
(Position:375.0030;Intensity:80;Charge:1),
(Position:375.0551;Intensity:3000;Charge:1),
(Position:375.9300;Intensity:1500;Charge:1),
(Position:376.112;Intensity:400;Charge:1),
(Position:376.4857;Intensity:1500;Charge:1),
(Position:376.661;Intensity:400;Charge:1),
(Position:376.8261;Intensity:600;Charge:1),
(Position:377.7525;Intensity:60;Charge:1),
(Position:378.703;Intensity:250;Charge:1),
(Position:378.776;Intensity:100;Charge:1),
(Position:378.8956;Intensity:250;Charge:1),
(Position:379.1168;Intensity:250;Charge:1),
(Position:379.542;Intensity:150;Charge:1),
(Position:379.6080;Intensity:400;Charge:1),
(Position:379.8596;Intensity:120;Charge:1),
(Position:379.9590;Intensity:250;Charge:1),
(Position:380.0626;Intensity:400;Charge:1),
(Position:380.6425;Intensity:200;Charge:1),
(Position:380.7777;Intensity:80;Charge:1),
(Position:381.4046;Intensity:600;Charge:1),
(Position:381.4141;Intensity:60;Charge:1),
(Position:381.6923;Intensity:1500;Charge:1),
(Position:382.1509;Intensity:2500;Charge:1),
(Position:382.2263;Intensity:150;Charge:1),
(Position:382.5391;Intensity:80;Charge:1),
(Position:382.5529;Intensity:2500;Charge:1),
(Position:382.6966;Intensity:1500;Charge:1),
(Position:382.8909;Intensity:1200;Charge:1),
(Position:383.5310;Intensity:1000;Charge:1),
(Position:384.0346;Intensity:120;Charge:1),
(Position:384.1526;Intensity:500;Charge:1),
(Position:384.2137;Intensity:800;Charge:1),
(Position:384.4346;Intensity:120;Charge:1),
(Position:384.7891;Intensity:80;Charge:1),
(Position:385.105;Intensity:200;Charge:1),
(Position:385.1909;Intensity:120;Charge:1),
(Position:385.7466;Intensity:2500;Charge:1),
(Position:386.0306;Intensity:150;Charge:1),
(Position:386.1005;Intensity:10000;Charge:1),
(Position:386.6619;Intensity:150;Charge:1),
(Position:386.8311;Intensity:60;Charge:1),
(Position:387.3599;Intensity:250;Charge:1),
(Position:387.4859;Intensity:150;Charge:1),
(Position:387.9117;Intensity:250;Charge:1),
(Position:387.9672;Intensity:2000;Charge:1),
(Position:388.7383;Intensity:4000;Charge:1),
(Position:388.8150;Intensity:200;Charge:1),
(Position:388.9615;Intensity:300;Charge:1),
(Position:389.6760;Intensity:800;Charge:1),
(Position:390.0812;Intensity:1200;Charge:1),
(Position:390.4051;Intensity:400;Charge:1),
(Position:390.7586;Intensity:250;Charge:1),
(Position:391.7840;Intensity:80;Charge:1),
(Position:392.1368;Intensity:600;Charge:1),
(Position:392.4022;Intensity:1200;Charge:1),
(Position:392.9032;Intensity:1200;Charge:1),
(Position:393.1409;Intensity:2000;Charge:1),
(Position:394.9892;Intensity:60;Charge:1),
(Position:395.1071;Intensity:60;Charge:1),
(Position:395.2282;Intensity:50;Charge:1),
(Position:395.3720;Intensity:50;Charge:1),
(Position:395.545;Intensity:16;Charge:3),
(Position:395.7574;Intensity:60;Charge:1),
(Position:395.780;Intensity:250;Charge:1),
(Position:396.7736;Intensity:60;Charge:1),
(Position:396.984;Intensity:11;Charge:3),
(Position:397.0380;Intensity:100;Charge:1),
(Position:397.061;Intensity:9;Charge:3),
(Position:397.8866;Intensity:80;Charge:1),
(Position:398.055;Intensity:10;Charge:3),
(Position:398.2897;Intensity:40;Charge:1),
(Position:398.5083;Intensity:50;Charge:1),
(Position:399.5243;Intensity:60;Charge:1),
(Position:399.8522;Intensity:200;Charge:1),
(Position:399.9183;Intensity:40;Charge:1),
(Position:400.6373;Intensity:400;Charge:1),
(Position:401.0847;Intensity:60;Charge:1),
(Position:401.566;Intensity:80;Charge:1),
(Position:402.3004;Intensity:100;Charge:1),
(Position:403.656;Intensity:10;Charge:3),
(Position:404.1780;Intensity:50;Charge:1),
(Position:404.6956;Intensity:4000;Charge:1),
(Position:405.425;Intensity:11;Charge:3),
(Position:406.4742;Intensity:1500;Charge:1),
(Position:406.8124;Intensity:50;Charge:1),
(Position:406.9126;Intensity:50;Charge:1),
(Position:407.2887;Intensity:1200;Charge:1),
(Position:407.7780;Intensity:40;Charge:1),
(Position:408.215;Intensity:12;Charge:3),
(Position:410.1894;Intensity:40;Charge:1),
(Position:410.8648;Intensity:40;Charge:1),
(Position:411.9706;Intensity:150;Charge:1),
(Position:412.206;Intensity:10;Charge:3),
(Position:412.318;Intensity:11;Charge:3),
(Position:412.394;Intensity:11;Charge:3),
(Position:412.8772;Intensity:40;Charge:1),
(Position:413.3224;Intensity:400;Charge:1),
(Position:413.5842;Intensity:80;Charge:1),
(Position:413.8164;Intensity:40;Charge:1),
(Position:413.893;Intensity:15;Charge:3),
(Position:414.052;Intensity:13;Charge:3),
(Position:414.165;Intensity:9;Charge:3),
(Position:414.4584;Intensity:200;Charge:1),
(Position:414.5038;Intensity:800;Charge:1),
(Position:415.5069;Intensity:40;Charge:1),
(Position:415.5671;Intensity:50;Charge:1),
(Position:415.613;Intensity:9;Charge:3),
(Position:415.7971;Intensity:60;Charge:1),
(Position:416.590;Intensity:18;Charge:3),
(Position:416.609;Intensity:9;Charge:3),
(Position:416.801;Intensity:13;Charge:3),
(Position:417.3920;Intensity:50;Charge:1),
(Position:417.544;Intensity:13;Charge:3),
(Position:417.6089;Intensity:60;Charge:1),
(Position:417.6812;Intensity:50;Charge:1),
(Position:417.8770;Intensity:50;Charge:1),
(Position:418.2933;Intensity:120;Charge:1),
(Position:418.6070;Intensity:50;Charge:1),
(Position:418.8218;Intensity:120;Charge:1),
(Position:418.8975;Intensity:120;Charge:1),
(Position:419.2611;Intensity:80;Charge:1),
(Position:419.6511;Intensity:40;Charge:1),
(Position:419.9487;Intensity:150;Charge:1),
(Position:420.0278;Intensity:40;Charge:1),
(Position:420.3213;Intensity:300;Charge:1),
(Position:420.5168;Intensity:40;Charge:1),
(Position:420.7881;Intensity:80;Charge:1),
(Position:421.1529;Intensity:80;Charge:1),
(Position:421.186;Intensity:9;Charge:3),
(Position:421.7371;Intensity:400;Charge:1),
(Position:422.0548;Intensity:100;Charge:1),
(Position:422.3401;Intensity:50;Charge:1),
(Position:422.346;Intensity:11;Charge:3),
(Position:422.7146;Intensity:50;Charge:1),
(Position:422.8614;Intensity:200;Charge:1),
(Position:423.436;Intensity:11;Charge:2),
(Position:423.4794;Intensity:100;Charge:1),
(Position:423.675;Intensity:13;Charge:3),
(Position:423.7129;Intensity:250;Charge:1),
(Position:423.981;Intensity:9;Charge:3),
(Position:424.0003;Intensity:50;Charge:1),
(Position:424.494;Intensity:12;Charge:3),
(Position:424.8621;Intensity:50;Charge:1),
(Position:425.1315;Intensity:200;Charge:1),
(Position:425.1984;Intensity:300;Charge:1),
(Position:425.9514;Intensity:40;Charge:1),
(Position:426.1672;Intensity:800;Charge:1),
(Position:427.2355;Intensity:250;Charge:1),
(Position:427.2961;Intensity:1200;Charge:1),
(Position:427.460;Intensity:12;Charge:3),
(Position:428.092;Intensity:12;Charge:3),
(Position:428.3607;Intensity:1200;Charge:1),
(Position:428.737;Intensity:14;Charge:3),
(Position:429.2669;Intensity:80;Charge:1),
(Position:429.806;Intensity:16;Charge:3),
(Position:430.0443;Intensity:250;Charge:1),
(Position:430.599;Intensity:18;Charge:3),
(Position:430.9113;Intensity:1200;Charge:1),
(Position:431.157;Intensity:20;Charge:3),
(Position:431.6298;Intensity:150;Charge:1),
(Position:432.490;Intensity:9;Charge:3),
(Position:432.6977;Intensity:1500;Charge:1),
(Position:435.3958;Intensity:80;Charge:1),
(Position:437.0999;Intensity:80;Charge:1),
(Position:437.327;Intensity:9;Charge:3),
(Position:437.337;Intensity:9;Charge:3),
(Position:437.354;Intensity:11;Charge:3),
(Position:437.376;Intensity:14;Charge:3),
(Position:437.404;Intensity:18;Charge:3),
(Position:437.7159;Intensity:800;Charge:1),
(Position:438.4776;Intensity:3000;Charge:1),
(Position:439.699;Intensity:9;Charge:3),
(Position:440.5987;Intensity:1200;Charge:1),
(Position:441.6362;Intensity:300;Charge:1),
(Position:442.084;Intensity:12;Charge:3),
(Position:442.8542;Intensity:600;Charge:1),
(Position:443.226;Intensity:9;Charge:3),
(Position:446.2904;Intensity:400;Charge:1),
(Position:446.7805;Intensity:120;Charge:1),
(Position:447.7273;Intensity:80;Charge:1),
(Position:448.3427;Intensity:80;Charge:1),
(Position:448.3510;Intensity:200;Charge:1),
(Position:449.0999;Intensity:50;Charge:1),
(Position:452.9883;Intensity:50;Charge:1),
(Position:458.511;Intensity:11;Charge:2),
(Position:464.8734;Intensity:30;Charge:1),
(Position:473.8096;Intensity:30;Charge:1),
(Position:486.1099;Intensity:50;Charge:1),
(Position:487.2678;Intensity:120;Charge:1),
(Position:487.3497;Intensity:60;Charge:1),
(Position:487.9570;Intensity:30;Charge:1),
(Position:489.2120;Intensity:100;Charge:1),
(Position:489.2858;Intensity:250;Charge:1),
(Position:490.4678;Intensity:30;Charge:1),
(Position:492.0365;Intensity:150;Charge:1),
(Position:492.1876;Intensity:500;Charge:1),
(Position:492.529;Intensity:12;Charge:2),
(Position:495.8981;Intensity:1500;Charge:1),
(Position:499.189;Intensity:11;Charge:2),
(Position:500.3257;Intensity:80;Charge:1),
(Position:500.331;Intensity:18;Charge:2),
(Position:500.560;Intensity:11;Charge:2),
(Position:500.7107;Intensity:30;Charge:1),
(Position:500.7513;Intensity:100;Charge:1),
(Position:501.3465;Intensity:60;Charge:1),
(Position:501.6340;Intensity:30;Charge:1),
(Position:501.983;Intensity:12;Charge:2),
(Position:503.204;Intensity:11;Charge:2),
(Position:503.217;Intensity:25;Charge:1),
(Position:503.711;Intensity:12;Charge:2),
(Position:504.3161;Intensity:150;Charge:1),
(Position:505.1227;Intensity:30;Charge:1),
(Position:505.3043;Intensity:30;Charge:1),
(Position:507.6163;Intensity:25;Charge:1),
(Position:510.215;Intensity:18;Charge:2),
(Position:510.237;Intensity:15;Charge:2),
(Position:511.1781;Intensity:150;Charge:1),
(Position:511.249;Intensity:9;Charge:3),
(Position:512.878;Intensity:9;Charge:3),
(Position:513.512;Intensity:40;Charge:1),
(Position:514.0683;Intensity:40;Charge:1),
(Position:514.0894;Intensity:100;Charge:1),
(Position:514.579;Intensity:11;Charge:2),
(Position:515.089;Intensity:12;Charge:2),
(Position:515.3345;Intensity:25;Charge:1),
(Position:515.756;Intensity:12;Charge:3),
(Position:516.371;Intensity:30;Charge:1),
(Position:516.7720;Intensity:80;Charge:1),
(Position:516.8926;Intensity:2500;Charge:1),
(Position:517.0337;Intensity:80;Charge:1),
(Position:517.047;Intensity:12;Charge:2),
(Position:517.3035;Intensity:500;Charge:1),
(Position:519.2900;Intensity:50;Charge:1),
(Position:519.3789;Intensity:80;Charge:1),
(Position:519.6388;Intensity:200;Charge:1),
(Position:520.053;Intensity:10;Charge:3),
(Position:520.6031;Intensity:30;Charge:1),
(Position:521.6631;Intensity:25;Charge:1),
(Position:521.7726;Intensity:150;Charge:1),
(Position:521.830;Intensity:18;Charge:2),
(Position:522.8317;Intensity:60;Charge:1),
(Position:522.8605;Intensity:1000;Charge:1),
(Position:522.895;Intensity:13;Charge:2),
(Position:523.4396;Intensity:250;Charge:1),
(Position:523.712;Intensity:10;Charge:3),
(Position:524.477;Intensity:18;Charge:3),
(Position:524.941;Intensity:13;Charge:2),
(Position:525.269;Intensity:13;Charge:2),
(Position:526.172;Intensity:18;Charge:2),
(Position:526.180;Intensity:13;Charge:3),
(Position:526.565;Intensity:11;Charge:2),
(Position:526.8021;Intensity:100;Charge:1),
(Position:527.1004;Intensity:1200;Charge:1),
(Position:527.1824;Intensity:800;Charge:1),
(Position:527.384;Intensity:9;Charge:3),
(Position:527.445;Intensity:14;Charge:3),
(Position:527.795;Intensity:15;Charge:3),
(Position:528.3259;Intensity:30;Charge:1),
(Position:528.377;Intensity:16;Charge:3),
(Position:528.5091;Intensity:60;Charge:1),
(Position:528.630;Intensity:12;Charge:3),
(Position:529.959;Intensity:11;Charge:3),
(Position:530.140;Intensity:12;Charge:3),
(Position:530.3774;Intensity:25;Charge:1),
(Position:530.408;Intensity:14;Charge:3),
(Position:530.766;Intensity:11;Charge:2),
(Position:530.824;Intensity:10;Charge:3),
(Position:531.236;Intensity:9;Charge:3),
(Position:531.771;Intensity:13;Charge:2),
(Position:532.422;Intensity:10;Charge:3),
(Position:532.5659;Intensity:150;Charge:1),
(Position:532.9520;Intensity:800;Charge:1),
(Position:533.0013;Intensity:300;Charge:1),
(Position:533.4382;Intensity:100;Charge:1),
(Position:534.108;Intensity:14;Charge:2),
(Position:534.1413;Intensity:80;Charge:1),
(Position:534.2509;Intensity:500;Charge:1),
(Position:534.837;Intensity:11;Charge:3),
(Position:535.526;Intensity:12;Charge:3),
(Position:536.525;Intensity:12;Charge:3),
(Position:536.636;Intensity:25;Charge:1),
(Position:536.896;Intensity:40;Charge:1),
(Position:536.955;Intensity:10;Charge:3),
(Position:537.145;Intensity:50;Charge:1),
(Position:537.2983;Intensity:400;Charge:1),
(Position:537.696;Intensity:11;Charge:3),
(Position:538.487;Intensity:60;Charge:1),
(Position:538.856;Intensity:14;Charge:2),
(Position:539.4667;Intensity:40;Charge:1),
(Position:539.736;Intensity:12;Charge:2),
(Position:539.8628;Intensity:300;Charge:1),
(Position:540.356;Intensity:15;Charge:2),
(Position:540.562;Intensity:60;Charge:1),
(Position:540.7277;Intensity:250;Charge:1),
(Position:541.241;Intensity:30;Charge:1),
(Position:541.671;Intensity:60;Charge:1),
(Position:542.558;Intensity:60;Charge:1),
(Position:542.934;Intensity:30;Charge:2),
(Position:543.1204;Intensity:250;Charge:1),
(Position:543.150;Intensity:13;Charge:2),
(Position:543.6034;Intensity:100;Charge:1),
(Position:544.8385;Intensity:200;Charge:1),
(Position:545.697;Intensity:25;Charge:1),
(Position:545.7125;Intensity:120;Charge:1),
(Position:546.745;Intensity:16;Charge:2),
(Position:546.846;Intensity:20;Charge:2),
(Position:548.383;Intensity:16;Charge:2),
(Position:549.536;Intensity:14;Charge:2),
(Position:549.9043;Intensity:25;Charge:1),
(Position:550.2992;Intensity:20;Charge:1),
(Position:550.773;Intensity:18;Charge:2),
(Position:550.8308;Intensity:30;Charge:1),
(Position:551.231;Intensity:12;Charge:2),
(Position:553.060;Intensity:12;Charge:2),
(Position:554.630;Intensity:13;Charge:2),
(Position:557.1165;Intensity:30;Charge:1),
(Position:557.4388;Intensity:60;Charge:1),
(Position:558.8306;Intensity:120;Charge:1),
(Position:561.7203;Intensity:200;Charge:1),
(Position:562.6102;Intensity:20;Charge:1),
(Position:564.697;Intensity:12;Charge:2),
(Position:566.4086;Intensity:50;Charge:1),
(Position:572.147;Intensity:11;Charge:3),
(Position:574.578;Intensity:9;Charge:3),
(Position:575.798;Intensity:10;Charge:3),
(Position:576.4588;Intensity:20;Charge:1),
(Position:578.523;Intensity:11;Charge:2),
(Position:583.555;Intensity:18;Charge:3),
(Position:585.038;Intensity:9;Charge:3),
(Position:585.624;Intensity:10;Charge:3),
(Position:586.3978;Intensity:30;Charge:1),
(Position:587.789;Intensity:9;Charge:3),
(Position:588.665;Intensity:13;Charge:2),
(Position:589.354;Intensity:15;Charge:3),
(Position:590.031;Intensity:9;Charge:3),
(Position:590.446;Intensity:16;Charge:2),
(Position:591.5753;Intensity:30;Charge:1),
(Position:592.060;Intensity:9;Charge:3),
(Position:592.177;Intensity:10;Charge:3),
(Position:593.133;Intensity:18;Charge:3),
(Position:595.396;Intensity:10;Charge:3),
(Position:595.527;Intensity:14;Charge:3),
(Position:595.735;Intensity:14;Charge:2),
(Position:596.336;Intensity:18;Charge:2),
(Position:596.41;Intensity:30;Charge:2),
(Position:596.728;Intensity:13;Charge:2),
(Position:597.013;Intensity:9;Charge:3),
(Position:598.098;Intensity:12;Charge:3),
(Position:598.267;Intensity:9;Charge:3),
(Position:598.8614;Intensity:30;Charge:1),
(Position:599.074;Intensity:12;Charge:3),
(Position:600.120;Intensity:18;Charge:3),
(Position:603.269;Intensity:9;Charge:3),
(Position:603.426;Intensity:16;Charge:3),
(Position:603.823;Intensity:13;Charge:3),
(Position:605.039;Intensity:11;Charge:3),
(Position:605.586;Intensity:11;Charge:3),
(Position:605.804;Intensity:9;Charge:3),
(Position:606.7161;Intensity:40;Charge:1),
(Position:610.3848;Intensity:30;Charge:1),
(Position:613.8312;Intensity:40;Charge:1),
(Position:613.9393;Intensity:40;Charge:1),
(Position:614.943;Intensity:30;Charge:2),
(Position:615.094;Intensity:20;Charge:2),
(Position:615.169;Intensity:9;Charge:3),
(Position:617.145;Intensity:9;Charge:3),
(Position:617.687;Intensity:15;Charge:2),
(Position:618.697;Intensity:9;Charge:3),
(Position:618.827;Intensity:7;Charge:3),
(Position:619.3271;Intensity:40;Charge:1),
(Position:619.650;Intensity:7;Charge:3),
(Position:619.714;Intensity:6;Charge:3),
(Position:620.309;Intensity:6;Charge:3),
(Position:620.476;Intensity:5;Charge:3),
(Position:621.5148;Intensity:30;Charge:1),
(Position:622.1000;Intensity:30;Charge:1),
(Position:623.2450;Intensity:40;Charge:1),
(Position:624.010;Intensity:20;Charge:2),
(Position:624.8045;Intensity:20;Charge:1),
(Position:624.929;Intensity:80;Charge:2),
(Position:625.4284;Intensity:30;Charge:1),
(Position:626.154;Intensity:5;Charge:3),
(Position:629.624;Intensity:6;Charge:3),
(Position:630.706;Intensity:15;Charge:2),
(Position:633.372;Intensity:12;Charge:2),
(Position:635.957;Intensity:5;Charge:3),
(Position:638.551;Intensity:15;Charge:2),
(Position:639.5369;Intensity:20;Charge:1),
(Position:640.1768;Intensity:30;Charge:1),
(Position:641.3419;Intensity:20;Charge:1),
(Position:641.867;Intensity:20;Charge:2),
(Position:642.3124;Intensity:20;Charge:1),
(Position:643.2621;Intensity:30;Charge:1),
(Position:644.821;Intensity:20;Charge:2),
(Position:645.816;Intensity:200;Charge:2),
(Position:649.6776;Intensity:60;Charge:1),
(Position:651.785;Intensity:20;Charge:2),
(Position:654.8047;Intensity:20;Charge:1),
(Position:659.4734;Intensity:20;Charge:1),
(Position:667.9833;Intensity:40;Charge:1),
(Position:685.707;Intensity:15;Charge:1),
(Position:694.713;Intensity:15;Charge:1),
(Position:706.939;Intensity:20;Charge:2),
(Position:713.291;Intensity:15;Charge:1),
(Position:716.6418;Intensity:25;Charge:1),
(Position:718.9294;Intensity:80;Charge:1),
(Position:720.9367;Intensity:30;Charge:1),
(Position:722.650;Intensity:12;Charge:2),
(Position:730.998;Intensity:50;Charge:2),
(Position:731.965;Intensity:5;Charge:3),
(Position:732.216;Intensity:6;Charge:3),
(Position:732.272;Intensity:40;Charge:2),
(Position:737.849;Intensity:20;Charge:2),
(Position:744.7797;Intensity:30;Charge:1),
(Position:746.444;Intensity:20;Charge:2),
(Position:749.7123;Intensity:40;Charge:1),
(Position:751.3113;Intensity:60;Charge:1),
(Position:758.813;Intensity:15;Charge:1),
(Position:771.383;Intensity:15;Charge:2),
(Position:778.273;Intensity:30;Charge:1),
(Position:783.437;Intensity:40;Charge:1),
(Position:792.335;Intensity:5;Charge:3),
(Position:793.9314;Intensity:80;Charge:1),
(Position:794.8170;Intensity:60;Charge:1),
(Position:800.1139;Intensity:80;Charge:1),
(Position:804.8260;Intensity:60;Charge:1),
(Position:808.7399;Intensity:50;Charge:1),
(Position:822.267;Intensity:150;Charge:1),
(Position:823.314;Intensity:5;Charge:3),
(Position:823.405;Intensity:5;Charge:3),
(Position:823.771;Intensity:9;Charge:3),
(Position:823.901;Intensity:8;Charge:3),
(Position:824.124;Intensity:6;Charge:3),
(Position:832.9342;Intensity:120;Charge:1),
(Position:833.4198;Intensity:20;Charge:1),
(Position:839.0075;Intensity:120;Charge:1),
(Position:847.0731;Intensity:30;Charge:1),
(Position:851.6408;Intensity:15;Charge:1),
(Position:856.584;Intensity:5;Charge:3),
(Position:866.4277;Intensity:60;Charge:1),
(Position:869.1007;Intensity:150;Charge:1),
(Position:879.579;Intensity:12;Charge:1),
(Position:882.665;Intensity:12;Charge:1),
(Position:886.939;Intensity:20;Charge:1),
(Position:900.203;Intensity:15;Charge:1),
(Position:1021.912;Intensity:15;Charge:1),
(Position:1047.252;Intensity:13;Charge:1),
(Position:1112.285;Intensity:21;Charge:1),
(Position:1137.719;Intensity:14;Charge:1),
(Position:1142.545;Intensity:52;Charge:1),
(Position:1144.225;Intensity:87;Charge:1),
(Position:1159.676;Intensity:91;Charge:1),
(Position:1161.075;Intensity:255;Charge:1),
(Position:1164.145;Intensity:160;Charge:1),
(Position:1169.318;Intensity:230;Charge:1),
(Position:1178.649;Intensity:160;Charge:1),
(Position:1188.609;Intensity:580;Charge:1),
(Position:1188.733;Intensity:225;Charge:1),
(Position:1197.633;Intensity:1030;Charge:1),
(Position:1264.217;Intensity:15;Charge:1),
(Position:1288.328;Intensity:14;Charge:1),
(Position:1356.875;Intensity:17;Charge:1),
(Position:1424.014;Intensity:30;Charge:1),
(Position:1428.901;Intensity:24;Charge:1),
(Position:1429.629;Intensity:14;Charge:1),
(Position:1431.260;Intensity:16;Charge:1),
(Position:1440.450;Intensity:96;Charge:1),
(Position:1444.623;Intensity:20;Charge:1),
(Position:1451.620;Intensity:72;Charge:1),
(Position:1455.904;Intensity:50;Charge:1),
(Position:1456.993;Intensity:14;Charge:1),
(Position:1483.048;Intensity:40;Charge:1),
(Position:1505.588;Intensity:37;Charge:1),
(Position:1521.171;Intensity:28;Charge:1),
(Position:1529.876;Intensity:94;Charge:1),
(Position:1533.959;Intensity:16;Charge:1),
(Position:1562.594;Intensity:30;Charge:1),
(Position:1563.624;Intensity:25;Charge:1),
(Position:1572.789;Intensity:14;Charge:1),
(Position:1577.373;Intensity:41;Charge:1),
(Position:1581.745;Intensity:28;Charge:1),
(Position:1644.931;Intensity:13;Charge:1),
(Position:1649.119;Intensity:20;Charge:1),
(Position:1886.180;Intensity:105;Charge:1),
(Position:1899.219;Intensity:47;Charge:1),
(Position:1911.890;Intensity:25;Charge:1),
(Position:1979.728;Intensity:22;Charge:1),
(Position:2238.693;Intensity:14;Charge:1),
(Position:2262.602;Intensity:21;Charge:1),
(Position:2622.919;Intensity:38;Charge:1),
(Position:2666.649;Intensity:17;Charge:1)
);

{ The lines below are measured lines from:
  An atlas of optogalvanic transitions in Neon.
  RAL report, RAL-91-069 (1991)
  S.H. Ashworth and J.M. Brown.
 with the lines from the NIST database (above) removed, and a few lines
 measured at Bristol added.
 The positions are in air nm; intensities are all arbitrary.
}

const
NeOptogalvanic: array [0..206] of TAtomicLine = (
(Position:  272.4011;Intensity:1;Charge:0),
(Position:  272.4388;Intensity:1;Charge:0),
(Position:  272.5329;Intensity:1;Charge:0),
(Position:  273.6725;Intensity:1;Charge:0),
(Position:  274.4945;Intensity:1;Charge:0),
(Position:  274.8936;Intensity:1;Charge:0),
(Position:  274.9878;Intensity:1;Charge:0),
(Position:  275.6178;Intensity:1;Charge:0),
(Position:  275.8475;Intensity:1;Charge:0),
(Position:  276.175;Intensity:1;Charge:0),
(Position:  276.2528;Intensity:1;Charge:0),
(Position:  276.588;Intensity:1;Charge:0),
(Position:  277.036;Intensity:1;Charge:0),
(Position:  277.0849;Intensity:1;Charge:0),
(Position:  277.1402;Intensity:1;Charge:0),
(Position:  277.2574;Intensity:1;Charge:0),
(Position:  277.9501;Intensity:1;Charge:0),
(Position:  278.1177;Intensity:1;Charge:0),
(Position:  278.149;Intensity:1;Charge:0),
(Position:  278.2165;Intensity:1;Charge:0),
(Position:  279.2907;Intensity:1;Charge:0),
(Position:  284.3765;Intensity:1;Charge:0),
(Position:  285.054;Intensity:1;Charge:0),
(Position:  288.0049;Intensity:1;Charge:0),
(Position:  288.2189;Intensity:1;Charge:0),
(Position:  294.9167;Intensity:1;Charge:0),
(Position:  295.1228;Intensity:1;Charge:0),
(Position:  298.3002;Intensity:1;Charge:0),
(Position:  302.0992;Intensity:1;Charge:0),
(Position:  302.6186;Intensity:1;Charge:0),
(Position:  312.6396;Intensity:1;Charge:0),
(Position:  343.4344;Intensity:1;Charge:0),
(Position:  343.7168;Intensity:1;Charge:0),
(Position:  344.7339;Intensity:1;Charge:0),
(Position:  400.0291;Intensity:1;Charge:0),
(Position:  400.0516;Intensity:1;Charge:0),
(Position:  401.9601;Intensity:1;Charge:0),
(Position:  403.6937;Intensity:1;Charge:0),
(Position:  404.1108;Intensity:1;Charge:0),
(Position:  404.5055;Intensity:1;Charge:0),
(Position:  416.9734;Intensity:1;Charge:0),
(Position:  419.3008;Intensity:1;Charge:0),
(Position:  419.5774;Intensity:1;Charge:0),
(Position:  420.6838;Intensity:1;Charge:0),
(Position:  421.6257;Intensity:1;Charge:0),
(Position:  422.0493;Intensity:1;Charge:0),
(Position:  422.1988;Intensity:1;Charge:0),
(Position:  423.7612;Intensity:1;Charge:0),
(Position:  425.7028;Intensity:1;Charge:0),
(Position:  426.5887;Intensity:1;Charge:0),
(Position:  426.8915;Intensity:1;Charge:0),
(Position:  428.2954;Intensity:1;Charge:0),
(Position:  428.8059;Intensity:1;Charge:0),
(Position:  429.04;Intensity:1;Charge:0),
(Position:  429.2564;Intensity:1;Charge:0),
(Position:  429.3165;Intensity:1;Charge:0),
(Position:  429.6818;Intensity:1;Charge:0),
(Position:  430.411;Intensity:1;Charge:0),
(Position:  430.471;Intensity:1;Charge:0),
(Position:  430.9109;Intensity:1;Charge:0),
(Position:  430.9684;Intensity:1;Charge:0),
(Position:  431.3283;Intensity:1;Charge:0),
(Position:  431.7057;Intensity:1;Charge:0),
(Position:  432.0706;Intensity:1;Charge:0),
(Position:  432.1331;Intensity:1;Charge:0),
(Position:  432.2605;Intensity:1;Charge:0),
(Position:  432.6604;Intensity:1;Charge:0),
(Position:  432.6854;Intensity:1;Charge:0),
(Position:  433.0228;Intensity:1;Charge:0),
(Position:  433.7302;Intensity:1;Charge:0),
(Position:  434.0976;Intensity:1;Charge:0),
(Position:  434.142;Intensity:1;Charge:0),
(Position:  434.28;Intensity:1;Charge:0),
(Position:  434.5225;Intensity:1;Charge:0),
(Position:  434.8374;Intensity:1;Charge:0),
(Position:  435.0623;Intensity:1;Charge:0),
(Position:  435.0998;Intensity:1;Charge:0),
(Position:  435.3647;Intensity:1;Charge:0),
(Position:  435.7697;Intensity:1;Charge:0),
(Position:  435.9271;Intensity:1;Charge:0),
(Position:  435.9546;Intensity:1;Charge:0),
(Position:  436.4245;Intensity:1;Charge:0),
(Position:  436.7044;Intensity:1;Charge:0),
(Position:  436.8794;Intensity:1;Charge:0),
(Position:  437.1443;Intensity:1;Charge:0),
(Position:  437.5267;Intensity:1;Charge:0),
(Position:  437.5717;Intensity:1;Charge:0),
(Position:  438.1541;Intensity:1;Charge:0),
(Position:  438.2265;Intensity:1;Charge:0),
(Position:  438.2815;Intensity:1;Charge:0),
(Position:  438.459;Intensity:1;Charge:0),
(Position:  439.0363;Intensity:1;Charge:0),
(Position:  439.0888;Intensity:1;Charge:0),
(Position:  440.1635;Intensity:1;Charge:0),
(Position:  440.201;Intensity:1;Charge:0),
(Position:  440.3285;Intensity:1;Charge:0),
(Position:  440.346;Intensity:1;Charge:0),
(Position:  440.3685;Intensity:1;Charge:0),
(Position:  440.7059;Intensity:1;Charge:0),
(Position:  441.254;Intensity:1;Charge:0),
(Position:  441.6082;Intensity:1;Charge:0),
(Position:  441.6357;Intensity:1;Charge:0),
(Position:  441.6982;Intensity:1;Charge:0),
(Position:  442.9229;Intensity:1;Charge:0),
(Position:  444.13;Intensity:1;Charge:0),
(Position:  444.23;Intensity:1;Charge:0),
(Position:  444.2525;Intensity:1;Charge:0),
(Position:  444.285;Intensity:1;Charge:0),
(Position:  444.5549;Intensity:1;Charge:0),
(Position:  444.6149;Intensity:1;Charge:0),
(Position:  448.0573;Intensity:1;Charge:0),
(Position:  451.0814;Intensity:1;Charge:0),                       
(Position:  452.1608;Intensity:1;Charge:0),
(Position:  453.2194;Intensity:1;Charge:0),
(Position:  453.287;Intensity:1;Charge:0),
(Position:  454.6709;Intensity:1;Charge:0),
(Position:  454.8014;Intensity:1;Charge:0), // 4 Mar 08 Was 454.801 
(Position:  454.8452;Intensity:1;Charge:0), // 4 Mar 08 Was 454.8461
(Position:  455.0031;Intensity:1;Charge:0), // 4 Mar 08 Was 455.0037
(Position:  455.0496;Intensity:1;Charge:0), // 4 Mar 08 Was 455.0488
(Position:  455.1013;Intensity:1;Charge:0),
(Position:  455.7883;Intensity:1;Charge:0), // 4 Mar 08 Was 455.7845
(Position:  455.8704;Intensity:1;Charge:0), // 4 Mar 08 Was 455.8646
(Position:  456.3264;Intensity:1;Charge:0), // 4 Mar 08 Was 456.3276
(Position:  456.4663;Intensity:1;Charge:0), // 4 Mar 08 Was 456.4652
(Position:  457.5988;Intensity:1;Charge:0), // Added 4 Mar 08
(Position:  458.645;Intensity:1;Charge:0),
(Position:  458.8102;Intensity:1;Charge:0),
(Position:  458.8627;Intensity:1;Charge:0),
(Position:  458.8802;Intensity:1;Charge:0),
(Position:  460.2491;Intensity:1;Charge:0),
(Position:  460.5495;Intensity:1;Charge:0),
(Position:  460.6371;Intensity:1;Charge:0),
(Position:  462.3349;Intensity:1;Charge:0),
(Position:  462.475;Intensity:1;Charge:0),        
(Position:  464.2877;Intensity:1;Charge:0), // Added 4 Mar 08
(Position:  464.4149;Intensity:1;Charge:0), // Added 4 Mar 08
(Position:  466.2121;Intensity:1;Charge:0), // 4 Mar 08 Was 466.2158
(Position:  466.4035;Intensity:1;Charge:0), // 4 Mar 08 Was 466.4058
(Position:  466.5434;Intensity:1;Charge:0),
(Position:  467.6911;Intensity:1;Charge:0),
(Position:  467.7258;Intensity:1;Charge:0), // Added 4 Mar 08
(Position:  467.7336;Intensity:1;Charge:0),
(Position:  467.8785;Intensity:1;Charge:0), // 4 Mar 08 Was 467.8762
(Position:  468.2391;Intensity:1;Charge:0), // 4 Mar 08 Was 468.2387
(Position:  468.3465;Intensity:1;Charge:0), // Added 4 Mar 08
(Position:  468.3987;Intensity:1;Charge:0), // 4 Mar 08 Was 468.4013
(Position:  469.3665;Intensity:1;Charge:0),
(Position:  471.1044;Intensity:1;Charge:0),
(Position:  472.5847;Intensity:1;Charge:0),
(Position:  474.9852;Intensity:1;Charge:0),
(Position:  475.1028;Intensity:1;Charge:0),
(Position:  476.6223;Intensity:1;Charge:0),
(Position:  477.4869;Intensity:1;Charge:0),
(Position:  478.3389;Intensity:1;Charge:0),
(Position:  479.9431;Intensity:1;Charge:0),
(Position:  480.0705;Intensity:1;Charge:0),
(Position:  480.5403;Intensity:1;Charge:0),
(Position:  484.2043;Intensity:1;Charge:0),
(Position:  484.3966;Intensity:1;Charge:0),
(Position:  484.539;Intensity:1;Charge:0),
(Position:  484.5464;Intensity:1;Charge:0),
(Position:  484.5565;Intensity:1;Charge:0),
(Position:  486.0749;Intensity:1;Charge:0),
(Position:  490.5405;Intensity:1;Charge:0),
(Position:  492.7533;Intensity:1;Charge:0),
(Position:  495.9434;Intensity:1;Charge:0),
(Position:  499.4041;Intensity:1;Charge:0),
(Position:  502.7199;Intensity:1;Charge:0),
(Position:  505.0599;Intensity:1;Charge:0),
(Position:  510.0579;Intensity:1;Charge:0),
(Position:  510.1879;Intensity:1;Charge:0),
(Position:  510.3478;Intensity:1;Charge:0),
(Position:  512.3821;Intensity:1;Charge:0),
(Position:  512.9319;Intensity:1;Charge:0),
(Position:  514.5939;Intensity:1;Charge:0),
(Position:  516.6307;Intensity:1;Charge:0),
(Position:  516.8906;Intensity:1;Charge:0),
(Position:  519.6497;Intensity:1;Charge:0),
(Position:  519.8497;Intensity:1;Charge:0),
(Position:  523.1361;Intensity:1;Charge:0),
(Position:  534.0637;Intensity:1;Charge:0),
(Position:  534.5739;Intensity:1;Charge:0),
(Position:  539.8458;Intensity:1;Charge:0),
(Position:  539.9483;Intensity:1;Charge:0),
(Position:  556.418;Intensity:1;Charge:0),
(Position:  556.433;Intensity:1;Charge:0),
(Position:  568.5179;Intensity:1;Charge:0),
(Position:  568.7529;Intensity:1;Charge:0),
(Position:  569.3105;Intensity:1;Charge:0),
(Position:  570.108;Intensity:1;Charge:0),
(Position:  570.9305;Intensity:1;Charge:0),
(Position:  576.0098;Intensity:1;Charge:0),
(Position:  576.2651;Intensity:1;Charge:0),
(Position:  576.3777;Intensity:1;Charge:0),
(Position:  588.9235;Intensity:1;Charge:0),
(Position:  589.3509;Intensity:1;Charge:0),
(Position:  596.6004;Intensity:1;Charge:0),
(Position:  602.6025;Intensity:1;Charge:0),
(Position:  605.2372;Intensity:1;Charge:0),
(Position:  605.6866;Intensity:1;Charge:0),
(Position:  609.2072;Intensity:1;Charge:0),
(Position:  611.4901;Intensity:1;Charge:0),
(Position:  613.451;Intensity:1;Charge:0),
(Position:  615.409;Intensity:1;Charge:0),
(Position:  618.6032;Intensity:1;Charge:0),
(Position:  630.6915;Intensity:1;Charge:0)
);

procedure MakeLineList(const AName: string; const Lines: array of TAtomicLine; Units: TResonanceUnits);
var
  i: Integer;
begin
  SetLength(AtomicLineLists, Length(AtomicLineLists)+1);
  with AtomicLineLists[High(AtomicLineLists)] do begin
    IgnoreIntensities := False;
    LineList := TLineList.Create(nmVac,False);
    Name := AName;
    for i := 0 to High(Lines) do
      with Lines[i] do
        LineList.AddLine(Position, Intensity, 0.0, clNone, Units);
  end;
end;

var
  i: Integer;
initialization
  MakeLineList('Ne', Neon, nmVac);
  MakeLineList('Optogalvanic Ne', NeOptogalvanic, nmAir);
  AtomicLineLists[1].LineList.AddIn(AtomicLineLists[0].LineList);
  AtomicLineLists[1].IgnoreIntensities := True;
  MakeLineList('Fe', Iron, nmVac);
finalization
  for i := 0 to High(AtomicLineLists) do
    AtomicLineLists[i].LineList.Free;
end.
